#ifndef CLUSTER_NEIGHBORHOOD_CPP
#define CLUSTER_NEIGHBORHOOD_CPP

#include <algorithm>

#include "ClusterNeighborhood.hh"

namespace Cluster { namespace Neighborhood {

using namespace std;

/**
 ** Helper class to sort a list of locations according to their distance to
 ** a given reference. Clusters equidistant to the center are sorted by
 ** their cluster id.
 **/
template <class CoordinateT>
class DistanceSorter
{
public:
    typedef ClusterLocation<CoordinateT> Location;
    typedef CoordinateT Distance;
    DistanceSorter(const Location &reference) : reference_(reference) {}
    inline bool operator()(const Location &lhs, const Location &rhs) const 
    {
        const Distance dl = distanceSquared(lhs, reference_);
        const Distance dr = distanceSquared(rhs, reference_);
        if (dl == dr)
            return lhs.getId() < rhs.getId();
        else
            return dl < dr;
    }
private:
    inline Distance distanceSquared(const Location &l1, const Location &l2) const
    {
        const CoordinateT dx = l1.getX() - l2.getX();
        const CoordinateT dy = l1.getY() - l2.getY();
        return dx*dx + dy*dy;
    }
    const Location &reference_;
};

/**
 ** Finder constructor
 **/
template<class CoordinateT>
Finder<CoordinateT>
    ::Finder(LocationList &tile, const pair<CoordinateT, CoordinateT> &binSize)
        : filter_(tile, binSize), tile_(tile)
{
}

/**
 ** Finder getNeighborhood
 **/
template<class CoordinateT>
void 
Finder<CoordinateT>
    ::getNeighborhood(const LocationIterator center, const CoordinateT radius, LocationList &neighborhood) const
{
    neighborhood.resize(0);
    typename Filter<CoordinateT>::iterator rawNeighborhood = filter_.getNeighborhood(center, radius);
    while (tile_.end() != rawNeighborhood)
    {
        neighborhood.push_back(*rawNeighborhood);
        ++rawNeighborhood;
    }
    sort(neighborhood.begin(), neighborhood.end(), DistanceSorter<CoordinateT>(*center));
}

/**
 ** Translate a coordinate into a positive integer indice to be used to access
 ** the elements of a vector.
 **
 ** Note: ALL negative coordinates are translated to 0.
 **/
template <class CoordinateT>
class HashFunction
{
public:
    HashFunction(const CoordinateT &binSize) : binSize_(binSize) {}
    unsigned int operator()(const CoordinateT &coordinate) const
    {
        return coordinate > 0 ? coordinate/binSize_ : 0;
    }
private:
    const CoordinateT binSize_;
};

/**
 ** Class HashSorter sorts clusters according to their hashed coordinates.
 ** The orser is first on the hashed value of X, then on the hash value of Y.
 **/
template<class CoordinateT>
class HashSorter
{
public:
    typedef ClusterLocation<CoordinateT> Location;
    HashSorter(const pair<HashFunction<CoordinateT>, HashFunction<CoordinateT> > &hash) : hash_(hash) {}
    inline bool operator()(const Location &lhsCluster, const Location &rhsCluster) const
    {
        const unsigned int lhsBinX = hash_.first(lhsCluster.getX());
        const unsigned int rhsBinX = hash_.first(rhsCluster.getX());
        if (lhsBinX == rhsBinX)
        {
            return hash_.second(lhsCluster.getY()) < hash_.second(rhsCluster.getY());
        }
        else
        {
            return lhsBinX < rhsBinX;
        }
    }
private:
    const pair<HashFunction<CoordinateT>, HashFunction<CoordinateT> > &hash_;
};

/**
 ** Bin the coordinates into the cells of a grid.
 **
 ** Note: modifies (sort) the location list.
 **
 ** Each bin is just an iterator on the first location belonging to the bin. The
 ** end of the bin is the beginning of the next bin.
 ** A location belongs to a cell if its hashed coordinates are equal to the (x,y)
 ** position of the grid.
 **/
template <class CoordinateT>
class Grid : public vector<typename Finder<CoordinateT>::LocationIterator>
{
public:
    typedef ClusterLocation<CoordinateT> Location;
    typedef vector<Location> LocationList;
    typedef typename LocationList::const_iterator LocationIterator;
    typedef vector<typename Finder<CoordinateT>::LocationIterator> GridType;
    Grid(LocationList &tile, const pair<CoordinateT, CoordinateT> binSize);
    inline const LocationList &getTile() const {return tile_;}
    // return the grid position of the cell containing location (x, y)
    inline pair<unsigned int, unsigned int> getGridPosition(const CoordinateT &x, const CoordinateT &y) const
    {
        return pair<unsigned int, unsigned int>(min(hash_.first (x), dimensions_.first -1),
                                                min(hash_.second(y), dimensions_.second-1));
    }
    inline pair<unsigned int, unsigned int> getGridPosition(const Location &location) const 
    {
        return getGridPosition(location.getX(), location.getY());
    }
    // return the index of the cell at the given grid position
    inline unsigned int getCellIndex(const pair<unsigned int, unsigned int> &gridPosition) const
    {
        return gridPosition.first * dimensions_.second + gridPosition.second;
    }
    inline unsigned int getCellIndex(const Location &location) const
    {
        return getCellIndex(getGridPosition(location));
    }
private:
    const pair<HashFunction<CoordinateT>, HashFunction<CoordinateT> > hash_;
    const LocationList &tile_;
    const pair<unsigned int, unsigned int> dimensions_;
    LocationList &sortTile(LocationList &tile)
    {
        sort(tile.begin(), tile.end(), HashSorter<CoordinateT>(hash_));
        return tile;
    }
    pair<unsigned int, unsigned int> computeDimensions() const;
};

/**
 ** Grid constructor.
 **/
template <class CoordinateT>
Grid<CoordinateT>::Grid(LocationList &tile, const pair<CoordinateT, CoordinateT> binSize)
    : hash_(HashFunction<CoordinateT>(binSize.first), HashFunction<CoordinateT>(binSize.second))
    , tile_(sortTile(tile))
    , dimensions_(computeDimensions())
{

    // oversize for a final empty bin -- coding convenience
    this->resize(dimensions_.first * dimensions_.second + 1);
    // initialize each bin to the first location belonging to that bin
    // empty bins are those equal to the next bin
    LocationIterator location = tile_.begin();
    for (typename Grid<CoordinateT>::iterator bin = this->begin(); this->end() != bin; ++bin)
    {
        while(tile_.end() != location && getCellIndex(*location) < (unsigned int)(bin - this->begin()))
        {
            ++location;
        }
        (*bin) = location;
    }
}

/**
 ** Grid::computeDimensions
 **/
template <class CoordinateT>
pair<unsigned int, unsigned int>
Grid<CoordinateT>::computeDimensions() const
{
    pair<CoordinateT, CoordinateT> boundingBox(0,0);
    for (LocationIterator location = tile_.begin(); location != tile_.end(); ++location)
    {
	if (location->getX() > boundingBox.first)
        {
            boundingBox.first = location->getX();
        }
	if (location->getY() > boundingBox.second)
        {
            boundingBox.second = location->getY();
        }
    }
    return pair<unsigned int, unsigned int>(1+hash_.first(boundingBox.first), 1+hash_.second(boundingBox.second));
}

/**
 ** Iterator class for the Prefilter. This is not a nested class to allow
 ** implicit instanciation of function tempates on this iterator (nested
 ** classes are nondeducible contexts [14.8.2.4/4]).
 **/
template<class CoordinateT>
class PreFilterIterator;

/**
 ** Provides an initial coarse pre-filtering of the clusters to speed-up the 
 ** identification of the neighborhoods.
 **
 ** Implementation:
 ** uses a grid to bin the locations
 **/
template<class CoordinateT>
class PreFilter
{
public:
    typedef ClusterLocation<CoordinateT> Location;
    typedef vector<Location> LocationList;
    typedef typename LocationList::const_iterator LocationIterator;
    typedef PreFilterIterator<CoordinateT> iterator;
    // Note: modifies the cluster location list
    PreFilter(LocationList &tile, const pair<CoordinateT, CoordinateT> binSize)
        : grid_(tile, binSize) {};
    iterator getNeighborhood(const LocationIterator &center, const CoordinateT &radius) const
    {
        return iterator(grid_, center, radius);
    };
private:
    const Grid<CoordinateT> grid_;
};

/**
 ** Class PreFilterIterator
 ** iterates ober the list of cluster locations (ie the tile), skipping most 
 ** of the cluster locations outside the circle (center, radius).
 **/
template<class CoordinateT>
class PreFilterIterator : protected vector<ClusterLocation<CoordinateT> >::const_iterator
{   
public:
    typedef ClusterLocation<CoordinateT> Location;
    typedef vector<Location> LocationList;
    typedef typename LocationList::const_iterator LocationIterator;
    typedef Grid<CoordinateT> GridType;
    typedef PreFilterIterator<CoordinateT> iterator;
    inline bool operator==(const LocationIterator &rhs) const {return *static_cast<const LocationIterator *>(this) == rhs;}
    inline bool operator!=(const LocationIterator &rhs) const {return !(*this==rhs);}
    inline iterator &operator++(); // prefix ++ only
    inline const Location &operator*() const { return **(static_cast<const LocationIterator *>(this)); }
    //inline operator LocationIterator() const {return location_;}
    PreFilterIterator(const GridType &grid, const LocationIterator &center, const CoordinateT &radius);
private:
    const GridType &grid_;
    const LocationIterator &center_;
    const CoordinateT &radius_;
protected:
    inline const LocationList &getTile() const {return grid_.getTile();};
    inline bool lastLocation() const {return getTile().end() == *static_cast<const LocationIterator *>(this);}
    inline const LocationIterator &getCenter() const {return center_;}
    inline const CoordinateT &getRadius() const {return radius_;}
    void nextValidPosition();
private:
    const pair<unsigned int, unsigned int> firstPosition_;
    const pair<unsigned int, unsigned int> lastPosition_;
    pair<unsigned int, unsigned int> gridPosition_;
    LocationIterator location_;
    bool lastLocationInCell() const
    {
        return lastLocation() || grid_[grid_.getCellIndex(gridPosition_) + 1] == *static_cast<const LocationIterator *>(this);
    }
};

/**
 ** operator== from a LocationIterator
 **/
template<class CoordinateT>
inline bool operator==(const typename PreFilterIterator<CoordinateT>::LocationIterator &lhs,
                       const PreFilterIterator<CoordinateT> &rhs)
{
    return rhs == lhs;
}

/**
 ** operator != from a LocationIterator
 **/
template<class CoordinateT>
inline bool operator!=(const typename PreFilterIterator<CoordinateT>::LocationIterator &lhs,
                       const PreFilterIterator<CoordinateT> &rhs)
{
    return !(lhs == rhs);
}

/**
 ** PreFilter<CoordinateT, HashFunctionT>::iterator
 **     ::iterator(const PreFilter<CoordinateT, HashFunctionT> &preFilter,
 **                const LocationIterator referenceLocation)
 **/
template<class CoordinateT>
PreFilterIterator<CoordinateT>::PreFilterIterator(const Grid<CoordinateT> &grid,
                                                  const LocationIterator &center,
                                                  const CoordinateT &radius)
    : grid_(grid)
    , center_(center)
    , radius_(radius)
    , firstPosition_(grid_.getGridPosition(center_->getX()-radius_, center_->getY()-radius_))
    , lastPosition_ (grid_.getGridPosition(center_->getX()+radius_, center_->getY()+radius_))
{
    gridPosition_  = firstPosition_;
    *static_cast<LocationIterator *>(this) = grid_[grid_.getCellIndex(gridPosition_)];
    nextValidPosition();
}

/**
 ** PreFilter<CoordinateT, HashFunctionT>::iterator::operator ++() 
 **/
template<class CoordinateT>
inline PreFilterIterator<CoordinateT> &
PreFilterIterator<CoordinateT>:: operator++()
{
    ++(*static_cast<LocationIterator *>(this));
    nextValidPosition();
    return *this;
}

/**
 ** PreFilter<CoordinateT, HashFunctionT>::iterator::operator ++() 
 **/
template<class CoordinateT>
void PreFilterIterator<CoordinateT>::nextValidPosition()
{
    while (lastLocationInCell() && gridPosition_.first <= lastPosition_.first)
    {
        if (lastPosition_.second > gridPosition_.second) // increment Y
        {
            gridPosition_.second++;
        }
        else // increment X and reset Y
        {
            gridPosition_.first++;
            gridPosition_.second = firstPosition_.second;
        }
        if (gridPosition_.first <= lastPosition_.first)
        {
            *(static_cast<LocationIterator *>(this)) = grid_[grid_.getCellIndex(gridPosition_)];
        }
        else
        {
            *(static_cast<LocationIterator *>(this)) = grid_.getTile().end();
        }
    }
}

/**
 ** FilterIterator 
 **/
template<class CoordinateT>
class FilterIterator;

/**
 ** Filter.
 **/
template<class CoordinateT>
class Filter
{
public:
    typedef ClusterLocation<CoordinateT> Location;
    typedef vector<Location> LocationList;
    typedef typename LocationList::const_iterator LocationIterator;
    Filter(LocationList &tile, pair<CoordinateT, CoordinateT> binSize) 
        : preFilter_(tile, binSize), tile_(tile) {}
    // iterator on the original list of cluster locations, skipping all irrelevant elements.
    typedef FilterIterator<CoordinateT>  iterator;
    iterator getNeighborhood(const LocationIterator center, const CoordinateT &radius) const 
    {
        return iterator(preFilter_, tile_, center, radius);
    }
private:
    const PreFilter<CoordinateT> preFilter_;
    const LocationList &tile_;
};

/**
 ** Filter::iterator
 ** Same as the PreFilter::iterator, except that it is an exact filter
 **/
template<class CoordinateT>
class FilterIterator : public PreFilterIterator<CoordinateT>
{
public:
    typedef ClusterLocation<CoordinateT> Location;
    typedef vector<Location> LocationList;
    typedef typename LocationList::const_iterator LocationIterator;
    typedef FilterIterator<CoordinateT> iterator;
    inline iterator &operator++();
    FilterIterator(const PreFilter<CoordinateT> &preFilter, 
                   const LocationList &tile,
                   const LocationIterator &center, 
                   const CoordinateT &radius) 
        : PreFilterIterator<CoordinateT>(preFilter.getNeighborhood(center, radius))
    {
        nextValidPosition();
    }
protected:
    inline void nextValidPosition();
    inline CoordinateT squaredDistance();
};

/**
 ** FilterIterator nextValidPosition
 ** Overloaded to get an exact filter
 **/
template<class CoordinateT>
void FilterIterator<CoordinateT>::nextValidPosition()
{
    const CoordinateT squaredRadius = this->getRadius() * this->getRadius();
    while ((!this->lastLocation()) && (*static_cast<LocationIterator *>(this) == this->getCenter() || squaredDistance() >= squaredRadius))
    {
        ++(*static_cast<PreFilterIterator<CoordinateT> *>(this));
    }
}

/**
 ** FilterIterator operator++
 ** Overloaded to get an exact filter
 **/
template<class CoordinateT>
FilterIterator<CoordinateT> &
FilterIterator<CoordinateT>::operator++()
{
    ++(*static_cast<PreFilterIterator<CoordinateT> *>(this));
    nextValidPosition();
    return *this;
}

/**
 ** FilterIterator squaredDistance
 **/
template <class CoordinateT>
inline CoordinateT FilterIterator<CoordinateT>::squaredDistance()
{
    const CoordinateT dx = this->getCenter()->getX() - (*this)->getX();
    const CoordinateT dy = this->getCenter()->getY() - (*this)->getY();;
    return dx*dx + dy*dy;
}

}} // namespace Cluster { namespace Neighborhood {

#endif // #ifndef CLUSTER_NEIGHBORHOOD_CPP

