/*
 * Copyright (c) Illumina 2008
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
#include <cstdio>
#include <cassert>
#include "LinAlg.h"
#include "CycleMatrix.h"
#include "MatrixEstimator.h"

CycleMatrix::CycleMatrix(const int numChannels) 
{
    m_numChannels = numChannels;
}

void CycleMatrix::clear()
{
    for (iterator i = begin(); end() != i; ++i) 
        delete [] *i;
    std::vector< float * >::clear();
}

CycleMatrix::~CycleMatrix()
{
    clear();
}

int CycleMatrix::estimate(const std::vector<float * > & cycleData,
                          const int                     numObjects,
                          const int                     numChannels)
{
    m_numChannels              = numChannels;
    MatrixEstimator *estimator = new MatrixEstimator();
    int              errorCode = 0;
    for (int i = 0; i < int(cycleData.size()); i++) {
        float *matrixData = new float [ numChannels * numChannels ];
        errorCode =  estimator->estimateMatrix(matrixData, cycleData[i],
                                               numObjects, numChannels);
        push_back(matrixData);
    }
    delete estimator;
    return errorCode;
}

/**
 * Inverts all matrices in the vector.
 */
int CycleMatrix::invert()
{
    int     errorCode = 0;
    float * inverse_matrix = new float [ m_numChannels * m_numChannels ];
    for (int i = 0; i < int(size()); i++) {
        float *matrix = at(i);
        if (matrix != 0) {
            matinvert(matrix, inverse_matrix, m_numChannels);
            at(i) = inverse_matrix;
            inverse_matrix = matrix;
        }
        else
            errorCode = i+1;
    }
    delete [] inverse_matrix;
    return errorCode;
}

int CycleMatrix::print(const TileInfo    & tileInfo,
                       const ImageRegion & imageRegion,
                       FILE              * fout)
{
    int exitCode = 0;
    for (int k = 0; k < int(size()); k++) {
        fprintf(fout, "%i\t%i\t%i\t%i\t%i\t%i\t%i\t",
                tileInfo.Lane+1, tileInfo.TileIndex+1,
                imageRegion.XMin, imageRegion.XMin+imageRegion.XSize,
                imageRegion.YMin, imageRegion.XMin+imageRegion.YSize, k+1);
        int i, j;
        for (i = 0; i < m_numChannels; i++) {
            for (j = 0; j < m_numChannels; j++)
                printf(" %5.2f", at(k)[i*m_numChannels+j]);
            if (i == m_numChannels - 1)
                fprintf(fout, "\n");
            else
                fprintf(fout, "\t");
        }
    }
    return exitCode;
}

int CycleMatrix::write(const char filename[],
                       const TileInfo    & tileInfo,
                       const ImageRegion & imageRegion)
{
    int exitCode = 0;
    FILE *fout = fopen(filename, "w");
    if (!fout)
        return 1;
    exitCode = print(tileInfo, imageRegion, fout);
    if (exitCode == 0)
        exitCode = fclose(fout);
    return exitCode;
}

int CycleMatrix::read(const char filename[])
{
    return 0;
}

int CycleMatrix::apply(float     ** cycledata,
                       const int    numObjects,
                       const int    numChannels,
                       const int    numDataCycles,
                       const int    firstCycle) const
{
    assert(numChannels == m_numChannels);
    assert(int(size())-firstCycle >= numDataCycles);

    int errorCode = 0;
    float *array2  = new float [ numObjects * numChannels];
    for (int i = 0; i < numDataCycles; i++) {
        float *matrix = at(firstCycle+i);
        if (matrix != 0) {
            float * const array = cycledata[i];
            matrix_vectortranspose_multiply(array2, matrix, array,
                                            numChannels, numObjects);
            cycledata[i] = array2;
            array2       = array;
        }
        else
            errorCode = i+firstCycle+1;
    }
    delete [] array2;
    return errorCode;
}

int CycleMatrix::applyInverse(float     ** cycledata,
                              const int    numObjects,
                              const int    numChannels,
                              const int    numDataCycles,
                              const int    firstCycle) const
{
    assert(numChannels == m_numChannels);
    assert(int(size())-firstCycle >= numDataCycles);

    int errorCode = 0;
    float *inverse_matrix = new float [ m_numChannels * m_numChannels ];
    float *array2  = new float [ numObjects * numChannels];
    for (int i = 0; i < numDataCycles; i++) {
        float *matrix = at(firstCycle+i);
        if (matrix != 0) {
            matinvert(matrix, inverse_matrix, m_numChannels);
            float * const array = cycledata[i];
            matrix_vectortranspose_multiply(array2, inverse_matrix, array,
                                            numChannels, numObjects);
            cycledata[i] = array2;
            array2 = array;
        }
        else
            errorCode = i+firstCycle+1;
    }
    delete [] array2;
    delete [] inverse_matrix;
    return errorCode;
}

