/*
 * Copyright (c) Illumina 2008
 * Author: Come Raczy
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#include <algorithm>
#include <math.h>
#include <limits>

#include "CycleMatrixCorrector.hh"

/**
 ** Compute the determinant of a square matrix of size n.
 ** 
 ** Uses a recursive decomposition into minors.
 **
 ** n: the size of the matrix.
 ** data: the matrix
 ** buffer: temporary storage of n(n-1)(2n-1)/6 elements
 **/
template <class T>
T determinant (const T *data, const unsigned int n, T *buffer)
{
    if (0 == n) { return 0; }
    if (1 == n) { return *data; }
    T result = 0;
    for (unsigned int minor = 0; n > minor; ++minor)
    {
        // copy the minor into the buffer
        unsigned int offset = 0;
        for (unsigned int i = 0; n > i; ++i)
        {
            if (minor != i)
            {
                for (unsigned int j = 1; n > j; ++j)
                {
                    buffer[offset] = data[i*n + j];
                    ++offset;
                }
            }
        } 
        // compute the determinant of the minor
        const T minorDeterminant = determinant(buffer, n-1, buffer + (n-1)*(n-1));
        // update the result
        result += (minor % 2 ? -1 : 1) * data[minor * n] * minorDeterminant;
    }
    return result;
}

void CycleMatrixCorrector::scale(const unsigned int numChannels)
{
    float *buffer = new float[numChannels*(numChannels-1)*(2*numChannels-1)];
    for (iterator matrix = begin(); end() != matrix; ++matrix)
    {
        const float d = determinant(*matrix, numChannels, buffer);
        if (d > std::numeric_limits<float>::epsilon( ))
        {
            const unsigned int numElements = numChannels * numChannels;
            for (unsigned int i = 0; numElements > i; ++i)
            {
                *matrix[i] /= d;
            }
        }
    }
    delete [] buffer;
}

int CycleMatrixCorrector::estimate(const std::vector<float * > & cycleData,
                                   const int                     numObjects,
                                   const int                     numChannels)
{
    for (size_t cycle = 0; cycleData.size() > cycle; ++cycle)
    {
        push_back(estimate(cycleData[cycle], numObjects, numChannels));
    }
    return 0;
}

// usual helper method
template<class T> 
T op_sum ( T l, T r) { return l + r; }

/**
 ** Estimate the cross-talk for one cycle.
 **
 ** - classifies the spots according to the index of the brightest
 **   channel;
 **
 ** - compute the average intensity of each channel for each class;
 **
 ** Note: the returned pointer is a dynamically allocated matrix which
 ** needs a delete[] in the client code (currently it is in the
 ** destructor of the parent class which is a bad idea since that
 ** destructor is not virtual).
 **
 **/
float *CycleMatrixCorrector::estimate(const float       *clusterIntensityList,
                                   const unsigned int numberOfClusters,
                                   const unsigned int numberOfChannels)
{
    float *matrixData = new float [ numberOfChannels * numberOfChannels ];
    std::fill_n(matrixData, numberOfChannels * numberOfChannels, 0.0f);
    // accumulate the intensities and residual intensities for each channel separately
    std::vector<unsigned int> clusterCountByChannel(numberOfChannels, 0);
    for (unsigned int cluster = 0; numberOfClusters > cluster; ++cluster)
    {
        const float * const clusterIntensity = clusterIntensityList + cluster * numberOfChannels; 
        const float * const brightestChannel = std::max_element(clusterIntensity, clusterIntensity + numberOfChannels);
        if (*brightestChannel > 0.0)
        {
            const unsigned int channelNumber = brightestChannel - clusterIntensity;
            float * const      rowData       = matrixData + channelNumber * numberOfChannels;
            std::transform(clusterIntensity, clusterIntensity + numberOfChannels, rowData, rowData, op_sum<float>);
            ++clusterCountByChannel[channelNumber];
        }
    }
    // compute the average intensity and average residual intensities for each channel
    for (size_t row = 0; numberOfChannels > row; ++row)
    {
        if (0 != clusterCountByChannel[row])
        {
            for (size_t col = 0; numberOfChannels > col; ++col)
            {
                matrixData[row*numberOfChannels + col] /= (float)clusterCountByChannel[row];
            }
        }
    }
    // transpose the matrix
    for (size_t row = 0; numberOfChannels > row; ++row)
    {
        for (size_t col = row + 1; numberOfChannels > col; ++col)
        {
            const size_t i = row * numberOfChannels + col;
            const size_t j = col * numberOfChannels + row;
            const float tmp = matrixData[i];
            matrixData[i] = matrixData[j];
            matrixData[j] = tmp;
        }
    }
    return matrixData;
}
