/*
 * Matrix estimator.
 *
 * Copyright (c) Illumina 2007
 * Author:       Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#include "MatrixEstimator.h"

#include "estimate_xtalk_mx.h"
#include "scale_normalise_xtalk_mx.h"
#include "optimise_xtalk_mx.h"
#include "savinio_exceptions.h"

MatrixEstimator::MatrixEstimator() : m_Intensities(0), m_NumClusters(0), 
                       m_NumChannels(4), m_ErrorMessage("")
{
}

MatrixEstimator::~MatrixEstimator()
{
    if (m_Intensities)
        delete [] m_Intensities;
}

void MatrixEstimator::setIntensities(const float * intensities,
                                     const int     numClusters, 
                                     const int     numChannels)
{
    m_NumClusters = numClusters;
    m_NumChannels = numChannels;
    m_Intensities = new float [ m_NumClusters * m_NumChannels];
    memcpy(m_Intensities, intensities,
           m_NumClusters * m_NumChannels* sizeof(float));
}

#if 0
int MatrixEstimator::estimateMatrix(float * matrix)
{
    return estimateMatrix(matrix, m_Intensities, m_NumClusters, m_NumChannels);
}
#endif

int MatrixEstimator::estimateMatrix(float       * matrix, 
                                    const float * intensities,
                                    const int     numClusters, 
                                    const int     numChannels)
{
    int    numFilteredClusters = 0;
    float *filteredIntensities = new float [ numClusters * numChannels ];

    // remove objects with zero intensities
    filterIntensities(filteredIntensities, &numFilteredClusters,
                      intensities, numClusters, numChannels);

    // input validation
    if (numFilteredClusters < MIN_NO_CLUSTERS) {
        delete [] filteredIntensities;
        char   tmp[10];
        snprintf(tmp, 10, "%d", numFilteredClusters);
        m_ErrorMessage = "Number of clusters too low: " + std::string(tmp);
        return errorNotEnoughClusters;
    }


    float *temporaryIntensities = new float[ numFilteredClusters * numChannels ];
    Matrix mx1(numChannels, numChannels);
    Matrix mx2(numChannels, numChannels);
    Matrix mx_unscaled;


    // initial estimate of the unscaled cross-talk matrix
    mx1.identity_matrix();

    applyXTalkMatrix(filteredIntensities, numFilteredClusters, numChannels,
                     mx1, temporaryIntensities);

    MatrixSamplingParameters params;
    params.percentageXUpper = PERCENTAGE_UPPER_X_S1_ITER1;
    params.percentageXLower = PERCENTAGE_LOWER_X_S1_ITER1;
    params.percentileY      = PERCENTILE_LOWER_Y_S1_ITER1; 

    estimateCrossTalkMatrix(mx2.data, params,
                            temporaryIntensities, numFilteredClusters, numChannels);

    // final estimate of the unscaled cross-talk matrix
    mx_unscaled = mx1 * mx2;

#if 1
    applyXTalkMatrix(filteredIntensities, numFilteredClusters, numChannels,
                     mx_unscaled, temporaryIntensities);

    MatrixSamplingParameters params2;
    params2.percentageXUpper = PERCENTAGE_UPPER_X_S1_ITER2;
    params2.percentageXLower = PERCENTAGE_LOWER_X_S1_ITER2;
    params2.percentileY      = PERCENTILE_LOWER_Y_S1_ITER2;

    estimateCrossTalkMatrix(mx2.data, params2,
                            temporaryIntensities, numFilteredClusters, numChannels);

    mx_unscaled = mx_unscaled * mx2;
#endif

    // scaling and normalising the cross-talk matrix 
    int errorCode = 0;
    errorCode = normaliseMatrix(matrix, mx_unscaled.data,
                                filteredIntensities, 
                                numFilteredClusters, numChannels);

    delete [] temporaryIntensities;
    delete [] filteredIntensities;

    return errorCode;
}

int MatrixEstimator::normaliseMatrix(float       * outputMatrix,
                                     const float * inputMatrix,
                                     const float * intensities,
                                     const int     numClusters,
                                     const int     numChannels)
{
    try {
        Matrix mx_input(numChannels, numChannels);
        Matrix mx_output;
        memcpy(mx_input.data, inputMatrix,
               numChannels * numChannels * sizeof(float));
        scaleNormaliseXTalkMatrix(intensities, numClusters, numChannels,
                                  mx_input, mx_output);
        memcpy(outputMatrix, mx_output.data,
              numChannels * numChannels * sizeof(float));
    }
    catch (Exception02 e) {
        m_ErrorMessage = "Scaling failure of cross-talk matrix";
        return errorScalingFailure;
    }
    return errorNone;
}

int MatrixEstimator::estimateCrossTalkMatrix(float       * matrix,
                                             const MatrixSamplingParameters & params,
                                             const float * intensities,
                                             const int     numClusters,
                                             const int     numChannels)
{
    Matrix mx;
    optimiseXTalkMatrix(intensities, numClusters, numChannels, mx, params);
    memcpy(matrix, mx.data,
           numChannels * numChannels * sizeof(float));
    return 0;
}

void MatrixEstimator::filterIntensities(float       * outputIntensities,
                                        int         * outputNumClusters,
                                        const float * inputIntensities,
                                        const int     inputNumClusters,
                                        const int     numChannels)
{
    filter_intensities(inputIntensities, outputIntensities, numChannels,
                       inputNumClusters, outputNumClusters);
}
