/*
 * Copyright (c) Solexa 2005, 2006
 *           (c) Illumina 2007
 * Author:      Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#include <cstdio>
#include <iostream>
#include <cassert>
#include <cstring>
#include <algorithm>
#include "PhasingEstimator.h"
#include "Stats.h"

/*
 * apply purity filtering to the data set
 */
void filter_reads(const std::vector<float *> & cycledata,
                  int         * nclusters,
                  const int     numChannels,
                  const float   CHASTITY,
                  const int     PURE_BASES)
{
    int i, inew, k, j;
    int npurebases = PURE_BASES;
    if (npurebases > int(cycledata.size()))
        npurebases = int(cycledata.size());
    inew = 0;
    std::vector<int> lookup(*nclusters);
    for (i = 0; i < *nclusters; i++) {
        lookup[i] = -1;
        float purity = 1.;
        for (k = 0; k < npurebases; k++) {
            float *cyclesignal = &(cycledata[k][i*numChannels]);
            float max1 = 1.;
            float max2 = 1.;
            int   imax1 = -1;
            int   imax2 = -1;
            // find brightest base
            for (j = 0; j < numChannels; j++) 
                if (cyclesignal[j] > max1) {
                    max1 = cyclesignal[j]; imax1 = j;
                }
            // find second brightest base
            for (j = 0; j < numChannels; j++) 
                if ((j != imax1) && (cyclesignal[j] > max2)) {
                    max2 = cyclesignal[j]; imax2 = j;
                }
            if ((imax1 == -1) || (imax2 == -1)) {
                purity = 0.; break;
            }
            purity = max1/(max1+max2);
            if (purity < CHASTITY)
                break;

        }
        if (purity < CHASTITY)
            continue;
        lookup[inew] = i;
        inew++;
    }
    *nclusters = inew;
    for (k = 0; k < int(cycledata.size()); k++)
        for (i = 0; i < *nclusters; i++)
            if (i != lookup[i])
                memcpy(&(cycledata[k][i*numChannels]), 
                       &(cycledata[k][lookup[i]*numChannels]), numChannels*sizeof(float));
}

/* 
 * Perform some base-calling based on the intensity values.
 * The bases are encoded as number from 0-3 for bases A, C, G, T
 * respectively (-1 for no base-call).
 * This could be replaced by sequences read from a sequence file.
 */ 
void call_bases(const std::vector<float *> &cycledata, int nclusters,
                const int numChannels, int *bases)
{
    int i, j, k;
    int ncycles = (int)cycledata.size();

    for (k = 0; k < ncycles; k++) {
        for (i = 0; i < nclusters; i++) {
            float max = 0.f;
            int idx = -1;
            for (j = 0; j < numChannels; j++)
                if (cycledata[k][i*numChannels+j] > max) {
                    max = cycledata[k][i*numChannels+j];
                    idx = j;
                }
            bases[k*nclusters+i] = idx;
        }
    }
}

/*
 * Attempt to estimate ratio of signal from current cycle to some
 * other cycle, for each base separately.
 */
float *estimate_phasing(const std::vector<float *> &cycledata,
                        const int nclusters,
                        int *sequences,
                        const int step)
{
    const int numChannels = 4;
    int   ncycles = int(cycledata.size());
    int   lastcycle  = std::min(ncycles, ncycles - step);
    int   firstcycle = std::max(0, - step);
    int   ndim = lastcycle-firstcycle;
    float *phasing = new float [ numChannels * ndim];
    int   cycle;
    float follow[numChannels], nofollow[numChannels];
    int   count[numChannels], nocount[numChannels];

    for (cycle = firstcycle; cycle < lastcycle; cycle++) {
        int i, j;
        for (j = 0; j < numChannels; j++) {
            follow[j] = nofollow[j] = 0.f;
            count[j] = nocount[j] = 0;
        }
        for (i = 0; i < nclusters; i++) {
            /* skip clusters with zero intensity */
            //assert(cycle*nclusters+i < nclusters*lastcycle);
            //assert((cycle+step)*nclusters+i < nclusters*lastcycle);
            if ((sequences[cycle*nclusters+i] == -1) ||
                (sequences[(cycle+step)*nclusters+i] == -1)) {
                continue;
            }
            assert(cycle < lastcycle);
            for (j = 0; j < numChannels; j++) {
                /* skip clusters with the current base-call */
                if (sequences[cycle*nclusters+i] == j)
                    continue;
                float value = cycledata[cycle][i*numChannels+j];
                if (sequences[(cycle+step)*nclusters+i] == j) {
                    // previous base hast the same base-call
                    follow[j] += value;
                    count[j] += 1;
                }
                else {
                    // previous base has different base-call
                    nofollow[j] += value;
                    nocount[j] += 1;
                }
            }
        }
        for (j = 0; j < numChannels; j++) {
            if (count[j] > 0)
                follow[j] /= count[j];
            if (nocount[j] > 0)
                nofollow[j] /= nocount[j];
            /* 
             * subtract general signal background (previous base different)
             * from signal left over from previous (next) cycle
             */
            phasing[j+(cycle-firstcycle)*numChannels] = follow[j] - nofollow[j];
        }
    }
    return phasing;
}

/*
 * format phasing and prephasing estimates for output
 */
void print_phasing(const float *phasing,
                   const float *prephasing,
                   const int ncycles,
                   const int numChannels,
                   const int phasing_firstcycle,
                   const int phasing_lastcycle,
                   const int prephasing_firstcycle,
                   const int prephasing_lastcycle)
{
    int i, j;
    float *values1 = new float [numChannels];
    float *values2 = new float [numChannels];
    char line[1024];
    for (i = 0; i < ncycles; i++) {
        for (j = 0; j < numChannels; j++) {
            values1[j] = 0.;
            values2[j] = 0.;
        }
        if ((i >= phasing_firstcycle) && (i < phasing_lastcycle))
            memcpy(values1, &phasing[(i-phasing_firstcycle)*numChannels],
                   numChannels * sizeof(float));
        if ((i >= prephasing_firstcycle) && (i < prephasing_lastcycle))
            memcpy(values2, &prephasing[(i-prephasing_firstcycle)*numChannels],
                   numChannels * sizeof(float));
        snprintf(line, sizeof(line),
                 "%i\t%6.4f %6.4f\t%6.4f %6.4f %6.4f %6.4f\t%6.4f %6.4f %6.4f %6.4f\n",
                 i + 1, Stats::average(values1, numChannels), 
                        Stats::average(values2, numChannels),
                 values1[0], values1[1], values1[2], values1[3],
                 values2[0], values2[1], values2[2], values2[3]);
        printf(line);
    }
    delete [] values1;
    delete [] values2;
}

/*************************************************************************/

PhasingEstimator::PhasingEstimator(int numChannels) : m_CycleIntensities(), m_NumClusters(0),
                                       m_NumChannels(numChannels), m_Sequences(0),
                                       m_NumMinClusters(500),
                                       m_NumMaxCycles(12), m_NumMinCycles(6)
{
    m_ErrorMessage = "";
}

PhasingEstimator::~PhasingEstimator()
{
    int i;
    for (i = 0; i < int(m_CycleIntensities.size()); i++)
        delete [] m_CycleIntensities[i];
    if (m_Sequences)
        delete [] m_Sequences;
}

void PhasingEstimator::setMinimumClusterNumber(const int numClusters)
{
    m_NumMinClusters = numClusters;
}

void PhasingEstimator::setMaximumFittedCycles(int num)
{
    m_NumMaxCycles = num;
}

void PhasingEstimator::setIntensities(const std::vector<float *> &cycleIntensities,
                                      const int                   numClusters,
                                      const int                   numCycles)
{
    m_NumClusters      = numClusters;
    int nCycles = numCycles;
    if ((nCycles < 0) || (nCycles > (int)(cycleIntensities.size())))
        nCycles = cycleIntensities.size();
    int i;
    for (i = 0; i < nCycles; i++) {
        float *intensities = new float [ m_NumClusters * m_NumChannels ];
        memcpy(intensities, cycleIntensities[i],
               m_NumClusters * m_NumChannels * sizeof(float));
        m_CycleIntensities.push_back(intensities);
    }

    m_Sequences        = new int [ m_NumClusters * m_CycleIntensities.size() ];
    call_bases(m_CycleIntensities, m_NumClusters, m_NumChannels, m_Sequences);
}

void PhasingEstimator::filterChastity(const float CHASTITY,
                                      const int   PURE_BASES)
{
    filter_reads(m_CycleIntensities, &m_NumClusters, m_NumChannels,
                 CHASTITY, PURE_BASES);
    call_bases(m_CycleIntensities, m_NumClusters, m_NumChannels, m_Sequences);
}

/* normalise cluster intensities such that the brightest base 
 * has a value of 1
 */    
void PhasingEstimator::normalise() 
{
    int i, j, k;
    for (k = 0;  k < (int)m_CycleIntensities.size(); k++) {
        for (i = 0; i < m_NumClusters; i++) {
            float norm = 1.;
            for (j = 0; j < m_NumChannels; j++)
                if (m_CycleIntensities[k][i*m_NumChannels+j] > norm)
                    norm = m_CycleIntensities[k][i*m_NumChannels+j];
            for (j = 0; j < m_NumChannels; j++)
                m_CycleIntensities[k][i*m_NumChannels+j] /= norm;
        }
    }
}

float *PhasingEstimator::estimateCycles(const int step)
{
    return estimate_phasing(m_CycleIntensities, m_NumClusters, m_Sequences, step);
}

int PhasingEstimator::estimateCycles(float **phasing,
                                     float **prePhasing)
{
    if (m_NumClusters < m_NumMinClusters) {
        *phasing    = 0;
        *prePhasing = 0;
        m_ErrorMessage = "Not enough clusters to estimate phasing";
        return errorNotEnoughClusters;
    }
    normalise(); // TODO: This should really move into setIntensities()
                 // kept here at the moment for exact compatibility with 
                 // older versions
    *phasing    = estimateCycles(-1);
    *prePhasing = estimateCycles( 1);
    return errorNone;
}

float *get_phasing_cycle(float *array, int numCycles, int numChannels)
{
    float *result = new float [numChannels];
    int   j;
    for (j = 0; j < numChannels; j++)
        result[j] = 0.;
    return result;
}


int PhasingEstimator::estimate(float *phasing,
                               float *prePhasing,
                               const float *phasingCycles,
                               const float *prePhasingCycles)
{
    *phasing    = 0.;
    *prePhasing = 0.;
    int numCycles = m_NumMaxCycles;
    if (numCycles > int(m_CycleIntensities.size()))
        numCycles = int(m_CycleIntensities.size());
    if (numCycles < m_NumMinCycles) {
        m_ErrorMessage = "Not enough cycles available to estimate phasing";
        return errorNotEnoughCycles;
    }
    {
        float gradient, intercept;
        float *x = new float [ numCycles ];
        float *y = new float [ numCycles ];
        int   i;
        for (i = 1; i < numCycles - 1; i++) {
            x[i-1] = float(i+2);
            y[i-1] = Stats::average(&phasingCycles[i*m_NumChannels], m_NumChannels);
            //printf("%f %f\n", x[i-1], y[i-1]);
        }
        Stats::linregress(x, y, numCycles - 2, & gradient, & intercept);
        *phasing = gradient;
        for (i = 1; i < numCycles - 1; i++) {
            x[i-1] = float(i+1);
            y[i-1] = Stats::average(&prePhasingCycles[i*m_NumChannels], m_NumChannels);
            //printf("%f %f\n", x[i-1], y[i-1]);
        }
        Stats::linregress(x, y, numCycles - 2, & gradient, & intercept);
        *prePhasing = gradient;

        delete [] x;
        delete [] y;
    }
    return errorNone;
}

int PhasingEstimator::estimate(float *phasing,
                               float *prePhasing)
{
    float *phasingCycles     = 0;
    float *prePhasingCycles  = 0;
    int   errorCode = 0;

    errorCode = estimateCycles(&phasingCycles, &prePhasingCycles);
    if (errorCode == errorNone)
        errorCode = estimate(phasing, prePhasing, phasingCycles, prePhasingCycles);
    delete [] phasingCycles;
    delete [] prePhasingCycles;
    return errorCode;
}


void PhasingEstimator::printCycles(const float *phasingCycles,
                                   const float *prePhasingCycles)
{
    int ncycles = int(m_CycleIntensities.size());
    int phasing_step = -1;
    int phasing_lastcycle  = std::min(ncycles, ncycles-phasing_step);
    int phasing_firstcycle = std::max(0, -phasing_step);
    int prephasing_step = 1;
    int prephasing_lastcycle  = std::min(ncycles, ncycles-prephasing_step);
    int prephasing_firstcycle = std::max(0, -prephasing_step);
    print_phasing(phasingCycles, prePhasingCycles,
                  int(m_CycleIntensities.size()), m_NumChannels,
                  phasing_firstcycle, phasing_lastcycle,
                  prephasing_firstcycle, prephasing_lastcycle);
}

