/*
 * Copyright (c) 2006-2007 Solexa
 *               2007      Illumina
 * Authors: Klaus Maisinger, Lorenzo Milazzo
 *
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

/*
 * Defines a model for the so-called prephasing/phasing effect, 
 * and create correction factors for the effect.
 *
 * By `phasing` we mean an effect that is caused by incomplete incorporation
 * cycles or incomplete deblocking and results in a given fraction of molecules
 * in a cluster lagging behind the current incorporation cycle.
 * (Similarly, `prephasing` refers to molecules running ahead for the current
 * incorporation cycle).
 */

#if defined(__BORLANDC__)
#include <stdlib.h>
#include <string.h>
#include <math.h>
#else
#include <cstdlib>
#include <cstring>
#include <cmath>
#endif
#include <vector>
#include <cstdio>
#include <cassert>

#include "PhasingModel.h"
#include "LinAlg.h"


/*
 * Generate a `phasing sub-matrix`.
 * The phasing sub-matrix (or, in the case of single run on a given cluster, the
 * phasing matrix) is a <ncycles>*<ncycles>-matrix that contains in the r-th
 * row the relative fraction of molecules that are expected to be in the j-th
 * cycle after r cycles of incorporation; without prephasing, all elements will
 * be zero for j>r.
 * Each row of this matrix is referred to the `phasing window function` for the
 * corresponding cycle.
 */
void definePhasingSubMatrix(float *matrix, int mx_dim,
                            int smx_dim_start, int smx_dim_end,
                            float phasingrate, float prephasingrate)
{ 
    // row and column indeces
    int r, c;
    // first row, second row, last row
    int r0, r1, rm;


    r0 = smx_dim_start-1;
    r1 = smx_dim_start;
    rm = smx_dim_end;

    matrix[r0*mx_dim+r0] = 1.f-phasingrate-prephasingrate;
    matrix[r0*mx_dim+r1] = prephasingrate;
    for (r = r1; r < rm; r++) {
        matrix[r*mx_dim+r0] = matrix[(r-1)*mx_dim+r0] * phasingrate;
        matrix[r*mx_dim+r1] = matrix[(r-1)*mx_dim+r1] * phasingrate +
                              matrix[(r-1)*mx_dim+r0] * (1.f-phasingrate-prephasingrate);
                     
        for (c = (smx_dim_start+1); c < smx_dim_end; c++) {
            matrix[r*mx_dim+c] = matrix[(r-1)*mx_dim+c] * phasingrate +
                                 matrix[(r-1)*mx_dim+c-1] * (1.f-phasingrate-prephasingrate) +
                                 matrix[(r-1)*mx_dim+c-2] * prephasingrate;
        }
    }
}


/*
 * Generate a `phasing matrix`.
 *
 * In the case of single run on a given cluster, the `phasing matrix` coincides
 * with  the `phasing sub-matrix`.
 * In the case of multiple runs on a given cluster, the `phasing matrix` contains
 * two or more `phasing sub-matrices`.
 *
 */
void getPhasingMatrix(float *matrix, int ncycles,
                      const std::vector<BasecallParameters> & reads)
{
    int i, j;
    int mx_dim;
    int smx_dim_start;
    int smx_dim_end;

    // validation
    assert(ncycles >= 1);
    for (i = 0; i < int(reads.size()); i++)
        assert(reads.at(i).FirstCycle < ncycles);

    mx_dim = ncycles;

    // initialising the matrix
    for (i = 0; i < (mx_dim*mx_dim); i++) matrix[i] = 0.;

    j = 0;
    for (i = 0; i < static_cast<int>(reads.size()); i++) {
        smx_dim_start = reads.at(i).FirstCycle + 1;
        if (i == int(reads.size()-1)) {
            smx_dim_end = mx_dim;
        }
        else {
            j = i+1;
            smx_dim_end = reads.at(j).FirstCycle;
        }
        // defining the phasing sub-matrix
        definePhasingSubMatrix(matrix, mx_dim, smx_dim_start, smx_dim_end,
                               reads[i].PhasingRate, reads[i].PrephasingRate);
    }
}


/*
 * Invert the phasing matrix generated by the function `getPhasingMatrix()`
 * and select the last row of this matrix to generate the inverse window 
 * function.
 */
void getPhasingInvWindow(float *matrix, int ncycles, int cycle, float *window) 
{
    float *inverse = new float [ ncycles*ncycles ];

#if 0
    char msg[10024];
    sprint_matrix(msg, sizeof(msg), matrix, ncycles, ncycles);
    printf("%s\n", msg);
#endif

    matinvert(matrix, inverse, ncycles);

#if 0
    sprint_matrix(msg, sizeof(msg), inverse, ncycles, ncycles);
    printf("%s\n", msg);
#endif

    int i;
    for (i = 0; i < ncycles; i++)
        window[i] = inverse[(cycle-1)*ncycles+i];
    delete [] inverse;
}


/*************************************************************************/
PhasingModel::PhasingModel()
{
}

PhasingModel::PhasingModel(const struct BasecallParameters & params)
{
    addRead(params);
}

PhasingModel::PhasingModel(const struct BasecallParameters & params,
                           std::vector<int> new_runs)
{
    int i;
    for (i = 0; i < static_cast<int>(new_runs.size()); i++) {
        BasecallParameters param;
        param = params;
        param.FirstCycle = new_runs[i];
        addRead(param);
    }
}

PhasingModel::PhasingModel(const std::vector<BasecallParameters> &params)
{
    int i;
    for (i = 0; i < static_cast<int>(params.size()); i++)
        addRead(params[i]);
}

PhasingModel::PhasingModel(const struct BasecallParameters *params,
                           int numParams)
{
    int i;
    for (i = 0; i < numParams; i++)
        addRead(params[i]);
}

PhasingModel::~PhasingModel()
{
    Reads.clear();
}

void PhasingModel::addRead(const struct BasecallParameters & params)
{
    Reads.push_back(params);
}

/*
 * Given <ncycles> scan cycles (incorporation or cleavage/deblock) of sequencing,
 * an <ncycles>-dim array <is_deblocked> of flags for deblocked cycles,
 * a phasing rate <phasing_rate>, find the most important <nwindow>
 * elements of the phasing window function <window> and a lookup array
 * <window_cycle> to identify the cycles that contribute to the window 
 * function. 
 * The arrays <window> and <window_cycle> must be allocated prior to 
 * the function call and contain at least <nwindow> elements; allocating
 * <ncycles> elements is a safe option.
 */
std::vector<PhasingWindow> PhasingModel::getPhasingCorrection(int cycle, int ncycles,
                                                              int *is_deblocked)
{
    int nallcycles = ncycles;
    int ncurcycles = cycle;
    int nallinccycles = 0;
    int ncurinccycles = 1;
    int i;
    std::vector<PhasingWindow> window;
    int *inc_cycle_index = new int [ nallcycles ];
 	for (i = 0; i < nallcycles; i++)
 		inc_cycle_index[i] = 0;
    // count all incorporation cycles; if current cycle is deblocked, then
    // count the current cycle and omit the next incorporation cycle 
    for (i = 0; i < nallcycles; i++) { 
        if ((i == cycle-1) ||
            ((!is_deblocked[i]) && 
             (!((i-1 == cycle-1) && is_deblocked[cycle-1])) ) ) {
            inc_cycle_index[nallinccycles] = i;
            nallinccycles ++;
        }
    }
#if defined(DEBUG_BASECALL)
    printf("Index ");
    for (i = 0; i < nallcycles; i++) 
        printf("%i %i  ", i, inc_cycle_index[i]);
    printf("\n");	
#endif

    for (i = 0; i < ncurcycles-1; i++) 
        if (!is_deblocked[i])
            ncurinccycles ++;

#if defined(DEBUG_BASECALL)
    printf ("cycle counts %i %i inc %i %i\n", ncurcycles, nallcycles,
            ncurinccycles, nallinccycles); 
    printf("phasing %f %f\n", Reads[0].PhasingRate, Reads[0].PrephasingRate);   
#endif
    // create the "phasing matrix". This matrix contains for each
    // nominal incorporation cycle (rows) the fraction of molecules
    // that is present in any other cycle (columns).
    // By definition, the (sum of all elements in each row) <= 1.
    // This function needs to be replaced for a different phasing
    // model. 
    float *phasing_matrix = new float [ nallinccycles * nallinccycles];
    getPhasingMatrix(phasing_matrix, nallinccycles, Reads);

#if 0
    print_matrix(phasing_matrix, nallinccycles, nallinccycles);
#endif
                       
    // create the full inverse window function for the current cycle
    float *full_window = new float [ nallinccycles ];
    getPhasingInvWindow(phasing_matrix, nallinccycles, ncurinccycles,
                        full_window);

    delete [] phasing_matrix;
    
#if defined(DEBUG_BASECALL)
    printf("Full window: ");
    for (i = 0; i < nallinccycles; i++)
        printf("%5.3f ", full_window[i]);
    printf("\n");
#endif

    // copy all values of window function above a minimum threshold
    float threshold = 0.01f;
    int *window_inc_cycle = new int [ nallinccycles ];
    for (i = 0; i < nallinccycles; i++) {
        if (fabs(full_window[i]) > threshold) {
            PhasingWindow w;
            window_inc_cycle[window.size()] = i;
            w.value = full_window[i];
            w.cycle = inc_cycle_index[i];
            window.push_back(w);
#if defined(DEBUG_BASECALL)
            printf("%i %i %i %i %f\n", window.size()-1, i, inc_cycle_index[i],
                   w.cycle, w.value);
#endif
        }
    }
    
    delete [] window_inc_cycle;
    delete [] inc_cycle_index;
    delete [] full_window;
    return window;
}  

