/*
 * Copyright (c) 2005-2007 Solexa
 * Author: Klaus Maisinger
 *
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <stdlib.h>
#include <math.h>
#else
#include <cstdlib>
#include <cmath>
#endif
#include <cassert>
#include <algorithm>

#include "ProbabilityIntegration.h"
#include "Random.h"
#include "LinAlg.h"
 
 /*
 * This class helps to determine the likelihood of a given base-call
 * by integrating the posterior probabilities over
 * the region in signal space where this base has the largest signal.
 * This space is delineated by 6 hypersurfaces (A=C, A=G, etc.).
 * Gaussian errors are assumed.
 * The integration uses an analytical integration over a spherical Gaussian
 * intersected by a plane. The intensity space is transformed to a new
 * representation in which the likelihood is spherical (the metric is based 
 * on the covariance matrix).
 * 
 * As a fall-back solution, a Monte-Carlo method can be used:
 * Random samples from the likelihood distribution are drawn
 * and the relative base-call frequencies for each base are counted.
 * To speed up computations, a set of random Gaussian deviates is 
 * buffered in memory.
 */
ProbabilityIntegration::ProbabilityIntegration(int nnsamples, int nncol,
                                               float max_score_)
{
    nsamples = nnsamples;
    ncol = nncol;
    max_score = max_score_;
    samples = new float [nsamples*ncol];
    drawSamples();
}

ProbabilityIntegration::~ProbabilityIntegration()
{
    delete [] samples;
}
    
void ProbabilityIntegration::drawSamples(unsigned int seed)
{
    // draw samples from a Gaussian normal distribution.
    // recycle these to speed up computation.
    srand(seed);
    multivariate_normal(samples, nsamples*ncol);
}

/*
 * These functions are extracted from ProbabilityIntegrator::integrate()
 * method. gcc seems to find it easier to optimise this way.
 */
void initialise_samples(float *samples2, float *signal, int ncol, int nsamples)
{
    int i, l;
    for (l = 0; l < nsamples; l++) 
        for (i = 0; i < ncol; i++)
            samples2[l*ncol+i] = signal[i];
}


/*
 * Obtain sample with appropriate covariance structure.
 * The code is a bit opaque - it's essentially a repeated matrix multiplication,
 * with loops reordered.
 * gcc seems to have problems optimising more readable versions.
 */
void add_error(float *samples2, float *samples,
               float *error_covariance_decomposition, int ncol, int nsamples)
{
    int i, j, l;
    float fac, *s1, *s2;
    // obtain sample with the appropriate covariance structure 
    for (i = 0; i < ncol; i++) 
        for (j = 0; j < ncol; j++) {
            fac = error_covariance_decomposition[i*ncol+j];
            s1 = samples +j;
            s2 = samples2+i;
            for (l = 0; l < nsamples*ncol; l+=ncol) {
                s2[l] += fac * s1[l];
            }
        }
}

/*
 * Count number of hits for each base
 */
void count_hits(float *samples2, float *count, int ncol, int nsamples)
{
    int l, i;
    for (i = 0; i < ncol; i++)
        count[i] = 0.;
    for (l = 0; l < nsamples; l++) {
        int imax = 0;
        int l2 = l*ncol;
        float max = samples2[l2+0];
        
        // find brightest base
        for (i = 1; i < ncol; i++)
            if (samples2[l2+i] > max) {
                max = samples2[l2+i]; imax = i;
            }
            
        count[imax] += 1.f;
    }
}


/*
 * Integrate the signal over the <nncol>-dimensional hyperspace
 * corresponding the <nncol> bases. The most likely signal vector
 * is given by the <nncol>-dimensional vector <signal>.
 * The covariance matrix C of the errors on <signal> is given
 * by its decomposition C= D D^T, where D is specified
 * by the <nncol>*<nncol>-matrix <error_covariance_decomposition>.
 * (In the case where D is lower triangular, this decomposition
 * is called Cholesky decomposition.)
 * The function returns the <nncol> quality scores <quality_scores>,
 * which are defined as 10*log10(p/(1-p)), where p is the 
 * probability of the corresponding base.
 */
void ProbabilityIntegration::integrate(float *signal, int nncol,
                                       float *error_covariance_decomposition,
                                       float *quality_scores)
{
    assert(ncol == nncol);

    // if signal is identical to zero, all calls are equally likely
    int i;
    int zeroflag = 1;
    for (i = 0; i < ncol; i++)
        if (signal[i] > 0.)
            zeroflag = 0;

    float *pvec = new float[ncol];
    if (zeroflag) {
        for (i = 0; i < ncol; i++)
            pvec[i] = 1./ncol;
    }

    if (! zeroflag) {
        float *dvec = new float [ncol - 1];
        float *inverse = new float [ncol * ncol];
        
        matinvert(error_covariance_decomposition, inverse, ncol);

        // Calculate the distance from the nearest hyperplane
        // in the 4-d intensity space that separates the current base-call 
        // from a different base-call. Say the base-call is A, then the
        // 3 hyperplanes are given by A=C, A=G, A=T. The distance
        // is calculated in a coordinate system in which the likelihood
        // is spherical.
        float sn = calculateDistanceVector(ncol, signal, 
                                           error_covariance_decomposition,
                                           inverse, dvec);
        delete [] inverse;

        if (sn >  3.) { // if signal-to-noise is sufficient
            int imax = 0;
            float max = signal[imax];
            for (i = 1; i < ncol; i++)
                if (signal[i] > max) {
                    imax = i;
                    max = signal[i];
                }
            pvec[imax] = 1.;
            int j = 0;
            // Convert the distance to the hyperplane to a probability by 
            // an integration of a spherical Gaussian, with the tail removed
            // in one dimension.
            for (i = 0; i < ncol; i++) {
                if (i != imax) {
                    pvec[i] = 1.- PofD(dvec[j]);
                    pvec[imax] -= pvec[i];
                    j++;
                }
            }
             
#if 0
            printf("l %f d %f %f %f pvec %f %f %f %f\n", sn,
                   dvec[0], dvec[1], dvec[2], pvec[0], pvec[1], pvec[2], pvec[3]);
#endif
        }
        else {
            // if signal-to-noise is poor, revert to stochastic integration
            float *count = new float [ncol];
            float *samples2 = new float [ncol*nsamples];

            // use a stochastic integration by drawing random samples 
            // from the posterior probability distribution
            // these are separate subroutines to make profiling easier;
            // also gcc seems to find it easier to optimise this way.

            // initialise each sample with the signal vector
            initialise_samples(samples2, signal, ncol, nsamples);
            // add noise obtained from the covariance decomposition
            add_error(samples2, samples, error_covariance_decomposition, 
                      ncol, nsamples);
            // count total number of hits for each base
            count_hits(samples2, count, ncol, nsamples);
            // avoid singular values 0 and 1
            for (i = 0; i < ncol; i++) {
                if ((int)count[i] == 0)
                    count[i] += 1.f;
                if ((int)count[i] == nsamples)
                    count[i] -= 1.f;
            }
            for (i = 0; i < ncol; i++)
                pvec[i] = count[i] / nsamples;
            delete [] samples2;
            delete [] count;
    
        }
        delete [] dvec;
    }
    const float tiny      = 1.e-6;
    for (i = 0; i < ncol; i++) {
        if (pvec[i] < tiny) 
            quality_scores[i] = -max_score;
        else if  (1.- pvec[i] < tiny)
            quality_scores[i] = max_score;
        else
            quality_scores[i] = 10.f * log10(pvec[i] / (1.-pvec[i]));
        quality_scores[i] = std::max(quality_scores[i], -max_score);
        quality_scores[i] = std::min(quality_scores[i],  max_score);
    }
    
    delete [] pvec;

}
