/*
 * Copyright (c) Solexa   2005-2007
 *               Illumina 2007
 * Authors: Klaus Maisinger
 *
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

/*
 * The core routines of the base-caller are found in this file.
 */

#include <vector>
#include <cassert>
#include <cstring>
#include <algorithm>

#include "BaseCallConfig.h"
#include "PhasingModel.h"
#include "ProcessBaseCall.h"
#include "ProbabilityIntegration.h"
#include "BaseCallConfig.h"
#include "DNACodes.h"
#include "CycleMatrix.h"

#include "Maths.h"
#include "LinAlg.h"
#include "Stats.h"
#include "Tools.h"
#include "Message.h"
#include "IntensityData.h"


ProcessBaseCall::ProcessBaseCall(const int numCycles)
{ 
    m_NumCycles   = numCycles;
    m_NumChannels = BaseCallConfig::numBases;
    m_BaseOrder = new unsigned char [ m_NumChannels ];
    for (int k = 0; k < m_NumChannels; k++)
        m_BaseOrder[k] = BaseUpperCase[k];
    m_DeblockOrder = new int [ m_NumCycles ];
    for (int i = 0; i < m_NumCycles; i++)
        m_DeblockOrder[i] = 0;
}

ProcessBaseCall::~ProcessBaseCall()
{
    delete [] m_BaseOrder;
    delete [] m_DeblockOrder;
}

void ProcessBaseCall::setCycleNumber(const int numCycles)
{
    m_NumCycles = numCycles;
    delete [] m_DeblockOrder;
    m_DeblockOrder = new int [ m_NumCycles ];
    for (int i = 0; i < m_NumCycles; i++)
        m_DeblockOrder[i] = 0;
}

void ProcessBaseCall::setDeblockingOrder(const int * deblockOrder,
                                         const int   numCycles)
{
    if (numCycles >= 0)
        m_NumCycles = numCycles;
    delete [] m_DeblockOrder;
    m_DeblockOrder = new int [ m_NumCycles ];
    for (int i = 0; i < m_NumCycles; i++)
        m_DeblockOrder[i] = deblockOrder[i];
}

void ProcessBaseCall::setPhasingModel(const PhasingModel &pm)
{
    m_PhasingModel = pm;
}

void ProcessBaseCall::setBaseOrder(const unsigned char * baseOrder,
                                   const int             numChannels)
{
    assert(numChannels <= 4);
    m_NumChannels = numChannels;
    for (int i = 0; i < m_NumChannels; i++)
        m_BaseOrder[i] = baseOrder[i];
}


/**
 * Goes through intensity values and makes sure that each of the 4 values
 * for a cluster is non-zero (hasn't shifted off the image); otherwise
 * set to zero (no base-call).
 */
void validate_signal(float     * signal,
                     float     * noise,
                     const int   nobj,
                     const int   ncol)
{
    int k, i;
    for (k = 0; k < nobj; k++) {
        if (Stats::min(&noise[k*ncol], ncol) <= 0.) {
             for (i = 0; i < ncol; i++) {
                 signal[k*ncol+i] = 0.;
                 noise[k*ncol+i] = 0.;
            }
        }
    }
}


/**
 * Goes through intensity values from all cycles used for phasing correction
 * and makes sure that every single one of them is set to zero (no base-call)
 * if the current cycle is zero (cluster has shifted off the image).
 */
void validate_signals(std::vector<float * > & signal,
                      std::vector<float * > & noise,
                      const int               nobj,
                      const int               ncol,
                      const int               index)
{
    int k, l, i;
    int nw = int(signal.size());
    for (k = 0; k < nobj; k++) {
        if (Stats::min(&noise[index][k*ncol], ncol) <= 0.)
            for (l = 0; l < nw; l++)
               for (i = 0; i < ncol; i++) {
                   signal[l][k*ncol+i] = 0.;
                   noise [l][k*ncol+i] = 0.;
               }
    }
}


/**
 * Estimates the average signal per cycle; this can be used to normalise
 * the intensities
 */
std::vector<float> estimateAverageSignals(const std::vector<float * > & signal,
                                          const std::vector<float * > & noise,
                                          const int                     nobj,
                                          const int                     ncol)
{
    assert(signal.size() == noise.size());
    int k, l, i;
    std::vector<int>   counts(signal.size());
    std::vector<float> average(signal.size());
    for (l = 0; l < int(counts.size()); l++) {
        average[l] = 0.;
        counts[l]  = 0;
    }
    for (k = 0; k < nobj; k++) {
        for (l = 0; l < int(counts.size()); l++)
            if (Stats::min(&noise[l][k*ncol], ncol) > 0.) {
                counts[l]  ++;
                for (i = 0; i < ncol; i++)
                    average[l] += signal[l][k*ncol+i];
            }
    }
    for (l = 0; l < int(counts.size()); l++)
        if (counts[l] > 0)
            average[l] /= counts[l];
    return average;
}   


/**
 * Calculates the noise covariance matrix based on the matrix and 
 * phasing corrections.
 */
void calculateCovariance(float     *covariance, 
                         const int  ndim,
                         const int  k,
                         const std::vector<PhasingWindow> & window,
                         const std::vector<float * >      & matrices,
                         const std::vector<float * >      & allnoise)
{
    int i, j, l, m, i2, j2, k2;
    k2 = k * ndim;
    initialise_array(covariance, ndim*ndim);
    for (l = 0; l < int(window.size()); l++) {
        float window2 = window[l].value * window[l].value;
        float *noise  = allnoise[l];
        float *matrix = matrices[l];
        for (i = 0; i < ndim; i++) {
            i2 = i*ndim;
            for (j = 0; j < ndim; j++) {
                j2 = j * ndim;
                for (m = 0; m < ndim; m++)
                    covariance[i*ndim+j] += window2
                                * matrix[i2+m] * matrix[j2+m] 
                                * noise[k2+m] * noise[k2+m];
            }
        }
    }
}


/**
 * Estimates relative base probabilities:
 * - For each cluster, obtain the noise covariance matrix of the 
 *   matrix-corrected intensities. The intensity covariance matrix
 *   is assumed to be diagonal with elements given by the error estimates
 *   obtained from the image analysis. Then apply matrix correction 
 *   to these values to obtain the covariance of the matrix-corrected data.
 * - Integrate the likelihood over the 4-d volumes defined by the 4 possible
 *   base calls (I_A >= I_X, X=C,G,T; I_C >= I_X, X=A,G,T; etc.)
 */
void estimateProbabilities(const std::vector<float * > &allsignal,
                           const std::vector<float * > &allnoise,
                           const std::vector<float * > &matrices,
                           const std::vector<PhasingWindow> &window,
                           float     * signal,
                           float     * probabilities,
                           const int   nobj,
                           const int   ncol)
{
    int i, k;
    float *covariance = new float [ncol*ncol];

    ProbabilityIntegration *integrator 
              = new ProbabilityIntegration(BaseCallConfig::numMCSamples, ncol,
                                           BaseCallConfig::maxQScore);
    for (k = 0; k < nobj; k++) {
        /* 
         * calculate noise covariance matrix
         */
        calculateCovariance(covariance, ncol, k,
                            window, matrices, allnoise);

        /*
         * Decompose covariance matrix and integrate 4d signal space
         * to obtain relative base probabilities
         */
        float *original = new float [ncol*ncol];
        for (i = 0; i < ncol*ncol; i++)
            original[i] = covariance[i];
            if (Stats::max(covariance, ncol*ncol) > 0.)
                if (cholesky_decompose(covariance, ncol)) {
                    char msg[10000];
                    int  msg_pos = 0;
                    error("Covariance matrix not positive definite.");
                    msg_pos = sprintf(msg, "Covariance: ");
                    snprint_matrix(msg+msg_pos, sizeof(msg) - msg_pos,
                                   original, ncol, ncol);
                    error(msg);
                }
        delete [] original;

        integrator->integrate(&signal[k*ncol], ncol,
                              covariance, &probabilities[k*ncol]);
    }
    delete [] covariance;
    delete integrator;
}

/**
 * Calls bases for each cycle and object/cluster in a list.
 * Uses the intensities in each colour channel to determine the sequence.
 * Probability estimates are adapted to avoid base-calls with scores less
 * then a non-called base.
 *
 * @param intensity     [numObjects * numChannels] array of intensities for
 *                      all objects in the observed channel
 * @param qualityscores [numObjects * numChannels] array of the background noise
 *                      dispersions at the molecule positions
 * @param numObjects    number of molecules/colonies for which a base
 *                      is to be determined
 * @param numChannels   number of colour channels observed
 * @param baseOrder     [numChannels] array of bases for each colour channel
 * @param sequence      [numObjects] array containing a base code for each 
 *                      object. The array has to be allocated
 *                      by the caller.      
 */
void callBases(const float         * intensity,
               const int             numObjects,
               const int             numChannels,
               const unsigned char * baseOrder,
               float               * qualityScores,
               unsigned char       * sequence)
{
    assert(numChannels > 0);
    assert(numObjects  >= 0);

    /* call bases based on intensity values */
    int i, j;
    for (i = 0; i < numObjects; i++) {
        sequence[i] = '.';
        int   maxIndex     = -1;
        float maxIntensity =  0.;
        for (j = 0; j < numChannels; j++)
            if (intensity[i*numChannels+j] > maxIntensity) {
                maxIntensity = intensity[i*numChannels+j];
                maxIndex     = j;
            }

        if (maxIndex >= 0)
            sequence[i] = baseOrder[maxIndex];
    }
}

/**
 * Overrides quality scores when they don't agree with the intensity based
 * base calls. (The quality scores may be based on stochastic integration
 * and can deviate from the base call.)
 */
void adjustScores(const float         * intensity,
                  const int             numObjects,
                  const int             numChannels,
                  const unsigned char * baseOrder,
                  float               * qualityScores,
                  unsigned char       * sequence)
{
    int i, j;
    for (i = 0; i < numObjects; i++) {
        // if no call, set all probabilities to 1/4
        if (sequence[i] == '.') {
            for (j = 0; j < numChannels; j++)
                qualityScores[i*numChannels+j] = -5; // 10.f * log10(p / (1.-p)), p =1/4
            continue;
        }
        // check if the base call agrees with the highest score; if not,
        // average probabilities of both and assign the same probability to both.
        int   maxIntensityIndex = -1;
        for (j = 0; j < numChannels; j++)
            if (sequence[i] == baseOrder[j])
                maxIntensityIndex = j;
        assert(maxIntensityIndex >= 0);
        while (Stats::max(& qualityScores[i*numChannels], numChannels) > 
               qualityScores[i*numChannels + maxIntensityIndex]) {
            // identify highest score
            int   maxScoreIndex = 0;
            float maxScore      = qualityScores[i*numChannels];
            for (j = 1; j < numChannels; j++)
                if (qualityScores[i*numChannels+j] > maxScore) {
                    maxScore = qualityScores[i*numChannels+j];
                    maxScoreIndex = j;
                }
            // average probabilities
            float p = 1 - 1./(1. + pow(10., qualityScores[i*numChannels+maxScoreIndex]/10.))
                    + 1.- 1./(1. + pow(10., qualityScores[i*numChannels+maxIntensityIndex]/10.));
            float score = 10.f * log10((p/2.) / (1. - (p/2.)));
            qualityScores[i*numChannels+maxScoreIndex]     = score;
            qualityScores[i*numChannels+maxIntensityIndex] = score;
       }
    }
}

/**
 * Calculates the number of cycles around the current cycle that are needed
 * to do an optimal phasing correction. Since phasing is a cumulative effect,
 * the number of cycles needed around the current cycle depends on the total
 * number of cycles and the phasing rates. A symmetric window is assumed,
 * so the number returned fits the expression 2*n+1, n integer.
 */
#if 0
int ProcessBaseCall::getWindowSize(const int nCycles)
{
    int numCycles = nCycles;
    if (numCycles == 0)
        numCycles = m_NumCycles;
    assert(numCycles > 0);
    int windowSize = 0;
    int cycle;
    for (cycle = numCycles - 1; (cycle >= 0); cycle--) {
        std::vector<PhasingWindow> window = m_PhasingModel.getPhasingCorrection(cycle+1, numCycles,
                                                                        m_DeblockOrder);
        int length   = int(window.size());
        int distance = Maths::max(cycle - window[0].cycle,
                                  window[length-1].cycle - cycle);
        if (2 * distance + 1 > windowSize)
            windowSize = 2 * distance + 1;
    }
    return windowSize;
}
#endif
/**
 * Calculates the number of cycles around the current cycle that are needed
 * to do an optimal phasing correction. Since phasing is a cumulative effect,
 * the number of cycles needed around the current cycle depends on the total
 * number of cycles and the phasing rates. A symmetric window is assumed,
 * so the number returned fits the expression 2*n+1, n integer.
 */
int ProcessBaseCall::getWindowSize(const int nCycles)
{
    int numCycles = nCycles;
    if (numCycles == 0)
        numCycles = m_NumCycles;
    assert(numCycles > 0);
    int windowSize = 0;
    int cycle;
    // loop contains arbitrary cut-off of 5 cycles - anything shorter is 
    // to unlikely to have an effect
    for (cycle = numCycles - 1; (cycle >= numCycles-5) && (cycle >= 0); cycle--) {
        std::vector<PhasingWindow> window = m_PhasingModel.getPhasingCorrection(cycle+1, numCycles,
                                                                        m_DeblockOrder);
        int length   = int(window.size());
        int distance = std::max(cycle - window[0].cycle,
                                window[length-1].cycle - cycle);
        if (2 * distance + 1 > windowSize)
            windowSize = 2 * distance + 1;
    }
    return windowSize;
}

/**
 * Goes through intensities and noise estimates and eliminates any clusters
 * for which less than 4 values have been detected (i.e. clusters that lie
 * on the edge of the image).
 */
void ProcessBaseCall::validateIntensities(float     ** intensities,
                                          float     ** noise,
                                          const int    numObjects,
                                          const int    numChannels,
                                          const int    numCycles)
{
    int k;
    for (k = 0; k < numCycles; k++)
        validate_signal(intensities[k], noise[k], numObjects, numChannels);
}


void ProcessBaseCall::applyMatrixCorrection(float             ** intensities,
                                            const CycleMatrix &  matrices,
                                            float             ** matrix_signal,
                                            const int            numClusters,
                                            const int            numChannels,
                                            const int            numCycles,
                                            const int            firstCycle)
{
    int i;
    for (i = 0; i < numCycles; i++)
        memcpy(matrix_signal[i], intensities[i], 
               numClusters * numChannels * sizeof(float));
    matrices.apply(matrix_signal, numClusters, numChannels, numCycles, firstCycle);
}


/**
 * Implements a phasing correction.
 * Identifies a set of cycles that the current cycle has leaked signal to,
 * calculates an inverse "window function" (a filter to reverse the effect
 * of phasing) for all cycles that contribute significantly, loads the
 * corresponding intensity files and applies the filter correction.
 */
void ProcessBaseCall::applyPhasingCorrection(float                ** matrixSignal,
                                             float                ** noise,
                                             std::vector<float * > & inverseMatrices,
                                             const int               numObjects, 
                                             const int               numChannels,
                                             const int               cycle,
                                             float                 * dephasedSignal,
                                             float                 * probabilities,
                                             const int               numDataCycles,
                                             const int               firstDataCycle)
{
    int lastDataCycle = firstDataCycle + numDataCycles - 1;
    int numCycles     = std::max(firstDataCycle + numDataCycles,
                                 m_NumCycles);
    assert(firstDataCycle <= cycle);
    assert(lastDataCycle  >= cycle);

    /* 
     * Construct the "phasing correction", a function that tells us 
     * in what ratios to add signals from different incorporation cycles.
     * generate a list of all relevant cycles.
     */
    std::vector<PhasingWindow> origWindow = m_PhasingModel.getPhasingCorrection(cycle+1, numCycles,
                                                                    m_DeblockOrder);
    std::vector<PhasingWindow> window;
    int l;
    for (l = 0; l < int(origWindow.size()); l++)
        if ((origWindow[l].cycle >= firstDataCycle) &&
            (origWindow[l].cycle <= lastDataCycle))
            window.push_back(origWindow[l]);

    int window_index = 0;                   // current cycle
    for (l = 0; l < int(window.size()); l++)
        if (cycle == window[l].cycle)
            window_index = l;

#if defined(DEBUG_BASECALL)
    printf("Window %2i %2i %f %f\n", 
           m_NumCycles, nwindow, m_PhasingModel.phasing_rate, m_PhasingModel.prephasing_rate);
    for (l = 0; l < nwindow; l++)
        printf("%2i %2i %g\n", l, window_cycles[l], window[l]);
#endif // defined(DEBUG_BASECALL)

    /*
     * Now construct a signal vector containing the signal from all cycles
     * that have a non-negligible contribution to the phasing/prephasing.
     */
    CycleData            windowSignal;
    CycleData            windowNoise;
    std::vector<float *> windowMatrices;

    for (l = 0; l < int(window.size()); l++) {
        float *this_signal    = new float [numObjects*numChannels];
        float *this_noise     = new float [numObjects*numChannels];

        memcpy(this_signal, matrixSignal[window[l].cycle-firstDataCycle], numObjects*numChannels*sizeof(float));
        memcpy(this_noise,  noise       [window[l].cycle-firstDataCycle], numObjects*numChannels*sizeof(float));
        // verify that all objects have been remapped
        windowSignal.push_back(this_signal);
        windowNoise.push_back(this_noise);
        windowMatrices.push_back(inverseMatrices[window[l].cycle]);
    } 

    /*
     * check that every cluster is remapped in every cycle the phasing
     * correction is applied to (i.e. the cluster hasn't shifted out of the
     * image frame.
     */
    validate_signals(windowSignal, windowNoise, numObjects, numChannels, window_index);

#if 1
    /*
     * rescale window function by average signal in each cycle to take
     * out effects of signal decay and possible focus issues.
     */
    std::vector<float> average_signals =
        estimateAverageSignals(windowSignal, windowNoise, numObjects, numChannels); 
    for (l = 0; l < int(window.size()); l++)
#if 0
        if ((average_signals[l] > 0.) &&
            (average_signals[l] > average_signals[window_index]))
#else
        if (average_signals[l] > 0.)
#endif
            window[l].value *= (average_signals[window_index] / average_signals[l]);
#endif // #if

    /* 
     * calculate signal by applying the phasing correction
     */
    int i;
    for (i = 0; i < numObjects*numChannels; i++)
        dephasedSignal[i] = 0.f;
    for (l = 0; l < int(window.size()); l++) {
        for (i = 0; i < numObjects*numChannels; i++)
            dephasedSignal[i] += window[l].value * windowSignal[l][i];
    }

    /*
     * compute confidence estimates on base calls
     */
    estimateProbabilities(windowSignal, windowNoise, windowMatrices, window,
                          dephasedSignal, probabilities, numObjects, numChannels);
}

/**
 * Chooses the dominant base based on the matrix-corrected,
 * phasing-corrected intensities.
 * Attempts to estimate probabilities by performing an integration
 * of the posterior distribution obtained from the noise covariance.
 *
 * @param matrixSignal    input array of size (numDataCycles) x (numObjects
 *                                                               *numChannels)
 *                        containing matrix corrected intensities
 * @param noise           input array containing (uncorrected) noise values
 * @param inverseMatrices contains the inverse frequency matrices
 * @param numObjects      number of clusters for current tile
 * @param numChannels     number of frequency channels,
 * @param cycle           current imaging cycle (0-indexed)
 * @param sequence        output array containing the sequence (dim numObjects)
 * @param qualityScores   output array containing the quality scores
 *                        (dimension numObjects * numChannels)
 * @param numDataCycles   number of imaging cycles in matrixSignal
 * @param firstDataCycle  first imaging cycle in matrixSignal (0-indexed)
 */
void ProcessBaseCall::call(float                ** matrixSignal,
                           float                ** noise,
                           std::vector<float * > & inverseMatrices,
                           const int               numObjects,
                           const int               numChannels,
                           const int               cycle,
                           unsigned char         * sequence,
                           float                 * qualityScores,
                           float                 * signal,
                           const int               numDataCycles,
                           const int               firstDataCycle)
{
    /*
     * estimate the relative base probabilities
     */
    applyPhasingCorrection(matrixSignal, noise, inverseMatrices,
                           numObjects, numChannels, cycle,
                           signal, qualityScores, numDataCycles, firstDataCycle);

    /*
     * assign bases (based on highest intensity; due to sampling errors
     * this may be different from the most likely base; but of course,
     * that is a rare case and indicates a very poor base call.
     */
    callBases(signal, numObjects, numChannels, m_BaseOrder,
              qualityScores, &sequence[cycle*numObjects]);

#if 1
    /*
     * adjust quality scores so that they agree with the calls.
     * Due to the stochastic integration, this is not automatically the case.
     */
    adjustScores(signal, numObjects, numChannels, m_BaseOrder,
                 qualityScores, &sequence[cycle*numObjects]);
#endif
}
