/*
 * Random number generation routines.
 * 
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#include <stdlib.h>
#else
#include <cmath>
#include <cstdlib>
#endif

/*
 * generate a uniform random sample drawn from the interval
 * [ 0, 1 ).
 */ 
float uniform()
{
    return (float)rand()/(RAND_MAX+1.0f);
}

/*
 * generate a uniform random sample drawn from the interval
 * [ <min>, <max> ).
 */ 
float uniform(float min, float max)
{
    return min + (max-min) * uniform();
}

/*
 * generate a normally distributed random sample (of unit variance)
 */
float normal()
{
    static int   sample_in_store = 0;
    static float sample = 0.;
    
    if (sample_in_store) {
        sample_in_store = 0;
        return sample;
    } 
    else {
        float sample1, sample2, squared, transform;
        do {
            sample1 = uniform(-1., 1.);
            sample2 = uniform(-1., 1.);
            squared = sample1*sample1 + sample2*sample2;
        } while ((squared >= 1.) || (squared == 0.));
        transform = sqrt(-2. * log(squared) / squared);
        
        sample_in_store = 1;
        sample = sample1 * transform;
        return sample2 * transform;        
    }
}

/*
 * generate a normally distributed random sample of mean
 * <mean> and standard deviation <sigma>
 */
float normal(float mean, float sigma)
{
    return mean + sigma * normal();
}

/*
 * generate a multivariate normally distributed random sample.
 * Each element is indenpendent with unit variance.
 */
void multivariate_normal(float *x, int n)
{
    int i;
    for (i = 0; i < n; i++)
        x[i] = normal();
}

/*
 * given the lower triangular matrix <covariance_cholesky>, 
 * which was obtained from a Cholesky decomposition of the covariance
 * matrix, generate a multivariate normally distributed sample with
 * the given covariance structure.
 */
void multivariate_normal(float *x, int n, float *covariance_cholesky)
{
    int i, j;
    float *y = new float [n];
    multivariate_normal(y, n);
    for (i = 0; i < n; i++) {
        x[i] = 0.;
        for (j = 0; j < n; j++)
            x[i] += covariance_cholesky[i*n+j] * y[j];
    }
    delete [] y;
} 
