/*
 * $RCSfile: estimate_xtalk_mx.cpp,v $
 * Author: Lorenzo Milazzo
 * $Id: estimate_xtalk_mx.cpp,v 1.2 2008/07/03 11:36:01 craczy Exp $
 *
 * Copyright (c) 2006, Solexa Ltd.
 *               2008, Illumina
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#include "estimate_xtalk_mx.h"
#include "optimise_xtalk_mx.h"
#include "savinio_common.h"
#include "scale_normalise_xtalk_mx.h"
#include "sampling.h"
#include "savinio_exceptions.h"

/**
 * Applies a cross-talk matrix correction:
 * (new data) = (cross-talk matrix)^-1 * (data)
 */
void applyXTalkMatrix(const float  * int_data,
                      const int      no_objects,
                      const int      no_cols,
                      const Matrix & mx,
                      float        * int_data_new)
{
    Matrix line(4, 1);
    Matrix line_new(4, 1);
    Matrix mx_inv;

    mx_inv = mx.inverse();

    for (int r = 0; r < no_objects; ++r) {
        for (int c = 0; c < no_cols; ++c) {
            line.data[c] = int_data[c + r * no_cols];
        }
        line_new = mx_inv * line;
        for (int c = 0; c < no_cols; ++c) {
            int_data_new[c + r * no_cols] = line_new.data[c];
        }
    }

}

/*
 * remove objects with zero intensities
 */
void filter_intensities(const float *input, float *output,
                        int ncols, int n_input, int *n_output)
{
    int i, j;
    *n_output = 0;
    for (i = 0; i < n_input; i++) {
        int flag = 0;
        for (j = 0; j < ncols; j++)
            if (input[i*ncols+j] != 0.) {
                flag = 1;
                break;
            }
        if (flag) {
            for (j = 0; j < ncols; j++)
                output[(*n_output) * ncols + j] = input[i*ncols+j];
            (*n_output) ++;
        }
    }
}


void estimateXTalkMatrix(const float * i_data_orig,
                         const int   no_objects_orig,
                         const int   no_cols,
                         Matrix    & mx_normalised)
{
    int    no_objects = 0;
    float *i_data = new float [no_objects_orig * no_cols];

    // remove objects with zero intensities
    filter_intensities(i_data_orig, i_data, no_cols,
                       no_objects_orig, &no_objects);

    // input validation
    if (no_objects < MIN_NO_CLUSTERS) {
        delete [] i_data;
        Exception01 e1((const char *)"", no_objects);
        throw e1;
    }

    // intensities (temporary)
    float *i_data_1 = new float[no_objects * no_cols];

    // cross-talk matrix (temporary)
    Matrix mx1, mx2;
    // cross-talk matrix (unscaled)
    Matrix mx_unscaled;


    // initial estimate of the unscaled cross-talk matrix
    mx1.identity_matrix();

    applyXTalkMatrix(i_data, no_objects, no_cols, mx1, i_data_1);

    MatrixSamplingParameters params;
    params.percentageXUpper = PERCENTAGE_UPPER_X_S1_ITER1;
    params.percentageXLower = PERCENTAGE_LOWER_X_S1_ITER1;
    params.percentileY      = PERCENTILE_LOWER_Y_S1_ITER1; 

    optimiseXTalkMatrix(i_data_1, no_objects, no_cols, mx2, params);

    // final estimate of the unscaled cross-talk matrix
    mx_unscaled = mx1 * mx2;

#if 1
    applyXTalkMatrix(i_data, no_objects, no_cols, mx_unscaled, i_data_1);

    MatrixSamplingParameters params2;
    params2.percentageXUpper = PERCENTAGE_UPPER_X_S1_ITER2;
    params2.percentageXLower = PERCENTAGE_LOWER_X_S1_ITER2;
    params2.percentileY      = PERCENTILE_LOWER_Y_S1_ITER2;
    
    optimiseXTalkMatrix(i_data_1, no_objects, no_cols, mx2, params2);
    
    mx_unscaled = mx_unscaled * mx2;
#endif

    // scaling and normalising the cross-talk matrix  
    scaleNormaliseXTalkMatrix(i_data, no_objects, no_cols,
                              mx_unscaled, mx_normalised);

    delete [] i_data_1;
    delete [] i_data;
}
