/*
 * $RCSfile: fitting.cpp,v $
 * Author: Lorenzo Milazzo
 * $Id: fitting.cpp,v 1.2 2008/07/03 11:36:01 craczy Exp $
 *
 * Copyright (c) 2006, Solexa Ltd.
 *               2008, Illumina
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
#include <cstring>
#include <cmath>
#include "fitting.h"
#include "Optimise.h"

/**
 * Contains the arguments for the function 'LAVFunc()';
 * @param no_el    number of elements of the distribution
 * @param data_x   distribution associated with the X axis
 * @param data_y   distribution associated with the Y axis
 */
struct args_lav {
    int    no_el;
    float *data_x;
    float *data_y;
};

/**
 * LAV function (Least Absolute Values, L1 norm)
 * @param  x  slope of the fitting line (independent variable)
 * @return L1 norm
 */
float LAVFunc(float x, void *p_args_lav_v)
{
    struct args_lav *p1 = (struct args_lav *)p_args_lav_v;

    float l1_norm = 0.0;
    for (int i = 0; i < p1->no_el; i++) {
        l1_norm += fabs(x * p1->data_x[i] - p1->data_y[i]);
    }

    return (l1_norm);
}

/**
 * Fits a straight line throught the orgin to the data using the L1 norm.
 */
void L1fit(const float * xaxis_data,
           const float * yaxis_data, 
           const int     no_bins,
           float       * slope)
{
    struct args_lav s_args_lav;
    s_args_lav.no_el = no_bins;
    s_args_lav.data_x = new float[no_bins];
    s_args_lav.data_y = new float[no_bins];
    memcpy(s_args_lav.data_x, xaxis_data, no_bins * sizeof(float));
    memcpy(s_args_lav.data_y, yaxis_data, no_bins * sizeof(float));
    void *p_args_lav_v = (void *)(&s_args_lav);

    // bracketing procedure
    float xbracket[3];
    float ydummy[3];
    xbracket[0] = (float)1 / 2.0;
    xbracket[1] = 2.f;
    xbracket[2] = 0.f;
    bracket(LAVFunc, p_args_lav_v,
            &xbracket[0], &xbracket[1], &xbracket[2],
            &ydummy[0], &ydummy[1], &ydummy[2]);

    // minimisation of the function associated to the LAV by using the Brent's method
    int   funcallcount;
    float xmin;
    float fmin;
    float xtol = 2.0e-3;
    brent(LAVFunc, p_args_lav_v, xbracket, xtol, &xmin, &fmin, &funcallcount);

    *slope = xmin;

    delete [] s_args_lav.data_x;
    delete [] s_args_lav.data_y;
}
