/*
 * $RCSfile: sampling.cpp,v $
 * Author: Lorenzo Milazzo
 * $Id: sampling.cpp,v 1.2 2008/07/03 11:36:01 craczy Exp $
 *
 * Copyright (c) 2006, Solexa Ltd.
 *               2008, Illumina
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#include <cassert>
#include <vector>
#include "Sorting.h"
#include "sampling.h"
#include "savinio_common.h"
#include "savinio_debugging.h"


/**
 * Derives a suitable y value from the binned y data.
 */
static float getYAxisData(const float * bin_data,
                          const int     b_start,
                          const int     b_end,
                          const MatrixSamplingParameters params)
{
    int     numElements = b_end - b_start;
    float * bin_data_t = new float[numElements];

    memcpy(bin_data_t, bin_data + b_start, numElements * sizeof(float));
    Sorting::sort(bin_data_t, size_t(numElements));

    int index_lbound = (int)(params.percentileY * numElements);
    float yaxis_data_el = bin_data_t[index_lbound];

    delete [] bin_data_t;

    return yaxis_data_el;
}


/**
 * Bins the intensities associated with the X and Y axes and
 * calculates the corresponding X,Y percentiles for each bin.
 * Note that there are can be <nmChannels> X-axes
 * with <nChannels-1> y-axes, one for each channel.
 */
void binAxes(int            base_index,
             const int      no_samples,
             const int      numChannels, 
             const float *  intensityData,
             float       *  xaxis_data,
             float       ** yaxis_data,
             const MatrixSamplingParameters params)
{
    /*
     * Find a suitable binning for the X axis data.
     */
    // make a copy of the intensity data for the channel currently used as X axis
    float *intensityBaseX = new float [no_samples];
    int j;
    for (j = 0; j < no_samples; j++)
        intensityBaseX[j] = intensityData[base_index + j * numChannels];

    // sort X axis data
    size_t *indices = new size_t [no_samples];
    for (j = 0; j < no_samples; j++)
        indices[j] = j;
    Sorting::sort_index(intensityBaseX, indices, size_t (no_samples));
    int index_lbound = (int)(params.percentileXLower * no_samples);
    int index_ubound = (int)(params.percentileXUpper * no_samples);

    float Xel_lbound = intensityBaseX[indices[index_lbound]];
    float Xel_ubound = intensityBaseX[indices[index_ubound]];
    if (params.percentageXUpper > 0.)
        Xel_ubound = params.percentageXUpper * Xel_ubound;
    if (params.percentageXLower > 0.)
        Xel_lbound = params.percentageXLower * Xel_ubound;

    for (int b = 0; b < params.numBins; b++) {
        xaxis_data[b] = Xel_lbound + (b + 0.5) * (Xel_ubound - Xel_lbound) / params.numBins;
    }

    /*
     * Sample the Y-axes,
     */
    // identify dimensions other than the current main axis
    std::vector<int> otherDimensions(numChannels-1);
    {
        int i = 0, j = 0;
        for (i = 0; i < numChannels; i++) 
            if (i != base_index) {
                otherDimensions[j] = i;
                j++;
            }
    }

    std::vector<float * > intensitiesBaseY(numChannels-1);
    for (j = 0; j < numChannels-1; j++)
        intensitiesBaseY[j] = new float[no_samples * numChannels];
    int *bin_occupancy = new int[ params.numBins ];

    int binOffset = 0;
    while (intensityBaseX[indices[binOffset]] < Xel_lbound)
        binOffset++;
    int allBinCount = 0;
    int binIndex    = binOffset;
    for (int b = 0; b < params.numBins; b++) {
        float Xel_max = Xel_lbound + (b + 1) * (Xel_ubound - Xel_lbound) / params.numBins;
        while (intensityBaseX[indices[binIndex]] < Xel_max) {
            for (int k = 0; k < numChannels-1; k++)
                intensitiesBaseY[k][binIndex-binOffset] = intensityData[otherDimensions[k] + indices[binIndex] * numChannels];
            binIndex++;
        }
        int binCount = binIndex - binOffset - allBinCount;
        bin_occupancy[b] = binCount;
        allBinCount += binCount;
    }

    int b_start = 0;
    int b_end   = 0;
    for (int b = 0; b < params.numBins; b++) {
        b_end = b_start + bin_occupancy[b];
        for (j = 0; j < numChannels - 1; j++)
            yaxis_data[j][b] = getYAxisData(intensitiesBaseY[j], b_start, b_end, params);
        b_start = b_end;
    }

    delete [] intensityBaseX;
    for (j = 0; j < numChannels-1; j++)
        delete [] intensitiesBaseY[j];
    delete [] indices;
    delete [] bin_occupancy;

#ifdef DEBUG_SAMPLING1
    std::cout << "\n>> debug - sampling1";
    std::cout << "\nX axis: " << std::endl;
    for (int i = 0; i < numBins; ++i) {
        std::cout << xaxis_data[i] << " ";
    }
    for (j = 0; j < numChannels - 1; j++) {
        std::cout << "\nY axis " << (j+1) << ":" << std::endl;
        for (int i = 0; i < numBins; ++i) {
            std::cout << yaxis_data[j][i] << " ";
        }
    }
    std::cout << "\n";
#endif // DEBUG_SAMPLING1
}


/**
 * Chooses the brightest intensities for each cluster.
 * Corresponding intensity values are kept in separate vectors for each base.
 */
void baseCall(const float          * input_data,
              const int              no_points,
              const int              no_cols,
              std::vector<float *> & data,
              std::vector<int>     & no_data)
{
    for (int i = 0; i < no_cols; i++)
        no_data[i] = 0;

    std::vector<float> objectIntensities(no_cols);
    for (int r = 0; r < no_points; ++r) {
        for (int i = 0; i < no_cols; ++i)
            objectIntensities[i] = input_data[r * no_cols + i];

        int   max_index = 0;
        float max       = objectIntensities[0];

        for (int i = 0; i < no_cols; i++)
            if (objectIntensities[i] > max) {
                max       = objectIntensities[i];
                max_index = i;
            }
        data[max_index][no_data[max_index]] = max;
        no_data[max_index] ++;
    }
}


/**
 * Filters out the very low intensities.
 */
void filterLowIntensities(const float * data,
                          const int     numData,
                          float       * dataFiltered,
                          int         * numDataFiltered)
{
    int   index_ubound   = int(PERCENTILE_UPPER_X_S3 * (numData + 1));
    float lowerThreshold = (PERCENTAGE_S3 * data[(index_ubound-1)]);

    int j = 0;
    for (int i = 0; i < numData; i++) {
        if (data[i] > lowerThreshold) {
            dataFiltered[j] = data[i];
            j++;
        }
    }

    *numDataFiltered = j;
}
