 /*
 * $RCSfile: scale_normalise_xtalk_mx.cpp,v $
 * Author: Lorenzo Milazzo
 * $Id: scale_normalise_xtalk_mx.cpp,v 1.2 2008/07/03 11:36:01 craczy Exp $
 *
 * Copyright (c) 2006, Solexa Ltd.
 *               2008, Illumina
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#include <cmath>
#include <cassert>
#include <vector>

#include "Optimise.h"
#include "Sorting.h"

#include "estimate_xtalk_mx.h"
#include "sampling.h"
#include "savinio_common.h"
#include "savinio_debugging.h"
#include "savinio_exceptions.h"
#include "scale_normalise_xtalk_mx.h"

/**
 * Contains the arguments for the function 'KSStatisticFunc()'
 * @param  data1    distribution 1 (reference distr.)
 * @param  data2    distribution 2
 * @param  p_func   pointer to the function 'scaledCumulDistrFunc()'
 */
struct args_ks {
    std::vector<float> data1;
    std::vector<float> data2;
    float (*p_func) (const float, const std::vector<float> &);
};

/**
 * Scaled cumulative distribution function (scaled cdf)
 */
float scaledCumulDistrFunc(const float                 value,
                           const std::vector<float > & data)
{
    int index = std::distance(data.begin(),
                              std::lower_bound(data.begin(), data.end(), value));
    return float(index) / data.size();
}


/**
 * KS statistic function
 * (Kolmogorov-Smirnov statistic, D-statistic)
 *
 *  @param  x  scale factor (independent variable)
 *  @return    KS statistic
 */
float KSStatisticFunc(float x,
                      void *p_args_ks_v)
{
    struct args_ks *p1 = (struct args_ks *)p_args_ks_v;

    float d = 0.f;
    int i;
    float ff, fo = 0.f, fn;
    for (i = 0; i < int(p1->data1.size()); i++) {
        fn = (float)(i + 1) / (p1->data1.size());
        ff = (*p1->p_func) (x * p1->data1[i], p1->data2);
        float dt = std::max(fabs(fo - ff), fabs(fn - ff));
        if (dt > d)
            d = dt;
        fo = fn;
    }

    return (d);
}


/**
 * Scales two distributions.
 */
static float scaleDistr(const float * distr1,
                        const int     no_el1,
                        const float * distr2,
                        const int     no_el2)
{
    // construct arguments for the KS statistic function
    struct args_ks s_args_ks;
    s_args_ks.data1.resize(no_el1);
    s_args_ks.data2.resize(no_el2);
    memcpy(&(s_args_ks.data1[0]), distr1, no_el1 * sizeof(float));
    memcpy(&(s_args_ks.data2[0]), distr2, no_el2 * sizeof(float));
    s_args_ks.p_func = scaledCumulDistrFunc;
    void *p_args_ks_v = (void *)(&s_args_ks);

    float xbracket[3];
    float ydummy[3];
    xbracket[0] = (float)1 / 2.0;
    xbracket[1] = 2.f;
    xbracket[2] = 0.f;
    bracket(KSStatisticFunc, p_args_ks_v,
            &xbracket[0], &xbracket[1], &xbracket[2],
            &ydummy[0], &ydummy[1], &ydummy[2]);

    int   funcallcount = 0;
    float xmin;
    float fmin;
    float xtol = 2.0e-3;
    brent(KSStatisticFunc, p_args_ks_v, xbracket, xtol, &xmin, &fmin, &funcallcount);

#if defined(DEBUG_MIN)
    std::cout << "\n>> debug - minimisation" << std::endl;
    std::cout.setf(std::ios::fixed, std::ios::floatfield);
    std::cout.precision(7);
    std::cout << "bracketing: " << std::endl;
    std::cout << " ax, bx, cx: " << xbracket[0] << " " << xbracket[1] << " " << xbracket[2] << std::endl;
    std::cout << " fa, fb, fc: " << ydummy[0] << " " << ydummy[1] << " " << ydummay[2] << std::endl;
    std::cout << "d statistic (minimum value): " << fmin << std::endl;
    std::cout << "scale factor: " << xmin << std::endl;
#endif // DEBUG_MIN

    return (xmin);
}


/**
 * Scales the columns of the he cross-talk matrix.
 * The scale factors are calculated by minimising a function associated to
 * the Kolmogorov-Smirnov statistic (KS statistic or D statistic)
 */
static void scaleXTalkMatrix(std::vector <float *>  & data_by_base,
                             const std::vector<int> & no_data_by_base,
                             const Matrix           & mx_unscaled,
                             Matrix                 & mx_scaled)
{
    int numChannels = int(no_data_by_base.size());
    assert(numChannels > 0);
    std::vector<float> scale_factors(numChannels, 1.);
    Matrix mx_inv;

    try {
        // selected data validation
        int flag = 0;
        for (int i = 0; i < numChannels; i++)
            if (no_data_by_base[i] < MIN_NO_SEL_DATA)
                flag = 1;
        if (flag) {
            Exception02 e2;
            throw e2;
        }

        for (int i = 1; i < numChannels; i++)
            scale_factors[i] =
                scaleDistr(data_by_base[0], no_data_by_base[0],
                           data_by_base[i], no_data_by_base[i]);

        mx_inv = mx_unscaled.inverse();
    }
    catch(Exception02 e2) {
        e2.display();
        mx_inv.identity_matrix();
    }

    for (int r = 0; r < mx_inv.rows; ++r) {
        for (int c = 0; c < mx_inv.cols; ++c) {
            mx_inv.data[c + r * mx_inv.cols] /= scale_factors[r];
        }
    }

    mx_scaled = mx_inv.inverse();

#if defined(DEBUG_SCALING)
      std::cout << "\n>> debug - scaling" << std::endl;

      std::cout.setf(std::ios::fixed, std::ios::floatfield);
      std::cout.precision(7);

      std::cout << "scale factors: "
                << scale_factors[0] << " " << scale_factors[1] << " "
                << scale_factors[2] << " " << scale_factors[3] << std::endl;

#endif // DEBUG_SCALING
}


/**
 * Normalises the cross-talk matrix to make its determinant equal to 1.
 */
void normaliseXTalkMatrix(const Matrix & matrixScaled,
                          Matrix       & matrixNormalised)
{
    float norm_factor = 1.0 / pow(fabs(matrixScaled.det()), 1.0 / matrixScaled.rows);
    matrixNormalised = matrixScaled * norm_factor;
}


void scaleNormaliseXTalkMatrix(const float * intensityData,
                               const int     numObjects,
                               const int     numChannels,
                               Matrix      & matrixUnscaled,
                               Matrix      & matrixNormalised)
{
    float *intensityDataScaled = new float[numObjects * numChannels];
    applyXTalkMatrix(intensityData, numObjects, numChannels,
                     matrixUnscaled, intensityDataScaled);

    std::vector<int>       numDataByBase(numChannels);
    std::vector<float  * > dataByBase(numChannels);
    for (int i = 0; i < numChannels; i++)
        dataByBase[i] = new float [ numObjects ];

    // preliminary base-calling
    baseCall(intensityDataScaled, numObjects, numChannels,
             dataByBase, numDataByBase);

    delete [] intensityDataScaled;

    // sort data for each channel
    for (int i = 0; i < numChannels; i++)
        Sorting::sort(dataByBase[i], size_t(numDataByBase[i]));

    // filter out low intensity values
    std::vector<int>      numDataByBaseFiltered(numChannels);
    std::vector<float * > dataByBaseFiltered(numChannels);
    for (int i = 0; i < numChannels; i++)
        dataByBaseFiltered[i] = new float [ numObjects ];

    for (int i = 0; i < numChannels; i++)
        filterLowIntensities(dataByBase[i], numDataByBase[i],
                             dataByBaseFiltered[i], &numDataByBaseFiltered[i]);

    for (int i = 0; i < numChannels; i++)
        delete [] dataByBase[i];

    // equalise scales of intensity distributions between bases
    Matrix matrixScaled;
    scaleXTalkMatrix(dataByBaseFiltered, numDataByBaseFiltered,
                     matrixUnscaled, matrixScaled);

    // normalise matrix to unit determinant
    normaliseXTalkMatrix(matrixScaled, matrixNormalised);

    for (int i = 0; i < numChannels; i++)
        delete [] dataByBaseFiltered[i];

#if defined(DEBUG_XTALK_MATRIX_2)
    std::cout << "\n>> debug - unscaled cross-talk matrix, last estimate" << std::endl;
    matrixUnscaled.display();
    std::cout << "\n>> debug - scaled cross-talk matrix" << std::endl;
    matrixScaled.display();
    std::cout << "\n>> debug - normalised cross-talk matrix" << std::endl;
    matrixNormalised.display();
#endif // DEBUG_XTALK_MATRIX_2
}
