/*
 * various disk IO and conversion routines for data analysis
 *
 * Time-stamp: <26 Jun 2006 17:00 BST IOProcess.cpp km>
 *
 * Copyright (c) Solexa 2003-2007
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#else
#include <cstdio>
#include <cstdlib>
#include <cstring>
#endif
#include <vector>
#include <string>
#include <sys/stat.h>
#if defined(HAVE_ZLIB)
#include <zlib.h>
#endif
#if defined(HAVE_BZLIB)
#include <bzlib.h>
#endif
#include <algorithm>
#include "IOProcess.h"
#include "Message.h"

int read_file(const char *filename,
              char      **buffer,
              long       *buffersize)
{
    FILE        *fp;
    int          errorno     = 0;
    struct stat  stbuf;
    size_t       nread    = 0;
    std::string  errormsg = "";
    if (stat(filename, &stbuf)) {
        errormsg = "Failed to stat " + std::string(filename) + ".";
        errorno  = 1; goto terminate;
    }
    fp = fopen(filename, "r");
    if (!fp) {
        errormsg = "Failed to open " + std::string(filename) + ".";
        errorno  = 1; goto terminate;
    }
    *buffersize = stbuf.st_size;
    *buffer     = new char [*buffersize + 1];
    nread = fread(*buffer, sizeof(char), (size_t)(*buffersize), fp);
    *(*buffer + *buffersize) = '\0';

    if (nread < (size_t)(*buffersize)) {
        delete [] *buffer; fclose(fp);
        errormsg = "Failed to read " + std::string(filename) + ".";
        errorno  = 2; goto terminate;
    }
    if (fclose(fp)) {
        delete [] *buffer;
        errormsg = "Failed to close " + std::string(filename) + ".";
        errorno  = 3;
    }
        
terminate:
    if (errorno != 0)
        error(errormsg.c_str());
    return errorno;
}

int read_gzip_file(const char   filename[],
                   char       **buffer,
                   long        *buffersize)
{
#if defined(HAVE_ZLIB)
    int                    errorno   = 0;
    long                   blocksize = 1000000;
    char                 * blockbuffer;
    std::vector<char * >   blocks;
    int                    nread     = 0;
    int                    i;
    std::string            errormsg  = "";

    gzFile fp = gzopen(filename, "rb");
    if (!fp) {
        errormsg = "Failed to open " + std::string(filename) + ".";
        errorno  = 1; goto terminate;
    }
    *buffersize = 0;
    nread = 0;
    do {
        blockbuffer = new char [blocksize];
        nread = gzread ( fp, blockbuffer, blocksize );
        if ( nread > 0) {
            *buffersize += nread;
            blocks.push_back(blockbuffer);
        }
        else {
            delete [] blockbuffer;
        }
        if (nread < 0) { // read error
            for (i = 0; i < int(blocks.size()); i++)
                delete [] blocks[i];
            blocks.clear();
        }
    } while (nread > 0);

    if ( nread < 0 ) {
        errormsg = "Failed to read " + std::string(filename) + ".";
        errorno  = 2; goto terminate;
    } 
    if (gzclose(fp)) {
        errormsg = "Failed to close " + std::string(filename) + ".";
        errorno  = 3;
    }
terminate:
    if ((! errorno) && (*buffersize > 0)) {
        *buffer = new char [*buffersize+1];
        for (i = 0; i < int(blocks.size()); i++) {
            memcpy(*buffer + i*blocksize, blocks[i], 
                   std::min(blocksize, *buffersize - i*blocksize));
            delete [] blocks[i];
        }
    }
    if (errorno != 0)
        error(errormsg.c_str());
    return errorno;
#else
    return 4;
#endif
}
    
int read_bzip2_file(const char   filename[],
                    char       **buffer,
                    long        *buffersize)
{
#if defined (HAVE_BZLIB)
    FILE                 * fp;
    int                    errorno   = 0;
    size_t                 nread     = 0;
    BZFILE               * b;
    const long             blocksize = 1000000;
    char                 * blockbuffer;
    std::vector<char * >   blocks;
    int                    bzerror;
    int                    i;
    std::string            errormsg  = "";

    fp = fopen(filename, "r");
    if (!fp) {
        errormsg = "Failed to open " + std::string(filename) + ".";
        errorno  = 1; goto terminate;
    }
    
    b = BZ2_bzReadOpen ( &bzerror, fp, 0, 0, NULL, 0 );
    if ( bzerror != BZ_OK ) {
        BZ2_bzReadClose ( &bzerror, b );
        errormsg = "Failed to open " + std::string(filename) + "for bzip2.";
        errorno  = 1; goto terminate;
    }

    bzerror = BZ_OK;
    *buffersize = 0;
    while ( bzerror == BZ_OK ) {
        blockbuffer = new char [blocksize];
        nread       = BZ2_bzRead ( &bzerror, b, blockbuffer, blocksize );
        if ( (bzerror == BZ_OK) || (bzerror == BZ_STREAM_END) ) {
            *buffersize += nread;
            blocks.push_back(blockbuffer);
        }
        else {
            delete [] blockbuffer;
            for (i = 0; i < int(blocks.size()); i++)
                delete [] blocks[i];
            blocks.clear();
        }
    }

    if ( bzerror != BZ_STREAM_END ) {
        BZ2_bzReadClose ( &bzerror, b );
        errormsg = "Failed to read " + std::string(filename) + ".";
        errorno  = 2; goto terminate;
    } else {
        BZ2_bzReadClose ( &bzerror, b );
    }
terminate:
    if ((! errorno) && (*buffersize > 0)) {
        *buffer = new char [*buffersize+1];
        for (i = 0; i < int(blocks.size()); i++) {
            memcpy(*buffer + i*blocksize, blocks[i], 
                   std::min(blocksize, *buffersize - i*blocksize));
            delete [] blocks[i];
        }
    }
    if (errorno != 0)
        error(errormsg.c_str());
    return errorno;
#else
    return 4;
#endif
}

int write_file(const char   filename[],
               const char * buffer,
               const long   buffersize)
{
    int          errorno     = 0;
    int          charwritten = 0;
    std::string  errormsg    = "";
    FILE        *fp          = fopen(filename, "w");
    if (!fp) {
        errormsg = "Failed to open " + std::string(filename) + " for writing.";
        errorno = 1; goto terminate;
    }
    charwritten = fwrite(buffer, sizeof(char), buffersize, fp);
    if (charwritten < buffersize) {
        errormsg = "Failed to write " + std::string(filename) + ".";
        errorno = 2;
    }
    if (fclose(fp)) {
        errormsg = "Failed to close " + std::string(filename) + ".";
        errorno = 3;
    }
terminate:
    if (errorno != 0)
        error(errormsg.c_str());
    return errorno;
}

int append_file(const char   filename[],
               const char * buffer,
               const long   buffersize)
{
    int          errorno     = 0;
    int          charwritten = 0;
    std::string  errormsg    = "";
    FILE        *fp          = fopen(filename, "w+");
    if (!fp) {
        errormsg = "Failed to open " + std::string(filename) + " for writing.";
        errorno = 1; goto terminate;
    }
    charwritten = fwrite(buffer, sizeof(char), buffersize, fp);
    if (charwritten < buffersize) {
        errormsg = "Failed to write " + std::string(filename) + ".";
        errorno = 2;
    }
    if (fclose(fp)) {
        errormsg = "Failed to close " + std::string(filename) + ".";
        errorno = 3;
    }
terminate:
    if (errorno != 0)
        error(errormsg.c_str());
    return errorno;
}

int write_gzip_file(const char   filename[],
                    const char * buffer,
                    const long   buffersize)
{
#if defined(HAVE_BZLIB)
    int          errorno     = 0;
    int          charwritten = 0;
    std::string  errormsg    = "";
    gzFile       fp          = gzopen(filename, "wb");
    if (!fp) {
        errormsg = "Failed to open " + std::string(filename) + " for writing.";
        errorno  = 1; goto terminate;
    }
    charwritten = gzwrite(fp, (char *)buffer, buffersize);
    if (charwritten < buffersize) {
        errormsg = "Failed to write to " + std::string(filename) + ".";
        errorno  = 2;
    }
    if (gzclose(fp)) {
        errormsg = "Failed to close " + std::string(filename) + ".";
        errorno  = 3;
    }
terminate:
    if (errorno != 0)
        error(errormsg.c_str());
    return errorno;
#else
    return 4;
#endif
}

int write_bzip2_file(const char   filename[],
                     const char * buffer,
                     const long   buffersize)
{
#if defined(HAVE_BZLIB)
    BZFILE      *b;
    int          bzerror;
    int          errorno   = 0;
    std::string  errormsg  = "";
    FILE        *fp        = fopen(filename, "w");
    if (!fp) {
        errormsg = "Failed to open " + std::string(filename) + " for writing.";
        errorno  = 1; goto terminate;
    }

    b = BZ2_bzWriteOpen( &bzerror, fp, 9, 0, 30 );
    if (bzerror != BZ_OK) {
        BZ2_bzWriteClose ( &bzerror, b, 0, 0, 0 );
        errormsg = "Failed to prepare " + std::string(filename) + " for bzip2.";
        errorno  = 1; goto terminate;
    }

    BZ2_bzWrite ( &bzerror, b, (void *)buffer, buffersize );
    if (bzerror == BZ_IO_ERROR) { 
        errormsg = "Failed to write to " + std::string(filename) + ".";
        errorno  = 2; 
    }

    BZ2_bzWriteClose( &bzerror, b, 0, 0, 0 );
    if (bzerror == BZ_IO_ERROR) {
        errormsg = "Failed to close " + std::string(filename) + ".";
        errorno  = 3;
    }
terminate:
    if (errorno != 0)
        error(errormsg.c_str());
    return errorno;
#else
    return 4;
#endif
}


long fread_uchar_array(const char filename[], unsigned char *data,
                       long nbytes, long offset)
{
    FILE *fp = fopen(filename, "r");
    if (fp) {
        if (offset > 0)
            fseek(fp, offset, SEEK_SET);
        long size = fread(data, sizeof(unsigned char), nbytes, fp);
        fclose(fp);
        return size;
    }
    return 0;
}

long fwrite_uchar_array(const char filename[], char mode[],
                        unsigned char *data, long len)
{
    FILE *fp = fopen(filename, mode);
    if (fp) {
        long size = fwrite(data, sizeof(unsigned char), len, fp);
        fclose(fp);
        return size;
    }
    return 0;
}

long fread_ushort_array(const char filename[], unsigned short *data,
                        long nelements, long offset)
{
    FILE *fp = fopen(filename, "r");
    if (fp) {
        if (offset > 0)
            fseek(fp, offset * sizeof(unsigned short), SEEK_SET);
        long size = fread(data, sizeof(unsigned short), nelements, fp);
        fclose(fp);
        return size;
    }
    return 0;
}

long fwrite_ushort_array(const char filename[], char mode[],
                         unsigned short *data, long nelem)
{
    FILE *fp = fopen(filename, mode);
    if (fp) {
        long size = fwrite(data, sizeof(unsigned short), nelem, fp);
        fclose(fp);
        return size;
    }
    return 0;
}

long fread_float_array(const char filename[], float *data, long len,
                       long offset)
{
    FILE *fp = fopen(filename, "r");
    if (fp) {
        if (offset > 0)
            fseek(fp, offset * sizeof(float), SEEK_SET);
        long size = fread(data, sizeof(float), len, fp);
        fclose(fp);
        return size;
    }
    return 0;
}

long fwrite_float_array(const char filename[], char mode[], float *data,
                        long len)
{
    FILE *fp = fopen(filename, mode);
    if (fp) {
        long size = fwrite(data, sizeof(float), len, fp);
        fclose(fp);
        return size;
    }
    return 0;
}

/*
 * return the filesize of a given file
 */
long file_size(char filename[])
{
    FILE *fp = fopen(filename, "r");
    if (fp) {
        fclose(fp);
        struct stat stbuf;
        stat(filename, &stbuf);
        return (long)stbuf.st_size;
    }
    return -1;
}

#if ! defined(_WIN32)

#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>

/*
 * create a new TCP/IP socket and connect to a server
 * at the given address and port
 */
int open_tcp_socket(const char address[],
                    const int  port)
{
    int                 sock;
    struct hostent    * hostinfo;
    struct sockaddr_in  serv_addr;

    // create socket
    sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (sock < 0) {
        perror("socket: Failed to create socket");
        exit(1);
    }

    // get host name
    hostinfo = gethostbyname(address);
    if (hostinfo == NULL) {
        fprintf(stderr, "Could not find host %s\n", address);
        exit(1);
    }

    // connect to server
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr = *((struct in_addr *)hostinfo->h_addr);
    serv_addr.sin_port = htons(port);
    if (connect(sock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
        perror("connect: Failed to connect with server");
        exit(1);
    }
    return sock;
}

/*
 * send a data array through a socket
 */
void send_uchar_file(const int           sock,
                     const char          filename[],
                     const unsigned char data[],
                     const long          datalen)
{
    char length[11];

    snprintf(length, 11, "%10d", (int)(strlen(filename)));
    send(sock, length, strlen(length), 0);
    send(sock, filename, strlen(filename), 0);
    snprintf(length, 11, "%10ld", datalen);
    send(sock, length, strlen(length), 0);
    send(sock, data, datalen, 0);
}

#endif
