/*
 * Copyright (c) Solexa   2006
 *               Illumina 2007
 * Author:
 *               Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
#ifndef INTENSITYDATA_H_
#define INTENSITYDATA_H_

#include <vector>
#include <map>
#include <string>
#include <stdexcept>
#include "ParameterTypes.h"
#include "PositionIndex.h"



class IntensityData 
{
public:
    typedef enum {
        compressionNone,
        compressionGzip,
        compressionBzip2
    } CompressionType;
private:
    std::string     m_Filename;
    CompressionType m_Compression;
    int             m_SkippedColumns;
    const bool      m_Ipar;
public:
    IntensityData(const char filename[],
                  CompressionType compression = compressionNone,
                  const bool ipar = false);
    virtual ~IntensityData();
    void setCompressionType(CompressionType compression = compressionNone);
    int  write(const float * intensities,
               const int     numClusters,
               const int     numChannels);
    int  write(const std::vector<float * > & signal,
               const int                     numClusters, 
               const int                     numChannels,
               const int                     lane = 0, 
               const int                     tile = 0, 
               const Position              * positions = 0);
    // Note: signal structure is (Cluster, Cycle, Channel)
    int read(float     * signal,
             int       * numClusters,
             int       * numChannels,
             const int   skipColumns = 0);
    // Note: signal structure is (Cycle, Cluster, Channel)
    int read(std::vector<float *> & signal,
             int                  * numClusters,
             int                  * numChannels,
             PositionIndex       ** indexPositions = 0,
             const unsigned int     lastCycle = 0,
             const unsigned int     firstCycle = 1);
};


int write_signal_array(const char                    filename[],
                       const std::vector<float * > & signal,
                       const int                     numClusters, 
                       const int                     numChannels,
                       const IntensityData::CompressionType
                                                     compression = IntensityData::compressionNone,
                       const int                     lane = 0, 
                       const int                     tile = 0, 
                       const Position              * positions = 0);
int write_signal_array(const char    filename[],
                       const float * intensity,
                       const int     numClusters,
                       const int     numChannels,
                       const IntensityData::CompressionType
                                     compression = IntensityData::compressionNone);

int read_signal_array(const char    filename[],
                      float       * signal,
                      int         * numClusters,
                      int         * numChannels,
                      const IntensityData::CompressionType
                                   compression = IntensityData::compressionNone,
                      const int    skipcols = 0);
/**
 ** Parameters:
 ** - lastCycle: the last cycle to include in the resulting data set. 
 **   If this parameter is set to 0, than all cycles are included (default=0).
 ** - firstCycle: first cycle to include in the the resulting data. Note,
 **   cycles are numbered from 1 (default=1).
 **/
std::vector<float * > read_signal_array(const char     *filename,
                                        int            *numClusters,
                                        int            *numChannels,
                                        int            *error,
                                        const IntensityData::CompressionType 
                                                    compression = IntensityData::compressionNone,
                                        PositionIndex **indexPositions = 0,
                                        const unsigned int lastCycle = 0,
                                        const unsigned int firstCycle = 1);



class UnsupportedCompressionException : public std::logic_error
{
public:
    UnsupportedCompressionException(const std::string &message) : std::logic_error(message) {}
};

/**
 ** The compression type should be completely removed. 
 ** This class is only provided as a convenience for refactoring the existing code
 **/
class Compression
{
public:
    Compression(const std::string &compressionName) throw (UnsupportedCompressionException);
    Compression(const Compression &compression) : compressionType_(compression.compressionType_) {}
    Compression &operator=(const Compression &compression) 
    {
        if (this != &compression)
        {
            compressionType_ = compression.compressionType_;
        }
        return *this;
    }
    inline operator IntensityData::CompressionType() const {return compressionType_;}
private:
    IntensityData::CompressionType compressionType_;
    static IntensityData::CompressionType getCompressionType(const std::string &compressionName) throw (UnsupportedCompressionException);
    static std::map<std::string, IntensityData::CompressionType> getCompressionTypeMap();
};

/**
 ** automation of the clean-up currently required after IntensityData::read
 **/
class CycleData : public std::vector<float *>
{
public:
    ~CycleData()
    {
        for (iterator i = begin(); end() != i; ++i)
        {
            delete [] *i;
            *i = 0;
        }
    }
};

#endif /*INTENSITYDATA_H_*/
