/*
 * Linear algebra routines: Cholesky decomposition, LU decomposition
 * and solution of the corresponding linear systems, matrix inversion
 *
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#include <stdio.h>
#else
#include <cmath>
#include <cstdio>
#endif
#include <cstring>
#include <cassert>
#include <iostream>
#include "LinAlg.h"

/* Prevent a SIGFPE overflow exception.
 * This should probably be a machine-/compiler specific constant.
 * This value is chosen arbitrarily.
 */
#define OVERFLOW_DIVISOR 1.e-20

void print_matrix(float *a, int nx, int ny)
{
    int i, j;
    for (i = 0; i < nx; i++) {
        if (i == 0)
            printf("[ [");
        else
            printf("\n  [");
        for (j = 0; j < ny; j++)
            printf("%f ", a[i*ny+j]);
        printf("]");
    }
    printf(" ]\n");
}

void print_imatrix(int *a, int n1, int n2)
{
    int i, j;
    for (i = 0; i < n1; i++) {
        for (j = 0; j < n2; j++)
            printf("%i ", a[i*n2+j]);
        printf("\n");
    }
}
    
/*
 * Invert input n*n matrix input and return the inverse as output.
 * Instead of solving A X = I directly (i.e. X=A^{-1}),
 * first perform an LU decomposition of A and then solve L U X = I using
 * two triangular systems.
 */
void matinvert(const float * input,
               float       * output,
               const int     n)
{
    int i, j;
    for (i = 0; i < n * n; i++)
        output[i] = input[i];

    /* divide by mean eigenvalue of the matrix to stabilise the inversion */
    float norm = 0.;
    for (i = 0; i < n; i++)
        norm += output[i * n + i];
    norm /= (float)n;
    for (i = 0; i < n * n; i++)
        output[i] /= norm;

    /* perform LU decomposition P A = L U to invert matrix */
    int *index = new int[n];
    lu_decompose_gauss_jordan_partial(output, n, index);

    /* create identity matrix I and solve L U X = P I */
    float *identity = new float[n * n];
    for (i = 0; i < n * n; i++)
        identity[i] = 0.;
    for (i = 0; i < n; i++)
        identity[i * n + i] = 1.;

    float *x = new float[n * n];
    lu_solve(output, index, n, identity, x, n);

    /* rescale matrix */
    for (j = 0; j < n; j++)
        for (i = 0; i < n; i++)
            output[i * n + j] = x[i * n + j] / norm;

    delete [] x;
    delete [] identity;
    delete [] index;

    return;
}

/*
 * Compute the Cholesky decomposition of the <n>*<n> symmetric,
 * positive definite matrix <a>. The matrix <a> is overwritten
 * by the lower triangular matrix L such that A = L L^T.
 * Returns 0 on success, 1 if the matrix is not positive definite.
 * 
 * The implemented version of the algorithm is the gaxpy Cholesky.
 * The computational cost is on the order of N^3/3 flops.
 */
 int cholesky_decompose(float *a, int n)
 {
    int j, i, k;
    float sq, sum;
    for (j = 0; j < n; j++) {
        if (j > 0) {
            for (i = j; i < n; i++) {
                sum = 0.;
                for (k = 0; k < j; k++)
                    sum += a[i*n+k] * a[j*n+k];
                a[i*n+j] -= sum;
            }
        }
        // check if decomposition failed
        if (a[j*n+j] <= 0.) {
            std::cerr << "matrix is not positive definite"; // throw exception?
            return 1;
        }
        sq = (float)sqrt(a[j*n+j]);
        for (i = j; i < n; i++)
            a[i*n+j] /= sq;
    }
    // reset upper triangle to zero
    for (i = 0; i < n-1; i++)
        for (j = i+1; j < n; j++)
            a[i*n+j] = 0.;
    return 0;
 }
 
 /*
 * LU decompose matrix <n>*<n> matrix <a> in place, using Gaussian elimination
 * with partial pivoting,
 * i.e. find lower and upper triangular matrices L and U such that
 * P A = L U, where A is the original matrix and P is a permutation
 * transformation.
 * On return, the upper and lower triangle of <a> store L and U respectively.
 * The diagonal elements belong to matrix U (L is scaled such that 
 * its diagonal elements are unity). index keeps track of row permutations P
 * performed to obtain the triangular shapes by keeping the index of the
 * original rows.
 * The parameter <d> is output as +1 or -1 depending whether the number of
 * row interchanges was even or odd, respectively.
 * 
 * Compare: Golub/van Loan, Matrix Computations, 3rd ed., 
 *             sec. 3.4.3 "Partial Pivoting Details", p. 112
 *          Numerical Recipes, routine ludcmp
 */ 
int lu_decompose_gauss_jordan_partial(float *a, int n, int *index, float *d)
{
    int i, j, k, l;
    float *scaling = new float [n];

    *d = 1.0f;

    // determine largest element in each row to give better estimate
    // for pivoting
    for (i = 0; i < n; i++)
        index[i] = i;
    for (i = 0; i < n; i++) {
        float max = (float)fabs(a[i*n+0]);
        for (j = 1; j < n; j++) {
            float abs_value = (float)fabs(a[i*n+j]);
            if (abs_value > max)
                max = abs_value;
        }
        if (fabs(max) < OVERFLOW_DIVISOR)
            goto singular_matrix;
        scaling[i] = 1. / max;
    }

    for (j = 0; j < n-1; j++) {
        // partial pivoting:
        // find largest element A_{ji} for i>=j, assuming scaling had occured
        float max = scaling[j] * (float)fabs(a[j*n+j]);
        int imax = j;
        for (i = j+1; i < n; i++)
            if (scaling[i] * (float)fabs(a[i*n+j]) > max) {
                max = scaling[i] * (float)fabs(a[i*n+j]);
                imax = i;
            }

        // if required, perform row permutation
        if (j != imax) {
            for (k = 0; k < n; k++) {
                float tmp = a[imax*n+k];
                a[imax*n+k] = a[j*n+k];
                a[j*n+k] = tmp;
            }
	    *d = -(*d);
            scaling[imax] = scaling[j]; // scaling[j] does not need to be updated
            int itmp = index[imax];
            index[imax] = index[j];
            index[j] = itmp;
        }

        if (fabs(a[j*n+j]) > OVERFLOW_DIVISOR) {
            // divide by the pivot element
            for (i = j+1; i < n; i++)
                a[i*n+j] /= a[j*n+j];
            // outer product Gaussian elimination
            for (i = j+1; i < n; i++)
                for (l = j+1; l < n; l++) 
                    a[i*n+l] -= a[i*n+j] * a[j*n+l];
        }
    }

    delete [] scaling;
    return 0;
singular_matrix:
    delete [] scaling;
#if 0
    std::cerr << "matrix is singular\n"; // throw exception?
#endif
    return 1;
}

int lu_decompose_gauss_jordan_partial(float *a, int n, int *index)
{
    float d;
    return lu_decompose_gauss_jordan_partial(a, n, index, &d);
}

/*
 * solve linear system A X = B, where A is a LU-decomposed n*n matrix, 
 * B and X are n*n2-matrices, and stores are the row permutations P applied
 * during the LU decomposition (i.e. P A = L U).
 */
int lu_solve(float *a, int *index, int n, float *b, float *x, int n2)
{
    int i, j, k;
    for (k = 0; k < n2; k++) {
        // compute x = P b
        for (i = 0; i < n; i++)
            x[i*n2+k] = b[index[i]*n2+k];

        // find first non-zero element in permuted vector P b to avoid
        // unnecessary summation over zero elements
        int i2 = 0;
        while ((i2+1 < n) && (x[i2*n2+k] == 0.))
            i2++;

        // forward subsitution (row version), solve L y = P b (and overwrite P b)
        // (the diagonal elements of L equal unity, so no division)
        for (i = i2+1; i < n; i++) {
            float sum = 0.;
            for (j = i2; j < i; j++)
                sum += a[i*n+j] * x[j*n2+k];
            x[i*n2+k] -= sum;
        }

        // back-substitution (row version), solve U x = y (and overwrite y)
        for (i = n-1; i >= 0; i--) {
            float sum = 0.;
            for (j = i+1; j < n; j++)
                sum += a[i*n+j] * x[j*n2+k];
            x[i*n2+k] -= sum;
            if (fabs(a[i*n+i]) < OVERFLOW_DIVISOR)
                return 1;
            x[i*n2+k] /= a[i*n+i];
        }
    }
    return 0;
}

#if 0
/*
 * DO NOT USE THIS ROUTINE! It has not been tested properly!
 * Principles taken from Golub/van Loan, Matrix computations, 3rd edition.
 */
static int lu_decompose_gauss_jordan_pivoting(float *a, int *p, int *q, int n)
{
    int i, j, k, m, l, itmp;
    float max, tmp;
    for (i = 0; i < n; i++) {
        p[i] = q[i] = i;
    }

    for (k = 0; k < n-1; k++) {
        // find largest matrix element
        m = l = k;
        max = (float)fabs(a[m*n+l]);
        for (i = k; i < n; i++)
            for (j = k; j < n; j++)
                if (fabs(a[i*n+j]) > max) {
                    max = (float)fabs(a[i*n+j]);
                    m = i;
                    l = j;
                }
        // interchange k-th and m-th rows
        if (k != m)
            for (j = 0; j < n; j++) {
                tmp = a[m*n+j];
                a[m*n+j] = a[k*n+j];
                a[k*n+j] = tmp;
            }
        itmp = p[k];
        p[k] = p[m];
        p[m] = itmp;
        // interchange k-th and l-th column
        if (k != l)
            for (i = 0; i < n; i++) {
                tmp = a[i*n+l];
                a[i*n+l] = a[i*n+k];
                a[i*n+k] = tmp;
            }
        itmp = q[k];
        q[k] = q[l];
        q[l] = itmp;
    
        if (a[k*n+k] != 0.) {
            for (i = k+1; i < n; i++)
                a[i*n+k] /= a[k*n+k];
            for (i = k+1; i < n; i++)
                for (j = k+1; j < n; j++)
                    a[i*n+j] -= a[i*n+k] * a[k*n+j];
        }
    }
    return 0;
}
#endif

/*
 * solves the linear system AX=B, where A is n*n, B is n*n2, using
 * LU decomposition.
 * On return, A is replaced by its inverse and B is overwritten 
 * to contain the solution matrix X.
 */ 
int gauss_jordan_inverse_solve(float *a, int n, float *b, int n2)
{
    int i;
    int *p = new int [n];
    for (i = 0; i < n; i++)
        p[i] = 0;
    int flag = lu_decompose_gauss_jordan_partial(a, n, p);
  
    if (!flag) {

        int j;
        // solve for b
        float *borig = new float  [n*n2];
        for (i = 0; i < n; i++)
            for (j = 0; j < n2; j++)
                borig[i*n2+j] = b[i*n2+j];
    
        flag = lu_solve(a, p, n, borig, b, n2);
        delete [] borig;

        if (!flag) {
            // replace matrix with inverse matrix
            float *identity = new float [n*n];
            float *aorig = new float [n*n];
            for (i = 0; i < n; i++) {
                for (j = 0; j < n; j++)
                    identity[i*n+j] = 0.;
                identity[i*n+i] = 1.;
            }
            for (i = 0; i < n; i++)
                for (j = 0; j < n; j++)
                    aorig[i*n+j] = a[i*n+j];
            flag = lu_solve(aorig, p, n, identity, a);
            delete [] identity;
            delete [] aorig;
        }
    }
    delete [] p;
    return flag;
}

/**
 * Applies a matrix vector multiplication to each row in input and output.
 * So O_{ij} = \sum_k A_{jk} V_{ik}.
 */
void matrix_vectortranspose_multiply(float       * output,
                                     const float * matrix,
                                     const float * vectortranspose,
                                     const int     numCols,
                                     const int     numRows)
{
    assert(numRows >= 0);
    assert(numCols > 0);
    if (matrix != 0) {
        int i, j, k;
        for (i = 0; i < numRows; i++)
            for (j = 0; j < numCols; j++) {
                output[i*numCols+j] = 0.;
                for (k = 0; k < numCols; k++)
                    output[i*numCols+j] += matrix[j*numCols+k]
                            * vectortranspose[i*numCols+k];
            }
    } 
    else
        memcpy(output, vectortranspose, numRows * numCols * sizeof(float));
}

/**
 * Matrix multiplication O_{ij} = \sum_k A_{ik} I_{kj}.
 */
void matrix_multiply(float       * output,
                     const float * matrix,
                     const float * input,
                     const int     numRows,
                     const int     numCols)
{
    assert(numRows >= 0);
    assert(numCols > 0);
    if (matrix != 0) {
        int i, j, k;
        for (i = 0; i < numRows; i++)
            for (j = 0; j < numCols; j++) {
                output[i*numCols+j] = 0.;
                for (k = 0; k < numCols; k++)
                    output[i*numCols+j] += matrix[i*numCols+k]
                            * input[k*numCols+j];
            }
    } 
    else
        memcpy(output, input, numRows * numCols * sizeof(float));
}

#if 0
/* test case for Cholesky decomposition */
int main(int argc, char *argv[])
 {
    int n = 4;
    int i;
    float *a = new float [n*n];
    for (i = 0; i < n*n; i++)
        a[i] = 0.;
    for (i = 0; i < n; i++)
        a[i*n+i] = (i+1)*(i+1);
    a[1*n+2] = a[2*n+1] = 1.;
        
    print_matrix(a, n, n);
    cholesky_decompose(a, n);
    print_matrix(a, n, n);
}
#endif
