/*
 * Copyright (c) 2004-2007 Solexa Ltd.
 *               2007      Illumina
 * Author:       Klaus Maisinger
 *
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
#if defined(__BORLANDC__)
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#else
#include <cstdlib>
#include <cstring>
#include <cassert>
#endif
#include "PrbData.h"
#include "Message.h"
#include "StringUtil.h"
#include "IOProcess.h"

int write_quality_file(const char filename[],
                       const std::vector<float * > & prob,
                       const int nobj,
                       const int ncol)
{
    int code = 0;
    int nprob = int(prob.size());
    FILE *probfile = fopen(filename, "wt");
    if (probfile) {
        int buffersize = nprob * ncol * 8;
        char *buffer = new char [buffersize];
        int n, i, k;
        for (n = 0; n < nobj; n++) {
            int idx = 0;
            for (i = 0; i < nprob; i++) {
                for (k = 0; k < ncol; k++) {
                    idx += snprintf(buffer+idx, 7, "%4.0f", prob[i][n*ncol+k]);
                    if (k != ncol-1)
                        idx += sprintf(buffer+idx, " ");
                }
                if (i != nprob-1)
                    idx += sprintf(buffer+idx, "\t");
            }
            sprintf(buffer+idx, "\n");
            if (fprintf(probfile, buffer) < (int)strlen(buffer)) {
                error("Failed writing to %s.", filename);
                code = 2;
            }
        }
        delete [] buffer;
        if (fclose(probfile)) {
            error("Failed to close %s.", filename);
            return 3;
        }
        return code;
    }
    else {
        error("Failed to open %s for writing.", filename);
        return 1;
    }
}

/***********************************************************************/
PrbData::PrbData(const char filename[])
{
    m_Filename = filename;
}

PrbData::~PrbData()
{
}

int PrbData::write(const std::vector<float * > & probabilities,
                   const int                     numClusters,
                   const int                     numChannels)
{
    return write_quality_file(m_Filename.c_str(),
                              probabilities, numClusters, numChannels);
}


int PrbData::read(std::vector<float *> & probabilities,
                   unsigned int        & numClusters, 
                   unsigned int        & numChannels) {
    char *prbBuffer= NULL;
    long buffersize;
    int status = read_file(m_Filename.c_str(), &prbBuffer, 
            &buffersize);
    if (status > 0) {
        exit(-1);    	
    }
    if (strlen(prbBuffer) > 0) {
	    std::string strin;
	    strin.append(prbBuffer, buffersize);
	    std::string convertedString;
	    parsePrb(m_Filename, strin, probabilities, numClusters,
	            numChannels);
    }
    delete[] prbBuffer;
    return status;
}

int PrbData::parsePrb(const std::string    & dataSourceName,
                      const std::string    & strin, 
                      std::vector<float *> & probabilities,
                      unsigned int         & numClusters, 
                      unsigned int         & numChannels) {
    unsigned int numCycles = 0;

    if (strin.length() == 0) { // Empty string
        error("parsePrb: Error empty string for parsePrb method.\n");
        exit(-1);
    }

    unsigned int clusterIndex = 0;
    vector<char*> inputLines;
    fastTokenizer(strin.c_str(), '\n', inputLines);

    vector<char*>::iterator iterLine;

    numClusters = inputLines.size();
    for (iterLine = inputLines.begin(); iterLine != inputLines.end(); iterLine++) {
        vector<char*> cycles;
        char *bufferBuffer = *iterLine;
        fastTokenizerInPlace(bufferBuffer, '\t', cycles);

        if (clusterIndex == 0) {
            numCycles = cycles.size();
        } else {
            assert(numCycles == cycles.size());
        }
        unsigned int cycleIndex;
        float * qualityValuesFloat= NULL;
        unsigned int qualityValuesFloatIndex;

        for (cycleIndex = 0; cycleIndex < numCycles; cycleIndex++) {
            char *buffer = cycles[cycleIndex];
            numChannels = 4;

            if (clusterIndex == 0) {
                qualityValuesFloat = new float[numChannels * numClusters];
                probabilities.push_back(qualityValuesFloat);
            } else {
                qualityValuesFloat = probabilities[cycleIndex];
            }
            qualityValuesFloatIndex = clusterIndex * numChannels;

            char *begin, *end;
            unsigned int channelIndex = 0;
            end = buffer;
            do {
                begin = end;
                qualityValuesFloat[qualityValuesFloatIndex] = (float)strtod(
                        begin, &end);
                qualityValuesFloatIndex++;
                channelIndex++;
            } while ((channelIndex < numChannels) && (end != NULL) && (end
                    != begin));

            if (channelIndex != 4) {
                error("parsePrb: Error parsing line: %s.", (*iterLine));
                error("parsePrb: can't find quality values for cycle %d cluster %d in file %s.",
                      cycleIndex, clusterIndex, dataSourceName.c_str());
                error("- could not parse field %s. Number of channels %d\n",
                      cycles[clusterIndex], numChannels);
                exit(-1);
            }
        }
        delete[] *iterLine;
        clusterIndex++;
    }
    return 0;
}

