/*
 * $RCSfile: Sorting.h,v $
 * Author: Klaus Maisinger
 * $Id: Sorting.h,v 1.1 2008/07/03 15:32:49 craczy Exp $
 *
 * Copyright (c) 2008, Illumina
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
#ifndef SORTING_H_
#define SORTING_H_

#if defined(__BORLANDC__)
#include <string.h>
#else
#include <cstring>
#endif
#include <vector>
#include <algorithm>

/**
 * Routines to map the STL sorting routines to arrays.
 */
namespace Sorting {
    /**
     * Sorts the array.
     */
    template<typename T>
    void sort (T            * array,
               const size_t   n);

    /**
     * Returns an index array showing the permutations needed to sort
     * the input array.
     */
    template<typename T>
    void sort_index(const T      * array,
                    size_t       * index,
                    const size_t   n);

    /**
     * Selects the <idx>th element from an array.
     */
    template<typename T>
    T select(const T      * array,
             const size_t   n,
             const size_t   idx);

    /**
     * Partitions an array such that the <idx>th element is in place, 
     * and all preceding elements are smaller and all subsequent elements
     * are larger.
     */
    template<typename T>
    void nth_element(T            * array,
                     const size_t   n,
                     const size_t   idx);


    /*****************************************************************/
    /*
     * Implementation details - g++ requires implementation in the header.
     */
    template<typename T>
    void sort (T            * array,
               const size_t   n)
    {
        std::vector<T> v(n);
        memcpy((void *)&v[0], (void *)array, n * sizeof(T));
        std::sort(v.begin(), v.end());
        memcpy(array, &v[0], n * sizeof(T));
    }

    template <typename T>
    class TmpNode
    {
    public:
        T      value; 
        size_t index;
    };

    template<typename T>
    bool operator< (const TmpNode<T> & lhs, const TmpNode<T> & rhs)
    {
        return lhs.value < rhs.value;
    }

    template<typename T>
    void sort_index(const T      * array,
                    size_t       * index,
                    const size_t   n)
    {

        std::vector<TmpNode<T> > v(n);
        size_t i;
        for (i = 0; i < n; i++) {
            v[i].value = array[i];
            v[i].index = i;
        }
        std::sort(v.begin(), v.end());
        for (i = 0; i < n; i++) {
            index[i] = v[i].index;
        }
    }

    template<typename T>
    void nth_element(T            * array,
                     const size_t   n,
                     const size_t   idx)
    {
        std::vector<T> v (n);
        memcpy(&v[0], array, n * sizeof(T));
        std::nth_element(v.begin(), v.begin()+idx, v.end());
        memcpy(array, &v[0], n * sizeof(T));    
    }

    template<typename T>
    T select(const T      * array,
             const size_t   n,
             const size_t   idx)
    {
        std::vector<T> v (n);
        memcpy(&v[0], array, n * sizeof(T));
        std::nth_element(v.begin(), v.begin()+idx, v.end());
        return (v[idx]);  
    }
}

#endif /*SORTING_H_*/
