/*
 * Statistical routines.
 *
 * Copyright (c) 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#else
#include <cmath>
#endif
#include <cassert>
#include "Maths.h"
#include "Stats.h"


int Stats::round(const float f)
{
    return (int)floor(f+0.5);
}

template <class T>
T Stats::max(const T *a, const int n)
{
    assert(n >= 1);
    T m = a[0];
    int i;
    for (i = 1; i < n; i++)
        if (a[i] > m)
            m = a[i];
    return m;
}

template <class T>
T Stats::min(const T *a, const int n)
{
    assert(n >= 1);
    int i;
    T m = a[0];
    for (i = 1; i < n; i++)
        if (a[i] < m)
            m = a[i];
    return m;
}

template<class T>
T Stats::average(const T *x, const int n)
{
    int i;
    T r = 0.;
    for (i = 0; i < n; i++)
        r += x[i];
    return (T)(r/n);
}

template <>
float Stats::average<float>(const float *x, const int n)
{
    int i;
    double r = 0.;
    for (i = 0; i < n; i++)
        r += x[i];
    return (float)(r/n);
}


/* 
 * Linear regression.
 * Returns gradient and intercept of the <ndim> data points given by the
 * <x> and <y> arrays.
 */
int Stats::linregress(float *x, float *y, int ndim,
                      float *gradient, float *intercept)
{
    assert(ndim > 1);
    *gradient = 0.f;
    *intercept = 0.f;
    if (fabs(max(x, ndim) - min(x, ndim)) < Maths::tiny)
        return 1;
    int i;
    double sum_x = 0.f, sum_y = 0.f, sum_xx = 0.f, sum_yx = 0.f;
    for (i = 0; i < ndim; i++) {
        sum_x += x[i];
        sum_xx += x[i] * x[i];
        sum_y += y[i];
        sum_yx += y[i] * x[i];
    }
    double denom = ((ndim * sum_xx) - (sum_x * sum_x));
    if (fabs(denom) < Maths::tiny)
        return 1;
    *gradient = (float)((ndim * sum_yx - sum_x * sum_y) / denom);
    *intercept = (float)((sum_y - (*gradient) * sum_x) / ndim);
    return 0;
}

