//  Time-stamp: <26 Jun 2006 16:57 BST Tiling.cpp km>
/*
 * Set the tiling structure and runfolder names for a data analysis run.
 * This object is used to generate suitable filenames for output files,
 * according to the run-folder specification. Internally, the image
 * analysis does not need to know about these conventions; tiles (positions
 * on the chip) are simply numbered sequentially.
 *
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <stdio.h>
#else
#include <cstdio>
#endif
#include "Tiling.h"

RunFolderFileName::RunFolderFileName(const char prefix[]) : nchannel(0), tileoffset(0)
{
    path = prefix;
    typestring[POSITION_FILE]  = "pos";
    typestring[INDEX_FILE]     = "idx",
    typestring[INTENSITY_FILE] = "int";
    typestring[NOISE_FILE]     = "nse";
    typestring[OFFSET_FILE]    = "off";
    typestring[CLUSTER_FILE]   = "clu";
    typestring[QC_FILE]        = "qcm";
    typestring[SIGNAL_FILE]    = "sig";
    typestring[SEQUENCE_FILE]  = "seq";
    typestring[PROBABILITY_FILE]= "prb";
    typestring[SIGNAL2_FILE]   = "sig2";
    typestring[SCORE_FILE]     = "score";
    typestring[RESCORE_FILE]   = "rescore";
}

RunFolderFileName::~RunFolderFileName()
{
    if (nchannel > 0) {
        delete [] channelidx; delete [] channelcount;
    }
}

/*
 * use <numLanes> lanes with indices given by the <numLanes>-dim 
 * array <laneIndex>. Each lane has <laneNumTiles> tiles.
 */
void RunFolderFileName::setLanes(const int *laneIndex,
                                 const int *laneNumTiles,
                                 const int  numLanes)
{
    nchannel     = numLanes;
    channelcount = new int [nchannel];
    channelidx   = new int [nchannel];
    int i;
    for (i = 0; i < nchannel; i++) {
        channelcount[i] = laneNumTiles[i];
        channelidx[i]   = laneIndex[i];
    }
}

/*
 * only use one lane with index <lane>. there are <numTiles> tiles,
 * counting starts from <tileIndex>
 */
void RunFolderFileName::setLane(const int lane,
                                const int tileIndex,
                                const int numTiles)
{
    nchannel = 1;
    channelcount = new int [nchannel];
    channelidx   = new int [nchannel];
    channelcount[0] = numTiles+tileIndex;
    channelidx[0]   = lane;
    tileoffset = tileIndex;
}

int RunFolderFileName::getLane(const int idx) 
{ 
    int i = 0, total = 0;
    while (i < nchannel) {
        total += channelcount[i];
        if (idx < total)
            return channelidx[i];
    }
    return 0; 
}

int RunFolderFileName::getTile(const int idx) { 
    int i = 0, total = 0;
    while (i < nchannel) {
        total += channelcount[i];
        if (idx < total)
            return idx-(total-channelcount[i]) + tileoffset;
    }
    return idx+tileoffset; 
}

int RunFolderFileName::getIndex(const int lane,
                                const int tile) {
    int i = 0, total = 0;
    while ((i < nchannel) && (channelidx[i] < lane))
        total += channelcount[i];
    return tile+total;
}

std::string RunFolderFileName::getTileFilename(const RunFolderFileName::filename_type type,
                                               const int lane, 
                                               const int tile,
                                               const int cycle,
                                               const int channel)
{
    char filename[FILENAME_MAX];
    if (type == QC_FILE) {
        snprintf(filename, FILENAME_MAX,
                 "%s_%01d_%04d_%02d_%s.xml", path.c_str(), lane + 1,
                 tile + 1, cycle + 1, typestring[type].c_str());
        return std::string(filename);
    }
    if (cycle < 0) {
        snprintf(filename, FILENAME_MAX,
                 "%s_%01d_%04d_%s.txt", path.c_str(), lane + 1,
                 tile + 1, typestring[type].c_str());
    }
    else {
        if (channel < 0)
            snprintf(filename, FILENAME_MAX,
                    "%s_%01d_%04d_%02d_%s.txt", path.c_str(),
                    lane + 1, tile + 1, 
                    cycle + 1, typestring[type].c_str());
        else
             snprintf(filename, FILENAME_MAX,
                    "%s_%01d_%04d_%02d_%1d_%s.txt", path.c_str(),
                    lane + 1, tile + 1, 
                    cycle + 1, channel + 1, typestring[type].c_str());
            
    }
    return std::string(filename);
}

std::string RunFolderFileName::getIndexFilename(const RunFolderFileName::filename_type type, 
                                                const int idx,
                                                const int cycle,
                                                const int channel)
{
    return getTileFilename(type, getLane(idx), getTile(idx), cycle, channel);
}

std::string RunFolderFileName::getFilePrefix()
{
    std::string s = path;
    return s;
}
