/* 
 * Time-stamp: <$Id: Use_Bases.cpp,v 1.2 2008/07/17 15:44:49 rshaw Exp $>
 * Functions to parse USE_BASES strings.
 *
 * Copyright (c) Illumina 2007
 * Author: Lukasz Szajkowski
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#include <stdlib.h> 
#include <stdio.h>
#include <cassert>
#include <algorithm>
#include <vector>
#include <string>
#include <cctype>
#include "Use_Bases.h"
#include "StringUtil.h"

using namespace std;

/*
 * The function expands bases the the useBases string.
 */
void expandLegacyUseBasesString(const std::string & useBases,
        const int unsigned readLength, std::string & useBasesOut) {
    unsigned int i;

    if (useBases.compare(CONST_ALL) == 0) {
        if (readLength <= 0) {
            fprintf(stderr, "USE_BASES '%s' specified but invalid read length %d given!",
            CONST_ALL.c_str(), readLength);
            exit(-1);
        }
        useBasesOut.append(readLength, 'Y');
    }
    // I don't know if the code below is still valid (lszajkowski 08.01.2008)
    else if (useBases.compare(CONST_ODD) == 0) {
        if (readLength <= 0) {
            fprintf(stderr, "USE_BASES '%s' specified but invalid read length %d given!",
            CONST_ODD.c_str(), readLength);
            exit(-1);
        }
        for (i = 0; i < readLength; i++) {
            useBasesOut.append(1, 'Y');
            useBasesOut.append(1, 'N');
        }
        useBasesOut.erase(useBasesOut.end());
    } else {
        int error = useBases.find_first_not_of(CONST_LEGAL_MAST);
        if (error >= 0) {
            fprintf(stderr, "USE_BASES string '%s' must be odd, all or YNYN...!",
            useBases.c_str());
            exit(-1);
        }
        useBasesOut.append(useBases);
    }
    return;
}
/*
 * The function parseUseBases parses a long form USE_BASES parameter 
 * into an array of indices
 * Inputs:
 * 1. useBases string - either:
 * 'all' - use all bases (probably an anachronism)
 * 'odd' - use odd numbered bases (definitely an anachronism)
 * Some combination of 'Y', 'y', 'N' and 'n'
 * 2. readLength - used in 'all' or 'odd' modes to generate a string
 * of the right length, also used to check there are enough 'Y/y's in
 * useBases - again probably an anachronism, set to 0 to ignore
 * 3. readPairMode (optional) Which read of a paired read - either
 * '0' - default 
 * '1' - first read: treat all 'y's as 'n's
 * '2' - second read: treat all 'Y's as 'N's
 */
void parseUseBases(const std::string & useBases, const int unsigned readLength,
        vector<unsigned int> & baseIndex, const int unsigned readPairMode) {
    std::string useBasesOut;
    
    unsigned int i;
    expandLegacyUseBasesString(useBases, readLength, useBasesOut);

    if (readPairMode == 1) {
        replaceTR(useBasesOut, "y", "N");
    } else if (readPairMode == 2) {
        replaceTR(useBasesOut, "Y", "N");
    }

    std::transform(useBasesOut.begin(), useBasesOut.end(), useBasesOut.begin(),
            (int(*)(int)) std::toupper);

    unsigned int length = useBasesOut.size();
    for (i = 0; i < length; i++) {
        if (useBasesOut[i] == 'Y') {
            baseIndex.push_back(i);
        }
    }

    unsigned int basedIndexSize = baseIndex.size();
    // Only check read length with use bases if positive read length given
    if (readLength > 0) {
        if (basedIndexSize < readLength) {
            fprintf(stderr, "Need %d bases of base information, only %d specified in %s",
            (int)useBasesOut.size(), basedIndexSize, useBasesOut.c_str());
        } else if (basedIndexSize > readLength) {
            for (i = basedIndexSize; i > readLength; i--) {
                baseIndex.pop_back();
            }
        }
    }
    return;
}

