/*
 * Implement a primitive XML parser that loosely follows the SAX API.
 * No validation is done; this is just a minimal implementation for the 
 * specific purposes of the analysis software.
 *
 * Copyright (c) Solexa   2005, 2006
 *               Illumina 2007
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
#include "XMLSAXParser.h"
#include <cstring>
#include <vector>
#include <string>
#include <cassert>

XMLSAXParser::XMLSAXParser()
{
}

XMLSAXParser::~XMLSAXParser()
{
}

void XMLSAXParser::setContentHandler(XMLContentHandler *hd) 
{
    handlers.push_back(hd);
}

std::vector<std::string> split(const std::string &s, const char delimiter = ' ')
{
    std::vector<std::string > result;
    int length = s.length();
    int i = 0;
    int first = 0;
    while (i < length) {
        while ((i < length) && (s[i] != delimiter))
            i++;
        if (i > first) {
            std::string part = s.substr(first, i-first);
            result.push_back(part);
        }
        first = i+1;
        i++;
    }
    return result;
}

void parseElement(const char content[], std::string &name, 
                  Attributes *attrs)
{
    std::string content2(content);
    std::vector<std::string> parts = split(content2, ' ');
    assert(parts.size() > 0);
    name = parts[0];
    int i;
    for (i = 1; i < int(parts.size()); i++) {
        std::vector<std::string> parts2 = split(parts[i], '=');
        assert(parts2.size() > 1);
        if (parts2[1][0] == '"')
            parts2[1].erase(0, 1);
        if (parts2[1][parts2[1].size()-1] == '"')
            parts2[1].erase(parts2[1].size()-1, 1);
        attrs->addAttribute(parts2[0], parts2[1]);
    }
}

void XMLSAXParser::parse(FILE *fh, XMLContentHandler *hd)
{
    if (hd != 0)
        handlers.push_back(hd);
    if (handlers.size() == 0)
        return;

    /* start document */
    std::list<XMLContentHandler *>::iterator handler;
    for (handler = handlers.begin(); handler != handlers.end(); ++handler)
        (*handler)->startDocument();
    char buffer [1000];
    char text[1000];
    char tag[1000];
    /* loop over lines in file */
    while (fgets(buffer, sizeof(buffer), fh)) {
        int tagopen = 0;
        int comment = 0;
        int starttag = 0;
        int endtag   = 0;
        int tagfirstpos = 0;
        int taglastpos  = 0;
        int textpos  = -1;
        int i, k;
        for (i = 0; i < (int)strlen(buffer); i++) {
            if (buffer[i] == '<') { /* tag detected */
                tagopen = 1;
                starttag = 1;
                tagfirstpos = i;
            }
            if ((buffer[i] == '/') && tagopen) {
                endtag = 1;  /* end tag detected */
                if (tagfirstpos == i-1) {
                    starttag = 0;
                    tagfirstpos = i;
                }
            }
            if ((buffer[i] == '?') && tagopen) {
                if (tagfirstpos == i-1)
                    comment = 1;
            }
            if (buffer[i] == '>') { /* tag closes */
                taglastpos = i;
                if (buffer[i-1] == '/') taglastpos = i-1;
                if (tagopen && (! comment)) {
                    memcpy(tag, &buffer[tagfirstpos+1], 
                           sizeof(char)*(taglastpos-tagfirstpos-1));
                    tag[taglastpos-tagfirstpos-1] = '\0';
                    std::string name;
                    Attributes *attrs = new Attributes();
                    parseElement(tag, name, attrs);
                    if (starttag) { /* notify handler of startElement */
                        for (handler = handlers.begin();
                             handler != handlers.end(); ++handler)
                            (*handler)->startElement(name.c_str(), attrs);
                        textpos=i+1;
                    }
                    else { /* notify handler of endElement */
                        if (textpos >= 0) { /* if there was text between tags,
                                                notify handler */
                            for (k = 0; k < tagfirstpos -textpos-1; k++)
                                text[k] = buffer[k+textpos];
                            text[tagfirstpos-textpos-1] = '\0';
                            for (handler = handlers.begin();
                                 handler != handlers.end(); ++handler)
                                (*handler)->characters(text, 0, i-textpos-3);
                            textpos = -1;
                        }
                    }
                    if (endtag) {
                        for (handler = handlers.begin();
                            handler != handlers.end(); ++handler)
                            (*handler)->endElement(name.c_str());
                    }
                    delete attrs;
                } 
                tagopen = 0;
                comment = 0;
            }
        }
    }
    /* end document */
    for (handler = handlers.begin(); handler != handlers.end(); ++handler)
        (*handler)->endDocument();
}
