#include <iostream>
#include <sstream>

using namespace std;

#include "RegistryName.hh"
#include "testGzPack.hh"

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( TestGzPack, registryName("GzPack"));

using namespace IoUtils;

void TestGzPack::createFile()
{
    gzFile fh = gzopen(fileName.c_str(), "w9");
    ostringstream dataLayout;
    dataLayout << "#CH" << numberOfChannels << ":OBJ" << numberOfClusters << "\n";
    gzputs(fh, dataLayout.str().c_str());
    for (unsigned int cycle = 0; cycle != numberOfCycles; ++cycle)
    {
        for (unsigned int cluster = 0; cluster != numberOfClusters; ++cluster)
        {
            ostringstream os;
            for (unsigned int channel = 0; channel != numberOfChannels; ++channel)
            {
                float value = 100.0*cycle + 10.0*cluster + channel;
                os << value << (channel == numberOfChannels-1 ? '\n' : '\t');
            }
            gzputs(fh, os.str().c_str());
        }
        ostringstream endCycle;
        endCycle << "# END CYCLE " << cycle + 1 << "\n";
        gzputs(fh, endCycle.str().c_str());
    }
    gzclose(fh);
}

void TestGzPack::setUp()
{
    fileName = "testGzPack.gz"; 
    numberOfChannels = 3;
    numberOfClusters = 4;
    numberOfCycles   = 5;
    createFile();
    reader = 0;
    try
    {
        reader = new GzPack::GzPackReader(fileName);
    }
    catch (logic_error &e)
    {
        cout << endl << e.what() << endl;
    }
}

void TestGzPack::tearDown()
{
    delete reader;
}

void TestGzPack::testConstructor()
{
    //CPPUNIT_ASSERT_THROW(GzPack::GzPackReader(fileName + ".bad"), std::ios_base::failure);
    CPPUNIT_ASSERT(!reader->fileError());
    CPPUNIT_ASSERT_EQUAL(numberOfChannels, reader->getNumberOfChannels());
    CPPUNIT_ASSERT_EQUAL(numberOfClusters, reader->getNumberOfClusters());
}

void TestGzPack::testIntensityData()
{
    CPPUNIT_ASSERT_EQUAL( false , reader->eof() );
    const float *intensityData1 = reader->getIntensityData();
    CPPUNIT_ASSERT_EQUAL((float)0.0, intensityData1[ 0]);
    CPPUNIT_ASSERT_EQUAL((float)1.0, intensityData1[ 1]);
    CPPUNIT_ASSERT_EQUAL((float)2.0, intensityData1[ 2]);
    CPPUNIT_ASSERT_EQUAL((float)10.0, intensityData1[ 3]);
    CPPUNIT_ASSERT_EQUAL((float)11.0, intensityData1[ 4]);
    CPPUNIT_ASSERT_EQUAL((float)31.0, intensityData1[10]);
    CPPUNIT_ASSERT_EQUAL((float)32.0, intensityData1[11]);
    intensityData1 = reader->getIntensityData();
    CPPUNIT_ASSERT_EQUAL((float)100.0, intensityData1[ 0]);
    CPPUNIT_ASSERT_EQUAL((float)101.0, intensityData1[ 1]);
    CPPUNIT_ASSERT_EQUAL((float)102.0, intensityData1[ 2]);
    CPPUNIT_ASSERT_EQUAL((float)110.0, intensityData1[ 3]);
    CPPUNIT_ASSERT_EQUAL((float)111.0, intensityData1[ 4]);
    CPPUNIT_ASSERT_EQUAL((float)131.0, intensityData1[10]);
    CPPUNIT_ASSERT_EQUAL((float)132.0, intensityData1[11]);
    CPPUNIT_ASSERT_EQUAL( false,  reader->eof() );
    // user storage in a float *
    float *intensityData2 = new float[reader->getNumberOfClusters() * reader->getNumberOfChannels()];
    reader->getIntensityData(intensityData2);
    CPPUNIT_ASSERT_EQUAL((float)200.0, intensityData2[ 0]);
    CPPUNIT_ASSERT_EQUAL((float)201.0, intensityData2[ 1]);
    CPPUNIT_ASSERT_EQUAL((float)202.0, intensityData2[ 2]);
    CPPUNIT_ASSERT_EQUAL((float)210.0, intensityData2[ 3]);
    CPPUNIT_ASSERT_EQUAL((float)211.0, intensityData2[ 4]);
    CPPUNIT_ASSERT_EQUAL((float)231.0, intensityData2[10]);
    CPPUNIT_ASSERT_EQUAL((float)232.0, intensityData2[11]);
    CPPUNIT_ASSERT_EQUAL( false,  reader->eof() );

    CPPUNIT_ASSERT_EQUAL( false,  reader->eof() );
    delete [] intensityData2;
    // user storage in a vector
    vector<float> intensityList(reader->getNumberOfClusters() * reader->getNumberOfChannels());
    reader->getIntensityData(intensityList.begin());
    CPPUNIT_ASSERT_EQUAL((float)300.0, intensityList[ 0]);
    CPPUNIT_ASSERT_EQUAL((float)301.0, intensityList[ 1]);
    CPPUNIT_ASSERT_EQUAL((float)302.0, intensityList[ 2]);
    CPPUNIT_ASSERT_EQUAL((float)310.0, intensityList[ 3]);
    CPPUNIT_ASSERT_EQUAL((float)311.0, intensityList[ 4]);
    CPPUNIT_ASSERT_EQUAL((float)331.0, intensityList[10]);
    CPPUNIT_ASSERT_EQUAL((float)332.0, intensityList[11]);
    intensityData1 = reader->getIntensityData();
    CPPUNIT_ASSERT_EQUAL( true,  reader->eof() );
}

void TestGzPack::testEof()
{
    const float *intensityData = 0;
    int currentCycle = 0;
    CPPUNIT_ASSERT_EQUAL(false, reader->eof());
    // read 3 cycles
    while (Z_NULL != (intensityData = reader->getIntensityData()) && 3 > currentCycle)
    {
        ++currentCycle;
        intensityData = NULL;
    }
    CPPUNIT_ASSERT_EQUAL(3, currentCycle);
    CPPUNIT_ASSERT(NULL != intensityData);
    CPPUNIT_ASSERT_EQUAL(false, reader->eof());
    intensityData = NULL;
    intensityData = reader->getIntensityData();
    CPPUNIT_ASSERT(NULL != intensityData);
    CPPUNIT_ASSERT_EQUAL(true, reader->eof());
}

#if 0
void TestGzPack::testMissingChannel()
{
    string m = "GzPackReader: Missing intensity value";
    GzPack::GzPackReader errorReader(createMissingChannelFile()); 
    try { errorReader.getIntensityData(); CPPUNIT_FAIL("Missed the missing intensity value");}
    catch(logic_error &e) {CPPUNIT_ASSERT_EQUAL(0, m.compare(0, m.length(), e.what())); }
}

void TestGzPack::testExcessChannel()
{
    string m = "GzPackReader: Excess intensity value";
    GzPack::GzPackReader errorReader(createExcessChannelFile());
    try {errorReader.getIntensityData(); CPPUNIT_FAIL("Missed the excess intensity value");}
    catch (logic_error &e) {CPPUNIT_ASSERT_EQUAL(0, m.compare(0, m.length(), e.what())); }
}

void TestGzPack::testMissingCluster()
{
    string m = "GzPackReader: Expected a float, got: #";
    GzPack::GzPackReader errorReader(createMissingClusterFile());
    try {errorReader.getIntensityData(); CPPUNIT_FAIL("Missed the missing cluster");}
    catch (logic_error &e) {CPPUNIT_ASSERT_EQUAL(0, m.compare(0, m.length(), e.what(), m.length())); }
}

std::string TestGzPack::createMissingChannelFile()
{
    const string errorFile = "missingChannel.gz";
    IoUtils::AutoGzFile fh(errorFile, "w6R");
    fh.puts("#2\t2\n1\t2\n3\n#\n1\t2\n3\t4\n#\n");
    return errorFile;
}

std::string TestGzPack::createMissingClusterFile()
{
    const string errorFile = "missingCluster.gz";
    IoUtils::AutoGzFile fh(errorFile, "w6R");
    fh.puts("#2\t2\n1\t2\n#\n1\t2\n#\n");
    return errorFile;
}

std::string TestGzPack::createExcessChannelFile()
{
    const string errorFile = "excessChannel.gz";
    IoUtils::AutoGzFile fh(errorFile, "w6R");
    fh.puts("#2\t2\n1\t2\n3\t4\t5\n#\n1\t2\n3\t4\n#\n");
    return errorFile;
}
#endif

