/*
 * Declaration of standard objects
 *
 * Copyright (c) Solexa 2005, 2006
 * Author:
 * Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <string.h>
#else
#include <cstring>
#endif
#include <cassert>
#include "ClusterObject.h"

// Default Constructor
ClusterObject::ClusterObject()
    : boundary(0)
    , nb(0)
    , area(0)
    , na(0)
    , is_split(0)
    , pass_number(0)
    , intensity(0.0f)
{
    min.x = max.x = 0;
    min.y = max.y = 0;

    max_pixel.x = 0;
    max_pixel.y = 0;
    max_pixel.value = 0.0f;

    position.x = 0.0f;
    position.y = 0.0f;
}

void ClusterObject::SetObjectBoundary(st_pixel * pBoundary,
                                      int nBoundary)
{
    assert(nBoundary > 0);
    this->nb = nBoundary;
    this->boundary = new st_pixel [nBoundary];
    memcpy(this->boundary, pBoundary, nBoundary * sizeof(st_pixel));
}

void ClusterObject::SetObjectArea(st_pixel * pArea,
                                  int nArea)
{
    assert(nArea > 0);
    this->na = nArea;
    this->area = new st_pixel [nArea];
    memcpy(this->area, pArea, nArea * sizeof(st_pixel));
}

/***************************************************************************
* Shift all coordinate references by (x0, y0)
****************************************************************************/
void ClusterObject::Shift(unsigned short x0,
                          unsigned short y0)
{
    this->min.x += x0;
    this->min.y += y0;
    this->max.x += x0;
    this->max.y += y0;
    this->max_pixel.x += x0;
    this->max_pixel.y += y0;
    int i;
    for (i = 0; i < this->nb; i++) {
        this->boundary[i].x += x0;
        this->boundary[i].y += y0;
    }
    for (i = 0; i < this->na; i++) {
        this->area[i].x += x0;
        this->area[i].y += y0;
    }
}

/***************************************************************************
* Create the smallest 2-d image that contains the pixels given
* for an object. All other pixels values are set to <default_value>
****************************************************************************/
float * ClusterObject::MakeNewImage(int *nx, int *ny, float default_value)
{
    int i;
    unsigned short xmin = this->min.x;
    unsigned short ymin = this->min.y;
    *nx = this->max.x - xmin + 1;
    *ny = this->max.y - ymin + 1;

    float *image = new float [ (*nx) * (*ny) ];
    if (image)
    {
        for (i = 0; i < (*nx) * (*ny); i++)
            (image)[i] = default_value;
        for (i = 0; i < this->na; i++) {
            int x = this->area[i].x - xmin;
            int y = this->area[i].y - ymin;
            (image)[x*(*ny)+y] = this->area[i].value;
        }
    }
    return image;
}

/***************************************************************************
* Debugging aid: write out image area as a table of pixel values
****************************************************************************/
void ClusterObject::Display()
{
    int   nx, ny;
    float *sp = this->MakeNewImage(&nx, &ny);
    if (sp)
    {
        display_float_image_region(sp, nx, ny, 0, 0, nx, ny, this->min.x, this->min.y);
        delete [] sp;
    }
}

void ClusterObject::clear()
{
    if (nb > 0)
    {
        delete [] boundary;
        boundary = 0;
    }
    nb = 0;

    if (na > 0)
    {
        delete [] area;
        area = 0;
    }
    na = 0;

    min.x = max.x = 0;
    min.y = max.y = 0;
}
