/*
 * Object positions and binary operations on them.
 *
 * Time-stamp: <26 Jun 2006 16:58 BST Positions.cpp km>
 * Copyright (c) Solexa   2003-2005, 2006
 *               Illumina 2008
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
/*
 * "Binary" operations check for the presence or absence of an object 
 * within a given distance on two different images. This is useful to 
 * answer questions such as "Is this object present on both images?",
 * regardless of the object's intensity.
 */

#include <fstream>
#include <cmath>
#include <algorithm>
#include "Stats.h"
#include "Positions.h"
#include "PositionData.h"
#include "ImageProcess.h"
#include "ProcessAlign.h"

using namespace std;


//---------------------------------------------------------------------------
// binary operations
//---------------------------------------------------------------------------

/**
 * Auxiliary function to fill all pixels within a maximum distance
 * of min_dist from the point (x0,y0) in an (nx,ny)-image array with 
 * a given value.
 */
static void helper_fill_pixels(int *array, const int nx, const int ny,
                               const int x0, const int y0,
                               const int value, const float min_dist)
{
    int size = (int)ceil(min_dist);
    int xmin = std::max(0, x0-size); int xmax = std::min(nx-1, x0+size);
    int ymin = std::max(0, y0-size); int ymax = std::min(ny-1, y0+size);
    int x, y, x2;
    float min_dist2 = min_dist * min_dist;
    for (x = xmin; x <= xmax; x++) {
        x2 = (x-x0)*(x-x0);
        for (y = ymin; y <= ymax; y++)
            if ((float)(x2+(y-y0)*(y-y0)) <= min_dist2)
                array[x*ny+y] = value;
    }
}

/**
 * Takes two lists of positions and returns a list that contains 
 * positions that appear in either the first or the second list.
 * the caller has to make sure enough memory has been reserved
 * for the returned list
 */
void Positions::bin_or(Position       * positions,  int       * count,
                       const Position * positions1, const int   count1,
                       const Position * positions2, const int   count2,
                       const int        nx,
                       const int        ny,
                       const float      min_dist)
{
    int i;
    // fill temporary image array
    int *temp = new int [nx*ny];
    for (i = 0; i < nx*ny; i++)
        temp[i] = 0;
    for (i = 0; i < count1; i++)
        helper_fill_pixels(temp, nx, ny, 
                           (int)positions1[i].X, (int)positions1[i].Y,
                           1, min_dist);

    // copy first list
    if (positions1 != positions)
        memcpy(positions, positions1, count1*sizeof(Position));
    *count = count1;

    // loop over second list
    int xpos, ypos;
    for (i = 0; i < count2; i++) {
        xpos = (int)positions2[i].X; ypos = (int)positions2[i].Y;
        if ((xpos >= 0) && (xpos < nx) && (ypos >= 0) && (ypos < ny)
            && (temp[xpos*ny+ypos] == 0)) {
            helper_fill_pixels(temp, nx, ny, xpos, ypos, 1, min_dist);
            positions[(*count)].X = (float)xpos;
            positions[(*count)].Y = (float)ypos;
            (*count)++;
        }
    }

    delete [] temp;
}


/**
 * Takes two lists of positions and returns a list of positions that
 * appear in the first, but not in the second list
 */
void Positions::bin_subtract(Position       * positions,  int       * count,
                             const Position * positions1, const int   count1,
                             const Position * positions2, const int   count2,
                             const int        nx,
                             const int        ny,
                             const float      min_dist)
{
    int i;
    int *temp = new int [nx*ny];
    for (i = 0; i < nx*ny; i++)
        temp[i] = 0;
    for (i = 0; i < count2; i++)
        helper_fill_pixels(temp, nx, ny,
                           (int)positions2[i].X, (int)positions2[i].Y,
                            1, min_dist);
    *count = 0;
    int xpos, ypos;
    for (i = 0; i < count1; i++) {
        xpos = (int)positions1[i].X; ypos = (int)positions1[i].Y;
        if ((xpos >= 0) && (xpos < nx) && (ypos >= 0) && (ypos < ny)
            && (temp[xpos*ny+ypos] == 0)) {
            positions[(*count)].X = (float)xpos;
            positions[(*count)].Y = (float)ypos;
            (*count)++;
        }
    }
    delete [] temp;
}

/**
 * Takes two lists of positions and return the a list of positions
 * that appear in both.
 */
void Positions::bin_and(Position       * positions,  int       * count,
                        const Position * positions1, const int   count1,
                        const Position * positions2, const int   count2,
                        const int        nx,
                        const int        ny,
                        const float      min_dist)
{
    int i;
    int *temp = new int [nx*ny];
    for (i = 0; i < nx*ny; i++)
        temp[i] = 0;
    for (i = 0; i < count2; i++)
        helper_fill_pixels(temp, nx, ny,
                           (int)positions2[i].X, (int)positions2[i].Y,
                           1, min_dist);
    *count = 0;
    int xpos, ypos;
    for (i = 0; i < count1; i++) {
        xpos = (int)positions1[i].X; ypos = (int)positions1[i].Y;
        if ((xpos >= 0) && (xpos < nx) && (ypos >= 0) && (ypos < ny)
            && (temp[xpos*ny+ypos] > 0)) {
            positions[(*count)].X = (float)xpos;
            positions[(*count)].Y = (float)ypos;
            (*count)++;
        }
    }
    delete [] temp;
}

//---------------------------------------------------------------------------
/**
 * Calculates the number of common objects in two images. The second
 * set of objects is shifted by the given offset.
 * @deprecated  may be removed in future versions.
 */
int Positions::calculate_object_overlap(const Position * positions1, const int n1,
                                        const Position * positions2, const int n2,
                                        const int nx,   const int ny,
                                        const int xoff, const int yoff,
                                        const float min_dist)
{
    if ((n1 == 0) || (n2 == 0))
        return 0;

    int        overlap = 0;
    int        n = n1+n2;
    Position * positions   = new Position [n];

    // the first argument to bin_and() determines the maximum overlap.
    // make sure that the max overlap is not greater than the smaller of
    // the numbers of objects
    if (n1 < n2)
        Positions::bin_and(positions, &overlap,
                           positions1, n1, positions2, n2, nx, ny, min_dist);
    else
        Positions::bin_and(positions, &overlap,
                           positions2, n2, positions1, n1, nx, ny, min_dist);

    delete [] positions;
    return overlap;
}

int Positions::calculate_object_overlap(const PixelValue * positions1, const int n1,
                                        const PixelValue * positions2, const int n2,
                                        const int nx, const int ny,
                                        const int xoff, const int yoff,
                                        const float min_dist)
{
    Position *pos1 = (Position *)fposition_from_fnoisepixel((st_fnoisepixel *)positions1, n1);
    Position *pos2 = (Position *)fposition_from_fnoisepixel((st_fnoisepixel *)positions2, n2);
    int result = 
      calculate_object_overlap(pos1, n1, pos2, n2, nx, ny,
                               xoff, yoff, min_dist);
    delete [] pos1; delete [] pos2;
    return result;
}


//---------------------------------------------------------------------------
void Positions::initialise()
{
    shiftX = shiftY = 0.;
    nx = ny = 0;
}

Positions::Positions()
{
    initialise();
}

/**
 * Copy constructor
 */
Positions::Positions(const Positions& temp)
{
    initialise();
    *this = temp;
}

Positions::Positions(const int n,
                     const int x,
                     const int y)
{
    initialise();
    resize(n);
    nx = x;
    ny = y;
}

Positions::~Positions()
{
}


/**
 * Saves list of positions to text file.
 */
int Positions::write(const char filename[]) const
{
    PositionData posfile(filename);
    return posfile.write(&at(0), int(size()));
}

/**
 * Reads a list of positions from a text file.
 */
int Positions::read(const char filename[])
{
    Position *data;
    int       num = 0;
    int       exitCode = 0;
    PositionData posfile(filename);
    exitCode = posfile.read(&data, &num);
    if (num > 0) {
        resize(num);
        memcpy(&at(0), data, num*sizeof(Position));
    }
    return exitCode;
}

/**
 * Shifts positions of each object in the list by a constant offset.
 */
void Positions::shift(const float x,
                      const float y)
{
    int i;
    for (i = 0; i < int(size()); i++) {
        at(i).X += x;
        at(i).Y += y;
    }
    shiftX = x;
    shiftY = y;
}

void Positions::align(Positions & reference,
                      const int         maxlag,
                      const float       fwhm)
{
    float xlag = 0., ylag = 0.;
    st_pixel *pos1 = pixel_from_fposition((st_fposition *)(&reference.at(0)),
                                          reference.size());
    st_pixel *pos2 = pixel_from_fposition((st_fposition *)(&at(0)),
                                          size());
    AlignmentRecord alignment;
    align_positions(pos1, reference.size(),
                    pos2, size(), nx, ny, fwhm,
                    maxlag, maxlag, &alignment);
    xlag = alignment.XOffset; ylag = alignment.YOffset;
    shift(-xlag, -ylag);
    delete [] pos1;
    delete [] pos2;
}

Positions *Positions::bin_or(const Positions &s2, float distance)
{
    Positions *result = new Positions(size()+s2.size(), nx, ny);
    int        num = 0;
    Positions::bin_or(&(result->at(0)), &num,
                      &at(0), int(size()),
                      &s2.at(0), int(s2.size()), nx, ny, distance);
    result->resize(num);
    return result;
}

Positions *Positions::bin_subtract(const Positions &s2, float distance)
{
    Positions *result = new Positions(size(), nx, ny);
    int        num = 0;
    bin_subtract(&(result->at(0)), &num,
                 &at(0), size(),
                 &s2.at(0), s2.size(), nx, ny, distance);
    result->resize(num);
    return result;
}

Positions *Positions::bin_and(const Positions &s2, float distance)
{
    Positions *result = new Positions(size(), nx, ny);
    int        num = 0;
    bin_and(&(result->at(0)), &num,
            &at(0), size(),
            &s2.at(0), s2.size(), nx, ny, distance);
    result->resize(num);
    return result;
}
