/*
 * Class for image alignment and intensity extraction.
 *
 * Time-stamp: <26 Jan 2007 17:18 GMT Process.cpp km>
 *
 * Performs automated alignment of iteratively passed images.
 * The individual images are processed elsewhere. This class simply
 * combines images taken in different colours, 
 * remaps coordinates suitably (image registration),
 * identifies common objects and extracts intensity informtion for multiple 
 * images. The output is written to disk.
 * 
 * Copyright (c) Solexa 2003, 2004, 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <ctype.h>
#include <math.h>
#include <string.h>
#else
#include <cctype>
#include <cmath>
#include <cstring>
#endif
#include <cstdio>
#include <cstdlib>
#include <cassert>
#include "DNACodes.h"
#include "IOProcess.h"
#include "FFT.h"
#include "ImageAnalysisConfig.h"
#include "IntensityData.h"
#include "Process.h"
#include "ProcessRemapping.h"
#include "QCImage.h"
#include "Tools.h"

//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
/*
 * Class that accumulates data from the camera and performs 
 * image registration and intensity extraction.
 */

DataProcessor::DataProcessor(const RunInfo & runInfo)
{
    memcpy(&run_info, &runInfo, sizeof(RunInfo));

    default_transforms   = new AlignmentTransform [run_info.NumChannels];
    bases                = new unsigned char [run_info.NumChannels];
    offset_tracker       = new OffsetTracker(run_info.NumTiles, 
                                             run_info.NumChannels);
    images               = new Image * [run_info.NumChannels];
    memset( images, 0, run_info.NumChannels * sizeof( Image * ) );
    max_object_number = run_info.NumChannels 
                       * run_info.ImageWidth * run_info.ImageHeight
                       / ImageAnalysisConfig::minPixelSpacePerObject;
    aligner = 0;
    useRemappedIntensityFlag    = true;
}

DataProcessor::~DataProcessor()
{
    delete [] images;
    delete [] default_transforms;
    delete [] bases;
    delete    offset_tracker;
}

void DataProcessor::SetTile(const TileInfo & tileInfo)
{
    memcpy(&current_tile, &tileInfo, sizeof(TileInfo));
}

void DataProcessor::SetParameters(const AnalysisParameters & params)
{
    memcpy(&analysis_params, &params, sizeof(AnalysisParameters));
}

void DataProcessor::SetAlignment(const Alignment * defaultAlignment)
{
    int i;
    for (i = 0; i < run_info.NumChannels; i++)
        default_transforms[i].copy(defaultAlignment[i],
                                   run_info.ImageWidth, run_info.ImageHeight);
}

void DataProcessor::setUseRemappedIntensityFlag(bool value)
{
    useRemappedIntensityFlag = value;
}


/**
 * Calculates the coordinate transforms needed
 * to align images back to the first cycle reference.
 *
 * WARNING: The caller MUST NOT deallocate any image data until
 * all 4 colours in a tile have been completed.
 */
void DataProcessor::AnalyseImage(Image          *image,
                                 const int       colour,
                                 const char      base,
                                 const Position *referencepositions,
                                 const int       referencecount)
{
    assert(image);
    assert(image->nx == run_info.ImageWidth);
    assert(image->ny == run_info.ImageHeight);
    assert(image->getData());
    assert(image->getData()->filter_image != 0);
    assert(image->getData()->noise_image  != 0);
    assert(image->getNumObjects() < 
           run_info.ImageWidth * run_info.ImageHeight / ImageAnalysisConfig::minPixelSpacePerObject);

    image->data->setSequencingInfo(base, current_tile.IsDeblocked);
    images[colour] = image;
    bases[colour]  = base;
    image->getData()->analyseEstimateFWHM();

    /*
     * image realignment
     */
    assert(default_transforms[colour].Region.XSize == run_info.ImageWidth);
    assert(default_transforms[colour].Region.YSize == run_info.ImageHeight);
    image->data->transform     = new AlignmentTransform;
    image->data->usedTransform = new AlignmentTransform;
    image->data->transform->copy(default_transforms[colour]);

    /* 
     * alignment on subregions of the image
     * to get an idea of image distortions 
     */
    if (current_tile.Cycle > 0) {
        aligner->alignRegionPixelPositions(image->getObjects(),
                                           image->getNumObjects(),
                                           image->data->transform);
    }
    image->data->usedTransform->copy(*(image->data->transform));
}

/**
 * Initialises the image alignment lookup table and cache Fourier transforms
 * for faster later access.
 */
void DataProcessor::SetReference(const Position * referencePositions,
                                 const int        referenceCount)
{
    aligner = new ProcessAlign(referencePositions, referenceCount, 
                               run_info.ImageWidth, run_info.ImageHeight,
                               analysis_params.ClusterFwhm);
    aligner->setMaximumOffsets(int(analysis_params.MaxStageOffset));
    aligner->setRegionParameters(2, 2);
}

void DataProcessor::ClearReference()
{
    if (aligner)
        delete aligner;
    aligner = 0;
}

int DataProcessor::validateOffsets()
{
    int i;
    int is_deblocked = images[0]->data->sequencing->is_deblocked;
    AlignmentTransform *position_transforms 
             = new AlignmentTransform [run_info.NumChannels];
    int *object_counts = new int [ run_info.NumChannels ]; 
    for (i = 0; i < run_info.NumChannels; i++) {
        position_transforms[i].copy(*(images[i]->data->usedTransform));
        object_counts[i] = images[i]->getNumObjects();
    }

    /* 
     * QC of image offsets 
     */

    /* QC for image realignment values */
    int valid_flag = 0;
    int valid_offsets = 0;

    valid_flag = validate_object_counts(object_counts, run_info.NumChannels, 200);
    for (i = 0; i < run_info.NumChannels; i++)
        if ((fabs(position_transforms[i].XOffset) 
                > analysis_params.MaxStageOffset) ||
            (fabs(position_transforms[i].YOffset) 
                > analysis_params.MaxStageOffset))
            valid_flag &= 0;

    if (valid_flag)
        valid_offsets = validate_offsets(position_transforms,
                                         run_info.NumChannels,
                                         analysis_params.MaxColourOffset,
                                         default_transforms);
    
    if ((valid_offsets == run_info.NumChannels) && (current_tile.Cycle > 0) 
                                                && (!(is_deblocked))
        && (valid_flag))
        offset_tracker->add_offsets(position_transforms, run_info.NumChannels,
                                    0);

    delete [] position_transforms;
    delete [] object_counts;

    return (valid_flag && valid_offsets);
}

bool sortPositionsbyCoordinate(const Position &lhs,
                               const Position &rhs)
{
    if (lhs.X == rhs.X)
        return (lhs.Y < rhs.Y);
    return (lhs.X < rhs.X);
}

void DataProcessor::BuildReference(PixelValue     ** objectLists,
                                   int             * numObjects,
                                   const int         numLists,
                                   Position       ** newReference,
                                   int             * numNewReference,
                                   const Position  * oldReference,
                                   const int         numOldReference)
{
    *numNewReference = 0;
    RemappingLookup *remapper = new RemappingLookup(run_info.ImageWidth,
                                                    run_info.ImageHeight);
    remapper->setReference(oldReference, numOldReference);
    remapper->setMaxNewPositions(max_object_number - numOldReference);
    int i;
    for (i = 0; i < numLists; i++)
        remapper->addNewPositions(objectLists[i], numObjects[i],
                                  analysis_params.ClusterDistance);
    remapper->findNewPositions(analysis_params.ClusterDistance);
    remapper->addNewPositionsToReference();
    *numNewReference = remapper->getNumberNewPositions();
    *newReference = new Position [*numNewReference];
#if 1
    std::vector<Position> sortPositions(*numNewReference);
    memcpy(&sortPositions[0], remapper->getNewPositions(), 
           *numNewReference * sizeof(Position));
    sort(sortPositions.begin(), sortPositions.end(), sortPositionsbyCoordinate);
    memcpy(*newReference, &sortPositions[0], 
           *numNewReference * sizeof(Position));
#else
    memcpy(*newReference, remapper->getNewPositions(), 
           *numNewReference * sizeof(Position));
#endif
    delete remapper;
    return;
}

void DataProcessor::getObjectListsCopy(PixelValue ** objectLists,
                                       int         * numObjects,
                                       const int     numLists)
{
    assert(numLists == run_info.NumChannels);
    int i;
    for (i = 0; i < run_info.NumChannels; i++) {
        numObjects[i]  = images[i]->getNumObjects();
        objectLists[i] = new PixelValue [numObjects[i]];
        memcpy(objectLists[i], images[i]->getObjects(),
               numObjects[i] * sizeof(PixelValue));
    }

    for (i = 0; i < run_info.NumChannels; i++)
        images[i]->getData()->usedTransform->transformReverse(objectLists[i], 
                                                              numObjects[i]);
}

void DataProcessor::BuildReference(Position       ** newReference,
                                   int             * numNewReference,
                                   const Position  * oldReference,
                                   const int         numOldReference)
{
    int i;
    int         * object_counts = new int          [run_info.NumChannels];
    PixelValue ** object_lists  = new PixelValue * [run_info.NumChannels];
    getObjectListsCopy(object_lists, object_counts, run_info.NumChannels);

    BuildReference(object_lists, object_counts, run_info.NumChannels,
                   newReference, numNewReference, oldReference, numOldReference);

    for (i = 0; i < run_info.NumChannels; i++)
        delete [] object_lists[i];
    delete [] object_counts;
    delete [] object_lists;
}

void DataProcessor::CompleteTile(float          * intensities,
                                 float          * noise,
                                 const Position * referencePositions,
                                 int            & numReference,
                                 const int        numElements,
                                 const char    ** error_message,
                                 int            * flag, 
                                 int            * refIndices)
{
    int i;
    int validFlag = 0;
    validFlag = validateOffsets();
    if (validFlag) 
         analyseTile(referencePositions, numReference,
                     intensities, noise, refIndices);
    else {
        for (i = 0; i < run_info.NumChannels * numReference; i++) {
            intensities[i] = 0.;
            noise[i] = 0.;
            refIndices[i] = -1;
        }
    }
    if (flag)
        *flag = validFlag;
}

/**
 * Performs the remapping of all colours and extracts the remapped intensities.
 */
void DataProcessor::analyseTile(const Position * referencePositions,
                                const int        referenceCount,
                                float          * intensities,
                                float          * noise,
                                int            * referenceIndices)
{
    int         * object_counts = new int          [run_info.NumChannels];
    PixelValue ** object_lists  = new PixelValue * [run_info.NumChannels];

    getObjectListsCopy(object_lists, object_counts, run_info.NumChannels);

    /*
     * remap molecules for the current position and produce array
     * with molecule intensities in all colours
     */
    int * indices = new int [referenceCount * run_info.NumChannels];
    int   i;
    for (i = 0; i < referenceCount*run_info.NumChannels; i++)
        indices[i] = -1;

    RemappingLookup *remapper = new RemappingLookup(run_info.ImageWidth,
                                                    run_info.ImageHeight);
    remapper->setReference(referencePositions, referenceCount);
    for (i = 0; i < run_info.NumChannels; i++) {
        remapper->remap(object_lists[i], object_counts[i],
                        analysis_params.ClusterDistance);
        memcpy(indices + i * referenceCount,
               remapper->getIndex(), referenceCount*sizeof(int));
    }
    delete remapper;
    if (referenceIndices != 0)
        memcpy(referenceIndices, indices,
               referenceCount * run_info.NumChannels * sizeof(int));

    if (useRemappedIntensityFlag) {
        for (i = 0; i < run_info.NumChannels; i++) {
            int col = get_base_index(bases[i], bases, run_info.NumChannels);
            remap_colour(referencePositions, indices + i * referenceCount,
                         referenceCount,
                         object_lists[i], object_counts[i],
                         intensities, noise, run_info.NumChannels, col);
        }
    }
    else {
        for (i = 0; i < referenceCount*run_info.NumChannels; i++)
            indices[i] = -1;
    }

    for (i = 0; i < run_info.NumChannels; i++)
        delete [] object_lists[i];
    delete [] object_counts;
    delete [] object_lists;

    /*
     * go back to images to read out missing intensity values
     * at the reference positions
     */
    for (i = 0; i < run_info.NumChannels; i++) {
        int col = get_base_index(bases[i], bases, run_info.NumChannels);
        complete_intensities(referencePositions,
                             indices + i * referenceCount, referenceCount,
                             run_info.NumChannels, col,
                             images[i]->getData()->filter_image,
                             images[i]->getData()->noise_image,
                             run_info.ImageWidth, run_info.ImageHeight,
                             intensities, noise,
                             analysis_params.ClusterFwhm, 0.51,
                             images[i]->getData()->usedTransform);
    }

    delete [] indices;
}


void DataProcessor::displayOffsets()
{
    if (offset_tracker)
        offset_tracker->display(default_transforms);
}
