/*
 * Image alignment routines.
 *
 * Time-stamp: <26 Jun 2006 16:56 BST ProcessAlign.cpp km>
 * Copyright (c) Solexa   2003-2005, 2006
 *               Illumina 2008
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
/*
 * Routines to realign images to a common coordinate frame.
 */

//---------------------------------------------------------------------------

#if defined(__BORLANDC__)
#include <math.h>
#include <string.h>
#else
#include <cmath>
#include <cstring>
#endif
#include <cassert>
#include <cstdio>
#include "FFT.h"
#include "Optimise.h"
#include "ProcessAlign.h"
#include "Tools.h"
#include "Stats.h"
#include "ImageUtil.h"
#include "Positions.h"


//---------------------------------------------------------------------------
// alignment routines
//---------------------------------------------------------------------------

// to make sure that index is in valid range [0, n)
#define WRAPPED_INDEX(i, n) (((i)+(n)) % (n)) 


/*
 * use Levenberg-Marquardt fit to interpolate peak in order to 
 * determine maximum to sub-pixel accuracy
 */
void fit_gaussian_maximum(float *array, int nx, int ny,
                          float max, float background, int imax, int jmax,
                          float *xmax, float *ymax, 
                          float *corr_min, float *corr_max, float *corr_width)
{
    const int nd1 = 3;
    const int na = 6;
    float x1v[nd1*nd1], x2v[nd1*nd1], yv[nd1*nd1], sig[nd1*nd1];
    float a[na];
    int ia[na];
    int i, j;

    // use nd1*nd1 pixel area around maximum as data
    for (i = 0; i < na; i++)
        ia[i] = 1;
    for (i = 0; i < nd1; i++) {
        int iidx = imax+(i-nd1/2);
        for (j = 0; j < nd1; j++) {
            int jidx = jmax+(j-nd1/2);
            int idx = i*nd1+j;
            x1v[idx] = (float)iidx; x2v[idx] = (float)jidx;
            yv[idx] = array[WRAPPED_INDEX(iidx, nx)*ny+WRAPPED_INDEX(jidx, ny)]; 
            sig[idx] = 1.;
        }
    }

    // set initial parameters for LM fit
    a[0] = max - background;
    a[5] = background;
    a[1] = (float)imax; a[2] = (float)jmax;
    // estimate width of Gaussian
    float yoff1 = 0.25*
        (array[WRAPPED_INDEX(imax-1, nx)*ny + WRAPPED_INDEX(jmax, ny)] +
         array[WRAPPED_INDEX(imax+1, nx)*ny + WRAPPED_INDEX(jmax, ny)] +
         array[WRAPPED_INDEX(imax, nx)*ny   + WRAPPED_INDEX(jmax-1, ny)] +
         array[WRAPPED_INDEX(imax, nx)*ny   + WRAPPED_INDEX(jmax+1, ny)]);
    float sigma = 1.f/sqrt(-2.f*log(fabs(yoff1-background)/a[0]));
    a[3] = a[4] = sigma;
  
    // LM fit
    float chisq = 0.;
    LM_2d(x1v, x2v, yv, sig, nd1*nd1, a, ia, na, &chisq, LM_gauss_2d_der);
    *xmax = a[1];
    *ymax = a[2];
    *corr_min = a[5];
    *corr_max = a[0]+a[5];
    *corr_width = sqrt(a[3]*a[3] + a[4]*a[4]);
}

/*
 * search global maximum of <nx>*<ny> array <corr> over an area 
 * centred around the initial values of <*xm>, <*ym>.
 * Maximal offsets are <xmaxoff> and <ymaxoff>.
 * Also returned are values are the minimum and maximum array
 * elements within this area.
 */
void find_correlation_maximum(float *corr, int nx, int ny,
                              int xmaxoff, int ymaxoff,
                              int *xm, int *ym,
                              float *min, float *max)
{
    int x, y;
    int x0 = *xm;
    int y0 = *ym;
    *max = corr[WRAPPED_INDEX(x0, nx)*ny+WRAPPED_INDEX(y0, ny)];
    *min = corr[WRAPPED_INDEX(x0, nx)*ny+WRAPPED_INDEX(y0, ny)];
    for (x = -xmaxoff+x0; x <= xmaxoff+x0; x++)
        for (y = -ymaxoff+y0; y <= ymaxoff+y0; y++) {
            int idx = WRAPPED_INDEX(x, nx)*ny+WRAPPED_INDEX(y, ny);
            if (corr[idx] > *max) {
                *max = corr[idx];
                *xm = x;
                *ym = y;
            }
            if (corr[idx] < *min)
                *min = corr[idx];
        }
}

/*
 * find the maximum of a correlation function and interpolate 
 * the maximum to sub-pixel accuracy
 */
void find_correlation_maximum(float *corr, int nx, int ny,
                              int xmaxlag, int ymaxlag,
                              float *xlag, float *ylag,
                              float *corr_min, float *corr_max,
                              float *corr_width)
{
    // find the maximum (and minimum) of the correlation function
    float min, max;
    int xm = Stats::round(*xlag), ym = Stats::round(*ylag);
    find_correlation_maximum(corr, nx, ny, xmaxlag, ymaxlag, 
                             &xm, &ym, &min, &max);

    // interpolate correlation function by Gaussian around maximum
    // to improve accuracy
    fit_gaussian_maximum(corr, nx, ny, max, min, xm, ym, xlag, ylag,
                         corr_min, corr_max, corr_width);
}

                              
/*
 * Calculate the correlation function for a range of possible image shifts.
 * Unlike for the FFT version, no periodic boundary conditions are assumed
 */
void correlate_images_real(float *image1, float *image2, int nx, int ny,
                           float *corr, 
                           int xmaxlag, int ymaxlag, int x0, int y0)
{
    int xmin, xmax, ymin, ymax;
    double sum = 0.;
    int a, b, x, y;
    for (a = -xmaxlag+x0; a <= xmaxlag+x0; a++)
        for (b = -ymaxlag+y0; b <= ymaxlag+y0; b++) {
            xmin = (a < 0 ? -a : 0); xmax = (a > 0 ? nx-a: nx);
            ymin = (b < 0 ? -b : 0); ymax = (b > 0 ? ny-b: ny);
            sum = 0;
            for (x = xmin; x < xmax; x++)
                for (y = ymin; y < ymax; y++)
                    sum +=  image1[x*ny+y] * image2[(x+a)*ny+(y+b)];
            corr[WRAPPED_INDEX(a, nx)*ny+WRAPPED_INDEX(b, ny)] = sum;
        }
}

/*
 * Align to images by computing their correlation function.
 * The correlation function is interpolated by a 2-d Gaussian 
 * to find maxima that do not lie on integer pixel shifts.
 * Unfortunately the lags of the correlation function are defined as
 * the negative of the way the image offsets are defined.
 */
void align_images(float *image1, float *image2,
                  int nx, int ny, int xmaxlag, int ymaxlag,
                  AlignmentRecord *alignment)
{
    if ((xmaxlag > nx/2) || (xmaxlag < 0))
        xmaxlag = nx/2;
    if ((ymaxlag > ny/2) || (ymaxlag < 0))
        ymaxlag = ny/2;

    // calculate the correlation function as a function of the lag by FFT
    float *corr = new float [nx*ny];
    correlate_images_real(image1, image2, nx, ny, corr, 
                          xmaxlag, ymaxlag, 
                          Stats::round(alignment->XOffset),
                          Stats::round(alignment->YOffset));

    find_correlation_maximum(corr, nx, ny, xmaxlag, ymaxlag,
                             &alignment->XOffset,
                             &alignment->YOffset,
                             &alignment->CorrelationMinimum,
                             &alignment->CorrelationMaximum,
                             &alignment->CorrelationWidth);

    delete [] corr;
}

/*
 * align images using FFTs
 */
void align_images_fft(float *image1, float *image2,
                      int nx, int ny, int xmaxlag, int ymaxlag,
                      AlignmentRecord *alignment)
{
    if ((xmaxlag > nx/2) || (xmaxlag < 0))
        xmaxlag = nx/2;
    if ((ymaxlag > ny/2) || (ymaxlag < 0))
        ymaxlag = ny/2;

    // calculate the correlation function as a function of the lag by FFT
    float *corr = new float [nx*ny];
    correlate_images(image2, image1, nx, ny, corr); // image order inverted
    find_correlation_maximum(corr, nx, ny, xmaxlag, ymaxlag,
                             &alignment->XOffset, 
                             &alignment->YOffset,
                             &alignment->CorrelationMinimum,
                             &alignment->CorrelationMaximum,
                             &alignment->CorrelationWidth);

    delete [] corr;
}

void align_fft_images(float *image1, float *image2, int nx, int ny, 
                      int xmaxlag, int ymaxlag,
                      AlignmentRecord *alignment)
{
    if ((xmaxlag > nx/2) || (xmaxlag < 0))
        xmaxlag = nx/2;
    if ((ymaxlag > ny/2) || (ymaxlag < 0))
        ymaxlag = ny/2;

    // calculate the correlation function as a function of the lag by FFT
    float *corr = new float [nx*ny];
    correlate_fft_images(image2, image1, nx, ny, corr); // image order inverted
    find_correlation_maximum(corr, nx, ny, xmaxlag, ymaxlag,
                             &alignment->XOffset, 
                             &alignment->YOffset,
                             &alignment->CorrelationMinimum,
                             &alignment->CorrelationMaximum,
                             &alignment->CorrelationWidth);

    delete [] corr;
}

/*
 * generates a gaussian PSF
 */
void create_gaussian_mask(float *mask, int n, int ny, float amp, float fwhm)
{
    int size = n/2;
    float sigma = fwhm / (2.f*sqrt(2.f*log(2.f)));
    float x2;  
    int x, y;

    for (x = 0; x < n; x++)
        for (y = 0; y < n; y++) {
            x2 = (float)((x-size)*(x-size)+(y-size)*(y-size));
            mask[x*n+y] = amp * exp(-x2 / (2 * sigma * sigma));
        }
}

/*
 * create a "clean" image by inserting objects with a perfect 
 * Gaussian point spread functions at the reference positions.
 */
void create_clean_image(float *image, int nx, int ny,
                        const st_pixel *positions, int npos, float fwhm,
                        int uniform)
{
    float sigma = fwhm / (2.f*sqrt(2.f*log(2.f)));
    int size = (int)ceil(2.f*sigma);
    int ns = 2*size+1;
    float *psf = new float[ ns*ns];
    int i;
 
    for (i = 0; i < nx*ny; i++)
        image[i]= 0.;
    create_gaussian_mask(psf, ns, ns, 1., fwhm);
    for (i = 0; i < npos; i++)
        add_image_region(image, nx, ny, psf, ns, ns,
                         positions[i].x-size, positions[i].y-size, 
                         (uniform ? 100.f : positions[i].value));
    delete [] psf;
}

void create_clean_fft_image(float *fourier, int nx, int ny,
                            st_pixel *positions, int npos,
                            float fwhm, int uniform)
{
    float *image = new float [nx*ny];
    create_clean_image(image, nx, ny, positions, npos, fwhm, uniform);
    fft_2d_r2c(image, fourier, nx, ny);
    delete [] image;
}

/*
 * align two lists of object positions
 */
void align_positions(const st_pixel * positions1, const int n1,
                     const st_pixel * positions2, const int n2,
                     const int x, const int y,
                     const float fwhm, const int xmaxlag, const int ymaxlag,
                     AlignmentRecord *alignment)
{
    if ((n1 == 0) || (n2 == 0))
        return;

    int nx = get_optimal_fft_length(x);
    int ny = get_optimal_fft_length(y);
    float *image1 = new float [nx*ny];
    float *image2 = new float [nx*ny];

    // initialise temporary arrays
    create_clean_image(image1, nx, ny, positions1, n1, fwhm, 1);
    create_clean_image(image2, nx, ny, positions2, n2, fwhm, 1);

    // align images
    if ((xmaxlag > 5) || (ymaxlag > 5))
        align_images_fft(image1, image2, nx, ny, xmaxlag, ymaxlag,
                         alignment);
    else
        align_images    (image1, image2, nx, ny, xmaxlag, ymaxlag,
                         alignment);
    
    delete [] image1;
    delete [] image2;
}

/*
 * align only objects within a subregion of the image whose corners are
 * given by the coordinates (<x1>, <y1>), (<x2>, <y2>)
 */
void align_region_positions(st_pixel *positions1, int n1,
                            st_pixel *positions2, int n2,
                            int x1, int y1, int x2, int y2,
                            float fwhm, int xmaxlag, int ymaxlag,
                            AlignmentRecord *alignment)
{
    int i;

    // discard unneeded positions and shift coordinate origin
    int region_n1 = 0;
    st_pixel *region_positions1 = new st_pixel [n1];
    for (i = 0; i < n1; i++) {
        int x = positions1[i].x;
        int y = positions1[i].y;
        if ((x >= x1) && (x < x2) && (y >= y1) && (y < y2)) {
            region_positions1[region_n1].x = x-x1;
            region_positions1[region_n1].y = y-y1;
            region_n1++;
        }
    }
        
    // discard unneeded positions and shift coordinate origin
    int region_n2 = 0;
    st_pixel *region_positions2 = new st_pixel [n2];
    for (i = 0; i < n2; i++) {
        int x = positions2[i].x;
        int y = positions2[i].y;
        if ((x >= x1) && (x < x2) && (y >= y1) && (y < y2)) {
            region_positions2[region_n2].x = x-x1;
            region_positions2[region_n2].y = y-y1;
            region_n2++;
        }
    }
    align_positions(region_positions1, region_n1,
                    region_positions2, region_n2, 
                    x2-x1, y2-y1, fwhm, xmaxlag, ymaxlag, alignment);
  
    delete [] region_positions1;
    delete [] region_positions2;
}

void align_region_positions(st_fposition *positions1, int n1,
                            st_fnoisepixel *positions2, int n2,
                            int x1, int y1, int x2, int y2,
                            float fwhm, int xmaxlag, int ymaxlag,
                            AlignmentRecord *alignment)
{
    st_pixel *pos1 = pixel_from_fposition(positions1, n1);
    st_pixel *pos2 = pixel_from_fnoisepixel(positions2, n2);

    align_region_positions(pos1, n1, pos2, n2, x1, y1, x2, y2,
                            fwhm, xmaxlag, ymaxlag, alignment);

    delete [] pos1;
    delete [] pos2;
}
//---------------------------------------------------------------------------
/**
 * Shifts pixel coordinates by specified offsets.
 * @param  pixels        pixel coordinates and values
 * @param  n             number of pixels
 * @param  xoff, yoff    image offsets
 */
void shiftPixels(PixelValue * pixels,
                 const int    n,
                 const float  xoff,
                 const float  yoff)
{
    int i;
    for (i = 0; i < n; i++) {
        pixels[i].X -= xoff;
        pixels[i].Y -= yoff;
    }
}

/**
 * Rescales pixel coordinates by specified factors, keeping the origin fixed.
 *
 * @param  pixels          pixel coordinates and values
 * @param  n               number of pixels
 * @param  xscale, yscale  image scale factors
 * @param  xorig, yorig    origin
 */
void scalePixels(PixelValue * pixels,
                 const int    n, 
                 const float  xscale = 1.f,
                 const float  yscale = 1.f, 
                 const float  xorig  = 0.f,
                 const float  yorig  = 0.f)
{
    float facx = xscale / (1.f + xscale);
    float facy = yscale / (1.f + yscale);
    int i;
    for (i = 0; i < n; i++) {
        pixels[i].X -= (pixels[i].X - xorig) * facx;
        pixels[i].Y -= (pixels[i].Y - yorig) * facy;
    }
}


void AlignmentTransform::transformReverse(PixelValue * pixels,
                                          const int    n) const
{
    shiftPixels(pixels, n, XOffset, YOffset);
    if ((XScale != 0.f) || (YScale != 0.f))
        scalePixels(pixels, n, XScale, YScale,
                    Region.XSize/2.f, Region.YSize/2.f);
}

void AlignmentTransform::copy(const Alignment & ref,
                              const int         nx,
                              const int         ny)
{
    XOffset            = ref.XOffset;
    YOffset            = ref.YOffset;
    XScale             = ref.XScale;
    YScale             = ref.YScale;
    Region.XSize       = nx;
    Region.YSize       = ny;
    Region.XMin        = 0;
    Region.YMin        = 0;
    nlocal             = 0;
}    

void AlignmentTransform::copy(const AlignmentTransform & ref)
{
    XOffset            = ref.XOffset;
    YOffset            = ref.YOffset;
    XScale             = ref.XScale;
    YScale             = ref.YScale;
    Region.XSize       = ref.Region.XSize;
    Region.YSize       = ref.Region.YSize;
    Region.XMin        = ref.Region.XMin;
    Region.YMin        = ref.Region.YMin;
    CorrelationMinimum = ref.CorrelationMinimum;
    CorrelationMaximum = ref.CorrelationMaximum;
    CorrelationWidth   = ref.CorrelationWidth;
    if (nlocal > 0)
       delete [] local_alignment;
    nlocal = ref.nlocal;
    if (nlocal > 0) {
        local_alignment = new AlignmentRecord [nlocal];
        memcpy(local_alignment, ref.local_alignment,
               nlocal * sizeof(AlignmentRecord));
    }
}

/**
 * Reads a set of aligment transforms from a text file.
 * This reads four transforms at a time; one for each channel.
 * The file format is:
 * Every line corresponds to one transform for one colour. 
 * The columns are x-offset, y-offset, x-scaling and y-scaling.
 * Lines starting with '#' are treated as comments
 */
int AlignmentTransform::load(const char  filename[],
                             Alignment * transforms,
                             const int   nmax)
{
    FILE *fp = fopen(filename, "r");
    char buffer[256];
    if (fp) {
        int count = 0, assigned = 0;
        while ((count < nmax) && (fgets(buffer, sizeof(buffer), fp))) {
            // remove comments
            int i;
            for (i = 0; i < (int)strlen(buffer); i++)
                if (buffer[i] == '#') {
                    buffer[i] = '\0';
                    break;
                }
            // parse line
            assigned = sscanf(buffer, "%f %f %f %f", 
                              &(transforms[count].XOffset), 
                              &(transforms[count].YOffset),
                              &(transforms[count].XScale),
                              &(transforms[count].YScale));
            if (assigned < 3)
                transforms[count].XScale = 0.f;
            if (assigned < 4)
                transforms[count].YScale = 0.f;
            if (assigned >= 2)
                count ++;
        }
        fclose(fp);
        return count;
    }
    return -1;
}

//---------------------------------------------------------------------------

/*
 * wrapper object to store a Fourier transformed reference image
 * for multiple image alignments
 */
ProcessAlign::ProcessAlign(const Position * refpositions,
                           const int        refcount,
                           const int        nx2,
                           const int        ny2,
                           const float      f)
{
    nx = nx2;
    ny = ny2;
    fnx = get_optimal_fft_length(nx);
    fny = get_optimal_fft_length(ny);
    fwhm = f;
    maxlagx = nx/4; maxlagy = ny/4;    reference_count = refcount;
    reference_positions = pixel_from_fposition((st_fposition *)refpositions, refcount);
    reference_fourier_image = new_fourier_array(fnx, fny);
    create_clean_fft_image(reference_fourier_image, fnx, fny,
                           reference_positions, reference_count, fwhm, 1);
}

ProcessAlign::~ProcessAlign()
{
    delete [] reference_positions;
    delete_fourier_array(reference_fourier_image);
}

void ProcessAlign::setMaximumOffsets(const int xlag,
                                     const int ylag)
{
    maxlagx = xlag;
    maxlagy = ylag;
    if (maxlagy < 0)
        maxlagy = maxlagx;
}

/**
 * Aligns a set of positions to the stored reference image.
 */
void ProcessAlign::alignPixelPositions(const PixelValue * positions,
                                       const int          count,
                                       AlignmentRecord *offsets,
                                       const float width)
{
    float f = width;
    if (f <= 0.)
        f = fwhm;
    st_pixel *positions2 = pixel_from_fnoisepixel((st_fnoisepixel *)positions, count);
    float *fourier_image = new_fourier_array(fnx, fny);
    create_clean_fft_image(fourier_image, fnx, fny, positions2, count, f, 1);
    align_fft_images(reference_fourier_image, fourier_image, fnx, fny, 
                     maxlagx, maxlagy, offsets);
    offsets->Region.XMin  = 0;
    offsets->Region.YMin  = 0;
    offsets->Region.XSize = nx;
    offsets->Region.YSize = ny;
    delete_fourier_array(fourier_image);
    delete [] positions2;
}

void ProcessAlign::setRegionParameters(const int numx,
                                       const int numy,
                                       const int sizex,
                                       const int sizey) 
{
    assert(numx > 1);
    assert(numy > 1);
    // implementation uses integer bit mask; do not overflow this
    assert(numx*numy < 8 * (int)sizeof(int) - 1); 
    numregionx = numx; numregiony = numy;
    if (sizex > 0) 
        regionsizex = sizex;
    else 
        regionsizex = nx / numregionx;
    if (sizey > 0) 
        regionsizey = sizey;
    else 
        regionsizey = ny / numregiony;
}

/**
 * Aligns subregions of the image separately. This is to identify 
 * image distortions that go beyond simple shifting.
 * The image is split into a set of regions and each tile 
 * is aligned separately.
 */ 
void ProcessAlign::alignRegionPixelPositions(const PixelValue   * positions,
                                             const int            count,
                                             AlignmentTransform * transform)
{
    float max_offset_distance = 3.f;
    st_pixel *pos = pixel_from_fnoisepixel((st_fnoisepixel *)positions, count);
    
    // initialise
    if (transform->nlocal > 0)
        delete [] transform->local_alignment; 
    transform->nlocal = numregionx * numregiony;
    transform->local_alignment = new AlignmentRecord [ transform->nlocal ];
    transform->XOffset = 0.f;
    transform->YOffset = 0.f;
    transform->XScale  = 0.f;
    transform->YScale  = 0.f;
    transform->Region.XSize   = nx;
    transform->Region.YSize   = ny;

    // perform local alignments
    int i, j;
    for (i = 0; i < numregionx; i++)
        for (j = 0; j < numregiony; j++) {
            int x1, x2, y1, y2;
            AlignmentRecord *ltransform 
                = &(transform->local_alignment[i*numregiony+j]);
            x1 = i*(nx-regionsizex)/(numregionx-1); x2 = x1 + regionsizex;
            y1 = j*(ny-regionsizey)/(numregiony-1); y2 = y1 + regionsizey;
            ltransform->Region.XMin = x1; 
            ltransform->Region.YMin = y1;
            ltransform->Region.XSize = regionsizex;
            ltransform->Region.YSize = regionsizey;
            ltransform->XOffset = 0.f;
            ltransform->YOffset = 0.f;
            ltransform->CorrelationMinimum = 0.f;
            ltransform->CorrelationMaximum = 0.f;
            ltransform->CorrelationWidth = 0.f;
            align_region_positions(reference_positions, reference_count,
                                   pos, count, x1, y1, x2, y2, fwhm,
                                   maxlagx, maxlagy, ltransform);
        }
 
    //exclude image regions with zero correlation maximum 
    int corrflag = 0;
    for (i = 0; i < transform->nlocal; i++)
        if (transform->local_alignment[i].CorrelationWidth > 0.f)
            corrflag |= (1<<i);

    // validate that alignments are mutually compatible within some tolerance.
    // this is to remove the effect of out-of-focus regions.
    int flag = AlignmentRecord::validate(transform->local_alignment,
                                         transform->nlocal,
                                         max_offset_distance, corrflag);

    // count surviving compatible regions
    int n = 0;
    for (i = 0; i < transform->nlocal; i++)
        if (flag & (1<<i))
            n++;

    if (n > 0) { // more than one alignment validated
        // Fit regression lines through x- and y-offsets; 
        // this assumes a linear stretch and no image rotation.
        // If linear regression fails, use the arithmetic mean instead.
        float *x = new float [n]; float *fx = new float [n];
        float *y = new float [n]; float *fy = new float [n];
        int i2, idx;
        idx = 0;
        for (i2 = 0; i2 < numregionx; i2++)
            for (j = 0; j < numregiony; j++) {
                i = i2*numregiony+j;
            if (flag & (1<<i)) { // exclude divergent alignments
                AlignmentRecord *ltransform 
                    = &(transform->local_alignment[i]);
                x[idx]  = ltransform->Region.XMin + regionsizex/2 
                          - transform->Region.XSize/2;
                fx[idx] = ltransform->XOffset;
                y[idx]  = ltransform->Region.YMin + regionsizey/2 
                          - transform->Region.YSize/2;
                fy[idx] = ltransform->YOffset;
                idx++;
            }
        }

        float gradient = 0., intercept = 0.;
        if ((n > 1) && (!Stats::linregress(x, fx, n, & gradient, &intercept))) {
            transform->XOffset = intercept;      
            transform->XScale  = gradient;
        }
        else {
            transform->XOffset = Stats::average(fx, n);
            transform->XScale  = 0.;
        }
        if ((n > 1) && (!Stats::linregress(y, fy, n, & gradient, &intercept))) {
            transform->YOffset = intercept; 
            transform->YScale  = gradient;
        }
        else {
            transform->YOffset = Stats::average(fy, n); 
            transform->YScale  = 0.;
        }
        delete [] x; delete [] fx; delete [] y; delete [] fy;
    }
    else {
        // To get to this point, the image basically has to be 
        // badly messed up.
        // Let's see if we can at least salvage one subregion of the image...

        // return extreme value as a flag if no subset of the image
        // could be aligned
        transform->XOffset = -(float)(transform->Region.XSize);
        transform->YOffset = -(float)(transform->Region.YSize);
        // if at least one region aligned, choose most likely alignment
        // based on the width of the correlation function
        // TODO: find better criterion
        // exclude correlation functions of width less than 2 - 
        // they are most likely dodgy
        corrflag = 0;
        for (i = 0; i < transform->nlocal; i++)
            if (transform->local_alignment[i].CorrelationWidth > 2.f)
                corrflag |= (1<<i);
        // now find smallest width
        if (corrflag) {
            float width = sqrt((float)(transform->Region.XSize 
                                     * transform->Region.YSize));
            for (i = 0; i < transform->nlocal; i++)
                if ((corrflag & (1<<i)) && 
                    (transform->local_alignment[i].CorrelationWidth < width)) {
                    width = transform->local_alignment[i].CorrelationWidth;
                    transform->XOffset = transform->local_alignment[i].XOffset;
                    transform->YOffset = transform->local_alignment[i].YOffset;                    
                 }
        }
    }
    delete [] pos;
}

/**
 * Takes a list of <numTtransforms> alignment transforms and
 * verifies if any of the x- and y-offsets are identical 
 * within a certain tolerance <maxOffsets>.
 * @param  maxOffsets  maximum valid xy offsets between colours
 * @param  includeFlag optional argument that contains a bit mask
 *                     of <numTransforms> bits which allows the user
 *                     to indicate which of the transforms are to be used;
 *                     transforms with zero bits are ignored.
 * 
 * @return an integer flag, whose first <ntrans> bits indicate
 *         whether the corresponding <transforms> element agrees
 *         with the majority vote for x,y offsets of the <transforms>
 *         (a bit value of 1 indicates agreement, 0 a misalignment)
 */ 
int AlignmentRecord::validate(AlignmentRecord * transforms,
                              const int         numTransforms,
                              const float       maxOffsets,
                              const int         includeFlag)
{
    int i, j;
    int flag = 0;
    int includeflag = includeFlag;

    // look at all transforms by default
    if (includeflag == -1) {
        includeflag = 0;
        for (i = 0; i < numTransforms; i++)
            includeflag |= (1<<i);
    }

    // number of transforms to be investigated
    int targetcount = 0;
     
    for (i = 0; i < numTransforms; i++)
        if (includeflag & (1<<i))
            targetcount ++;

    // if not more than one transform provided, exit
    if (targetcount <= 1)
        return includeflag;

    /* for each tranform, find find all other transforms that have 
     * similar offsets within a maximum distance of max_offset
     */
    int *flags = new int [numTransforms];
    for (i = 0; i < numTransforms; i++)
        flags[i] = 0;
    for (i = 0; i < numTransforms; i++) 
        for (j = i+1; j < numTransforms; j++)
            if ((includeflag & (1<<i)) && (includeflag & (1<<j)) &&
                (fabs(transforms[i].XOffset - transforms[j].XOffset) 
                 < maxOffsets) &&
                (fabs(transforms[i].YOffset - transforms[j].YOffset) 
                 < maxOffsets)) {
                    flags[i] |= (1<<i) | (1<<j);
                    flags[j] |= (1<<i) | (1<<j);
            }
    /* if there is more than one set of mutually incompatible offsets, 
     * no sensible alignment can be found
     */
    for (i = 0; i < numTransforms; i++) { 
        if (flags[i] != 0)
            for (j = i+1; j < numTransforms; j++)
                if ((flags[j] != 0) && (flags[j] & flags[i] == 0))
                    goto terminate;
    }

    // set up return flag to point at all aligned transforms
    for (i = 0; i < numTransforms; i++)
        if (flags[i] != 0)
            flag |= (1<<i);

  terminate:
    delete [] flags;
    return flag;
}
