/*
 * Time-stamp: <26 Jun 2006 17:04 BST ProcessAlign.h km>
 * Copyright (c) Solexa 2005, 2006
 * Author:
 * Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

//---------------------------------------------------------------------------

#ifndef _PROCESSALIGN_H
#define _PROCESSALIGN_H

#include "ImageUtil.h"
#include "ParameterTypes.h"

struct AlignmentRecord : public AlignmentStats {
public:
    virtual ~AlignmentRecord() {};
    static int validate(AlignmentRecord * transforms,
                        const int         numTransforms,
                        const float       maxOffset   = 5.,
                        const int         includeflag = -1);
};

void align_images(float *image1, float *image2,
                  int nx, int ny, int xmaxlag, int ymaxlag,
                  AlignmentRecord *alignment);
void align_images_fft(float *image1, float *image2,
                      int nx, int ny, int xmaxlag, int ymaxlag,
                      AlignmentRecord *alignment);
void align_fft_images(float *image1, float *image2, int nx, int ny, 
                      int xmaxlag, int ymaxlag,
                      AlignmentRecord *alignment);
void create_clean_fft_image(float *fourier, int nx, int ny,
                            st_pixel *positions, int npos,
                            float fwhm, int uniform);
void align_positions(const st_pixel *positions1, const int n1,
                     const st_pixel *positions2, const int n2,
                     const int nx, const int ny,
                     const float fwhm, const int xmaxlag, const int ymaxlag,
                     AlignmentRecord *alignment);

/**
 * Transforms coordinate systems between different images.
 */
class AlignmentTransform : public AlignmentRecord
{
 public:
    AlignmentRecord * local_alignment;
    int               nlocal;
    AlignmentTransform(const int x = 0,
                       const int y = 0)
        { XOffset = 0.f; XOffset = 0.f;
          XScale  = 0.f; YScale  = 0.f;
          Region.XMin  = 0; Region.YMin  = 0;
          Region.XSize = x; Region.YSize = y;
          CorrelationMinimum = 0.f; CorrelationMaximum = 0.f;
          CorrelationWidth   = 0.f;
          local_alignment = 0; nlocal = NumAlignmentRegions; }
    AlignmentTransform(const AlignmentTransform & ref)
    {
        AlignmentTransform();
        copy(ref);
    }
    ~AlignmentTransform()
    {
        if (nlocal > 0)
            delete [] local_alignment;
    }
    void copy(const AlignmentTransform & ref);
    void copy(const Alignment & ref, 
              const int         nx,
              const int         ny);
    void transformForward(float * x,
                          float * y) const
    {
        *x += (*x - Region.XSize/2) * XScale + XOffset;
        *y += (*y - Region.YSize/2) * YScale + YOffset;
    }
    void transformReverse(float * x,
                          float * y) const
    {
        *x = (*x + Region.XSize/2.f*XScale - XOffset) / (1.f+XScale);
        *y = (*y + Region.YSize/2.f*YScale - YOffset) / (1.f+YScale);
    }
    void transformReverse(PixelValue * pixel,
                          const int    numValues) const;
    static int load(const char   filename[],
                    Alignment  * transforms,
                    const int    nmax);
};


class ProcessAlign
{
    float    * reference_fourier_image;
    st_pixel * reference_positions;
    int        reference_count;
    float fwhm;
    int maxlagx, maxlagy;
    int nx, ny, fnx, fny;
    int numregionx, numregiony;
    int regionsizex, regionsizey;
 public:
    ProcessAlign(const Position * refpositions,
                 const int        refcount,
                 const int        nx2,
                 const int        ny2,
                 const float      fwhm);
    ~ProcessAlign();
    void setMaximumOffsets(const int xlag,
                           const int ylag = -1);
    void alignPixelPositions(const PixelValue * positions,
                             const int          numPositions,
                             AlignmentRecord  * offsets, 
                             const float        fwhm = -1.f);
    void setRegionParameters(const int numx,
                             const int numy,
                             const int sizex = -1,
                             const int sizey = -1); 
    void alignRegionPixelPositions(const PixelValue   * positions,
                                   const int            numPositions,
                                   AlignmentTransform * transform);
};

#endif /* _PROCESSALIGN_H */
