/*
 * Post-processing for detected image objects
 *
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#else
#include <cmath>
#endif
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <algorithm>
#include "Maths.h"
#include "Stats.h"
#include "Optimise.h"
#include "Tools.h"
#include "Sorting.h"
#include "ClusterObject.h"
#include "ProcessDetectedObjectStats.h"

/*
 * Create a lookup <array> of dimension <nx>*<ny>, such that
 * for each of the <nobj> <objects> the pixels that belong to this
 * object (are above the local detection threshold) are set to the 
 * index of the object in the array <objects>. The first object has
 * index 1 (not 0), the second 2 etc.
 */
void create_lookup_array(int *array, int nx, int ny,
                         ClusterObject *objects, int nobj)
{
    int i, j;
    for (i = 0; i < nx * ny; i++)
        array[i] = 0;
    for (i = 0; i < nobj; i++) {
        for (j = 0; j < objects[i].na; j++) {
            int x = objects[i].area[j].x;
            int y = objects[i].area[j].y;
            if ((x >= 0) && (x < nx) && (y >= 0) && (y < ny)) {
                int idx = x * ny + y;
                array[idx] = i+1;
            }
        }
    }
}

/*
 * Expand the lookup <array> of dimension <nx>*<ny>
 * created by the previous routine such 
 * that each empty pixel that is adjacent to one ore more pixels
 * belonging to exactly one object is set to the negative index 
 * of the object in the <nobj>-dim array <objects>.
 * Example: a pixel adjacent to object 2 is set to -2.
 */
void expand_lookup_array(int *array, int nx, int ny,
                         ClusterObject *objects, int nobj)
{
    int i, x, y;
    for (i = 0; i < nobj; i++) {
        /*
         * search minimal square region around current object
         */ 
        int xmin = std::max(0,    objects[i].min.x - 1);
        int xmax = std::min(nx-1, objects[i].max.x + 1);
        int ymin = std::max(0,    objects[i].min.y - 1);
        int ymax = std::min(ny-1, objects[i].max.y + 1);
        int list[4];
        int l, nl;
        int flag;
        for (x = xmin; x <= xmax; x++)
            for (y = ymin; y <= ymax; y++)
                if (array[x*ny+y] == 0) {
                    nl = 0;
                    int x2 = x-1;
                    int y2 = y;
                    
                    /* create list of the 4 nearest pixels */
                    if (x2 >= 0) {
                        list[nl] = array[x2*ny+y2]; nl++;
                    }
                    x2 = x+1;
                    if (x2 < nx) {
                        list[nl] = array[x2*ny+y2]; nl++;
                    }
                    x2 = x;
                    y2 = y-1;
                    if (y2 >= 0) {
                        list[nl] = array[x2*ny+y2]; nl++;
                    }
                    y2 = y+1;
                    if (y2 < ny) {
                        list[nl] = array[x2*ny+y2]; nl++;
                    }

                    /* check if pixel is adjacent to current cluster */
                    flag = 0;
                    for (l = 0; l < nl; l++)
                        if (list[l] == i+1)
                            flag = 1;
                    /* check if pixel is adjacent to any other cluster */
                    if (flag)
                        for (l = 0; l < nl; l++)
                            if ((list[l] != 0) && (list[l] != i+1)
                                && (list[l] != -i-1) && (list[l] != -(nobj+1)))
                                flag = 2;
                    
                    /* if pixel is adjacent to current cluster, set to negative index */ 
                    if (flag == 1)
                        array[x*ny+y] = -(i+1);
                    /* if pixel is adjacent to more than one cluster, 
                     * set to unused dummy value */
                    if (flag == 2)
                        array[x*ny+y] = -(nobj+1);
                }
    }
}

/*
 * Debugging routine to print out a subset of the lookup array
 */
void print_lookup_array(int *array, int nx, int ny,
                        int firstx, int firsty, int lastx, int lasty)
{
    if (lastx < firstx)
        lastx = nx;
    if (lasty < firsty)
        lasty = ny;
    int x, y;
    firstx = std::max(firstx, 0); lastx = std::min(lastx, nx-1);
    firsty = std::max(firsty, 0); lasty = std::min(lasty, ny-1);
    for (x = firstx; x < lastx; x++) {
        for (y = firsty; y < lasty; y++) {
            printf(" %4i", array[x*ny+y]);
        }
        printf("\n");
    }
}

/*
 * Debugging routine to print out a subset of the <array>.
 * Pixels that are activated in the lookup array are preceded by a '*'.
 */ 
void print_lookup_image_array(float *array, int *lookup, int nx, int ny,
                              int firstx, int firsty, int lastx, int lasty)
{
    if (lastx < firstx)
        lastx = nx;
    if (lasty < firsty)
        lasty = ny;
    int x, y;
    firstx = std::max(firstx, 0); lastx = std::min(lastx, nx-1);
    firsty = std::max(firsty, 0); lasty = std::min(lasty, ny-1);
    for (x = firstx; x < lastx; x++) {
        for (y = firsty; y < lasty; y++) {
            if (lookup[x*ny+y] != 0) {
                printf("x%4.0f", array[x*ny+y]);
            }
            else {
                printf(" %4.0f", array[x*ny+y]);
            }
        }
        printf("\n");
    }
}

/*
 * Estimate the image background value and the dispersion of the background
 * around each of the <nobj> <objects> for the <nx>*<ny>-dim. image <array>.
 * Pixels that are active in the <lookup> array are ignored
 * for the calculation of the background.
 * The results are returned by setting the corresponding fields in the
 * record <stats>.
 * The background is estimated by using rank statistics. The background is 
 * given by the median of the pixel distribution, and the noise is chosen such 
 * that 68% of the pixels lie within a 2-sigma interval.
 * The procedure is iterated and each time 3-sigma outliers are removed to
 * avoid large contaminations by residual low-level signal. 
 */             
void get_background_dispersion(float *array, int *lookup, int nx, int ny,
                               ClusterObject *objects, int nobj,
                               ObjectStatsRecord *stats, float *array2)
{
    int nf = 10;
    float *selected = new float [(2*nf+1)*(2*nf+1)];
    int i, x, y;
    for (i = 0; i < nobj; i++) {
        /*
         * Choose a window around the current pixel, select all pixels that 
         * are not thresholded by the object detection
         * and sort the resulting pixel list.
         */
        int xmin = std::max(objects[i].max_pixel.x - nf, 0);
        int xmax = std::min(objects[i].max_pixel.x + nf, nx-1);
        int ymin = std::max(objects[i].max_pixel.y - nf, 0);
        int ymax = std::min(objects[i].max_pixel.y + nf, ny-1);
#if 0
        print_lookup_array(lookup, nx, ny, xmin, ymin, xmax+1, ymax+1);
        print_lookup_image_array(array, lookup, nx, ny, xmin, ymin, xmax+1, ymax+1);
        display_float_image_region(array, nx, ny, xmin, ymin,
                                   xmax-xmin+1, ymax-ymin+1);
        display_float_image_region(array2, nx, ny, xmin, ymin,
                                   xmax-xmin+1, ymax-ymin+1);
#endif
        int count = 0;
        for (x = xmin; x <= xmax; x++)
            for (y = ymin; y <= ymax; y++) {
                if (lookup[x*ny+y] == 0) {
                    selected[count] = array[x*ny+y];
                    count++;
                }
            }
        if (count < 20)
            break;
        Sorting::sort(selected, size_t(count));

        /*
         * Determine rank statistics from sorted pixel values.
         * Iteratively discard 3-sigma outliers.
         */
        int imed, imax, imin;
        int fullcount = count;
        int prevcount = count;
        do {
            imed = count/2;
            imin = std::max(0, (int)Stats::round(count/2-0.34*count));
            imax = std::min(fullcount-1, (int)Stats::round(count/2+0.34*count));
            stats[i].background = selected[imed];
            stats[i].noise = (selected[imax]- selected[imin])/2.f;
#if 0
            printf("   %i %i %f %f\n",
                   objects[i].max_pixel.x, objects[i].max_pixel.y,
                   stats[i].background, stats[i].noise);
#endif
            prevcount = count;
            while ((count-1 >= 0) && 
                   (selected[count-1] > stats[i].background + 3.f*stats[i].noise ))
                count --;
        } while ((count >= 20) && (prevcount - count >= 2));

#if 0
        int idx = 0, oldidx = 0;
        for (x = 0; x < 50; x++) {
            float xval = (x-20)*stats[i].noise/5.+stats[i].background;
            while ((idx < count) && (selected[idx] < xval))
                idx++;
            printf("(%f %f %i),", (x-20)/5., xval, idx-oldidx);
            oldidx = idx;
        }
        printf("\n");
#endif
    }
    delete [] selected;
}

#if 0
#define TEST_PROCEDURE
#endif

void print_params(const char msg[], float *p, int np, float chisq = -1.)
{
    int i;
    printf(msg);
    for (i = 0; i < np; i++)
        printf(" %6.1f", p[i]);
    if (chisq < 0.)
        printf("\n");
    else
        printf(" %g\n", chisq);
}

int fit_gaussian(float *x1val, float *x2val, float *yval, float *sig,
                 int n, float *x, float *y, float *max,
                 float *xfwhm, float *yfwhm, 
                 float *background)
{
    const int np = 5;
    const int IDX_AMPLI = 0, IDX_X1ARG = 1, IDX_X2ARG = 2;
    const int IDX_X1SIG = 3, IDX_X2SIG = 3, IDX_CONST = 4;    
    float *p = new float [np];
    float *po = new float [np];
    int *ip = new int [np];
    int i, j;

    float minval, maxval;
    get_array_min_max(yval, n, &minval, &maxval);
#if defined(TEST_PROCEDURE)
    printf("<Values>\n");
    for (i = 0; i < n; i++)
        printf("%f %f %f %f\n", x1val[i], x2val[i], yval[i], sig[i]); 
    printf("</Values>\n");
#endif

    ip[IDX_X1ARG] = 1; po[IDX_X1ARG] = *x;
    ip[IDX_X2ARG] = 1; po[IDX_X2ARG] = *y;
    ip[IDX_X1SIG] = 1; po[IDX_X1SIG] = *xfwhm;
    ip[IDX_X2SIG] = 1; po[IDX_X2SIG] = *yfwhm;
    if (*background >= 0.) {
        ip[IDX_AMPLI] = 1; po[IDX_AMPLI] = maxval - *background;
        ip[IDX_CONST] = 0; po[IDX_CONST] = *background;
    }
    else {
        ip[IDX_AMPLI] = 1; po[IDX_AMPLI] = maxval - minval;
        ip[IDX_CONST] = 1; po[IDX_CONST] = minval;
    } 
    
    float best_chisq  = 1.e30f;
    int allfailed = 1;
    float factor = 2.3f;
#if defined(TEST_PROCEDURE)
    print_params("new object ", po, np);
#endif
    for (i = 0; i < 3; i++) {
        for (j = 0; j < np; j++)
            p[j] = po[j];
        p[IDX_X1SIG] = po[IDX_X1SIG] / pow(factor, i);
        p[IDX_X2SIG] = po[IDX_X2SIG] / pow(factor, i);

#if defined(TEST_PROCEDURE)
        print_params("params in ", p, np);
#endif

        float chisq = 0.;
        int failed = LM_2d(x1val, x2val, yval, sig, n, p, ip, np, &chisq,
                           LM_gauss_spherical_2d_der);
        if (!failed) {
#if defined(TEST_PROCEDURE)
            print_params("params out", p, np, chisq);
#endif
             if ((chisq < best_chisq) 
                  && (fabs(p[IDX_X1ARG]-po[IDX_X1ARG]) < 2.)
                  && (fabs(p[IDX_X2ARG]-po[IDX_X2ARG]) < 2.)
                  && (p[IDX_X1SIG] > 0.)
                  && (p[IDX_AMPLI] > 0.) && (p[IDX_AMPLI] < 2.*maxval)
                  && (p[IDX_AMPLI]+p[IDX_CONST] > 0.)) {
#if defined(TEST_PROCEDURE)
                print_params("params max", p, np, chisq);
#endif
                *max = p[IDX_AMPLI]; *x = p[IDX_X1ARG]; *y = p[IDX_X2ARG];
                *background = p[IDX_CONST];
                *xfwhm = fabs(p[IDX_X1SIG]); *yfwhm = fabs(p[IDX_X2SIG]);
                best_chisq = chisq;
             }
             else
                failed = 1;
        }
        allfailed = allfailed && failed;
    }
    if (allfailed) {
        *xfwhm = 0.; *yfwhm = 0.;
    }              
    delete [] p; delete [] po; delete [] ip;
    return allfailed;
}

int fit_gaussian_object(ClusterObject *object, float *image, int nx, int ny,
                        float *x, float *y, float *max,
                        float *xfwhm, float *yfwhm, 
                        float *background)
{ 
    *x = object->max_pixel.x;
    *y = object->max_pixel.y;
    *max = object->max_pixel.value;
    *xfwhm = std::max(1.f, float(object->max.x - object->min.x));
    *yfwhm = std::max(1.f, float(object->max.y - object->min.y));

    int i;
    int na = object->na;
    float *x1val = new float [ na ]; float *x2val = new float [ na ];
    float *yval = new float [ na ];  float *sig = new float [ na ];
    for (i = 0; i < na; i++) {
        int ix = object->area[i].x;
        int iy = object->area[i].y;
        x1val[i] = ix; x2val[i] = iy;
        yval[i] = image[ix*ny+iy]; sig[i] = object->max_pixel.value/10.f;
#if defined(TEST_PROCEDURE)
        printf("%6.1f %6.1f %6.1f %6.1f\n",
               x1val[i], x2val[i], yval[i], sig[i]);
#endif
    }
    int failed = fit_gaussian(x1val, x2val, yval, sig, na,
                              x, y, max, xfwhm, yfwhm, background);
    delete [] x1val; delete [] x2val; delete [] yval; delete [] sig;
    return failed;
}

int fit_gaussian_area(int *lookup, float *array, int nx, int ny,
                      ClusterObject *object, int idx, ObjectStatsRecord *stats,
                      float *array2)
{
    int x, y;
    int xmin = std::max(0,      object->min.x - 2);
    int xmax = std::min(nx - 1, object->max.x + 2);
    int ymin = std::max(0,      object->min.y - 2);
    int ymax = std::min(ny - 1, object->max.y + 2);
#if defined(TEST_PROCEDURE)
    printf("<Index> %i </Index>\n", idx);
    printf("<Coordinates> %i %i </Coordinates>\n",
           object->max_pixel.x, object->max_pixel.y);
    print_lookup_array(lookup, nx, ny, xmin, ymin, xmax+1, ymax+1);
    print_lookup_image_array(array, lookup, nx, ny, xmin, ymin, xmax+1, ymax+1);
    display_float_image_region(array, nx, ny, xmin, ymin,
                               xmax-xmin+1, ymax-ymin+1);
    display_float_image_region(array2, nx, ny, xmin, ymin,
                               xmax-xmin+1, ymax-ymin+1);
#endif
    int na2 = 0;
    int na = (xmax - xmin + 1)* (ymax - ymin + 1);
    float *x1val = new float [na];
    float *x2val = new float [na];
    float *yval = new float [na];
    float *sig = new float [na];
    for (x = xmin; x <= xmax; x++)
        for (y = ymin; y <= ymax; y++) {
            int tmp = lookup[x*ny+y];
            if ((tmp == 0) || (abs(tmp) == idx+1)) {
                x1val[na2] = x; x2val[na2] = y;
                sig[na2] = 10.f; yval[na2] = array[x*ny+y];
                na2++;
            }
        }

    float xv = object->max_pixel.x;
    float yv = object->max_pixel.y;
    float xwidth = object->max.x - object->min.x + 0.5f;
    float ywidth = object->max.y - object->min.y + 0.5f;
    xwidth = std::max(xwidth, ywidth);
    ywidth = std::max(xwidth, ywidth);
    stats->gaussian_background = stats->background;
    stats->gaussian_maximum = stats->maximum;
    stats->gaussian_xsigma = 0.;
    stats->gaussian_ysigma = 0.;

    int failed = fit_gaussian(x1val, x2val, yval, sig, na2, 
                              &xv, &yv, &stats->gaussian_maximum,
                              &xwidth, &ywidth, &stats->gaussian_background);
    if (!failed) {
        stats->gaussian_xcentre = xv;
        stats->gaussian_ycentre = yv;
        stats->gaussian_xsigma = xwidth;
        stats->gaussian_ysigma = ywidth;
        stats->gaussian_volume = 2.f*Maths::pi*stats->gaussian_maximum*xwidth*ywidth;
        stats->gaussian_chi2 = 0.f;
#if defined(TEST_PROCEDURE)
        printf("%4i %4i   %6.1f %6.1f %3.1f %6.1f %5.1f %6.1f    ",
               object->max_pixel.x, object->max_pixel.y,
               stats->xinter, stats->yinter,
               std::max((object->xmax-object->xmin+0.5),
                   (object->ymax-object->ymin+0.5)),
               stats->maximum, stats->noise, stats->background);
        printf("%6.1f %6.1f %3.1f %4.0f %6.0f    ",
               stats->xcentre, stats->ycentre, stats->width_sigma,
               stats->maximum, stats->integral); 
        printf("%6.1f %6.1f %3.1f %3.1f %4.0f %4.0f %6.0f\n",
               xv, yv, xwidth, ywidth,
               stats->gaussian_maximum, stats->gaussian_background,
               stats->gaussian_volume);
        printf("<Plot>\n");
        printf("plot 'data.txt' u 1:3 w lp, a*exp(-0.5*(x-b)*(x-b)/(s*s))+c t 'a=%4.0f, b=%6.1f, s=%3.1f (f=%3.1f), c=%4.0f', a=%4.0f, b=%6.1f, s=%3.1f, c=%4.0f\n",
               stats->gaussian_maximum, xv, xwidth, xwidth*(2.*sqrt(2.*log(2.))), stats->gaussian_background, stats->gaussian_maximum, xv, xwidth, stats->gaussian_background);
        printf("plot 'data.txt' u 2:3 w lp, a*exp(-0.5*(x-b)*(x-b)/(s*s))+c t 'a=%4.0f, b=%6.1f, s=%3.1f (f=%3.1f), c=%4.0f', a=%4.0f, b=%6.1f, s=%3.1f, c=%4.0f\n",
               stats->gaussian_maximum, yv, ywidth, ywidth*(2.*sqrt(2.*log(2.))), stats->gaussian_background, stats->gaussian_maximum, yv, ywidth, stats->gaussian_background);
        printf("</Plot>\n");
#undef TEST_PROCEDURE
#endif
    }
    delete [] x1val; delete [] x2val; delete [] yval; delete [] sig;
    return failed;
}

/*
 * Integrate the intensity given by the <nc>*<ny>-dim. <array>
 * over the area of the <nobj> <objects> and subtract the local image
 * background. The integration area is given by the array <lookup>.
 * 
 * Also calculates the centre of gravity of the objects and
 * the maximum value. The width of the object is estimated 
 * from the total integrated intensity and maximum by 
 * assuming a Gaussian shape.
 */
void integrate_objects(int *lookup, float *array, int nx, int ny,
                       ClusterObject *objects, int nobj, ObjectStatsRecord *stats)
{
    int i, x, y;
    double integral, maximum, xcentre, ycentre, sigma;
    float value;
    for (i = 0; i < nobj; i++) {
        int xmin = std::max(0,      objects[i].min.x - 1);
        int xmax = std::min(nx - 1, objects[i].max.x + 1);
        int ymin = std::max(0,      objects[i].min.y - 1);
        int ymax = std::min(ny - 1, objects[i].max.y + 1);

        integral = 0.;
        maximum = 0.;
        xcentre = 0.;
        ycentre = 0.;
        sigma = 0.;
        for (x = xmin; x <= xmax; x++)
            for (y = ymin; y <= ymax; y++)
                if (abs(lookup[x*ny+y]) == i+1) {
                    value = (array[x*ny+y] - stats[i].background);
                    integral += value;  /* integral */
                    if (value > maximum)
                        maximum = value; /* maximum */
                    xcentre += x * value;  /* centre of gravity */
                    ycentre += y * value;
                }
        if ((integral > 0.) && (maximum > 0.)) {
            xcentre /= integral;
            ycentre /= integral;
            /* The integral I of a 2-d Gaussian of amplitude A and
             * width <s> is given by I = 2 * pi * s^2 * A */ 
            sigma = sqrt(fabs(integral / (maximum * 2. * Maths::pi)));
        }
        else {
            integral = 0.;
            maximum = 0.;
            xcentre = objects[i].max_pixel.x;
            ycentre = objects[i].max_pixel.y;
        }
        stats[i].integral = integral;
        stats[i].xcentre = xcentre;
        stats[i].ycentre = ycentre;
        stats[i].maximum = maximum;
        stats[i].width_sigma = sigma; 
    }
}


void printObjectStats(FILE                    * fp,
                      const PixelValue        * positions,
                      const ObjectStatsRecord * stats,
                      const int                 nobj)
{
    int i;
    for (i = 0; i < nobj; i++) {
        fprintf(fp, "%6.1f %6.1f %7.1f %6.1f ",
                positions[i].X, positions[i].Y, positions[i].Intensity,
                positions[i].Noise);
        fprintf(fp, "%1i %2i %2i %3i %3i ",
                stats[i].properties & OBJECT_PROPERTY_IS_SPLIT,
                stats[i].xdim, stats[i].ydim, stats[i].area, stats[i].boundary);
        fprintf(fp, "%6.1f %6.1f %3.1f %5.0f %6.0f %4.0f %5.0f ",
                stats[i].xcentre, stats[i].ycentre,
                stats[i].width_sigma * (2. * sqrt(2.*log(2.))),
                stats[i].maximum, stats[i].integral, stats[i].noise,
                stats[i].background); 
        fprintf(fp, "%6.1f %6.1f %3.1f %3.1f %5.0f %6.0f %6.0f\n",
                stats[i].gaussian_xcentre, stats[i].gaussian_ycentre,
                stats[i].gaussian_xsigma * (2. * sqrt(2.*log(2.))),
                stats[i].gaussian_ysigma * (2. * sqrt(2.*log(2.))),
                stats[i].gaussian_maximum, stats[i].gaussian_volume,
                stats[i].gaussian_chi2);
    }
}

void printObjectStats(FILE                    * fp,
                      const float             * image,
                      const float             * noise,
                      const int                 nx,
                      const int                 ny,
                      st_pixel                * positions,
                      const ObjectStatsRecord * stats,
                      const int                 nobj)
{
   PixelValue *positions2 = (PixelValue *)fnoisepixel_from_pixel(positions, nobj,
                                                                  noise, nx, ny);
   printObjectStats(fp, positions2, stats, nobj);
   delete [] positions2; 
}
