/*
 * $RCSfile: ProcessRemapping.cpp,v $
 * Author: Klaus Maisinger
 * $Id: ProcessRemapping.cpp,v 1.2 2008/07/03 11:36:01 craczy Exp $
 * 
 * Copyright (c) Solexa 2005, 2006
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

/**
 * Given suitable coordinate transformations obtained from image alignment,
 * identify identical objects or positions on multiple images.
*/
#if defined(__BORLANDC__)
#include <string.h>
#include <math.h>
#include <stdio.h>
#else
#include <cstring>
#include <cmath>
#include <cstdio>
#endif
#include <cassert>
#include <algorithm>
#include "Stats.h"
#include "Tools.h"
#include "Sorting.h"
#include "ProcessRemapping.h"

/**
 * Sorts array of objects by signal-to-noise ratios (in descending order).
 */
static void sort_objects_by_snr(PixelValue *objects, int n)
{
    int i;
    float *signal_to_noise = new float [n];
    for (i = 0; i < n; i++) {
        signal_to_noise[i] = 0.;
        if (objects[i].Noise > 0.)
            signal_to_noise[i] = objects[i].Intensity / objects[i].Noise;
    }

    size_t *index = new size_t [n];
    for (i = 0; i < n; i++)
        index[i] = i;
    Sorting::sort_index(signal_to_noise, index, size_t(n));

    PixelValue *sort_objects = new PixelValue [n];
    for (i = 0; i < n; i++)
        sort_objects[n-1-i] = objects[index[i]];
    memcpy(objects, sort_objects, n*sizeof(PixelValue));
    delete [] sort_objects;
    delete [] index;
    delete [] signal_to_noise;
#if 0
    for (i = 0; i < n; i++) 
        printf("%i %f %f %f %f %f\n", i, 
               objects[i].x, objects[i].y,
               objects[i].value, objects[i].noise,
               objects[i].value / objects[i].noise);
#endif                    
}


/**
 * Assembles index arrays from lists of object positions acquired in different
 * detection channels.
 * The index arrays contain for each reference positions
 * the indices of the objects in the lists that remap to the 
 * reference positions.
 * Objects that do not remapped are considered as new objects and are
 * added to the newpositions array.
 *
 * @param reference       <nref>-dim. array of reference positions to be used
 *                        for remapping 
 * @param nref            dimension or <referencepositions> array
 * @param positions       <ncol>-dim array of pointers to list of objects
 *                        detected in each colour
 * @param npos            <ncol>-dim array containing the number of objects
 *                        detected in each colour
 * @param nnewmax         dimension of the pre-allocated arrays <newpositions>,
 *                        <newindices>
 * @param ncol            number of different colours observed
 * @param nx, ny          image dimensions
 * @param min_dist        minimum distance for two objects in different colours
 *                        to be considered separate
 * @param nnew            current number of objects in <newpositions> and
 *                        <newindices> arrays
 * @param refindex        <nref>*<ncol>-dim. arrays holding the indices
 *                        of the remapping objects in the <positions>
 *                        arrays
 * @param newpositions, newindex
 *                        lists of positions and indices for newly 
 *                        discovered objects 
 * @param nnew            number of newly  detected objects and written
 *                        to newpositions, newintensities, newsigma
 */

RemappingLookup::RemappingLookup(int x, int y) 
{ 
    nx = x; ny = y;
    lookup = 0; reference = 0; refindex = 0; nref = 0;
    newpositions = 0; nnew = 0; nmax = 0;
}

RemappingLookup::~RemappingLookup()
{ 
    clearNewPositions();
    clearReference(); 
}

void RemappingLookup::clearReference() 
{ 
    if (lookup)
        delete [] lookup;
    if (reference)
        delete [] reference;
    if (refindex)
        delete [] refindex;
}

void RemappingLookup::clearNewPositions() 
{ 
    if (nmax > 0) {
        delete [] newpositions;
        delete [] candidate_newpositions;
    }
    nmax = 0;
}

/**
 * Creates lookup table for reference object positions.
 */
void RemappingLookup::setReference(const Position * positions,
                                   const int        numPositions)
{
    clearReference();
    nref = numPositions;
    refindex  = new int      [nref];
    reference = new Position [nref];
    memcpy(reference, positions, nref*sizeof(Position));

    // create lookup array for fast nearest neighbour-search
    int j;
    lookup = new int[nx*ny];
    for (j = 0; j < nx*ny; j++)
        lookup[j] = -1;

    for (j = 0; j < nref; j++) {
        int x = Stats::round(reference[j].X);
        int y = Stats::round(reference[j].Y);
        // these checks should normally always be in bounds;
        // may be out of bounds when working with finite precision?
        x = std::min(nx-1, x); x = std::max(0, x);
        y = std::min(ny-1, y); y = std::max(0, y);
        lookup[x*ny+y] = j;
    }
}

void RemappingLookup::getNewPositions(Position  * positions,
                                      const int   nm)
{
    memcpy(positions, newpositions, std::min(nm, nnew));
}

/**
 * Preallocates memory for maximal number of new objects to be remapped.
 */ 
void RemappingLookup::setMaxNewPositions(int n)
{
    clearNewPositions();
    nmax = n;
    nnew = 0;
    ncandnew = 0;
    candidate_newpositions = new PixelValue [nmax];
    newpositions           = new Position   [nmax];
}

/**
 * Tentatively adds some new positions to existing reference.
 * This does not modify the reference, only identifies all positions that 
 * do not match the reference, and keeps them in a list.
 */
void RemappingLookup::addNewPositions(const PixelValue * positions,
                                      const int          npos,
                                      const float        min_dist)
{
    int i;
    int *new_index = new int [npos];
    remap(positions, new_index, npos, refindex, min_dist);
    for (i = 0; i < npos; i++)
        if ((new_index[i] < 0)  && (ncandnew < nmax)) {
            memcpy(&candidate_newpositions[ncandnew], &positions[i],
                   sizeof(PixelValue));
                   ncandnew++; 
        }
            
    delete [] new_index;
}

/**
 * Goes through list of candidate new detections (built by addNewPositions())
 * and identifies overlaps to create a list of unique new detections. 
 */
void RemappingLookup::findNewPositions(float min_dist)
{
    int      * new_refindex   = new int      [nref+ncandnew];
    Position * new_reference  = new Position [nref+ncandnew];
    int i;
    for (i = 0; i < nref+ncandnew; i++)
        new_refindex[i] = -1;

    memcpy(new_reference, reference, nref*sizeof(Position));
    memcpy(new_refindex,  refindex,  nref*sizeof(int));
    float *new_ref_nearest_distance = new float [nref+ncandnew];
    for (i = 0; i < nref+ncandnew; i++) 
        new_ref_nearest_distance[i] = 2.f * min_dist * min_dist;
    
#if 1
    /* sort candidate detections by signal-to-noise ratio
     * to make sure that the object position is determined from the clearest 
     * detection channel, not by some weak cross-talk
     */
    sort_objects_by_snr(candidate_newpositions, ncandnew);
#endif

    /* go through list and add candidate detections to reference lookup table
     */
    int j;
    for (j = 0; j < ncandnew; j++) {
        int index;
        float dist;
        remapPosition(candidate_newpositions[j], &index, &dist,
                      min_dist, new_reference, nref+ncandnew,
                      j, new_refindex, new_ref_nearest_distance);
        if ((index < 0) && (nnew < nmax)) {
            int ix = Stats::round(candidate_newpositions[j].X);
            int iy = Stats::round(candidate_newpositions[j].Y);
            if ((ix < 0) || (ix >= nx) || (iy < 0) || (iy >= ny))
                continue; /* object shifted out of image frame */
            lookup[ix*ny+iy] = nref+nnew;
            newpositions[nnew].X = candidate_newpositions[j].X;
            newpositions[nnew].Y = candidate_newpositions[j].Y;
            new_reference[nref+nnew].X =  newpositions[nnew].X;
            new_reference[nref+nnew].Y =  newpositions[nnew].Y;
            nnew++;
        }
        
    }
    delete [] new_ref_nearest_distance;
    delete [] new_refindex;
    delete [] new_reference;
}

/**
 * Merges positions detected by findNewPositions() into reference
 */
void RemappingLookup::addNewPositionsToReference()
{
    int i;
    int      * newrefindex  = new int      [nref+nnew];
    Position * newreference = new Position [nref+nnew];
    for (i = 0; i < nref+nnew; i++)
        newrefindex[i] = -1;
    
    memcpy(newreference,      reference,    nref*sizeof(Position));
    memcpy(newreference+nref, newpositions, nnew*sizeof(Position));
    delete [] reference;
    delete [] refindex;
    reference = newreference;
    refindex  = newrefindex;
    nref += nnew;
}

void RemappingLookup::remap(const PixelValue * positions,
                            const int          npos,
                            const float        min_dist)
{
    int *new_index = new int [npos];
    int i;
    for (i = 0; i < npos; i++)
        new_index[i] = -1;
    remap(positions, new_index, npos, refindex, min_dist);
    delete [] new_index;
}
                            
/**
 * Remaps a list of objects to the reference.
 * The array <new_index> holds, for each object, the index of the reference
 * object it remaps to.
 * The array <ref_index> holds, for each reference object, the index of the new
 * object it remaps to.
 * The index value of -1 indicates a failure to find a matching object.
 */
void RemappingLookup::remap(const PixelValue * positions,
                            int              * new_index,
                            const int          npos,
                            int              * ref_index,
                            float              min_dist)
{
    /* for each reference position, keep track of how close the nearest
     * remapping object was
     */
    int   i, j;
    float min_dist2 = min_dist * min_dist;
    float *ref_nearest_distance = new float [nref];
    float *new_nearest_distance = new float [npos];
    for (i = 0; i < nref; i++)
        ref_index[i] = -1;
    for (i = 0; i < npos; i++)
        new_index[i] = -1;
    
    for (i = 0; i < nref; i++) 
        ref_nearest_distance[i] = 2.f * min_dist2;
    for (i = 0; i < npos; i++) 
        new_nearest_distance[i] = 2.f * min_dist2;

    // now try to remap each object
    for (j = 0; j < npos; j++)
        remapPosition(positions[j], &new_index[j], &new_nearest_distance[j],
                      min_dist, reference, nref, j, ref_index, ref_nearest_distance);

    delete [] ref_nearest_distance;
    delete [] new_nearest_distance;
}

/**
 * Remaps a single position to the reference lookup table.
 */
void RemappingLookup::remapPosition(const PixelValue & position,
                                    int              * new_index, 
                                    float            * new_nearest_distance,
                                    const float        min_dist,
                                    const Position   * reference,
                                    const int          nref,
                                    const int          pos,
                                    int              * ref_index,
                                    float            * ref_nearest_distance)
{
    /* for each reference position, keep track of how close the nearest
     * remapping object was
     */
    float min_dist2 = min_dist * min_dist;
    *new_index = -1;
    *new_nearest_distance = 2.f * min_dist2;

    float xpos = position.X; int ix = Stats::round(xpos);
    float ypos = position.Y; int iy = Stats::round(ypos);
    if ((ix < 0) || (ix >= nx) || (iy < 0) || (iy >= ny))
            return; /* object shifted out of image frame */

    int x, y;
    int xmin = std::max(0, (int)ceil(xpos-min_dist-0.5));
    int ymin = std::max(0, (int)ceil(ypos-min_dist-0.5));
    int xmax = std::min(nx-1, (int)floor(xpos+min_dist+0.5));
    int ymax = std::min(ny-1, (int)floor(ypos+min_dist+0.5));
    assert(xmin >= 0); assert(xmax < nx); assert(xmin <= xmax);
    assert(ymin >= 0); assert(ymax < ny); assert(ymin <= ymax);

    for (x = xmin; x <= xmax; x++)
        for (y = ymin; y <= ymax; y++) {
            int k = lookup[x*ny+y];
            if (k >= 0) {
                float xdist, ydist, dist2;
                xdist = xpos - reference[k].X; 
                ydist = ypos - reference[k].Y;
                dist2 = xdist*xdist + ydist*ydist;
                if (dist2 > min_dist2)
                    continue;
                if ((pos >= 0) && (k < nref))
                    if (dist2 < ref_nearest_distance[k]) {
                        ref_index[k] = pos;
                        ref_nearest_distance[k] = dist2;
                    }
                if (dist2 < *new_nearest_distance) {
                    *new_index = k;
                    *new_nearest_distance = dist2;
                }
            }
        }
    return;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
                           
/**
 * Given a list of reference positions and data points with objects,
 * the routine assembles lists of intensities and noise dispersions
 * for those objects that remap to the reference positions.
 *
 * @params refpositions   <nref>-dim. array of reference positions to be used
 *                        for remapping 
 * @params refindices     <nref>-dim. array of indices that point
 *                        to any remapped entries in the <data> array
 * @params nref           dimension of <referencepositions> array
 * @params data           <ndata>-dim array of objects detected in the current
 *                        colour
 * @params ndata          number of data objects
 * @params ncol           number of different colours observed
 * @params icol           the column in the output file (0-A, 1-C, 2-G, 3-T)
 * @params intensities    preallocated <nref>*<ncol>-dim array to hold
 *                        the intensities for each molecule in all colours
 * @paramssigma           the noise dispersions at the molecule postions
 *                        corresponding to the remapped molecules
 */
void remap_colour(const Position   * refpositions,
                  const int        * refindices,
                  const int          nref,
                  const PixelValue * data,
                  const int          ndata,
                  float            * intensities,
                  float            * sigma,
                  const int          ncol,
                  const int          icol)
{
    int i, k;
    // find remapping molecules
    for (i = 0; i < nref; i++) {
        k = refindices[i];
        if (k >= 0) {
            intensities[i*ncol+icol] = data[k].Intensity; 
            sigma[i*ncol+icol]       = data[k].Noise; 
        }
        else {
            intensities[i*ncol+icol] = 0.; 
            sigma[i*ncol+icol]       = 0.; 
        }
    }
}


/**
 * Completes missing intensity values where no cluster has been found.
 */
void complete_intensities(const Position           * refpositions,
                          const int                * refindices, 
                          const int                  nref,
                          const int                  ncol,
                          const int                  icol, 
                          const float              * image,
                          const float              * noise_image,
                          const int                  nx,
                          const int                  ny,
                          float                    * intensities,
                          float                    * noise_dispersions,
                          const float                fwhm,
                          const float                min_dist,
                          AlignmentTransform * transform)
{
    int i;
    for (i = 0; i < nref; i++)
        if (refindices[i] < 0) {
            intensities[i*ncol+icol]       = 0.;
            noise_dispersions[i*ncol+icol] = 0.;
            // TODO: rescale coordinates
            float xpos = refpositions[i].X; 
            float ypos = refpositions[i].Y; 
            transform->transformForward(&xpos, &ypos);
            int x = Stats::round(xpos);
            int y = Stats::round(ypos);
#if 0
            if ((x < 0) || (x >= nx) || (y < 0) || (y >= ny))
                continue;
#endif
            // TODO: think of a good estimator for the intensity and noise values
            int xmin = std::max(0, (int)ceil(xpos-min_dist-0.5));
            int ymin = std::max(0, (int)ceil(ypos-min_dist-0.5));
            int xmax = std::min(nx-1, (int)floor(xpos+min_dist+0.5));
            int ymax = std::min(ny-1, (int)floor(ypos+min_dist+0.5));
            if ((xmax < 0) || (xmin >= nx) || (ymax < 0) || (ymin >= ny)) {
                //printf("Skipping %i\n", i);
                continue;
            }
            assert(xmin >= 0); assert(xmin < nx);
            assert(xmax >= 0); assert(xmax < nx);
            assert(ymin >= 0); assert(ymin < ny);
            assert(ymax >= 0); assert(ymax < ny);
            intensities[i*ncol+icol]       = image[xmin*ny+ymin];
            noise_dispersions[i*ncol+icol] = noise_image[xmin*ny+ymin];
            for (x = xmin; x <= xmax; x++)
                for (y = ymin; y <= ymax; y++)
                    if (image[x*ny+y] > intensities[i*ncol+icol]) {
                        intensities[i*ncol+icol] = image[x*ny+y];
                        noise_dispersions[i*ncol+icol] = noise_image[x*ny+y];
                    }
        }
}
