/*
 * QC module for image analysis
 *
 * Time-stamp: <26 Jun 2006 16:50 BST QCImage.cpp km>
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#else
#include <cmath>
#endif
#include <cstdio>
#include <cassert>
#include <algorithm>
#include "Maths.h"
#include "Optimise.h"
#include "FFT.h"
#include "QCImage.h"

#define MAX(x,y) ((x) > (y) ? (x) : (y))

/* 
 * given a 2d-array holding the 2d-Fourier transform of an image,
 * return the 1-d power spectrum and associated errors on the estimate.
 * The 1-d power spectrum is obtained by averaging the power over an 
 * annulus in the 2-dimensional plane.
 * ARGUMENTS:
 *    nx, ny           image dimensions
 *    real             <nx>*<ny> dimensional array containing the real parts 
 *                     of the Fourier transform
 *    imag             <nx>*<ny> dimensional array with imaginary parts
 *    ucellx, ucelly   dimensions of a pixel in Fourier space
 *    wss              normalisation factor, usually to take into account any
 *                     window functions
 *    ustart, ucell    left border of first interval and bin size 
 *                     for power spectrum
 *    u, uerr          <nps> centres of u-bins the power spectrum is sampled at
 *                     and associated errors
 *    ps, pserr        <nps> power spectrum estimates and their errors
 *    nps              number of points for the power spectrum
 */
void power_spectrum_1d_fourier(float *real, float *imag, int nx, int ny,
                               float ucellx, float ucelly,
                               float wss, float ustart, float ucell,
                               float *u, float *uerr, float *ps,
                               float *pserr, int nps)
{
    int   i, j, l;

    float *sum = new float[nps]; float *sum2 = new float[nps];
    int *num = new int[nps];

    for (l = 0; l < nps; l++) {
        sum[l] = sum2[l] = 0.; num[l] = 0;
        u[l] = uerr[l] = ps[l] = pserr[l] = 0.;
    }


    /* 
     * average complex image values over an annulus of width <ucell>.
     * assume Fourier transform has been normalised such that the phase centre 
     * is at array position (0, 0)
     */
    float udist, value, ux2, uy2;
    for (i = 1; i < nx/2; i++) {
        ux2 = (ucellx*i) * (ucellx*i);
        for (j = 1; j < ny/2; j++) {
            uy2 = (ucelly*j) * (ucelly*j);
            udist = sqrt(ux2 + uy2);
            int idx = (int)floor((udist - ustart) / ucell);
            if ((idx >= 0) && (idx < nps)) {
        
                // need to add only one half of Fourier plane because
                // of its Hermitian property
                int mapidx = i*ny+j;
                value = real[mapidx] * real[mapidx] + imag[mapidx] * imag[mapidx];
                if (value > 0.) {
                    num[idx] += 1; sum[idx] += value; sum2[idx] += value * value;
                }
                mapidx = (nx-i)*ny+j;
                value = real[mapidx] * real[mapidx] + imag[mapidx] * imag[mapidx];
                if (value > 0.) {
                    num[idx] += 1; sum[idx] += value; sum2[idx] += value * value;
                }
#if 0
                mapidx = i*ny+(ny-j);
                value = real[mapidx] * real[mapidx] + imag[mapidx] * imag[mapidx];
                if (value > 0.) {
                    num[idx] += 1; sum[idx] += value; sum2[idx] += value * value;
                }
                mapidx = (nx-i)*ny+(ny-j);
                value = real[mapidx] * real[mapidx] + imag[mapidx] * imag[mapidx];
                if (value > 0.) {
                    num[idx] += 1; sum[idx] += value; sum2[idx] += value * value;
                }
#endif
            }
        }
    }

    /* correct for window function by rescaling normalisation */
    for (l = 0; l < nps; l++) {
        sum[l] /= wss; sum2[l] /= wss*wss;
    }

    /* 
     * from the averaged image values now calculate a power spectrum 
     * and estimate the mean error on the power spectrum 
     */
    for (l = 0; l < nps; l++) {
        u[l] =  (ustart + ucell * (l+0.5f)); /* u-value at bin centre */
        uerr[l] = 0.5f * ucell;
        if (num[l] > 0) {
            ps[l] = sum[l] / num[l];  /* estimate the mean of C_l */
            if (num[l] > 1) {
                pserr[l] =
                    sum2[l]/num[l] - (sum[l]/num[l]) * (sum[l]/num[l]);
                                /* variance of C_l */
                pserr[l] /= num[l] - 1.f;/* variance of the estimate of the mean */
            }

            /* now form the square root of C_l */
            ps[l] = sqrt(ps[l]);
            if (num[l] > 1)
                pserr[l] = 0.5f / ps[l] * sqrt(pserr[l]);
                    /* sdev(\sqrt(c_l) \approx |d\sqrt(c_l)/dC_l| sdev(C_l)*/
        }
#if 0
        printf("%i %f %f %f %f", l, u[l], uerr[l], ps[l], pserr[l]);
        printf(" %f %f %i\n", sum[l], sum2[l], num[l]);
#endif
    }

    delete[] sum; delete[] sum2;
    delete[] num;
    return;
}

/*
 * given a power spectrum of <n> points <p> with associated errors <perr>
 * sampled at positions <u>, return an estimate for the best-fit Gaussian
 * model (plus a constant offset which is assumed to correspond to white
 * image noise) that fits the part of the power spectrum above <usigma_start>.
 *
 * Given a poor initial estimate for <usigma_start>, the dispersion
 * of the Gaussian in Fourier space, the resulting new estimate
 * may also be very poor.
 */
float fit_power_spectrum(float *u, float *p, float *perr, int n,
                         float usigma_start)
{
    int i;
    float offset = 0.;
    float amplitude = 0.;

    int no = 5;
    int imin = 0, imax = n-1;
    /* find index corresponding to the smallest non-zero scale */
    while ((p[imax] <= 0.) && (imax > 0))
        imax--;
    /* find initial noise estimate (the constant component in the spectrum)
       by averaging over a number of bins on small scales */
    if (imax < no)
        no = imax;
    for (i = imax-no+1; i < imax+1; i++)
        offset += p[i];
    offset /= no;
    /* discard all power spectrum points below a wavelength corresponding
     * to the FWHM in Fourier space. This is to avoid being influenced by
     * the large-scale image structure, which contains little information about
     * the resolution or the point-spread function.
     */ 
    while (u[imin] < usigma_start*sqrt(2.*log(2.)))
        imin++;
    if (imax+1-imin < 4)
        return usigma_start;
    if (imin < no+1)
        no = imin-1;
    /* average over some large-scale bins go get a (very rough) estimate
     * of the normalisation of the Gaussian
     */
    for (i = 1; i < no+1; i++)
        amplitude += p[i];
    amplitude /= no;

    /* copy the data points within the interval picked above to a new set
     * of arrays
     */
    int nd = imax+1-imin;
    float *x = new float [nd]; float *y = new float [nd];
    float *s = new float [nd];
    for (i = 0; i < nd; i++) {
        x[i] = u[i+imin]; y[i] = p[i+imin];
        /* This choice of the noise is quite ad-hoc. The main idea is to give
         * more emphasis to large scales (which ought to be dominated by the
         * signal). The errors <perr> are actually smaller on small scales,
         * as more Fourier modes can contribute to the average and the sample
         * variance is smaller. However, this puts undue emphasis on the noise
         * part of the spectrum; also, any systematic rather than random errors
         * (e.g. the beam profile not being perfectly Gaussian) may throw off
         * the best fit.
         */
        s[i] = 1.f/sqrt(y[i]);
#ifdef DEBUG
        printf("%3i %f %f %f\n", i, x[i], y[i], s[i]);
#endif
    }

    /*
     * now use a Levenberg-Marquardt routine to fit a Gaussian plus a
     * constant to the power spectrum. Since the amplitude is poorly 
     * constrained, iterate over different amplitudes to avoid local
     * minima and find the best fit.
     */
    float sbest = usigma_start;
    float chi2best = 1.e20f;
    for (i = -1; i < 2; i++) {
        float a[5] = { amplitude * pow(10.f, i), 0.f, usigma_start, offset, 0.f };
        int ia[5] = { 1, 0, 1, 1, 1 };
        int na = 4;
        float chisq;
  
#ifdef DEBUG
        printf("params %f %f %f %f %f\n", a[0], a[1], a[2], a[3], a[4]);
#endif
        LM_1d(x, y, s, nd, a, ia, na, &chisq, LM_gauss_1d_der);
        if ((chisq < chi2best) && (a[0] > 0.) && (a[2] > 0.) && (a[3] > 0.)) {
            chi2best = chisq;
            sbest = a[2];
        } 
#ifdef DEBUG
        printf("params %f %f %f %f %f\n", a[0], a[1], a[2], a[3], a[4]);
#endif
    }
  
    delete [] x;
    delete [] y;
    delete [] s;
    return sbest;
}

/*
 * This routine estimates the width of the instrumental point-spread-function
 * assuming that it has a Gaussian shape. (In other words, it determines
 * the optical resolution of the image.)
 * The method chosen here may not be entirely obvious: The width of the
 * point spread function is estimated in Fourier space rather than in 
 * pixel space. 
 * The limited instrumental resolution can be modelled as a convolution 
 * of the image pixels with the point-spread function. Each object 
 * in the image is blurred, and one could try to estimate the width
 * of the point-spread function by trying to deconvolve each object separately.
 * A possible approach might be to measure the width for each object 
 * and average over them (or rather choose the minimum if the objects are
 * resolved).
 * The approach taken here exploits the fact that a convolution in pixel
 * space corresponds to a simple multiplication in Fourier space.
 * The 1-d power spectrum is used to estimate the shape of the Fourier
 * transform of the point spread function. For an additional 
 * white instrumental noise component on the image, 
 * the power spectrum will show a flat constant offset on all scales.
 * This constant offset is subtracted (and could potentially be used to
 * given an estimate of the instrumental noise). A Gaussian is then fitted
 * to the remaining signal. [The latter  will not necessarily be perfectly 
 * Gaussian-shaped even if the PSF itself is. This is due to sample variance
 * and the potentially non-flat power spectrum of the signal itself.
 * Furthermore, the resolution may vary somewhat across the image.
 * On small scales, Gaussianity is usually a good enough approximation.]
 */
float estimate_fwhm_fourier(float *real, float *imag, int nx, int ny,
                            float fwhm_start)
{
    /*
     * recall that the cell size in Fourier space is given by
     * 1./(N * delta), where N is the image dimension and <delta>
     * is the pixel dimension in real space. Since all sizes are measured
     * directly in pixels, delta is set to 1.
     */
    float ucellx = 1.f/(nx), ucelly = 1.f/(ny);
    /* now choose bin size for power spectrum. Only sample inside
     * Nyquist interval 
     */
    float ucell = 2.f * std::max(ucellx, ucelly);
    int nps = std::min(nx/2, ny/2)/2; 
    /* normalisation for window function. not used here */
    float wss = 1.f; 

    float *uarray = new float [nps]; float *psarray = new float [nps];
    float *uerror = new float [nps]; float *pserror = new float [nps];
    power_spectrum_1d_fourier(real, imag, nx, ny, ucellx, ucelly,
                              wss, 1.5f * ucell, ucell,
                              uarray, uerror, psarray, pserror, nps);

    /* convert FWHM to sigma */
    float sigma = fwhm_start / (2.f*sqrt(2.f*log(2.f))); 
    /* convert width to a Gaussian in pixel space to the equivalent
       width in Fourier space */
    float usigma = 1.f/(2.f*Maths::pi*sigma);

#ifdef DEBUG
    printf("usigma fwhm 1: %f %f %f\n", usigma, sigma, fwhm_start); 
#endif

    /* fit Gaussian model to the power spectrum */
    usigma = fit_power_spectrum(uarray, psarray, pserror, nps, usigma);
    sigma = 1.f/(2.f*Maths::pi*usigma);
    float fwhm = sigma * (2.f*sqrt(2.f*log(2.f)));

#ifdef DEBUG
    printf("usigma fwhm 2: %f %f %f %f\n", usigma, sigma, fwhm,
           usigma*(sqrt(2.*log(2.)))); 
#endif

    delete [] uarray; delete [] psarray;
    delete [] uerror; delete [] pserror;
    return fwhm;
}

float estimate_fwhm_fourier(float *fft, int nx, int ny,
                            float fwhm_start)
{
    float *real = new float [nx*ny];
    float *imag = new float [nx*ny];
    convert_fft2realimag(fft, nx, ny, real, imag, 1);
    float fwhm = estimate_fwhm_fourier(real, imag, nx, ny, fwhm_start);
    delete [] real;
    delete [] imag;
    return fwhm;
}

/*
 * Fourier transform an image and estimate its resolution in 
 * Fourier space
 */
float estimate_fwhm(float *image, int nx, int ny, float fwhm_start)
{
    float *fft = new_fourier_array(nx, ny);
    fft_2d_r2c(image, fft, nx, ny);
    float fwhm = estimate_fwhm_fourier(fft, nx, ny, fwhm_start);
    delete_fourier_array(fft);
    return fwhm;
}
