/* -*- tab-width: 4 -*- */

/*
 * Main program for image analysis module in the Genome Analyzer
 * data analysis pipeline.
 * Copyright (c) 2005, 2006 Solexa
 *               2007, 2008 Illumina
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#include <cstdio>
#include <cassert>
#include <string>
#include <algorithm>
#include "Images.h"
#include "TileAnalyser.h"
#include "Process.h"
#include "Tiling.h"
#include "Logger.h"
#include "QCWriter.h"
#include "JournalParser.h"
#include "ImageProcess.h"
#include "PositionData.h"
#include "IntensityData.h"
#include "ParameterTypes.h"
#include "Message.h"
#include "XMLSAXParser.h"
#include "XMLImageConfigHandler.h"
#include "ImageAnalysisConfig.h"

struct ExtendedAnalysisParameters : public AnalysisParameters {
public:
    int  numDetectionCycles;
    bool useRemappedIntensityFlag;
    bool useInterpolatedPositionFlag;

    ExtendedAnalysisParameters()
		{
			numDetectionCycles = 1;
			useRemappedIntensityFlag = true;
			useInterpolatedPositionFlag = false;
		}
};

int snprintf_offsets(char s[], int len, char text[],
                     std::vector<AlignmentTransform *> transforms)
{
    int i;
    int msg_idx = 0;
    msg_idx = sprintf(s, "%-25s", text); 
    for (i = 0; i < int(transforms.size()); i++)
        msg_idx += snprintf(s+msg_idx, len-msg_idx, "%6.2f %6.2f  ", 
                           transforms[i]->XOffset,
                           transforms[i]->YOffset);
    return msg_idx;
}

/**
 * Performs data analysis for a single tile and cycle.
 */
void analyseCycle(ImageList                 ** image_list,
                  unsigned char              * base_list,
                  const int                    cycle,
                  TileInfo                   & tileInfo,
                  RunInfo                    & runInfo,
                  ExtendedAnalysisParameters & params,
                  int                        * is_deblocked,
                  TileAnalyser               * analyser,
                  Logger                     * logger)
{
    int  j;
    char msg[256];

    // load image files 
    for (j = 0; j < runInfo.NumChannels; j++) {
        if (image_list[cycle*runInfo.NumChannels+j]->load()) {
            warning("Image load failed. Using empty image instead.");
            image_list[cycle*runInfo.NumChannels+j]->createEmptyImages(
                    runInfo.ImageWidth, runInfo.ImageHeight,
                    runInfo.NumTiles);
        }
        assert((int)image_list[cycle*runInfo.NumChannels+j]->size() 
                == runInfo.NumTiles);
    }
    tileInfo.Cycle       = cycle;
    tileInfo.IsDeblocked = is_deblocked[cycle];
    analyser->SetTile(tileInfo);
    for (j = 0; j < runInfo.NumChannels; j++) {
        // single-image analysis
        Image *image = image_list[cycle*runInfo.NumChannels+j]->at(0);
        // image alignment
        image->useInterpolatedPositionFlag = params.useInterpolatedPositionFlag;
        analyser->GetAnalyser()->setUseRemappedIntensityFlag(params.useRemappedIntensityFlag);
        analyser->AnalyseImage(image, j, (char)base_list[j]);
        sprintf(msg, "Image %i %04d %3i %i: %c #obj %5d offset %6.2f %6.2f %7.4f %7.4f exp %3.1f %3.1f",
                tileInfo.Lane+1, tileInfo.TileIndex+1, cycle+1, j+1,
                (char)(base_list[j]),
                image->getNumObjects(), 
                image->data->transform->XOffset,
                image->data->transform->YOffset,
                image->data->transform->XScale,
                image->data->transform->YScale,
                image->properties->ExposureTime,
                image->data->estimated_fwhm);
        logger->info(msg);
    }

    std::vector<AlignmentTransform * > transforms;
    transforms.clear();
    for (j = 0; j < runInfo.NumChannels; j++)
        transforms.push_back(image_list[cycle*runInfo.NumChannels+j]->at(0)
                                               ->getData()->transform);
    char s[125];
    snprintf(s, sizeof(s),
            "Tile  %i %04i %3i: Original offsets:     ",
            tileInfo.Lane+1, tileInfo.TileIndex+1, cycle + 1);
    snprintf_offsets(msg, sizeof(msg), s, transforms);
    logger->info(msg);
}


void run_analysis(ImageList         ** image_list,
                  unsigned char      * base_list,
                  int                  firstCycle,
                  int                  numCycles,
                  TileInfo           & tileInfo,
                  RunInfo            & runInfo,
                  ExtendedAnalysisParameters & params,
                  const char           path_qc_output[],
                  const char           path_image_analysis[],
                  const char           sample[],
                  std::string          logfile,
                  const char           offset_filename[],
                  const std::string  & loadsuffix,
                  const std::string  & savesuffix,
                  IntensityData::CompressionType compression,
                  int                * is_deblocked,
                  std::vector<int>   & matrix_cycles)
{
    assert(numCycles > 0);
    assert(runInfo.NumChannels > 0);
    assert(runInfo.NumChannels <= 4);
    assert(runInfo.NumTiles == 1);
    assert((params.GaussFactor >= 3.));
    assert((params.ClusterFwhm     > 1.) && (params.ClusterFwhm < 10.));
    assert((params.ClusterDistance > 1.) && (params.ClusterDistance < 10.));
    assert((params.PearsonThreshold > 0.) && (params.PearsonThreshold < 1.));

    /* set up logging object */
    Logger *logger = new Logger(logfile.c_str());
    logger->add_handler(new LogHandler());
    char msg[256];
    sprintf(msg, "%s%s", ImageAnalysisConfig::softwareName,
                         ImageAnalysisConfig::softwareVersion);
    logger->set_source(msg); 
    logger->info(path_image_analysis);
    logger->info(logfile.c_str());

    int i, j;
    i = 0;
    while ((i < runInfo.NumChannels) && (image_list[i]->load()))
        i++; // failed to load image
    if (i == runInfo.NumChannels) {
        logger->error("Failed to load any first-cycle image; exiting...\n");
        delete logger;
        error("Failed to load any first-cycle image; exiting...\n");
        exit(1);
    }

    runInfo.NumTiles    = image_list[i]->size();
    runInfo.ImageWidth  = image_list[i]->at(0)->nx;
    runInfo.ImageHeight = image_list[i]->at(0)->ny;

    sprintf(msg, "Lane %i TileIndex %i Tile %s_%i_%04i " 
            "Cycles %i Channels %i Tiles %i",
            tileInfo.Lane+1, tileInfo.TileIndex+1, sample,
            tileInfo.Lane+1, tileInfo.TileIndex+1,
            numCycles, runInfo.NumChannels, runInfo.NumTiles);
    logger->info(msg);
    if (compression == IntensityData::compressionGzip)
        logger->info("Compressing output files with gzip.");
    if (compression == IntensityData::compressionBzip2)
        logger->info("Compressing output files with bzip2.");

    if (params.DetectionType == 1)
    {
        sprintf(msg, "Tile  %i %04i: nx ny minmatch fwhm dist %i %i %4.2f %5.1f %5.1f",
                tileInfo.Lane+1, tileInfo.TileIndex+1,
                runInfo.ImageWidth, runInfo.ImageHeight,
                params.PearsonThreshold, params.ClusterFwhm,
                params.ClusterDistance);
    }
    else
    {
        sprintf(msg, "Tile  %i %04i: nx ny thresh fwhm dist %i %i %5.1f %5.1f %5.1f",
                tileInfo.Lane+1, tileInfo.TileIndex+1,
                runInfo.ImageWidth, runInfo.ImageHeight,
                params.GaussFactor, params.ClusterFwhm,
                params.ClusterDistance
		);
    }
    logger->info(msg);
    sprintf(msg, "Tile  %i %04i: #detCycles %i remap %i interp %i filterType %i ClusterC %5.1f ClusterGamma %5.1f minObjectPixels %i",
            tileInfo.Lane+1, tileInfo.TileIndex+1,
            params.numDetectionCycles, 
            (params.useRemappedIntensityFlag ? 1 : 0),
            (params.useInterpolatedPositionFlag ? 1 : 0),
            params.FilterType,
            params.ClusterC, params.ClusterGamma, params.minObjectPixels);
    logger->info(msg);

    // initialise objects for creation of filenames
    std::string path = std::string(path_qc_output) + std::string(sample);
    RunFolderFileName *qc_name = new RunFolderFileName(path.c_str());
    qc_name->setLane(tileInfo.Lane, tileInfo.TileIndex, runInfo.NumTiles);

    // initialise objects for creation of filenames
    path = std::string(path_image_analysis) + std::string(sample);
    RunFolderFileName *intensity_name = new RunFolderFileName(path.c_str());
    intensity_name->setLane(tileInfo.Lane, tileInfo.TileIndex, runInfo.NumTiles);

    TileAnalyser *analyser = new TileAnalyser(runInfo);
    analyser->SetParameters(params);
    Alignment *default_transforms = new Alignment [runInfo.NumChannels];
    if (AlignmentTransform::load(offset_filename, default_transforms,
                                 runInfo.NumChannels) < 0) {
        sprintf(msg, "Could not open offset file %s.", offset_filename);
        memset(default_transforms, 0, sizeof(Alignment)*runInfo.NumChannels);
    }
    else {
        sprintf(msg, "Successfully read offset file %s", offset_filename);
    }
    logger->info(msg);
    analyser->SetAlignment(default_transforms);
    delete [] default_transforms;

    std::vector<AlignmentTransform * >  transforms;
    std::vector<float *>                intensities;
    std::vector<float *>                noisevalues;
    Position                          * referencePositions = 0;
    int cycle;
    int numObjects = 0;

    // Read in existing data files. This is only reqired for extensions of 
    // previous analyses.
    if (loadsuffix.size() > 0) {
        int numObjectsTmp, numChannelsTmp;
        std::string filename = qc_name->getIndexFilename(RunFolderFileName::POSITION_FILE);
        sprintf(msg, "Reading %s", filename.c_str());
        logger->info(msg);
        PositionData posfile(filename.c_str());
        int status = posfile.read(&referencePositions, &numObjects);
        if (status) 
            error("Failed to read position information from %s.",
                  filename.c_str());

        filename = intensity_name->getIndexFilename(RunFolderFileName::INTENSITY_FILE) 
                   + loadsuffix;
        IntensityData intfile(filename.c_str(), compression);
        status = intfile.read(intensities, &numObjectsTmp, &numChannelsTmp);
        if (status == 0) {
            if (numObjects != numObjectsTmp) {
                error("Number of objects in file %s does not match.",
                      filename.c_str());
                exit(1);
            }
            if ((numChannelsTmp != runInfo.NumChannels) && (numChannelsTmp != 0)) {
                error("Number of channels in file %s does not match.",
                      filename.c_str());
                exit(1);
            }
        } 
        filename = intensity_name->getIndexFilename(RunFolderFileName::NOISE_FILE) 
                   + loadsuffix;
        IntensityData nsefile(filename.c_str(), compression);
        status = nsefile.read(noisevalues, &numObjectsTmp, &numChannelsTmp);
        if (status == 0) {
            if (numObjects != numObjectsTmp) {
                error("Number of objects in file %s does not match.",
                      filename.c_str());
                exit(1);
            }
            if ((numChannelsTmp != runInfo.NumChannels) && (numChannelsTmp != 0)) {
                error("Number of channels in file %s does not match.",
                      filename.c_str());
                exit(1);
            }
        } 
        firstCycle  = intensities.size();
        sprintf(msg, "Starting from cycle %i\n", firstCycle+1); 
    }

    /*
     * Build the template of cluster positions
     */
    if (firstCycle == 0) {
        int numTemplateCycles = std::min(params.numDetectionCycles, numCycles);
        if (numTemplateCycles < params.numDetectionCycles) {
            std::string msg = "Fewer cycles available than required for building the template.";
            warning(msg.c_str());
            logger->warning(msg.c_str());
        }
        sprintf(msg, "Tile  %i %04d: Building cluster template from %i cycles.",
                tileInfo.Lane+1, tileInfo.TileIndex+1, numTemplateCycles);
        logger->info(msg);
        Position                   * tempReferencePositions = 0;
        std::vector<PixelValue * >   tempPositions;
        std::vector<int>             tempCounts;
        tempPositions.resize(numTemplateCycles * runInfo.NumChannels);
        tempCounts.resize   (numTemplateCycles * runInfo.NumChannels);

        for (cycle = firstCycle; cycle < numTemplateCycles; cycle++) {
            if (cycle > 0)
                analyser->SetReference(tempReferencePositions, numObjects);
            analyseCycle(image_list, base_list, cycle,
                         tileInfo, runInfo, params, is_deblocked,
                         analyser, logger);

            /*
             * Generate templates for cluster positions:
             */
            analyser->GetAnalyser()->getObjectListsCopy(&tempPositions[cycle*runInfo.NumChannels],
                                                        &tempCounts[cycle*runInfo.NumChannels],
                                                        runInfo.NumChannels);
            // Temporary reference for second cycle
            if ((cycle == 0) && (cycle != numTemplateCycles - 1)) {
                analyser->GetAnalyser()->BuildReference(&tempPositions[0],
                                             &tempCounts[0],
                                             runInfo.NumChannels,
                                             &tempReferencePositions, &numObjects);
            }
            /*
             * Generate the reference positions that are used for all cycles
             * and write them to a text file.
             */
            if (cycle == numTemplateCycles - 1) {
                analyser->GetAnalyser()->BuildReference(&tempPositions[0], &tempCounts[0],
                                             runInfo.NumChannels * numTemplateCycles,
                                             &referencePositions, &numObjects);
                sprintf(msg, "Tile  %i %04d: Found %i clusters.",
                        tileInfo.Lane+1, tileInfo.TileIndex+1, numObjects);
                logger->info(msg);

                analyser->SavePositions(tileInfo, path_qc_output,
                                        numObjects, referencePositions);
                analyser->SaveIndexPositions(tileInfo, path_qc_output,
                                             numObjects, referencePositions);
            }
        }
        for (int i = 0; i < int(tempPositions.size()); i++)
            delete [] tempPositions[i];
        if (cycle > 0) {
            analyser->GetAnalyser()->ClearReference();
            delete [] tempReferencePositions;
        }
    }

    /*
     *  Now extract intensities for new cycles.
     */
    sprintf(msg, "Tile  %i %04d: Extracting intensities.",
            tileInfo.Lane+1, tileInfo.TileIndex+1);
    logger->info(msg);
    for (cycle = firstCycle; cycle < numCycles; cycle++) {
        analyser->SetReference(referencePositions, numObjects);
        analyseCycle(image_list, base_list, cycle,
                     tileInfo, runInfo, params, is_deblocked,
                     analyser, logger);

        float *signal = new float [ runInfo.NumChannels*numObjects ];
        float *noise  = new float [ runInfo.NumChannels*numObjects ];
        Alignment    *alignments  = new Alignment    [ runInfo.NumChannels ];
        TileQCValues *qcValues    = new TileQCValues [ runInfo.NumChannels ];
        analyser->CompleteTile(signal, noise,
                               referencePositions, numObjects, numObjects,
                               alignments, qcValues, 0);
        int valid_flag = qcValues[0].OffsetsValidated;
        delete [] alignments;
        delete [] qcValues;
        intensities.push_back(signal);
        noisevalues.push_back(noise);

        char s[125];
        snprintf(s, sizeof(s),
                "Tile  %i %04i %3i: Validated offsets (%i):",
                tileInfo.Lane+1, tileInfo.TileIndex+1, cycle + 1,
                valid_flag);
        snprintf_offsets(msg, sizeof(msg), s, transforms);
        logger->info(msg);

        /* write output */
        if (cycle == 0) {
            for (j = 0; j < runInfo.NumChannels; j++) {
                image_list[cycle*runInfo.NumChannels+j]->at(0)->data->analyseObjectStats();
            }
            analyser->SaveClusterDetails(tileInfo, path_qc_output,
                                         numObjects);
        }
        // save individual intensity/noise files per cycle
        // only if current cycle is specified in <matrix_cycles>
        int save_flag = 0;
        for (j = 0; j < (int)matrix_cycles.size(); j++)
            if (cycle == matrix_cycles[j])
                save_flag = 1;
        if (save_flag) {
            IntensityData intfile(qc_name->getIndexFilename(RunFolderFileName::INTENSITY_FILE, 
                                                         0, cycle).c_str());
            intfile.write(signal, numObjects, runInfo.NumChannels);
            IntensityData nsefile(qc_name->getIndexFilename(RunFolderFileName::NOISE_FILE, 
                                                         0, cycle).c_str());
            nsefile.write(noise, numObjects, runInfo.NumChannels);
        }
        analyser->SaveQCFile(tileInfo, path_qc_output);

        /* clear images to reduce memory footprint */            
        for (j = 0; j < runInfo.NumChannels; j++)
            image_list[cycle*runInfo.NumChannels+j]->at(0)->clear();
    }
    logger->info("Writing output files...");
    IntensityData nsefile(intensity_name->getIndexFilename(RunFolderFileName::NOISE_FILE).c_str(),
                         compression);
    nsefile.write(noisevalues, numObjects, runInfo.NumChannels,
                  intensity_name->getLane(0), intensity_name->getTile(0),
                  referencePositions);
    std::string filename = intensity_name->getIndexFilename(RunFolderFileName::INTENSITY_FILE) 
                           + savesuffix;
    IntensityData intfile(filename.c_str(), compression);
    intfile.write(intensities, numObjects, runInfo.NumChannels,
                  intensity_name->getLane(0), intensity_name->getTile(0),
                  referencePositions);
    for (cycle = 0; cycle < numCycles; cycle++) {
        delete [] intensities[cycle];
        delete [] noisevalues[cycle];
    }
    delete [] referencePositions;
    // write out combined intensity and signal files
    analyser->GetAnalyser()->displayOffsets();

    logger->info("Run terminated successfully.");
    delete analyser;
    delete logger;
    delete qc_name;
    delete intensity_name;
}

int read_xml_parameters(const char filename[],
                        struct AnalysisParameters &params)
{
    FILE *fh = fopen(filename, "r");
    if (!fh) {
        error("Failed to open parameter file %s.", filename);
        return 0;
    }

    XMLImageConfigHandler handler(&params);
    XMLSAXParser          sax_parser;
    sax_parser.setContentHandler(&handler);
    sax_parser.parse(fh);
    fclose(fh);
    return 0;
}

int main(int argc, char *argv[]) 
{
    std::string offset_filename    = "default_offsets.txt";
    std::string parameter_filename = "";
    std::string tmp_prefix         = "tempseq";
    std::string tmp_prefix2        = "tempseq";
    std::string journal_filename   = "";
    char        tmp_filename[FILENAME_MAX] = "";
    char  sample[FILENAME_MAX]          = "s";
    IntensityData::CompressionType compression = IntensityData::compressionNone;
    int   ncycles     = 0;
    int   firstcycle  = 0;
    int   i, idummy;
    ImageList     **image_list;
    std::vector<std::string > filenames;
    unsigned char  *base_list;

    char *deblockedstring     = NULL;
    std::string logfile       = "";
    std::string loadsuffix    = "";
    std::string savesuffix    = "";
    std::vector<int> matrix_cycles;

    TileInfo tileInfo;
    tileInfo.Lane             = 0;
    tileInfo.TileIndex         = 0;

    RunInfo runInfo;
    runInfo.NumChannels       = 4;
    runInfo.NumTiles          = 1;
    runInfo.ImageWidth        = ImageAnalysisConfig::numDefaultImagePixels;
    runInfo.ImageHeight       = ImageAnalysisConfig::numDefaultImagePixels;

    ExtendedAnalysisParameters params;

    for (i = 1; i < argc; i++) {
        if (argv[i][0] == '-')
            switch (argv[i][1]) {
            case 'c':
                switch (argv[i][2]) {
                case 'o':
                  {
                    char comp[20];
                    sscanf(argv[i], "-compression=%s", comp);
                    if (!strcmp(comp, "gzip"))
                        compression = IntensityData::compressionGzip;
                    if (!strcmp(comp, "bzip2"))
                        compression = IntensityData::compressionBzip2;
                  } break;
                 case 'l': sscanf(argv[i], "-clusterC=%f", &params.ClusterC); 
                           break;
                } break;								
            case 'f': 
                switch (argv[i][2]) {
                case 'w': sscanf(argv[i], "-fwhm=%f", &params.ClusterFwhm);
                          break;
                case 'i': sscanf(argv[i], "-filter=%d", &params.FilterType); 
                          break;
                case 'l': {
					int flag = 0;
					sscanf(argv[i], "-flag=%d", &flag);
					switch (flag) {
						case 0: params.useRemappedIntensityFlag = true;
							params.useInterpolatedPositionFlag = false; break;
						case 1: params.useRemappedIntensityFlag = false;
							params.useInterpolatedPositionFlag = false; break;
						case 2: params.useRemappedIntensityFlag = true;
							params.useInterpolatedPositionFlag = true; break;
						case 3: params.useRemappedIntensityFlag = false;
							params.useInterpolatedPositionFlag = true; break;
						}
					} break;
                } break;
            case 't': 
                switch (argv[i][2]) {
                case 'h': sscanf(argv[i], "-threshold=%f", &params.GaussFactor); 
                          break;
                case 'i': 
                    if (sscanf(argv[i], "-tile=%i,%i", 
                               &(tileInfo.Lane), &(tileInfo.TileIndex)) != 2)
                        if (sscanf(argv[i], "-tile=%[^_]_%d_%d",
                                   sample, &(tileInfo.Lane), 
                                   &(tileInfo.TileIndex)) == 3) {
                            tileInfo.Lane --; tileInfo.TileIndex--;
                        }
                    break;
                case 'y': sscanf(argv[i], "-type=%d", &params.FilterType); 
                    break;
                } break;
            case 'm': 
                switch (argv[i][2]) {
                case 'a':
                    switch (argv[i][3]) {
                    case 'x': sscanf(argv[i], "-maxshift=%f", 
                                     &params.MaxStageOffset); break;
                    case 't': sscanf(argv[i], "-matrix=%i", &idummy); 
                        matrix_cycles.push_back(idummy-1); break;
                    } break;
                case 'i':
                    switch (argv[i][4]) {
					case 'm':
						sscanf(argv[i], "-minmatch=%f", &params.PearsonThreshold); break;
					case 'o':
						sscanf(argv[i], "-minobjpix=%i", &params.minObjectPixels); break;
					} break;
                }
                break;
            case 'o': if (sscanf(argv[i], "-offsets=%s", tmp_filename))
                          offset_filename = tmp_filename;
                      break;
            case 'g': sscanf(argv[i], "-gamma=%f", &params.ClusterGamma); 
                      break;
            case 'd': 
                switch (argv[i][2]) {
                case 'i': sscanf(argv[i], "-dist=%f", &params.ClusterDistance); break;
                case 'e': deblockedstring = new char [ strlen(argv[i]) + 1];
                          sscanf(argv[i], "-deblocked=%s", deblockedstring);
                          break;
                } break;
            case 'p': 
                if (sscanf(argv[i], "-prefix=%s", tmp_filename) == 1) {
                    tmp_prefix = tmp_filename;
                    if (tmp_prefix2 == "tempseq")
                        tmp_prefix2 = tmp_filename;
                }
                if (sscanf(argv[i], "-prefix2=%s", tmp_filename))
                    tmp_prefix2 = tmp_filename;

                if (sscanf(argv[i], "-params=%s", tmp_filename)) {
                     parameter_filename = tmp_filename;
                }
                break;
            case 's': 
                switch (argv[i][2]) {
                case 'a': 
                    switch(argv[i][3]) {
                    case 'm': sscanf(argv[i], "-sample=%s", sample); break;
                    case 'v':
                        if (sscanf(argv[i], "-savesuffix=%s", tmp_filename))
                            savesuffix = tmp_filename;
                        break;
                    } break;
                } break;
            case 'n':
                switch (argv[i][2]) {
                case 'x': sscanf(argv[i], "-nx=%i", &(runInfo.ImageWidth));
                    break;
                case 'y': sscanf(argv[i], "-ny=%i", &(runInfo.ImageHeight));
                    break;
                case 'c': sscanf(argv[i], "-nc=%i", &firstcycle); break;
                case 'd': sscanf(argv[i], "-nd=%i", &params.numDetectionCycles); break;
                } break;
            case 'l': 
                switch (argv[i][2]) {
                case 'o':
                    switch (argv[i][3]) {
                    case 'g':
                        if (sscanf(argv[i], "-logfile=%s", tmp_filename))
                            logfile = tmp_filename;
                        break;
                    case 'a':
                        if (sscanf(argv[i], "-loadsuffix=%s", tmp_filename))
                            loadsuffix = tmp_filename;
                        break;
                    } break;
                } break;
            }   
        else {
            journal_filename = argv[i];
        }
    }
		
	// xml files overrides all cmdline opts
	read_xml_parameters(parameter_filename.c_str(), params);
		
    if ((argc < 2) || (journal_filename == "")) {
        printf("Usage: run_image_analysis [options] <journal filename>\n"
            "Options:\n"
			"  -type=<n>                    set detection algorithm type 0,1\n"
			"  -filter=<n>                  set filter type 0,1,2,3\n"
			"  -fwhm=<x>                    set instrumental FWHM\n"
			"  -gamma=<x>                   set power structure of noise. Only used when filter=2 or 3\n"
			"  -clusterC=<x>                set filter parameter. Only used when filter=2 or 3\n"		 
			"  -threshold=<x>               set 'Gauss factor'\n"
			"  -minmatch=<x>                set 'Pearson threshold'\n"
			"  -maxshift=<n>                set maximum shift for alignment (in pixels)\n"
			"  -dist=<x>                    set distance parameter for colocalisation\n"
			"  -params=<filename>           set the path to an XML parameters file\n"	
			"  -prefix=<filename>           set prefix for internal filenames\n"
			"  -logfile=<filename>          set the path to a log file\n"
			"  -sample=<text>               set sample id (must not contain underscores)\n"
			"  -tile=<lane>,<tile>          set lane/tile (0 indexed)\n"
			"  -nx=<n>                      set image width\n"
			"  -ny=<n>                      set image height\n"
			"  -nc=<n>                      set \"first cycle\"\n"
			"  -nd=<n>                      set number of detection cycles\n"
			"  -minobjpix=<n>               set smallest number of pixels required to be considered an object\n"
			"  -flag=<bitmask>              0=no interp, no remap, 3=interp and remap\n"
			"  -tile=<sample>_<lane>_<tile> set sample/lane/tile (1 indexed)\n");
				 
        exit(0);
    }

    // parse journal file (a description of which images to load)
    if (parse_journal(journal_filename.c_str(), &runInfo, &ncycles, filenames,
                      &base_list) < 0) {
        error("Cannot open journal file %s.", journal_filename.c_str());
        exit(1);
    }
    image_list = new ImageList * [ static_cast<int>(filenames.size()) ];
    for (i = 0; i < static_cast<int>(filenames.size()); i++) {
        image_list[i] = new ImageList;
        image_list[i]->addFile(filenames[i].c_str());
    }

    ncycles += firstcycle;

    // parse deblock string
    int *is_deblocked = new int [ncycles];
    for (i = 0; i < ncycles; i++)
        is_deblocked[i] = 0;
    if (deblockedstring) {
        for (i = 0; i < std::min(ncycles, (int)strlen(deblockedstring)); i++)
            if (toupper(deblockedstring[i]) == 'D')
                is_deblocked[i] = 1;
        delete [] deblockedstring;
    }

    // create name for log file
    if (logfile.length() == 0)
        logfile = std::string(tmp_prefix) + std::string(ImageAnalysisConfig::softwareName)
                + std::string(".log");

    // run the image analysis
    run_analysis(image_list, base_list, firstcycle, ncycles,
                 tileInfo,  runInfo, params,
                 tmp_prefix.c_str(), tmp_prefix2.c_str(), sample, logfile,
                 offset_filename.c_str(), loadsuffix, savesuffix, compression,
                 is_deblocked, matrix_cycles);

    delete [] base_list;
    for (i = 0; i < runInfo.NumChannels * (ncycles-firstcycle); i++)
        delete image_list[i];
    delete [] image_list;
    delete [] is_deblocked;
    return 0;
}
