/*
 * Copyright (c) Solexa   2005-2007
 *               Illumina 2008
 * Authors:      Klaus Maisinger
 *
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
  */

//
// CVS - $Id: callBases.cpp,v 1.2 2008/07/16 21:44:21 mzerara Exp $
// Main program for the base-caller ("Bustard").


#include <string>
#include <vector>
#include <cassert>
#include <cstdio>
#include <algorithm>

#include "DNACodes.h"
#include "LinAlg.h"
#include "Logger.h"
#include "Message.h"
#include "Tiling.h"
#include "XMLSAXParser.h"

#include "ImageXMLReader.h"
#include "IntensityData.h"
#include "ParameterTypes.h"
#include "PositionData.h"
#include "Tools.h"

#include "BaseCallConfig.h"
#include "BaseCallParamsXMLReader.h"
#include "CycleMatrix.h"
#include "CycleMatrixCorrector.hh"
#include "FrequencyMatrix.h"
#include "PhasingModel.h"
#include "PrbData.h"
#include "ProcessBaseCall.h"
#include "SequenceData.h"

/**
 * Parses the QC XML file for information on bases, exposure times and
 * deblocked cycles.
 */
int read_xml_info(const char      filename[],
                  unsigned char   bases[],
                  float         * exposure_times,
                  const int       numChannels,
                  int           * is_deblocked)
{
    int i;
    assert(numChannels <= BaseCallConfig::numBases);
    *is_deblocked = 0;
    for (i = 0; i < numChannels; i++) {
        exposure_times[i] = 0.;
        bases[i]          = BaseUpperCase[i];
    }
        
    FILE *fh = fopen(filename, "r");
    if (!fh)
        return 1;

    ImageXMLReader *handler    = new ImageXMLReader();
    XMLSAXParser   *sax_parser = new XMLSAXParser();
    sax_parser->setContentHandler(handler);
    sax_parser->parse(fh);
    fclose(fh);

    int errorCode = 0;

    int ncol2 = handler->getCount();
    if (ncol2 != numChannels) {
        error("XML file %s has only %i images.", filename, ncol2);
        errorCode = 2;
    }

    if ((ncol2 == numChannels) || (ncol2 == 0)) {
        for (i = 0; i < ncol2; i++) {
            bases[i]          = handler->getBases()[i];
            exposure_times[i] = handler->getExposureTimes()[i];
            *is_deblocked     = handler->getDeblocked()[0];
        }
    }
    delete sax_parser;
    delete handler;
    return errorCode;
}

/**
 * Parses the parameter XML file for program options.
 */
void read_xml_parameters(const char filename[],
                         struct BasecallParameters *params)
{
    FILE *fh = fopen(filename, "r");
    if (!fh) {
        error("Failed to open parameter file %s.", filename);
        return;
    }

    BaseCallParamsXMLReader *handler    = new BaseCallParamsXMLReader();
    XMLSAXParser            *sax_parser = new XMLSAXParser();
    sax_parser->setContentHandler(handler);
    sax_parser->parse(fh);
    fclose(fh);
    handler->getParameters(params);
    delete sax_parser;
    delete handler;
}

/**
 * Selects the appropriate matrix from a list of matrices, depending on
 * which read the current cycle belongs to. This is needed for data with
 * multiple reads.
 */
int getResponseMatrixIndex(const std::vector<FrequencyMatrix * > response_matrices,
                           const std::vector<int>                start_cycles, 
                           const int                             current_cycle)
{
    int i;
    int maximum = 0;
    int index   = 0;
    for (i = 0; i < int(start_cycles.size()); i++) {
        if ((current_cycle >= start_cycles[i]) && 
            (response_matrices[i] != 0) &&
            (start_cycles[i] >= maximum))
        {
            maximum = start_cycles[i];
            index = i;
        }
    }

    return index;
}

/**
 * Sets up a vector of frequency response matrices, one per imaging cycle.
 */
int loadMatrices(std::vector<float * >                 & matrices,
                 const std::vector<FrequencyMatrix * > & response_matrices,
                 const std::vector<int>                & start_cycles,
                 const int                               numChannels,
                 const int                               numCycles,
                 const unsigned char                   * ubases,
                 const std::vector<float * >           & exposure)
{
    char *bases = new char [numChannels];
    int i;
    for (i = 0; i < numChannels; i++)
        bases[i] = char(ubases[i]);
    int l;
    for (l = 0; l < numCycles; l++) {
        float *matrix = 0;
        int    index  = getResponseMatrixIndex(response_matrices, 
                                               start_cycles, l);
        FrequencyMatrix *response_matrix = response_matrices[index];
        if (response_matrix)
            matrix = response_matrix->getMatrix(bases, exposure[l]);
        else 
            matrix = identity_matrix(numChannels);
        matrices.push_back(matrix);
    }
    delete [] bases;
    return 0;
}                                      

/**
 * Logs the frequency response matrices.
 */
void logMatrices(const std::vector<float * > & matrices,
                 const int                     numChannels,
                 Logger                      * logger)
{
    for (int i = 0; i < int(matrices.size()); i++) {
        char message[250];
        float * matrix = matrices.at(i);
        if (matrix != 0) {
            int pos = sprintf(message, "Matrix: ");
            snprint_matrix(message+pos, 250-pos, matrix, numChannels, numChannels);
            logger->info(message);
        }
    }
}

void call(ProcessBaseCall                      * base_caller, 
          RunFolderFileName                    * intensity_name,
          RunFolderFileName                    * sequence_name,
          int                                    ncycles,
          const struct RunInfo                 & runInfo,
          const struct TileInfo                & tileInfo,
          Position                             * positions,
          int                                    numPositions,
          unsigned char                        * bases, 
          const std::vector<float * >          & exposure,
          const std::vector<FrequencyMatrix * > & response_matrices,
          const std::vector<int>               & start_matrix_cycles,
          Logger                               * logger,
          const std::string                    & savesuffix,
          IntensityData::CompressionType         compression,
          IntensityData::CompressionType         sig2compression,
          const bool ipar, const bool iterativeMatrix)
{
    int  i;
    char msg[256];
    int  nobj, ncol, errorno;
    std::string           filename;
    std::vector<float * > intensities;
    std::vector<float * > noise;
    filename = intensity_name->getIndexFilename(RunFolderFileName::INTENSITY_FILE, 0);
    snprintf(msg, sizeof(msg), "Reading intensity file %s.", filename.c_str());
    logger->info(msg);
    IntensityData intfile(filename.c_str(), compression, ipar);

    Position *positions2;
    if (positions != 0) { // pos.txt file was previously found
        errorno = intfile.read(intensities, &nobj, &ncol, 0);
        positions2 = positions;
    }
    else {  // no pos.txt file - read positions from the intensity file
        PositionIndex *intPositions;
        errorno = intfile.read(intensities, &nobj, &ncol, &intPositions);
        numPositions = nobj;
        positions2   = new Position [ numPositions ];
        int i;
        for (i = 0; i < numPositions; i++) {
            positions2[i].X = float(intPositions[i].X);
            positions2[i].Y = float(intPositions[i].Y);
        }
        delete [] intPositions;
    }
    if (errorno != 0) {
        error("Failed to read intensity file %s.", filename.c_str());
        exit(1);
    }
    if (numPositions != nobj) {
        error("Unexpected number of clusters in file %s.", filename.c_str());
        exit(1);
    }
    if ((int(intensities.size()) < ncycles) &&
           (int(intensities.size()) != 0)) {
        error("Unexpected number of cycles in file %s.", filename.c_str());
        exit(1);
    }

    filename = intensity_name->getIndexFilename(RunFolderFileName::NOISE_FILE, 0).c_str();
    snprintf(msg, sizeof(msg), "Reading noise     file %s.", filename.c_str());
    logger->info(msg);
    IntensityData nsefile(filename.c_str(), compression, ipar);
    errorno = nsefile.read(noise, &nobj, &ncol);
    if (errorno != 0) {
        error("Failed to read noise file %s.", filename.c_str());
        exit(1);
    }
    if (numPositions != nobj) {
        error("Unexpected number of clusters in file %s.", filename.c_str());
        exit(1);
    }
    if ((int(intensities.size()) < ncycles) &&
           (int(intensities.size()) != 0)) {
        error("Unexpected number of cycles in file %s.", filename.c_str());
        exit(1);
    }

    if (ncol == 0)
        ncol = runInfo.NumChannels;
    if (intensities.size() == 0)
        for (i = 0; i < ncycles; i++) {
            intensities.push_back(0);
            noise.push_back(0);
        }

    for (i = int(intensities.size()) - 1; i >= ncycles; i--) {
        delete [] intensities[i];
        delete [] noise[i];
        intensities.pop_back();
        noise.pop_back();
    }

    CycleMatrix matrices(ncol);
    loadMatrices(matrices, response_matrices, start_matrix_cycles,
                 ncol, ncycles, bases, exposure);
    logMatrices(matrices, ncol, logger);

    std::vector<float * > probabilities(ncycles);
    std::vector<float * > matrix_signal(ncycles);
    std::vector<float * > signal(ncycles);
    unsigned char *sequence = new unsigned char [nobj * ncycles];

    // determine maximum window of cycles needed for phasing correction
    int windowSize = base_caller->getWindowSize();
    sprintf(msg, "Window size: %i", base_caller->getWindowSize());
    logger->info(msg);

    int cycle;
    for (cycle = 0; cycle < ncycles; cycle++) {
        matrix_signal[cycle] = new float [nobj*ncol];
    }
    matrices.invert();
    sprintf(msg, "Starting to base-call tile %i %04i...",
            tileInfo.Lane+1, tileInfo.TileIndex+1);
    logger->info(msg);

    // apply matrix correction
    base_caller->validateIntensities(&(intensities[0]), &(noise[0]),
                                     nobj, ncol, ncycles);
    base_caller->applyMatrixCorrection(&(intensities[0]),
                                       matrices, &(matrix_signal[0]),
                                       nobj, ncol, ncycles);
    if (iterativeMatrix)
    {
        logger->info("iterative matrix correction\n");
        CycleMatrixCorrector iterativeMatrix(ncol);
        iterativeMatrix.estimate(matrix_signal, nobj, ncol);
        iterativeMatrix.invert();
        iterativeMatrix.scale(ncol);
        base_caller->applyMatrixCorrection(&(matrix_signal[0]),
                                           iterativeMatrix, &(matrix_signal[0]),
                                           nobj, ncol, ncycles);
    }
    else
    {
        logger->info("simple matrix correction\n");
    }

    for (cycle = 0; cycle < ncycles; cycle++)
        delete [] intensities[cycle];
    
    for (cycle = 0; cycle < ncycles; cycle++) {
#if 0
        int firstDataCycle = 0;
        int numDataCycles  = ncycles;
#else
        // IPAR usage pattern
        int firstDataCycle = std::max(0, cycle - windowSize/2);
        int numDataCycles  = std::min(windowSize, ncycles - firstDataCycle);
#endif
        probabilities[cycle] = new float [nobj*ncol];
        signal[cycle]        = new float [nobj*ncol];

        sprintf(msg, "Tile %i %04i %3i: Calling bases...",
                tileInfo.Lane+1, tileInfo.TileIndex+1, cycle+1);
        logger->info(msg);
        //<<<<<<< run_base_caller.cpp
        //base_caller->call(intensities, noise, matrices, nobj, ncol, cycle,
        //                  sequence, probabilities[cycle],
        //                   signal[cycle], matrix_signal[cycle]);

        //delete [] matrix_signal[cycle];
        //if (cycle - windowSize >= 0) {
        //    delete [] intensities[cycle-windowSize];
        //    delete [] noise[cycle-windowSize];
        //    intensities[cycle-windowSize] = 0;
        //    noise[cycle-windowSize] = 0;
        //}
        //=======
        base_caller->call(&(matrix_signal[firstDataCycle]),
                          &(noise        [firstDataCycle]),
                          matrices, nobj, ncol, cycle,
                          sequence, probabilities[cycle], signal[cycle],
                          numDataCycles, firstDataCycle);
        // release memory asap to reduce memory footprint
        int finishedCycle = cycle - windowSize/2;
        if (finishedCycle >= 0) {
             delete [] matrix_signal[finishedCycle];
             delete [] noise[finishedCycle];
             matrix_signal[finishedCycle] = 0;
             noise[finishedCycle] = 0;
         }
        //>>>>>>> 1.37.2.11
    }


    /* save corrected intensities and probabilities to file */
    errorno = 0;
    if (!errorno) {
        std::string filename = sequence_name->getIndexFilename(RunFolderFileName::SIGNAL2_FILE, 0);
        snprintf(msg, sizeof(msg), "Writing %s.", filename.c_str());
        logger->info(msg);
        IntensityData sigfile(filename.c_str(), sig2compression);
        errorno = sigfile.write(signal, nobj, ncol,
                                tileInfo.Lane, tileInfo.TileIndex,
                                positions);
    }
    if (!errorno) {
        std::string filename = sequence_name->getIndexFilename(RunFolderFileName::PROBABILITY_FILE, 0);
        snprintf(msg, sizeof(msg), "Writing %s.", filename.c_str());
        logger->info(msg);
        PrbData prbfile(filename.c_str());
        errorno = prbfile.write(probabilities, nobj, ncol);
    }
    if (!errorno) {
        std::string filename = sequence_name->getIndexFilename(RunFolderFileName::SEQUENCE_FILE, 0);
        snprintf(msg, sizeof(msg), "Writing %s.",
                 filename.c_str());
        logger->info(msg);
        filename += savesuffix;
        SequenceData seqfile(filename.c_str());
        errorno = seqfile.write(sequence, nobj, ncycles,
                                tileInfo.Lane, tileInfo.TileIndex,
                                positions);
    }
    if (errorno) {
        error("Cannot recover from IO error! Exiting...");
        perror("Error: ");
        exit(1);
    }

    if (positions == 0)
        delete [] positions2;
    delete [] sequence;
    int l;
    for (l = 0; l < ncycles; l++) {
        //<<<<<<< run_base_caller.cpp
        //if (intensities[l] != 0)
        //    delete [] intensities[l];
        //if (noise[l] != 0)
        //    delete [] noise[l];
        //delete [] matrices[l];
        //=======
        if (noise[l] != 0)
            delete [] noise[l];
        if (matrix_signal[l] != 0)
            delete [] matrix_signal[l];
        //>>>>>>> 1.37.2.11
        delete [] signal[l];
        delete [] probabilities[l];
    }
}

/*
 * set up and run the base-caller
 */
void run_base_caller(const char path_image_analysis[],
                     const char path_basecaller[],
                     const char path_intensities[],
                     const char sample[],
                     const struct RunInfo            & runInfo,
                     const struct TileInfo           & tileInfo,
                     int                               ncycles,
                     const std::vector<BasecallParameters> & params,
                     const std::vector<std::string>  & matrices,
                     const std::vector<int>          & starts_matrices,
                     const std::string               & logfile,
                     const std::string               & savesuffix,
                     const IntensityData::CompressionType compression,
                     const IntensityData::CompressionType sig2compression,
                     const bool ipar, const bool iterativeMatrix)
{
    assert(ncycles > 0);
    assert(runInfo.NumChannels > 0);
    assert(runInfo.NumChannels <= BaseCallConfig::numBases);

    /* set up logger */
    Logger *logger = new Logger(logfile.c_str());
    logger->add_handler(new LogHandler());
    char msg[256];
    sprintf(msg, "%s%s", BaseCallConfig::softwareName,
                         BaseCallConfig::softwareVersion);
    logger->set_source(msg); 
    sprintf(msg, "> `Bustard` - base-caller");
    logger->info(msg);
    sprintf(msg, "Base-caller path: %s", path_basecaller);
    logger->info(msg);
    sprintf(msg, "Log file: %s", logfile.c_str());
    logger->info(msg);
    if (compression == IntensityData::compressionGzip)
        logger->info("Using gzip compressed intensity files.");
    if (compression == IntensityData::compressionBzip2)
        logger->info("Using bzip2 compressed intensity files.");
    if (sig2compression == IntensityData::compressionGzip)
        logger->info("Using gzip compressed sig2 files.");
    if (sig2compression == IntensityData::compressionBzip2)
        logger->info("Using bzip2 compressed sig2 files.");


    // load frequency response matrix
    std::vector<FrequencyMatrix*> response_matrices;
    response_matrices.reserve(matrices.size());

    int i;
    for (i = 0; i < (int)matrices.size(); i++) {
        response_matrices.push_back(new FrequencyMatrix());
        int num = response_matrices.at(i)->read(matrices.at(i).c_str());
        if (num) {
            sprintf(msg, "Successfully loaded %ix%i matrix from %s", 
                    num, num, matrices.at(i).c_str());
        }
        else {
            sprintf(msg, "Could not find matrix file %s", matrices.at(i).c_str());
            delete response_matrices.at(i);
            response_matrices.at(i) = NULL;
        }
        logger->info(msg);
    }

    for (i = 0; i < static_cast<int>(params.size()); i++) {
        sprintf(msg, "Using phasing=%6.4f, prephasing=%6.4f from cycle %2i", 
                params[i].PhasingRate, params[i].PrephasingRate,
                params[i].FirstCycle + 1);
        logger->info(msg);
    }

    // initialise objects for creation of filenames
    std::string path = std::string(path_image_analysis) + std::string(sample);
    RunFolderFileName *sequencer_name = new RunFolderFileName(path.c_str());
    sequencer_name->setLane(tileInfo.Lane, tileInfo.TileIndex, runInfo.NumTiles);

    // initialise objects for creation of filenames
    path = std::string(path_basecaller) + std::string(sample);
    RunFolderFileName *base_caller_name = new RunFolderFileName(path.c_str());
    base_caller_name->setLane(tileInfo.Lane, tileInfo.TileIndex, runInfo.NumTiles);

    // initialise objects for creation of filenames
    path = std::string(path_intensities) + std::string(sample);
    RunFolderFileName *intensity_name = new RunFolderFileName(path.c_str());
    intensity_name->setLane(tileInfo.Lane, tileInfo.TileIndex, runInfo.NumTiles);

    ProcessBaseCall *base_caller = new ProcessBaseCall();
    base_caller->setCycleNumber(ncycles);
    PhasingModel phasing(params);
    base_caller->setPhasingModel(phasing);

    /* 
     * load intensity files and call base-caller
     */
    {
        unsigned char bases[4] = { 'A', 'C', 'G', 'T' };
        int tile  = 0;
        int ncol  = runInfo.NumChannels;
        int nread = 0;
        int i;

        /* read cluster positions */
        Position  * referencepositions;
        std::string positionFileName = sequencer_name->getIndexFilename(RunFolderFileName::POSITION_FILE,
                                                                        tile);
        PositionData posfile(positionFileName.c_str());
        if (posfile.read(& referencepositions, & nread)) {
            if (ipar) {
                error("Failed to read positions from %s.",
                      positionFileName.c_str());
                exit(1);
            }
            else
                warning("Failed to read positions from %s.",
                        positionFileName.c_str());
                referencepositions = 0;
        }

        /* read exposure times and block info from XML file */
        int           *is_deblocked = new int [ncycles];
        unsigned char *allbases     = new unsigned char [ runInfo.NumChannels * ncycles ];
        float         *allexposures = new float         [ runInfo.NumChannels * ncycles ];
        int cycle;
        for (cycle = 0; cycle < ncycles; cycle++) {
            is_deblocked[cycle] = 0;
            for (i = 0; i < runInfo.NumChannels; i++) {
                allbases    [cycle*runInfo.NumChannels+i] = bases[i];
                allexposures[cycle*runInfo.NumChannels+i] = 0.;
            }
            std::string xmlFileName = sequencer_name->getIndexFilename(RunFolderFileName::QC_FILE,
                                                                       tile, cycle);
            if (read_xml_info(xmlFileName.c_str(),
                              allbases+cycle*runInfo.NumChannels,
                              allexposures+cycle*runInfo.NumChannels,
                              runInfo.NumChannels, &is_deblocked[cycle]))
                warning("Error reading XML file %s.", xmlFileName.c_str());
        }

        sprintf(msg, "Intensity file: %s",
                intensity_name->getIndexFilename(RunFolderFileName::INTENSITY_FILE, tile).c_str());
        logger->info(msg);
        sprintf(msg, "Noise file:     %s",
                intensity_name->getIndexFilename(RunFolderFileName::NOISE_FILE, tile).c_str());
        logger->info(msg);

        std::vector <float * > exposures;
        for (cycle = 0; cycle < ncycles; cycle++) {
            float *exposure_times = new float [ runInfo.NumChannels ];

            /* read intensity values */
            for (i = 0; i < ncol; i++) {
                int baseidx = get_base_index(allbases[cycle*runInfo.NumChannels+i],
                                             bases, ncol);
                assert((baseidx >= 0) && (baseidx < runInfo.NumChannels));
                exposure_times[baseidx] = allexposures[cycle*runInfo.NumChannels+i]; 
            }
            int offset = sprintf(msg, "Exposure:");
            for (i = 0; i < runInfo.NumChannels; i++)
                offset += sprintf(msg+offset, "  %c %6.3f",
                                  BaseUpperCase[i], exposure_times[i]);
            logger->info(msg);
            exposures.push_back(exposure_times);
        }
        /* call base-caller */
        base_caller->setDeblockingOrder(is_deblocked, ncycles);
        call(base_caller, intensity_name, base_caller_name,
             ncycles, runInfo, tileInfo, referencepositions, nread,
             (unsigned char *)BaseUpperCase, exposures,
             response_matrices, starts_matrices, logger,
             savesuffix, compression, sig2compression, ipar, iterativeMatrix);

        for (cycle = 0; cycle < ncycles; cycle++)
            delete [] exposures[cycle];
        for (i = 0; i < int(response_matrices.size()); i++)
            if (response_matrices.at(i))
                delete response_matrices.at(i);
        delete [] allexposures;
        delete [] allbases;
        delete [] referencepositions;
        delete [] is_deblocked;
    }

    delete base_caller;
    delete logger;
    delete base_caller_name;
    delete sequencer_name;
    delete intensity_name;
}


int main(int argc, char *argv[]) 
{
    bool ipar = false;
    bool iterativeMatrix = false;
    std::string path_prefix                 = "tempseq";
    std::string path_prefix2                = "tempseq";
    std::string path_prefix3                = "../tempseq";
    std::string sample                      = "s";
    char        tmp_filename[FILENAME_MAX] = "../tempseq";
    int         ncycles = 0;
    int         i, restart;
    std::string logfile;
    IntensityData::CompressionType compression = IntensityData::compressionNone;
    IntensityData::CompressionType sig2compression = IntensityData::compressionNone;
    std::string savesuffix    = "";

    std::vector<int>         starts_matrix;  // cycles for new matrix (0 indexed)
    std::vector<std::string> matrices;

    std::vector<BasecallParameters> params;
    params.resize(1);

    TileInfo tileInfo;
    tileInfo.TileIndex    = 0;
    tileInfo.Lane         = 0;

    RunInfo runInfo;
    runInfo.NumTiles      = 1;
    runInfo.NumChannels   = BaseCallConfig::numBases;

    for (i = 1; i < argc; i++) {
        if (argv[i][0] == '-')
            switch (argv[i][1]) {
            case '-': // IPAR
                {
                    if (!strcmp(argv[i], "--IPAR"))
                    {
                        ipar = true;
                    }
                } break;
            case 'c':
                {
                    char comp[20];
                    sscanf(argv[i], "-compression=%s", comp);
                    if (!strcmp(comp, "gzip"))
                        compression = IntensityData::compressionGzip;
                    if (!strcmp(comp, "bzip2"))
                        compression = IntensityData::compressionBzip2;
                } break;
            case 'i':
                {
                    if (!strcmp(argv[i], "-iterative-matrix"))
                    {
                        iterativeMatrix = true;
                    }
                } break;
            case 't':
                switch (argv[i][2]) {
                case 'i':
                    if ((sscanf(argv[i], "-tile=%i,%i", 
                               &(tileInfo.Lane), &(tileInfo.TileIndex)) == 2)
                        || (sscanf(argv[i], "-tile=%[^_]_%d_%d",
                                   tmp_filename, &(tileInfo.Lane), 
                                   &(tileInfo.TileIndex)) == 3)) {
                        sample = tmp_filename;
                        tileInfo.Lane --; tileInfo.TileIndex --;
                    }
                    break;
                } break;
            case 'm':
                switch (argv[i][2]) {
                case 'a':
                    switch (argv[i][3]) {
                    case 't':
                        if (sscanf(argv[i], "-matrix=%s", tmp_filename)) {
                            matrices.push_back(tmp_filename);
                            starts_matrix.push_back(0);
                        }
                        break;
                    } 
                    break;
                }
                break;
            case 'p':
                if (sscanf(argv[i], "-phasing=%f",    &(params[0].PhasingRate)))
                    break;
                if (sscanf(argv[i], "-prephasing=%f", &(params[0].PrephasingRate)))
                    break;
                if (sscanf(argv[i], "-prefix=%s", tmp_filename) == 1) {
                    path_prefix = tmp_filename;
                    if (path_prefix2 == "tempseq")
                        path_prefix2 = tmp_filename;
                    break;
                }
                if (sscanf(argv[i], "-prefix2=%s", tmp_filename)) {
                    path_prefix2 = tmp_filename;
                    break;
                }
                if (sscanf(argv[i], "-prefix3=%s", tmp_filename)) {
                    path_prefix3 = tmp_filename;
                    break;
                }
                if (sscanf(argv[i], "-params=%s", tmp_filename)) {
                    read_xml_parameters(tmp_filename, &(params[0]));
                }
                break;
            case 's':
                switch (argv[i][2]) {
                case 'a':
                    switch(argv[i][3]) {
                    case 'm':
                        if (sscanf(argv[i], "-sample=%s", tmp_filename))
                            sample = tmp_filename;
                        break;
                    case 'v':
                        if (sscanf(argv[i], "-savesuffix=%s", tmp_filename))
                            savesuffix = tmp_filename;
                        break;
                    } break;
                 case 'i':
                    {
                    char sig2comp[20];
                    sscanf(argv[i], "-sig2compression=%s", sig2comp);
                    if (!strcmp(sig2comp, "gzip"))
                        sig2compression = IntensityData::compressionGzip;
                    if (!strcmp(sig2comp, "bzip2"))
                        sig2compression = IntensityData::compressionBzip2;
                    } break;                    
                } break;
            case 'n':
                switch (argv[i][2]) {
                case 'c': sscanf(argv[i], "-nc=%i", &ncycles); break;
                case 'e':
                    if (strlen(argv[i]) > 5) {
                        switch (argv[i][5]) {
                        case 'm':
                            if (sscanf(argv[i], "-new-matrix=%i:%[^:]",
                                       &restart, tmp_filename)) {
                                matrices.push_back(tmp_filename);
                                starts_matrix.push_back(restart-1);
                            }
                            break;
                       case 'p': 
                            if (sscanf(argv[i], "-new-params=%i:%[^:]",
                                       &restart, tmp_filename) == 2) {
                                BasecallParameters param;
                                read_xml_parameters(tmp_filename, &param);
                                param.FirstCycle = restart - 1;
                                params.push_back(param);
                            }
                            break;
                       } break;
                    }
                }
            case 'l':
                {
                    char path[FILENAME_MAX] = "";
                    sscanf(argv[i], "-logfile=%s", path);
                    logfile = path;
                } break;
            }
    }

    if ((argc < 2)) {
        printf("Usage: run_base_caller [options]\n"
               "Options:\n"
               "  -I\n     Run on an IPAR folder\n"
               "  -matrix=<filename>\n     set the path for matrix file (single matrix)\n"
               "  -new-matrix=<filename>,<cycle>\n     set the path for matrix file and the cycle at which\n     the matrix is applied (multiple matrices)\n"
               "  -new-params=<filename>,<cycle>\n     set cycle at which the phasing/prephasing effect is reset (1 indexed)\n"
               "  -phasing=<x>\n     set phasing rate (between 0.0 and 0.1)\n"
               "  -prephasing=<x>\n     set prephasing rate (between 0.0 and 0.1)\n"
               "  -params=<filename>\n     read options from parameter XML file\n"
               "  -prefix=<path>\n     set path for image analysis output files\n"
               "  -prefix2=<path>\n     set prefix for output files\n"
               "  -prefix3=<path>\n     set prefix for input intensity files\n"
               "  -sample=<text>\n     set sample id (must not contain underscores)\n"
               "  -tile=<lane>,<tile>\n     set lane/tile (1 indexed)\n"
               "  -tile=<sample>_<lane>_<tile>\n     set sample/lane/tile (1 indexed)\n"
               "  -nc=<n>\n     set number of cycles sequenced\n"
               "  -logfile=<filename>\n     set the path for log file\n"
               "  -iterative-matrix\n     use the iterative matrix\n");
        exit(0);
    }

    for (i = 0; i < static_cast<int>(params.size()); i++) {
        assert((params[i].PhasingRate >= 0.)    && (params[i].PhasingRate < 0.1));
        assert((params[i].PrephasingRate >= 0.) && (params[i].PrephasingRate < 0.1));
    }

    if (logfile.length() == 0)
        logfile = path_prefix2 + std::string(BaseCallConfig::softwareName)
                               + std::string(".log");

    run_base_caller(path_prefix.c_str(),  path_prefix2.c_str(),
                    path_prefix3.c_str(), sample.c_str(),
                    runInfo, tileInfo, ncycles,
                    params, matrices, starts_matrix,
                    logfile, savesuffix, compression, sig2compression, 
                    ipar, iterativeMatrix);

    return 0;
}
