/*
 * Main program for per-cycle intensity correlation calculation.
 * 
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

/*
 * Estimate the relative correlation between signals from clusters in
 * different sequencing cycles
 */
 
#include <vector>
#include <string>
#include <iostream>
#include <memory>
#include <getopt.h>

#include "MatrixEstimator.h"
#include "PhasingEstimator.h"
#include "IntensityData.h"
#include "CycleMatrix.h"
#include "FrequencyMatrix.h"
#include "LinAlg.h"
#include "Message.h"
#include "CommandLine.hh"

const int   defaultMinClusters =   500;
const int   PURE_BASES = 12;
const float CHASTITY   = 0.6;


class PhasingEstimateOptions : public CommandLine::Options
{
public:
    PhasingEstimateOptions()
        : CommandLine::Options()
        , iparFlag(false)
        , firstCycle(1)
        , compressionName("none")
        , compression(Compression(compressionName))
        , intensityFileName("")
        , matrixFileName("")
    {
    }
    bool iparFlag;
    int  firstCycle;
private:
    std::string compressionName;
public:
    Compression compression;
    std::string intensityFileName;
    std::string matrixFileName;
private:
    virtual int getOptions(int argc, char * const argv[])
        throw (CommandLine::UnknownOptionException, 
               CommandLine::MissingOptionException, 
               CommandLine::InvalidArgumentException);
    virtual void processOptions() throw (CommandLine::InvalidArgumentException);
    virtual void getParameters(int argc, char * const argv[]) throw (CommandLine::InvalidArgumentException);
    virtual int usage(int argc, char * const argv[], const std::string &message = "");
};

int estimatePhasing(const PhasingEstimateOptions &options);

int main(int argc, char *argv[]) 
{
    PhasingEstimateOptions options;
    if (CommandLine::Options::RUN == options.init(argc, argv))
    {
        const int returnValue = estimatePhasing(options);
    }
}

void applyMatrixCorrection(std::vector<float *> &cycledata, const std::string &matrixFileName,
                           const unsigned int numChannels, const unsigned int numClusters);


int estimatePhasing(const PhasingEstimateOptions &options)
{
    IntensityData input(options.intensityFileName.c_str(), options.compression, options.iparFlag);
    CycleData cycledata;
    int nclusters;
    int numChannels;
    input.read(cycledata, &nclusters, &numChannels, 0, 20 + options.firstCycle - 1, options.firstCycle); 

    if (nclusters < defaultMinClusters) 
    {
        warning("Not enough clusters in %s to estimate phasing.",
                options.intensityFileName.c_str());
        return 0;
    }
    
    /* try to read frequency cross-talk matrix */
    if (!options.matrixFileName.empty()) 
    {
        applyMatrixCorrection(cycledata, options.matrixFileName, numChannels, nclusters);
    }

    //CycleMatrix matrixdata;
    //matrixdata.estimate(cycledata, nclusters, numChannels);
    //matrixdata.print();
    //matrixdata.applyInverse(&(cycledata[0]), nclusters, numChannels,
    //                        int(cycledata.size()));

    float *phasingCycles, *prephasingCycles;
    std::auto_ptr<PhasingEstimator> estimator(new PhasingEstimator(numChannels));
    estimator->setIntensities(cycledata, nclusters);
    estimator->filterChastity(CHASTITY, PURE_BASES);
    info("PF %i", estimator->getClusterNumber());
    int code = estimator->estimateCycles(&phasingCycles, &prephasingCycles);

    if (code == PhasingEstimator::errorNone) 
    {
        estimator->printCycles(phasingCycles, prephasingCycles);
        delete [] phasingCycles;
        delete [] prephasingCycles;
    }
    else 
    {
        warning("%s (%s).", estimator->getErrorMessage(), options.intensityFileName.c_str());
    }
    return 0;
}

void applyMatrixCorrection(std::vector<float *> &cycledata, const std::string &matrixFileName,
                           const unsigned int numChannels, const unsigned int numClusters)
{
    float *matrix = 0;
    std::auto_ptr<FrequencyMatrix> fmatrix(new FrequencyMatrix());
    //int num = fmatrix->read(options.matrixFileName.c_str());
    if (0 != fmatrix->read(matrixFileName.c_str())) {
        matrix = fmatrix->getInverseMatrix();
    }
    else {
        warning("Could not find matrix file %s.", matrixFileName.c_str());
    }
    // apply matrix correction
    if (0 != matrix) 
    {
        float *array2 = new float [ numClusters * numChannels];
        for (std::vector<float *>::iterator cycle = cycledata.begin(); cycledata.end() > cycle; ++cycle)
        {
            float * const array = *cycle;
            matrix_vectortranspose_multiply(array2, matrix, array, numChannels, numClusters);
            *cycle = array2;
            array2 = array;
        }
        delete [] array2;
        delete [] matrix;
        matrix = 0;
    }
}


using namespace std;
using namespace CommandLine;

int PhasingEstimateOptions::getOptions(int argc, char * const argv[])
        throw (UnknownOptionException, MissingOptionException, InvalidArgumentException)
{
    const char *shortOptionList = "hIf:c:m:";
    const struct option longOptionList[] =
    {   {"help",                      no_argument,            0, 'h'}
    ,   {"IPAR",                      no_argument,            0, 'I'}
    ,   {"firstcycle",          required_argument,            0, 'f'}
    ,   {"compression",         required_argument,            0, 'c'}
    ,   {"matrix",              required_argument,            0, 'm'}
    ,   {0,       0,                 0,          0}
    };
    int longOptionIndex;
    int option;
    // Note: "case 'h'" returns immediately
    while (-1 != (option = getopt_long(argc, argv, shortOptionList, longOptionList, &longOptionIndex)))
    {
        switch (option)
        {
            case 'I':
                iparFlag = true;
                compressionName = "gzip";
                break;
            case 'f':
                firstCycle = Options::convert<unsigned int>(optarg);
                break;
            case 'c':
                compressionName = optarg;
                break;
            case 'm':
                matrixFileName = optarg;
                break;
            case 'h':
                return usage(argc, argv);
                break;
            case '?':
            case ':':
            default:
                throw UnknownOptionException("");
                break;
        };
    }
    return RUN;
}

void PhasingEstimateOptions::processOptions() throw (InvalidArgumentException)
{
    try
    {
        compression = Compression(compressionName);
    }
    catch(const UnsupportedCompressionException &e)
    {
        throw InvalidArgumentException(e.what());
    }
    
}

void PhasingEstimateOptions::getParameters(int argc,  char * const argv[]) throw (InvalidArgumentException)
{
    if (argc == optind + 1)
    {
        intensityFileName = argv[optind];
        ++optind;
    }
    else
    {
        throw InvalidArgumentException(string("The phasing estimate requires exactly one intensity file"));
    }
}

int PhasingEstimateOptions::usage(int argc, char * const argv[], const std::string &message)
{
    if (0 != message.size())
    {
        error(message.c_str());
    }
    const string program = argv[0];
    const string padding = string(program.length(), ' ');
    ostringstream os;
    os << argv[0] 
       << " [--help] [--IPAR] [--firstcycle <first cycle>] [--compression <compression>] [--matrix <file name>] <intensity file>" << endl
       << endl
       << "-h prints this help and exits" << endl
       << "-I IPAR data format. [false]" << endl
       << "-f discard cycles before the indicated first cycle. [1]" << endl
       << "-c set the compression type to 'none', 'gzip' or 'bzip2'. [none]" << endl
       << "-m set the name of the matrix file if matrix correction is rquired. ['']" << endl
       << endl
       << "Generate the phasing estimates for the given intensity file." <<endl
       << endl;
    info(os.str().c_str());
    return NO_RUN;
}

