// -*- mode: c++; indent-tabs-mode: nil; c-basic-offset: 4; -*-
/*****************************************************************************/
//
// Copyright (c) Illumina 2008
// Author: R. J. Shaw
// This source file is covered by the "Solexa Public Source License"
// agreement and bound by the terms therein.
//
/*****************************************************************************/
//
// Extracts a set of quality predictors for a phred-style quality calibration 
// from the run-folder. The predictors used are currently:
// - cycle number
// - 1. - chastity of current base
// - 1. - chastity of first 12 bases in read
// - -Qbustard
//
/*****************************************************************************/

#include <string>
#include <vector>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cassert>

#include "EQP_Options.h"
#include "IntensityData.h"
#include "Message.h"
#include "PrbData.h"
#include "SequenceData.h"
#include "QualityData.h"
#include "QPredictorTypes.h"
#include "QPredictorFunctions.h"

/*****************************************************************************/

enum {ERR_OK = 0, ERR_USAGE = 1, ERR_OUT_OF_MEMORY = 2, ERR_FILE_READ = 3,
      ERR_QUALITY_PREDICTORS = 4, ERR_FILE_WRITE = 5};

typedef std::vector<unsigned int> Cycle_Ind_Vec;
typedef Cycle_Ind_Vec::const_iterator Cycle_Ind_Vec_CIter;

typedef std::vector<float*> Sig_Ptr_Vec;
typedef Sig_Ptr_Vec::const_iterator Sig_Ptr_Vec_CIter;

typedef std::vector<float> Sig_Vec;
typedef Sig_Vec::const_iterator Sig_Vec_CIter;

typedef std::vector<float*> Prob_Ptr_Vec;
typedef Prob_Ptr_Vec::const_iterator Prob_Ptr_Vec_CIter;

typedef std::vector<float> Unchastity_Vec;
typedef Unchastity_Vec::iterator Unchastity_Vec_Iter;
typedef Unchastity_Vec::const_iterator Unchastity_Vec_CIter;

typedef std::vector<unsigned int> Len_Vec;
typedef Len_Vec::iterator Len_Vec_Iter;
typedef Len_Vec::const_iterator Len_Vec_CIter;



/*****************************************************************************/

bool extract_quality_predictors(const Sig_Ptr_Vec& sig_ptr_vec,
                                const Prob_Ptr_Vec& prob_ptr_vec,
                                const unsigned char* seq_ch_arr_ptr,
                                const PositionIndex* position_index_ptr,
                                int num_clusters, int num_channels,
                                long lo_cluster_ind,
                                long hi_cluster_ind,
                                unsigned int read_lo_cycle_num,
                                unsigned int read_hi_cycle_num,
                                unsigned int num_early_cycles,
                                const Quality_Data::QPTypes& qptypes,
                                QPBufferType& qp_buffer)
{
    const int curr_num_clusters(hi_cluster_ind - lo_cluster_ind + 1);
    const int 
        num_cycles_used(read_hi_cycle_num - read_lo_cycle_num + 1);

    const unsigned n_qp_lines(curr_num_clusters * num_cycles_used);
    const unsigned n_predictors(qptypes.size());

    qp_buffer.init(n_predictors,n_qp_lines);

    /// \todo check that high cycle doesn't go over sig2 range
    QPREDICTOR::FuncInitData idata(read_lo_cycle_num,
                                   read_hi_cycle_num,
                                   num_clusters,
                                   num_channels,
                                   num_early_cycles,
                                   sig_ptr_vec,
                                   prob_ptr_vec,
                                   (const char*) seq_ch_arr_ptr,
                                   position_index_ptr);

    QPREDICTOR::FuncSet qpfuncs(idata);

    unsigned qp_buffer_i(0); 
    for (long cluster_ind(lo_cluster_ind);
         cluster_ind <= hi_cluster_ind; ++cluster_ind) {

        for(unsigned i(0);i<n_predictors;++i){
            const QPREDICTOR::index_t qpi(qptypes[i]);
            if(qpfuncs.get_func_ptr(qpi) != 0){
                qpfuncs.get_func_ptr(qpi)->get_cluster_values(cluster_ind,qp_buffer.value[i].begin()+qp_buffer_i);
            }
        }

        const char* seq_ch_ptr = (const char*) (seq_ch_arr_ptr + cluster_ind);

        // Iterate over cycles.
        // [The sequence data has all 1st cycles together, then all 2nd, etc. 
        //  -> need to increase its ptr by num_clusters per cycle.]
        //
        for(unsigned i(read_lo_cycle_num-1);i<read_hi_cycle_num;++i){
            qp_buffer.called_base[qp_buffer_i] = *(seq_ch_ptr+i*num_clusters);
            ++qp_buffer_i;
        }
    }

    return true;
}

/*****************************************************************************/

int process_block_of_clusters(const Sig_Ptr_Vec& sig_ptr_vec,
                              const Prob_Ptr_Vec& prob_ptr_vec,
                              const unsigned char* seq_ch_arr_ptr,
                              const PositionIndex* position_index_ptr,
                              int num_clusters, int num_channels,
                              long lo_cluster_ind,
                              long hi_cluster_ind,
                              const std::string& read_msg_str,
                              unsigned int read_lo_cycle_num,
                              unsigned int read_hi_cycle_num,
                              const EQP_Options& eqp_options,
                              Quality_Data& quality_data)
{
    QPBufferType qp_buffer;

    if (!extract_quality_predictors(sig_ptr_vec, prob_ptr_vec,
                                    seq_ch_arr_ptr,
                                    position_index_ptr,
                                    num_clusters,
                                    num_channels,
                                    lo_cluster_ind,
                                    hi_cluster_ind,
                                    read_lo_cycle_num,
                                    read_hi_cycle_num,
                                    eqp_options.my_num_early_cycles,
                                    eqp_options.my_qptypes,
                                    qp_buffer)){
        error("Failed to extract quality predictors%s",
              read_msg_str.c_str());
        return ERR_QUALITY_PREDICTORS;
    }

     const int status = quality_data.write(qp_buffer);

    if (status != 0) {
        error("Failed to write quality predictors%s to `%s'",
              read_msg_str.c_str(), quality_data.file_path_str().c_str());
        return ERR_FILE_WRITE;
    }

    return ERR_OK;
}

/*****************************************************************************/

int process_clusters(const Sig_Ptr_Vec& sig_ptr_vec,
                     const Prob_Ptr_Vec& prob_ptr_vec,
                     const unsigned char* seq_ch_arr_ptr,
                     const PositionIndex* position_index_ptr,
                     int num_clusters, int num_channels,
                     const std::string& read_msg_str,
                     unsigned int read_lo_cycle_num,
                     unsigned int read_hi_cycle_num,
                     const EQP_Options& eqp_options,
                     Quality_Data& quality_data)
{
    int status(ERR_OK);
    static const long clusters_per_block(10000);

    long hi_cluster_ind(0);
    long lo_cluster_ind(0);
    int num_blocks
        = static_cast<int>((num_clusters - 1) / clusters_per_block + 1);

    for (int block_ind(0); block_ind < num_blocks; block_ind++) {
        lo_cluster_ind = block_ind * clusters_per_block;
        hi_cluster_ind = lo_cluster_ind + clusters_per_block - 1;

        if (hi_cluster_ind > num_clusters - 1) {
            hi_cluster_ind = num_clusters - 1;
        }

        status = process_block_of_clusters(sig_ptr_vec, prob_ptr_vec,
                                           seq_ch_arr_ptr,
                                           position_index_ptr,
                                           num_clusters, num_channels,
                                           lo_cluster_ind,
                                           hi_cluster_ind,
                                           read_msg_str,
                                           read_lo_cycle_num,
                                           read_hi_cycle_num,
                                           eqp_options,
                                           quality_data);

        if (status != ERR_OK) {
            break;
        }
    }

    return status;
}

/*****************************************************************************/

int main(int num_args, char* arg_cstr_arr[])
{
    const std::string app_name(arg_cstr_arr[0]);

    EQP_Options eqp_options;

    if (eqp_options.init(num_args, arg_cstr_arr)
        != CommandLine::Options::RUN) {
        return ERR_USAGE;
    }

    const unsigned int expected_num_cycles(eqp_options.total_num_cycles());

    IntensityData intensity_data(eqp_options.my_sig2_path_str.c_str(),
                                 eqp_options.my_sig2_compr);
    PrbData prb_data(eqp_options.my_prb_path_str.c_str());
    SequenceData seq_data(eqp_options.my_seq_path_str.c_str());

    Sig_Ptr_Vec sig_ptr_vec;
    int num_clusters(0);
    int num_chans(0);
    PositionIndex* position_index_ptr(0);

    int status = intensity_data.read(sig_ptr_vec, &num_clusters, &num_chans,
                                     &position_index_ptr);

    if (status != 0) {
        error("Failed to read intensity data from `%s'",
              eqp_options.my_sig2_path_str.c_str());
        return ERR_FILE_READ;
    }

    const int num_cycles(sig_ptr_vec.size());

    // Check value found if an expected value was specified (i.e. is non-zero)
    // BUT only if some clusters were actually found -> non-empty sig2 file.
    if ((num_clusters != 0)
        && (expected_num_cycles != 0)
        && ((unsigned int) num_cycles != expected_num_cycles)) {
        error("Number of cycles (%d) of intensity data found in `%s' "
              "does not match total number of cycles specified (%u)",
              num_cycles, eqp_options.my_sig2_path_str.c_str(),
              expected_num_cycles);
        return ERR_FILE_READ;
    }

    Prob_Ptr_Vec prob_ptr_vec;
    unsigned int num_prb_clusters(0);
    unsigned int num_prb_chans(0);

    status = prb_data.read(prob_ptr_vec, num_prb_clusters, num_prb_chans);

    if (status != 0) {
        error("Failed to read probability data from `%s'",
              eqp_options.my_prb_path_str.c_str());
        return ERR_FILE_READ;
    }

    if (num_prb_clusters != num_clusters) {
        error("Mismatch in cluster numbers between `%s' (%u) and `%s' (%u)",
              eqp_options.my_sig2_path_str.c_str(), num_clusters,
              eqp_options.my_prb_path_str.c_str(), num_prb_clusters);
        return ERR_FILE_READ;
    }

    // Check value found if an expected value was specified (i.e. is non-zero)
    // BUT only if some clusters were actually found -> non-empty prb file.
    if ((num_clusters != 0)
        && (expected_num_cycles != 0)
        && (prob_ptr_vec.size() != expected_num_cycles)) {
        error("Number of cycles (%u) of probability data found in `%s' "
              "does not match total number of cycles specified (%u)",
              prob_ptr_vec.size(), eqp_options.my_prb_path_str.c_str(),
              expected_num_cycles);
        return ERR_FILE_READ;
    }

    unsigned char* seq_ch_arr_ptr
        = new unsigned char[num_clusters * num_cycles];

    if (seq_ch_arr_ptr == 0) {
        error("Out of memory : sequence data");
        return ERR_OUT_OF_MEMORY;
    }


    // N.B. : Silently pads.
    status = seq_data.read(seq_ch_arr_ptr, num_clusters, num_cycles);

    if (status != 0) {
        error("Failed to read sequence data from `%s'",
              eqp_options.my_seq_path_str.c_str());
    }


    // Extract the quality predictors (possibly per read) and write them to
    // one or more qval files (one per read if multi-read).

    EQP_Options::File_Path_Str_List_CIter
        qval_path_str_citer(eqp_options.my_qval_path_str_list.begin());

    for (EQP_Options::Read_Num_Vec_CIter read_num_citer 
             = eqp_options.my_read_num_vec.begin();
         read_num_citer != eqp_options.my_read_num_vec.end();
         ++read_num_citer, ++qval_path_str_citer) {

        const unsigned int read_num(*read_num_citer);
        unsigned int read_lo_cycle_num(0);
        unsigned int read_hi_cycle_num(0);

        Quality_Data quality_data(*qval_path_str_citer,
                                  eqp_options.my_qptypes,
                                  eqp_options.my_qval_compr);

        std::string read_msg_str;

        if (read_num != 0) {
            std::ostringstream ostrm;
            ostrm << " for read " << read_num;
            read_msg_str = ostrm.str();
        }

        if (!eqp_options.get_read_cycle_range(read_num, read_lo_cycle_num,
                                              read_hi_cycle_num)) {
            error("Failed to get cycle range%s", read_msg_str.c_str());
        }

        if ((status = process_clusters(sig_ptr_vec,
                                       prob_ptr_vec,
                                       seq_ch_arr_ptr,
                                       position_index_ptr,
                                       num_clusters,
                                       num_chans,
                                       read_msg_str,
                                       read_lo_cycle_num,
                                       read_hi_cycle_num,
                                       eqp_options,
                                       quality_data)) != ERR_OK) {
            return status;
        }
    }


    /// \todo cts leaking sig2 (and maybe prb?) data. In main() you can obviously get
    /// away with it, but the file parsers need to start returning objects
    /// which clean themselves up.
    delete [] position_index_ptr;

    return ERR_OK;
}

/*****************************************************************************/

