/*****************************************************************************/
// Copyright (c) Illumina 2008
// Author: Richard Shaw
// This source file is covered by the Illumina Genome Analyzer Software
// License Agreement and bound by the terms therein.
/*****************************************************************************/

#include <string>
#include <iostream>

#include "Align_Data.h"
#include "Align_Data_Factory.h"
#include "Align_File_Data.h"
#include "Export_File_Data.h"
#include "Message.h"
#include "PAS_Options.h"
#include "Read_Align_Info.h"
#include "Tile_Mgr.h"
#include "Tile_Stats.h"
#include "Tiling.h"

/*****************************************************************************/

// Identifier to embed in source and binary files
static char* produce_align_stats_id = "$Id: produceAlignStats.cpp,v 1.5 2008/07/23 13:41:57 rshaw Exp $";

enum { ERR_OK=0, ERR_ARGS=1, ERR_READ_FILE=2, ERR_WRITE_FILE=3,
       ERR_PROCESSING=4 };

typedef std::list<Align_Data*> Align_Data_Ptr_List;
typedef Align_Data_Ptr_List::iterator Align_Data_Ptr_List_Iter;

typedef std::list<Tile_Stats::Quality_Type> Quality_Type_List;
typedef std::list<Tile_Stats::Quality_Type>::const_iterator
Quality_Type_List_CIter;

/*****************************************************************************/

void get_output_filename(const unsigned int lane_num,
                         const unsigned int tile_num,
                         const Tile_Stats::Quality_Type quality_type,
                         std::string& output_filename)
{
    RunFolderFileName run_folder_file_name("./s");
    RunFolderFileName::filename_type
        file_name_type(RunFolderFileName::SCORE_FILE);

    switch (quality_type) {
    case Tile_Stats::QUALITY_RAW: 
        file_name_type = RunFolderFileName::SCORE_FILE;
        break;
    case Tile_Stats::QUALITY_FILTERED:
        file_name_type = RunFolderFileName::RESCORE_FILE;
        break;
    }

    output_filename
    = run_folder_file_name.getTileFilename(file_name_type, lane_num - 1,
                                           tile_num - 1);
}

/*****************************************************************************/

bool process_one_tile(Align_Data_Ptr_List& align_data_ptr_list,
                      bool& data_remaining,
                      const Quality_Type_List& quality_type_list,
                      Tile_Mgr& tile_mgr,
                      bool input_type_is_export,
                      const std::string speced_output_filename)
{
    // Want to read all the data for one tile at the same time.
    // It may be spread over multiple reads and hence export files.
    // Then want to write it all out.

    data_remaining = false;
    const unsigned int num_reads(align_data_ptr_list.size());
    Tile_Stats tile_stats(num_reads);

    for (Align_Data_Ptr_List_Iter 
             align_data_ptr_iter(align_data_ptr_list.begin());
         align_data_ptr_iter != align_data_ptr_list.end();
         ++align_data_ptr_iter) {

        Read_Align_Info read_align_info;
        bool at_end_of_tile(false);
        bool at_eof(false);

        while ((*align_data_ptr_iter)->
               get_next_read_align_info(read_align_info,
                                        at_end_of_tile,
                                        at_eof)) {
            try {
                tile_stats.add_read_align_info(read_align_info);
            }
            catch (...) {
                error("Failed to add read align info to tile stats.");
                return false;
            }

            if (at_end_of_tile) {
                if (align_data_ptr_iter == align_data_ptr_list.begin()) {
                    // First file sets the consensus.
                    data_remaining = !at_eof;
                } else if (at_eof == data_remaining) {
                    error("Tile representation is not consistent across "
                          "reads.");
                    data_remaining = false;
                    return false;
                }

                break;
            }
        }
    }

    tile_stats.derive_stats();

    for (Quality_Type_List_CIter quality_type_citer(quality_type_list.begin());
         quality_type_citer != quality_type_list.end(); ++quality_type_citer) {
        Tile_Stats::Quality_Type quality_type(*quality_type_citer);
        std::string output_filename;

        if (input_type_is_export) {
            get_output_filename(tile_stats.lane_num(),
                                tile_stats.tile_num(),
                                quality_type,
                                output_filename);
        } else {
            output_filename = speced_output_filename;
        }

        File_Buffer file_buffer(output_filename,
                                File_Buffer::compressionNone);

        if (!tile_stats.write_score_file(quality_type, file_buffer)) {
            error("Failed to write stats for tile.");
        }
    }

    if (input_type_is_export) {
        if (!tile_mgr.report_tile(tile_stats.lane_num(),
                                  tile_stats.tile_num())) {
            return false;
        }
    }

    return true;
}

/*****************************************************************************/

bool write_bad_tile_files(const Tile_Mgr& tile_mgr,
                          const Quality_Type_List& quality_type_list,
                          const unsigned int num_reads)
{
    unsigned int lane_num(0);
    Tile_Mgr::Tile_Num_List missing_tile_num_list;

    tile_mgr.get_missing_tile_num_list(lane_num, missing_tile_num_list);

    for (Tile_Mgr::Tile_Num_List_CIter
             tile_num_citer(missing_tile_num_list.begin());
         tile_num_citer != missing_tile_num_list.end(); ++tile_num_citer) {
        for (Quality_Type_List_CIter
                 quality_type_citer(quality_type_list.begin());
             quality_type_citer != quality_type_list.end();
             ++quality_type_citer) {
            Tile_Stats::Quality_Type quality_type(*quality_type_citer);
            std::string output_filename;

            get_output_filename(lane_num, *tile_num_citer, quality_type,
                                output_filename);
            File_Buffer file_buffer(output_filename,
                                    File_Buffer::compressionNone);
            File_Buffer::Line_Vec line_vec;

            for (unsigned int read_num(1); read_num <= num_reads; ++read_num) {
                line_vec.push_back(Tile_Stats::hdr_str(lane_num,
                                                       *tile_num_citer,
                                                       quality_type,
                                                       read_num));
            }

            if (num_reads > 1) {
                // All Reads : read_num 0
                line_vec.push_back(Tile_Stats::hdr_str(lane_num,
                                                       *tile_num_citer,
                                                       quality_type, 0));
            }

            if (!file_buffer.write_some_lines(line_vec)) {
                error("Failed to write statistics set to %s",
                      file_buffer.file_path_str().c_str());
                return false;
            }
        }
    }

    return true;
}

/*****************************************************************************/

int main(int num_args, char* arg_cstr_arr[])
{
    const std::string app_name(arg_cstr_arr[0]);
    int ret_val(ERR_OK);


    // Process command line.

    PAS_Options pas_options;

    if (pas_options.init(num_args, arg_cstr_arr)
        != CommandLine::Options::RUN) {
        return ERR_ARGS;
    }

    Tile_Mgr tile_mgr(pas_options.my_tile_list_file_path_str);
    const bool input_type_is_export(pas_options.my_input_type
                                    == Align_Data_Factory::EXPORT_INPUT);


    // Create a list of data ptrs from the file list (e.g. one per read).

    Align_Data_Factory align_data_factory;
    Align_Data_Ptr_List align_data_ptr_list;
    const unsigned int num_reads(pas_options.my_input_path_str_list.size());

    for (PAS_Options::Path_Str_List_CIter
             input_path_str_citer(pas_options.my_input_path_str_list.begin());
         input_path_str_citer != pas_options.my_input_path_str_list.end();
         ++input_path_str_citer) {

        Align_Data* align_data_ptr
            = align_data_factory.get_align_data(pas_options.my_input_type,
                                                *input_path_str_citer,
                                                pas_options.my_input_compr);
        align_data_ptr_list.push_back(align_data_ptr);
    }


    // Process the tiles one at a time (combining reads if necessary).

    bool data_remaining(true);
    Quality_Type_List quality_type_list;
    quality_type_list.push_back(Tile_Stats::QUALITY_RAW);

    if (input_type_is_export) {
        quality_type_list.push_back(Tile_Stats::QUALITY_FILTERED);
    }

    while (data_remaining) {
        if (!process_one_tile(align_data_ptr_list, data_remaining,
                              quality_type_list, tile_mgr,
                              input_type_is_export,
                              pas_options.my_speced_score_path_str)) {
            ret_val = ERR_PROCESSING;
            break;
        }
    }


    // Write null files for each unrepresented tile (export file input only).

    if (input_type_is_export) {
        if (!write_bad_tile_files(tile_mgr, quality_type_list, num_reads)) {
            ret_val = ERR_PROCESSING;
        }
    }

    return ret_val;
}

/*****************************************************************************/

