/*
 * Copyright (c) Solexa 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */
#ifndef BINNEDTABLE_H_
#define BINNEDTABLE_H_

#include <vector>
#include <cassert>
#include <cstdio>

/*
 * This container is used to hold a multi-dimensional binned table
 * and translates the indices from the multiple dimensions to one 
 * single one-dimensional vector. It also stores the bin boundaries.
 */
template <class T>
class BinnedTable
{
public:
    std::vector<std::vector<float> > binnings;
    std::vector<T> table;
    virtual int getBinIndex(std::vector<float>params)
    {
        std::vector<int> indices = getBinIndices(params);
        return getBinIndex(indices);
    }
    virtual int getBinIndex(std::vector<int>indices)
    {
        int idx = 0;
        int i;
        for (i = indices.size()-1; i >= 0; i--)
            idx = idx * (binnings[i].size()-1) + indices[i];
        return idx; 
    }
    virtual std::vector<int> getBinIndices(std::vector<float>params)
    {
        assert((binnings.size() == params.size()));
        std::vector<int> result(binnings.size());
        int i;
        for (i = 0; i < (int)(params.size()); i++) {
            float value = params[i];
            int idx = 0;
            while ((idx+1 < (int)(binnings[i].size())) && (value > binnings[i][idx+1])) {
               idx ++;
            }
            if (idx >= (int)(binnings[i].size()) - 1)
                idx = binnings[i].size() - 2;
            result[i] = idx;
        }
        return result;
    }
    virtual std::vector<int> getBinIndices(int binidx)
    {
        std::vector<int> result(binnings.size());
        assert((binidx < (int)(table.size())));
        int i;
        int idx = binidx;
        for (i = 0; i < (int)(binnings.size()); i++) {
            int size = binnings[i].size()-1;
            int idx2 = idx % size;
            idx /= size;
            result[i] = idx2;
        }
        return result;
    }
    virtual std::vector<float> getBinParams(int binidx)
    {
        std::vector<int>   indices = getBinIndices(binidx);
        return getBinParams(indices);
    }
    virtual std::vector<float> getBinParams(std::vector<int> indices)
    {
        std::vector<float> params(binnings.size());
        int i;
        for (i = 0; i < (int)(binnings.size()); i++) {
            params[i] = binnings[i][indices[i]+1];
        }
        return params;
    }
public:
    BinnedTable() {};
    virtual ~BinnedTable() {};
    virtual int size() { return table.size(); }
    virtual void generateTable()
    {
        int size = 1;
        int i;
        for (i = 0; i < (int)(binnings.size()); i++) {
            size *= (binnings[i].size() - 1);
        }
        table.clear();
        table.resize(size);
    }
    virtual int addParameter(std::vector<float> binning)
    {
        assert(table.size() == 0);
        binnings.push_back(binning);
        int i;
#if 1
        fprintf(stderr, "Added parameter %i:", (int)(binnings.size()));
        for (i = 0; i < (int)(binning.size()); i++)
            fprintf(stderr, " %f", binning[i]);
        fprintf(stderr, "\n");
#endif
        return 0;
    }
};

#endif /*BINNEDTABLE_H_*/
