/*****************************************************************************/
// Copyright (c) Illumina 2007
// Author: Richard Shaw
// This source file is covered by the "Solexa Public Source License"
// agreement and bound by the terms therein.
/*****************************************************************************/

#include <sstream>

#include <cstdlib>

#include "Message.h"

#include "DataSetFactory.h"

/*****************************************************************************/

Data_Set_Factory::Data_Set_Factory(const std::string& qval_file_path_str,
                                   File_Buffer::Compression_Type
                                   qval_file_compr,
                                   const std::string& qref_file_path_str,
                                   File_Buffer::Compression_Type
                                   qref_file_compr)
    : my_qval_file_buffer(qval_file_path_str, qval_file_compr),
      my_qref_file_buffer(qref_file_path_str, qref_file_compr),
      my_need_qref_file_flag(!qref_file_path_str.empty())
{
    ;
}

/*****************************************************************************/

Data_Set_Factory::~Data_Set_Factory()
{
    ;
}

/*****************************************************************************/
// Read a DataSet from the file specified at construction. 
// See comments in the header file and the documentation for the data format.

bool Data_Set_Factory::get_next_data_set(DataSet*& data_set_ptr, bool& at_eof)
{
    bool ret_val(true);
    at_eof = false;

    data_set_ptr = new DataSet; // Caller must delete.
    const unsigned int max_num_records_per_data_set(DataSet::MAX_NUM_RECORDS);
    const unsigned int alloc_size(DataSet::ALLOC_SIZE);
    unsigned int num_records(0);
    unsigned int line_num(0);

    int num_params(0);
    BaseAlignment read;

    // If we do not need the actual qref base, the value will stay as `.'.
    read.base_aligned = '.';

    std::string qval_line;
    
    bool qval_at_eof(false);
    bool qref_at_eof(false);

    while ((ret_val = my_qval_file_buffer.read_line(qval_line, qval_at_eof))) {
        ++line_num;

        if ((qval_line.size() == 0)
            || (qval_line[0] == '#') || (qval_line[0]=='>')) {

            if (qval_at_eof) {
                break;
            } else {
                continue;
            }
        }

        if (my_need_qref_file_flag) {
            if (!read_ref_base(read.base_aligned, qref_at_eof)) {
                error("Failed to read base from `%s' "
                      "corresponding to `%s' in `%s'",
                      my_qref_file_buffer.file_path_str().c_str(),
                      qval_line.c_str(),
                      my_qval_file_buffer.file_path_str().c_str());
                return false;
            }
        }

        // Determine number of columns.
        if (num_params == 0) {
            std::stringstream ss(qval_line);
            std::string s;

            while (ss >> s)
                ++num_params;

            num_params -= 1;  // last column is called base
            data_set_ptr->values.resize(num_params);
        }

        // Reserve some memory in advance.
        if ((num_records % alloc_size) == 0) {
            for (unsigned int i(0); i < data_set_ptr->values.size(); i++)
                data_set_ptr->values[i]
                    .reserve(((num_records / alloc_size) + 1) * alloc_size);
        }

        // Parse line and store its fields.
        std::stringstream ss(qval_line);
        std::vector<std::string> fields;
        std::string s;

        while (ss >> s) {
            fields.push_back(s);
        }

        if ((int)(fields.size()) != (num_params + 1)) {
            error("Invalid record (num %u at line %u): `%s'",
                  num_records, line_num, qval_line.c_str());
            
            return false;
        }

        for (int i(0); i < num_params; i++) {
            const float value(atof(fields[i].c_str()));
            data_set_ptr->values[i].push_back(value);
        }

        read.base_read = fields[fields.size() - 1].c_str()[0];
        data_set_ptr->reads.push_back(read);

        if (++num_records == max_num_records_per_data_set) {
            // The DataSet is `full', so return it.
            break;
        }

        if (qval_at_eof || qref_at_eof) {
            // Just break out of the loop if we hit the end of a file.
            // Sort out the other details outside the loop.
            break;
        }
    }

    // Fix any false mismatch of file ends by attempting a final
    // read from the file that appears not to have ended
    // - but only if both files are actually being used.
    if (my_need_qref_file_flag) {
        if (qref_at_eof && !qval_at_eof) {
            my_qval_file_buffer.read_line(qval_line, qval_at_eof);
        } else if (!qref_at_eof && qval_at_eof) {
            read_ref_base(read.base_aligned, qref_at_eof);
        }
    }

    // Returning false indicates an error only if eof is not set.
    if (!(ret_val || qval_at_eof)) {
        error("Read from `%s' failed",
              my_qval_file_buffer.file_path_str().c_str());
        return false;
    }

    // If two files are being used, has one ended before the other?
    if (my_need_qref_file_flag && (qval_at_eof != qref_at_eof)) {
        
        // FIXME : Check for permissible junk at the end of the file
        //         that has not reached EOF.
        error("Ran out of matching lines in `%s' for lines in `%s'",
              (qval_at_eof
               ? my_qval_file_buffer.file_path_str().c_str()
               : my_qref_file_buffer.file_path_str().c_str()),
              (qval_at_eof
               ? my_qref_file_buffer.file_path_str().c_str()
               : my_qval_file_buffer.file_path_str().c_str()));
        return false;
    }

    at_eof = qval_at_eof;
    return true;
}

/*****************************************************************************/

bool Data_Set_Factory::read_ref_base(char& ref_base, bool& at_eof)
{
    bool ret_val(false);
    std::string qref_line;

    while (my_qref_file_buffer.read_line(qref_line, at_eof)) {
        if ((qref_line.size() == 0)
            || (qref_line[0] == '#') || (qref_line[0]=='>')) {
            continue;
        }

        ref_base = qref_line[0];
        ret_val = true;
        break;
    }

    return ret_val;
}

/*****************************************************************************/

