// -*- mode: c++; indent-tabs-mode: nil; c-basic-offset: 4; -*-
/*****************************************************************************/
// Copyright (c) Illumina 2008
// Author: Richard Shaw
// This source file is covered by the Illumina Genome Analyzer Software 
// License Agreement and bound by the terms therein.
/*****************************************************************************/

#include <string>
#include <map>
#include <sstream>

#include <getopt.h>

#include "File_Buffer.h"
#include "Message.h"
#include "StringUtil.h"

#include "EQP_Options.h"
#include "QPredictorTypes.h"

using namespace CommandLine;

using namespace std;

/*****************************************************************************/

EQP_Options::EQP_Options()
    : CommandLine::Options(),
      my_num_early_cycles(my_default_num_early_cycles),
      my_multi_read_flag(false),
      my_sig2_compr_name("none"),
      my_sig2_compr(IntensityData::compressionNone),
      my_qval_compr_name("none"),
      my_qval_compr(File_Buffer::compressionNone)
{
    my_default_qptypes.push_back(QPREDICTOR::UNPURITY_NHOOD1);
    my_default_qptypes.push_back(QPREDICTOR::SHIFTED_UNPURITY);
    my_default_qptypes.push_back(QPREDICTOR::SHIFTED_SUM_DIFF_UNCHASTITY);
}

/*****************************************************************************/

bool EQP_Options::get_read_cycle_range(unsigned int read_num,
                                       unsigned int& read_lo_cycle_num,
                                       unsigned int& read_hi_cycle_num)
{
    read_lo_cycle_num = 1;

    if (read_num == 0) {
        // Not considering reads separately, so use all cycles.

        read_hi_cycle_num = 0;

        for (Orig_Read_Lens_List_CIter
                 orig_read_len_citer(my_orig_read_lens_list.begin());
             orig_read_len_citer != my_orig_read_lens_list.end();
             ++orig_read_len_citer) {
            read_hi_cycle_num += *orig_read_len_citer;
        }

        return true;
    }

    if (read_num > my_orig_read_lens_list.size()) {
        return false;
    }

    Orig_Read_Lens_List_CIter
        orig_read_len_citer(my_orig_read_lens_list.begin());
    read_hi_cycle_num = *orig_read_len_citer++;

    for (unsigned int tmp_read_num(1); tmp_read_num < read_num;
         ++tmp_read_num) {
        read_lo_cycle_num = read_hi_cycle_num + 1;
        read_hi_cycle_num += *orig_read_len_citer++;
    }

    return true;
}

/*****************************************************************************/
// Simply uses the get_read_cycle_range method with read_num 0,
// i.e. no division into separate reads.

unsigned int EQP_Options::total_num_cycles()
{
    unsigned int read_num(0);
    unsigned int read_lo_cycle_num(0);
    unsigned int read_hi_cycle_num(0);

    get_read_cycle_range(read_num, read_lo_cycle_num, read_hi_cycle_num);
    return read_hi_cycle_num;
}

/*****************************************************************************/

int EQP_Options::getOptions(int argc, char* const argv[])
    throw (CommandLine::UnknownOptionException, 
           CommandLine::MissingOptionException, 
           CommandLine::InvalidArgumentException)
{
    static const char* short_option_cstr = "hl:e:mz:Z:x:";

    static const struct option long_option_arr[] =
        { {"help",                    no_argument, 0, 'h'},
          {"orig_read_lengths", required_argument, 0, 'l'},
          {"early_cycles",      required_argument, 0, 'e'},
          {"multi_read",              no_argument, 0, 'm'},
          {"sig2_compression",  required_argument, 0, 'z'},
          {"qval_compression",  required_argument, 0, 'Z'},
          {"predictors",        required_argument, 0, 'x'},
          {0, 0, 0, 0}
        };

    if(argc<=1) return usage(argc,argv);

    int long_option_index(0);
    int option(0);

    // Note: case 'h' returns immediately.
    while (-1 != (option = getopt_long(argc, argv,
                                       short_option_cstr, long_option_arr,
                                       &long_option_index))) {
        switch (option) {
        case 0:
            break;
        case 'h':
            return usage(argc, argv);
            break;
        case 'l':
            my_orig_read_lengths_str = optarg;
            break;
        case 'e':
            my_num_early_cycles = Options::convert<unsigned int>(optarg);
            break;
        case 'm':
            my_multi_read_flag = true;
            break;
        case 'z':
            my_sig2_compr_name = optarg;
            break;
        case 'Z':
            my_qval_compr_name = optarg;
            break;
        case 'x':
            my_qptypes_str = optarg;
            break;

        case '?':
        case ':':
        default:
            throw UnknownOptionException("");
            break;
        }
    }

    return RUN;
}

/*****************************************************************************/
// FIXME : Either use File_Buffer::CompressionType in IntensityData
//         or at least move this functionality there.

bool resolve_compression(const std::string& compr_name,
                         IntensityData::CompressionType& compr_type,
                         std::string& msg_str)
{
    typedef std::map<std::string, IntensityData::CompressionType> 
        Str_Compr_Map;
    typedef Str_Compr_Map::const_iterator Str_Compr_Map_CIter;
    Str_Compr_Map str_compr_map;

    str_compr_map["none"] = IntensityData::compressionNone;
    str_compr_map["gzip"] = IntensityData::compressionGzip;
    str_compr_map["bzip2"] = IntensityData::compressionBzip2;

    Str_Compr_Map_CIter match_str_compr_citer(str_compr_map.find(compr_name));

    if (str_compr_map.end() == match_str_compr_citer) {
        msg_str += "compression must be";

        for (Str_Compr_Map_CIter str_compr_citer = str_compr_map.begin();
             str_compr_citer != str_compr_map.end(); ++str_compr_citer) {
            msg_str += (((str_compr_citer == str_compr_map.begin())
                         ? " " : " or ")
                        + str_compr_citer->first);
        }

        return false;
    }

    compr_type = match_str_compr_citer->second;
    return true;
}



static
bool
check_no_repeat_qptypes(const Quality_Data::QPTypes& qptypes, std::string& msg_str){

    using namespace QPREDICTOR;

    std::vector<bool> qptypes_mask(SIZE,false);
    const unsigned n_qptypes(qptypes.size());
    for (unsigned i(0);i<n_qptypes;++i){
        const index_t qpi(qptypes[i]);
        if (qptypes_mask[qpi]) {
            msg_str += string("predictor `") + get_label(qpi) + 
                        "` is repeated in predictor list.";
            return false;
        }

        qptypes_mask[qpi] = true;
    }
    return true;
}



/*****************************************************************************/

bool EQP_Options::parse_qual_fields(std::string& msg_str)
{
    my_qptypes.clear();

    // Return default list if predictors not specified.
    if (my_qptypes_str.empty()) {
        my_qptypes = my_default_qptypes;
        bool status = check_no_repeat_qptypes(my_qptypes,msg_str);
        return status;
    }

    using namespace QPREDICTOR;

    /// \todo get rid of vector copy:
    ///
    vector<string> qptype_str_vec = tokenize_str(my_qptypes_str, ":");

    const unsigned n_qptype_strs(qptype_str_vec.size());
    for(unsigned i(0);i<n_qptype_strs;++i){
        const index_t qpi(get_index(qptype_str_vec[i].c_str()));

        if (qpi == SIZE) {
            msg_str += ("Unknown predictor name: "+qptype_str_vec[i]+"\n");
            return false;
        }

        my_qptypes.push_back(qpi);
    }

    bool status = check_no_repeat_qptypes(my_qptypes,msg_str);
    return status;
}


/*****************************************************************************/

void EQP_Options::processOptions() 
    throw (CommandLine::InvalidArgumentException)
{
    // Parse colon-separated original read lengths list.

    typedef std::vector<std::string> Str_Vec;
    typedef Str_Vec::const_iterator Str_Vec_CIter;

    if (!my_orig_read_lengths_str.empty()) {
        unsigned int total_orig_read_len(0);
        Str_Vec read_len_str_vec = tokenize_str(my_orig_read_lengths_str, ":");
        unsigned int read_num(1);

        for (Str_Vec_CIter read_len_str_citer = read_len_str_vec.begin();
             read_len_str_citer != read_len_str_vec.end();
             ++read_len_str_citer, ++read_num) {
            const unsigned int orig_read_len
                = Options::convert<unsigned int>(*read_len_str_citer);
            total_orig_read_len += orig_read_len;
            my_orig_read_lens_list.push_back(orig_read_len);

            // If we are processing multiple reads separately, construct a 
            // list of them (assuming read numbers increment from 1).
            if (my_multi_read_flag) {
                my_read_num_vec.push_back(read_num);
            }
        }
    }


    // If we are not processing multiple reads separately, the list will
    // contain a single invalid read number (same convention as 
    // parseUseBases).
    if (!my_multi_read_flag) {
        my_read_num_vec.push_back(0);
    }

    if (my_multi_read_flag && (my_orig_read_lens_list.size() < 2)) {
        std::string msg_str("A list of per-read original read lengths (for "
                            "multiple reads) must be supplied when the "
                            "--multi_read option is specified.");
        throw InvalidArgumentException(msg_str);
    }


    // Parse compression types.

    std::string msg_str("sig2 ");

    if (!resolve_compression(my_sig2_compr_name, my_sig2_compr, msg_str)) {
        throw InvalidArgumentException(msg_str);
    }

    msg_str = "qval ";

    if (!resolve_compression(my_qval_compr_name, my_qval_compr, msg_str)) {
        throw InvalidArgumentException(msg_str);
    }


    // Parse predictor list.

    msg_str = "";

    if (!parse_qual_fields(msg_str)) {
        throw InvalidArgumentException(msg_str);
    }
}

/*****************************************************************************/

void EQP_Options::getParameters(int argc, char* const argv[])
        throw (CommandLine::InvalidArgumentException)
{
    bool sig2_file_found(false);
    bool prb_file_found(false);
    bool seq_file_found(false);

    while (argc > optind) {
        const std::string arg_str(argv[optind++]);

        if (!sig2_file_found) {
            my_sig2_path_str = arg_str;
            sig2_file_found = true;
        } else if (!prb_file_found) {
            my_prb_path_str = arg_str;
            prb_file_found = true;
        } else if (!seq_file_found) {
            my_seq_path_str = arg_str;
            seq_file_found = true;
        } else {
            my_qval_path_str_list.push_back(arg_str);
        }
    }

    if (!sig2_file_found) {
        throw MissingOptionException("Sig2 file not specified.");
    }

    if (!prb_file_found) {
        throw MissingOptionException("Prb file not specified.");
    }

    if (!seq_file_found) {
        throw MissingOptionException("Seq file not specified.");
    }

    if (my_multi_read_flag) {
        if (my_qval_path_str_list.size() == 1) {
            throw MissingOptionException("Only one output (qval) file "
                                         "supplied though --multi_read "
                                         "was specified.");
        }

        if (my_qval_path_str_list.size() != my_orig_read_lens_list.size()) {
            throw MissingOptionException("Number of original read lengths "
                                         "specified does not match number "
                                         "of output (qval) files.");
        }
    } else if (my_qval_path_str_list.size() > 1) {
        throw InvalidArgumentException("Multiple output (qval) files "
                                       "were specified without "
                                       "--multi_read option.");        
    } else if (my_qval_path_str_list.empty()) {
        throw MissingOptionException("No qval file specified.");
    }
}

/*****************************************************************************/

int EQP_Options::usage(int /*num_args*/, 
                       char* const arg_cstr_arr[],
                       const std::string& msg_str)
{
    if (0 != msg_str.size())
    {
        error(msg_str.c_str());
    }

    const std::string app_name(arg_cstr_arr[0]);
    std::ostringstream ostrm;

    ostrm << "Usage : " << app_name << " -h|--help" << std::endl
          << "   OR : " << app_name << std::endl
          << "[-l|--orig_read_lengths <ORIG_READ_LENGTHS> [-m|--multi_read]]"
          << std::endl
          << "[-e|--early_cycles <EARLY_CYCLES>] " << std::endl
          << "[-m|--multi_read] " << std::endl
          << "[-z|--sig2_compression <gzip|bzip2>] " << std::endl
          << "[-Z|--qval_compression <gzip|bzip2>] " << std::endl
          << "[-x|--predictors <PREDICTOR_LIST>] " << std::endl
          << "<sig2_file> <prb_file> <seq_file> " << std::endl
          << "<qval_file | read1_qval_file read2_qval_file ...>" << std::endl
          << std::endl
          << "where :-" << std::endl
          << "ORIG_READ_LENGTHS is a colon-separated list " << std::endl
          << "EARLY_CYCLES is the number of cycles over which " << std::endl
          << "  max_early_* are calculated (default " 
          << my_default_num_early_cycles << ")" << std::endl
          << "PREDICTOR_LIST is a colon-separated list drawn from the "
          << std::endl
          << "  following valid predictor names :-" << std::endl;

    using namespace QPREDICTOR;

    for(unsigned i(0);i<SIZE;++i){
        ostrm << get_label(i) << "\n";
    }

    ostrm << "\n"
          << "The default list is :-\n";

    const unsigned n_def_types(my_default_qptypes.size());
    ostrm << "  ";
    for(unsigned i(0);i<n_def_types;++i){
        if(i) ostrm << ":";
        ostrm << get_label(my_default_qptypes[i]);
    }
    ostrm << "\n";

    info(ostrm.str().c_str());
    return NO_RUN;
}

/*****************************************************************************/
