/*****************************************************************************/
// Copyright (c) Illumina 2008
// Author: Richard Shaw
// This source file is covered by the Illumina Genome Analyzer Software 
// License Agreement and bound by the terms therein.
/*****************************************************************************/

#include <string>
#include <map>
#include <sstream>

#include <getopt.h>

#include "File_Buffer.h"
#include "Message.h"
#include "Tile_List.h"

#include "QCal_Options.h"

using namespace CommandLine;

/*****************************************************************************/

QCal_Options::QCal_Options()
    : CommandLine::Options(),
      my_qval_suffix_str("_qval.txt"),
      my_qref_suffix_str("_qref.txt"),
      my_qval_compr_name("none"),
      my_qref_compr_name("none"),
      my_cfg_path_str("QualityCalibration.xml"),
      my_qval_compr(File_Buffer::compressionNone),
      my_qref_compr(File_Buffer::compressionNone)
{
    ;
}

/*****************************************************************************/

int QCal_Options::getOptions(int argc, char* const argv[])
    throw (CommandLine::UnknownOptionException, 
           CommandLine::MissingOptionException, 
           CommandLine::InvalidArgumentException)
{
    const char* short_option_cstr = "hS:s:Z:z:c:T:t:";

    const struct option long_option_arr[] =
        { {"help",                   no_argument, 0, 'h'},
          {"qval_suffix",      required_argument, 0, 'S'},
          {"qref_suffix",      required_argument, 0, 's'},
          {"qval_compression", required_argument, 0, 'Z'},
          {"qref_compression", required_argument, 0, 'z'},
          {"cfg",              required_argument, 0, 'c'},
          {"tile_list_file",   required_argument, 0, 'T'},
          {"tile_prefix",      required_argument, 0, 't'},
          {0, 0, 0, 0}
        };

    int long_option_index(0);
    int option(0);

    // Note: case 'h' returns immediately.
    while (-1 != (option = getopt_long(argc, argv,
                                       short_option_cstr, long_option_arr,
                                       &long_option_index))) {
        switch (option) {
        case 0:
            break;
        case 'h':
            return usage(argc, argv);
            break;
        case 'S':
            my_qval_suffix_str = optarg;
            break;
        case 's':
            my_qref_suffix_str = optarg;
            break;
        case 'Z':
            my_qval_compr_name = optarg;
            break;
        case 'z':
            my_qref_compr_name = optarg;
            break;
        case 'c':
            my_cfg_path_str = optarg;
            break;
        case 'T':
            my_tile_list_file_path_str = optarg;
            break;
        case 't':
            my_tile_prefix_str = optarg;
            break;

        case '?':
        case ':':
        default:
            throw UnknownOptionException("");
            break;
        }
    }

    return RUN;
}

/*****************************************************************************/

void QCal_Options::processOptions()
        throw (CommandLine::InvalidArgumentException)
{
    std::string msg_str("qval ");

    if (!resolve_compression(my_qval_compr_name, my_qval_compr, msg_str)) {
        throw InvalidArgumentException(msg_str);
    }

    msg_str = "qref ";

    if (!resolve_compression(my_qref_compr_name, my_qref_compr, msg_str)) {
        throw InvalidArgumentException(msg_str);
    }

    if (my_tile_list_file_path_str.empty() != my_tile_prefix_str.empty()) {
        msg_str = ("Either both or neither of --tile_list_file and "
                   "--tile_prefix must be specified.");
        throw InvalidArgumentException(msg_str);
    }
}

/*****************************************************************************/

void QCal_Options::getParameters(int argc, char* const argv[])
        throw (CommandLine::InvalidArgumentException)
{
    // Support specification of qval files either by tile list file
    // OR by explicit listing but not both - on the assumption that
    // an attempt to do both is probably a user error.

    if (!my_tile_list_file_path_str.empty()) {
        Tile_List tile_list(my_tile_list_file_path_str);
        Tile_List::Tile_Name_List tile_name_list;

        if (tile_list.find_tiles(my_tile_prefix_str, tile_name_list,
                                 my_qval_suffix_str)) {
            my_qval_file_name_list
                = static_cast<File_Name_List>(tile_name_list);
        } else {
            const std::string msg_str
                = ("Failed to read tile list from `"
                   + my_tile_list_file_path_str + "' for prefix `"
                   + my_tile_prefix_str + "'.");
            throw InvalidArgumentException(msg_str.c_str());
        }

        if (argc > optind) {
            const std::string msg_str
                ("Attempt to list one or more qval files explicitly after"
                 " specifying qval files by tile list file.");
            throw InvalidArgumentException(msg_str.c_str());
        }

        return;
    }

    const unsigned int qval_suffix_len(my_qval_suffix_str.length());

    while (argc > optind) {
        const std::string arg_str(argv[optind++]);
        bool suffix_ok(true);

        if (arg_str.length() >= qval_suffix_len) {
            suffix_ok = (arg_str.compare(arg_str.length() - qval_suffix_len,
                                         qval_suffix_len,
                                         my_qval_suffix_str) == 0);
        } else {
            suffix_ok = false;
        }

        if (!suffix_ok) {
            throw InvalidArgumentException(arg_str 
                                           + " does not include suffix `"
                                           + my_qval_suffix_str + "'");
        }

        my_qval_file_name_list.push_back(arg_str);
    }

    if (my_qval_file_name_list.empty()) {
        throw InvalidArgumentException("At least one qval file must be "
                                       "specified using a tile list or "
                                       "by being explicitly listed.");
    }
}

/*****************************************************************************/

int QCal_Options::usage(int num_args, char* const arg_cstr_arr[],
                        const std::string& msg_str)
{
    if (0 != msg_str.size()) {
        error(msg_str.c_str());
    }

    const std::string app_name(arg_cstr_arr[0]);
    std::ostringstream ostrm;

    ostrm << "Usage : " << app_name << " -h|--help" << std::endl
          << "   OR : " << app_name 
          << "[-S|--qval_suffix _qval.txt]" << std::endl
          << "[-s|--qref_suffix _qref.txt]" << std::endl
          << "[-Z|--qval_compression none|gzip|bzip2] " << std::endl
          << "[-z|--qref_compression none|gzip|bzip2] " << std::endl
          << "-c|--cfg <cfg_file.xml> " << std::endl
          << "< <qvalue-file1> [<qvalue-file2> ...]" << std::endl
          << "  | -T|--tile_list_file <tiles.txt> " << std::endl
          << "    -t|--tile_prefix <lane_prefix e.g. s_3> >" << std::endl
          << std::endl
          << "The qval_suffix and qref_suffix options (with default values "
          << std::endl
          << "as shown above) are used to generate qref filenames from the "
          << std::endl
          << "specified qval filenames." << std::endl
          << std::endl 
          << "In addition, however, if the qval files are specified by means "
          << std::endl
          << "of a tile list file plus lane prefix filter, the qval_suffix "
          << std::endl
          << "is appended to the tile names to generate the qval filenames."
          << std::endl
          << "In this case, non-default suffix values may be required, e.g. "
          << std::endl
          << "_1_qval.txt and _1_qref.txt for read 1 in eland pair analysis."
          << std::endl;

    info(ostrm.str().c_str());
    return NO_RUN;
}

/*****************************************************************************/
