// -*- mode: c++; indent-tabs-mode: nil; c-basic-offset: 4; -*-
//
// Copyright 2008 Illumina, Inc
// Author: Chris Saunders
//
// This source file is covered by the Illumina Genome Analyzer Software
// License Agreement and bound by the terms therein.
//

#include <algorithm>

#include <cmath>

#include "QPredictorFunctions.h"
#include "QPredictorMotif5p13p1.h"
#include "QPredictorMotif5p23p2.h"


static
float 
calc_unpurity(const float* sig,
              const unsigned n_channels)
{
    float max(sig[0]);
    float sum(sig[0]);

    for(unsigned i(1);i<n_channels;++i){
        const float s(sig[i]);
 
        if (s > max) {
            max = s;
        }
        sum += s;
    }

    if((max <= 0.) || (sum <= 0.)) return 1.;

    return 1.- max/sum;
}


static
float 
calc_unchastity(const float* sig,
                const unsigned n_channels)
{
    float max(sig[0]);
    float max2(sig[1]);

    for(unsigned i(1);i<n_channels;++i){
        const float s(sig[i]);
 
        if (s > max) {
            max2 = max;
            max = s;
        } else if (s > max2) {
            max2 = s;
        }
    }

    const float sum(max+max2);
    
    if((max <= 0.) || (sum <= 0.)) return 1.;

    return 1.- max/sum;
}


static
float 
calc_shifted_unpurity(const float* sig,
                      const unsigned n_channels)
{
    float max(sig[0]);
    float min(sig[0]);
    float sum(sig[0]);

    for(unsigned i(1);i<n_channels;++i){
        const float s(sig[i]);
 
        if      (s>max) { max = s; }
        else if (s<min) { min = s; }
        
        sum += s;
    }

    const float shift(1.-min);
    max += shift;
    sum += shift*n_channels;

    return 1.- max/sum;
}


static
float 
calc_shifted_unchastity(const float* sig,
                        const unsigned n_channels)
{
    float max(sig[0]);
    float max2(sig[1]);
    float min(sig[0]);

    for(unsigned i(1);i<n_channels;++i){
        const float s(sig[i]);
 
        if (s>max){
            max2 = max;
            max = s;
        } else if (s>max2) {
            max2 = s;
        }

        if (s<min) { min=s; }
    }

    const float shift(1.-min);
    max  += shift;
    max2 += shift;

    return 1.- max/(max+max2);
}

#if 0
static
float
calc_shifted_sum_diff_unpurity(const float* sig,
                               const unsigned n_channels)
{
    float max(sig[0]);
    float min(sig[0]);
    float sum(sig[0]);

    for(unsigned i(1);i<n_channels;++i){
        const float s(sig[i]);
 
        if (s>max) { max = s; }
        if (s<min) { min = s; }

        sum += s;
    }

    const float shift(1.-min);
    const float shifted_sum = sum+(shift*n_channels);

    return (sum-(2.*max))/shifted_sum;
}
#endif

static
float 
calc_shifted_sum_diff_unchastity(const float* sig,
                                 const unsigned n_channels)
{
    float max(sig[0]);
    float max2(sig[1]);
    float min(sig[0]);
    float sum(sig[0]);

    for(unsigned i(1);i<n_channels;++i){
        const float s(sig[i]);
 
        if (s>max){
            max2 = max;
            max = s;
        } else if (s>max2) {
            max2 = s;
        }

        if (s<min) { min=s; }

        sum += s;
    }

    const float shift(1.-min);
    sum += (shift*n_channels);

    return (max2-max)/sum;
}


// 
//

using namespace QPREDICTOR;



struct CycleFunc : public FuncBase {
    CycleFunc(const FuncInitData& d) : FuncBase(d) {}
    
    virtual 
    void get_cluster_values(const unsigned,
                            std::vector<float>::iterator fval) {
        
        for(unsigned i(0); i < _cycle_range; ++i){            
            *fval++ = (i+1);
        }
    }
};


/// if is_bounded is false, signal decay reports values below zero, a result
/// of a cycle having higher intensity than the first base.
///
struct SignalDecayFunc : public FuncBase {
    SignalDecayFunc(const FuncInitData& d,
                    const bool is_bounded) : FuncBase(d), _is_bounded(is_bounded) {}
    
    virtual 
    void get_cluster_values(const unsigned cluster_id,
                            std::vector<float>::iterator fval) {
        
        const unsigned cluster_offset(cluster_id*_data.n_channels);

        float cycle1_max_sig(0.);

        for(int i(_data.low_cycle-1); i < (int)_data.high_cycle; ++i){
            float* cycle_sig(_data.sig[i]+cluster_offset);
            
            const float max_sig(*std::max_element(cycle_sig,cycle_sig+_data.n_channels));
            if(i == (int)(_data.low_cycle-1) ) {
                cycle1_max_sig = max_sig;
            }

            float val;
            if ( cycle1_max_sig <= 0. || max_sig <= 0. ) val = 1.;
            else                                         val = 1.- (max_sig/cycle1_max_sig);

            if(_is_bounded) val = std::max(val,static_cast<float>(0.));
            *fval++ = val;
        }
    }
private:
    const bool _is_bounded;
};



struct HomopolLenFunc : public FuncBase {
    HomopolLenFunc(const FuncInitData& d) : FuncBase(d), _len(_cycle_range) {}

    virtual
    void get_cluster_values(const unsigned cluster_id,
                            std::vector<float>::iterator fval) {

        std::fill(_len.begin(),_len.end(),0);
        unsigned n(1);
        char last_c(0);
        for(int i(static_cast<int>(_cycle_range)-1);i>=0;--i){
            const unsigned ci(i+_data.low_cycle-1);
            const char c(_data.seq[ci*_data.n_clusters+cluster_id]);
            if(i==(int)(_cycle_range-1)) {
                last_c=c;
            } else if(c==last_c){
                n++;
            } else {
                _len[i+1]=n;
                n=1;
                last_c=c;
            }

            if(i==0){ _len[i]=n; }
        }

        for(unsigned i(0);i<_cycle_range;++i){
            if(_len[i] != 0) n=_len[i];
            *fval++ = n;
        }
    }

private:
    std::vector<unsigned> _len;
};



struct RawUnqualityFunc : public FuncBase {
    RawUnqualityFunc(const FuncInitData& d) : FuncBase(d) {}
    
    virtual 
    void get_cluster_values(const unsigned cluster_id,
                            std::vector<float>::iterator fval) {
        
        const unsigned cluster_offset(cluster_id*_data.n_channels);
        
        for(int i(_data.low_cycle-1); i < (int)(_data.high_cycle); ++i){
            float* cycle_prob(_data.prob[i]+cluster_offset);
            
            *fval++ = -static_cast<int>(*std::max_element(cycle_prob,cycle_prob+_data.n_channels));
        }
    }
};



struct Sig2Func : public FuncBase {

    typedef float(*score_func_t)(const float*, const unsigned);

    Sig2Func(const FuncInitData& d,
             score_func_t sf) : FuncBase(d), _sig2_score_func(sf) {}
    
    virtual 
    void get_cluster_values(const unsigned cluster_id,
                            std::vector<float>::iterator fval) {
        
        const unsigned cluster_offset(cluster_id*_data.n_channels);
        
        for(int i(_data.low_cycle-1); i < (int)_data.high_cycle; ++i){
            float* cycle_sig(_data.sig[i]+cluster_offset);
            
            *fval++ = _sig2_score_func(cycle_sig,_data.n_channels);
        }
    }

private:
    score_func_t _sig2_score_func;
};



struct UnpurityFunc : public Sig2Func {
    UnpurityFunc(const FuncInitData& d) : Sig2Func(d,calc_unpurity) {}
};

struct UnchastityFunc : public Sig2Func {
    UnchastityFunc(const FuncInitData& d) : Sig2Func(d,calc_unchastity) {}
};

struct ShiftedUnpurityFunc : public Sig2Func {
    ShiftedUnpurityFunc(const FuncInitData& d) : Sig2Func(d,calc_shifted_unpurity) {}
};

struct ShiftedUnchastityFunc : public Sig2Func {
    ShiftedUnchastityFunc(const FuncInitData& d) : Sig2Func(d,calc_shifted_unchastity) {}
};

struct ShiftedSumDiffUnchastityFunc : public Sig2Func {
    ShiftedSumDiffUnchastityFunc(const FuncInitData& d) 
        : Sig2Func(d,calc_shifted_sum_diff_unchastity) {}
};



static
bool
is_gc(const char c){
    switch(c) {
    case 'C':
    case 'G':
    case 'c':
    case 'g': return true;
    default : return false;
    }
}

static
bool
is_at(const char c){
    switch(c) {
    case 'A':
    case 'T':
    case 'a':
    case 't': return true;
    default : return false;
    }
}


struct GCContentFunc : public FuncBase {
    GCContentFunc(const FuncInitData& d) : FuncBase(d) {}

    virtual
    void get_cluster_values(const unsigned cluster_id,
                            std::vector<float>::iterator fval) {

        unsigned ngc(0),nat(0);
        for(int i(_data.low_cycle-1); i < (int)_data.high_cycle; ++i){
            const char c(_data.seq[i*_data.n_clusters+cluster_id]);
            if     (is_gc(c)) { ++ngc; }
            else if(is_at(c)) { ++nat; }
        }

        float val;
        if( (ngc+nat) > 0 ) {
            val = static_cast<float>(ngc)/static_cast<float>(ngc+nat);
        } else {
            val = 1.;
        }

        for(unsigned i(0);i<_cycle_range;++i){ 
            *fval++ = val;
        }
    }
};


static
unsigned
base_id(const char c){
    switch(c) {
    case 'A': return 0;
    case 'C': return 1;
    case 'G': return 2;
    case 'T': return 3;
    default : return 4;
    }
}


struct UnEntropyFunc : public FuncBase {
    UnEntropyFunc(const FuncInitData& d) : FuncBase(d) {}

    virtual
    void get_cluster_values(const unsigned cluster_id,
                            std::vector<float>::iterator fval) {

        static const float invlog2(1./std::log(2.));

        for(unsigned i(0);i<(n_bases+1);++i){
            counts[i] = 0.;
        }

        for(int i(_data.low_cycle-1); i < (int)_data.high_cycle; ++i){
            const char c(_data.seq[i*_data.n_clusters+cluster_id]);
            counts[base_id(c)] += 1.;
        }

        const float norm(1./static_cast<float>(_cycle_range-counts[n_bases]));

        float unentropy(0.);
        for(unsigned i(0);i<n_bases;++i){
            if(counts[i] <= 0.) continue;
            const float p(counts[i]*norm);
            unentropy += p*std::log(p);
        }
        unentropy *= invlog2;

        for(unsigned i(0); i < _cycle_range; ++i){ 
            *fval++ = unentropy;
        }
    }

    enum { n_bases = 4 };
    float counts[n_bases+1];
};



struct EdginessFunc : public FuncBase {
    EdginessFunc(const FuncInitData& d) : FuncBase(d) {
        /// \todo get procedure from TM for finding center X and Y from
        ///       pipeline files, esp for GA2
        _center_x=500.;
        _center_y=500.;
    }

    virtual
    void get_cluster_values(const unsigned cluster_id,
                            std::vector<float>::iterator fval) {

        const float x(static_cast<float>(_data.pos[cluster_id].X));
        const float y(static_cast<float>(_data.pos[cluster_id].Y));

        const float edginess(std::max((std::abs(_center_x-x)/_center_x),
                                      (std::abs(_center_y-y)/_center_y)));

        for(unsigned i(0); i < _cycle_range; ++i){ 
            *fval++ = edginess;
        }
    }

private:
    float _center_x;
    float _center_y;
};



/// returns the worst (max) score in the first n_early_cycles cycles
///
template <typename SCORE_FUNC>
struct EarlyFunc : public FuncBase {
    EarlyFunc(const FuncInitData& d) 
        : FuncBase(d),_scores(_data.n_early_cycles), _sfd(d) {
        _sfd.high_cycle=_sfd.n_early_cycles;
        _sf.reset(new SCORE_FUNC(_sfd));
    }

    virtual
    void get_cluster_values(const unsigned cluster_id,
                            std::vector<float>::iterator fval) {
        
        _sf->get_cluster_values(cluster_id,_scores.begin());

        const float fmax(*std::max_element(_scores.begin(),_scores.end()));
        for(unsigned i(0); i < _cycle_range; ++i){
            *fval++ = fmax;
        } 
   }

private:
    std::vector<float> _scores;
    FuncInitData _sfd;
    std::auto_ptr<SCORE_FUNC> _sf;
};



/// returns the worst (max) score from a neighborhood of +/- nhood_size cycles about the current cycle
///
template <typename SCORE_FUNC>
struct NHoodFunc : public FuncBase {
    NHoodFunc(const FuncInitData& d,
              const unsigned nhood_size) 
        : FuncBase(d), _nhood_size(static_cast<int>(nhood_size)),
          _scores(_cycle_range), _sf(d)  {}

    virtual 
    void get_cluster_values(const unsigned cluster_id,
                            std::vector<float>::iterator fval) {
        
        _sf.get_cluster_values(cluster_id,_scores.begin());

        for(int i(0); i < static_cast<int>(_cycle_range); ++i){
            float fmax(_scores[i]);
            for(int j(i-_nhood_size);j<=(i+_nhood_size); j++){
                if(j == i || j<0 || j>=(int)_cycle_range) continue;
                fmax=std::max(fmax,_scores[j]);
            }
            
            *fval++ = fmax;
        } 
   }

private:
    const int _nhood_size;
    std::vector<float> _scores;
    SCORE_FUNC _sf;
};



template <typename MOTIF_DATA>
struct SeqMotifFunc : public FuncBase {
    SeqMotifFunc(const FuncInitData& d) 
        : FuncBase(d), _motif(new char[_mf.crange]) {}

    ~SeqMotifFunc() { delete [] _motif; }

    virtual
    void get_cluster_values(const unsigned cluster_id,
                            std::vector<float>::iterator fval) {

        for(int i(_data.low_cycle-1); i < (int)_data.high_cycle; ++i){
            for(int j(0);j<_mf.crange;++j){
                const int ci(j+i-_mf.c5p);
                char c('N');
                if( (ci >= (int)(_data.low_cycle-1)) && (ci < (int)_data.high_cycle) ){
                    c=_data.seq[ci*_data.n_clusters+cluster_id];
                }
                _motif[j]=c;
            }

            *fval++ = get_score();
        }
    }

private:
    float
    get_score() const {
        typename MOTIF_DATA::motif_map_t::const_iterator mi(_mf.motif_scores.find(_motif));
        if(mi == _mf.motif_scores.end()) return 0.;
        else                             return mi->second;
    }


    static const MOTIF_DATA _mf;
    char* _motif;
};

template <typename T> const T SeqMotifFunc<T>::_mf;



namespace QPREDICTOR {

    FuncSet::FuncSet(const FuncInitData& d) : _data(d) {
        _qpms[IN_READ_CYCLE_NUM].reset(new CycleFunc(_data));
        _qpms[SIGNAL_DECAY].reset(new SignalDecayFunc(_data,true));
        _qpms[UNBOUNDED_SIGNAL_DECAY].reset(new SignalDecayFunc(_data,false));
        _qpms[HOMOPOL_LEN].reset(new HomopolLenFunc(_data));
        _qpms[RAW_UNQUALITY].reset(new RawUnqualityFunc(_data));
        _qpms[UNCHASTITY].reset(new UnchastityFunc(_data));
        _qpms[UNPURITY].reset(new UnpurityFunc(_data));
        _qpms[SHIFTED_UNCHASTITY].reset(new ShiftedUnchastityFunc(_data));
        _qpms[SHIFTED_UNPURITY].reset(new ShiftedUnpurityFunc(_data));
        _qpms[SHIFTED_SUM_DIFF_UNCHASTITY].reset(new ShiftedSumDiffUnchastityFunc(_data));
        _qpms[GC_CONTENT].reset(new GCContentFunc(_data));
        _qpms[UNENTROPY].reset(new UnEntropyFunc(_data));
        _qpms[EDGINESS].reset(new EdginessFunc(_data));
        _qpms[MAX_EARLY_UNCHASTITY].reset(new EarlyFunc<UnchastityFunc>(_data));
        _qpms[MAX_EARLY_UNPURITY].reset(new EarlyFunc<UnpurityFunc>(_data));
        _qpms[MAX_EARLY_SHIFTED_UNCHASTITY].reset(new EarlyFunc<ShiftedUnchastityFunc>(_data));
        _qpms[MAX_EARLY_SHIFTED_UNPURITY].reset(new EarlyFunc<ShiftedUnpurityFunc>(_data));
        _qpms[MAX_EARLY_SHIFTED_SUM_DIFF_UNCHASTITY].reset(new EarlyFunc<ShiftedSumDiffUnchastityFunc>(_data));
        _qpms[MAX_LOCAL_UNCHASTITY].reset(new NHoodFunc<UnchastityFunc>(_data,2));
        _qpms[UNCHASTITY_NHOOD1].reset(new NHoodFunc<UnchastityFunc>(_data,1));
        _qpms[UNPURITY_NHOOD1].reset(new NHoodFunc<UnpurityFunc>(_data,1));
        _qpms[SHIFTED_UNCHASTITY_NHOOD1].reset(new NHoodFunc<ShiftedUnchastityFunc>(_data,1));
        _qpms[SHIFTED_UNPURITY_NHOOD1].reset(new NHoodFunc<ShiftedUnpurityFunc>(_data,1));
        _qpms[SHIFTED_SUM_DIFF_UNCHASTITY_NHOOD1].reset(new NHoodFunc<ShiftedSumDiffUnchastityFunc>(_data,1));
        _qpms[CONTEXT_5P1_3P1].reset(new SeqMotifFunc<Motif5P13P1Data>(_data));
        _qpms[CONTEXT_5P2_3P2].reset(new SeqMotifFunc<Motif5P23P2Data>(_data));
    }
}

