// -*- mode: c++; indent-tabs-mode: nil; c-basic-offset: 4; -*-
/*****************************************************************************/
//
// Copyright (c) Illumina 2008
// Author: R. J. Shaw
// This source file is covered by the "Solexa Public Source License"
// agreement and bound by the terms therein.
//
/*****************************************************************************/

#include <cassert>

#include "File_Buffer.h"
#include "Maths.h"
#include "StringUtil.h"

#include "QualityData.h"

/*****************************************************************************/

int Quality_Data::write(const QPBufferType& qp_buffer)
{
    int err_num(0);

    // Calculate the required size of buffer and allocate it.
    const unsigned n_predictors(qp_buffer.value.size());
    const unsigned n_lines(qp_buffer.called_base.size());

    // limit for %.6g float write for each predictor (doubles could have a three digit exp and would be 13):
    static const unsigned max_expected_qp_width(12);
    static const unsigned other_line_width(2); // ...for the basecall and '\n'

    const unsigned max_expected_line_width((max_expected_qp_width+1)*n_predictors+other_line_width);
    const unsigned file_buf_size(max_expected_line_width*n_lines);

    char* file_buf_ptr = new char[file_buf_size + 1];
    char* buf_ptr(file_buf_ptr);
                                 
    // default c++ new should throw an exception before this occurs:
    //
    assert(file_buf_ptr != 0);
    if (file_buf_ptr == 0) {
        static const int out_of_memory(51); // well above write_file_* ranges.
        return out_of_memory;
    }

    // write the quality predictors buffer to a string buffer.
    for (unsigned i(0); i<n_lines; ++i) {
        for (unsigned j(0);j<n_predictors;++j){
            buf_ptr += sprintf(buf_ptr, "%.6g\t",qp_buffer.value[j][i]);
        }

        assert(qp_buffer.called_base[i] != '\0');
        buf_ptr += sprintf(buf_ptr,"%c\n",qp_buffer.called_base[i]);
    }

    const int num_useful_bytes(buf_ptr - file_buf_ptr);


    // Write the buffer to the file (with compression if required).

    if (!my_file_buffer.write_buffer(file_buf_ptr, num_useful_bytes)) {
        // FIXME : central error codes?
        err_num = 1;
    }

    delete [] file_buf_ptr;
    return err_num;
}

/*****************************************************************************/

#if 0
int Quality_Data::read_chunk(Quality_Predictors_Vec& quality_predictors_vec)
{
    return read_quality_file_chunk(my_file_buffer, quality_predictors_vec);
}

/*****************************************************************************/

unsigned int Quality_Data::get_field_width(const Field field) const
{
    Field_Width_Map_CIter field_width_citer = my_field_width_map.find(field);

    assert(field_width_citer != my_field_width_map.end());

    if (field_width_citer == my_field_width_map.end()) {
        return 0;
    }

    return field_width_citer->second;
}

/*****************************************************************************/
// FIXME : assumes the traditional 4 predictors.
//         Currently Quality_Data::read_chunk is not used anyway.

int read_quality_file_chunk(File_Buffer& file_buffer,
                            Quality_Predictors_Vec& quality_predictors_vec)
{
    quality_predictors_vec.clear();

    int err_num(0);
    bool at_eof(false);

    File_Buffer::Line_Vec line_vec;

    if (!file_buffer.read_some_lines(line_vec, at_eof)) {
        err_num = 1;
        return err_num;
    }

    // Parse out the predictors.

    Quality_Predictors qual_preds;

    for (File_Buffer::Line_Vec_CIter line_citer(line_vec.begin());
         line_citer != line_vec.end(); ++line_citer) {
        const std::string line_str(*line_citer);

        if (line_str.empty()) {
            continue;
        }

        if (line_str[0] == '#') {
            // Skip line as comment (but only if `#' is very first character).
            continue;
        }

        if (sscanf(line_str.c_str(), "%d %f %f %d %c",
                   &(qual_preds.in_read_cycle_num),
                   &(qual_preds.unchastity),
                   &(qual_preds.max_early_unchastity),
                   &(qual_preds.raw_unquality),
                   &(qual_preds.called_base)) == 5) {
            quality_predictors_vec.push_back(qual_preds);
        } else {
            err_num = 42; // above the read_file range
            break;
        }
    }

    return err_num;
}
#endif

/*****************************************************************************/
