// -*- mode: c++; indent-tabs-mode: nil; c-basic-offset: 4; -*-
/*
 * Copyright (c) Solexa 2006, 2007
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

/*
 * Phred-like quality calibration table.
 * Based on "Base-Calling of Automated Sequncer Traces Using Phred. 
 * II. Error Probabilities.", B. Ewing and P. Green, 1998,
 * Genome Research, 8, 186-194.
 * Many thanks go to Pablo Alvarez and Will Brockman, who have first
 * applied phred recalibration to Solexa data and provided guidance 
 * on the subject of quality scores. The choice of calibration parameters
 * in the current implementation is based on their work.
 */

#include <cstdio>
#include <cstdlib>
#include <cassert>
#include <sstream>
#include <iostream>
#include <fstream>
#include "QualityTable.h"

QualityTable::QualityTable()
{
}

QualityTable::~QualityTable()
{
}

void QualityTable::printDataTable()
{
    data_table.print();
}

void QualityTable::printErrorTable()
{
    error_table.print();
}

int QualityTable::addParameter(std::vector<float> binning)
{
    data_table.addParameter(binning);
    return 0;
}

/*
 * Create the rule set for the quality table by iteratively computing 
 * the error table, finding the bin with the lowest error rate and removing
 * all corresponding data from the table.
 */
void QualityTable::create()
{
    int bin;
    error_table.create(&data_table);
    QualityRule rule;
    while ((bin = error_table.findMaximalBin()) > -1) {
        rule.params = error_table.getBinParams(bin);
        float error = (1.+error_table.table[bin].error) /
                      (1.+error_table.table[bin].count); 
        rule.setErrorProbability(error);
        //        rule.error = error_table.table[bin].error;
        //        rule.count = error_table.table[bin].count;
        rule_table.push_back(rule);
        error_table.discardBins(bin);
        error_table.computeCounts(&data_table);
        error_table.computeScores();
    }
}

/*
 * Write the rule set to disk.
 */
int QualityTable::write(FILE *fp)
{
    int i, j;
    for (i = 0; i < (int)(rule_table.size()); i++) {
        char buffer[256];
        int offset = 0;
        std::vector<float> params;
        for (j = 0; j < (int)(rule_table[i].params.size()); j++) {            
            offset += snprintf(&(buffer[offset]), sizeof(buffer)-offset, 
                               "%.3f\t", 
                               rule_table[i].params[j]);
        }
        offset += snprintf(&(buffer[offset]), sizeof(buffer)-offset,
                           "%i\t%i\t%6.4f", 
                           rule_table[i].getPhredScore(),
                           rule_table[i].getSolexaScore(),
                           rule_table[i].getErrorProbability());
#if 0
        offset += snprintf(&(buffer[offset]), sizeof(buffer)-offset, "%8i %8i", 
                           rule_table[i].error, rule_table[i].count);
#endif
        offset += snprintf(&(buffer[offset]), sizeof(buffer)-offset, "\n");
        fprintf(fp, buffer);
    }
    return 0;
}

int QualityTable::write(const char filename[])
{
    int code = 0;
    FILE *fp = fopen(filename, "w");
    if (!fp)
        return -1;
    code = write(fp);
    fclose(fp);
    return code;
}

void QualityTable::print()
{
    write(stdout);
}

/*
 * Read the quality table (the rule set) from disk.
 */
int QualityTable::read(const char filename[])
{
    std::string line;
    std::ifstream file(filename);

    if (!file.is_open()) {
        std::cerr << "Failed to open `" << filename << "'" << std::endl;
        return -1;
    }

    int noparams(0);
    while (getline(file, line)) {
        if (line.size()==0 || line[0]=='#' || line[0]=='>')
            continue;

        std::stringstream ss(line);
        std::vector<std::string>fields;
        std::string s;
        while (ss >> s)
            fields.push_back(s);
        if (fields.size() < 4) {
            std::cerr << "Bad line.\n";
             continue;
        }
        if (noparams == 0)
            noparams = fields.size()-3;
        if ((int)(fields.size())-3 != noparams) {
            std::cerr << "Bad line.\n";
            continue;
        }

        float param;
        QualityRule rule;
        for (int i = 0; i < noparams; i++){
            char * ea;
            char ** e(&ea);
            param = strtof(fields[i].c_str(),e);
            if(*e == fields[i].c_str()) {
                std::cerr << "Invalid floating point threshold values in qtable file";
                abort();
            }
            rule.params.push_back(param);
        }

        // use the phred-score, or whatever is in the phred-score column:
        //
        rule.score = atoi(fields[noparams].c_str());
        rule_table.push_back(rule);
    }

    return 0;
}

/*
 * apply the rule set from the current quality table to the data set
 */
void QualityTable::applyScore(DataSet *data, std::vector<int>& scores)
{
    int no_params = data->getNumberParameters();
    if (no_params == 0)
        return;

    int no_data = data->getNumberValues();
    int i, j, k;
    std::vector<float> params(no_params);
    scores.resize(no_data);
    std::vector<std::vector<float> > *values = data->getParameterValues();
    for (i = 0; i < no_data; i++) {
        for (j = 0; j < no_params; j++)
            params[j] = (*values)[j][i];
#if 0
        printf("***");
        for (j = 0; j < no_params; j++)
            printf("%6.2f ", params[j]);
        printf("\n");
#endif
        //        float min = 1.;
        int   imin = rule_table.size()-1;
        int   flag = 1;
        for (k = 0; k < (int)(rule_table.size()); k++) {
            flag = 1;
            for (j = 0; j < no_params; j++)
                if (params[j] >= rule_table[k].params[j])
                    flag = 0;
#if 0
            if (flag && (rule_table[k].getErrorProbability() < min)) {
                min = rule_table[k].getErrorProbability();
                imin = k;
            }
#else
            if (flag) {
                imin = k;
                //                min = rule_table[k].getErrorProbability();
                break;
            }
#endif
#if 0
            for (j = 0; j < no_params; j++)
                printf("%6.2f ", rule_table[k].params[j]);
            printf("flag %i %i %f\n", flag, imin, min);
#endif
        }
        if (imin >= 0) {
            scores[i] = rule_table[imin].score;
        }
        else { 
          // continue convention of reporting random Q-score for empty
          // qtable (this should be an error!)
          static const int random_phred_score(static_cast<int>(std::floor(-10.*std::log10(0.25)+0.5)));
          scores[i] = random_phred_score;
        }
    }
}

void QualityTable::addData(DataSet &data)
{
    data_table.addData(data);
}

void QualityTable::addData(DataSet *data)
{
    data_table.addData(*data);
}
