// -*- mode: c++; indent-tabs-mode: nil; c-basic-offset: 4; -*-
/*
 * Copyright (c) Solexa 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */
 
/*
 * This object holds the set of rules that make up a quality table.
 * The quality table can be read from or written to disk.
 * The table is constructed in three steps:
 * 1. A data set (i.e. a list of parameters that are monotonic in data quality)
 *    is read. 
 * 2. Then the raw data are binned into a lookup table (data_table).
 * 3. The raw data table can be integrated to produce a second lookup
 *    table (error_table) which holds the integrated error counts and 
 *    corresponding error rates. This table is used to read off the
 *    rules for the quality table.
 */
 
#ifndef QUALITYTABLE_H_
#define QUALITYTABLE_H_

#include <cstdio>
#include <vector>
#include <cmath>
#include "ErrorTable.h"

class QualityRule {
public:
    std::vector<float> params;

    // cts - Ultimately we would like to only store trial and error counts here,
    // and calculate any of phred, solexa or phred lower bound scores from these
    // counts. However to use phred-score based qtables in the pipeline 1.1 
    // EAS release in a way that allows us to read any previously generated 
    // qtable, we will temporarily read the phred Q-value column from the qtable
    // and report this as QualityApplies output. 
    //
    int score;

    float p_error;
#if 0
    long count;
    long error;
    void  setPhredScore(int score) 
    { 
        p_error = float(pow(10., (double(score)/-10.))); 
    }
    void  setSolexaScore(int score)
    {
        p_error = float(1. / (1. + pow(10., double(score)/10.)));
    }
#endif
    void  setErrorProbability(float p) { p_error = p; }

    int   getPhredScore() const 
    { 
        return int(floor(-10.*log10(p_error)+0.5)); 
    }
    float getErrorProbability() const { return p_error; }
    int   getSolexaScore() const
    { 
        return int(floor(10.*log10((1.-p_error)/(p_error))+0.5));
    }
};

class QualityTable {
    std::vector<std::vector<float> > binnings;
public:
    DataTable data_table;
    ErrorTable error_table;
    std::vector<QualityRule> rule_table;
    virtual int write(FILE *fp);
public:
    QualityTable();
    virtual ~QualityTable();
    virtual int addParameter(std::vector<float> binning);
    virtual void create();
    virtual int  write(const char filename[]);
    virtual int  read(const char filename[]);
    virtual void print();
    virtual void applyScore(DataSet *data, std::vector<int>& scores); 
    virtual void printDataTable();
    virtual void printErrorTable();
    virtual void addData(DataSet &set);
    virtual void addData(DataSet *set);
};

#endif /*QUALITYTABLE_H_*/
