/*****************************************************************************/
// Copyright (c) Illumina 2008
// Author: Richard Shaw
// This source file is covered by the Illumina Genome Analyzer Software
// License Agreement and bound by the terms therein.
/*****************************************************************************/

#include <cctype>
#include <cstdlib>

#include "Message.h"
#include "String_Types.h"
#include "StringUtil.h"

#include "Export_File_Data.h"

/*****************************************************************************/
// N.B. Unlike StringUtil tokenize_str, this returns empty fields.

unsigned int split_on_tab(const std::string str, Str_Vec& str_vec)
{
    const char separ_ch('\t');
    std::string::size_type last_pos(0);
    std::string::size_type pos(0);

    for (std::string::const_iterator ch_citer(str.begin());
         ch_citer != str.end(); ++ch_citer, ++pos) {
        if (*ch_citer == separ_ch) {
            str_vec.push_back(str.substr(last_pos, (pos - last_pos)));
            last_pos = pos + 1;
        }
    }

    if (pos >= last_pos) {
        str_vec.push_back(str.substr(last_pos, (pos - last_pos)));
    }

    return str_vec.size();
}

/*****************************************************************************/

bool apply_mismatch(const std::string& read_str,
                    const std::string& mismatch_str,
                    std::string& ref_str)
{
    unsigned int read_len_left(read_str.length());
    ref_str = read_str;
    std::string::iterator ref_ch_iter(ref_str.begin());
    std::string::const_iterator mismatch_ch_citer(mismatch_str.begin());

    unsigned int count(0);

    for ( ; mismatch_ch_citer != mismatch_str.end(); ++mismatch_ch_citer) {
        const char mismatch_ch(*mismatch_ch_citer);

        if (isdigit(mismatch_ch)) {
            (count *= 10) += (mismatch_ch - '0');
        } else {
            if (count > 0) {
                if (read_len_left >= count) {
                    read_len_left -= count;
                } else {
                    return false;
                }

                ref_ch_iter += count;
                count = 0;
            }

            if (read_len_left) {
                --read_len_left;
            } else {
                return false;
            }

            *ref_ch_iter++ = mismatch_ch;
        }
    }

    if (count != read_len_left) {
        return false;
    }

    return true;
}

/*****************************************************************************/

Export_File_Data::Export_File_Data(const std::string& data_file_name,
                                   const File_Buffer::Compression_Type
                                   compression)
    : Align_Data(data_file_name, compression)
{
    ;
}

/*****************************************************************************/

Export_File_Data::~Export_File_Data()
{
    ;
}

/*****************************************************************************/

bool Export_File_Data::get_next_read_align_info(Read_Align_Info&
                                                read_align_info,
                                                bool& at_eof)
{
    std::string export_file_line_str;
    bool data_line_found(false);

    while (!data_line_found) {
        if (!my_file_buffer.read_line(export_file_line_str, at_eof)) {
            return false;
        }

        if (export_file_line_str.empty()
            || (export_file_line_str[0] == '#')) {
            continue;
        }

        data_line_found = true;

        Str_Vec export_field_str_vec;

        if (split_on_tab(export_file_line_str, export_field_str_vec)
            != NumberOfEntries) {
            error("Found unexpected number of fields (%u) in line `%s'",
                  export_field_str_vec.size(), export_file_line_str.c_str());

            return false;
        }

        read_align_info.my_lane_num 
            = atoi(export_field_str_vec[LANE_NUM_FIELD_IND].c_str());
        read_align_info.my_tile_num 
            = atoi(export_field_str_vec[TILE_NUM_FIELD_IND].c_str());

        // Single-read export files have an empty read number field.
        const std::string 
            read_num_str(export_field_str_vec[READ_NUM_FIELD_IND]);
        read_align_info.my_read_num
            = (read_num_str.empty() ? 1 : atoi(read_num_str.c_str()));

        read_align_info.my_passed_filter_flag
            = (export_field_str_vec[PASSED_FILTER_FIELD_IND] == "Y");

        read_align_info.my_read_str = export_field_str_vec[READ_SEQ_FIELD_IND];


        // Aligned?
        if ((read_align_info.my_aligned_flag 
             = !export_field_str_vec[ALIGN_POS_FIELD_IND].empty())) {
            if (!apply_mismatch(export_field_str_vec[READ_SEQ_FIELD_IND],
                                export_field_str_vec[MISMATCH_FIELD_IND],
                                read_align_info.my_ref_str)) {
                error("Alignment descriptor `%s' "
                      "is incompatible with read `%s'",
                      export_field_str_vec[MISMATCH_FIELD_IND].c_str(),
                      export_field_str_vec[READ_SEQ_FIELD_IND].c_str());
                return false;
            }

            read_align_info.my_align_score
            = atoi(export_field_str_vec[ALIGN_SCORE_FIELD_IND].c_str());
        }
    }

    return true;
}

/*****************************************************************************/

char Export_File_Data::blank_base()
{
    return 'N';
}

/*****************************************************************************/
