/*****************************************************************************/
// Copyright (c) Illumina 2008
// Author: Richard Shaw
// This source file is covered by the Illumina Genome Analyzer Software
// License Agreement and bound by the terms therein.
/*****************************************************************************/

#include <sstream>

#include <getopt.h>

#include "Message.h"
#include "Tile_List.h"

#include "PAS_Options.h"

using namespace CommandLine;

/*****************************************************************************/

PAS_Options::PAS_Options()
    : CommandLine::Options(),
      my_input_type(Align_Data_Factory::EXPORT_INPUT),
      my_input_compr_name("none"),
      my_input_compr(File_Buffer::compressionNone),
      my_tile_list_file_path_str("tiles.txt")
{
    ;
}

/*****************************************************************************/

PAS_Options::~PAS_Options()
{
    ;
}

/*****************************************************************************/

int PAS_Options::getOptions(int argc, char* const argv[])
    throw (CommandLine::UnknownOptionException, 
           CommandLine::MissingOptionException, 
           CommandLine::InvalidArgumentException)
{
    const char* short_option_cstr = "heaz:o:T:";

    const struct option long_option_arr[] =
        { {"help",                   no_argument, 0, 'h'},
          {"export_input",           no_argument, 0, 'e'},
          {"align_input",            no_argument, 0, 'a'},
          {"input_compr",      required_argument, 0, 'z'},
          {"score_file",       required_argument, 0, 'o'},
          {"tile_list_file",   required_argument, 0, 'T'},
          {0, 0, 0, 0}
        };

    int long_option_index(0);
    int option(0);

    bool input_type_specified(false);

    // Note: case 'h' returns immediately.
    while (-1 != (option = getopt_long(argc, argv,
                                       short_option_cstr, long_option_arr,
                                       &long_option_index))) {
        switch (option) {
        case 0:
            break;
        case 'h':
            return usage(argc, argv);
            break;
        case 'e':
            // Input type must be specified exactly once (<1 checked below).
            if (input_type_specified) {
                const std::string
                    msg_str("Multiple specification of input type");
                throw InvalidArgumentException(msg_str);
            }

            my_input_type = Align_Data_Factory::EXPORT_INPUT;
            input_type_specified = true;
            break;
        case 'a':
            // Input type must be specified at most once (<1 checked below).
            if (input_type_specified) {
                const std::string
                    msg_str("Multiple specification of input type");
                throw InvalidArgumentException(msg_str);
            }

            my_input_type = Align_Data_Factory::ALIGN_INPUT;
            input_type_specified = true;
            break;
        case 'z':
            my_input_compr_name = optarg;
            break;
        case 'o':
            my_speced_score_path_str = optarg;
            break;
        case 'T':
            my_tile_list_file_path_str = optarg;
            break;

        case '?':
        case ':':
        default:
            throw UnknownOptionException("");
            break;
        }
    }

    if (!input_type_specified) {
        throw MissingOptionException("No input type specified.");
    }

    return RUN;
}

/*****************************************************************************/

void PAS_Options::processOptions()
    throw (CommandLine::InvalidArgumentException)
{
    std::string msg_str("input ");

    if (!resolve_compression(my_input_compr_name, my_input_compr, msg_str)) {
        throw InvalidArgumentException(msg_str);
    }

    if (my_speced_score_path_str.empty()
        != (my_input_type == Align_Data_Factory::EXPORT_INPUT)) {
        if (my_speced_score_path_str.empty()) {
            msg_str = "Align file input specified with no score file.";
        } else {
            msg_str = "Score file unexpected for export file input.";
        }

        throw InvalidArgumentException(msg_str);
    }
}

/*****************************************************************************/

void PAS_Options::getParameters(int argc, char* const argv[])
    throw (CommandLine::InvalidArgumentException)
{
    while (argc > optind) {
        const std::string arg_str(argv[optind++]);
        my_input_path_str_list.push_back(arg_str);
    }

    // Expecting exactly 1 align input file or 1 or 2 export input files.

    const unsigned int num_input_files(my_input_path_str_list.size());

    if (num_input_files == 0) {
        throw InvalidArgumentException("No input files specified.");
    }

    const unsigned int num_unexpected_args((my_input_type
                                            == Align_Data_Factory::ALIGN_INPUT)
                                           ? (num_input_files - 1)
                                           : ((num_input_files == 1)
                                              ? 0
                                              : (num_input_files - 2)));
    std::string msg_str("");

    switch (num_unexpected_args) {
    case 0:
        return;
    case 1:
        msg_str = "Unexpected argument `";
        break;
    default:
        msg_str = "Multiple unexpected arguments starting with `";
    }

    const unsigned int
        num_expected_args(num_input_files - num_unexpected_args);
    unsigned int arg_ind(0);

    for (Path_Str_List_CIter path_str_citer(my_input_path_str_list.begin());
         path_str_citer != my_input_path_str_list.end();
         ++path_str_citer, ++arg_ind) {
        if (arg_ind == num_expected_args) {
            msg_str += ((*path_str_citer) + "'.");
        }
    }
         
    throw InvalidArgumentException(msg_str);
}

/*****************************************************************************/

int PAS_Options::usage(int num_args, char* const arg_cstr_arr[],
                       const std::string& msg_str)
{
    if (0 != msg_str.size()) {
        error(msg_str.c_str());
    }

    const std::string app_name(arg_cstr_arr[0]);
    std::ostringstream ostrm;

    ostrm << "Usage : " << app_name << " -h|--help" << std::endl
          << "   OR : " << app_name << std::endl
          << " -e|--export_input " << std::endl
          << "    | <-a|--align_input -o|--score_file <score_file>>"
          << std::endl
          << "[-z|--input_compr <gzip|bzip2>]" << std::endl
          << "[-T|tile_list_file <tiles.txt>" << std::endl
          << "<input_file> [<input_file2>]" << std::endl;

    info(ostrm.str().c_str());
    return NO_RUN;
}

/*****************************************************************************/
