/*****************************************************************************/
// Copyright (c) Illumina 2008
// Author: Richard Shaw
// This source file is covered by the Illumina Genome Analyzer Software
// License Agreement and bound by the terms therein.
/*****************************************************************************/

#include <cassert>
#include <sstream>

#include "Message.h"
#include "Tile_Stats.h"

/*****************************************************************************/

const unsigned int Tile_Stats::our_all_reads_read_num = 0;

/*****************************************************************************/

Tile_Stats::Tile_Stats(const unsigned int num_reads)
    : my_lane_num(0), my_tile_num(0), my_num_reads(num_reads),
      my_curr_read_num(0), my_curr_read_num_cycles(0),
      my_non_blank_filter(my_non_blank_mask),
      my_all_stats_set_ptr(0), my_passed_filter_stats_set_ptr(0)
{
    // If there are multiple reads, also include all reads (read num 0).
    unsigned int low_read_num((num_reads > 1) ? our_all_reads_read_num : 1);

    my_quality_type_list.push_back(QUALITY_RAW);
    my_quality_type_list.push_back(QUALITY_FILTERED);

    std::string key_str;

    for (Quality_Type_List_CIter
             quality_type_citer(my_quality_type_list.begin());
         quality_type_citer != my_quality_type_list.end();
         ++quality_type_citer) {
        for (unsigned int read_num(low_read_num);
             read_num <= my_num_reads; ++read_num) {
            make_key_str(*quality_type_citer, read_num, key_str);
            my_key_stats_set_ptr_map[key_str] = new Stats_Set;

            if (read_num == our_all_reads_read_num) {
                my_key_stats_set_ptr_map[key_str]->mark_as_all_reads();
            }
        }
    }
}

/*****************************************************************************/

Tile_Stats::~Tile_Stats()
{
    for (Key_Stats_Set_Ptr_Map_Iter
             pair_iter(my_key_stats_set_ptr_map.begin());
         pair_iter != my_key_stats_set_ptr_map.end(); ++pair_iter) {
        delete pair_iter->second;
        pair_iter->second = 0;
    }
}

/*****************************************************************************/

void Tile_Stats::add_read_align_info(const Read_Align_Info& read_align_info)
{
    if (my_lane_num == 0) {
        // First read.
        if ((read_align_info.my_lane_num < MIN_LANE_NUM)
            || (read_align_info.my_lane_num > MAX_LANE_NUM)) {
            error("Supplied lane number %u is outside valid range %u - %u",
                  read_align_info.my_lane_num, MIN_LANE_NUM, MAX_LANE_NUM);
            throw std::range_error("");
        }

        if (read_align_info.my_tile_num < MIN_TILE_NUM) {
            error("Supplied tile number %u is less than minimum %u",
                  read_align_info.my_tile_num, MIN_TILE_NUM);
            throw std::range_error("");
        }

        if (read_align_info.my_read_num < 1) {
            error("Supplied read num %u is less than 1",
                  read_align_info.my_read_num);
            throw std::range_error("");
        }

        my_lane_num = read_align_info.my_lane_num;
        my_tile_num = read_align_info.my_tile_num;
    } else {
        if (read_align_info.my_lane_num != my_lane_num) {
            error("Unexpected lane number %u (expected %u)",
                  read_align_info.my_lane_num, my_lane_num);
            throw std::range_error("");
        }

        if (read_align_info.my_tile_num != my_tile_num) {
            error("Unexpected tile number %u (expected %u)",
                  read_align_info.my_tile_num, my_tile_num);
            throw std::range_error("");
        }
    }

    const std::string& read_str(read_align_info.my_read_str);
    const std::string& ref_str(read_align_info.my_ref_str);

    if (read_align_info.my_read_num != my_curr_read_num) {
        handle_new_read(read_align_info.my_read_num,
                        read_str.length());
    }

    if (read_str.length() != my_curr_read_num_cycles) {
        error("Unexpected read sequence length %u (expected %u)",
              read_str.length(), my_curr_read_num_cycles);
        throw std::range_error("");
    }


    // Update non-blank totals by cycle.

    std::for_each(read_str.begin(), read_str.end(), my_non_blank_filter);
    my_all_stats_set_ptr->apply_non_blank_mask(my_non_blank_mask);

    if (read_align_info.passed_filter()) {
        my_passed_filter_stats_set_ptr->
            apply_non_blank_mask(my_non_blank_mask);
    }



    if (!read_align_info.is_aligned()) {
        // Nothing more to do.
        return;
    }

    // From this point the read is known to be aligned.

    if (ref_str.length() != my_curr_read_num_cycles) {
        error("Unexpected ref sequence length %u (expected %u)",
              ref_str.length(), my_curr_read_num_cycles);
        throw std::range_error("");
    }

    my_one_read_cum_errs_mask = 0;
    std::string::const_iterator read_ch_citer(read_str.begin());
    std::string::const_iterator ref_ch_citer(ref_str.begin());
    unsigned int read_cum_num_errs(0);
    unsigned int row_start_ele_ind(0);

    for (unsigned int cycle_num(1); cycle_num <= my_curr_read_num_cycles;
         ++cycle_num, ++read_ch_citer, ++ref_ch_citer,
         row_start_ele_ind += Cum_Errs::max_to_count) {
        const char read_ch(*read_ch_citer);
        const char ref_ch(*ref_ch_citer);

        if ((read_ch != the_blank_ch) && (ref_ch != the_blank_ch)) {
            if (read_ch != ref_ch) {
                ++read_cum_num_errs;
            }

            if (read_cum_num_errs <= Cum_Errs::max_to_count) {
                const unsigned int 
                    start_ind(row_start_ele_ind + read_cum_num_errs);
                const unsigned int
                    num_eles(Cum_Errs::max_to_count - read_cum_num_errs);
                my_one_read_cum_errs_mask[std::slice(start_ind, num_eles, 1)]
                    = 1;
            }
        }

        ++(my_all_stats_set_ptr->base_calls()(read_ch, ref_ch, cycle_num));

        if (read_align_info.passed_filter()) {
            ++(my_passed_filter_stats_set_ptr->base_calls()(read_ch, ref_ch,
                                                            cycle_num));
        }
    }

    my_all_stats_set_ptr->add_align_score(read_align_info.my_align_score);
    my_all_stats_set_ptr->cum_errs().vals() += my_one_read_cum_errs_mask;

    if (read_align_info.passed_filter()) {
        my_passed_filter_stats_set_ptr->
            add_align_score(read_align_info.my_align_score);
        my_passed_filter_stats_set_ptr->cum_errs().vals()
            += my_one_read_cum_errs_mask;
    }

    // FIXME : implement pattern stats
}

/*****************************************************************************/

void Tile_Stats::derive_stats()
{
    if (my_num_reads > 1) {
        derive_all_reads_stats();
    }

    for (Key_Stats_Set_Ptr_Map_Iter
             pair_iter(my_key_stats_set_ptr_map.begin());
         pair_iter != my_key_stats_set_ptr_map.end(); ++pair_iter) {
        pair_iter->second->derive_stats();
    }
}

/*****************************************************************************/

bool Tile_Stats::write_score_file(Quality_Type quality_type,
                                  File_Buffer& file_buffer) const
{
    std::string key_str;
    const bool multi_read(my_num_reads > 1);
    const unsigned int num_stat_sets(my_num_reads
                                     + (multi_read ? 1 : 0));

    for (unsigned int stats_set_num(1); stats_set_num <= num_stat_sets;
         ++stats_set_num) {
        unsigned int read_num(stats_set_num);
    
        if (stats_set_num > my_num_reads) {
            // All-reads stats
            read_num = 0;
        }

        File_Buffer::Line_Vec line_vec;
        line_vec.push_back(hdr_str(my_lane_num, my_tile_num,
                                   quality_type, read_num));

        if (!file_buffer.write_some_lines(line_vec)) {
            error("Failed to write statistics set to %s",
                  file_buffer.file_path_str().c_str());
            return false;
        }

        make_key_str(quality_type, read_num, key_str);

        Key_Stats_Set_Ptr_Map_CIter
            ele_citer(my_key_stats_set_ptr_map.find(key_str));
        Stats_Set* stats_set_ptr(ele_citer->second);

        if (!stats_set_ptr->write(file_buffer)) {
            return false;
        }
    }

    return true;
}

/*****************************************************************************/

const std::string Tile_Stats::hdr_str(unsigned int lane_num,
                                      unsigned int tile_num,
                                      Quality_Type quality_type,
                                      unsigned int read_num)
{
    std::stringstream hdr_ostrm;
    hdr_ostrm << "# Lane " << lane_num << " : Tile " << tile_num
              << " : " << qual_str(quality_type) << " : ";

    if (read_num == 0) {
        hdr_ostrm << "All Reads";
    } else {
        hdr_ostrm << "Read " << read_num;
    }

    return hdr_ostrm.str();
}

/*****************************************************************************/

const std::string Tile_Stats::qual_str(Quality_Type quality_type)
{
    std::string str;

    switch (quality_type) {
    case QUALITY_RAW: str = "Raw"; break;
    case QUALITY_FILTERED: str = "Quality Filtered"; break;
    }

    return str;
}

/*****************************************************************************/

void Tile_Stats::make_key_str(Quality_Type quality_type,
                              const unsigned int read_num,
                              std::string& key_str) const
{
    std::ostringstream ostrm;

    switch (quality_type) {
    case QUALITY_RAW: ostrm << "raw"; break;
    case QUALITY_FILTERED: ostrm << "QF"; break;
    }

    ostrm << ": " << read_num;
    key_str = ostrm.str();
}

/*****************************************************************************/

void Tile_Stats::handle_new_read(const unsigned int read_num,
                                 const unsigned int num_cycles)
{
    if ((read_num < 1) || (read_num > my_num_reads)) {
        error("Read number found (%u) is outside expected range (1 - %u)",
              read_num, my_num_reads);
        throw std::range_error("");
    }

    my_curr_read_num = read_num;
    my_curr_read_num_cycles = num_cycles;
    my_non_blank_mask.resize(num_cycles);
    my_one_read_cum_errs_mask.resize(num_cycles * Cum_Errs::max_to_count);

    std::string key_str;

    make_key_str(QUALITY_RAW, read_num, key_str);
    my_all_stats_set_ptr = my_key_stats_set_ptr_map[key_str];
    my_all_stats_set_ptr->set_num_cycles(num_cycles);

    make_key_str(QUALITY_FILTERED, read_num, key_str);
    my_passed_filter_stats_set_ptr = my_key_stats_set_ptr_map[key_str];
    my_passed_filter_stats_set_ptr->set_num_cycles(num_cycles);
}

/*****************************************************************************/

void Tile_Stats::derive_all_reads_stats()
{
    assert(my_num_reads > 1);
    std::string key_str;

    for (Quality_Type_List_CIter
             quality_type_citer(my_quality_type_list.begin());
         quality_type_citer != my_quality_type_list.end();
         ++quality_type_citer) {

        make_key_str(*quality_type_citer, our_all_reads_read_num, key_str);
        Stats_Set* all_reads_stats_set_ptr = my_key_stats_set_ptr_map[key_str];

        unsigned int total_num_cycles(0);

        for (unsigned int read_num(1); read_num <= my_num_reads; ++read_num) {
            make_key_str(*quality_type_citer, read_num, key_str);
            Stats_Set* read_stats_set_ptr = my_key_stats_set_ptr_map[key_str];
            total_num_cycles += read_stats_set_ptr->num_cycles();
        }

        all_reads_stats_set_ptr->set_num_cycles(total_num_cycles);
        unsigned int lo_cycle_num(1);
        unsigned int hi_cycle_num(1);

        for (unsigned int read_num(1); read_num <= my_num_reads; ++read_num) {
            make_key_str(*quality_type_citer, read_num, key_str);
            Stats_Set* read_stats_set_ptr = my_key_stats_set_ptr_map[key_str];
            hi_cycle_num = lo_cycle_num + read_stats_set_ptr->num_cycles() - 1;

            all_reads_stats_set_ptr->add_align_score(*read_stats_set_ptr);
            all_reads_stats_set_ptr->
                base_calls().replace_cycles(lo_cycle_num, hi_cycle_num,
                                            read_stats_set_ptr->base_calls());

            lo_cycle_num = hi_cycle_num + 1;
        }
    }
}

/*****************************************************************************/
