#!/bin/sh

# PROJECT: RAMM
# MODULE: gatherDiagnosticData.sh
# AUTHOR: Gery Vessere
# Copyright (c) 2008 Illumina

# this script packages pipeline files into a zip archive and deposits it into a directory
# from where it will be processed further for ftp-ing by the sendDiagnosticData.sh script
# this is ported from the Windows CopyUtility by Mitch Shiver, you will see some
# setting names that are relics from his scripts and are not handled here.

if [ -z $DEBUG ]
then
	DEBUG=0
fi
E_BADARGS=65

umask a+w 

if [ ! -n "$2" ]
then
  echo "Usage: `basename $0` <report type> <gerald folder>

<report type> is one of AnalysisFilesOnlyReport LocalFilesOnlyReport PerformanceLinkReport PerformanceRunReport ProblemReport.
<gerald folder> is the path to the GERALD folder.
"
  exit $E_BADARGS
fi 


# find out which kind of pipeline version we are on
# and set the folder name that will receive the conf files
THISSCRIPTLOC=$(cd `dirname "$0"` && pwd)
LOC=`echo $THISSCRIPTLOC | grep "bin" | wc -l`

REPORTTYPE=$1
RUNFOLDER=$(cd "$2" && pwd) # gets the absolute path of this folder
GATHER_SCRIPT=1
cd $RUNFOLDER


if [ -h ${RUNFOLDER} ]
then
	log "the gerald folder specified cannot be a symbolic link" 1
fi

MACHINENAME=`uname -n`

# a complicated way to get a log file name...
# in case the scripts get renamed
[[ `basename $0` =~ "(.+).sh" ]]
SCRIPT_BASE=${BASH_REMATCH[1]}

CSU_LOG=$SCRIPT_BASE.log


# prepare for human readable option text conversion to variable name 
IFS="
"
cat $0 | egrep "^#" | grep "ILMN_CSU_" | tr -d "#" > /tmp/conv.tmp

function log
{
	DATE=`date +"%F %T"`
	
        if [ ! -e "$CSU_LOG" ]
       	then 
                touch "$CSU_LOG"
        fi
	echo "gatherDiagnosticData:$DATE:$MACHINENAME:$1" >> $CSU_LOG

	# also output to STDOUT if wanted
	if [[ -n "$2" && "$2" = "1" ]]
	then
		echo "gatherDiagnosticData: $1"
	fi
}

# loads configuration from a file containing "'param' 0/1" lines
# where params are human readable and converts this into the 
# corresponding environment variables (all should begin with
# ILMN_CSU) that have the same
function loadConf
{
	FILE=$1
	ENVNAMESFILE='confVariables.tmp'
	touch $ENVNAMESFILE

	for choice in yes no; 
	do
		for confSetting in $( cat $FILE | perl -pe 's/"\s0/=no/' | perl -pe 's/"\s1/=yes/' | tr -d '"' | egrep "=\s*$choice\s*$" | cut -d"=" -f1 ); do
			TMP=`cat /tmp/conv.tmp | grep "ILMN_CSU" | grep $confSetting | cut -f2`
			if [ ! "$TMP" = "" ]
			then
				VAL=0
				if [ $choice = 'yes' ]
				then 
					VAL=1	
				fi
				echo "$TMP=$VAL" >> $ENVNAMESFILE
			fi
		done;
	done

	# load the variables into the environment and get rid of the temp file
	source $ENVNAMESFILE
	rm -f $ENVNAMESFILE
}

# does the opposite of loadconf and stores into the conf file specified
# these two functions come in handy when updates to the configuration
# need to be done
function storeConf
{
	OLDIFS="$IFS"
	IFS="
	 "
	CONFFILE=$1
	rm -f $CONFFILE
	CONFVARS=${!ILMN_CSU*}
	touch $CONFFILE
	
	for ENV in `echo $CONFVARS`;
	do
		paramName=`cat /tmp/conv.tmp | grep $ENV | cut -f1`
		val=$(eval "echo $`echo $ENV`")
		
		if [ -n "$paramName" ]
		then
			if [ "$val" = "1" ]
			then
				echo "$paramName=yes" >> $CONFFILE
			else
				echo "$paramName=no" >> $CONFFILE
			fi
		fi;
	
	done;
	IFS="$OLDIFS"
}


# takes a runfolder and tries to take a guess at which is the most
# likely value for the FIRECREST/BUSTARD/GERALD variables
function findPaths
{
	RUNFOLDER=$1
	[[ $RUNFOLDER =~ "(.+)\/(.+)\/(.+)\/(.+)\/(.+)$" ]]
	RUNFOLDERNAME=`basename ${BASH_REMATCH[1]}`
	FIRECREST_PATH=${BASH_REMATCH[1]}/${BASH_REMATCH[2]}/${BASH_REMATCH[3]}
	BUSTARD_PATH=$FIRECREST_PATH/${BASH_REMATCH[4]}
	GERALD_PATH=$BUSTARD_PATH/${BASH_REMATCH[5]}

	if [[ -n "$DEBUG" && "$DEBUG" -ge 1 ]]
	then
		echo "FireCrest Path:" $FIRECREST_PATH
		echo "Bustard Path:" $BUSTARD_PATH
		echo "Gerald Path:" $GERALD_PATH
	fi

	if [[ $GERALD_PATH = $RUNFOLDER/// ]]
	then
		# now, this may never happen since the logic changed
		log "cannot find the paths of folders based on the run folder you specified, did you run the complete analysis on this runfolder ?" 1
		exit 
	fi

	if [ -h ${BASH_REMATCH[1]} ]
	then
		log "the runfolder above the gerald folder you specified cannot be a symbolic link" 1
		exit
	fi

}


case "$REPORTTYPE" in
"AnalysisFilesOnlyReport"|"LocalFilesOnlyReport"|"PerformanceLinkReport"|"PerformanceRunReport"|"ProblemReport")
echo ""
;;
*)
log "invalid report type: please make sure to respect case when entering the report type
valid report names are: AnalysisFilesOnlyReport LocalFilesOnlyReport PerformanceLinkReport PerformanceRunReport ProblemReport" 1
exit 
;;
esac


# prepare common variables

findPaths $RUNFOLDER




# configure path for folder to be synced through FTP
# data repository can be configured for all pcs and/or for a particular pc only
# the pc specific setting will always take precedence over the global setting.
# if a .conf file is found in the current path it takes precedence over
# the one that sits in the same folder as the RAMM executable
 
# function getRepos
# {
	# REPOS_ALL=`cat $1 2>&1 | egrep "^\s*Data repository all\s*=" | cut -d"=" -f2` 2>&1 > /dev/null
	# REPOS_PC=`cat $1 2>&1 | grep "Repository\sfor\s$MACHINENAME" | cut -d"=" -f2` 2>&1 > /dev/null
	# REPOS_ALL=${REPOS_ALL%/}
	# REPOS_PC=${REPOS_PC%/}
# }


# checks folder exists and is writable
function checkFolder
{
	if [[ ! -d $1 ]]
	then
		log "$SCRIPT_BASE was not previously configured or some of the configured folders ($1) are not accessible on this machine. Please run the RAMMSetup.sh script in your Pipeline installation." 1
		exit
	fi
	if [[ ! -w $1 ]]
	then
		log "Directory $1 is not writable for this machine/user, please make the folder writable using the 'chmod a+w $1' or 'chmod g+w' (if you can restrict your pipeline users to a specific group name); also if $1 is on a mounted partition, make sure it is mounted read/writable. ${SCRIPT_BASE}.sh will stop execution" 1
		exit
	fi
}

# takes a variable name and expands its content if it contains
# bracketted generic directives
function expandDirectives
{
	variableName=$1
	# [pipelineBin] -> $THISSCRIPTLOC
	f=$(eval "echo $`echo $variableName`")
	f=${f/\[pipelineBin\]/`echo $THISSCRIPTLOC`}
	eval "$variableName=\"$f\""

	# [geraldFolder] -> $GERALD_PATH
	f=$(eval "echo $`echo $variableName`")
	f=${f/\[geraldFolder\]/`echo $GERALDPATH`}
	eval "$variableName=\"$f\""

	# [runEmailServer] -> 
	#f=$(eval "echo $`echo $variableName`")
	#f=${f/\[runEmailServer\]/`echo 'emailServer'`}
	#eval "$variableName=\"$f\""
	
	# [runDomain] -> 
	#f=$(eval "echo $`echo $variableName`")
	#f=${f/\[runDomain\]/`echo 'emailDomain'`}
	#eval "$variableName=\"$f\""
	
	# [runEmails] -> 
	#f=$(eval "echo $`echo $variableName`")
	#f=${f/\[runEmails\]/`echo 'emails'`}
	#eval "$variableName=\"$f\""

}

function readSettings
{
	SETTINGS_FILE=$1
	FTP_FOLDER=`cat $SETTINGS_FILE 2> /dev/null | grep 'FTP Folder' | cut -d"=" -f2`
	SYSCONF_FOLDER=`cat $SETTINGS_FILE 2> /dev/null | grep 'System Configs Folder' | cut -d"=" -f2`
	TRANSFERRED_FOLDER=`cat $SETTINGS_FILE 2> /dev/null | grep 'Transferred Files Folder' | cut -d"=" -f2`
	
	FTPCONFIGS_FOLDER=`cat $SETTINGS_FILE 2> /dev/null | grep 'FTP Configs Folder' | cut -d"=" -f2`
	REPORTCONF_FOLDER=`cat $SETTINGS_FILE 2> /dev/null | grep 'Report Configs Folder' | cut -d"=" -f2`
	FTPSYNC_SCRIPT=`cat $SETTINGS_FILE 2> /dev/null | grep 'FTP Synchronization Script' | cut -d"=" -f2`
	SMTP_SERVER=`cat $SETTINGS_FILE 2> /dev/null | grep 'SMTP Server' | cut -d"=" -f2`
	DOMAIN=`cat $SETTINGS_FILE 2> /dev/null | grep 'Your Domain' | cut -d"=" -f2`
	EMAIL_TO=`cat $SETTINGS_FILE 2> /dev/null | grep 'Emails' | cut -d"=" -f2`

	for variable in FTP_FOLDER SYSCONF_FOLDER TRANSFERRED_FOLDER FTPCONFIGS_FOLDER REPORTCONF_FOLDER FTPSYNC_SCRIPT SMTP_SERVER DOMAIN EMAIL_TO
	do
		expandDirectives $variable
	done	

	if [[ $GATHER_SCRIPT -eq 1 ]]
	then
		for folder in SYSCONF_FOLDER REPORTCONF_FOLDER
		do
			checkFolder $(eval "echo $`echo $folder`")
		done
                mkdir -p $FTP_FOLDER 2> /dev/null
		chmod 777 $FTP_FOLDER 2> /dev/null

	else
		# for the sync script
		for folder in SYSCONF_FOLDER FTPCONFIGS_FOLDER
		do
			checkFolder $(eval "echo $`echo $folder`")
		done
		
		# sync script only: create "Transferred" folder
		mkdir -p "${REPORTCONF_FOLDER}" 2> /dev/null
		chmod a+rw "${REPORTCONF_FOLDER}" 2> /dev/null
		mkdir -p "${TRANSFERRED_FOLDER}" 2> /dev/null
		chmod a+rw "${TRANSFERRED_FOLDER}" 2> /dev/null
		mkdir -p "${FTP_FOLDER}" 2> /dev/null
		chmod a+rw "${FTP_FOLDER}" 2> /dev/null
        fi
}

RAMM_SETTINGS=$SCRIPT_BASE.conf
readSettings $THISSCRIPTLOC/../share/RAMM/$RAMM_SETTINGS

mkdir -p "${FTP_FOLDER}/$MACHINENAME/" 2>&1 > /dev/null
chmod a+rw "${FTP_FOLDER}/$MACHINENAME/" 2> /dev/null

# if [ -n $EMAILSETTINGS ]
# then
	# echo "$EMAILSETTINGS" > $FTP_FOLDER/$MACHINENAME/.email
# fi

TEMP_FOLDER=/tmp

STORE_NAME=${RUNFOLDERNAME}_Files
STORE_PATH=${TEMP_FOLDER}/$STORE_NAME

if [ -d "$STORE_PATH" ]
then
	rm -Rf $STORE_PATH
fi

#fi


# do FTP Sync before reading and updating configuration settings. That's because runs
# tend to be so long that we don't want to see updates to reports configurations  taking
# too long to take effect
if [ ! -z "$FTPSYNC_SCRIPT" ]
then
	# run the command in this file
	# this contains the command to run the sync script on single installs
	`echo $FTPSYNC_SCRIPT`
fi

# configure the report conf folder location

REPORTCONF_MACHINE_FOLDER=${REPORTCONF_FOLDER}/$MACHINENAME
if [ ! -d "${REPORTCONF_MACHINE_FOLDER}" ]
then
	mkdir -p "${REPORTCONF_MACHINE_FOLDER}"
	chmod a+rw "${REPORTCONF_FOLDER}" 2> /dev/null
fi




# read mail configuration settings

if [ ! -e "${REPORTCONF_MACHINE_FOLDER}/$REPORTTYPE.conf" ]
then
	OLDIFS="$IFS"
	IFS=" "

	# create the config file when it doesn't exist
	# here are the default config parameters
	case "$REPORTTYPE" in
	"AnalysisFilesOnlyReport")
		setVars='ILMN_CSU_DATAGATHER ILMN_CSU_FTPZIPFILE ILMN_CSU_SYSCONF ILMN_CSU_EVENTSCRIPTLOG ILMN_CSU_ROBOCOPYLOG ILMN_CSU_PIPELINESUMMARY ILMN_CSU_ALLGRAPHIMAGES ILMN_CSU_CALLGRAPHIMAGES ILMN_CSU_ALLRESULTS ILMN_CSU_SCORERESULT ILMN_CSU_RESCORERESULT ILMN_CSU_MATRIXSUMMARY ILMN_CSU_OFFSETSUMMARY ILMN_CSU_PHASINGSUMMARY ILMN_CSU_QCMFILES ILMN_CSU_PROMPTDATALOC ILMN_CSU_PROMPTORIGINATOR'
		;;
	"LocalFilesOnlyReport")
		setVars='ILMN_CSU_DATAGATHER ILMN_CSU_FTPZIPFILE ILMN_CSU_SYSCONF ILMN_CSU_CURRENTCONFIG ILMN_CSU_DATACOLLECTIONLOGS ILMN_CSU_GALAXYRUNCONF ILMN_CSU_NFNLOG ILMN_CSU_NETCOPYLOG ILMN_CSU_OSLOGS ILMN_CSU_RUNPROCESSESREPORT ILMN_CSU_INSTALLRESULTS ILMN_CSU_PROMPTORIGINATOR'
		;;
	"PerformanceLinkReport")
		setVars='ILMN_CSU_DATAGATHER ILMN_CSU_FTPZIPFILE ILMN_CSU_SYSCONF ILMN_CSU_RUNPARAMS ILMN_CSU_RUNCONFIG ILMN_CSU_RECIPE ILMN_CSU_AFCALIBRATIONRESULTS ILMN_CSU_RUNLOGS ILMN_CSU_GALAXYRUNCONF ILMN_CSU_RUNBROWSERLOG ILMN_CSU_GCLOG ILMN_CSU_PIPELINESUMMARY ILMN_CSU_ALLGRAPHIMAGES ILMN_CSU_CALLGRAPHIMAGES ILMN_CSU_ALLRESULTS ILMN_CSU_SCORERESULT ILMN_CSU_RESCORERESULT ILMN_CSU_MATRIXSUMMARY ILMN_CSU_OFFSETSUMMARY ILMN_CSU_PHASINGSUMMARY ILMN_CSU_QCMFILES ILMN_CSU_PROMPTDATALOC '
		;;
	"PerformanceRunReport")
		setVars='ILMN_CSU_DATAGATHER ILMN_CSU_PIPELINESUMMARY ILMN_CSU_ALLRESULTS ILMN_CSU_SCORERESULT ILMN_CSU_RESCORERESULT ILMN_CSU_PIPELINEDETAILS ILMN_CSU_FTPZIPFILE ILMN_CSU_SYSCONF ILMN_CSU_RUNPARAMS ILMN_CSU_RUNCONFIG ILMN_CSU_RECIPE ILMN_CSU_AFCALIBRATIONRESULTS ILMN_CSU_RUNLOGS ILMN_CSU_GALAXYRUNCONF ILMN_CSU_RUNBROWSERLOG ILMN_CSU_GCLOG ILMN_CSU_MATRIXSUMMARY ILMN_CSU_OFFSETSUMMARY ILMN_CSU_PHASINGSUMMARY ILMN_CSU_QCMFILES '
		;;
	"ProblemReport")
		setVars='ILMN_CSU_DATAGATHER ILMN_CSU_FTPZIPFILE ILMN_CSU_SYSCONF ILMN_CSU_RUNPARAMS ILMN_CSU_CURRENTCONFIG ILMN_CSU_RUNCONFIG ILMN_CSU_RECIPE ILMN_CSU_DATACOLLECTIONLOGS ILMN_CSU_AFCALIBRATIONRESULTS ILMN_CSU_AFCALIBRATIONIMAGES ILMN_CSU_RUNLOGS ILMN_CSU_GALAXYRUNCONF ILMN_CSU_EVENTSCRIPTLOG ILMN_CSU_ROBOCOPYLOG ILMN_CSU_NFNLOG ILMN_CSU_NETCOPYLOG ILMN_CSU_RUNBROWSERLOG ILMN_CSU_RUNBROWSERREPORT ILMN_CSU_GCLOG ILMN_CSU_OSLOGS ILMN_CSU_RUNPROCESSESREPORT ILMN_CSU_INSTALLRESULTS ILMN_CSU_PIPELINESUMMARY ILMN_CSU_ALLGRAPHIMAGES ILMN_CSU_CALLGRAPHIMAGES ILMN_CSU_ALLRESULTS ILMN_CSU_SEQRESULTS ILMN_CSU_SCORERESULT ILMN_CSU_RESCORERESULT ILMN_CSU_MATRIXSUMMARY ILMN_CSU_OFFSETSUMMARY ILMN_CSU_PHASINGSUMMARY ILMN_CSU_QCMFILES ILMN_CSU_PROMPTDATALOC ILMN_CSU_PROMPTORIGINATOR ILMN_CSU_PIPELINEDETAILS '
		;;
	esac

	if [[ "$DEBUG" -ge 1 ]]
	then
		log "actions for $REPORTTYPE: $setVars" 1
	fi


	# set all variables to 0
	for var in `cat /tmp/conv.tmp | cut -f2 | tr "\n" " "`;
	do
		eval "$var=0; export $var"
	done 

	# set the variables that need to be set to 1
	for var in echo $setVars;
	do
		eval "$var=1; export $var"
	done
	
	IFS="$OLDIFS"
	storeConf ${REPORTCONF_MACHINE_FOLDER}/$REPORTTYPE.conf	
fi

	# load the config file
	loadConf ${REPORTCONF_MACHINE_FOLDER}/$REPORTTYPE.conf

	# check if the config file needs updating
	# update it if necessary and acknowledge that update was done
	
	MACHINE_UPDATE_FILE=$MACHINENAME\_$REPORTTYPE\_DataSetUpdate
	ALL_UPDATE_FILE=All_$REPORTTYPE\_DataSetUpdate

	if [[ ! -d "${SYSCONF_FOLDER}/ACK" ]]
	then
		mkdir -p "${SYSCONF_FOLDER}/ACK" 2> /dev/null
		chmod a+rw "${SYSCONF_FOLDER}/ACK" 2> /dev/null
	fi


	# machine update files take precedence over all update files
	# I will use acknowledgement
	# files for machine updates, and the timestamp on those
	# will be compared to the timestamp of the update file
	# so that more than one update can be made to a machine config
	# which doesn't seem to be possible with Mitch's script
	if [[ -e "${SYSCONF_FOLDER}/ACK/ACK_${MACHINE_UPDATE_FILE}" && -e "${SYSCONF_FOLDER}/${MACHINE_UPDATE_FILE}" && "${SYSCONF_FOLDER}/ACK/ACK_${MACHINE_UPDATE_FILE}" -nt "${SYSCONF_FOLDER}/${MACHINE_UPDATE_FILE}" ]]
	then
		log "update already done" 1
		# do nothing since the machine update has already been done and an all update file isn't considered
	elif [[ -e "${SYSCONF_FOLDER}/${MACHINE_UPDATE_FILE}" ]]
	then
		UPDATE_FILE=${SYSCONF_FOLDER}/${MACHINE_UPDATE_FILE}
		TOUCH=${SYSCONF_FOLDER}/ACK/ACK_${MACHINE_UPDATE_FILE}
	elif [[ -e "${SYSCONF_FOLDER}/ACK/ACK_$MACHINENAME_$ALL_UPDATE_FILE" && "${SYSCONF_FOLDER}/ACK/ACK_$MACHINENAME_$ALL_UPDATE_FILE" -nt "${SYSCONF_FOLDER}/$ALL_UPDATE_FILE" ]]
	then
		log "update already done" 1
		# do nothing as well because the all_update has already been done on this machine
	elif [[ -e "${SYSCONF_FOLDER}/$ALL_UPDATE_FILE" ]]
	then
		UPDATE_FILE=${SYSCONF_FOLDER}/$ALL_UPDATE_FILE
		TOUCH=${SYSCONF_FOLDER}/ACK/ACK_$MACHINENAME_$ALL_UPDATE_FILE
	else
		log "no updates available for reports" 0
	fi
	
	# do the actual updating of config work
	
	# echo "update file: $UPDATE_FILE
# touch: $TOUCH
# update folder: $SYSCONF_FOLDER
# machine update file: ${MACHINE_UPDATE_FILE}
# all update file: $ALL_UPDATE_FILE";
	
	if [[ -n "${UPDATE_FILE}" && "${UPDATE_FILE}" -nt "$TOUCH" ]]
	then
		loadConf ${UPDATE_FILE} 
		storeConf ${REPORTCONF_MACHINE_FOLDER}/$REPORTTYPE.conf
		touch $TOUCH
		log "updated configuration for report $REPORTTYPE" 1
	fi


# now do the actual copying work	

IFS=" "

# take care of the prompt variables first
PROMPTVARS=${!ILMN_CSU_PROMPT*}
if [ ! -d "${STORE_PATH}" ]
then
	mkdir "${STORE_PATH}" 2> /dev/null
	chmod a+rw "${STORE_PATH}" 2> /dev/null
fi

README=${STORE_PATH}/README.txt
rm -f $README
touch $README

echo "Report date: `date`
run folder: $RUNFOLDER
************************
 " >> $README

if [ ${ILMN_CSU_DATAGATHER} -eq 1 ]
then
	for ENV in `echo $PROMPTVARS`;
	do
	case $ENV in
	"ILMN_CSU_PROMPTORIGINATOR")
		val=$(eval "echo $`echo $ENV`")

	        if [ "$val" = "1" ]
	        then

			read -p "please enter your name: " NAME
			read -p "please enter your email address: " EMAIL
			read -p "please enter the description of your problem (on one line): " PROBLEM
		
		        echo "User Name: " $NAME >> $README
		        echo "User Email Address: " $EMAIL >> $README
		        echo "Problem Description: " $PROBLEM >> $README
		fi
	;;
	esac          
	done

fi


CONFVARS=${!ILMN_CSU*}

for ENV in `echo $CONFVARS`;
do
	# grab the parameters for the copy
	# findSpec will be arguments to find
	# dest will be arguments to ln
	copySpec=`cat /tmp/conv.tmp | grep $ENV`
	label=`echo $copySpec | cut -f1`
	OLDIFS="$IFS"
	export IFS=" "

	findSpecs=( `echo $copySpec | cut -f3 | tr " " "@" | tr ":" " "` )        
	dehydrationCommands=(`echo $copySpec | cut -f4 | tr " " "@" | tr ":" " "`)
	dests=(`echo $copySpec | cut -f5 | tr " " "@" | tr ":" " "`)
	IFS="$OLDIFS"
	
	val=$(eval "echo $`echo $ENV`")
	
	if [ "$val" = "1" ]
        then		
		if [[ -n "$findSpecs" && "$findSpecs" != "" ]]
		then
			case $ENV in

			
			# not a good idea to ask for location of run_folder under linux where the script may 
			# run in a non-interactive environment, this should be specified on the command line
			"ILMN_CSU_PROMPTDATALOC" | "ILMN_CSU_PROMPTORIGINATOR")
				# do nothing
				;;
			*)	
				i=0
				specsCount=${#findSpecs[*]}

				if [[ $DEBUG -ge 3 ]]
				then
					echo "specCounts: $specsCount"
				fi
				
				if [ $ILMN_CSU_DATAGATHER -eq 1 ]
				then
					while [ $i -lt $specsCount ]; do
						findSpec=`echo ${findSpecs[$i]} | tr "@" " "`
						dest=`echo ${dests[$i]} | tr "@" " "`
						dehydrationCommand=`echo ${dehydrationCommands[$i]} | tr "@" " "`
						
						
						if [[ $DEBUG -ge 2 ]]
						then
							echo "findSpec: $findSpec"
							echo "dest: $dest" 
							echo "dehydrationCommand: $dehydrationCommand"
						fi
						
						IFS="
"				
						# do the actual finding and copying of files (changed to a mere linking operation)
						rm -f copyErrors.tmp 
						files=$( eval find `echo $findSpec` 2>> /dev/null )
						
						if [ $specsCount -gt 1 ]
						then
							c=`expr $i + 1`
							spec=" spec #$c"						
						else
							spec=""
						fi
						
						log "gathering run data for $label$spec" 1
			
						for file in $files
						do
							if [ ! -d $file ]
							then
								d=$( eval echo "$STORE_PATH/`echo $dest`" )
								mkdir -p $d 2>&1
								chmod a+rw "$d" 2>&1
				
								if [[ $DEBUG -ge 2 ]]
								then
									log "file copy simulation..." 1 
									
								else
									if [ -n "$dehydrationCommand" ]
									then
										eval "cat $file | $dehydrationCommand" > "$d/`basename $file`"  2>&1
									else
										ln -s "$file" "$d/`basename $file`" 2>&1
									fi
								fi
							fi
							
						done >> copyErrors.tmp
					
						if [ -z "$files" ]
						then
							log "no files found for this copy spec: $findSpec" 0 
						fi
						
						# process errors into the log file
						if [ -s copyErrors.tmp ]
						then
							log ">>> some errors occured while copying files" 1
							echo "consult $CSU_LOG for more details"
							log `cat copyErrors.tmp` 
						
							# copy README errors to the zip file
							if [ ! -e $README ]
							then
								touch $README
							fi	
							cat copyErrors.tmp >> $README
						fi	
						rm -f copyErrors.tmp
						
						let i=i+1
					done	
				fi
			esac
#else	
			#log "nothing to do" 1;
		fi
	fi
done;

log "Finished gathering files for the $REPORTTYPE report" 1


log "zipping run files into the ${FTP_FOLDER}/$MACHINENAME/ folder (this may take a while...)" 1
#cd `dirname ${STORE_PATH}` > /dev/null
	
ZIPNAME=${STORE_NAME}

if [ $DEBUG -ge 1 ]
then 
	# for Illumina internal testing purposes only
	# this appends some info to the zip file name
	export IPARRUN=`expr "$RUNFOLDER" : ".*IPAR"`
	export RUNTYPE=`echo "$RUNFOLDER" | perl -ne '/(SR|SFPE|DFPE)/; print $1;'`
	
	
	if [ $IPARRUN -ne 0 ]
	then
		export IPAR="-IPAR"
	else
		export IPAR=""
	fi
	
	BZ=`echo "$ZIPNAME" | perl -ne '/^(.+?)_(.+?)_(.+?)_(.+?)(_.+?)?$/; print "$1_$2"'`
	EZ=`echo "$ZIPNAME" | perl -ne '/^(.+?)_(.+?)_(.+?)_(.+?)(_.+?)?$/; print "\_$3_$4$5"'`
	CY=`echo "$RUNFOLDER" | perl -ne '/(C\d-\d\d?)/; print "$1"'`
	ZIPNAME="${BZ}-${RUNTYPE}-${CY}${IPAR}${EZ}"
	
	log "zip file name is $ZIPNAME" 1
fi

if [ "$ILMN_CSU_DATAGATHER" -eq 1 ]
then
	if [ "$ILMN_CSU_FTPZIPFILE" -eq 1 ]
	then
		cd $STORE_PATH
		zip -r ${FTP_FOLDER}/$MACHINENAME/$ZIPNAME.zip * > /dev/null
		cd - > /dev/null
	else
		FORREVIEW_FOLDER="${TRANSFERRED_FOLDER}/../forReview"
		if [ ! -d "${FORREVIEW_FOLDER}" ]
		then
			mkdir "${FORREVIEW_FOLDER}" 2> /dev/null
			chmod a+rw "${FORREVIEW_FOLDER}" 2> /dev/null
		fi
		FORREVIEW_FOLDER=$(cd "${FORREVIEW_FOLDER}" && pwd)
		log "zip file was placed under ${FORREVIEW_FOLDER}/$STORE_NAME.zip and will not be sent through ftp automatically. You may review its contents and send it manually through FTP using the following credentials: 
-ftp site: ftp-hayward.illumina.com
-login: fas
-password: 1fas6781
-folder: Reports" 1
		cd ${STORE_NAME}
		zip -r ${FORREVIEW_FOLDER}/$ZIPNAME.zip ${STORE_NAME}/* > /dev/null
		cd - > /dev/null
	fi
fi

rm -Rf ${STORE_PATH}
	
	
# do some clean up
rm /tmp/conv.tmp


# check if ftp sync can be done now
if [ -e "${FTPSYNC_SCRIPT}" ] 
then
	# run the command in this file
	# this contains the command to run the sync script on single installs
	`echo ${FTPSYNC_SCRIPT}`
fi


# if adding parameters to this script, make sure the variable names begin with
# ILMN_CSU since the logic of the script depends on this.
# you need to add the setting in question to the default settings of the
# reports you wish to add this setting to, finally you will need to delete
# the .conf file (a new default will be created automatically)
# the variable names in the second column could be randomly generated, but it doesn't make
# things easy to configure the standard report contents.

# column #1: name of the human readable setting in the config file
# column #2: variable associated to the setting in RAMM
# column #3: find specification
# column #4: filter specification usually a perl command executed on each of the files found with find
# column #5: destination in zip file

#Enable Data Gather	ILMN_CSU_DATAGATHER			
#FTP Zipped File	ILMN_CSU_FTPZIPFILE		
#Prompt user for data location	ILMN_CSU_PROMPTDATALOC			
#Prompt user for originator data	ILMN_CSU_PROMPTORIGINATOR				
#Run Parameters (.PARAMS)	ILMN_CSU_RUNPARAMS	$FIRECREST_PATH/.params -name ".params":$FIRECREST_PATH/../.params -name ".params":$FIRECREST_PATH/../../*.params -name "*params"		Pipeline/RunParams/FC:Pipeline/RunParams/Data:Pipeline/RunParams/RunFolder
#Run Configuration Files (.XML)	ILMN_CSU_RUNCONFIG	$FIRECREST_PATH/../../IPAR_Config -name "*":$FIRECREST_PATH/../../Config/ -name "*.xml":$FIRECREST_PATH/../../ReadPrep1/Config -name "*.xml":$FIRECREST_PATH/../../ReadPrep2/Config -name "*.xml"		RunConfig/IPAR:RunConfig:RunConfig/ReadPrep1:RunConfig/ReadPrep2
#Recipe (.XML)	ILMN_CSU_RECIPE	$FIRECREST_PATH/../.. -name "Recipe*xml"		.
#AutoFocus Calibration Result (.TXT)	ILMN_CSU_AFCALIBRATIONRESULTS	$FIRECREST_PATH/../../Calibration* -name "AFCalResult.txt"		Calibration
#RunLogs (.XML)	ILMN_CSU_RUNLOGS	$FIRECREST_PATH/../../RunLog* -name "RunLog*xml":$FIRECREST_PATH/../../ReadPrep1 -name "RunLog*xml":$FIRECREST_PATH/../../ReadPrep2 -name "RunLog*xml"		.:ReadPrep1RunLogs:ReadPrep2RunLogs
#RunBrowser Logs (.XML)	ILMN_CSU_RUNBROWSERLOG	$FIRECREST_PATH/../RunBrowser -name "s*bro.xml"		RunBrowser
#Run Details Excel File (.XLS)	ILMN_CSU_PIPELINEDETAILS	$FIRECREST_PATH/../.. -name "*run_details*.xls"		Pipeline
#Pipeline Summary (.HTM)	ILMN_CSU_PIPELINESUMMARY	$GERALD_PATH -name "Summary.htm"		Pipeline
#'All' Graph Images (.PNG)	ILMN_CSU_ALLGRAPHIMAGES	$GERALD_PATH -name "s_?_all.png"		Pipeline/All
#'Call' Graph Images (.PNG)	ILMN_CSU_CALLGRAPHIMAGES	$GERALD_PATH -name "s_?_call.png"		Pipeline/Call
#'All' Result Files (.TXT)	ILMN_CSU_ALLRESULTS	$GERALD_PATH -name "s_*_all.txt":$BUSTARD_PATH/SignalMeans -name "s_*_all.txt"		Pipeline/All:Pipeline/All
#'Score' Result Files (.TXT)	ILMN_CSU_SCORERESULT	$GERALD_PATH -name "s_?_????_score.txt"	tr "\n" "|" | perl -pe 's/^.*?(Breakdown of errors.*?)Error.*$/$1/i' | tr "|" "\n"	Pipeline/Score
#'Rescore' Result Files (.TXT)	ILMN_CSU_RESCORERESULT	$GERALD_PATH -name "s_?_????_rescore.txt"	tr "\n" "|" | perl -pe 's/^.*?(Breakdown of errors.*?)Error.*$/$1/i' | tr "|" "\n"	Pipeline/Rescore
#Matrix Summary Files (.TXT)	ILMN_CSU_MATRIXSUMMARY	$FIRECREST_PATH/Matrix/:$BUSTARD_PATH/Matrix/ -name "s_??_summary.txt"		Pipeline/Matrix:Pipeline/Matrix
#Offset Summary Files (.TXT)	ILMN_CSU_OFFSETSUMMARY	$FIRECREST_PATH/Offsets/ -name "offset_summary.txt"		Pipeline/Offsets
#Phasing Summary Files (.TXT)	ILMN_CSU_PHASINGSUMMARY	$BUSTARD_PATH/Phasing/ -name "s_??_phasing.txt"		Pipeline/Phasing
#'QCM' Files (.XML)	ILMN_CSU_QCMFILES	$FIRECREST_PATH/Firecrest/ -name "s_*_qcm.xml"		Pipeline/QCM
