package Gerald::Common;
use Exporter;
@ISA=("Exporter");
@EXPORT=qw(&getShortRunFolderName &getRunFolderName &getTileList &gnuplotImage &getReadID &expandUseBasesString &expandLegacyUseBasesString &parseUseBases &pad_seq_with_masked_cycles &getCycleMap &updateCycleOffsets &compareVersions &parseAnalysisDirName &getPhasingOptions &getReadStartCycles &getReadLengths &getAutoFilename &getPhasingFilename &convertProbToQual &convertProbCorrectToPhred &convertQualToProbCorrect &convertQualToProbWrong &getQualFromChar &getMatchDetails &apply_mismatch &getExactScoreForRead &adjustScoreForIncorrectBase &getWorstBases &getAlignmentScore &scoreAlignFromNhood &bestAlignFromList &pickBestSingleReadAlignment &getMeanBaseQuality &expandSymbolicQualityString &sigpipe_handler &conditionally_init_sigpipe_handler &conditionally_report_sigpipes &trim_external_Ns &lane_prefix &lane_num &get_tiles_by_lane &get_expected_tile_nums &determine_missing_tile_nums &getSmtFilter);

#------------------------------------------------------------------------------

# PROJECT: GERALD
# MODULE:  Common.pm
# AUTHOR:  A. J. Cox
#
# Copyright (c) 2007 Solexa
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# Functions common to the GERALD module
use warnings;
use strict;
use POSIX;

{
    my $cvs_id='@(#) $Id: Common.pm,v 1.3 2008/07/22 17:38:19 craczy Exp $';
    my $cvs_version='$Revision: 1.3 $';
    # Make CVS tag into a displayable string
#    $version=$cvs_version;
#    $version=~s/\$//g;
#    $version=~s/Revision//;
#    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope

my $pathGnuplot = "gnuplot";
my $pathConvert = "convert";


# Make a short run folder name to use in the read ID
# Strip off date and try to shorten other fields
sub getShortRunFolderName
{
    die unless((@_==1)||(@_==2));
    my ($longRunFolderName, $readPairMode)=@_;
    return "s" unless defined($longRunFolderName);
    $longRunFolderName=~s/^\d{6}_//; # strip date
    $longRunFolderName=~s/^slxa-//; # strip unnecessary machine name
    $longRunFolderName=~s/_0+/_/; # strip leading zeroes
# remove any _FCxxxx idiocy at end
# TC 18.6.7 - also remove any _R1 or _R2 suffixes. Issue is that both read 1
# and read 2 end up suffixed _R1 because everything gets named after the first
# run folder
    $longRunFolderName=~s/_[Ff][Cc]\d+(_[Rr][12])?$//; 

    if (defined($readPairMode))
    {
	die "Invalid value $readPairMode for readPairMode - must be 0,1 or 2"
	    unless (($readPairMode eq '0')
		    ||($readPairMode eq '1')
		    ||($readPairMode eq '2'));
	$longRunFolderName.="-$readPairMode"
	    if (($readPairMode eq '1')||($readPairMode eq '2'));
    } # if
    return $longRunFolderName;
} # sub getShortRunFolderName


# extract the name of the run-folder from a directory path within the 
# run-folder
sub getRunFolderName($) {
    my $path = shift;
    my $name = undef;

    if ($path =~ m#/(\d{6}_[^/_]+_\d{4}[^/]*)/#) {
	$name = $1;
    }
    if ($path =~ m#/(\d{8})/#) {
	$name = $1;
    }
    return $name;
}

# get a list of filenames from a Gerald folder with a certain suffix
#  indexed by lanes and tile
sub getTileList($$) {
    my $folder = shift;
    my $suffix = shift;

    my $file;
    my %results;
    opendir (DIR, $folder) or die "can't open ${folder}.\n";
    while (defined($file = readdir (DIR))) {
	next unless ($file=~/s_([1-9])_([0-9]{4})${suffix}/);
	$results{$1}{$2} = $file;
    }

    return %results;
}

sub gnuplotImage($$) {
    my $gnuplotInputFile = shift;
    my $imageFile        = shift;
    system ("${pathGnuplot} ${gnuplotInputFile} > ${gnuplotInputFile}.ps");
    system ("${pathConvert} ps:${gnuplotInputFile}.ps $imageFile");
    unlink ("${gnuplotInputFile}.ps");
}


# Make a read ID of the form b2_537:2:8:222:333
# For paired reads use b2_537:2:8:222:333-1 or b2_537:2:8:222:333-2
# Input - output of getShortRunFolderName plus lane, tile, x, y
# Optional readPairMode appends '-1' or '-2' to readID
sub getReadID
{
    die unless ((@_==6)||(@_==7));
    my ($shortRunFolderName,$lane,$tile,$x,$y,$readPairMode,$indexBases)=@_;
    my $readID="${shortRunFolderName}:${lane}:${tile}:${x}:${y}";
    if (defined($indexBases))
    {
	$indexBases=~s/\./N/g;
	$readID.='#'.$indexBases;
    } # if

    if (defined($readPairMode))
    {
	die "Invalid value $readPairMode for readPairMode - must be 0,1 or 2"
	    unless (($readPairMode eq '0')
		    ||($readPairMode eq '1')
		    ||($readPairMode eq '2'));
	$readID.="/$readPairMode"
	    if (($readPairMode eq '1')||($readPairMode eq '2'));
    } # if
    return $readID;
} # sub getReadID



# Convert a USE_BASES parameter from short to long form
# eg nY5N2y3 becomes nYYYYYNNyyy
sub expandUseBasesString
{
    my ($useBasesString)=@_;
    my $a;
    while($useBasesString=~/([YyNnIi])(\d+)/) 
    { 
	$a=$1 x $2; 
	$useBasesString=~s/$&/$a/; 
    } # while
    return $useBasesString;
} # sub expandUseBasesString

#------------------------------------------------------------------------------

sub expandLegacyUseBasesString($$)
{
    my $useBases = shift;
    my $readLength = shift;

    if ($useBases eq "all")
    {
	die "$0: USE_BASES 'all' specified but invalid read length $readLength given!"
	    if ($readLength<=0); 
        $useBases= "Y" x $readLength;
    } # if
    elsif ($useBases eq "odd")
    {
	die "$0: USE_BASES 'odd' specified but invalid read length $readLength given!"
	    if ($readLength<=0); 
        $useBases= "YN" x ($readLength);
        $useBases=substr($useBases,0,length($useBases)-1);
    } # elsif
    else
    {
        die "Use bases string $useBases must be odd, all or YNYN..."
            unless ($useBases=~/^[YNyn]+$/);
    } # else

    return $useBases;
}

#------------------------------------------------------------------------------

# sub parseUseBases
# Parse a long form USE_BASES parameter into an array of indices
# Inputs:
# 1. USE_BASES string - either: 
# 'all' - use all bases (probably an anachronism)
# 'odd' - use odd numbered bases (definitely an anachronism)
# Some combination of 'Y', 'y', 'N' and 'n'
# 2. read length - used in 'all' or 'odd' modes to generate a string
# of the right length, also used to check there are enough 'Y/y's in 
# USE_BASES - again probably an anachronism, set to 0 to ignore
# 3. (optional) Which read of a paired read - either
# '1' - first read: treat all 'y's as 'n's
# '2' - second read: treat all 'Y's as 'N's
sub parseUseBases
{
    die unless ((@_==2)||(@==3));
    my ($useBases, $readLength, $readPairMode)=@_;
    $readPairMode="0" unless (defined($readPairMode));
    my @baseIndex;

    $useBases = expandLegacyUseBasesString($useBases, $readLength);

    if ($readPairMode eq "1")
    {
	$useBases=~tr/y/N/;
    } # if
    elsif ($readPairMode eq "2")
    {
	$useBases=~tr/Y/N/;
    } # elsif

    $useBases = uc ($useBases);


    my @ub=split(//,$useBases);
  #  my @baseIndex;
    for (my $i=0;$i<@ub;$i++)
    {
        push (@baseIndex, $i) if ($ub[$i] eq "Y"); 
    } # if

    # Only check read length with use bases if positive read length given
    if ($readLength>0)
    {
	die "Need $readLength bases of base information, only ", 
	scalar(@baseIndex), " specified in $useBases"
	    unless (scalar(@baseIndex)>=$readLength);
	@baseIndex=splice(@baseIndex,0,$readLength);
    } # if
    return @baseIndex;
} # sub parseUseBases

#------------------------------------------------------------------------------
# Pads a supplied sequence string with the specified character at all 
# positions previously masked out by a USE_BASES mask.
# The USE_BASES mask must be specified in the array-of-cycle-indices format
# returned by parseUseBases in conjunction with the original read length
# (the same as the length of the USE_BASES string).
#
# [The expectation is that USE_BASES will be pre-parsed before calling this
#  function many times.]

sub pad_seq_with_masked_cycles($$$$)
{
    my $seq_str = shift;
    my $read_used_cycle_inds_ref = shift;
    my $read_orig_length = shift;
    my $pad_ch = shift;

    if (length($seq_str) != scalar(@$read_used_cycle_inds_ref)) {
	my $seq_str_len = length($seq_str);
	my $expected_length = scalar(@$read_used_cycle_inds_ref);

	die("Sequence `$seq_str' is of length $seq_str_len "
	    . "- expected length $expected_length");
    }

    my $padded_str = $pad_ch x $read_orig_length;
    my @padded_str_chs = split(//, $padded_str);
    my @seq_chs = split(//, $seq_str);
    @padded_str_chs[@$read_used_cycle_inds_ref] = @seq_chs;
    $padded_str = join('', @padded_str_chs);

    return $padded_str;
}

#------------------------------------------------------------------------------
# Assumes fully expanded USE_BASES, i.e. only [NYny]* with `Y's strictly 
# before `y's.
# Note that, for convenience, @cycle_map is indexed by (1-offset) cycle number 
# so the first element (cycle 0) is undefined and should be ignored.

sub getCycleMap($$)
{
    my $use_bases_str = shift;
    my $cycle_map_ref = shift;

    my $read_num = 0;
    my $max_read_num = 0;

    my @cycle_nums;

    my @use_bases_chars = split(//, $use_bases_str);
    my $base_ind = 1;

    foreach my $use_bases_char (@use_bases_chars) {
	$read_num = 0;

	if ($use_bases_char eq 'Y') {
	    $read_num = 1;

	    if ($max_read_num == 2) {
		warn "Found `Y' after `y' in use_bases $use_bases_str"
		    . " after base $base_ind\n";
		return 0;
	    }
	} elsif ($use_bases_char eq 'y') {
	    $read_num = 2;

	    if ($max_read_num == 0) {
		warn "Found `y' before any `Y' in use_bases $use_bases_str"
		    . " after base $base_ind\n";
		return 0;
	    }
	} elsif (!(($use_bases_char eq 'N') || ($use_bases_char eq 'n'))) {
	    warn "Found unexpected char `$use_bases_char' "
		. "in use_bases $use_bases_str after base $base_ind\n";
	    return 0;
	}

	if ($read_num > $max_read_num) {
	    $max_read_num = $read_num;
	}

	if (!defined($cycle_nums[$read_num])) {
	    $cycle_nums[$read_num] = 0;
	}

	$cycle_map_ref->[$base_ind]{read_num} = $read_num;
	$cycle_map_ref->[$base_ind]{cycle_num} = ++$cycle_nums[$read_num];

	++$base_ind;
    }

    return $max_read_num;
}

#------------------------------------------------------------------------------
# @cycle_offsets is indexed by 1-offset $read_num.

sub updateCycleOffsets($$$)
{
    my $cycle_offsets_ref = shift;
    my $curr_read_num = shift;
    my $num_cycles_in_curr_read = shift;

    if (defined($cycle_offsets_ref->[$curr_read_num + 1])) {
        return;
    }

    if ($curr_read_num == 1) {
        $cycle_offsets_ref->[$curr_read_num] = 0;
    }

    if (!defined($cycle_offsets_ref->[$curr_read_num])) {
        die("Cannot calculate cycle offset for read $curr_read_num "
            . "as offset for previous read is unknown\n");
    }

    $cycle_offsets_ref->[$curr_read_num + 1]
        = $cycle_offsets_ref->[$curr_read_num] + $num_cycles_in_curr_read;
}

#------------------------------------------------------------------------------
# Takes refs to two version triples (major, minor, release).
# Returns : (v2 > v1) : 1; (v2 < v1) : -1; (v2 == v1) : 0
#           or undef if either specified array is not a triple.

sub compareVersions($$)
{
    my $ver1_ref = shift;
    my $ver2_ref = shift;

    my @ver1 = @$ver1_ref;
    my @ver2 = @$ver2_ref;

    my $num_ver_pieces = 3;

    if ((scalar(@ver1) != $num_ver_pieces) 
	|| (scalar(@ver1) != $num_ver_pieces)) {
	return undef;
    }

    for (my $piece_ind = 0; $piece_ind < $num_ver_pieces; ++$piece_ind) {
	my $ver1_piece = $ver1[$piece_ind];
	my $ver2_piece = $ver2[$piece_ind];

	if ($ver2_piece < $ver1_piece) {
	    return -1;
	} elsif ($ver2_piece > $ver1_piece) {
	    return 1;
	}
    }

    return 0;
}

#------------------------------------------------------------------------------
# Parse the name of an analysis directory Firecrest / Bustard / Gerald,
# returning information extracted as a hash.
# Possible keys : software_name, major_ver, minor_ver, release, user, date,
#                 first_cycle, last_cycle

sub parseAnalysisDirName($$)
{
    my $dir_name = shift;
    my $dir_info_ref = shift;

    my @dir_pieces = split('_', $dir_name);

    if (@dir_pieces < 3) {
	return 0;
    }

    my $date = $dir_pieces[-2];
    my @date_pieces = split('-', $date);

    if (@date_pieces != 3) {
	return 0;
    }

    $dir_info_ref->{user} = $dir_pieces[-1];
    $dir_info_ref->{date} = $date;

    my $soft_name_ver = $dir_pieces[-3];

    if ($soft_name_ver =~ /([a-zA-Z]+)(\d+)\.(\d+)\.(\d+)/) {
	$dir_info_ref->{software_name} = $1;
	$dir_info_ref->{major_ver} = $2;
	$dir_info_ref->{minor_ver} = $3;
	$dir_info_ref->{release} = $4;
    } else {
	$dir_info_ref->{software_name} = $soft_name_ver;
    }

    if (@dir_pieces > 3) {
	my $cycle_str = $dir_pieces[-4];

	if ($cycle_str =~ /C(\d+)-(\d+)/) {
	    $dir_info_ref->{first_cycle} = $1;
	    $dir_info_ref->{last_cycle} = $2;
	}
    }

    return 1;
}

#------------------------------------------------------------------------------
# Parse a Bustard .params file (i.e. one in a parent Firecrest folder)
# returning a reference to the Run substructure for the specified
# Bustard directory.

sub getBustardRunInfo($$$$$) {
    my $xml_file_path = shift;
    my $bustard_dir = shift;
    my $rel_xpath_str = shift;
    my $force_array_keys_ref = shift;
    my $key_attr_ref = shift;

    if (! -f $xml_file_path) {
	return undef;
    }

    my $xml_ref = new XML::Simple(suppressempty => '',
				  XMLDecl => 1);

    my @force_array_keys = ('Run', @$force_array_keys_ref);
    my %key_attr_map = %$key_attr_ref;
    $key_attr_map{'Run'} = 'Name';

    my $cfg_hash_ref = $xml_ref->XMLin($xml_file_path,
				       forcearray => \@force_array_keys,
				       keyattr => \%key_attr_map);

    if (!defined($cfg_hash_ref)) {
        return undef;
    }

    my @rel_xpath_dirs = split('/', $rel_xpath_str);
    my @layer_names = ('Run', $bustard_dir, @rel_xpath_dirs);
    my $curr_hash_ref = $cfg_hash_ref;
    my $next_hash_ref;

    foreach my $layer_name (@layer_names) {
	my $next_hash_ref = $curr_hash_ref->{$layer_name};

	if (!defined($next_hash_ref)) {
	    return undef;
	}

	$curr_hash_ref = $next_hash_ref;
    }

    return $curr_hash_ref;
}

#------------------------------------------------------------------------------
# Parse a Bustard .params file (i.e. one in a parent Firecrest folder)
# to extract a hash of phasing options indexed by read and then by option.

sub getPhasingOptions($$$)
{
    my $xml_file_path = shift;
    my $bustard_dir = shift;
    my $options_ref = shift;

    my $rel_xpath_str = 'BaseCallParameters/Phasing';
    my @force_array_keys = ('Phasing');
    my %key_attr_map = ('Phasing' => 'Read');

    my $curr_hash_ref = getBustardRunInfo($xml_file_path, $bustard_dir,
					  $rel_xpath_str,
					  \@force_array_keys, \%key_attr_map);

    if (!defined($curr_hash_ref)) {
	return 0;
    }

    %$options_ref = %$curr_hash_ref;

    return 1;
}

#------------------------------------------------------------------------------
# Parse a Bustard .params file (i.e. one in a parent Firecrest folder)
# to extract a hash of SMT filter parameters
sub getSmtFilter($$$)
{
    my $paramsfile   = shift;
    my $bustarddir   = shift;
    my $filterRef    = shift;
    my $paramsRef    = getBustardRunInfo($paramsfile, $bustarddir, 'RunParameters', [], {});
    return 0 unless defined($paramsRef);
    $filterRef->{filter}    = $paramsRef->{SmtFilter};
    $filterRef->{relation}  = $paramsRef->{SmtRelation};
    $filterRef->{threshold} = $paramsRef->{SmtThreshold};
    return 1;
}

#------------------------------------------------------------------------------
# Parse a Bustard .params file (i.e. one in a parent Firecrest folder)
# to extract the value of the chastity threshold
sub getChastityThreshold($$)
{
    my $paramsfile   = shift;
    my $bustarddir   = shift;
    my $paramsRef    = getBustardRunInfo($paramsfile, $bustarddir, 'RunParameters', [], {});
    return $paramsRef->{ChastityThreshold};
}


#------------------------------------------------------------------------------
# Parse a Bustard .params file (i.e. one in a parent Firecrest folder)
# to extract an array of start cycles (e.g. two for Paired End reads)
# for the specified Bustard directory.

sub getReadStartCycles($$$) {
    my $xml_file_path = shift;
    my $bustard_dir = shift;
    my $read_start_cycles_ref = shift;

    my $rel_xpath_str = 'RunParameters/Reads';
    my @force_array_keys = ('Reads');
    my %key_attr_map = ('Reads' => 'Index');

    my $curr_hash_ref = getBustardRunInfo($xml_file_path, $bustard_dir,
					  $rel_xpath_str,
					  \@force_array_keys, \%key_attr_map);

    if (!defined($curr_hash_ref)) {
	return 0;
    }

    foreach my $read_key (keys %$curr_hash_ref) {
	my $read_hash_ref = $curr_hash_ref->{$read_key};
	my $first_cycle = $read_hash_ref->{'FirstCycle'};
	my $last_cycle = $read_hash_ref->{'LastCycle'};

	if (!(defined($first_cycle))) {
	    return 0;
	}

	push(@$read_start_cycles_ref, $first_cycle);
    }

    return 1;
} 

#------------------------------------------------------------------------------
# Parse a Bustard .params file (i.e. one in a parent Firecrest folder)
# to extract an array of read lengths (e.g. two for Paired End reads)
# for the specified Bustard directory.

sub getReadLengths($$$) {
    my $xml_file_path = shift;
    my $bustard_dir = shift;
    my $read_lengths_ref = shift;

    my $rel_xpath_str = 'RunParameters/Reads';
    my @force_array_keys = ('Reads');
    my %key_attr_map = ('Reads' => 'Index');

    my $curr_hash_ref = getBustardRunInfo($xml_file_path, $bustard_dir,
					  $rel_xpath_str,
					  \@force_array_keys, \%key_attr_map);

    if (!defined($curr_hash_ref)) {
	return 0;
    }

    foreach my $read_key (keys %$curr_hash_ref) {
	my $read_hash_ref = $curr_hash_ref->{$read_key};
	my $first_cycle = $read_hash_ref->{'FirstCycle'};
	my $last_cycle = $read_hash_ref->{'LastCycle'};

	if (!(defined($first_cycle) && defined($last_cycle))) {
	    return 0;
	}

	my $read_length = ($last_cycle - $first_cycle) + 1;

	push(@$read_lengths_ref, $read_length);
    }

    return 1;
}

#------------------------------------------------------------------------------
# Translation of def getAutoFilename (makefile.py).

sub getAutoFilename($$$$)
{
    my $option_ref = shift;
    my $lane_num = shift;
    my $prefix = shift;
    my $suffix = shift;

    my $filename;

    if ($option_ref->{AutoFlag} == 1) {
	if ($option_ref->{AutoLane} == 0) {
	    $filename = sprintf("%ss_%02i%s",
			       $prefix, $option_ref->{Cycle}, $suffix);
	} else {
	    $filename = sprintf("%ss_%i_%02i%s",
			       $prefix, $option_ref->{AutoLane},
			       $option_ref->{Cycle}, $suffix);
	}
    } elsif ($option_ref->{AutoFlag} == 2) {
	$filename = sprintf("%ss_%i_%02i%s",
			   $prefix, $lane_num, $option_ref->{Cycle}, $suffix);
    } else {
	$filename = $option_ref->{Path};
    }

    return $filename;
}

#------------------------------------------------------------------------------
# Translation of def getPhasingFilename (makefile.py).

sub getPhasingFilename($$$)
{
    my $phasing_dir = shift;
    my $option_ref = shift;
    my $lane_num = shift;

    my $filename = getAutoFilename($option_ref, $lane_num,
				   "${phasing_dir}/", "_phasing.xml");

    if (($option_ref->{AutoFlag} == 0) && (!$option_ref->{Path})) {
        $filename = "params${lane_num}.xml";
    }

    return $filename;
}

#------------------------------------------------------------------------------

# *** Utilities for converting between quality values and probabilities ***

my $to10Base10=10.0/log(10.0); 

# Convert prob(base call=X), X=A,C,G,T into a Solexa quality value
# 0.01 ~= Q20, 0.001 =~ Q30 etc.
# Same formula works for all 4 base possibilities
# Moreover our QV and the Phred QV round to the same integer for
# base calls of Q10 and above
sub convertProbToQual($)
{
    my ($p)=@_;
    my $toBase10=10.0/log(10.0); # NB converts log(x) -> 10*log10(x)
    return $to10Base10*log($p/(1-$p));
} # convertProbToQual

# Convert probability that a called bases is correct into a Phred quality value
# 0.01 = Q20, 0.001 = Q30 etc.
# This formula only makes sense for a called base
sub convertProbCorrectToPhred($)
{
    my ($p)=@_;
    my $toBase10=-10.0/log(10.0); # NB converts log(x) -> 10*log10(x)
    return -$to10Base10*log(1-$p);
} # convertProbCorrectToPhred

# Convert a Solexa quality value for X into the prob that base=X
sub convertQualToProbCorrect($)
{
    my ($qv)=@_;
    return 1.0/(1+10**(-$qv/10.0));
} # convertQualToProbCorrect

# Convert a Solexa quality value for X into the prob that base != X
sub convertQualToProbWrong($)
{
    my ($qv)=@_;
    return convertQualToProbCorrect(-$qv);
} # convertQualToProbWrong

sub getQualFromChar($)
{
    my ($c)=@_;
    return ord($c)-64;
} # getQualFromChar

# sub expandSymbolicQualityString
# Expand a symbolic format quality string into a string of ASCII numbers
sub expandSymbolicQualityString($)
{
    my $qv;
    for (my $i=0;$i<length($_[0]);$i++)
    {
	$qv.=sprintf("%d ",getQualFromChar(substr($_[0],$i,1)));
    } # for
    return $qv;
} # convertSymbolicQualityString($)


# *** Functions shared between pickBestAlignment.pl and pickBestPair.pl ***

# sub getMatchDetails
# Parse match descriptor into its constituent fields, eg
# BAC_plus_vector.fa:45265R18T13C3 goes to
# BAC_plus_vector.fa, 45265, R, 18T13C3
sub getMatchDetails($$)
{
    my ($details,$chrom)=@_;
    if ($details=~s/^(.+)://)
    {
	$chrom=$1;
    } # if
    
    if ($details=~/^(-?\d+)([RF])([0123456789ACGTN]+)$/)
    {
#	$chrom1=$1; $pos1=$2; $strand1=$3; $type1=$4;
	return ($chrom, $1, $2, $3);
    } # if
    else
    {
	die "$0: Unable to parse $details\n";
    } # else
} # sub getMatchDetails

#------------------------------------------------------------------------------

sub apply_mismatch($$$$)
{
    my $new_seq = shift;
    my $num_cycles = shift;
    my $mismatch_str = shift;
    my $line_num = shift;

    my $ref_bases_str = 'ACGTN';

    my @mismatch_fields = split(/([$ref_bases_str])/, $mismatch_str);
    my $pos = 0;

    foreach my $mismatch_field (@mismatch_fields) {
	next if ($mismatch_field eq '');

	if ($pos >= $num_cycles) {
	    die("Mismatch string `$mismatch_str' specifies position beyond "
		. "end of sequence (line $line_num).\n");
	}

	if ($mismatch_field =~ /[$ref_bases_str]/) {
	    substr($new_seq, $pos++, 1) = $mismatch_field;
	} elsif ($mismatch_field =~ /\d+/) {
	    $pos += $mismatch_field;
	} else {
	    die("Unexpected field `$mismatch_field' in mismatch "
		. "string `$mismatch_str' (line $line_num)\n");
	}
    }

    return $new_seq;
}

#------------------------------------------------------------------------------

# sub getExactScoreForRead
# Get score to give to read if all bases correct - a log probability
# TC 8.8.7 - cache answers to avoid repeated computation
{
    my %cache;
    sub getExactScoreForRead
    {
	my ($qv)=@_;
	my $score=0;
	my $b;
	for (my $i=0;$i<length($qv);$i++)
	{
	    $b=substr($qv,$i,1);
	    unless (defined($cache{$b}))
	    {
		$cache{$b}=log(convertQualToProbCorrect(getQualFromChar($b)));
	    } # unless
	    $score+=$cache{$b};
	} # for	
	return $score;
    } # getExactScoreForRead
} # scope of cache
# sub adjustScoreForIncorrectBase
# Get adjustment to apply to exact match score if a base of this quality 
# matches incorrectly TBD hash this to avoid repeated computation
# TC 8.8.7 - cache answers to avoid repeated computation
{
    my %cache;
    sub adjustScoreForIncorrectBase
    {
	my ($c)=@_;
	unless (defined($cache{$c}))
	{
	    my $p=convertQualToProbCorrect(getQualFromChar($c));
	    $cache{$c}=log((1-$p)/3)-log($p);
	} # unless
	return $cache{$c};
    } # adjustScoreForIncorrectBase
}


# sub getWorstBases
# Get the score adjustments for the three lowest quality bases that were 
# aligned
sub getWorstBases($$)
{
    my ($qv,$readLengthELAND)=@_;
    # print "$qv $readLengthELAND\n";
    $qv=substr($qv,0,$readLengthELAND);
    my @qvals=sort(split(//,$qv));
    # print "@qvals\n";
    for (my $i=0; $i<3; $i++)
    {
	$qvals[$i]=adjustScoreForIncorrectBase($qvals[$i]);
    }
    return ($qvals[0],$qvals[1],$qvals[2]);
} # sub getWorstBases

# sub getMeanBaseQuality
# Get the mean base quality for a read - uses output of getExactScoreForRead
sub getMeanBaseQuality($$)
{
    my ($exactScore, $qv)=@_;
    return convertProbCorrectToPhred(exp($exactScore/length($qv)));
} # sub getMeanBaseQuality




# sub getAlignmentScore
# Work out alignment score (=log(probability)) of alignment given qvals
# Inputs:
# qv=quality value ASCII string eg ^^^^^^^^^^^^^^^^^ZW^V
# type=alignment descriptor, eg 20A4c1A
# thisScore=log-probability score for exact match
# Output:
# Adjusted log-probability score for inexact match
sub getAlignmentScore($$$)
{ 
#    die unless (@_==3);
    my ($qv,$type,$thisScore) =@_;
    my $posInRead=0;
    while ($type=~/(\d*)[ACGT]/g)
    {
	$posInRead++;
	$posInRead+=$1 unless ($1 eq "");
	$thisScore+=
	    adjustScoreForIncorrectBase(substr($qv,$posInRead-1,1));
    } # while
    return $thisScore;
} # sub getAlignmentScore

# sub scoreAlignFromNhood
# Estimate confidence for an alignment based on quality values and
# neighbourhood information
# Need to adjust the neighbourhood to account for the fact we already
# have an alignment for the best match and possibly some others too
# Inputs:
# qv=quality value ASCII string eg ^^^^^^^^^^^^^^^^^ZW^V
# nhood=neighbourhood string eg 1:0:0
# bestScore=highest probability out of reads found so far
# totalScoreSoFar=summed probabilities for reads considered so far
# readsUsedSoFar=number of reads used so far
# readLengthELAND=read length used in originating ELAND alignment (usually 32)
# Output:
# Estimated prob of alignment being wrong in 'Phred' format, ie Q20=0.99%
sub scoreAlignFromNhood($$$$$$$)
{
#    die unless (@_==7);
    my ($qv, $nhood, $bestScore, 
	$totalScoreSoFar, 
	$readsUsedSoFar,
	$readLengthELAND,
	$thisScore)=@_;
    my $tsf=$totalScoreSoFar; # TBD delete
#    my $thisScore=getExactScoreForRead($qv);
    my @nbors=split(':',$nhood);
    my @worstBases=getWorstBases($qv,$readLengthELAND);

    for my $nbor (@nbors)
    {
	while (($readsUsedSoFar>0)&&($nbor>0))
	{
	    $readsUsedSoFar--;
	    $nbor--;
	} # while
    } # for @nbors
    # print "Adjusted nhood $nbors[0]:$nbors[1]:$nbors[2]\n";

    # add in some 3-neighbours to get approx score in (1,0,0),(0,1,0) and 
    # (0,0,1) cases - arbitrary value for time being
    push @nbors,10; 
    unshift (@worstBases,0);

    for (my $i=0; $i<4; $i++)
    {
	$thisScore+=$worstBases[$i];
	$totalScoreSoFar+=$nbors[$i]*exp($thisScore);
    } # for

    # print "SCORE: $bestScore $totalScoreSoFar ";
    if ($bestScore==$totalScoreSoFar)
    {
	die; # should never be here
#	$bestScore=$bestPossibleScore;
    } # if
    else
    {
	$bestScore
	    =floor(convertProbCorrectToPhred($bestScore/$totalScoreSoFar));
    } # else

    return $bestScore;

} # scoreAlignFromNhood

# sub bestAlignFromList
# Pick best alignment from a list of possibilities and work out its 
# relative probability with respect to the rest
# Inputs:
# qv=quality value ASCII string eg ^^^^^^^^^^^^^^^^^ZW^V
# nhood=neighbourhood string eg 1:0:0
# details=details of aligning reads
# readLengthELAND=read length used in originating ELAND alignment (usually 32)
# Output:
# Estimated prob of alignment being wrong in 'Phred' format, ie Q20=0.99%
sub bestAlignFromList($$$$$)
{
#    die unless (@_==5);
    my($qv, $nhood, $details, $readLengthELAND, $exactScore )=@_;
    my $totalScore=0;
    my $bestScore=undef;
#    my $exactScore=getExactScoreForRead($qv);
    my $thisScore;
    my ($chrom, $pos, $strand, $type);
    my ($bestChrom, $bestPos, $bestStrand, $bestType);
    

    # print "bestAlignFromList $qv $nhood $details $exactScore ", exp($exactScore), "\n";

    my @s=split(",",$details);

    for (my $i=0;$i<@s;$i++)
    {
	($chrom, $pos, $strand, $type)=getMatchDetails($s[$i], $chrom);
	
#	$thisScore=$exactScore;

	$thisScore=getAlignmentScore($qv,$type,$exactScore);
	 # print "$s[$i] $chrom $pos $strand $type $thisScore ";
	$thisScore=exp($thisScore);
	 # print "$thisScore\n";

	$totalScore+=$thisScore;
	
	$bestScore=$thisScore-1 unless (defined($bestScore));
	if ($thisScore>$bestScore)
	{
	    $bestScore=$thisScore;
	    ($bestChrom, $bestPos, $bestStrand, $bestType)
		=($chrom, $pos, $strand, $type);
	} # if
    } # for

    # print "SCORE: $bestScore $totalScore\n";

    $bestScore=scoreAlignFromNhood
	    ($qv, $nhood, $bestScore, 
	     $totalScore, scalar(@s), $readLengthELAND, $exactScore);

    # print "SCORE NHOOD-ADJUSTED: $bestScore\n";
    
    return ($bestChrom, $bestPos, $bestStrand, $bestType, $bestScore); 
} # bestAlignFromList

# sub pickBestSingleReadAlignment
# Take in read quality value string and a set of alignment details
# and output the details of the best alignment
# Inputs: 
# qv=quality value string
# nhood=neighbourhood string eg 1:0:0
# details=match details string
# readLengthELAND=read length used in originating ELAND alignment (usually 32)
# Outputs:
# chrom=chromosome
# pos=position in chromosome (must be >0, else no match)
# strand=F or R
# type=match description
# score=estimated probability alignment is correct (Phred format, ie Q20=99%)
# TC 19.10.7 - pos==undef is the signal that no match is found
# pos<=0 can occasionally occur for wraparounds
sub pickBestSingleReadAlignment($$$$$)
{
#    die unless (@_==5);
    my ($qv, $nhood, $details, $readLengthELAND, $exactScore)=@_;
    my ($chrom, $pos, $strand, $type, $score);

    # score for exact match
#    $exactScore=getExactScoreForRead($qv); 


    if ($details eq '-')
    { # no match found for read
#	print "$readID aaa$nhood 0 F - 0\n";
#	next;
	$pos=undef;
	return ($chrom, $pos, $strand, $type, $score);
    } # if
    elsif ($details!~/,/)
    { # just one match found for read, score it
	($chrom, $pos, $strand, $type)=getMatchDetails($details, $chrom);
	
	# score for exact match
#	$score=getExactScoreForRead($qv); 
	$score=$exactScore; 
	# adjust to score for actual match
	$score=getAlignmentScore($qv,$type,$score); 
	# convert from score to prob
	$score=exp($score);
	# convert prob to score relative to neighbours
	$score=scoreAlignFromNhood
	    ($qv, $nhood, $score, $score, 1, $readLengthELAND,
	     $exactScore );

#	print "$readID $chrom $pos $strand $type $score\n";
    } # if
    else
    { # multiple matches found for read, pick best and score it
	($chrom, $pos, $strand, $type, $score)
	    =bestAlignFromList( $qv, $nhood, $details, $readLengthELAND,
				$exactScore );
#	print "$readID $chrom $pos $strand $type $score\n";
    } # else
    return ($chrom, $pos, $strand, $type, $score);
} # sub pickBestSingleReadAlignment



# Uncomment this and run to demo quality value conversion
#for (my $qv=-5;$qv<=60;$qv++)
#{
#    my $p1=convertQualToProbCorrect($qv);
#    my $p2=convertQualToProbWrong($qv);
#    my $qw=convertProbToQual($p2/3); # assume 3 wrong bases equiprobable
#    printf ("%d %10.7f %10.7f %10.7f %10.7f %d %d %10.7f\n", 
#            $qv, $p1, $p2, convertProbToQual($p2), $qw, floor($qw), 
#            floor($qw)-$qv, convertProbCorrectToPhred($p1) );
#} # for

#------------------------------------------------------------------------------
# A set of functions to do SIGPIPE debugging.
#------------------------------------------------------------------------------

my $sigpipe_count = 0;

sub sigpipe_handler($)
{
    ++$sigpipe_count;
}

#------------------------------------------------------------------------------

sub conditionally_init_sigpipe_handler()
{
    my $catch_pipe_sigs = (defined($ENV{'ILMN_CATCH_PIPE_SIGS'})
			   ? $ENV{'ILMN_CATCH_PIPE_SIGS'}
			   : 0);

    if ($catch_pipe_sigs != 0) {
	$SIG{PIPE} = \&sigpipe_handler;
    }
}

#------------------------------------------------------------------------------

sub conditionally_report_sigpipes()
{
    if ($sigpipe_count != 0) {
	print STDERR "Warning: SIGPIPE count : $sigpipe_count";
    }
}

#------------------------------------------------------------------------------

sub trim_external_Ns($$$)
{
    my $read_seq = shift;
    my $trim_start_pos_ref = shift;
    my $trim_len_ref = shift;

    # To persuade a simple split to do what is wanted, we append an 'N'
    # to the string, do the split and then subtract 1 one from the trailing
    # N count.
    my @chunks = split(/[^N]/, ($read_seq . 'N'));
    $$trim_start_pos_ref = length($chunks[0]);
    my $trail_n_count = length($chunks[$#chunks]) - 1;

    if (scalar(@chunks) == 1) {
	# This should not happen as an all-Ns read should have failed quality
	# control and not reached here at all but handle the case anyway.
	$$trim_start_pos_ref = 0;
	$$trim_len_ref = 0;
	return '';
    }

    $$trim_len_ref = length($read_seq) - $$trim_start_pos_ref - $trail_n_count;

    return substr($read_seq, $$trim_start_pos_ref, $$trim_len_ref);
}

#------------------------------------------------------------------------------

sub lane_prefix($)
{
    my $lane_num = shift;
    return 's_' . $lane_num;
}

#------------------------------------------------------------------------------

sub lane_num($)
{
    my $lane_prefix = shift;

    if ($lane_prefix =~ /^s_(\d+)$/) {
	return $1;
    }

    return 0;
}

#------------------------------------------------------------------------------

sub get_tiles_by_lane($$$)
{
    my ($tile_file_path, $lane_prefix, $tile_list_ref) = @_;

    open(TILE_FILE, $tile_file_path) or return 0;

    while (my $line = <TILE_FILE>) {
	next if ($line =~ m/^\#/);

	my @curr_tiles = split(" ", $line);

	foreach my $tile (@curr_tiles) {
	    if ($tile =~ /^${lane_prefix}/) {
		push @$tile_list_ref, $tile;
	    }
	}
    }

    close(TILE_FILE);

    return 1;
}

#------------------------------------------------------------------------------

sub get_expected_tile_nums($$$)
{
    my $tile_file_path = shift;
    my $lane_num = shift;
    my $expected_tile_nums_ref = shift;

    my @tile_list;

    if (!get_tiles_by_lane($tile_file_path, lane_prefix($lane_num),
			   \@tile_list)) {
	return 0;
    }

    @$expected_tile_nums_ref
	= map { my @arr = split('_', $_); $arr[-1] } @tile_list;

    return 1;
}

#------------------------------------------------------------------------------

sub determine_missing_tile_nums($$$$)
{
    my $tile_file_path = shift;
    my $tile_freqs_ref = shift;
    my $lane_num = shift;
    my $missing_tile_nums_ref = shift;

    my @expected_tile_nums = ();

    if (!get_expected_tile_nums($tile_file_path, $lane_num,
                                \@expected_tile_nums)) {
        die("Failed to get expected tile nums from file `$tile_file_path'.");
    }

    my @missing_tile_nums = ();

    foreach my $expected_tile_num_str (@expected_tile_nums) {
        # Strip any leading zeros to match keys in %tile_freqs.
        my $expected_tile_num = int($expected_tile_num_str);

        if (!defined($tile_freqs_ref->{$expected_tile_num})) {
	    push(@missing_tile_nums, $expected_tile_num);
	}
    }

    @$missing_tile_nums_ref = @missing_tile_nums;
    return 1;
}

#------------------------------------------------------------------------------

1;
