#!/usr/bin/env perl

# Copyright (c) 2006 Solexa
# Authors: 
# M. Parkinson, L. J. Davies
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

use warnings;
use strict;
use Cwd;

my $version;
my $id;
{
    my $cvs_id='@(#) $Id: Signal_Means_Tile.pl,v 1.1 2008/07/04 10:22:23 craczy Exp $';
    my $cvs_version='$Revision: 1.1 $';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope


# Plot average intensity versus sequencing cycle

my $usage=<<"END";
Usage: $0 qualityFilterPath qualityCriterion sigFile seqFile qhgFile
END

die $usage unless (@ARGV == 5);

my $numChannels=4;
my $qualityFilter=shift(@ARGV);
my $qcCrit=shift(@ARGV);
my $sigFile = shift(@ARGV);
my $seqFile = shift(@ARGV);
my $qhgFile = shift(@ARGV);

my ($laneNum,$sigPipe,$seqPipe,$cycleNum,$sigLine,$seqLine);
my ($thisCall,$thisCallKey,$baseCountKey,$numBase,$numCalled);
my ($numClusters,$numFilteredClusters);
my (@array,@sig,@seq);
my (@all,@bcall,@call);
my @bases = qw(A C G T);

my %thisCallArray =
    ( 'A'=>0,
      'C'=>1,
      'G'=>2,
      'T'=>3 );

my $num_bases = @bases;
my $num_header_cols = 4;

#print STDERR "Filtering $sigFile and $seqFile with $qhgFile\n\n";

$sigPipe="cat $sigFile | perl $qualityFilter -stats \'$qcCrit\' $qhgFile |";
$seqPipe="cat $seqFile | perl $qualityFilter -stats \'$qcCrit\' $qhgFile |";


#print  STDERR "$sigPipe\n";
#print  STDERR "$seqPipe\n";

open (SIG_PIPE, $sigPipe) 
    || die "Error: failed to open pipe $sigPipe: $!";
open (SEQ_PIPE, $seqPipe) 
    || die "Error: failed to open pipe $seqPipe: $!";


my $num_cycles = 0;

while (defined($sigLine=<SIG_PIPE>)) {
    $seqLine = <SEQ_PIPE>;

    if ($sigLine =~ m/^\#/) {
	if ($sigLine =~ m/^\# PASSED ([0-9]+) OF ([0-9]+)/) {
	    $numFilteredClusters = $1;
	    $numClusters = $2;
	}

	next;
    }

    @sig = split('\t',$sigLine);
    @seq = split('\t',$seqLine);

    $laneNum = $sig[0];
#	print STDERR $sigLine;

    # Check signal and sequence headers match.

    if ($laneNum ne $seq[0]){
	die "The sig and seq lines are not from same lane $laneNum $seq[0]\n";
    }

    if($sig[1] ne $seq[1]){
	die "The sig and seq lines are not from same tile $sig[1] $seq[1]\n";
    }

    if($sig[2] ne $seq[2]){
	die "The sig and seq lines don't have same x-coord $sig[2] $seq[2]\n";
    }

    if($sig[3] ne $seq[3]){
	die "The sig and seq lines don't have same y-coord $sig[3] $seq[3]\n";
    }

    $num_cycles = ((@sig) - $num_header_cols);


    # Produce statistics per cycle (for the tile).

    for($cycleNum = 1; $cycleNum <= $num_cycles; $cycleNum++)
    {
	### Cycle information
	@array = split(' ', $sig[$num_header_cols + $cycleNum - 1]);
	    
	if (@array==0)
	{
	    warn "Failed to find intensity values for cycle $cycleNum\n";
	    last;
	}

	## intensities for all bases
	$all[$cycleNum]{A} += $array[0];
	$all[$cycleNum]{C} += $array[1];
	$all[$cycleNum]{G} += $array[2];
	$all[$cycleNum]{T} += $array[3];
	$all[$cycleNum]{Count}++;

	## called bases
	$thisCall=substr($seq[4], ($cycleNum - 1), 1);

	$bcall[$cycleNum]{$thisCall}++;

	## intensities for (valid) called bases
	if($thisCall =~ m/[ACGT]/)
	{
	    $call[$cycleNum]{$thisCall}
	    += $array[$thisCallArray{$thisCall}]; # sum
	    $thisCallKey='Count'.$thisCall;
	    $call[$cycleNum]{$thisCallKey}++; # count
	} # if
    } # for
} # while

close (SIG_PIPE)
    || die "Error: failed to close signal pipe $sigPipe: $!";
close (SEQ_PIPE)
    || die "Error: failed to close sequence pipe $seqPipe: $!";

unless (defined($laneNum))
{
    # No intensity info was found: output a set of zero intensities 
    warn "Warning: no intensity information obtained from pipe '$sigPipe'";
    # Set laneNum to -1 rather than guess
    $laneNum=-1;
} # unless

### Output statistics.
##  Print header lines.
print "# Clusters: Filtered $numFilteredClusters Original $numClusters\n";
print "#\n";
print "#Lane\tCycle\tAll A\tAll C\tAll G\tAll T\tCall A\tCall C\tCall G\tCall T\tNum A\tNum C\tNum G\tNum T\tNum X\n";

for($cycleNum = 1; $cycleNum <= $num_cycles; $cycleNum++){
    
    print "$laneNum\t$cycleNum\t";

    ## Intensities averaged over all (called/not called) bases

    $all[$cycleNum]{Count} = 0 unless defined ($all[$cycleNum]{Count});

    foreach (@bases) {
	$all[$cycleNum]{$_} = 0 unless defined ($all[$cycleNum]{$_});

	if ($all[$cycleNum]{Count} != 0){
	    $all[$cycleNum]{$_} /= $all[$cycleNum]{Count};
	}
	printf ("%0.1f\t", $all[$cycleNum]{$_});
    }


    ## Intensities averaged over called bases

    foreach (@bases) {
	$baseCountKey='Count'.$_;
	$call[$cycleNum]{$baseCountKey} = 0 
	    unless defined ($call[$cycleNum]{$baseCountKey});

	$call[$cycleNum]{$_} = 0 unless defined ($call[$cycleNum]{$_});

	if ($call[$cycleNum]{$baseCountKey} != 0){
	    $call[$cycleNum]{$_} /= $call[$cycleNum]{$baseCountKey};
	}
	printf ("%0.1f\t", $call[$cycleNum]{$_});
    }


    ## Numbers of bases called by bases and number not called.

    $numCalled = 0;

    foreach (@bases) {
	$bcall[$cycleNum]{$_} = 0 unless defined ($bcall[$cycleNum]{$_});
	$numBase = $bcall[$cycleNum]{$_};

	printf("%d\t", $numBase);
	$numCalled += $numBase;
    }

    printf("%d", $all[$cycleNum]{Count} - $numCalled);

    print "\n";
}
