#!/usr/bin/env perl


# Jerboa produces an experimental summary sheet in html format
# The html prints some summary statistics
# The html provides links to other information
#
# Copyright (c) 2006 Solexa; 2007, 2008 Illumina
# Author L. J. Davies
# Extended by R. J. Shaw.
#
# This source file is covered by the "Illumina Genome Analyzer
# Software License Agreement" and bound by the terms therein.

# The program is called using directory name

use warnings; 
use strict;
use Cwd;
use XML::Simple;
use File::Basename;
use File::Spec;
use Data::Dumper;

use FindBin;
use lib "$FindBin::Bin/../lib/perl";
use Gerald::Common;

#------------------------------------------------------------------------------

my $version;
my $id;
{
    my $cvs_id='@(#) $Id: jerboa.pl,v 1.10 2008/07/23 15:18:30 mzerara Exp $';
    my $cvs_version='$Revision: 1.10 $';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope

#------------------------------------------------------------------------------

sub info
{
    my $message = shift @_;
    print STDERR "Info: ", basename($0), $message, "\n";
}

#------------------------------------------------------------------------------

sub getBustardPath($$);

#------------------------------------------------------------------------------

my $maxNumLanes=8;
my $numChannels=4;
my $ref_read_num = 1; # Use this to index values that are read-independent.

# Next def makes sure all unknown fields are denoted the same way
my $unknown='unknown';

my $seqFileRegex='s_[1-8]_[0-9][0-9][0-9][0-9]';
my $seqSuffix='_seq.txt';
my $alignSuffix='_align.txt';
my $scoreSuffix='_score.txt';
my $realignSuffix='_realign.txt';
my $rescoreSuffix='_rescore.txt';
my $allSuffix='_all.txt';

# Intensities less than minIntensity get ignored in dropoff calculations
my $minIntensity=5; 

# Read bustard path from xml file rather than from argument list
my $configDir = File::Spec->rel2abs($ARGV[0]);
my $configXMLFile = File::Spec->catdir($configDir, "config.xml");

my $bustardPath;
getBustardPath(\$bustardPath, $configXMLFile);
my $firecrestPath="$bustardPath/..";
my $runFolderPath="$bustardPath/../../..";
my $signal_means_path = "${bustardPath}/SignalMeans";


my $xml = new XML::Simple();

my $not_avail_str = 'N/A';

my $num_paired_reads = 2;
my $all_reads_key = 'AllReads';
my $pair_key_separ_str = ':';

my $Orientation_str = 'Orientation';
my @orient_display_map = (['Fm', 'F-: &gt; R2 R1 &gt;'],
                          ['Fp', 'F+: &gt; R1 R2 &gt;'],
                          ['Rm', 'R-: &lt; R2 R1 &gt;'],
                          ['Rp', 'R+: &gt; R1 R2 &lt;']);
my %orient_short_display_map = ('Fm' => 'F-',
                                'Fp' => 'F+',
                                'Rm' => 'R-',
                                'Rp' => 'R+');
my $nominal_orient_key = 'Nominal';
my $pc_nominal_orient_key = 'NominalOrientationPercent';

my $Insert_Size_str = 'InsertSize';

my $num_orient_ok_small_insert_key = 'NominalOrientationButSmallInsert';
my $num_orient_ok_large_insert_key = 'NominalOrientationButLargeInsert';
my $num_pairs_ok_key = 'PairsOK';

my $pc_suffix = 'Percent';
my $pc_orient_ok_small_insert_key
    = $num_orient_ok_small_insert_key . $pc_suffix;
my $pc_orient_ok_large_insert_key
    = $num_orient_ok_large_insert_key . $pc_suffix;
my $pc_pairs_ok_key = $num_pairs_ok_key . $pc_suffix;

my %insert_num_pc_key_map = ($num_orient_ok_small_insert_key
			     => $pc_orient_ok_small_insert_key,
			     $num_orient_ok_large_insert_key
			     => $pc_orient_ok_large_insert_key,
			     $num_pairs_ok_key
			     => $pc_pairs_ok_key);
my $insert_pc_format = "%0.1f";

my @insert_stat_display_map = (['Median', 'Median'],
                               ['LowSD', 'Below-median SD'],
                               ['HighSD', 'Above-median SD'],
                               ['Min', 'Low thresh.'],
                               ['Max', 'High thresh.']);

my @inserts_display_map = ([$num_orient_ok_small_insert_key, 'Too small'],
			   [$num_orient_ok_large_insert_key, 'Too large'],
			   [$num_pairs_ok_key, 'Orientation and size OK']);

my $Unique_key = 'SingleAlignmentFound';
my $Unique_str = 'Unique';
my $Rescuable_key = 'ManyAlignmentsFound';
my $Rescuable_str = 'Rescuable';
my $Repeat_key = 'Repeat';
my $Repeat_str = $Repeat_key;
my $Repeat_Masked_key = 'RM';
my $Repeat_Masked_str = 'Repeat Masked';
my $Not_Matched_key = 'NM';
my $Not_Matched_str = 'Not Matched';
my $Low_Quality_key = 'QC';
my $Low_Quality_str = 'Low Quality';
my @align_outcome_display_map = ([$Unique_key, $Unique_str,],
				 [$Rescuable_key, $Rescuable_str],
                                 [$Repeat_key, $Repeat_str],
				 [$Repeat_Masked_key, $Repeat_Masked_str],
                                 [$Not_Matched_key, $Not_Matched_str],
                                 [$Low_Quality_key, $Low_Quality_str]);
my $Read_str = 'Read';
my $All_Reads_key = 'AllReads';
my $Total_str = 'Total';
my $pc_prefix = 'pc';

my $paired_not_done_key = 'NoPairedAlignmentDone';
my $paired_not_done_str = 'Paired alignment not performed';

my $unique_pair_align_key = 'UniquePairedAlignment';
my $multi_pair_align_key = 'ManyPairedAlignments';

my $Lost_str = 'Lost';
my $Rescued_Unique_key = 'RescuedUnique';
# my $Position_Change_key = 'PositionChange';

my @paired_align_display_map = ([$Unique_key, $Unique_str],
				[$Rescuable_key, $Rescuable_str]);

my @paired_reads_display_map = (["${Read_str}1", 'Read 1'],
				["${Read_str}2", 'Read 2'],
				[$All_Reads_key, 'All Reads']);

my @paired_effects_display_map = ([$Rescued_Unique_key,
				   'Uniqueness Rescue (% of rescuable)']
#				  [$Position_Change_key, 'Position Change']);
				  );

#------------------------------------------------------------------------------

sub param_file_is_legacy_style($)
{
    my $bustard_dir = shift;

    my %dir_info;

    if (!parseAnalysisDirName($bustard_dir, \%dir_info)) {
	warn "Failed to parse bustard directory name : $bustard_dir\n";
	exit(1);
    }

    if (!(defined($dir_info{major_ver}) && defined($dir_info{minor_ver})
	  && defined($dir_info{release}))) {
	warn "Failed to extract bustard version from : $bustard_dir\n";
	exit(1);
    }			       

    my @bustard_ver = ($dir_info{major_ver}, $dir_info{minor_ver},
		       $dir_info{release});
    my @new_param_bustard_ver = (1, 9, 0);

    my $style_is_legacy 
	= (compareVersions(\@new_param_bustard_ver, \@bustard_ver) < 0);

    return $style_is_legacy;
}

#------------------------------------------------------------------------------
# Read XML file into a hash. Return 1 if successfully completed
# Check for zero padding at end of file
sub readFileXML
{
    my ($fileName)=@_;
    return undef unless (-f $fileName);
    unless(open (FILE, $fileName)) 
    { 
	warn "Failed to open $fileName: $!"; return undef; 
    } 
    my $text="";
    while (<FILE>)
    {
	$text.=$_;
    } # while
    close (FILE);

    my $l=length($text);

    while (($l!=0)&&(substr($text,$l-1,1) eq "\0"))
    {
	$l--;
    } # while

    if ($l==0)
    {
	warn "file $fileName seems to be completely blank";
	return undef;
    } # if   

    $text=substr($text,0,$l) unless ($l==length($text));

    my $data = $xml->XMLin($text);

    return $data;
} # sub readFileXML

#------------------------------------------------------------------------------
# Take a series of n points (xi, yi)
# Convert yi to log(yi)
# Fit a best fit line to obtain average slope in x linear - y log space
# Convert this to a best fit percent loss per cycle

# Formula used for best fit is y ~ ax+b
# a=(sigma(xi yi) - n * mean x * mean y)/(sigma(xi xi) - n * mean x * mean x)  
# b = mean y - a * mean x (b not worked out)

# This is formula 18.34 From Jeffrey 2nd edition ISBN 0-442-30728-4 
# (& 000's of other maths books)

# Version in code uses fact that mean (xi) = sigma (xi) / n


sub logLinearFit
{
    my ($x, $y)=@_;
    die unless (scalar(@{$x})==scalar(@{$y}));
    my ($x_sum, $y_sum, $xx_sum, $xy_sum, $i, $ylog, $a, $b)
    =(0,0,0,0,0,0,0,0);
    my $numDataPoints=0;

    for ($i=0; $i<@{$x}; $i++)
    {
#        print "$i ".$x->[$i]." ".$y->[$i]."\n";
	next unless ($y->[$i]>=$minIntensity);
	++$numDataPoints;

        $ylog=log($y->[$i]);
        $x_sum+=$x->[$i];
        $y_sum+=$ylog;
        $xx_sum+=($x->[$i]*$x->[$i]);
        $xy_sum+=($x->[$i]*$ylog);
    } # for
#    print "$i\n";

    unless ($numDataPoints>=2)
    {
	warn "Not enough valid data points to do fit";
	return 100.0;
    } # unless
    
    # Work out gradient of best fit line
    $a=($xy_sum-($x_sum*$y_sum/$i))/($xx_sum-($x_sum*$x_sum/$i));

    # Convert to a percent loss per cycle
    return (100*(1-exp($a)));
} # sub logLinearFit

#------------------------------------------------------------------------------

sub key_matched($$)
{
    my $match_key_fields_ref = shift;
    my $key_fields_ref = shift;

    my @match_key_fields = @{$match_key_fields_ref};
    my $num_match_key_fields = scalar(@match_key_fields);

    my @key_fields = @{$key_fields_ref};

    if (scalar(@key_fields) != $num_match_key_fields) {
	return 0;
    }

    my $matched = 1;

    for (my $field_ind = 0; $field_ind < $num_match_key_fields;
	 ++$field_ind) {
	if (!($key_fields[$field_ind]
	      =~ /$match_key_fields[$field_ind]/)) {
	    $matched = 0;
	    last;
	}
    }

    return $matched;
}

#------------------------------------------------------------------------------

sub calc_total_by_field($$$$)
{
    my $hash_ref = shift;
    my $match_key_fields_ref = shift;
    my $wild_field_ind = shift;
    my $separ_str = shift;

    foreach my $key_str (keys %{$hash_ref}) {
	my @key_fields = split(/$separ_str/, $key_str);
	next if (!key_matched($match_key_fields_ref, \@key_fields));

	my @total_key_fields = @key_fields;
	splice(@total_key_fields, $wild_field_ind, 1, $Total_str);
	my $total_key = join($separ_str, @total_key_fields);

	if (!defined($hash_ref->{$total_key})) {
	    $hash_ref->{$total_key} = 0;
	}

	my $val = $hash_ref->{$key_str};

	if (!defined($val) || ($val eq $not_avail_str)) {
	    $val = 0;
	}

	$hash_ref->{$total_key} += $val;
    }
}

#------------------------------------------------------------------------------

sub calc_percent_by_field($$$$)
{
    my $hash_ref = shift;
    my $match_key_fields_ref = shift;
    my $total_key = shift;
    my $separ_str = shift;

    my $divisor = $hash_ref->{$total_key};

    if (!defined($divisor)) {
	warn("Divisor ($total_key) not found when calculating percentage!");
        $divisor = 0;
    }

    $divisor /= 100.0;

    foreach my $key_str (keys %{$hash_ref}) {
	my @key_fields = split(/$separ_str/, $key_str);
	next if (!key_matched($match_key_fields_ref, \@key_fields));

	my $pc_key_str = join($separ_str, ($pc_prefix, $key_str));
        $hash_ref->{$pc_key_str} = (($divisor != 0)
                                    ? ($hash_ref->{$key_str} / $divisor)
                                    : 0);
    }
}

#------------------------------------------------------------------------------

sub getBustardPath($$) {
   my $bustardPathRef = shift;
   my $configFile = shift;
   my $xml = XML::Simple->new();
   my $xmlData = $xml->XMLin($configFile);

   if (!defined($xmlData)) {
     warn "Failed to parse $configFile as XML\n";
     exit(1);
   }
     $$bustardPathRef = $xmlData->{'ChipWideRunParameters'}{"EXPT_DIR"} or die "Can't find EXPT_DIR in xml file!";
}

sub extract_run_info($$$$$)
{
    my $appName = shift;
    my $bustardPath = shift;
    my $runFolderRef = shift;
    my $machineNameRef = shift;
    my $bustardDirRef = shift;

    my $runFolder = undef;
    my $machineName = undef;
    
    my @array = split('/', $bustardPath);
    $$bustardDirRef = $array[-1];

#   Check for valid as-agreed-in-spec-months ago run-folder name
#   or eight digit run-folder name (grrr....)

    for (my $i = 0; $i < @array; $i++) {
	next if (($array[$i] !~ m/0000\d{4}/) 
		 && ($array[$i] !~ m/^\d{6}_(.*)_\d{4}$/)) ;
	$runFolder = $array[$i];
	$machineName = $1 if defined ($1);
	last;
    }

#   Last ditch attempt at finding a run folder name

    $runFolder = $array[-4]
	if ((defined($array[-4])) && (!defined($runFolder)));

# Give up...
    $runFolder = $unknown unless (defined($runFolder));

    info(": deduced run folder name $runFolder");

    unless (defined($machineName)) {
	my $paramsFile = "${runFolderPath}/${runFolder}.params";

	print STDERR $appName, 
	": looking in $paramsFile for machine name\n";
	my $xmlData;

	if (defined($xmlData = readFileXML($paramsFile))) {
#	    print Dumper($xmlData);

	    if (defined($xmlData->{run}{instrument})) {
		$machineName = $xmlData->{run}{instrument};	
		print STDERR $appName, 
		": found machine name $machineName\n";
	    } else {
		warn "Failed to find instrument name in $paramsFile\n";
	    }
	}
    }

    $machineName = $unknown unless (defined($machineName));

    $$runFolderRef = $runFolder;
    $$machineNameRef = $machineName;
}

#------------------------------------------------------------------------------

sub init_lane_results($$$$$)
{
    my $lane_results_ref = shift;
    my $lane_num_reads_ref = shift;
    my $lane_num_tiles_ref = shift;
    my $maxNumLanes = shift;
    my $num_available_reads = shift;


    my %undef_hash;

    for (my $lane_num = 1; $lane_num <= $maxNumLanes; ++$lane_num) {
        # The number of reads for a lane is initialised to the number
        # available but may be reduced later according to analysis type.
        $lane_num_reads_ref->[$lane_num] = $num_available_reads;
        $lane_num_tiles_ref->[$lane_num] = 0;

        for (my $read_num = 1; $read_num <= $num_available_reads;
             ++$read_num) {
	    # Force $lane_results_ref->[$read_num][$lane] to exist before use.
	    %{$lane_results_ref->[$read_num][$lane_num]} = %undef_hash;
	    my $lane_read_ref = $lane_results_ref->[$read_num][$lane_num];

	    $lane_read_ref->{phasingapplied} = undef;
	    $lane_read_ref->{prephasingapplied} = undef;
	    $lane_read_ref->{phasingmeasured} = undef;
	    $lane_read_ref->{prephasingmeasured} = undef;
	    $lane_read_ref->{tileCountRaw} = 0;
	    $lane_read_ref->{tileCountPF} = 0;

	    # set to 1 if any _(re)score.txt successfully opened for 
	    # tile in that lane.
	    $lane_read_ref->{aligned} = 0;

	    foreach my $thisStat 
                (qw/template type length lengths_list purity eland/) {
		$lane_read_ref->{$thisStat} = $unknown;
	    }
	}
    }
}

#------------------------------------------------------------------------------

sub store_measured_phasings($$$$$$)
{
    my $lane_results_ref = shift;
    my $bustard_path = shift;
    my $num_lanes = shift;
    my $num_reads = shift;
    my $phasing_options_ref = shift;
    my $param_is_legacy = shift;

    my $phasing_path = "$bustard_path/Phasing";
    my $xml_data;

    my $measured_phasing_file;

    for (my $lane_num = 1; $lane_num <= $num_lanes; ++$lane_num) {
	for (my $read_num = 1; $read_num <= $num_reads; ++$read_num) {
	    my $lane_read_ref = $lane_results_ref->[$read_num][$lane_num];

	    if (!$param_is_legacy) {
		my $curr_options_ref = $phasing_options_ref->{$read_num};

		# FIXME : error?
		next if (!defined($curr_options_ref));

		# Take a copy so AutoFlag can be overridden.
		my %curr_options = %{$curr_options_ref};
		$curr_options{'AutoFlag'} = 2; # `use measured per lane'

		$measured_phasing_file = getPhasingFilename($phasing_path,
							    \%curr_options,
							    $lane_num);

	    } else {
		# Look for lane-specific measured phasing values 
		# (pipeline ver >= 0.1.7).
		$measured_phasing_file = sprintf("%s/s_%d_phasing.xml",
						 $phasing_path, $lane_num);
	    }

	    if (defined($xml_data = readFileXML($measured_phasing_file))) {
		#print STDERR Dumper($xml_data), "\n";;
		if (defined($xml_data->{Phasing})) {
		    $lane_read_ref->{phasingmeasured} = $xml_data->{Phasing};
		}

		if (defined($xml_data->{Prephasing})) {
		    $lane_read_ref->{prephasingmeasured}
		    = $xml_data->{Prephasing};
		}
	    }
	}
    }
}

#------------------------------------------------------------------------------

sub legacy_read_applied_phasing_file($$$$$)
{
    my $app_name = shift;
    my $phasings_applied_file = shift;
    my $phasing_ref = shift;
    my $prephasing_ref = shift;
    my $bustard_dir = shift;

    info(": looking in $phasings_applied_file for phasing information");

    my $xmlData;

    if (defined($xmlData = readFileXML($phasings_applied_file))) {
	if (defined($xmlData->{Run})) {
	    # FIXME : Use forcearray and keyattr in readFileXML to avoid 
	    #         the ARRAY / HASH division. 

	    my $run_ref = undef;

	    if (ref($xmlData->{Run}) eq 'ARRAY') {
		for (my $i = 0; $i < scalar(@{$xmlData->{Run}}); $i++) {
		    if ($xmlData->{Run}[$i]{Name} eq $bustard_dir) {
			$run_ref = $xmlData->{Run}[$i];
			last;
		    }
		}
	    } elsif (ref($xmlData->{Run}) eq 'HASH') {
		$run_ref = $xmlData->{Run};
	    }

	    $$phasing_ref = $run_ref->{Parameters}[0]
	    if (defined($run_ref->{Parameters}[0]));

	    $$prephasing_ref = $run_ref->{Parameters}[1]
	    if (defined($run_ref->{Parameters}[1]));

	    printf STDERR "%s: found phase=%5.3f prephase=%5.3f\n",
	    $app_name, $$phasing_ref, $$prephasing_ref;
	}
    }
}

#------------------------------------------------------------------------------

sub store_applied_phasings($$$$$$$$)
{
    my $app_name = shift;
    my $lane_results_ref = shift;
    my $num_lanes = shift;
    my $num_reads = shift;
    my $bustard_path = shift;
    my $bustard_params_file = shift;
    my $phasing_options_ref = shift;
    my $param_is_legacy = shift;

    my $phasing_applied = undef;
    my $prephasing_applied = undef;
    my $phasings_applied_file = $bustard_params_file;
    my $bustard_dir = basename($bustard_path);

    if ($param_is_legacy) {
	legacy_read_applied_phasing_file($app_name, $phasings_applied_file,
					 \$phasing_applied,
					 \$prephasing_applied,
					 $bustard_dir);
    }

    my $phasing_dir = 'Phasing';
    my $xml_data;
    my $curr_options_ref;
    my %legacy_options;

    if ($param_is_legacy) {
	$legacy_options{'AutoFlag'} = 0;
	$legacy_options{'Path'} = '';
	$curr_options_ref = \%legacy_options;
    }

    for (my $lane_num = 1; $lane_num <= $num_lanes; ++$lane_num) {
	for (my $read_num = 1; $read_num <= $num_reads; ++$read_num) {
	    my $lane_read_ref = $lane_results_ref->[$read_num][$lane_num];

	    if ($param_is_legacy) {
		if ($phasing_applied > 0.) {
		    $lane_read_ref->{phasingapplied} = $phasing_applied;
		}

		if ($prephasing_applied > 0.) {
		    $lane_read_ref->{prephasingapplied} = $prephasing_applied;
		}
	    }

	    # Applied phasing files now provided by getPhasingFilename.

	    if (!$param_is_legacy) {
		$curr_options_ref = $phasing_options_ref->{$read_num};

		# FIXME : error?
		next if (!defined($curr_options_ref));
	    }

	    my $applied_phasing_file 
		= join('/', ($bustard_path,
			     getPhasingFilename($phasing_dir,
						$curr_options_ref,
						$lane_num)));

	    if (defined($xml_data = readFileXML($applied_phasing_file))) { 
		if (defined($xml_data->{Phasing})) {
		    $lane_read_ref->{phasingapplied} = $xml_data->{Phasing};
		}

		if (defined($xml_data->{Prephasing})) {
		    $lane_read_ref->{prephasingapplied}
		    = $xml_data->{Prephasing};
		}
	    }
	}
    }
}

#------------------------------------------------------------------------------

sub read_sample_file($$$$)
{
    my $appName = shift;
    my $sampleFile = shift;
    my $laneResultsRef = shift;
    my $num_reads = shift;

    my $xmlData;

    info(": looking in $sampleFile for sample information");

    if (defined($xmlData = readFileXML($sampleFile))) {
	info(": found sample file $sampleFile, trying to parse");

#    $xmlData=undef;
#    $xmlData=$xml->XMLin($sampleFile);
	for (my $i = 0; $i < scalar(@{$xmlData->{Lane}}); $i++) {
	    print STDERR $appName, 
	    ": found sample ID ", $xmlData->{Lane}[$i]{SampleID},
	    " for lane ", $xmlData->{Lane}[$i]{LaneNumber}, "\n";

#	print $xmlData->{Lane}[$i]{LaneNumber}
#	    , " -- ", $xmlData->{Lane}[$i]{SampleID}, "\n";
	    for (my $read_num = 1; $read_num <= $num_reads; ++$read_num) {
		$laneResultsRef->
		    [$read_num][$xmlData->{Lane}[$i]{LaneNumber}]{sample}
		= $xmlData->{Lane}[$i]{SampleID};
	    }
	}
    } else {
	warn "Unable to find file $sampleFile";
	for (my $lane = 1; $lane <= $maxNumLanes; $lane++) {
	    for (my $read_num = 1; $read_num <= $num_reads; ++$read_num) {
		$laneResultsRef->[$read_num][$lane]{sample} = $unknown;
	    }
	}
    }
}

#------------------------------------------------------------------------------

sub use_bases_to_read_lengths_list($)
{
    my $use_bases_str = shift;

    my @base_counts = ();

    my $curr_base_count = ($use_bases_str =~ tr/Y/Y/);
    push (@base_counts, $curr_base_count);

    $curr_base_count = ($use_bases_str =~ tr/y/y/);
    
    if ($curr_base_count > 0) {
        push (@base_counts, $curr_base_count);
    }

    my $list_str = join(', ', @base_counts);
    return $list_str;
}

#------------------------------------------------------------------------------

sub extract_lane_cycle_map($$$$$)
{
    my $xml_hash_ref = shift;
    my $lane_cycle_map_ref = shift;
    my $lane_results_ref = shift;
    my $num_lanes = shift;
    my $lane_num_reads_ref = shift;

    my %num_lane_reads_found;

    for (my $lane_num = 1; $lane_num <= $num_lanes; ++$lane_num) {
	$num_lane_reads_found{$lane_num} = 0;
    }

    # USE_BASES may be specified chip-wide (default for all lanes) ...

    my $chip_wide_ref = $xml_hash_ref->{'ChipWideRunParameters'};
    my $chipwide_use_bases_str = "";

    if (defined($chip_wide_ref)) {
	my $use_bases_str = $chip_wide_ref->{'USE_BASES'};

        if (defined($use_bases_str)) {
            my @cycle_map = ();
            $chipwide_use_bases_str = $use_bases_str;

            if (!($use_bases_str =~/^[YNyn]+$/)) {
                my $read_length = $chip_wide_ref->{'READ_LENGTH'};

                if (defined($read_length)) {
                    $use_bases_str = expandLegacyUseBasesString($use_bases_str,
                                                                $read_length);
                }
            }

            my $num_reads_found = getCycleMap($use_bases_str, \@cycle_map);

            if ($num_reads_found == 0) {
                warn("Failed to parse Chip Wide USE_BASES");
                return 0;
            }

            my $lengths_list = use_bases_to_read_lengths_list($use_bases_str);

            for (my $lane_num = 1; $lane_num <= $num_lanes; ++$lane_num) {
                @{$lane_cycle_map_ref->{$lane_num}} = @cycle_map;
                $num_lane_reads_found{$lane_num} = $num_reads_found;

                $lane_results_ref->[$ref_read_num][$lane_num]{lengths_list}
                = $lengths_list;
            }
        }
    }

    # ... or individually for each lane
    # ... or with default overridden for one or more lanes ...

    my $lane_spec_ref = $xml_hash_ref->{'LaneSpecificRunParameters'};

    if (defined($lane_spec_ref)) {
	my $use_bases_ref = $lane_spec_ref->{'USE_BASES'};
	my $read_length_ref = $lane_spec_ref->{'READ_LENGTH'};

	if (defined($use_bases_ref)) {
	    for (my $lane_num = 1; $lane_num <= $num_lanes; ++$lane_num) {
		# No analysis -> no need for USE_BASES - even if specified
		next if ($lane_results_ref->[$ref_read_num][$lane_num]{type}
			 eq 'NONE');

		my $use_bases_str = $use_bases_ref->{"s_$lane_num"};

		if (defined($use_bases_str) and not ref($use_bases_str)) {
		    # FIXME : code duplication
		    if (!($use_bases_str =~/^[YNyn]+$/)) {
			my $read_length = $read_length_ref->{"s_$lane_num"};

			if (defined($read_length)) {
			    $use_bases_str
				= expandLegacyUseBasesString($use_bases_str,
							     $read_length);
			} else
			{
                            $use_bases_str = $chipwide_use_bases_str; 
			}

		    }

		    my @cycle_map = ();
		    my $num_reads_found 
			= getCycleMap($use_bases_str, \@cycle_map);

		    if ($num_reads_found == 0) {
			warn "Failed to parse USE_BASES for lane $lane_num";
			return 0;
		    }

		    @{$lane_cycle_map_ref->{$lane_num}} = @cycle_map;
		    $num_lane_reads_found{$lane_num} = $num_reads_found;

                    $lane_results_ref->[$ref_read_num][$lane_num]{lengths_list}
                    = use_bases_to_read_lengths_list($use_bases_str);
		}
	    }
	}
    }


    # ... but each lane for which alignment is done must have a USE_BASES 
    #     specified somehow.

    my $is_ok = 1;

    for (my $lane_num = 1; $lane_num <= $num_lanes; ++$lane_num) {
        my $this_lane_is_ok = 1;
	my $analysis_type
	    = $lane_results_ref->[$ref_read_num][$lane_num]{type};

	# No analysis -> no need for USE_BASES.
	next if ($analysis_type eq 'NONE');

	my $num_reads_found = $num_lane_reads_found{$lane_num};

        # Do not check number of reads used for sequence[_pair] but do
        # proceed to the setting of length below, so yield will be set.
        unless (($analysis_type eq 'SEQUENCE')
                || ($analysis_type eq 'SEQUENCE_PAIR')) {
            my $num_reads_used = $lane_num_reads_ref->[$lane_num];

            if ($num_reads_found == 0) {
                warn("Did not find USE_BASES for lane $lane_num "
                     . "(or chip-wide).");
                $this_lane_is_ok = 0;
            } elsif ($num_reads_found != $num_reads_used) {
                warn("Expected $num_reads_used reads but found "
                     . "$num_reads_found for lane $lane_num (or chip-wide).");
                $this_lane_is_ok = 0;
            }
        }

        if (!$this_lane_is_ok) {
            $is_ok = 0;
        }

        # While all the required values are still accessible, set `length's.
        if ($this_lane_is_ok) {
            my $read_len_list
                = $lane_results_ref->[$ref_read_num][$lane_num]{lengths_list};
            my $read_num = 0;

            foreach my $read_len (split(/, /, $read_len_list)) {
                if (++$read_num > $num_reads_found) {
                    die("Read length list for lane $lane_num has more "
                        . "entries than expected ($num_reads_found)");
                }

                $lane_results_ref->[$read_num][$lane_num]{length} = $read_len;
            }
        }
    }

    return $is_ok;
}

#------------------------------------------------------------------------------

sub read_config_file($$$$$$$)
{
    my $configFile = shift;
    my $lane_results_ref = shift;
    my $maxNumLanes = shift;
    my $num_available_reads = shift;
    my $lane_num_reads_ref = shift;
    my $max_reads_used_ref = shift;
    my $lane_cycle_map_ref = shift;

    info(": looking in $configFile for target information");
    my %configKeywords = ( "ANALYSIS"    => "type",
			   "GENOME_FILE" => "template",
			   "ELAND_GENOME" => "eland",
			   "QF_PARAMS"   => "purity",
                           "CHASTITY_THRESHOLD" => "chastity_thresh" );

    my $xmlData = readFileXML($configFile);

    if (!defined($xmlData)) {
	warn "Failed to parse $configFile as XML\n";
	exit(1);
    }

    info(": found config file $configFile, trying to parse");

    if (defined($xmlData->{ChipWideRunParameters})) {
        foreach my $keyword (keys %configKeywords) {
            if (defined($xmlData->{ChipWideRunParameters}{$keyword})) {
                for (my $lane = 1; $lane <= $maxNumLanes; $lane++) {
                    for (my $read_num = 1; $read_num <= $num_available_reads;
                         ++$read_num) {
                        $lane_results_ref->
                            [$read_num][$lane]{$configKeywords{$keyword}}
                        = $xmlData->{ChipWideRunParameters}{$keyword};
                    }
                }
            }
        }
    }
    
    if (defined($xmlData->{LaneSpecificRunParameters})) {
        my %data = %{$xmlData->{LaneSpecificRunParameters}};

        foreach my $keyword (keys %configKeywords) {
            if (defined($data{$keyword})) {
                foreach my $key (keys %{$data{$keyword}}) {
                    my ($lane) = $key =~ m/[^_]+_(\d+)/;

                    for (my $read_num = 1; $read_num <= $num_available_reads;
                         ++$read_num) {
                        $lane_results_ref->
                            [$read_num][$lane]{$configKeywords{$keyword}} 
                        = $data{$keyword}{$key};
                    }
                }
            }
        }
    }

    # no output for ANALYSIS none, use ELAND_GENOME as Eland target
    for (my $lane_num = 1; $lane_num <= $maxNumLanes; ++$lane_num) {
        my $analysis_type
	    = uc($lane_results_ref->[$ref_read_num][$lane_num]{type});

        # Only eland_pair and sequence_pair use more than 1 read 
        # no matter how many are available.
        if (($analysis_type ne "ELAND_PAIR") 
	    && ($analysis_type ne "SEQUENCE_PAIR")) {
            $lane_num_reads_ref->[$lane_num] = 1;
        }

        my $num_reads_used = $lane_num_reads_ref->[$lane_num];

        if ($num_reads_used > $$max_reads_used_ref) {
            $$max_reads_used_ref = $num_reads_used;
        }

        for (my $read_num = 1; $read_num <= $num_available_reads;
             ++$read_num) {
	    my $lane_read_ref = $lane_results_ref->[$read_num][$lane_num];
            $lane_read_ref->{type} = $analysis_type;

            if ($analysis_type eq "NONE") {
                foreach my $value (qw/length template/) {
                    $lane_read_ref->{$value} = $unknown;
                }
            } elsif (($analysis_type eq "SEQUENCE")
                     || ($analysis_type eq "SEQUENCE_PAIR")) {
                foreach my $value (qw/template/) {
                    $lane_read_ref->{$value} = $unknown;
                }
            } elsif (($analysis_type eq "ELAND")
		     || ($analysis_type eq "ELAND_EXTENDED") 
		     || ($analysis_type eq "ELAND_PAIR")) {
                $lane_read_ref->{template} = basename($lane_read_ref->{eland});
            }
        }
    }

    # Extract lane cycle map (also sets lengths_list and length values)
    # (needs to be after extraction and upper-casing of analysis types).

    if (!extract_lane_cycle_map($xmlData, $lane_cycle_map_ref,
				$lane_results_ref, $maxNumLanes,
                                $lane_num_reads_ref)) {
	exit(1);
    }
}

#------------------------------------------------------------------------------

sub read_chip_file($$$)
{
    my $appName = shift;
    my $chipFile = shift;
    my $chipNameRef = shift;

    info(": looking in $chipFile for chip information");
    my $xmlData;

    if (defined($xmlData = readFileXML($chipFile))) {
	info(": found chip file $chipFile, trying to parse");

	if (defined ($xmlData->{ChipID})) {
	    print STDERR $appName, 
	    ": found chip ID ", $xmlData->{ChipID}, "\n";
	    $$chipNameRef = $xmlData->{ChipID};
	} else {
	    warn "Unable to find field ChipID in file $chipFile"; 
	}
    }
#else
#{
#    warn "Unable to find file $chipFile";
#} # else

}

#------------------------------------------------------------------------------

sub read_tiles_file($$)
{
    my $tiles_file = shift;
    my $tiles_ref = shift;

    if (!open(TILES, $tiles_file)) {
	warn "Failed to open tile list file `", $tiles_file, "'.\n";
	exit(0);
    }

    while (my $line = <TILES>) {
	next if ($line =~ m/^\#/);

	my @curr_tiles = split(" ", $line);
	push @$tiles_ref, @curr_tiles;
    }

    close(TILES);
}

#------------------------------------------------------------------------------
# The _all.txt file is assumed to contain all of the cycles seen by Bustard,
# i.e. not masked by a USE_BASES string. The calculated statistics, however,
# must be based on the masked cycles. In addition, there may be more than
# a single read and statistics must be calculated separately for each.
#
# These requirements are met by mapping the combined cycle number to a
# read number and a cycle number within masked read using %lane_cycle_map.
# A read number of 0 indicates a masked out cycle so such a cycle is skipped.

sub process_all_file($$$$$$$)
{
    my $file_name = shift;
    my $tile_ind = shift;
    my $curr_tile_lane = shift;
    my $lane_results_ref = shift;
    my $results_ref = shift;
    my $lane_cycle_map_ref = shift;
    my $num_reads = shift;

    my @array;

    my $thisIntensity;
    my @cyc2to10num;
    my @cyc2to10inten;
    my @cyc10to20num;
    my @cyc10to20inten;

    my ($combined_cycle_num, $read_num, $cycle_num);

    # Initialise tile statistics.
    # (These are also the final values if the tile _all file cannot be opened.)
    for ($read_num = 1; $read_num <= $num_reads; ++$read_num) {
	my $tile_read_ref = $results_ref->[$read_num][$tile_ind];
	$tile_read_ref->{oneSig} = 0;
	$tile_read_ref->{sigAvg1to3} = 0;
	$tile_read_ref->{sigAvg2to4} = 0;
	$tile_read_ref->{sig20AsPctOf1} = 0;
	$tile_read_ref->{sigLoss2to10} = undef;
	$tile_read_ref->{sigLoss10to20} = undef;
    }

    my $raw_cluster_count = 0;
    my $PF_cluster_count = 0;

    # Read _all file (one line per combined cycle after the comments)
    # and accumulate statistics.
    
    if (defined(open(IN, $file_name))) {
	@cyc2to10num = ();
	@cyc2to10inten = ();
	@cyc10to20num = ();
	@cyc10to20inten = ();

	while (<IN>) {
	    if ($_ =~ m/^\#/) {
		if ($_ 
		    =~ m/^\# Clusters: Filtered ([0-9]+) Original ([0-9]+)/) {

		    $PF_cluster_count = $1;
		    $raw_cluster_count = $2;
		}

		next;
	    }

	    @array = split('\t', $_);
	    $combined_cycle_num = $array[1];

	    $read_num = $lane_cycle_map_ref->
            {$curr_tile_lane}[$combined_cycle_num]{read_num};

	    next if ($read_num == 0); # i.e. the cycle is masked out.

	    my $tile_read_ref = $results_ref->[$read_num][$tile_ind];

	    $cycle_num = $lane_cycle_map_ref->
            {$curr_tile_lane}[$combined_cycle_num]{cycle_num};

	    $thisIntensity = (($array[2] + $array[3] + $array[4] + $array[5]) 
			      / $numChannels);

	    if (($cycle_num >= 2) && ($cycle_num <= 4)) {
		$tile_read_ref->{sigAvg2to4} += $thisIntensity;
	    }
	    
	    if (($cycle_num >= 2) && ($cycle_num <= 10)) {
		push @cyc2to10num, $cycle_num;
		push @cyc2to10inten, $thisIntensity;
	    }

	    if (($cycle_num >= 10) && ($cycle_num <= 20)) {
		push @cyc10to20num, $cycle_num;
		push @cyc10to20inten, $thisIntensity;
	    }

	    if ($cycle_num =~ /^[123]$/) {
		$tile_read_ref->{sigAvg1to3} += $thisIntensity;

		if ($cycle_num eq '1') {
		    $tile_read_ref->{oneSig} += $thisIntensity;
		}
	    } elsif ($cycle_num eq '20') {
		$tile_read_ref->{sig20AsPctOf1} += $thisIntensity;

		# Skip remainder of file once all needed intensities read.
		# NB assumes cycle 20 data always comes after the other cycles
		#    of interest and (for multiple reads) that cycle 20 data
		#    for the last read always comes after that for the others. 
		last if ($read_num == $num_reads); 
	    }
	} # while loop over _all.txt lines corresponding to combined cycles


	# Finish calculating multi-cycle statistics.

	for ($read_num = 1; $read_num <= $num_reads; ++$read_num) {
	    my $tile_read_ref = $results_ref->[$read_num][$tile_ind];

	    if ($tile_read_ref->{oneSig} != 0) {
		$tile_read_ref->{sig20AsPctOf1}
		= ($tile_read_ref->{sig20AsPctOf1}
		   / $tile_read_ref->{oneSig}) * 100;
	    } else {
		$tile_read_ref->{sig20AsPctOf1} = undef;
	    }
	
	    if (defined($tile_read_ref->{sigAvg1to3})) {
		$tile_read_ref->{sigAvg1to3} /= 3;
	    }

	    if (defined($tile_read_ref->{sigAvg2to4})) {
		$tile_read_ref->{sigAvg2to4} /= 3;
	    }

	    $tile_read_ref->{sigLoss2to10}
	    = logLinearFit(\@cyc2to10num, \@cyc2to10inten);
	    $tile_read_ref->{sigLoss10to20}
	    = logLinearFit(\@cyc10to20num, \@cyc10to20inten);
	}
    } else {
	warn("`$file_name' not found.");
    }


    # Store cluster count statistics (same values for all reads).

    for ($read_num = 1; $read_num <= $num_reads; ++$read_num) {
	my $tile_read_ref = $results_ref->[$read_num][$tile_ind];

	$tile_read_ref->{clusterCountRaw} = $raw_cluster_count;
	$tile_read_ref->{clusterCountPF} = $PF_cluster_count;

	if ($PF_cluster_count != 0) {
	    $lane_results_ref->[$read_num][$curr_tile_lane]{tileCountPF}++;
	}

	$tile_read_ref->{PF_percent} 
	= (($raw_cluster_count != 0)
	   ? ($PF_cluster_count / $raw_cluster_count * 100) 
	   : 0);
    }
}

#------------------------------------------------------------------------------

sub process_score_file($$$$$$$$$$$)
{
    my $file_name = shift;
    my $infoContentKey = shift;
    my $errorKey = shift;
    my $pc_align_key = shift;
    my $av_align_score_key = shift;
    my $tile_ind = shift;
    my $curr_tile_lane = shift;
    my $num_reads = shift;
    my $results_ref = shift;
    my $lane_results_ref = shift;
    my $cluster_count = shift;

    my $info_content_line = undef;
    my $num_unique_aligns = undef;
    my $total_unique_align_score = undef;

    for (my $read_num = 0; $read_num <= $num_reads; ++$read_num) { # 0 -> All
        my $tile_read_ref = $results_ref->[$read_num][$tile_ind];
        # Should these all be set to undef rather than to 0,
        # i.e. should attempts to divide 0 by 0 be reported as 0 or unknown?
        $tile_read_ref->{$errorKey} = 0;
        $tile_read_ref->{$infoContentKey} = 0;
        $tile_read_ref->{$pc_align_key} = 0;
        $tile_read_ref->{$av_align_score_key} = 0;
    }

    my $read_num = 1;
    my $highest_read_num = $read_num;
    my $tile_read_ref = $results_ref->[$read_num][$tile_ind];

    if (defined(open(IN, $file_name))) {
        for (my $read_num = 1; $read_num <= $num_reads; ++$read_num) {
            $lane_results_ref->[$read_num][$curr_tile_lane]{aligned} = 1;
        }

        my $in_info_content = 0;

	while (<IN>) {
#	    print $_; die;
            
            if ($_ =~ /^\# Lane.*Read/) {
                if ($_ =~ /^\# Lane.*Read (\d+)/) {
                    $read_num = $1;

                    if ($read_num > $num_reads) {
                        die("Expected $num_reads reads but found "
                            . "Read $read_num in $file_name\n");
                    }

                    if ($read_num > $highest_read_num) {
                        $highest_read_num = $read_num;
                    }
                } else {
                    $read_num = 0; # All Reads
                }

                $tile_read_ref = $results_ref->[$read_num][$tile_ind];
            } elsif ($_ =~ m/ percent error rate/) {
		my @array = split(' ',$_);
		$tile_read_ref->{$errorKey} = $array[0];
            } elsif ($_ =~ /unique alignments : (\d+) \(total score (\d+)\)/) {
                $num_unique_aligns = $1;
                $total_unique_align_score = $2;

                if ($cluster_count != 0) {
                    $tile_read_ref->{$pc_align_key}
                    = ($num_unique_aligns / $cluster_count) * 100;

                    $tile_read_ref->{$av_align_score_key}
                    = $total_unique_align_score / $cluster_count;
                }
	    } elsif ($_ =~ /^\~/) {
		$info_content_line = $_;
                $in_info_content = 1;
	    } elsif ($in_info_content) {
                # Now at the line after the last info content for a stats set.
                # Process that last info content line.
                $in_info_content = 0;

                chomp($info_content_line);
                $info_content_line =~ s/^\~//;
                my @array = split('\t', $info_content_line);

                $tile_read_ref->{$infoContentKey} = $array[4] / $array[0]
                    if ((@array == 7) && ($array[0] != 0));
	    }

	} # while <IN>

	close(IN);

        if ($highest_read_num < $num_reads) {
            die("Expected $num_reads but highest Read found in "
                . " $file_name was Read $highest_read_num\n");
        }
    }
}

#------------------------------------------------------------------------------
# Sum all leaf values under an XML node handling arbitrary levels of
# nesting using recursion.

# forward declaration apparently needed for recursion in Perl.
sub sum_vals_under_node($);

sub sum_vals_under_node($)
{
    my $xml_ref = shift;

    if (!ref($xml_ref)) {
	# There is no nesting; this is already a leaf value.
	return $xml_ref;
    }

    my $sum = 0;
    my @sub_tags = keys (%{$xml_ref});

    foreach my $sub_tag (@sub_tags) {
	my $val = $xml_ref->{$sub_tag};

	if (!ref($val)) {
	    # a leaf value
	    $sum += $val;
	} elsif (ref($val) eq 'HASH') {
	    # more nesting
	    $sum += sum_vals_under_node($val);
	} else {
	    die "Found unexpected reference type in XML structure\n";
	}
    }

    return $sum;
}

#------------------------------------------------------------------------------

sub process_pair_info_file($$)
{
    my $pair_results_ref = shift;
    my $lane_num = shift;

    my $pair_info_fn = "s_${lane_num}_pair.xml";
    my $xml_data = readFileXML($pair_info_fn);

    if (defined($xml_data)) {
	my %undef_hash;
	%{$pair_results_ref->{$lane_num}} = %undef_hash;
	my $lane_pair_ref = $pair_results_ref->{$lane_num};
	$lane_pair_ref->{$paired_not_done_key} = 0; # force to exist

	my $inserts_data = $xml_data->{$Insert_Size_str};

	if (defined($inserts_data)) {
            my @insert_stat_strs = map { $_->[0] } @insert_stat_display_map;

	    foreach my $stat_str (@insert_stat_strs) {
		$lane_pair_ref->{"$Insert_Size_str:$stat_str"}
		= (defined($inserts_data->{$stat_str})
		   ? $inserts_data->{$stat_str}
		   : $not_avail_str);
	    }
	}

	my $orient_data = $xml_data->{$Orientation_str};

	if (defined($orient_data)) {
	    # Insert size stats

            my @orientation_keys = map { $_->[0] } @orient_display_map;

	    foreach my $stat_key (@orientation_keys, $nominal_orient_key) {
		$lane_pair_ref->{"$Orientation_str:$stat_key"}
		= (defined($orient_data->{$stat_key})
		   ? $orient_data->{$stat_key}
		   : 0);
	    }

	    # Insert frequency stats

	    my @insert_freq_keys = map { $_->[0] } @inserts_display_map;

	    foreach my $stat_key (@insert_freq_keys) {
		# not provided so derived instead
		next if ($stat_key eq $num_pairs_ok_key);

		$lane_pair_ref->{"$Insert_Size_str:$stat_key"}
		= (defined($orient_data->{$stat_key})
		   ? $orient_data->{$stat_key}
		   : 0);

		my $pc_key = $insert_num_pc_key_map{$stat_key};

		$lane_pair_ref->{"$Insert_Size_str:$pc_key"}
		= (defined($orient_data->{$pc_key})
		   ? sprintf($insert_pc_format, $orient_data->{$pc_key})
		   : '0.0');
	    }
	}

	my $reads_data = $xml_data->{'Reads'};

	if (defined($reads_data)) {
            my @align_outcome_keys 
                = map { $_->[0] } @align_outcome_display_map;

	    foreach my $read1_outcome_key (@align_outcome_keys) {
		my $key1_str = $Read_str . "1" . $read1_outcome_key;
		my $read1_data = $reads_data->{$key1_str};

		if (defined($read1_data)) {
		    foreach my $read2_outcome_key (@align_outcome_keys) {
			my $key2_str = $Read_str . "2" . $read2_outcome_key;
			my $read2_data = $read1_data->{$key2_str};

			if (defined($read2_data)) {
			    # Find all leaf values at arbitrary levels of 
			    # nesting under this node and sum them to produce 
			    # the main entry of interest.
			    $lane_pair_ref->{"$key1_str:$key2_str"}
			    = sum_vals_under_node($read2_data);

			    # A few special cases at this level.
			    my @inner_keys = keys (%{$read2_data});

			    foreach my $inner_key (@inner_keys) {
				if (($inner_key eq $unique_pair_align_key)
				    || ($inner_key eq $multi_pair_align_key)) {
				    $lane_pair_ref->
				    {"$key1_str:$key2_str:$inner_key"}
				    = sum_vals_under_node($read2_data->
							  {$inner_key});
				} elsif ($inner_key eq $paired_not_done_key) {
				    $lane_pair_ref->{$paired_not_done_key} = 1;
				}
			    }
			} else {
			    $lane_pair_ref->{"$key1_str:$key2_str"} = 0;
			}
		    }
		} else {
		    foreach my $read2_outcome_key (@align_outcome_keys) {
			my $key2_str = $Read_str . "2" . $read2_outcome_key;
			$lane_pair_ref->{"$key1_str:$key2_str"} = 0;
		    }
		}
	    }
	}
    }
}

#------------------------------------------------------------------------------
# Break down the unique paired alignments according to which of the individual
# read alignments were non-unique but `rescuable' - read1 only, read2 only
# or both reads.
#
# Calculate the percentages relative to the total in each rescuable category.

sub derive_uniqueness_rescue_stats($$)
{
    my $lane_pair_ref = shift;
    my $separ_str = shift;

    my $total = 0;
    my $total_divisor = 0;

    # The $bitmask is just a device to represent each read separately and
    # then both reads.
    for (my $bitmask = 1; $bitmask < (1 << $num_paired_reads); ++$bitmask) {
	my $all_rescued = 1;
	my @key_parts = ();
	my $new_key = '';
	
	for (my $read_num = 1; $read_num <= $num_paired_reads; ++$read_num) {
	    my $rescued = ($bitmask & (1 << ($read_num - 1)));

	    if ($rescued) {
		$new_key = $Read_str . $read_num;
	    } else {
		$all_rescued = 0;
	    }

	    push(@key_parts, $Read_str . $read_num . ($rescued
						      ? $Rescuable_key
						      : $Unique_key));
	}

	if ($all_rescued) {
	    $new_key = $All_Reads_key;
	}

	$new_key = join($separ_str, ($new_key, $Rescued_Unique_key));
	my $new_pc_key = join($separ_str, ($pc_prefix, $new_key));
	my $base_key = join($separ_str, @key_parts);
	my $val = $lane_pair_ref->{join($separ_str, ($base_key,
						     $unique_pair_align_key))};

	my $divisor = $lane_pair_ref->{$base_key};
	my $pc_val;

	if (defined($val) && defined($divisor)) {
	    $pc_val = (($divisor > 0)
		       ? ($val / $divisor * 100.0)
		       : 0);
	} else {
	    $pc_val = 0;
	    $val = 0 if !defined($val);
	}

	$lane_pair_ref->{$new_key} = $val;
	$lane_pair_ref->{$new_pc_key} = $pc_val;
	$total += $val;
	$total_divisor += (defined($divisor) ? $divisor : 0);
    }

    my $total_key = join($separ_str, ($Total_str, $Rescued_Unique_key));
    $lane_pair_ref->{$total_key} = $total;
    $lane_pair_ref->{join($separ_str, ($pc_prefix, $total_key))}
    = (($total_divisor > 0)
       ? ($total / $total_divisor * 100.0)
       : 0);
}

#------------------------------------------------------------------------------
# By `mispairing' is meant that one read has been aligned (does not matter
# whether uniquely or otherwise) but its partner has not (whether because
# it was low quality - too many unknown bases - or simply could not be matched
# to the reference).
#
# The percentages calculated are relative to total number of aligned partner
# reads (not total number of read pairs).  

sub derive_mispairing_stats($$)
{
    my $lane_pair_ref = shift;
    my $separ_str = shift;

    for (my $ref_read_num = 1; $ref_read_num <= 2; ++$ref_read_num) {
        my $lost_read_num = ($ref_read_num == 1) ? 2 : 1;

        my $count = 0;
	my $divisor = 0;

        foreach my $ref_key ($Unique_key, $Rescuable_key, $Repeat_key) {
            foreach my $lost_key($Not_Matched_key, $Low_Quality_key) {

                my $read1_key = "${Read_str}1" . (($ref_read_num == 1)
                                                  ? $ref_key
                                                  : $lost_key);
                my $read2_key = "${Read_str}2" . (($ref_read_num == 2)
                                                  ? $ref_key
                                                  : $lost_key);

                my $val = $lane_pair_ref->{join($separ_str,
                                                ($read1_key, $read2_key))};

                if (defined($val)) {
                    $count += $val;
                }
            }

	    my @key_strs = ($Read_str . $ref_read_num . $ref_key, $Total_str);
	    
	    if ($ref_read_num != 1) {
		@key_strs = reverse @key_strs;
	    }

	    my $val = $lane_pair_ref->{join($separ_str, @key_strs)};

	    if (defined($val)) {
		$divisor += $val;
	    }
        }

	my $key = $Read_str . $lost_read_num . $Lost_str;
	my $pc_key = join($separ_str, ($pc_prefix, $key));

        $lane_pair_ref->{$key} = $count;
	$lane_pair_ref->{$pc_key} = (($divisor != 0)
                                     ? ($count / $divisor * 100.0)
                                     : 0.0);
    }
}

#------------------------------------------------------------------------------

sub derive_insert_freq_stats($)
{
    my $lane_pair_ref = shift;

    my $nominal_orient =
	$lane_pair_ref->{"$Orientation_str:$nominal_orient_key"};

    my $ok_freq = (defined($nominal_orient)
                   ? $lane_pair_ref->{"$Orientation_str:$nominal_orient"}
                   : 0);

    foreach my $bad_freq_key ($num_orient_ok_small_insert_key,
			      $num_orient_ok_large_insert_key) {
	my $key = "$Insert_Size_str:$bad_freq_key";
	my $bad_freq = (defined($lane_pair_ref->{$key})
			? $lane_pair_ref->{$key}
			: 0);
	$ok_freq -= $bad_freq;
    }

    $lane_pair_ref->{"$Insert_Size_str:$num_pairs_ok_key"} = $ok_freq;


    # Calc total num inserts and use as denominator in percentage.

    my $total_inserts = 0;
    my @orientation_keys = map { $_->[0] } @orient_display_map;

    foreach my $stat_key (@orientation_keys) {
        my $num_inserts = $lane_pair_ref->{"$Orientation_str:$stat_key"};
	$total_inserts += (defined($num_inserts) ? $num_inserts : 0);
    }

    my $pc_key = $insert_num_pc_key_map{$num_pairs_ok_key};

    $lane_pair_ref->{"$Insert_Size_str:$pc_key"} 
    = sprintf($insert_pc_format, (($total_inserts != 0)
                                  ? (100.0 * $ok_freq / $total_inserts)
                                  : 0.0));
}

#------------------------------------------------------------------------------

sub derive_pair_stats($)
{
    my $lane_pair_ref = shift;
    my $paired_align_done = !$lane_pair_ref->{$paired_not_done_key};

    # Individual Alignment Outcomes
    calc_total_by_field($lane_pair_ref,
                        ["^${Read_str}1", "^${Read_str}2"],
                        1, $pair_key_separ_str);
    calc_total_by_field($lane_pair_ref,
                        ["^${Read_str}1", "^(${Read_str}2|${Total_str})"],
                        0, $pair_key_separ_str);
    calc_percent_by_field($lane_pair_ref,
                          ["^(${Read_str}1|${Total_str})",
                           "^(${Read_str}2|${Total_str})"],
                          "${Total_str}:${Total_str}",
                          $pair_key_separ_str);


    # Breakdown of UniquePairedAlignment, ManyPairedAlignments
    if ($paired_align_done) {
	foreach my $pair_align_key ($unique_pair_align_key,
				    $multi_pair_align_key) {
	    calc_total_by_field($lane_pair_ref,
				["^${Read_str}1", "^${Read_str}2",
				 "^$pair_align_key\$"],
				1, $pair_key_separ_str);
	    calc_total_by_field($lane_pair_ref,
				["^${Read_str}1",
				 "^(${Read_str}2|${Total_str})",
				 "^$pair_align_key\$"],
				0, $pair_key_separ_str);

	    calc_percent_by_field($lane_pair_ref,
				  ["^(${Read_str}1|${Total_str})",
				   "^(${Read_str}2|${Total_str})",
				   "^${pair_align_key}\$"],
				 "${Total_str}:${Total_str}:${pair_align_key}",
				  $pair_key_separ_str);
	}
    }

    # Uniqueness rescue
    if ($paired_align_done) {
	derive_uniqueness_rescue_stats($lane_pair_ref, $pair_key_separ_str);
    }

    # Mispairing
    derive_mispairing_stats($lane_pair_ref, $pair_key_separ_str);

    # Orientations
    my @orientation_strs = map { $_->[0] } @orient_display_map;
    my $orient_regex = '^(' . join('|', @orientation_strs) . ')$';
    calc_total_by_field($lane_pair_ref,
                        ["^${Orientation_str}\$", $orient_regex],
                        1, $pair_key_separ_str);
    calc_percent_by_field($lane_pair_ref,
                          ["^${Orientation_str}\$", $orient_regex],
                          "${Orientation_str}:${Total_str}",
                          $pair_key_separ_str);

    # Insert size freqs
    if ($paired_align_done) {
	derive_insert_freq_stats($lane_pair_ref);
    }
}

#------------------------------------------------------------------------------
# Some statistics are relevant only if alignment has been performed.
# Some statistics are based on all clusters (raw), some only on 
# purity-filtered clusters (PF).

sub get_stat_names_by_type($$$$$)
{
    my $is_aligned = shift;
    my $stats_raw_ref = shift;
    my $stats_PF_ref = shift;
    my $align_stats_ref = shift;
    my $total_stats_read_ref = shift;

    my @prealign_stats_raw = (qw/clusterCountRaw/);
    my @align_stats_raw = qw/infoContentRaw errorRaw/;

    my @prealign_stats_PF = qw/clusterCountPF PF_percent oneSig sig20AsPctOf1 sigAvg1to3 sigAvg2to4 sigLoss2to10 sigLoss10to20/;
    my @align_stats_PF = qw/errorPF pctAlignedPF avAlignScorePF infoContentPF/;

    if ($is_aligned) {
	@$stats_raw_ref = (@prealign_stats_raw, @align_stats_raw);
	@$stats_PF_ref = (@prealign_stats_PF, @align_stats_PF);
    } else {
	@$stats_raw_ref = @prealign_stats_raw;
	@$stats_PF_ref = @prealign_stats_PF;
    }

    @$align_stats_ref = (@align_stats_raw, @align_stats_PF);

    # Some chip total statistics (e.g. clusters) should be
    # counted only for one read (e.g. read1);
    # others (e.g. yield) should be summed over reads
    # [the convention is that read0 means all reads].
    %$total_stats_read_ref = ('yield' => 0,
			      'clusterCountRaw' => 1,
			      'clusterCountPF' => 1);
}

#------------------------------------------------------------------------------

sub calc_lane_results($$$$$$)
{
    my $laneResultsRef = shift;
    my $resultsRef = shift;
    my $maxNumLanes = shift;
    my $lane_num_reads_ref = shift;
    my $lane_type_ref = shift;
    my $numTiles = shift;

    my (@stats, @stats_raw, @stats_PF, @align_stats,
	%total_stats_read);

    my $thisVal;
    my $numSamples;

    for (my $lane_num = 1; $lane_num <= $maxNumLanes; $lane_num++) {
        my $num_reads = $lane_num_reads_ref->[$lane_num];

        for (my $read_num = 1; $read_num <= $num_reads; ++$read_num) {
	    my $lane_read_ref = $laneResultsRef->[$read_num][$lane_num];

            # Some statistics are relevant only if alignment has been performed
            # - assumed to correspond to the presence or absence of _realign 
            # files for tiles in this lane.
            my $is_aligned
                = ($lane_read_ref->{aligned} != 0);

            get_stat_names_by_type($is_aligned, \@stats_raw, \@stats_PF,
                                   \@align_stats, \%total_stats_read);

            if (!$is_aligned) {
                # Alignment apparently not (yet) done, so undefine all the 
                # alignment stats so that they are reported as `unknown' 
                # not zero.
                foreach my $thisStat (@align_stats) {
                    $lane_read_ref->{$thisStat} = undef;
                }
            }

            @stats = (@stats_raw, @stats_PF);

            if (exists $lane_type_ref->{$lane_num}) {
                for (my $tileInd = 0; $tileInd < $numTiles; $tileInd++) {
		    my $tile_read_ref = $resultsRef->[$read_num][$tileInd];

		    next if (!defined($tile_read_ref->{lane}));
                    next if ($tile_read_ref->{lane} != $lane_num);

		    # New policy (2007-10-23) is to ignore tiles with zero 
		    # clusters passing QF for ALL stats - not just QF ones.

                    if ($tile_read_ref->{PF_percent} != 0) {
                        foreach my $thisStat (@stats) {
			    $thisVal = 0;

			    if (defined($tile_read_ref->{$thisStat})) {
				$thisVal = $tile_read_ref->{$thisStat};
			    }

                            $lane_read_ref->{$thisStat}{mean} += $thisVal;
                            $lane_read_ref->{$thisStat}{sumsq} += ($thisVal
								   * $thisVal);
                        }
                    }
                } # for tileInd

                foreach my $thisStat (@stats) {
                    # Need to average % of purity filtered (PF) sequences and 
                    # any raw stats over all tiles.
                    # Other stats should only be averaged over tiles that
                    # contained at least 1 sequence that passed the filter.

                    $numSamples = $lane_read_ref->{tileCountPF};

		    # New policy (2007-10-23) is to ignore tiles with zero 
		    # clusters passing QF for ALL stats - not just QF ones.
                    # foreach ('PF_percent', @stats_raw) {
                    #    if ($thisStat eq $_) {
                    #        $numSamples = $lane_read_ref->{tileCountRaw};
                    #    }
                    # }

#	            print STDERR "$thisStat $numSamples $lane_num", "\n";

                    my $stat_ref = $lane_read_ref->{$thisStat};

                    if (defined($stat_ref)) {
                        if ($numSamples >= 2) {
                            # NB : ((sum *  sum) / N) == ((mean * mean) * N)
                            $stat_ref->{stdev}
                            = ($stat_ref->{sumsq}
                               - (($stat_ref->{mean} * $stat_ref->{mean})
                                  / $numSamples));

                            $stat_ref->{stdev} /= ($numSamples - 1);

			    if ($stat_ref->{stdev} < 0) {
				$stat_ref->{stdev} = 0;
			    }

                            $stat_ref->{stdev} = sqrt($stat_ref->{stdev});
                        } else {
                            $stat_ref->{stdev} = undef;
                        }

                        if ($numSamples > 0) {
                            $stat_ref->{mean} /= $numSamples;
                        } else {
                            $stat_ref->{mean} = undef;
                        }
                    }
                } # foreach (@stats)

                # Yield (bases) stats are obtained from clusterCountPF by 
                # simply multiplying by tileCountPF and by read length.
                if (defined($lane_read_ref->{clusterCountPF})
		    && ($lane_read_ref->{length} ne $unknown)) {
                    $lane_read_ref->{yield}{total}
		    = ($lane_read_ref->{clusterCountPF}{mean}
		       * $lane_read_ref->{tileCountPF}
		       * $lane_read_ref->{length});
                } else {
                    $lane_read_ref->{yield}{total} = 0;
                }

#	    $laneResultsRef->[$lane_num]{runFolder}=$runFolder;
#	    $laneResultsRef->[$lane_num]{machine}=$machineName;
#	    $chipResults{runFolder}=$runFolder;
#	    $chipResults{machine}=$machineName;
	
            } # if
        } # for read_num
    } #  for lane_num
}

#------------------------------------------------------------------------------

sub calc_chip_results($$$$$$)
{
    my $lane_results_ref = shift;
    my $num_lanes = shift;
    my $lane_num_reads_ref = shift;
    my $max_reads_used = shift;
    my $chip_stats_by_read_ref = shift;
    my $chip_totals_ref = shift;

    my (@stats_raw, @stats_PF, @align_stats,
	%total_stats_read);

    # Alignment stats are filtered out below if necessary (by lane),
    # so ask for all @stats_raw and @stats_PF by specifying $is_aligned as 1.
    get_stat_names_by_type(1, \@stats_raw, \@stats_PF, \@align_stats, 
			   \%total_stats_read);

    my @stats = (@stats_raw, @stats_PF);

    foreach my $curr_stat (@stats) {
	# Does this stat require alignment to be meaningful?
	my $stat_requires_alignment = 0;

	foreach (@align_stats) {
	    if ($curr_stat eq $_) {
		$stat_requires_alignment = 1;
		last;
	    }
	}

	# Is this stat purity filtered?
	my $stat_is_PF = 0;

	foreach ('PF_percent', @stats_raw) {
	    if ($curr_stat eq $_) {
		$stat_is_PF = 1;
		last;
	    }
	}

	my @read_total_tiles_to_count = ();
	my @read_chip_sums = ();

        for (my $read_num = 1; $read_num <= $max_reads_used; ++$read_num) {
            $read_chip_sums[$read_num] = 0;
            $read_total_tiles_to_count[$read_num] = 0;
        }

        for (my $lane_ind = 1; $lane_ind <= $num_lanes; ++$lane_ind) {
            my $num_reads = $lane_num_reads_ref->[$lane_ind];

            for (my $read_num = 1; $read_num <= $num_reads; ++$read_num) {
                my $lane_read_ref = $lane_results_ref->[$read_num][$lane_ind];

                # Some statistics are relevant only if alignment has been 
                # performed - assumed to correspond to the presence or absence
                # of _realign files for tiles in this lane and read.
                my $is_aligned = ($lane_read_ref->{aligned} != 0);
                next if ($stat_requires_alignment && !$is_aligned);

                next if (!defined($lane_read_ref->{$curr_stat}));

#               New policy (2007-10-23) is to ignore tiles with zero clusters 
#               passing QF for ALL stats - not just QF ones.
#                my $num_tiles_to_count
#                    = $lane_read_ref->{$stat_is_PF 
#                                           ? 'tileCountPF'
#                                           : 'tileCountRaw'};
		my $num_tiles_to_count = $lane_read_ref->{'tileCountPF'};

                next if ($num_tiles_to_count == 0);

                # The lane means are over relevant number of tiles,
                # so have to be multipled by those to get the original sums.
                $read_chip_sums[$read_num]
                    += ($lane_read_ref->{$curr_stat}{mean}
                        * $num_tiles_to_count);

                my $cluster_stat 
                    = ($stat_is_PF ? 'clusterCountPF' : 'clusterCountRaw');
                my $num_clusters_to_count
                    = ($lane_read_ref->{$cluster_stat}{mean}
                       * $num_tiles_to_count);

                if (!defined($lane_read_ref->{$cluster_stat}{total})) {
                    $lane_read_ref->{$cluster_stat}{total}
                    = $num_clusters_to_count;
                }

                $read_total_tiles_to_count[$read_num] += $num_tiles_to_count;
            } # read_num
        } # lane_ind

        for (my $read_num = 1; $read_num <= $max_reads_used; ++$read_num) {
            $chip_stats_by_read_ref->[$read_num]{$curr_stat}{by_tile}
            = (($read_total_tiles_to_count[$read_num] != 0)
               ? ($read_chip_sums[$read_num] 
                  / $read_total_tiles_to_count[$read_num])
               : undef);
        }
    } # stat

    # For some stats, totals across lanes and (perhaps) reads are meaningful.
    foreach my $curr_stat (keys %total_stats_read) {
	my $chip_sum = 0;
	my $total_stat_read = $total_stats_read{$curr_stat};

        for (my $lane_ind = 1; $lane_ind <= $num_lanes; $lane_ind++) {
            my $num_reads = $lane_num_reads_ref->[$lane_ind];

            for (my $read_num = 1; $read_num <= $num_reads; ++$read_num) {
		# Some chip total statistics (e.g. clusters) should be
		# counted only for one read (e.g. read1);
		# others (e.g. yield) should be summed over reads
		# [the convention is that read0 means all reads].
		next if (($total_stat_read != 0)
			 && ($total_stat_read != $read_num));

		my $lane_read_ref = $lane_results_ref->[$read_num][$lane_ind];

                # Need this or the next check will cause it to be defined.
                next if (!defined($lane_read_ref->{$curr_stat}));

                if (defined($lane_read_ref->{$curr_stat}{total})) {
                    $chip_sum += $lane_read_ref->{$curr_stat}{total};
                }
            }
        }

	$chip_totals_ref->{$curr_stat} = $chip_sum;
    }
}

#------------------------------------------------------------------------------

sub print_header($$$)
{
    my $id = shift;
    my $runFolder = shift;
    my $machineName = shift;

    print "<!--RUN_TIME " . scalar(localtime) . " -->\n";
    print "<!--SOFTWARE_VERSION $id-->\n";

    print "<html>\n<body>\n\n";
    print "<a name=\"Top\"><h2><title>$runFolder Summary</title></h2></a>\n";
    print "<h1>Summary Information For Experiment $runFolder "
	. "on Machine $machineName</h1>\n";
}

#------------------------------------------------------------------------------

sub print_chip_summary($$$)
{
    my $machineName = shift;
    my $runFolder = shift;
    my $chipName = shift;

    print "<h2><br></br>Chip Summary<br></br></h2>\n";

    print "<table border=\"1\" cellpadding=\"5\">\n";
    printf("<tr><td>Machine</td><td>%s</td></tr>\n", uc($machineName));
    printf("<tr><td>Run Folder</td><td>%s</td></tr>\n", $runFolder);
    printf("<tr><td>Chip ID</td><td>%s</td></tr>\n", $chipName); # TBD
    print "</table>\n";
}

#------------------------------------------------------------------------------

sub print_chip_results_summary($)
{
    my $chip_totals_ref = shift;

    print "<h2><br></br>Chip Results Summary<br></br></h2>\n";

    print "<table border=\"1\" cellpadding=\"5\">\n<tr>\n";
    print "<td>Clusters</td>\n";
    print "<td>Clusters (PF)</td>\n";
    print "<td>Yield (kbases)</td>\n";
    print "</tr>\n";

    print "<tr>";

    for my $stat (qw/clusterCountRaw clusterCountPF yield/) {
	my $val = $chip_totals_ref->{$stat};

	if ($stat eq 'yield') {
	    $val = int ($val / 1000);
	}

	print "<td>$val</td>\n";
    }

    print "</tr>\n";
    print "</table>\n";
}

#------------------------------------------------------------------------------

sub print_lane_parameter_summary($$)
{
    my $laneResultsRef = shift;
    my $maxNumLanes = shift;

    print "<h2><br></br>Lane Parameter Summary<br></br></h2>\n";

    print "<table border=\"1\" cellpadding=\"5\">\n<tr>\n";
    print "<td>Lane</td>\n";
    print "<td>Sample ID</td>\n";
    print "<td>Sample Target</td>\n";
    print "<td>Sample Type</td>\n";
    print "<td>Length</td>\n";
    print "<td>Filter</td>\n";
    print "<td>Chast. Thresh.</td>\n";
    print "<td>Num Tiles</td>\n";
    print "<td>Tiles</td>\n";
    print "</tr>\n";

    my $read_num = 1; # These parameters should be independent of read num. 

    for (my $lane = 1; $lane <= $maxNumLanes; $lane++) {
	my $lane_read_ref = $laneResultsRef->[$read_num][$lane];

	print "<tr>\n";
	print "<td>$lane</td>\n";
	print "<td>$lane_read_ref->{sample}</td>\n";
	print "<td>$lane_read_ref->{template}</td>\n";
	print "<td>$lane_read_ref->{type}</td>\n";
	print "<td>$lane_read_ref->{lengths_list}</td>\n";
	print "<td>$lane_read_ref->{purity}</td>\n";
        if (defined($lane_read_ref->{chastity_thresh})) {
             print "<td>$lane_read_ref->{chastity_thresh}</td>\n";
        } else {
             print "<td>$unknown</td>\n";
        }
	print "<td>$lane_read_ref->{tileCountRaw}</td>\n";
	print "<td><a href=\"#Lane$lane\">Lane $lane</a></td>\n";
	print "</tr>\n";
    }

    print "</table>\n";
}

#------------------------------------------------------------------------------

sub print_lane_results_summary($$$$$$)
{
    my $lane_results_ref = shift;
    my $maxNumLanes = shift;
    my $lane_type_ref = shift;
    my $chip_stats_by_read_ref = shift;
    my $read_num = shift;
    my $lane_num_reads_ref = shift;

    my @int_stats = qw/clusterCountRaw clusterCountPF oneSig/;
    my @float_stats 
        = qw/sig20AsPctOf1 PF_percent pctAlignedPF avAlignScorePF errorPF/;

    my %stat_formats;

    foreach my $stat (@int_stats) {
	$stat_formats{$stat} = '%d';
    }

    foreach my $stat (@float_stats) {
	$stat_formats{$stat} = '%3.2f';
    }

    my @stats = (@int_stats, @float_stats);

    print  "<table border=\"1\" cellpadding=\"5\">\n";

    print "<tr>\n";
    print "<td colspan=\"2\">Lane Info</td>\n";
    print "<td colspan=\"8\">Tile Mean +/- SD for Lane</td>\n";
    print "</tr>\n";

    print "<tr>\n";
    print  "<td>Lane </td>\n";
    print  "<td>Lane Yield (kbases) </td>\n";
    print  "<td>Clusters (raw)</td>\n";
    print  "<td>Clusters (PF) </td>\n";
    print  "<td>1st Cycle Int (PF) </td>\n";
    print  "<td>% intensity after 20 cycles (PF) </td>\n";
    print  "<td>% PF Clusters </td>\n";
    print  "<td>% Align (PF) </td>\n";
    print  "<td>Alignment Score (PF) </td>\n";
    print  "<td> % Error Rate (PF) </td>\n";
    print  "</tr>\n";

    for (my $lane = 1; $lane <= $maxNumLanes; $lane++) {
        # Do not display a lane if this read number is inappropriate for it.
        next if ($lane_num_reads_ref->[$lane] < $read_num);

	if (exists $lane_type_ref->{$lane}) {
            my $lane_read_ref = $lane_results_ref->[$read_num][$lane];

	    print "<tr>\n";
	    print "<td>$lane</td>\n";

	    my $yield_kb = int($lane_read_ref->{yield}{total} / 1000);
	    print "<td>$yield_kb</td>\n";

	    foreach my $thisStat (@stats) {
		if (defined($lane_read_ref->{$thisStat})) {
		    my $form = $stat_formats{$thisStat};

		    # SD undefined here -> single valid tile -> show SD as 0.
		    printf("<td>$form +/- $form</td>\n",
			   $lane_read_ref->{$thisStat}{mean},
			   (defined($lane_read_ref->{$thisStat}{stdev})
			    ? $lane_read_ref->{$thisStat}{stdev} : 0));
		} else {
		    printf ("<td>${unknown}</td>\n");
		}
	    }

	    print  "</tr>\n";
	}
    }

    print "<tr><td colspan=\"13\">Tile mean across chip</td></tr>\n";

    print "<tr>\n";
    print "<td>Av.</td>\n";
    print "<td></td>\n";

    foreach my $thisStat (@stats) {
	if (defined($chip_stats_by_read_ref->
		    [$read_num]{$thisStat}{by_tile})) {
	    my $form = $stat_formats{$thisStat};
	    
	    printf("<td>$form</td>\n",
		   $chip_stats_by_read_ref->[$read_num]{$thisStat}{by_tile});
	} else {
	    printf ("<td>${unknown}</td>\n");
	}
    }

    print  "</tr>\n";
    print  "</table>\n";
}

#------------------------------------------------------------------------------

sub print_expanded_lane_results_summary($$$$$)
{
    my $lane_results_ref = shift;
    my $maxNumLanes = shift;
    my $lane_type_ref = shift;
    my $read_num = shift;
    my $lane_num_reads_ref = shift;

    print  "<table border=\"1\" cellpadding=\"5\">\n<tr>\n\n";

    print "<tr><td colspan=\"2\">Lane Info</td>\n";
    print "<td colspan=\"2\">Phasing Info</td>\n";
    print "<td colspan=\"2\">Raw Data (tile mean)</td>\n";
    print "<td colspan=\"7\">Filtered Data (tile mean)</td></tr>\n";

    print "<td>Lane </td>\n";
    print "<td>Clusters (tile mean) (raw)</td>\n";
    print "<td>% Phasing </td>\n";
    print "<td>% Prephasing </td>\n";
    print "<td>% Error Rate (raw) </td>\n";
    print "<td> Equiv Perfect Clusters (raw) </td>\n";

    print "<td>% retained </td>\n";

    print "<td>Cycle 2-4 Av Int (PF) </td>\n";
    print "<td>Cycle 2-10 Av % Loss (PF) </td>\n";
    print "<td>Cycle 10-20 Av % Loss (PF) </td>\n";

#print  "<td>Av 1-3 Cycle Int </td>\n";
#print  "<td>% intensity after 20 cycles </td>\n";

    print "<td>% Align (PF) </td>\n";
    print "<td>% Error Rate (PF) </td>\n";
    print "<td> Equiv Perfect Clusters (PF) </td>\n";

    print "</tr>\n";

    for (my $lane = 1; $lane <= $maxNumLanes; $lane++) {
        # Do not display a lane if this read number is inappropriate for it.
        next if ($lane_num_reads_ref->[$lane] < $read_num);

	if (exists $lane_type_ref->{$lane}) {
            my $lane_read_ref = $lane_results_ref->[$read_num][$lane];

	    print  "<tr>\n";
	    print "<td>$lane</td>\n";
	    printf("<td>%d</td>\n", $lane_read_ref->{clusterCountRaw}{mean});

	    foreach my $thisStat (qw/phasingapplied prephasingapplied/) {
		if (defined ($lane_read_ref->{$thisStat})) {
		    printf("<td>%6.4f</td>\n",
			   $lane_read_ref->{$thisStat} * 100.);
		} else {
		    print "<td>$unknown</td>\n";
		}
	    }

	    if (defined($lane_read_ref->{errorRaw})) {
		printf("<td>%3.2f</td>\n",
		       $lane_read_ref->{errorRaw}{mean});
	    } else {
		print "<td>$unknown</td>\n";
	    }

	    if (defined($lane_read_ref->{infoContentRaw})) {
		printf("<td>%d</td>\n",
		       $lane_read_ref->{infoContentRaw}{mean});
	    } else {
		print "<td>$unknown</td>\n";
	    }

	    if (defined($lane_read_ref->{PF_percent})) {
		printf("<td>%3.2f</td>\n",
		       $lane_read_ref->{PF_percent}{mean});
	    } else {
		print "<td>$unknown</td>\n";
	    }

	    if (defined($lane_read_ref->{sigAvg2to4})) {
		# SD undefined here -> single valid tile -> show SD as 0.
		printf("<td>%d +/- %d</td>\n",
		       $lane_read_ref->{sigAvg2to4}{mean},
		       (defined($lane_read_ref->{sigAvg2to4}{stdev})
			? $lane_read_ref->{sigAvg2to4}{stdev} : 0));
	    } else {
		print "<td>$unknown</td>\n";
	    }

	    foreach my $thisStat (qw/sigLoss2to10 sigLoss10to20/) {
		if (defined($lane_read_ref->{$thisStat})) {
		    # SD undefined here -> single valid tile -> show SD as 0.
		    printf("<td>%3.2f +/- %3.2f</td>\n",
			   $lane_read_ref->{$thisStat}{mean},
			   (defined($lane_read_ref->{$thisStat}{stdev})
			    ? $lane_read_ref->{$thisStat}{stdev} : 0));
		} else {
		    print "<td>$unknown</td>\n";
		}
	    }

# printf ("<td>%3.2f</td>\n",$laneResultsRef->[$laneInd]{sigAvg1to3}{mean});
# printf ("<td>%3.2f</td>\n",$laneResultsRef->[$laneInd]{sig20AsPctOf1}{mean});

	    foreach my $thisStat (qw/pctAlignedPF errorPF/) {
		if (defined($lane_read_ref->{$thisStat})) {
		    printf("<td>%3.2f</td>\n",
			   $lane_read_ref->{$thisStat}{mean});
		} else {
		    print "<td>$unknown</td>\n";
		}
	    }

	    if (defined($lane_read_ref->{infoContentPF})) {
		printf("<td>%d</td>\n",
		       $lane_read_ref->{infoContentPF}{mean});
	    } else {
		print "<td>$unknown</td>\n";
	    }

	    print  "</tr>\n";
	} # if
    } # for lane

    print "</table>\n";
}

#------------------------------------------------------------------------------

sub print_monotemplate_info($$)
{
    my $lane_num = shift;
    my $gerald_path = shift; 

    my $monotemplate  = "${gerald_path}/s_${lane_num}_monotemplate.txt";

    if (!open(TEMPLATES, "$monotemplate")) {
	print "<p> No template summary available</p>\n";
    } else {  ###### if data exists
	my @templates = ();
		
	print "<h3>Monotemplate Summary</h3>\n";

	print "<table border=\"1\" cellpadding=\"5\">\n\n\n";
	print "<tr>\n";
	print "<td>Lane </td>\n";
	print "<td>Template </td>\n";
	print "<td>Count </td>\n";
	print "<td>Percent </td>\n";
	print "<td>True 1st Cycle Intensity </td>\n";
	print "<td>Av Error Rate </td>\n";
	print "<td> % Perfect </td>\n";
	print "</tr>\n";
		
	my $template_count = 0;
	my $totalTemplates = 0;

	while (<TEMPLATES>) {
	    next if ($_ =~ m/\#/);
	    my @array = split('\t',$_);

	    my $status = 0;
	    ### check if new template
	    if ($template_count == 0) {
		$templates[$template_count]{lane} = $array[0];
		$templates[$template_count]{tag} = $array[2];
		$templates[$template_count]{number} = $array[3];
		$totalTemplates += $array[3];

		if ($array[2] =~ m/^A/) {
		    $templates[$template_count]{intensity} = $array[4];
		} elsif($array[2] =~ m/^C/){
		    $templates[$template_count]{intensity} = $array[5];
		} elsif($array[2] =~ m/^G/){
		    $templates[$template_count]{intensity} = $array[6];
		} elsif($array[2] =~ m/^T/){
		    $templates[$template_count]{intensity} = $array[7];
		}

		$templates[$template_count]{error} 
		= $array[12] + $array[13] + $array[14] 
		    + $array[15] + $array[16];

		$templates[$template_count]{count}++;
		$templates[$template_count]{perfect} = $array[17];
		$template_count++;
	    } else {
		# Not new template

		for (my $i = 0; $i < $template_count; $i++) {
		    if ($templates[$i]{tag} =~ m/$array[2]/) {
			## see this one before
			$templates[$i]{error} = $templates[$i]{error}
			+ $array[12] + $array[13] + $array[14]
			    + $array[15] + $array[16];

			if ($templates[$i]{perfect} > $array[17]) {
			    $templates[$i]{perfect} = $array[17];
			}

			$templates[$i]{count}++;
			$status++;
		    }
		}

		if ($status == 0) {
		    ## new template 
		    $templates[$template_count]{lane} = $array[0];
		    $templates[$template_count]{tag} = $array[2];
		    $templates[$template_count]{number} = $array[3];

		    $totalTemplates += $array[3];

		    if ($array[2] =~ m/^A/) {
			$templates[$template_count]{intensity} = $array[4];
		    } elsif ($array[2] =~ m/^C/) {
			$templates[$template_count]{intensity} = $array[5];
		    } elsif ($array[2] =~ m/^G/) {
			$templates[$template_count]{intensity} = $array[6];
		    } elsif ($array[2] =~ m/^T/) {
			$templates[$template_count]{intensity} = $array[7];
		    }

		    $templates[$template_count]{error} 
		    = $array[12] + $array[13] + $array[14] 
			+ $array[15] + $array[16];

		    $templates[$template_count]{count}++;
		    $templates[$template_count]{perfect} = $array[17];
		    $template_count++;
		} # if
	    } # else
	} # while
		
	my $percentTemplate;

	for (my $i = 0; $i < $template_count; $i++) {
	    $percentTemplate = 0;
	    $percentTemplate
		= 100.0 * $templates[$i]{number} / $totalTemplates
		unless ($totalTemplates == 0);

	    print "<tr>\n";
	    print "<td>$templates[$i]{lane}</td>\n";
	    print "<td>$templates[$i]{tag}</td>\n";
	    print "<td>$templates[$i]{number}</td>\n";
	    printf "<td>%5.2f</td>\n", $percentTemplate;
		    
#		        print "<td>$templates[$i]{intensity}</td>\n";
	    printf "<td>%d</td>\n", $templates[$i]{intensity};

	    $templates[$i]{error}  /= $templates[$i]{count}
	    unless ($templates[$i]{count} == 0);

#		        print "<td>$templates[$i]{error}</td>\n";
	    printf "<td>%5.2f</td>\n", $templates[$i]{error};

	    print "<td>$templates[$i]{perfect}</td>\n";
	    print "</tr>\n";
	}
		
	print  "</table>\n";
    }
    
    print "<b><br></br>Monotemplate IVC plot: </b>\n";
    print "<p> <a href='Monotemplate.htm' target=\"_blank\"> "
	. "Monotemplate.htm\n </a></p>\n";
}

#------------------------------------------------------------------------------

sub print_standard_pair_table($$$$)
{
    my $lane_pair_ref = shift;
    my $display_map_ref = shift;
    my $key3_str = shift;
    my $separ_str = shift;

    print "<table border=\"1\" cellpadding=\"5\">\n\n\n";
    print "<tr>\n";
    print "<th>Read 1 \\ Read 2</th>\n";

    my @stat_labels = map { $_->[1] } @{$display_map_ref};

    foreach my $col_label (@stat_labels, $Total_str) {
	print "<th>$col_label</th>\n";
    }

    print "</tr>\n";

    my @stat_key_strs = map { $_->[0] } @{$display_map_ref};
    my $read1_ind = 0;

    foreach my $read1_key_str (@stat_key_strs, $Total_str) {
        my $key1_str = $Total_str;
        my $row_label = $Total_str;

        if ($read1_key_str ne $Total_str) {
            $key1_str = $Read_str . "1" . $read1_key_str;
            $row_label = $stat_labels[$read1_ind++];
        }

	print "<tr>\n";
	print "<th>$row_label</th>";

	foreach my $read2_key_str (@stat_key_strs, $Total_str) {
	    my $key2_str = (($read2_key_str ne $Total_str)
			    ? $Read_str . "2" . $read2_key_str
			    : $Total_str);

            my @key_strs = ($key1_str, $key2_str);

            if ($key3_str) {
                push(@key_strs, $key3_str);
            }

            my $key_str = join($separ_str, @key_strs);
	    my $val = $lane_pair_ref->{$key_str};
            my $pc_val = $lane_pair_ref->{join($separ_str,
                                               ($pc_prefix, $key_str))};

	    $val = 0 unless defined($val);
	    $pc_val = 0 unless defined($pc_val);

	    printf("<td align=\"right\">%d (%0.1f%%)</td>",
		   $val, $pc_val);
	}

	print "</tr>\n";
    }

    print "</table>\n";
}

#------------------------------------------------------------------------------

sub tabulate_alignment_outcomes($$)
{
    my $lane_pair_ref = shift;
    my $separ_str = shift;

    print "<h3>Individual Alignments</h3>\n";

    # Suppress Repeat Masked row and column by filtering the display map
    # unless either of them have non-zero totals.
    my $RM_found = 0;

    for (my $read_num = 1; $read_num <= $num_paired_reads; ++$read_num) {
	my @key_parts = ($Read_str . $read_num . $Repeat_Masked_key,
			 $Total_str);
	if ($read_num > 1) {
	    @key_parts = reverse(@key_parts);
	}

	if ($lane_pair_ref->{join($separ_str, @key_parts)} > 0) {
	    $RM_found = 1;
	}
    }

    my @display_map = grep { (($_->[0] ne $Repeat_Masked_key) 
			      || $RM_found) } @align_outcome_display_map;


    # Print the table.

    print_standard_pair_table($lane_pair_ref, \@display_map, '', $separ_str);
}

#------------------------------------------------------------------------------

sub tabulate_rel_orient_stats($$)
{
    my $lane_pair_ref = shift;
    my $nominal_orient_str_ref = shift;

    print "<h3>Relative Orientation Statistics</h3>";

    print "<table border=\"1\" cellpadding=\"5\">\n\n\n";
    print "<tr>\n";

    my $nominal_orient_str 
        = $lane_pair_ref->{"$Orientation_str:$nominal_orient_key"};
    $$nominal_orient_str_ref = $nominal_orient_str;
    my @orient_labels = map { $_->[1] } @orient_display_map;

    foreach my $label (@orient_labels, $Total_str) {
	print "<th>$label</th>";
    }

    print "</tr>\n";
    print "<tr>\n";

    foreach my $stat_str ((map { $_->[0] } @orient_display_map), $Total_str) {
	my $val = $lane_pair_ref->{"$Orientation_str:$stat_str"};

        if ($stat_str ne $Total_str) {
            my $pc_val 
                = $lane_pair_ref->{"$pc_prefix:$Orientation_str:$stat_str"};
            $val = sprintf("%d (%0.1f%%)", $val, $pc_val);

            if ($stat_str eq $nominal_orient_str) {
                $val = "<b>$val</b>";
            }
        }

	print "<td align=\"right\">$val</td>";
    }

    print "</tr>\n";
    print "</table>\n";
}

#------------------------------------------------------------------------------

sub tabulate_insert_size_stats($$)
{
    my $lane_pair_ref = shift;
    my $nominal_orient_str = shift;

    my $orient_str = $orient_short_display_map{$nominal_orient_str};
    print("<h3>Insert Size Statistics "
          . "(for relative orientation $orient_str)</h3>");

    print "<table border=\"1\" cellpadding=\"5\">\n\n\n";
    print "<tr>\n";

    my @insert_stat_labels = map { $_->[1] } @insert_stat_display_map;

    foreach my $label (@insert_stat_labels) {
	print "<th>$label</th>";
    }

    print "</tr>\n";
    print "<tr>\n";

    my @insert_stat_strs = map { $_->[0] } @insert_stat_display_map;

    foreach my $stat_str (@insert_stat_strs) {
        my $val = $lane_pair_ref->{"$Insert_Size_str:$stat_str"};
	print "<td align=\"right\">$val</td>";
    }

    print "</tr>\n";
    print "</table>\n";
}

#------------------------------------------------------------------------------

sub tabulate_insert_freq_stats($)
{
    my $lane_pair_ref = shift;

    print("<h3>Insert Statistics (% of individually uniquely alignable pairs)</h3>");

    print "<table border=\"1\" cellpadding=\"5\">\n\n\n";
    print "<tr>\n";

    my @insert_freq_keys = map { $_->[0] } @inserts_display_map;
    my @insert_freq_labels = map { $_->[1] } @inserts_display_map;

    foreach my $label (@insert_freq_labels) {
	print "<th>$label</th>";
    }

    print "</tr>\n";
    print "<tr>\n";

    foreach my $key (@insert_freq_keys) {
	my $val = $lane_pair_ref->{"$Insert_Size_str:$key"};
	my $pc_key = $insert_num_pc_key_map{$key};
	my $pc_val = $lane_pair_ref->{"$Insert_Size_str:$pc_key"};

	my $disp_str = sprintf("%d (%0.1f%%)", $val, $pc_val);

	if ($key eq $num_pairs_ok_key) {
	    $disp_str = "<b>$disp_str</b>";
        }

	print "<td align=\"right\">$disp_str</td>";
    }

    print "</tr>\n";
    print "</table>\n";
}

#------------------------------------------------------------------------------

sub print_pair_results($$)
{
    my $pair_results_ref = shift;
    my $lane_num = shift;

    my $lane_pair_ref = $pair_results_ref->{$lane_num};

    print "<h2>Lane $lane_num : Pair Summary</h2>\n";

    if (!defined($lane_pair_ref)) {
	print "<p>No pair summary available</p>\n";
	return;
    }

    my $nominal_orient_str = '';

    tabulate_rel_orient_stats($lane_pair_ref, \$nominal_orient_str);

    tabulate_insert_size_stats($lane_pair_ref, $nominal_orient_str);
    tabulate_insert_freq_stats($lane_pair_ref);
}

#------------------------------------------------------------------------------

sub print_tile_results_by_lane($$$$$$$$)
{
    my $lane_results_ref = shift;
    my $tile_results_ref = shift;
    my $maxNumLanes = shift;
    my $lane_type_ref = shift;
    my $numTiles = shift;
    my $lane_num_reads_ref = shift;
    my $gerald_path = shift;
    my $pair_results_ref = shift;

    my $aligned = 0;

    for (my $lane_num = 1; $lane_num <= $maxNumLanes; ++$lane_num) {
	if (exists $lane_type_ref->{$lane_num}) {
	    my $analysis_type = $lane_type_ref->{$lane_num};

	    # For paired reads, show the paired results up front.
	    if ($analysis_type =~ m/ELAND_PAIR/) {
		print_pair_results($pair_results_ref, $lane_num);
	    }

            my $num_reads = $lane_num_reads_ref->[$lane_num];

            for (my $read_num = 1; $read_num <= $num_reads; ++$read_num) {
		my $name_str = (($read_num > 1) 
				? "Lane${lane_num}Read${read_num}" 
				: "Lane${lane_num}");
                my $read_str = ($num_reads > 1) ? " : Read $read_num" : '';
                print "<a name=\"${name_str}\"><h2><br></br>"
		. "Lane ${lane_num}${read_str}<br></br></h2></a>\n";

                my $lane_read_ref = $lane_results_ref->[$read_num][$lane_num];
                $aligned = $lane_read_ref->{aligned};

                print "<table border=\"1\" cellpadding=\"5\">\n<tr>\n";
                print "<td>Lane </td>\n";
                print "<td>Tile </td>\n";
                print "<td>Clusters (raw)</td>\n";
                print "<td>Av 1st Cycle Int (PF) </td>\n";
                print "<td>Av % intensity after 20 cycles (PF) </td>\n";
                print "<td>% PF Clusters </td>\n";
                print "<td>% Align (PF) </td>\n";
                print "<td>Av Alignment Score (PF) </td>\n";
                print "<td>% Error Rate (PF) </td>\n";
                print "</tr>\n";

                for (my $tileInd = 0; $tileInd < $numTiles; $tileInd++) {
                    my $tile_read_ref
                        = $tile_results_ref->[$read_num][$tileInd];

		    next if (!defined($tile_read_ref->{lane}));
                    next if ($tile_read_ref->{lane} != $lane_num);
                    print "<tr>\n";

		    my $open = "<td>";
		    my $close = "</td>";

                    print "${open}$tile_read_ref->{lane}${close}\n";
                    print "${open}$tile_read_ref->{tile}${close}\n";

		    my $percent_PF = $tile_read_ref->{PF_percent};

		    if ($percent_PF == 0) {
			$open .= '[';
			$close = ']' . $close;
		    }

                    print "${open}$tile_read_ref->{clusterCountRaw}${close}\n";
                    printf("${open}%3.2f${close}\n", $tile_read_ref->{oneSig});

		    if (defined($tile_read_ref->{sig20AsPctOf1})) {
			printf("${open}%3.2f${close}\n",
			       $tile_read_ref->{sig20AsPctOf1});
		    } else {
			printf("${open}${not_avail_str}${close}\n");
		    }

                    printf("${open}%3.2f${close}\n", $percent_PF);

                    if ($aligned) {
                        printf("${open}%4.1f${close}\n",
                               $tile_read_ref->{pctAlignedPF});
                        printf("${open}%4.1f${close}\n",
                               $tile_read_ref->{avAlignScorePF});
                        print "${open}$tile_read_ref->{errorPF}${close}\n";
                    } else {
                        print "${open}$unknown${close}\n";
                        print "${open}$unknown${close}\n";
                        print "${open}$unknown${close}\n";
                    }

                    print "</tr>\n";
                }

                print "</table>\n";
            } # for read_num

	    # Some analysis modes have extra output.

	    if ($analysis_type =~ m/DEFAULT/) {
		my $cov_graph = "s_". $lane_num . "_cov.png";
	    
		print "<b><br></br>Coverage plot: </b>\n";
		print "<p> <a href='Plots/${cov_graph}' target=\"_blank\"> "
		    . "$cov_graph\n </a></p>\n";
		print "<td><a href=\"#Top\">Back to top</a></td>\n";
	    } elsif ($analysis_type =~ m/MONOTEMPLATE/) {
		print_monotemplate_info($lane_num, $gerald_path);
	    }

	    if (!($analysis_type =~ m/NONE/)) {
		print "<td><a href=\"#Top\">Back to top</a></td>\n";
	    }
	} # if exists ($lane_type...
    } # for lane_num
}

#------------------------------------------------------------------------------

### Check the input parameters

if (@ARGV != 1) {
    print STDERR "$0 Gerald_dir\n";
    die;
} #  if @ARGV

my $appName = basename($0);
my $geraldPath = $ARGV[0];

# Following line enables you to do e.g. "jerboa.pl ."
$geraldPath = File::Spec->rel2abs($geraldPath);

## Check to see if directory name ends with '/'
if ($geraldPath !~ m/\/$/) {
    $geraldPath = $geraldPath . "/";
}

my ($runFolder, $machineName, $bustard_dir);
extract_run_info($appName, $bustardPath, \$runFolder, \$machineName,
		 \$bustard_dir);

my $param_is_legacy = param_file_is_legacy_style($bustard_dir);
my $bustard_params_file = "${firecrestPath}/.params";
my @read_start_cycles = ();

my %phasing_options;

if (!$param_is_legacy) {
    # FIXME : Multiple reads of $bustard_params_file.
    if (!getPhasingOptions($bustard_params_file, $bustard_dir,
			   \%phasing_options)) {
	warn "Failed to get phasing options\n";
	exit(1);
    }

    if (!getReadStartCycles($bustard_params_file, $bustard_dir,
			    \@read_start_cycles)) {
	warn "Failed to extract read start cycles from $bustard_params_file\n";
	exit(1);
    }
} else {
    # legacy
    @read_start_cycles = (1);
}

my $num_available_reads = @read_start_cycles;

my @laneResults;
my @lane_num_reads;
my @lane_num_tiles;

init_lane_results(\@laneResults, \@lane_num_reads, \@lane_num_tiles,
                  $maxNumLanes, $num_available_reads);

store_measured_phasings(\@laneResults, $bustardPath,
			$maxNumLanes, $num_available_reads,
			\%phasing_options, $param_is_legacy);

store_applied_phasings($appName, \@laneResults,
		       $maxNumLanes, $num_available_reads,
		       $bustardPath, $bustard_params_file,
		       \%phasing_options, $param_is_legacy);

my $sampleFile = "${runFolderPath}/samples.xml";
read_sample_file($appName, $sampleFile, \@laneResults, $num_available_reads);

# Look for reference, filter, USE_BASES in config.xml file.
my $configFile = File::Spec->catdir($geraldPath, "config.xml");
my $max_reads_used = 1;
my %lane_cycle_map;

read_config_file($configFile, \@laneResults, $maxNumLanes,
                 $num_available_reads, \@lane_num_reads, \$max_reads_used,
                 \%lane_cycle_map);

my $exptdir      = $bustardPath;
my @dirs         = File::Spec->splitdir($exptdir);
my $bustarddir   = $dirs[$#dirs];
my $firecrestdir = File::Spec->catdir(@dirs[0 .. ($#dirs - 1)]);
my $paramsfile   = File::Spec->catfile($firecrestdir, ".params");
my $chastityThreshold = Gerald::Common::getChastityThreshold($paramsfile, $bustarddir);

@{$laneResults["chastity_thresh"]} = $chastityThreshold if $chastityThreshold;;

my $chipFile = "${runFolderPath}/chip.xml";
my $chipName = $unknown;
read_chip_file($appName, $chipFile, \$chipName);

my @tiles = ();
my $tilesFile = File::Spec->catdir($geraldPath, "tiles.txt");
print STDERR $appName, ": reading tile list from ${tilesFile}\n";
read_tiles_file($tilesFile, \@tiles);

if (@tiles == 0)
{
    warn "No tilenames found in `$tilesFile'\n";
    exit(1);
}

# Ensure sensible (or at least consistent) file ordering
@tiles = sort(@tiles);

my @results;
my %lane_type;

my $tileInd = 0;
my $cycle = 0;
my $raw_cluster_count = 0;
my $PF_cluster_count = 0;

my %undef_hash;

foreach my $tile_name (@tiles) {

    ## Determine Lane and Tile

    my @array = split("_", $tile_name);
    my $curr_tile_lane = $array[1];
    my $tile_id = $array[2];
    my $num_reads = $lane_num_reads[$curr_tile_lane];

    ++$lane_num_tiles[$curr_tile_lane];

    # If there is more than one read, there should also be `All Reads' -> 0.
    my $lo_read_num = (($num_reads > 1) ? 0 : 1);

    # Initialise tile values for the lane-specific number of reads.
    for (my $read_num = $lo_read_num; $read_num <= $num_reads; ++$read_num) {
	# Force $results[$read_num][$tileInd] to exist before referencing it.
	%{$results[$read_num][$tileInd]} = %undef_hash;
        my $tile_read_ref = $results[$read_num][$tileInd];

        $tile_read_ref->{lane} = $curr_tile_lane;
        $tile_read_ref->{tile} = $tile_id;

	my $lane_read_ref = $laneResults[$read_num][$curr_tile_lane];

        $lane_read_ref->{tileCountRaw}++;

        # Set some defaults - mostly for Eland.

        $tile_read_ref->{template} = $lane_read_ref->{template};
        $tile_read_ref->{length} = $lane_read_ref->{length};
        $tile_read_ref->{type} = $lane_read_ref->{type};
        $tile_read_ref->{clusterCountRaw} = 0;
        $tile_read_ref->{errorPF} = 0;

	# Set lane type from single reads (i.e. read_num > 0) if available.
	if (($read_num > 0) && defined($tile_read_ref->{type})) {
	    if ($tile_read_ref->{type} ne "NONE") {
		$lane_type{$curr_tile_lane} = $tile_read_ref->{type};
	    }
	}
    }

    # Process _all.txt file for current tile.
    process_all_file("${signal_means_path}/${tile_name}${allSuffix}",
		     $tileInd, $curr_tile_lane, 
		     \@laneResults, \@results,
		     \%lane_cycle_map, $num_reads);

    my $tile_ref_read_ref = $results[$ref_read_num][$tileInd];
    $raw_cluster_count = (defined($tile_ref_read_ref->{clusterCountRaw})
                          ? $tile_ref_read_ref->{clusterCountRaw}
                          : 0);
    $PF_cluster_count = (defined($tile_ref_read_ref->{clusterCountPF})
                         ? $tile_ref_read_ref->{clusterCountPF}
                         : 0);

    # Process _rescore.txt and _score.txt file for current tile.
    process_score_file("${geraldPath}/${tile_name}${rescoreSuffix}",
                       'infoContentPF', 'errorPF',
                       'pctAlignedPF', 'avAlignScorePF',
                       $tileInd, $curr_tile_lane, $num_reads,
                       \@results, \@laneResults, $PF_cluster_count);

    process_score_file("${geraldPath}/${tile_name}${scoreSuffix}",
                       'infoContentRaw', 'errorRaw',
                       'pctAlignedRaw', 'avAlignScoreRaw',
                       $tileInd, $curr_tile_lane, $num_reads,
                       \@results, \@laneResults, $raw_cluster_count);

    $tileInd++;
}

my $numTiles = $tileInd;

my %pair_results;

# Eland Pair analysis produces an extra XML file.
for (my $lane_num = 1; $lane_num <= $maxNumLanes; ++$lane_num) {
    my $analysis_type = $laneResults[$ref_read_num][$lane_num]{type};

    if (($lane_num_tiles[$lane_num] == 0)
        && ($analysis_type ne "NONE")) {
        warn("Analysis $analysis_type but no tiles specified for "
             . "lane $lane_num");
        next;
    }

    if ($analysis_type eq "ELAND_PAIR") {
        process_pair_info_file(\%pair_results, $lane_num);
        derive_pair_stats(\%{$pair_results{$lane_num}});
    }
}


### Outputting the Lane Information

# Calculate the Lane Results.
calc_lane_results(\@laneResults, \@results, $maxNumLanes, \@lane_num_reads,
                  \%lane_type, $numTiles);

# Calculate weighted averages of statistics across the chip
# (by relevant tiles and by relevant clusters) and totals where relevant.
my @chip_stats_by_read = ();
my %chip_totals = ();
calc_chip_results(\@laneResults, $maxNumLanes,
                  \@lane_num_reads, $max_reads_used,
                  \@chip_stats_by_read, \%chip_totals);

# Now ready to start printing the page

# Print some basic information.
print_header($id, $runFolder, $machineName);

# Print the chip parameter summary.
print_chip_summary($machineName, $runFolder, $chipName);

# Print the chip results summary.
print_chip_results_summary(\%chip_totals);

# Print the lane parameter summary.
print_lane_parameter_summary(\@laneResults, $maxNumLanes);

# Print a lane summary table for each read used by at least one lane.

for (my $read_num = 1; $read_num <= $max_reads_used; ++$read_num) {
    my $read_str = ($max_reads_used > 1) ? " : Read $read_num" : '';
    print "<h2><br></br>Lane Results Summary${read_str}<br></br></h2>\n";

    print_lane_results_summary(\@laneResults, $maxNumLanes, \%lane_type,
                               \@chip_stats_by_read, $read_num,
                               \@lane_num_reads);
}

# Print an expanded lane summary table for each read used by at least one lane.

for (my $read_num = 1; $read_num <= $max_reads_used; ++$read_num) {
    my $read_str = ($max_reads_used > 1) ? " : Read $read_num" : '';
    print  "<h2><br></br>Expanded Lane Summary$read_str<br></br></h2>\n";

    print_expanded_lane_results_summary(\@laneResults, $maxNumLanes,
                                        \%lane_type, $read_num,
                                        \@lane_num_reads);
}

# Print links to image index pages.

print  "<b><br></br>IVC Plots</b>\n";
print  "<p> <a href='IVC.htm' target=\"_blank\"> IVC.htm\n </a></p>\n";

print  "<b><br></br>All Intensity Plots</b>\n";
print  "<p> <a href='All.htm' target=\"_blank\"> All.htm\n </a></p>\n";

print  "<b><br></br>Error graphs: </b>\n";
print  "<p> <a href='Error.htm' target=\"_blank\"> Error.htm\n </a></p>\n";

print  "<b><br></br>Error Curves: </b>\n";
print  "<p> <a href='Perfect.htm' target=\"_blank\"> Perfect.htm\n </a></p>\n";

print  "<td><a href=\"#Top\">Back to top</a></td>\n";

# Print the tile results tables.
print_tile_results_by_lane(\@laneResults, \@results, $maxNumLanes,
                           \%lane_type, $numTiles, \@lane_num_reads,
			   $geraldPath, \%pair_results);

# Final entry in HTML file
print  "</body>\n</html>\n";

#------------------------------------------------------------------------------
