#!/usr/bin/env perl

# PROJECT: GERALD
# MODULE:  pickBestPair.pl
# AUTHOR:  A. J. Cox
#
# Copyright (c) 2007 Solexa
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.


use warnings;
use strict;
#use File::Basename;

use FindBin;
use lib "$FindBin::Bin/../lib";
use Gerald::Common;
use POSIX;
use XML::Simple;
use Data::Dumper;

my $version;
my $id;
{
    my $cvs_id='@(#) $Id: pickBestPair.pl,v 1.2 2008/07/08 15:12:10 craczy Exp $';
    my $cvs_version='$Revision: 1.2 $';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope

# parameters to control program behaviour - passed in via command line
# as --param-name=value
my %controlParams =
(

 'min-percent-unique-pairs'=>10,
 # min percent of clusters for which each half can be uniquely aligned
 # without recourse to pairing information. Such clusters are used to estimate
 # the distribution of insert sizes. NB the percentage of such clusters
 # reflects not only data quality but also the repetitiveness of the alignment
 # target

 'min-percent-consistent-pairs'=>70,
 # min percent of the uniquely aligning clusters that have the same orientation
 # We would expect this to be close to 100%, if not a problem with the 
 # sample prep is indicated

 'add-shadow-to-singleton-threshold'=>1000000,
 # If the aligned half of an 'singleton' read has a single read alignment 
 # score of at least 'min-single-read-alignment-score' AND the base quality
 # of the non-aligned 'shadow' half of the read exceeds this, then
 # output a zero-score alignment of the shadow - intention of this feature is 
 # to facilitate 
 # localized reassembly at a later date. Default value of 1000000 effectively
 # switches this feature off by default

 'min-paired-read-alignment-score'=>0,
 # If a read-pair has a paired alignment score of at least this, output a 
 # alignment for each half of the pair

 'min-single-read-alignment-score'=>0,
 # If a read is not part of a paired alignment but has a single read
 # score of at least this, output an alignment for that read

 
 'circular'=>'N'
 # If '--circular' is specified, assume all alignment targets are circular
 # genomes for the purposes of assigning pairs. 
); # %controlParams

my $usage=<<"END";
Usage: $0 
[ --min-percent-unique-pairs=value ] 
[ --min-percent-consistent-pairs=value ]
[ --add-shadow-to-singleton-threshold=value ]
[ --min-single-read-alignment-score=value ]
[ --min-paired-read-alignment-score=value ]
[ --circular[=genome1Name[:size1][,genome2Name[:size2][,...] ] ] ]
ELAND_read_length1 ELAND_read_length2 - read lengths for read 1 & 2
read1_eland_extended.txt read2_eland_extended.txt - alignments for reads 1 & 2
read1_qv.txt read2_qv.txt - quality values for reads 1 & 2
filt.txt - filter mask (Y=read passed filter, N=failed)
read1_saf.txt read2_saf.txt (best alignments for reads 1 & 2)
anomaly.txt (reads with anomalous alignments)
[ genomesize.xml ] - genome size details, needed iff '--circular' is set
END



#while ($ARGV[0]=~/^\-\-(.*)=(.*)$/)
while($ARGV[0]=~/^\-\-(.*?)(=(.*))?$/)
{
    my $param=$1;
    my $value=$3;
    $value='Y' unless (defined($value));
    unless (defined($controlParams{$param}))
    {
	die "${usage}Error: Invalid parameter $param specified!\n";
    } # unless
    else
    {
	print STDERR "$0: setting $param to $value\n";
	$controlParams{$param}=$value;
    } # else
    shift(@ARGV);
} # while


die $usage unless ((@ARGV==10)||(@ARGV==11));

#my $readLengthELAND1=shift(@ARGV); 
#my $readLengthELAND2=shift(@ARGV);

my ($readLengthELAND1,$readLengthELAND2,
    $elandExtendedFile1,$elandExtendedFile2,
    $qualityValueFile1,$qualityValueFile2,
    $filtFile,
    $safFile1,$safFile2,
    $anomalyFile,
    $genomeSizeFile)=@ARGV;

my ($l1, $l2, $filt);
#my @s1;
#my @s2;
my @m1;my @m2;
my ($match1,$chrom1,$pos1,$strand1,$type1,$bestMatch1,$bestChrom1,$bestScore1);
my ($match2,$chrom2,$pos2,$strand2,$type2,$bestMatch2,$bestChrom2,$bestScore2);

my ($matchSingle1,$chromSingle1,$posSingle1,$strandSingle1,$typeSingle1);
my ($matchSingle2,$chromSingle2,$posSingle2,$strandSingle2,$typeSingle2);


my ($qv1, $readID1, $read1, $nhood1, $details1);
my ($qv2, $readID2, $read2, $nhood2, $details2);

# storage array for candidate read pairs
my @pairsFound;

# hash for read pair information (insert size etc)
my %pairProperties;

# by default, do a paired alignment. If pairing is too poor, switch off
my $doPairedAlignment=1;

my ($scoreSingle1, $scoreSingle2);
#my ($nhood1, $nhood2);
my $numClusters=0;
my $numFilteredClusters=0;
my $numPairsTotal=0; 
#my $numPairsGood=0;
my $pairsSoFar=0;
my $bestInsert1=0; 
my $bestInsert2=0;

my $rl1=undef;
my $rl2=undef;
my $insertLength;
my %insertHist;
#my $median=undef;
#my $insertLow=undef;
#my $insertHigh=undef;



# This array lists the four possible types of orientation we might see.
# Notation is as follows:
# Suppose read 1 hits to genome on forward strand, where and with what
# orientation does read 2 hit to genome?
# first char = orientation of second read
# second char: p ('plus') if second read pos > first read pos, else m ('minus')
# All preps seen so far are type 'Rp'
my @orientation=
(
 'Rp', # reads point towards each other on genome - all preps like this so far 
 'Fp', # read 1 points toward read 2 on genome
 'Fm', # read 2 points toward read 1 on genome
 'Rm'  # reads point away from each other on genome
);

# This hash maps the various possible orienations of the two reads 
# to one of the four underlying orientation types above

# first char = orientation of first read
# second char = orientation of second read
# third char: p ('plus') if second read pos > first read pos, else m ('minus')
my %orientation=
(
 'FRp'=>'Rp',
 'RFm'=>'Rp',
 'FFp'=>'Fp',
 'RRm'=>'Fp',
 'FFm'=>'Fm',
 'RRp'=>'Fm',
 'RFp'=>'Rm',
 'FRm'=>'Rm');


# This hash maps each orientation type to the orientation type it 'looks like'
# if the alignments 'wrap around' on a circular genome, eg an Rp-type paired
# read with an insert that crosses the boundary will give an Rm-type alignment
# with an abnormal insert size. Confused? Drink some more coffee.

my %antiOrientation=
(
 'Rp'=>'Rm',
 'Fp'=>'Fm',
 'Rm'=>'Rp',
 'Fm'=>'Fp'
 );


my $orientation;
#my $nominalOrientation;

# The sizes of any alignment targets that are to be treated as circular
# are stored here - reference to hash
my $circular=undef;


my $noAlignmentDoneScore=-1;

#
# --- Subroutine Definitions ---
#

# getInsertLength
sub getInsertLength($$$$$$)
{
    my ( $pos1, $strand1, $rl1, $pos2, $strand2, $rl2 ) = @_;  
#    die unless (@_==6);


#    if ($strand1 eq 'F')
    if ($pos1<$pos2) 
    {
	$insertLength=$pos2-$pos1+$rl2;
    } # if
    else
    {
	$insertLength=$pos1-$pos2+$rl1;
    } # else
} # sub getInsertLength

# Classify the orientation of the read pair
sub getOrientation($$$$)
{
    # NB This may not be the optimal way to do this when the insert length
    # is shorter than the read length, however such reads sequence into the
    # adapter anyway so need handling with caution
    # Tony C 9.10.7

    my ( $pos1, $strand1, $pos2, $strand2 ) = @_;
#    die unless (@_==4);
    my $q=$strand1.$strand2.(($pos2>$pos1)?'p':'m');
#    print "$pos1 $strand1 $pos2 $strand2 $q $orientation{$q}\n";
    return $orientation{$q};
} # sub getOrientation

sub estimateInsertSizeDist
{

# Estimating insert size distribution
# Get median not mean - less sensitive to dud inserts from random alignments
# Assume insert sizes are normally distributed near the median
# 34.13% of a normal distribution is within 1 SD each side of mean
# Estimate SD empirically each side of median
# Chuck away anything not within 3 SDs of median

    my $sdLow=0.5-0.3413;
    my $sdHigh=0.5+0.3413;
    my $lastInsert;
    my $fractionSoFar;
    my $numOutliersSmall=0;
    my $numOutliersLarge=0;

    my $numPairsGood=$pairProperties{Orientation}
    {$pairProperties{Orientation}{Nominal}};

    for $insertLength ( sort {$a<=>$b} 
			keys %{$insertHist
			       {$pairProperties{Orientation}{Nominal}}} )
    {
#	$pairsSoFar += $insertHist{$insertLength};
	$pairsSoFar += $insertHist{
	    $pairProperties{Orientation}{Nominal}}{$insertLength};
	$fractionSoFar=$pairsSoFar/$numPairsGood; # TBD check for numPairs!=0
	if (($fractionSoFar>=$sdLow)
	    &&(!defined($pairProperties{InsertSize}{LowSD})))
	{
	    $pairProperties{InsertSize}{LowSD}=$lastInsert;
	} # if
	elsif (($fractionSoFar>0.5)
	       &&(!defined($pairProperties{InsertSize}{Median})))
	{
	    $pairProperties{InsertSize}{Median}=$lastInsert;
#	    $median=$lastInsert;
	} # elsif
	elsif (($fractionSoFar>=$sdHigh)
	       &&(!defined($pairProperties{InsertSize}{HighSD})))
	{
	    $pairProperties{InsertSize}{HighSD}=$lastInsert;
#	    $insertHigh=$lastInsert;
	} # elsif
	
#	print "$insertLength $insertHist{$insertLength} $pairsSoFar ", 100.0*$pairsSoFar/$numPairs, "\n";
	
	$lastInsert=$insertLength;
    } # for

#    print "$insertLow $median $insertHigh\n";
#    $pairProperties{Median}

    $pairProperties{InsertSize}{LowSD}
    = $pairProperties{InsertSize}{Median}-$pairProperties{InsertSize}{LowSD};
    $pairProperties{InsertSize}{HighSD}
    -= $pairProperties{InsertSize}{Median};

    $pairProperties{InsertSize}{Min}
    =$pairProperties{InsertSize}{Median}
	- (3*$pairProperties{InsertSize}{LowSD});

    $pairProperties{InsertSize}{Max}
    =$pairProperties{InsertSize}{Median}
	+ (3*$pairProperties{InsertSize}{HighSD});


    for $insertLength ( keys %{$insertHist
			       {$pairProperties{Orientation}{Nominal}}} )
    {
	if ($insertLength<$pairProperties{InsertSize}{Min})
	{
	    $numOutliersSmall+=
	    $insertHist{$pairProperties{Orientation}{Nominal}}{$insertLength};
	} # if
	elsif ($insertLength>$pairProperties{InsertSize}{Max})
	{
	    $numOutliersLarge+=
	    $insertHist{$pairProperties{Orientation}{Nominal}}{$insertLength};
	} # elsif
    } # for

    $pairProperties{Orientation}{NominalOrientationButSmallInsert}
    = $numOutliersSmall;

    $pairProperties{Orientation}{NominalOrientationButLargeInsert}
    = $numOutliersLarge;

    $pairProperties{Orientation}{NominalOrientationButSmallInsertPercent}
    = sprintf("%-5.5f",100.0*$numOutliersSmall/$numPairsTotal);

    $pairProperties{Orientation}{NominalOrientationButLargeInsertPercent}
    = sprintf("%-5.5f",100.0*$numOutliersLarge/$numPairsTotal);



#    print STDERR "$0: Pair distribution: lower=$insertLow median=$median upper=$insertHigh\n";
} # sub estimateInsertSizeDist

sub adjustPairedNhood($)
{
#    die unless (@_==1);
    my ($nhood)=@_;
    my @nbors=split(':',$nhood);
    if ($nbors[0]!=0)
    {
	$nhood='1:0:0';
    } # if
    elsif ($nbors[1]!=0)
    {
	$nhood='0:1:0';
    } # elsif
    else
    {
	$nhood='0:0:1';
    } # else
    return $nhood;
} # adjustPairedNhood


# sub getScoreFor3Neighbour
# get best possible score for a 3-neighbour of a read
sub getScoreFor3Neighbour($$$)
{
#    die unless (@_==2);
    my ($qv, $readLengthELAND, $thisScore)=@_;
    my @worstBases=getWorstBases($qv, $readLengthELAND);
#    my $thisScore=getExactScoreForRead($qv);
    for my $adj (@worstBases)
    {
	$thisScore+=$adj;
    } # for
    return $thisScore;
} # sub getScoreFor3Neighbour



# sub bestPairedAlignFromList
# get and score best paired alignment from list
sub bestPairedAlignFromList($$$$$$$)
{
#    die unless (@_==7);
    my ($qv1, $qv2, $readLengthELAND1, $readLengthELAND2,
	$exactScore1, $exactScore2 )=@_; 
    my @pairsFound=@{$_[-1]};
    die unless (@pairsFound>1);
    my ($details1, $details2);
    my ($chrom1, $pos1, $strand1, $type1);
    my ($chrom2, $pos2, $strand2, $type2);

#    my $exactScore1=getExactScoreForRead($qv1);
#    my $exactScore2=getExactScoreForRead($qv2);

    my $thisScore;
    my $bestScore=-1;
    my $bestPair=0;

    my $totalScore=
	getScoreFor3Neighbour($qv1, $readLengthELAND1, $exactScore1)
	+ getScoreFor3Neighbour($qv2, $readLengthELAND2, $exactScore2);
#    print getScoreFor3Neighbour($qv1, $readLengthELAND1), " ",
#    getScoreFor3Neighbour($qv2, $readLengthELAND2), "\n";

    $totalScore=10*exp($totalScore); # Add in 10 3-neighbours
#    print "ts=$totalScore\n";


    for (my $i=0; $i<@pairsFound; $i++)
    {
	($details1, $chrom1, $details2, $chrom2) 
	    = split(' ',$pairsFound[$i]);
	die unless (defined($chrom2));
	($chrom1, $pos1, $strand1, $type1)
	    =getMatchDetails($details1, $chrom1);
	($chrom2, $pos2, $strand2, $type2)
	    =getMatchDetails($details2, $chrom2);
	$thisScore 
	    = getAlignmentScore($qv1,$type1,$exactScore1)
	    + getAlignmentScore($qv2,$type2,$exactScore2);
	$thisScore=exp($thisScore);
	$totalScore+=$thisScore;
	if ($thisScore>$bestScore)
	{
	    $bestScore=$thisScore;
	    $bestPair=$i;
	} # if
    } # for

    if (@pairsFound!=1)
    {
	($details1, $chrom1, $details2, $chrom2) 
	    = split(' ',$pairsFound[$bestPair]);
	die unless (defined($chrom2));
	($chrom1, $pos1, $strand1, $type1)
	    =getMatchDetails($details1, $chrom1);
	($chrom2, $pos2, $strand2, $type2)
	    =getMatchDetails($details2, $chrom2);
    } # if

    if ($bestScore==$totalScore)
    {
	$bestScore=200;
#	print "@pairsFound $bestScore $totalScore\n";
#	die; # should never be here
    } # if
    else
    {
	$bestScore=floor(convertProbCorrectToPhred($bestScore/$totalScore));
    } # else

    return ($chrom1, $pos1, $strand1, $type1, 
	     $chrom2, $pos2, $strand2, $type2,
	    $bestScore );
} # sub bestPairedAlignFromList

# sub parseCircular
# Sort out the genome size information necessary to handle circular
# genomes properly
sub parseCircular
{
    return if ($controlParams{'circular'} eq 'N');

    my @genomeList;
    my %genomeHash;
    my $genome;

    @genomeList=split(',',$controlParams{'circular'})
	unless ($controlParams{'circular'} eq 'Y');


    # If no list of genomes is given then we definitely need to read
    # in XML file of genome sizes
    my $wantDefault=((@genomeList>0)?'N':'Y');

    for my $genome (@genomeList)
    {
#	print $genome, "\n";
	if ($genome=~/(.*?):(.*)?$/)
	{
	    my $param=$1;
	    my $value=$2;
	    
	    if ($value<=0)
	    {
		warn "$0: nonsensical size $value specified for genome $param";
	    } # if
	    else
	    {
#		print "$param $value\n";
		$genomeHash{$param}=$value;
	    } # else
	} # if
	else
	{ # specified a genome but not size, need to get it from defaults
	    $genomeHash{$genome}=-1;
	    $wantDefault='Y';
	} # else
    } # genomeList
	
    if ($wantDefault eq 'Y')
    {
	# read in XML file
	unless (defined($genomeSizeFile))
	{
	    print STDERR 
		"Warning: --circular specified but no genome size file given!\n";
	    print STDERR 
		"Circularization will not be done\n";
	    return;
	} # unless
	unless (-e $genomeSizeFile )
	{
	    print STDERR 
		"Warning: genome size file $genomeSizeFile does not exist!\n";
	    print STDERR 
		"Circularization will not be done\n";
	    return;
	} # unless

	my $xml = new XML::Simple();
	$circular=XMLin($genomeSizeFile);

    } # if wantDefault

    # if a list of genomes to circularize is specified, only keep those
    if (@genomeList>0)
    {
#	print "fred\n";
	for $genome (keys %{$circular})
	{
#	    print "$genome\n";
	    unless (defined($genomeHash{$genome}))
	    {
#		    delete($circular->{$genome});
		$circular->{$genome}=undef;
	    } # unless
	    elsif  ($genomeHash{$genome}!=-1)
	    { # if user has specified a genome size then use it
		$circular->{$genome}=$genomeHash{$genome};
	    } # elsif
	} # for
	for $genome (keys %genomeHash)
	{
	    if ($genomeHash{$genome}==-1)
	    {
		unless (defined($circular->{$genome}))
		{
		    print STDERR "Warning: Circularization specified for genome $genome, but genome size neither specified nor found in defaults\n";
		} # unless
		else
		{
		    print STDERR "Circularization specified for genome $genome, using default genome size of $circular->{$genome}\n";
		} # unless
	    } # if
	    else
	    {
		print STDERR "Circularization specified for genome $genome, using genome size of $genomeHash{$genome}\n";
		$circular->{$genome}=$genomeHash{$genome}; 
	    } # else
	} # for
    } # if (@genomeList>0)

#    print Dumper($circular),"\n";    
 
#    die;

} # sub parseCircular

# sub printAnomalyLine
# Print details of anomalous reads
sub printAnomalyLine
{
    die unless ((@_==9)||(@_==1));
    my ($readID, $read1, $read2, $qv1, $qv2, 
	$nhood1, $details1, $nhood2, $details2)=@_;

    # use read 1 identifier but trim off '/1' at end
#    $readID=~s/\-1$//;
    $readID=~s/\/1$//;

    unless (defined($details2))
    {
	print ANOM "$readID\tOK\n";
    } # unless
    else
    {
	print ANOM "$readID\t$read1\t$read2\t$qv1\t$qv2";
	if ($details1 ne '-')
	{
	    print ANOM "\t$details1";
	} # if
	else
	{
	    print ANOM "\t$nhood1";
	} # else

	if ($details2 ne '-')
	{
	    print ANOM "\t$details2";
	} # if
	else
	{
	    print ANOM "\t$nhood2";
	} # else
	print ANOM "\n";
    } # else   
} # sub printAnomalyLine

# sub printAlignmentLine
# Print a line of alignment details
sub printAlignmentLine
{
    die unless ((@_==10)||(@_==3));
#    my ($fh, $readID, $chrom, $pos, $strand, $type, 
#	$scoreSingle, $scorePair, $strandOfPair, $offsetOfPair)=@_;
    my $fh=shift;

    if ((@_==9)
	&&($_[2]<=0)
	&&(defined($circular))
	&&(defined($circular->{$_[1]})))
    {
	$_[2]+=$circular->{$_[1]};
    } # if

    $"="\t";
    print $fh "@_\n";

#    if (@_==9)
#    { 
#    print $fh "$readID\t$chrom\t$pos\t$strand\t$type\t$scoreSingle\t$scorePair\t$strandOfPair\t$offsetOfPair\n"; 
#	    print BEST1 "$readID1\t$chrom1\t$pos1\t$strand1\t$type1\t$scoreSingle1\t$bestScore1\t$strand2\t$bestInsert1\n";
} # printAlignmentLine


# sub makeKey
# make an XML element name based on match type (unique, repeat, QC or no match)
sub makeKey
{
    die unless(@_==3);
#    my ($pos, $nhood, $read)=@_;
    my ($nhood, $details, $read)=@_;
    my $key;
    if ($details eq '-')
    {
	if ($nhood=~/:/)
	{
	    $key="Read${read}Repeat";
	}
	else
	{
	    $key="Read${read}$nhood";	
	} # else

    }
    elsif ($details =~/,/)
    {
	$key="Read${read}ManyAlignmentsFound";
    }
    else
    {
	$key="Read${read}SingleAlignmentFound";
    }

    return $key;
} # sub makeKey

# Handle output of a singleton (one read aligns, other doesn't) cluster
# Returns number of alignments output
# 2: Good alignment of singleton, shadow has high base quality so a
# zero-score alignment of shadow produced (so its sequence is available)
# 1: Good alignment of singleton, shadow has low base quality
# 0: Singleton alignment is not reliable
sub printSingletonAlignment($$$$$$$$$$$$$)
{
    my ($fhAligned,
	$readIDAligned, $nhoodAligned,
	$chromSingleAligned, $posSingleAligned, $strandSingleAligned,
	$typeSingleAligned, $scoreSingleAligned,
	$exactScoreShadow, $qvShadow,
	$fhShadow,
	$readIDShadow, $nhoodShadow)=@_;
	
    if ($scoreSingleAligned
	>=$controlParams{'min-single-read-alignment-score'})
    { # print alignment if of sufficient quality
	printAlignmentLine
	    ($fhAligned, $readIDAligned, 
	     $chromSingleAligned, $posSingleAligned, $strandSingleAligned,
	     $typeSingleAligned, $scoreSingleAligned, 0, 'N', 0);
	if (getMeanBaseQuality($exactScoreShadow, $qvShadow)
	    >=$controlParams{'add-shadow-to-singleton-threshold'})
	{ #  add null alignment of shadow if quality good enough
	    printAlignmentLine
		($fhShadow, $readIDShadow, $chromSingleAligned, 
		 $posSingleAligned, 
		 $strandSingleAligned, '-', 
		 $noAlignmentDoneScore, 0, 'N', 0);
	    return 2; 
#		    $pairProperties{Reads}{$key1}{$key2}
#		    {SingletonRead2}{AlignmentOKShadowAdded}+=$passedFiltering;
	} # if
	else
	{
	    printAlignmentLine($fhShadow, $readIDShadow, $nhoodShadow);
#		    $pairProperties{Reads}{$key1}{$key2}
#		    {SingletonRead2}{AlignmentOK}+=$passedFiltering;
	    return 1;
	} # else
    } # if
    else
    {
	printAlignmentLine($fhAligned, $readIDAligned, $nhoodAligned);
	printAlignmentLine($fhShadow, $readIDShadow, $nhoodShadow);
	return 0;
#		$pairProperties{Reads}{$key1}{$key2}
#		{SingletonRead2}{AlignmentPoor}+=$passedFiltering;
    } # else
 
} # printSingletonAlignment

# Handle printing of alignments in case where either:
# i. No pairing was done (because there weren't enough clusters having
# unique alignments of each read to enable insert size info to be obtained)
# - in this case the data is effectively treated as two sets of unrelated
# single reads. 
# ii. There were alignments of the individual reads, but none formed a
# 'feasible pairing', i.e. had an appropriate distance and orientation
# with respect to one another.
# return codes:
# 0 - no alignments printed
# 1 - read 1 aligned, read 2 didn't
# 2 - 

sub printUnpairedAlignments($$$$$$$$$$$$$$$$$$$$$$$$)
{
    my ($fh1,
	$readID1, $nhood1,
	$chromSingle1, $posSingle1, $strandSingle1,
	$typeSingle1, $scoreSingle1,
	$exactScore1, $qv1,
	$key1,
	$fh2,
	$readID2, $nhood2,
	$chromSingle2,$posSingle2,$strandSingle2,
	$typeSingle2, $scoreSingle2,
	$exactScore2, $qv2,
	$key2,
	$outcome,$passedFiltering)=@_;


    printAnomalyLine($readID1, $read1, $read2, $qv1, $qv2,
		     $nhood1, $details1, $nhood2, $details2);
    
    if ((defined($posSingle1))&&
	($scoreSingle1>=
	 $controlParams{'min-single-read-alignment-score'}))
    { # good quality alignment read 1
	if ((defined($posSingle2))&&
	    ($scoreSingle2>=
	    $controlParams{'min-single-read-alignment-score'}))
	{ # good quality single alignment both reads
	    if ($chromSingle1 ne $chromSingle2)
	    {
		$bestInsert1="$chromSingle2:$posSingle2";
		$bestInsert2="$chromSingle1:$posSingle1";
	    } # if
	    else
	    {
		$bestInsert1=$posSingle2-$posSingle1;
		$bestInsert2=-$bestInsert1;
	    } # else
	    
	    printAlignmentLine
		($fh1, $readID1, $chromSingle1, 
		 $posSingle1, $strandSingle1,
		 $typeSingle1, $scoreSingle1, 
		 0, $strandSingle2, $bestInsert1);
	    
	    printAlignmentLine
		($fh2, $readID2, $chromSingle2, 
		 $posSingle2, $strandSingle2,
		 $typeSingle2, $scoreSingle2, 
		 0, $strandSingle1, $bestInsert2);


	    $pairProperties{Reads}{$key1}{$key2}
	    {$outcome}{BothAlignmentsOK}
	    +=$passedFiltering;

	} # if
	else
	{ # read 2 poor - treat as singleton read 1
	    $pairProperties{Reads}{$key1}{$key2}
	    {$outcome}{Read2Poor}
	    +=$passedFiltering;
	    
	    printSingletonAlignment
		($fh1,
		 $readID1, $nhood1,
		 $chromSingle1, $posSingle1, $strandSingle1,
		 $typeSingle1, $scoreSingle1,
		 $exactScore2, $qv2,
		 $fh2,
		 $readID2, $nhood2);
	} # else
    } # if
    elsif ((defined($posSingle2))&&
	   ($scoreSingle2
	    >=$controlParams{'min-single-read-alignment-score'}))
    { # read 1 poor, read 2 good - treat as singleton read 2
	$pairProperties{Reads}{$key1}{$key2}
	{$outcome}{Read1Poor}
	+=$passedFiltering;
	
	printSingletonAlignment
	    ($fh2,
	     $readID2, $nhood2,
	     $chromSingle2,$posSingle2,$strandSingle2,
	     $typeSingle2, $scoreSingle2,
	     $exactScore1, $qv1,
	     $fh1,
	     $readID1, $nhood1);
	
    } # elsif
    else
    { # both reads have poor alignment scores
	$pairProperties{Reads}{$key1}{$key2}
	{BothAlignButNoFeasiblePair}{$outcome}
	+=$passedFiltering;
	printAlignmentLine(*BEST1, $readID1, $nhood1);
	printAlignmentLine(*BEST2, $readID2, $nhood2);
    } # else
    
} # printUnpairedAlignments

#
# --- Initialize ---
# 

parseCircular();


open (READ1, $elandExtendedFile1) 
    || die "Failed to open $elandExtendedFile1: $!";
open (READ2, $elandExtendedFile2) 
    || die "Failed to open $elandExtendedFile2: $!";
open (QV1, $qualityValueFile1) 
    || die "Failed to open $qualityValueFile1: $!";
open (QV2, $qualityValueFile2) 
    || die "Failed to open $qualityValueFile2: $!";
open (FILT, $filtFile) 
    || die "Failed to open $filtFile: $!";

#
# --- Pass 1: estimate insert from clusters with unique alignment of each read
#



while (defined($l1=<READ1>))
{
    $numClusters++;
    $l2=<READ2>;
    die "Error: $elandExtendedFile1 & $elandExtendedFile2 need same num lines"
	unless (defined($l2));
    $filt=<FILT>;
    die "Error: $elandExtendedFile1 & $filtFile need same num lines"
	unless (defined($filt));
    next unless ($filt=~/^Y/);
    $numFilteredClusters++;



#    next if (($details1 eq '-')||($details2 eq '-'));
#    next if (($details1=~/,/)||($details2=~/,/));

    # TC 7.8.7 - changed to speed up, only bother parsing read 2 line 
    # if read 1 line is any good

    chomp($l1);
    ($readID1, $read1, $nhood1, $details1)=split("\t",$l1);

    next if (($details1 eq '-')||($details1=~/,/));
#    next if (($details1=~/,/)||($details2=~/,/));

    chomp($l2);

    ($readID2, $read2, $nhood2, $details2)=split("\t",$l2);

    next if (($details2 eq '-')||($details2=~/,/));



    $rl1=length($read1) unless (defined($rl1));
    $rl2=length($read2) unless (defined($rl2));

    ($chrom1, $pos1, $strand1, $type1)=getMatchDetails($details1,$chrom1);
    ($chrom2, $pos2, $strand2, $type2)=getMatchDetails($details2,$chrom2);

    
    if ($chrom1 eq $chrom2)
    {
	$orientation=getOrientation( $pos1, $strand1, $pos2, $strand2 );
	$pairProperties{Orientation}{$orientation}++;
	$numPairsTotal++;
	$insertLength
	    =getInsertLength( $pos1, $strand1, $rl1, $pos2, $strand2, $rl2 );
#	$pairProperties{InsertSize}{$orientation}+=$insertLength;
	$insertHist{$orientation}{$insertLength}++;
    } # if

} # while

# add read length parameters to XML output
$pairProperties{Length}{Read1}{SeedLengthForELAND}=$readLengthELAND1;
$pairProperties{Length}{Read2}{SeedLengthForELAND}=$readLengthELAND2;
$pairProperties{Length}{Read1}{Total}=$rl1;
$pairProperties{Length}{Read2}{Total}=$rl2;

    
$pairProperties{Pairs}{ClustersTotal}=$numClusters;
$pairProperties{Pairs}{ClustersUsedToComputeInsert}=$numPairsTotal;
$pairProperties{Pairs}{ClustersPassedFiltering}=$numFilteredClusters;


if ($numPairsTotal!=0)
{

    $pairProperties{Pairs}{InitialUniquePairsPercent}
    = sprintf("%-5.5f",100.0*$numPairsTotal/$numClusters);
 
    my ($thisOr, $maxOr) =(0,0);
    for $orientation( @orientation )
    {
	if (defined($pairProperties{Orientation}{$orientation}))
	{
#	    $pairProperties{InsertSize}{$orientation}
#	    /=$pairProperties{Orientation}{$orientation};
	} # if
	else
	{
	    $pairProperties{Orientation}{$orientation}=0;
#	    $pairProperties{InsertSize}{$orientation}=0;
	} # else  
 
	if ($pairProperties{Orientation}{$orientation}>$maxOr)
	{
	    $maxOr=$pairProperties{Orientation}{$orientation};
	    $pairProperties{Orientation}{Nominal}=$orientation;
	} # if

    } # for

#    $pairProperties{Orientation}{Confidence}
#    = sprintf("%8.5f", (100.0*$maxOr/$pairProperties{Pairs}{Used}));

    if ($maxOr!=$numPairsTotal)
    {
	$pairProperties{Orientation}{NominalOrientationPercent}
	= sprintf("%-5.5f",100.0*$maxOr/$numPairsTotal);
# uncomment next to print as Q-score	
#	= sprintf("%-5.5f", 
#		  -10*log(($numPairsTotal-$maxOr)/$numPairsTotal)
#		  /log(10));
    } # if
    else
    {
	$pairProperties{Orientation}{NominalOrientationPercent}=100;
    } # else

    print STDERR 
	"$0: Found $numPairsTotal initial pairs from $numClusters clusters\n";

    if ($pairProperties{Pairs}{InitialUniquePairsPercent}
	   <$controlParams{'min-percent-unique-pairs'})
    {
	print STDERR "$0: only ", 
	$pairProperties{Pairs}{InitialUniquePairsPercent},
	" percent of clusters with unique alignment of each read, need ",
	$controlParams{'min-percent-unique-pairs'}, 
	" percent to do paired alignment\n";
	$doPairedAlignment=undef;
    } # if
    elsif ($pairProperties{Orientation}{NominalOrientationPercent}
	    <$controlParams{'min-percent-consistent-pairs'})
    {
	print STDERR "$0: only ", 
	$pairProperties{Orientation}{NominalOrientationPercent},
	" percent of reads paired consistently, need ",
	$controlParams{'min-percent-consistent-pairs'}, 
	" percent to do paired alignment\n";
	$doPairedAlignment=undef;
    } # elsif
    else
    {
#	$nominalOrientation=$pairProperties{Orientation}{Nominal};
	estimateInsertSizeDist();
    } # else

} # if
else
{
    print STDERR "$0: No qualifying pairs found, can't estimate insert size\n";
    $pairProperties{Orientation}{NominalOrientationPercent} = 0;
    $pairProperties{Orientation}{InitialUniquePairsPercent} = 0;
    $pairProperties{Orientation}{Nominal} = 'Rp';
    $doPairedAlignment=undef;
} # else

#print XMLout(\%pairProperties,NoAttr=>1,RootName=>'ReadPairProperties');

die unless (defined($pairProperties{Orientation}{Nominal}));

print STDERR "$0: Nominal orientation of this sample appears to be ", 
    $pairProperties{Orientation}{Nominal}, "\n";



unless (defined($doPairedAlignment))
{
    print STDERR 
	"$0: warning - data will be treated as two sets of unpaired reads\n";
} # unless



#
# --- Second pass: pair up alignments using insert size information
#


seek READ1,0,0;
seek READ2,0,0;
seek FILT,0,0;

open (BEST1, ">$safFile1") || die "Failed to open $safFile1: $!";
open (BEST2, ">$safFile2") || die "Failed to open $safFile2: $!";
open (ANOM,  ">$anomalyFile") || die "Failed to open $anomalyFile: $!";


my $numWraparounds=0;
my $numResolved1=0;
my $numResolved2=0;
my $unique1=0;
my $unique2=0;
my $key1;
my $key2;
my $exactScore1;
my $exactScore2;
my $passedFiltering;
my $numOutputAlignments;

while ($l1=<READ1>)
{
    $l2=<READ2>;
    $qv1=<QV1>;
    $qv2=<QV2>;
    $filt=<FILT>;
    die unless ((defined($l2))
		&&(defined($qv1))
		&&(defined($qv2))
		&&(defined($filt)));
    chomp($l1);
    chomp($l2);
    chomp($qv1);
    chomp($qv2);
    $passedFiltering=(($filt=~/^Y/)?1:0);


#    @s1=split("\t",$l1);	
#    @s2=split("\t",$l2);	

    ($readID1, $read1, $nhood1, $details1)=split("\t",$l1);
    ($readID2, $read2, $nhood2, $details2)=split("\t",$l2);

    $exactScore1=getExactScoreForRead($qv1);
    $exactScore2=getExactScoreForRead($qv2);
    
    ($chromSingle1, $posSingle1, $strandSingle1, $typeSingle1, $scoreSingle1)
	=pickBestSingleReadAlignment
	($qv1, $nhood1, $details1, $readLengthELAND1, $exactScore1 );

    ($chromSingle2, $posSingle2, $strandSingle2, $typeSingle2, $scoreSingle2)
	=pickBestSingleReadAlignment
	($qv2, $nhood2, $details2, $readLengthELAND2, $exactScore2 );

#    $key1=makeKey($posSingle1,$nhood1,1);
#    $key2=makeKey($posSingle2,$nhood2,2);
    $key1=makeKey($nhood1,$details1,1);
    $key2=makeKey($nhood2,$details2,2);

    unless (defined($doPairedAlignment))
    {

	printUnpairedAlignments 
	    (*BEST1,
	     $readID1, $nhood1,
	     $chromSingle1, $posSingle1, $strandSingle1,
	     $typeSingle1, $scoreSingle1,
	     $exactScore1, $qv1,
	     $key1,
	     *BEST2,
	     $readID2, $nhood2,
	     $chromSingle2,$posSingle2,$strandSingle2,
	     $typeSingle2, $scoreSingle2,
	     $exactScore2, $qv2,
	     $key2,
	     'NoPairedAlignmentDone',$passedFiltering);
	next;
    } # unless

    # pickBestSingleReadAlignment returns 0 if no match found
    if (!defined($posSingle1))
    { # No match to first read
	printAnomalyLine($readID1, $read1, $read2, $qv1, $qv2, 
			 $nhood1, $details1, $nhood2, $details2);
	# print ANOM "$readID1 $nhood1 $nhood2\n";
	if (!defined($posSingle2))
	{ # No match to second read either
	    $pairProperties{Reads}{$key1}{$key2}{NoMatchToEither}+=$passedFiltering;
#	    print BEST1 "$readID1\tNM\n";
#	    print BEST2 "$readID2\tNM\n";
	    printAlignmentLine(*BEST1, $readID1, $nhood1);
	    printAlignmentLine(*BEST2, $readID2, $nhood2);
	} # if
	else
	{
#	    $pairProperties{Reads}{$key1}{$key2}
#	    {SingletonRead2}+=$passedFiltering;
	    $numOutputAlignments=printSingletonAlignment
	    (*BEST2,
	     $readID2, $nhood2,
	     $chromSingle2, $posSingle2, $strandSingle2,
	     $typeSingle2, $scoreSingle2,
	     $exactScore1, $qv1,
	     *BEST1,
	     $readID1, $nhood1);

	    if ($passedFiltering==1)
	    {
		if ($numOutputAlignments==2)
		{
		    $pairProperties{Reads}{$key1}{$key2}
		    {SingletonRead2}{AlignmentOKShadowAdded}++;
		} # if
		elsif($numOutputAlignments==1)
		{
		    $pairProperties{Reads}{$key1}{$key2}
		    {SingletonRead2}{AlignmentOK}++;
		} # elsif
		else
		{
		    $pairProperties{Reads}{$key1}{$key2}
		    {SingletonRead2}{AlignmentPoor}++;
		} # else
	    } # if

	} # else
    } # if
    elsif (!defined($posSingle2))
    { # no match to second read
#	$pairProperties{Reads}{$key1}{$key2}{SingletonRead1}+=$passedFiltering;
	printAnomalyLine($readID1, $read1, $read2, $qv1, $qv2, 
			 $nhood1, $details1, $nhood2, $details2);
#	print ANOM "$readID1 $nhood1 $nhood2\n";

	$numOutputAlignments=printSingletonAlignment
	(*BEST1,
	 $readID1, $nhood1,
	 $chromSingle1, $posSingle1, $strandSingle1,
	 $typeSingle1, $scoreSingle1,
	 $exactScore2, $qv2,
	 *BEST2,
	 $readID2, $nhood2);

	if ($passedFiltering==1)
	{
	    if ($numOutputAlignments==2)
	    {
		$pairProperties{Reads}{$key1}{$key2}
		{SingletonRead1}{AlignmentOKShadowAdded}++;
	    } # if
	    elsif ($numOutputAlignments==1)
	    {
		$pairProperties{Reads}{$key1}{$key2}
		{SingletonRead1}{AlignmentOK}++;
	    } # elsif
	    else
	    {
		$pairProperties{Reads}{$key1}{$key2}
		{SingletonRead1}{AlignmentPoor}++;
	    } # else
	} # if

    } # elsif        
    else
    {
	@m1=split(',',$details1);
	@m2=split(',',$details2);

	@pairsFound=();

	for $match1 (@m1)
	{
	    ($chrom1, $pos1, $strand1, $type1)
		=getMatchDetails($match1,$chrom1);
	    for $match2 (@m2)
	    {
#		($chrom1, $pos1, $strand1, $type1)
#		     =getMatchDetails($match1,$chrom1);
		($chrom2, $pos2, $strand2, $type2)
		     =getMatchDetails($match2,$chrom2);
		if ($chrom1 eq $chrom2)
		{
		    $orientation=getOrientation( $pos1, $strand1, 
						 $pos2, $strand2 );
		    if ($orientation eq $pairProperties{Orientation}{Nominal}) 
		    {
			$insertLength=getInsertLength
			( $pos1, $strand1, $rl1, $pos2, $strand2, $rl2 );
			if (($insertLength>=$pairProperties{InsertSize}{Min})
			    &&($insertLength<=$pairProperties{InsertSize}{Max}))
			{
			    push @pairsFound, 
			    "$match1 $chrom1 $match2 $chrom2";
			} # if
		    }
		    elsif ((defined($circular))
			   &&(defined($circular->{$chrom1}))
			   &&($orientation eq 
			      $antiOrientation
			      {$pairProperties{Orientation}{Nominal}}))
		    {
			if ($pos1<$pos2) 
			{ 
			    $insertLength=getInsertLength
				( $pos1+$circular->{$chrom1}, $strand1, $rl1, 
				  $pos2, $strand2, $rl2 );
			} # if
			else
			{ 
			    $insertLength=getInsertLength
				( $pos1, $strand1, $rl1, 
				  $pos2+$circular->{$chrom1}, $strand2, $rl2 );
			} # else

			if (($insertLength>=$pairProperties{InsertSize}{Min})
			    &&($insertLength<=$pairProperties{InsertSize}{Max}))			{
			    $numWraparounds++;
			    push @pairsFound, 
			    "$match1 $chrom1 $match2 $chrom2";
#			    print STDERR "$chrom1 $pos1 $pos2\n";
			} # if
		    } # elsif
		} # if chrom1 eq chrom2
	    } # for @m2
	} # for @m1

	if (@pairsFound==0)
	{
#	    $pairProperties{Reads}{$key1}{$key2}
#	    {BothAlignButNoFeasiblePair}+=$passedFiltering;


	    printUnpairedAlignments 
		(*BEST1,
		 $readID1, $nhood1,
		 $chromSingle1, $posSingle1, $strandSingle1,
		 $typeSingle1, $scoreSingle1,
		 $exactScore1, $qv1,
		 $key1,
		 *BEST2,
		 $readID2, $nhood2,
		 $chromSingle2,$posSingle2,$strandSingle2,
		 $typeSingle2, $scoreSingle2,
		 $exactScore2, $qv2,
		 $key2,
		 'BothAlignButNoFeasiblePair',$passedFiltering);


	} # if pairsFound=0
 	elsif (@pairsFound>=1)
	{ # at least one decent pair found
#	    print $pairsFound[0], "\n";

	    if (@pairsFound==1)
	    { # single best pair found
		($bestMatch1, $chrom1, $bestMatch2, $chrom2)
		    =split(' ',$pairsFound[0]);
		die unless (defined($chrom2));
		($chrom1, $pos1, $strand1, $type1)
		    =getMatchDetails($bestMatch1,$chrom1);
		($chrom2, $pos2, $strand2, $type2)
		    =getMatchDetails($bestMatch2,$chrom2);

		my $dummy;
		$nhood1=adjustPairedNhood($nhood1);
		$nhood2=adjustPairedNhood($nhood2);

		($dummy, $dummy, $dummy, $dummy, $bestScore1)
		    =pickBestSingleReadAlignment
		    ($qv1, $nhood1, $bestMatch1, $readLengthELAND1, 
		     $exactScore1 );

		($dummy, $dummy, $dummy, $dummy, $bestScore2)
		    =pickBestSingleReadAlignment
		    ($qv2, $nhood2, $bestMatch2, $readLengthELAND2, 
		     $exactScore2 );

		# add up bestScore1 and bestScore2 
		# - this is 1 alignment of 2 reads so both must get same score
		$bestScore1+=$bestScore2;

		$bestInsert1=$pos2-$pos1;
		$bestInsert2=-$bestInsert1;
		

		$pairProperties{Reads}{$key1}{$key2}
		{UniquePairedAlignment}+=$passedFiltering;

	    } # if pairsFound==1
	    else
	    { # many decent pairs found
		($chrom1, $pos1, $strand1, $type1, 
		 $chrom2, $pos2, $strand2, $type2,
		 $bestScore1 )
		    =bestPairedAlignFromList
		($qv1, $qv2, $readLengthELAND1, $readLengthELAND2, 
		 $exactScore1, $exactScore2, 
		 \@pairsFound);
		
		$bestInsert1=$pos2-$pos1;
		$bestInsert2=-$bestInsert1;

		$pairProperties{Reads}{$key1}{$key2}
		{ManyPairedAlignments}+=$passedFiltering;
	    } # else

	
	    if ($bestScore1
		    >=$controlParams{'min-paired-read-alignment-score'})
	    { # TBD - don't output alignment if score is bad
		printAlignmentLine
		    (*BEST1, $readID1, $chrom1, $pos1, $strand1, $type1, 
		     $scoreSingle1, $bestScore1, $strand2, $bestInsert1);
		printAlignmentLine
		    (*BEST2, $readID2, $chrom2, $pos2, $strand2, $type2,
		     $scoreSingle2, $bestScore1, $strand1, $bestInsert2);
		printAnomalyLine($readID1);
	    }
	    else
	    {
		printAlignmentLine(*BEST1, $readID1, $nhood1);
		printAlignmentLine(*BEST2, $readID2, $nhood2);
	    } # else

	    # print ANOM "OK\n";
	    $numResolved1+=(@m1>1);
	    $numResolved2+=(@m2>1);
	    $unique1+=(@m1==1);
	    $unique2+=(@m2==1);



	} # else
    } #else

} # while READ1

#
# --- Tidy up and produce XML output
#

close (READ1);
close (READ2);
close (QV1);
close (QV2);
close (FILT);
close (BEST1);
close (BEST2);
close (ANOM);


print STDERR "$unique1 read 1s were unique\n";
print STDERR "Resolved a further $numResolved1 read 1s with pairing information\n";
print STDERR "$unique2 read 2s were unique\n";
print STDERR "Resolved a further $numResolved2 read 2s with pairing information\n";
if (defined($circular))
{
 print STDERR 
     "An extra $numWraparounds possible pairings were found using circularization\n";
} # if

$pairProperties{ControlParametersUsed}=\%controlParams;

if (defined($circular))
{
    $pairProperties{CircularizationParametersUsed}=$circular;
} # if


print STDERR "Outputting XML summary\n";
print XMLout(\%pairProperties,
	     NoAttr=>1,
	     KeyAttr=>[],
	     XMLDecl=>1,
	     RootName=>'ReadPairProperties');
#	     GroupTags=>{Reads=>'R1'});
#	     GroupTags=>{'Pairs'=>'R1'});

# --- Thank you and goodnight ---
