#!/usr/bin/env perl
use warnings;
use strict;

use FindBin;
use lib "$FindBin::Bin/../lib";
use Gerald::Common;

#------------------------------------------------------------------------------
# File: score.pl
# Copyright (c) 2005, 2006 Solexa
# Author: A. J. Cox
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# Description:
# Use output of PhageAlign.cpp to estimate error rate at each cycle.
# This in turn is used to generate log-likelihood scores which
# can be fed back into PhageAlign.cpp. This process is iterated
# to convergence to obtain maximum likelihood error rates.
#------------------------------------------------------------------------------

my $version;
my $id;
{
    my $cvs_id='@(#) $Id: score.pl,v 1.3 2008/07/22 08:54:57 rshaw Exp $';
    my $cvs_version='$Revision: 1.3 $';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope

#------------------------------------------------------------------------------
# Define sets of bases and their possible combinations.

my @bases = qw/A C G T/;

# Conversion factor
my $toBase10 = 1.0 / log(10.0);

# Input types
my $INPUT_TYPE_ALIGN = 1;
my $INPUT_TYPE_SAF = 2;
my $INPUT_TYPE_SAF_MULTI = 3;

# Filter type strings
my $raw_str = 'raw';
my $QF_str = 'QF';

# Read type string
my $all_reads_str = 'all_reads';

# Used to find the names of any tiles not represented in the export file(s).
# Empty (re)score.txt files should be generated for such tiles.
my $tile_file_path = './tiles.txt';

#------------------------------------------------------------------------------

sub print_most_common_words($$$$$$);
sub print_most_common_blank_masks($$$$$);

#------------------------------------------------------------------------------

sub get_read_ref_pairs_by_type($$$$)
{
    my $blank_base = shift;
    my $correct_pairs_ref = shift;
    my $err_pairs_ref = shift;
    my $blank_read_pairs_ref = shift;

    @$correct_pairs_ref = qw/ AA CC GG TT /;
    @$err_pairs_ref = qw/ AC AG AT CA CG CT GA GC GT TA TC TG/;
    @$blank_read_pairs_ref = map {"$blank_base$_"} @bases;

#    @$all_pairs_ref = (@$correct_pairs_ref, @$err_pairs_ref,
#                       @$blank_read_pairs_ref);
}

#------------------------------------------------------------------------------
# The hash_seq frequency table grows very large even for a single tile,
# so at the end of each tile we extract just the information that is needed,
# storing it in the form of the final output text in common_words_texts 
# - again by tile - and throw away the raw statistics.

sub handle_tile_end($$$$$$$$)
{
    my $input_is_saf = shift;
    my $tile_num = shift;
    my $multiple_reads = shift;
    my $read_num = shift;

    my $hash_seq_ref = shift;
    my $hash_blank_ref = shift;
    my $common_words_texts_ref = shift;
    my $blank_base = shift;

    my @data_types = ($raw_str);

    if ($input_is_saf) {
        push (@data_types, $QF_str);

        @data_types = map { join(':', ($tile_num, $_)) } @data_types;

        if ($multiple_reads) {
            @data_types = map { join(':', ($_, $read_num)) } @data_types;
        }
    }

    my $max_num_to_list = 30;
    my $max_num_per_line = 6;

    foreach my $data_type (@data_types) {
        my $cache_str = '';

        print_most_common_words(\$cache_str, $data_type, $hash_seq_ref,
                                $max_num_to_list, $max_num_per_line,
                                $blank_base);

        print_most_common_blank_masks(\$cache_str, $data_type, $hash_blank_ref,
                                      $max_num_to_list, $max_num_per_line);

        # Store what is actually needed.
        $common_words_texts_ref->{$data_type} = $cache_str;
    }

    # Clear the large tables - the whole point of this exercise.
    %{$hash_seq_ref} = ();
    %{$hash_blank_ref} = ();
}

#------------------------------------------------------------------------------
# Requires reads to be supplied in ascending order.

sub read_input($$$$$$$$$$$$$$$$$$$$$$$)
{
    my $input_is_saf = shift;
    my $blank_base = shift;
    my $multiple_reads = shift;
    my $num_reads_ref = shift;
    my $header_strs_ref = shift;
    my $all_bases_ref = shift;
    my $hash_seq_ref = shift;
    my $hash_blank_ref = shift;
    my $common_words_texts_ref = shift;
#    my $cyc_bases_ref = shift;
    my $all_hash_ref = shift;
    my $cyc_hash_ref = shift;
    my $cycle_counts_ref = shift;
    my $unique_align_counts_ref = shift;
    my $align_score_sums_ref = shift;
    my $base_counts_ref = shift;
    my $blank_counts_ref = shift;
    my $err_counts_ref = shift;
    my $lane_num_ref = shift;
    my $tile_freqs_ref = shift;
    my $cum_err_counts_by_cycle_ref = shift;
    my $max_errs_to_count_by_cycle = shift;
    my $qual_val_freqs_ref = shift; # used only for SAF
    my $errs_by_qual_val_ref = shift; # used only for SAF

    # FIXME : Is aligned_num_fields not 7 for _*align file?
    my ($aligned_num_fields, $not_aligned_num_fields,
        $read_seq_field_ind, $strand_dir_field_ind, $align_score_field_ind)
        = ($input_is_saf
           ? (22, 22, 8, 13, 15)
           : (6, 3, 0, 4, 1)); 

    my $lane_field_ind = 2;       # only for SAF
    my $tile_field_ind = 3;       # only for SAF

    my $num_aligns_field_ind = 2; # only for _align; single align in SAF export
    my $read_num_field_ind = 7;   # only for SAF multiple
    my $qual_field_ind = 9;       # only for SAF
    my $align_chrom_field_ind = 10; # only for SAF
    my $align_pos_field_ind = 12; # used only for SAF

    my $ref_seq_field_ind = 5;    # only for _align; SAF has mismatch
    my $mismatch_field_ind = 14;  # only for SAF; _align has full ref seq
    my $filtered_field_ind = 21;  # only for SAF

    my @cycle_offsets = ();
    my $cycle_offset = 0;
    my $num_cycles = 0;

    my ($lane_num, $tile_num, $last_tile_num) = (0, 0, 0);

    # If there are multiple reads, the read num is parsed and
    # $highest_read_num will be updated; otherwise just set it to 1 here.
    my $highest_read_num = ($multiple_reads ? 0 : 1);
    my $read_num = 1;
    my $last_read_num = 1;

    my @data_types = ();
    my @last_data_types = ();

    my $line_num = 0;

    while (<STDIN>) {
        ++$line_num;

        if (/^\#/o) {
            push @$header_strs_ref, $_;
            next;
        }

	chomp;
        my @fields = ($input_is_saf ? split(/\t/o) : split(/\s+/o));
        my $read_seq = $fields[$read_seq_field_ind];
        $num_cycles = length($read_seq);
	my $qual_seq = ($input_is_saf ? $fields[$qual_field_ind] : '');

        @data_types = ();
	my @qv_data_types = ();

        if ($input_is_saf) {
            my $read_str = '';

            # Could do checking, e.g. that lane is same for all reads.
            $lane_num = $fields[$lane_field_ind];
	    $tile_num = $fields[$tile_field_ind];
            ++$tile_freqs_ref->{$tile_num};

            if ($multiple_reads) {
		$read_num = $fields[$read_num_field_ind];

		if (!defined($read_num)) {
		    die("Read number undefined at line $line_num\n");
		}

		$read_str = ":$read_num";

		if ($read_num > $highest_read_num) {
		    $highest_read_num = $read_num;
		    updateCycleOffsets(\@cycle_offsets,
				       $read_num, $num_cycles);
		}
            }

            # We now know the tile num so we can check if it has changed
            # since the last input line and hence need to call handle_tile_end.
            # To handle the case of a single-tile lane in eland_pair, we also 
            # need to check for change in read num (extracted above as well).
            # Also check we are not on the first line, i.e. that the last vals
            # have really been set from a previous data line.
            if ((($tile_num != $last_tile_num)
                 || ($read_num != $last_read_num))
                && ($line_num > 1)) { 

                # Pass the info for the tile that ended in the previous line.
                handle_tile_end($input_is_saf, $last_tile_num,
                                $multiple_reads, $last_read_num,
                                $hash_seq_ref, $hash_blank_ref,
                                $common_words_texts_ref, $blank_base);
            }

            $last_tile_num = $tile_num;
            $last_read_num = $read_num;


            # On with setting up the @data_types list.

            @data_types = ($raw_str . $read_str);
	    @qv_data_types = ($raw_str . $read_str);

            if ($multiple_reads) {
                push(@data_types, join(':', ($raw_str, $all_reads_str)));
            }

            my $filtered = $fields[$filtered_field_ind];

            if (defined($filtered) && ($filtered eq 'Y')) {
                push(@data_types, $QF_str . $read_str);
		push(@qv_data_types, $QF_str . $read_str);

                if ($multiple_reads) {
                    push(@data_types, join(':', ($QF_str, $all_reads_str)));
                }
            }

            @data_types = map { join(':', ($tile_num, $_)) } @data_types;
        } else {
            @data_types = ($raw_str);
        }

        my $blank_read_seq = $read_seq;
        
        # The 'N' -> '.' is necessary because export files represent unknown
        # base as 'N' (the original code was for align files where it is '.').
        # For export files, therefore, translating all known bases to 'N'
        # without translating the original Ns would map all traces to all Ns.
        $blank_read_seq =~ tr/ACGTacgtN/NNNNNNNN./;

        foreach my $data_type (@data_types) {
            $cycle_offset = (($data_type =~ /${all_reads_str}$/)
                             ? $cycle_offsets[$read_num] : 0);

            for (my $cycle_ind = 0; $cycle_ind < $num_cycles; ++$cycle_ind) {
                my $read_base = substr($read_seq, $cycle_ind, 1);

                $all_bases_ref->
                {$data_type}[$cycle_offset + $cycle_ind]{$read_base}++;
            }

            $hash_seq_ref->{$data_type}{$read_seq}++;
            $hash_blank_ref->{$data_type}{$blank_read_seq}++;

            # Might want to check for variation in length -> error?
            $cycle_counts_ref->{$data_type} = $cycle_offset + $num_cycles;
        }

	if (!$input_is_saf) {
	    # export file lines have a fixed number of fields.
	    next if (@fields == $not_aligned_num_fields);
	}

        die "Found fewer than $aligned_num_fields fields in line $line_num\n" 
            unless (@fields >= $aligned_num_fields);

	# Skip line if (SAF and not aligned) or (_align and multiple aligns).
	# For SAF, $fields[$align_chrom_field_ind] could be `NM' or `QC' or
	# `RM' or `\d+:\d+:\d+' -> use simpler test of align pos presence.
	next if ($input_is_saf
		 ? ($fields[$align_pos_field_ind] eq '')
		 : ($fields[$num_aligns_field_ind] != 1));

        my $ref_seq;

        if ($input_is_saf) {
            my $mismatch_str = $fields[$mismatch_field_ind];

            # `-' -> partner aligned but this read did not -> not counted
            next if ($mismatch_str eq '-');

            $ref_seq = apply_mismatch($read_seq, $num_cycles,
                                      $mismatch_str, $line_num);
        } elsif ($fields[$strand_dir_field_ind] eq 'R') {
            $ref_seq = reverse($fields[$ref_seq_field_ind]);
            $ref_seq =~ tr/ACGT/TGCA/;
        } else {
            $ref_seq = $fields[$ref_seq_field_ind];
        }

        # At this stage we have a unique alignment.
        # _*align :  the line has been skipped if ($num_aligns != 1).
        # SAF export : only uniq aligns and partner-only aligned
        #              but latter are skipped above.
        # The summing of scores from only unique alignments is consistent
        # across analysis modes but differs from legacy behaviour of jerboa
        # for legacy analysis modes : Approved by AC.

        foreach my $data_type (@data_types) {
            ++$unique_align_counts_ref->{$data_type};
            $align_score_sums_ref->{$data_type}
            += $fields[$align_score_field_ind];

            $cycle_offset = (($data_type =~ /${all_reads_str}$/)
                             ? $cycle_offsets[$read_num] : 0);

	    my $num_errs_in_read_up_to_curr_cycle = 0;

            for (my $cycle_ind = 0; $cycle_ind < $num_cycles; $cycle_ind++) {
		my $overall_cycle_ind = $cycle_offset + $cycle_ind;
                my $read_base = substr($read_seq, $cycle_ind, 1);
                my $ref_base = substr($ref_seq, $cycle_ind, 1);
                my $base_pair = $read_base . $ref_base;

                ++$base_counts_ref->{$data_type};
                $blank_counts_ref->{$data_type} += ($read_base eq $blank_base);

		if (($read_base ne $blank_base)
		    && ($ref_base ne $blank_base)) {
		    if ($read_base ne $ref_base) {
			++$err_counts_ref->{$data_type};
			++$num_errs_in_read_up_to_curr_cycle;
		    }

		    for (my $num = $num_errs_in_read_up_to_curr_cycle;
			 $num < $max_errs_to_count_by_cycle; ++$num) {
			$cum_err_counts_by_cycle_ref->
			{$data_type}[$overall_cycle_ind][$num]++;
		    }
		}

                $all_hash_ref->{$data_type}{$base_pair}++;
                $cyc_hash_ref->
                {$data_type}[$overall_cycle_ind]{$base_pair}++;
                $cyc_hash_ref->
                {$data_type}[$overall_cycle_ind]{$read_base}++;
            }
        }

	# Quality value statistics for export (cf. reportQuality.pl)
	if ($input_is_saf) {
	    my $qual_is_symbolic = 1;
	    my @qual_vals;

	    if (length($qual_seq) != $num_cycles) {
		$qual_is_symbolic = 0;
		@qual_vals = split(/ /, $qual_seq);
                my $num_qual_vals = @qual_vals;

		die("Unexpected format of quality field `$qual_seq' "
                    . "($num_qual_vals quality values, $num_cycles cycles)")
		    if ($num_qual_vals != $num_cycles);
	    }

	    for (my $cycle_ind = 0; $cycle_ind < $num_cycles; $cycle_ind++) {
                my $read_base = substr($read_seq, $cycle_ind, 1);
                my $ref_base = substr($ref_seq, $cycle_ind, 1);

		# Ignore any bases that are N in read or ref.
		# Unlike for a score file, there is no requirement to count
		# uncalled bases for a quality report.
		next if (($read_base eq 'N') || ($ref_base eq 'N'));

		my $qual_val = ($qual_is_symbolic
				? (ord(substr($qual_seq, $cycle_ind, 1)) - 64)
				: $qual_vals[$cycle_ind]);

		foreach my $data_type (@qv_data_types) {
		    if (!defined($qual_val_freqs_ref->
				 {$data_type}{$qual_val})) {
			$qual_val_freqs_ref->{$data_type}{$qual_val} = 0;
		    }

		    ++$qual_val_freqs_ref->{$data_type}{$qual_val};

		    if (!defined($errs_by_qual_val_ref->
				 {$data_type}{$qual_val})) {
			$errs_by_qual_val_ref->{$data_type}{$qual_val} = 0;
		    }

		    if ($read_base ne $ref_base) {
			++$errs_by_qual_val_ref->{$data_type}{$qual_val};
		    }
		}
	    }
	}
    } # loop over file lines

    if ($line_num > 0) {
        # Catch the end of the last (or only) tile - assuming some input.
        handle_tile_end($input_is_saf, $tile_num,
                        $multiple_reads, $read_num,
                        $hash_seq_ref, $hash_blank_ref,
                        $common_words_texts_ref, $blank_base);
    }

    $$num_reads_ref = $highest_read_num;
    $$lane_num_ref = $lane_num;
}

#------------------------------------------------------------------------------

sub print_header($$)
{
    local *OSTRM = shift;
    my $header_strs_ref = shift;

    foreach my $header_str (@$header_strs_ref) {
        print OSTRM $header_str;
    }
}

#------------------------------------------------------------------------------

sub print_summary_stats($$$$$$$)
{
    local *OSTRM = shift;
    my $num_bases = shift;
    my $num_blanks = shift;
    my $num_errs = shift;
    my $num_unique_aligns = shift;
    my $align_score_sum = shift;
    my $expected_num_cycles = shift;

    print OSTRM "$num_bases bases of sequence found\n";
    print OSTRM "$num_errs were errors\n";
    print OSTRM "$num_blanks were blanks\n";

    # Try to measure error rate by looking at scores

    if (($num_bases > 0) && (($num_bases - $num_blanks) > 0)) {
#        printf(OSTRM "%5.2f percent error rate\n",
#               (100.0 * $num_errs) / $num_bases);
#        printf(OSTRM "%5.2f percent blank rate\n",
#               (100.0 * $num_blanks) / ($num_blanks + $num_bases));

        printf(OSTRM "%5.2f percent error rate\n",
               (100.0 * $num_errs) / ($num_bases - $num_blanks));
        printf(OSTRM "%5.2f percent blank rate\n",
               (100.0 * $num_blanks) / ($num_bases));
    }

    print OSTRM "\nunique alignments : $num_unique_aligns ",
    "(total score $align_score_sum)\n";
    print OSTRM "cycles : $expected_num_cycles\n";
}

#------------------------------------------------------------------------------

sub print_errors_by_cycle($$$$$)
{
    local *OSTRM = shift;
    my $data_type = shift;
    my $num_cycles = shift;
    my $cyc_hash_ref = shift;
    my $blank_base = shift;

    my @correct_pairs;
    my @err_pairs;
    my @blank_read_pairs;

    get_read_ref_pairs_by_type($blank_base, \@correct_pairs,
                               \@err_pairs, \@blank_read_pairs);

    print OSTRM "\nBreakdown of errors by cycle\n";

    my $correctTotal;
    my $errTotal;
    my $blankTotal;

    my ($non_blank_total, $total);

    my $blankRate;
    my $errorRate;

    print OSTRM "Cycle:\tErr pc:\tBlank pc:\n";

    for (my $cycle_ind = 0; $cycle_ind < $num_cycles; ++$cycle_ind) {
        $correctTotal = 0;
        $errTotal = 0;
        $blankTotal = 0;

        foreach (@correct_pairs) {
            $correctTotal += $cyc_hash_ref->{$data_type}[$cycle_ind]{$_}
            if defined ($cyc_hash_ref->{$data_type}[$cycle_ind]{$_});
        }

        foreach (@err_pairs) {
            $errTotal += $cyc_hash_ref->{$data_type}[$cycle_ind]{$_}
            if defined ($cyc_hash_ref->{$data_type}[$cycle_ind]{$_});
        }

        foreach (@blank_read_pairs) {
            $blankTotal += $cyc_hash_ref->{$data_type}[$cycle_ind]{$_}
            if defined ($cyc_hash_ref->{$data_type}[$cycle_ind]{$_});
        }

        # NB error rate is defined as (#errors) / (#non-blank bases)
        # blank rate is defined as (#blanks) / (#all bases)

        $non_blank_total = $correctTotal + $errTotal;
        $total = $non_blank_total + $blankTotal;

        $errorRate = (($non_blank_total != 0)
                      ? ((100.0 * $errTotal) / $non_blank_total)
                      : 0);

        $blankRate = (($total != 0)
                      ? ((100.0 * $blankTotal) / $total)
                      : 0);


        printf(OSTRM "%d\t%5.2f\t%5.2f\n",
	       $cycle_ind + 1, $errorRate, $blankRate);
#       print OSTRM $cycle_ind+1, " $correctTotal $errTotal $blankTotal\n";

    } #for cycle_ind
}

#------------------------------------------------------------------------------

sub print_err_percents_by_bases($$$$$$)
{
    local *OSTRM = shift;
    my $data_type = shift;
    my $row_is_ref = shift;
    my $read_includes_blanks = shift;
    my $all_hash_ref = shift;
    my $read_bases_ref = shift;

    my $blanks_str = ($read_includes_blanks
                      ? '(including blanks)'
                      : '(excluding blanks)');

    my $title = 'Error rate relative to ' . ($row_is_ref
                                             ? "reference base $blanks_str"
                                             : "sequenced base");

    my $subtitle = ($row_is_ref
                    ? '(Given a reference base, what is it sequenced as?)'
                    : '(Given a sequenced base, what was it really?)');

    my $ref_label = 'Really:';
    my $read_label = 'Read as:';

    my $row_label = ($row_is_ref ? $ref_label : $read_label);
    my $col_label = ($row_is_ref ? $read_label : $ref_label);

    my @row_bases = (($row_is_ref || !$read_includes_blanks)
                     ? @bases
                     : @$read_bases_ref);

    my @col_bases = (!($row_is_ref && $read_includes_blanks)
                     ? @bases
                     : @$read_bases_ref);

    printf(OSTRM "\n$title");
    printf(OSTRM "\n$subtitle\n\n");

    print OSTRM "$row_label\t$col_label\n";

    foreach my $col_base (@col_bases) {
        print OSTRM "\t${col_base} pc";
    }

    print OSTRM "\n";

    foreach my $row_base (@row_bases) {
        print OSTRM "$row_base";
        my $tot = 0;
        my $val = 0;

        foreach my $col_base (@col_bases) {
            my ($read_base, $ref_base) = ($row_is_ref
                                          ? ($col_base, $row_base)
                                          : ($row_base, $col_base));
            my $base_pair = $read_base . $ref_base;

            $all_hash_ref->{$data_type}{$base_pair} = 0
            unless defined($all_hash_ref->{$data_type}{$base_pair});

            $tot += $all_hash_ref->{$data_type}{$base_pair};
        }

        foreach my $col_base (@col_bases) {
            my ($read_base, $ref_base) = ($row_is_ref
                                          ? ($col_base, $row_base)
                                          : ($row_base, $col_base));
            my $base_pair = $read_base . $ref_base;

            $val = (($tot != 0)
                    ? (100.0 * $all_hash_ref->{$data_type}{$base_pair} / $tot)
                    : 0);
            printf OSTRM "\t%5.2f", $val;
        }

        print OSTRM "\n";
    }
}

#------------------------------------------------------------------------------

sub print_errs_by_base($$$$)
{
    local *OSTRM = shift;
    my $data_type = shift;
    my $all_hash_ref = shift;
    my $read_bases_ref = shift;

    print OSTRM "\nBreakdown of errors by nucleotide\n\n";
    print OSTRM "Read As:\tReally:\n";
    foreach my $ref_base (@bases) { print OSTRM "\t$ref_base pc" };
    print OSTRM "\n";

    foreach my $read_base (@$read_bases_ref) {
        print OSTRM "$read_base";

#       $tot = 0;
#       foreach $ref_base (@bases) {
#	    $val=$all_hash_ref->{$data_type}{$read_base . $ref_base};
#	    $val=0 unless (defined($val));
#	    $tot+=$val;
#       }

        foreach my $ref_base (@bases) {
            my $val = $all_hash_ref->{$data_type}{$read_base . $ref_base};
            $val = 0 unless (defined($val));
            printf OSTRM "\t$val";
        }

        print OSTRM "\n";
    }
}

#------------------------------------------------------------------------------

sub print_errs_by_cycle_and_base($$$$$)
{
    local *OSTRM = shift;
    my $data_type = shift;
    my $cyc_hash_ref = shift;
    my $num_cycles = shift;
    my $read_bases_ref = shift;

    print OSTRM "\nFull breakdown of errors by cycle and nucleotide\n";
    print OSTRM "Cycle:\tRead As:\tReally:\n\t";

    foreach my $ref_base (@bases) { print OSTRM "\t$ref_base ct" };
    print OSTRM "\n";

    for (my $cycle_ind = 0; $cycle_ind < $num_cycles; ++$cycle_ind) {
        foreach my $read_base (@$read_bases_ref) {
            printf OSTRM "%d\t%s", $cycle_ind + 1, $read_base;

            foreach my $ref_base (@bases) {
                my $base_pair = $read_base . $ref_base;

                $cyc_hash_ref->{$data_type}[$cycle_ind]{$base_pair} = 0
                    unless defined ($cyc_hash_ref->
                                    {$data_type}[$cycle_ind]{$base_pair});

                my $val = $cyc_hash_ref->{$data_type}[$cycle_ind]{$base_pair};

                die unless (defined($val));
                printf OSTRM "\t$val";
            }

            print OSTRM "\n";
        }
    }
}

#------------------------------------------------------------------------------

sub print_err_fractions_by_cycle_and_base($$$$$)
{
    local *OSTRM = shift;
    my $data_type = shift;
    my $cyc_hash_ref = shift;
    my $num_cycles = shift;
    my $read_bases_ref = shift;

    print OSTRM "\nError rate relative to reference by cycle/nucleotide\n";
    print OSTRM "Cycle:\tRead As:\tReally:\n\t";
    foreach my $ref_base (@bases) { print OSTRM "\t$ref_base ct" };
    print OSTRM "\n";

    for (my $cycle_ind = 0; $cycle_ind < $num_cycles; ++$cycle_ind) {
        foreach my $read_base (@$read_bases_ref) {
            printf OSTRM "@%d\t%s", $cycle_ind + 1, $read_base;

            foreach my $ref_base (@bases) {
                my $tot = 0;

                map { 
                    $tot += $cyc_hash_ref->
                    {$data_type}[$cycle_ind]{$_ . $ref_base}
                } @$read_bases_ref;

                my $val = $cyc_hash_ref->
                {$data_type}[$cycle_ind]{$read_base . $ref_base};
#	        $val=0 unless (defined($val));
#	        die unless (defined($val));
                $val /= $tot unless ($tot == 0);
                printf OSTRM "\t%6.3f", $val;
            }

            print OSTRM "\n";
        }
    }
}

#------------------------------------------------------------------------------

sub print_info_content_by_cycle($$$$$)
{
    local *OSTRM = shift;
    my $data_type = shift;
    my $all_bases_ref = shift;
    my $cyc_hash_ref = shift;
    my $num_cycles = shift;

    my $priorEntropy = 2;
    my $toBase2 = 1.0 / log(2.0);

    my $thisEntropy;
    my $thisProb;
    my $thisInfo; 
    my $thisAligned;
    my $thisTotal;
    my $infoSoFar;
    my $alignedSoFar;
    my $totalSoFar;

    print OSTRM "\nInformation Content By Cycle\n";
    print OSTRM "\tBases this cycle\t\tBases so far\n";
    print OSTRM "Cycle:\tEquiv info:\tAlign:\tTotal:\t"
	. "Equiv info:\tAlign:\tTotal:\n";

    for (my $cycle_ind = 0; $cycle_ind < $num_cycles; ++$cycle_ind) {
        $thisInfo=0;

        for my $j (@bases) {
            # work out entropy contribution for letter $j
            $all_bases_ref->{$data_type}[$cycle_ind]{$j} = 0
                unless defined($all_bases_ref->{$data_type}[$cycle_ind]{$j});
            #   $numRef->[$i] += $all_bases_ref->{$data_type}[$cycle_ind]{$j};
        
            next unless (defined ($cyc_hash_ref->{$data_type}[$cycle_ind]{$j})
                         && ($cyc_hash_ref->{$data_type}[$cycle_ind]{$j}
                             != 0));

            $thisEntropy = 0;

            for my $k (@bases) { 
                # work out entropy of one instance of letter $j
                next unless (defined($cyc_hash_ref->
                                     {$data_type}[$cycle_ind]{$j . $k})
                             && ($cyc_hash_ref->
                                 {$data_type}[$cycle_ind]{$j . $k} != 0));
                $thisProb = ($cyc_hash_ref->{$data_type}[$cycle_ind]{$j . $k}
                             / $cyc_hash_ref->{$data_type}[$cycle_ind]{$j});
                $thisEntropy += ($thisProb * log($thisProb));
#                print OSTRM "$thisProb\n";
            } # for k

            $thisEntropy *= $toBase2; # convert units to bits
            $thisEntropy += $priorEntropy; # convert to an info gain
            $thisEntropy /= 2.0; # convert units to bases from bits
#            print OSTRM "$i $j $thisEntropy\n";
            $thisInfo += ($cyc_hash_ref->{$data_type}[$cycle_ind]{$j}
                          * $thisEntropy);
        } # for j

        $thisAligned=0;
        $thisTotal=0;

        for my $j (@bases) {
            $thisAligned += $cyc_hash_ref->{$data_type}[$cycle_ind]{$j}
            if (defined($cyc_hash_ref->{$data_type}[$cycle_ind]{$j}));

            $thisTotal += $all_bases_ref->{$data_type}[$cycle_ind]{$j}
            if (defined ($all_bases_ref->{$data_type}[$cycle_ind]{$j}));
        } # for $j


#    $thisAligned
#        =$cycHash[$i]{A}+$cycHash[$i]{C}+$cycHash[$i]{G}+$cycHash[$i]{T};
#    $thisTotal
#        =$allBases[$i]{A}+$allBases[$i]{C}+$allBases[$i]{G}+$allBases[$i]{T};

        $infoSoFar += $thisInfo;
        $alignedSoFar += $thisAligned;
        $totalSoFar += $thisTotal;

        printf(OSTRM "~%d\t%8.2f\t%d\t%d\t%8.2f\t%d\t%d\n",
               $cycle_ind + 1,
	       $thisInfo, $thisAligned, $thisTotal,
               $infoSoFar, $alignedSoFar, $totalSoFar);

    } # for cycle_ind
}

#------------------------------------------------------------------------------

sub print_most_common_patterns($$$)
{
    local *OSTRM = shift;
    my $data_type = shift;
    my $common_words_texts_ref = shift;

    my $text = $common_words_texts_ref->{$data_type};

    if (defined($text)) {
        print OSTRM $text;
    }
}

#------------------------------------------------------------------------------
# Print most common words (with <= 2 blanks) by rank.
# N.B. This function actually prints into a cache string,
#      subsequently printed by print_most_common_patterns. (stringstream?)

sub print_most_common_words($$$$$$)
{
    my $cache_str_ref = shift;
    my $data_type = shift;
    my $hash_seq_ref = shift;
    my $max_num_to_list = shift;
    my $max_num_per_line = shift;
    my $blank_base = shift;

    my $cache_str = $$cache_str_ref;

    my $print_blank = '.';
    my $maxGaps = 2;

    my %countNum = ();
    my %countSeq = ();

    if (defined($hash_seq_ref->{$data_type})) {
        foreach my $seq (sort keys %{$hash_seq_ref->{$data_type}}) {

            # Would prefer to use variables but this is unpleasant and slow :-
            # $_ = $seq;
            # eval "tr/$blank_base/$print_blank/";
            # my $print_seq = $_;
            my $print_seq = $seq;
            $print_seq =~ tr/N/./;

            my $num_gaps = 0;

            while ($print_seq =~ /\./g) { ++$num_gaps; };
            next if ($num_gaps > $maxGaps);

            $countSeq{$hash_seq_ref->{$data_type}{$seq}} .= "$print_seq ";
            $countNum{$hash_seq_ref->{$data_type}{$seq}}++;
        }
    }

    my $numToList = $max_num_to_list;
    $numToList = keys %countSeq if ((keys %countSeq) < $max_num_to_list);
    $cache_str .= 
	"\nThe $numToList most common words with $maxGaps blanks or less:\n";
    $cache_str .= "Ranking\tOccurrences\tWords\n";

    my $rank = 0;

    foreach my $i (sort {$b <=> $a} keys %countSeq) {
	++$rank;
	last if ($rank > $max_num_to_list);

	if ($countNum{$i} > $max_num_per_line) {
	    $cache_str .= "$rank\t$i\t$countNum{$i} sequences\n";
	} else {
	    $cache_str .= "$rank\t$i\t$countSeq{$i}\n";
	}
    }

    $$cache_str_ref = $cache_str;
}

#------------------------------------------------------------------------------
# Print most common patterns of blank/non blank bases by rank.
# FIXME : other than filtering, duplicates print_most_common_words.

# N.B. This function actually prints into a cache string,
#      subsequently printed by print_most_common_patterns. (stringstream?)

sub print_most_common_blank_masks($$$$$)
{
    my $cache_str_ref = shift;
    my $data_type = shift;
    my $hash_blank_ref = shift;
    my $max_num_to_list = shift;
    my $max_num_per_line = shift;

    my $cache_str = $$cache_str_ref;

    my %countNum = ();
    my %countSeq = ();

    if (defined($hash_blank_ref->{$data_type})) {
        foreach my $seq (sort keys %{$hash_blank_ref->{$data_type}}) {
            $countSeq{$hash_blank_ref->{$data_type}{$seq}} .= "$seq ";
            $countNum{$hash_blank_ref->{$data_type}{$seq}}++;
        }
    }

    my $numToList = $max_num_to_list;
    $numToList = keys %countSeq if ((keys %countSeq) < $max_num_to_list);
    $cache_str .= ("\nThe $numToList most common blank patterns "
                   . "(N=any nonblank character)");
    $cache_str .= "\nRanking\tOccurrences\tWords\n";

    my $rank = 0;

    foreach my $i (sort {$b <=> $a} keys %countSeq) {
	++$rank;
	last if ($rank > $max_num_to_list);

	if ($countNum{$i} > $max_num_per_line) {
	    $cache_str .= "$rank\t$i\t$countNum{$i} sequences\n";
	} else {
	    $cache_str .= "$rank\t$i\t$countSeq{$i}\n";
	}
    }

    $$cache_str_ref = $cache_str;
}

#------------------------------------------------------------------------------

sub print_likelihood_scores($$$$$)
{
    local *OSTRM = shift;
    my $data_type = shift;
    my $cyc_hash_ref = shift;
    my $num_cycles = shift;
    my $read_bases_ref = shift;

    # score you would get if each outcome is equiprobable
    # give this score in event of no information
    # (maybe better to average scores from other cycles)
    my $neutralScore = int(100 * $toBase10 * log(1 / (4 - 1)));


    print OSTRM "\nLog likelihood scores\n";
    print OSTRM "Cycle:\tRead As:\tReally:\n\t";
    foreach my $ref_base (@bases) { print OSTRM "\t$ref_base" };
    print OSTRM "\n";

    for (my $cycle_ind = 0; $cycle_ind < $num_cycles; ++$cycle_ind) {
	foreach my $read_base (@$read_bases_ref) {
	    printf OSTRM ">%d\t%s", $cycle_ind + 1, $read_base;

	    my $tot = 0;

	    foreach my $ref_base (@bases) {
		my $val = $cyc_hash_ref->
                {$data_type}[$cycle_ind]{$read_base . $ref_base};
		$val = 0 unless (defined($val));
		$tot += $val;
	    }

	    foreach my $ref_base (@bases) {
		my $val = $cyc_hash_ref->
                {$data_type}[$cycle_ind]{$read_base . $ref_base};
		$val = 0 unless (defined($val));
		my $matchScore = $neutralScore;
		$matchScore = int(100 * $toBase10 * log($val / ($tot - $val)))
		    unless (($tot == $val) || ($val == 0));
		print OSTRM "\t$matchScore";
	    }

	    print OSTRM "\n";
	} # loop over bases read
    } # loop over cycles
}

#------------------------------------------------------------------------------

sub print_cum_errors_by_cycle($$$$$)
{
    local *OSTRM = shift;
    my $data_type = shift;
    my $cum_err_counts_by_cycle_ref = shift;
    my $num_cycles = shift;
    my $max_errs_to_count_by_cycle = shift;

    print OSTRM "\nCumulative errors by cycle\n";
    print OSTRM "Cycle:";

    my $separ_str = "\t";

    for (my $num = 1; $num <= $max_errs_to_count_by_cycle; ++$num) {
	print OSTRM $separ_str, $num;
    }

    print OSTRM "\n";

    for (my $cycle_ind = 0; $cycle_ind < $num_cycles; ++$cycle_ind) {
	printf OSTRM "!%d", $cycle_ind + 1;

	for (my $num = 1; $num <= $max_errs_to_count_by_cycle; ++$num) {
	    if (!defined($cum_err_counts_by_cycle_ref->
			 {$data_type}[$cycle_ind][$num - 1])) {
		$cum_err_counts_by_cycle_ref->
		{$data_type}[$cycle_ind][$num - 1] = 0;
	    }

	    print OSTRM $separ_str,
	    $cum_err_counts_by_cycle_ref->{$data_type}[$cycle_ind][$num - 1];
	}

	print OSTRM "\n";
    }

    print OSTRM "\n";
}

#------------------------------------------------------------------------------
# If the data type is one that has been QFed, it is possible that no entries
# were found for it (no clusters passed QF) and thus the number of cycles
# could not be determined. If so, attempt to use the number of cycles for
# the corresponding raw data type. 

sub get_expected_num_cycles($$$)
{
    my $cycle_counts_ref = shift;
    my $data_type = shift;
    my $num_cycles = shift;

    if ($num_cycles == 0) {
	my @data_type_fields = split(/:/, $data_type);
	my $QF_field_ind = 1;
	my $QF_field = $data_type_fields[$QF_field_ind];

	if (defined($QF_field)) {
	    if ($QF_field eq $QF_str) {
		splice(@data_type_fields, $QF_field_ind, 1, $raw_str);
		my $raw_data_type = join(':', @data_type_fields);
		$num_cycles = (defined($cycle_counts_ref->{$raw_data_type})
			       ? $cycle_counts_ref->{$raw_data_type} : 0);
	    }
	}
    }

    return $num_cycles;
}

#------------------------------------------------------------------------------

sub filename_for_data_type($$)
{
    my $data_type = shift;
    my $lane_num = shift;

    my %qual_type_str_map;
    $qual_type_str_map{$raw_str} = '_score.txt';
    $qual_type_str_map{$QF_str} = '_rescore.txt';

    my @data_type_fields = split(/:/, $data_type);

    # Ignore read number.

    my $tile_num = $data_type_fields[0];
    my $qual_str = $qual_type_str_map{$data_type_fields[1]};

    my $filename = sprintf("s_%d_%04d%s", $lane_num, $tile_num, $qual_str);

    return $filename;
}

#------------------------------------------------------------------------------

sub is_all_reads_data_type($)
{
    my $data_type = shift;
    my @data_type_fields = split(/:/, $data_type);
    my $is_all_reads = 0;

    if (@data_type_fields > 2) {
        my $read_field = $data_type_fields[-1];
        $is_all_reads = ($read_field eq $all_reads_str);
    }

    return $is_all_reads;
}

#------------------------------------------------------------------------------

sub title_for_data_type($$)
{
    my $data_type = shift;
    my $lane_num = shift;

    my %qual_type_title_map;
    $qual_type_title_map{$raw_str} = 'Raw';
    $qual_type_title_map{$QF_str} = 'Quality Filtered';

    my @data_type_fields = split(/:/, $data_type);

    my $lane_str = 'Lane ' . $lane_num;
    my $tile_str = 'Tile ' . $data_type_fields[0];
    my $qual_title_str = $qual_type_title_map{$data_type_fields[1]};

    my $read_str = '';

    if (@data_type_fields > 2) {
        my $read_field = $data_type_fields[-1];
        $read_str = (($read_field eq $all_reads_str)
                     ? ' : All Reads'
                     : " : Read $read_field");
    }

    return join(' : ', ($lane_str, $tile_str, $qual_title_str)) . $read_str;
}

#------------------------------------------------------------------------------

sub write_bad_tile_files($$$$)
{
    my $tile_freqs_ref = shift;
    my $lane_num = shift;
    my $per_tile_data_types_ref = shift;
    my $header_strs_ref = shift;

    my @expected_tile_nums = ();

    if (!get_expected_tile_nums($tile_file_path, $lane_num,
                                \@expected_tile_nums)) {
        print STDERR "Failed to get expected tile nums.\n";
        exit 1;
    }

    my $last_filename = '';

    foreach my $expected_tile_num_str (@expected_tile_nums) {
        # Strip any leading zeros to match keys in %tile_freqs.
        my $expected_tile_num = int($expected_tile_num_str);

        if (!defined($tile_freqs_ref->{$expected_tile_num})) {
            foreach my $data_type (map { join(':', ($expected_tile_num, $_)) }
                                   @$per_tile_data_types_ref) {
                my $filename = filename_for_data_type($data_type, $lane_num);
                my $title_str = title_for_data_type($data_type, $lane_num);

                if ($filename ne $last_filename) {
                    open(OSTRM, ">$filename");
                    $last_filename = $filename;
                }

                print_header(*OSTRM, $header_strs_ref);
                print OSTRM "# $title_str\n";
            }
        }
    }    
}

#------------------------------------------------------------------------------

sub write_one_stat_set($$$$$$$$$$$$$$$$$$)
{
    local *OSTRM = shift;
    my $input_is_saf = shift;
    my $blank_base = shift;
    my $lane_num = shift;
    my $data_type = shift;
    my $cycle_counts_ref = shift;
    my $unique_align_counts_ref = shift;
    my $align_score_sums_ref = shift;
    my $base_counts_ref = shift;
    my $blank_counts_ref = shift;
    my $err_counts_ref = shift;
    my $header_strs_ref = shift;
    my $cyc_hash_ref = shift;
    my $all_hash_ref = shift;
    my $all_bases_ref = shift;
    my $common_words_texts_ref = shift;
    my $cum_err_counts_by_cycle_ref = shift;
    my $max_errs_to_count_by_cycle = shift;

    my $num_cycles = (defined($cycle_counts_ref->{$data_type})
                      ? $cycle_counts_ref->{$data_type} : 0);
    my $num_unique_aligns = (defined($unique_align_counts_ref->{$data_type})
                             ? $unique_align_counts_ref->{$data_type} : 0);
    my $align_score_sum = (defined($align_score_sums_ref->{$data_type})
                           ? $align_score_sums_ref->{$data_type} : 0);
    my $num_bases = (defined($base_counts_ref->{$data_type})
                     ? $base_counts_ref->{$data_type} : 0);
    my $num_blanks = (defined($blank_counts_ref->{$data_type})
                      ? $blank_counts_ref->{$data_type} : 0);
    my $num_errs = (defined($err_counts_ref->{$data_type})
                    ? $err_counts_ref->{$data_type} : 0);

    my @read_bases = ($blank_base, @bases);
    my $read_bases_ref = \@read_bases;

    my $expected_num_cycles = get_expected_num_cycles($cycle_counts_ref,
						      $data_type,
						      $num_cycles);

    my $is_all_reads = 0;

    print_header(*OSTRM, $header_strs_ref);

    if ($input_is_saf) {
        my $title_str = title_for_data_type($data_type, $lane_num);
        print OSTRM "# $title_str\n";

	# Statistics by cycle for All Reads would simply be duplicates of
	# those for the individual reads, so suppress them using this flag.
	$is_all_reads = is_all_reads_data_type($data_type);
    }

    print_summary_stats(*OSTRM, $num_bases, $num_blanks, $num_errs,
                        $num_unique_aligns, $align_score_sum,
			$expected_num_cycles);

    if (!$is_all_reads) {
	print_errors_by_cycle(*OSTRM, $data_type, $num_cycles, $cyc_hash_ref,
			      $blank_base);
    }

    if ($num_bases == 0) {
        print STDERR
            "Warning: did not find any sequences to process ($data_type)\n";
#    exit (0);
    } elsif ($num_bases == $num_blanks) {
        print STDERR
            "Warning: no nonblank bases of sequence to process ($data_type)\n";
        next;
    }

    my $row_is_ref = 1;
    my $read_includes_blanks = 1;
    print_err_percents_by_bases(*OSTRM, $data_type, $row_is_ref,
                                $read_includes_blanks, $all_hash_ref,
                                $read_bases_ref);

    $row_is_ref = 1;
    $read_includes_blanks = 0;
    print_err_percents_by_bases(*OSTRM, $data_type, $row_is_ref,
                                $read_includes_blanks, $all_hash_ref,
                                $read_bases_ref);

    $row_is_ref = 0;
    $read_includes_blanks = 1;
    print_err_percents_by_bases(*OSTRM, $data_type, $row_is_ref,
                                $read_includes_blanks, $all_hash_ref,
                                $read_bases_ref);

    print_errs_by_base(*OSTRM, $data_type, $all_hash_ref, $read_bases_ref);

    if (!$is_all_reads) {
	print_errs_by_cycle_and_base(*OSTRM, $data_type,
				     $cyc_hash_ref, $num_cycles,
				     $read_bases_ref);
	print_err_fractions_by_cycle_and_base(*OSTRM, $data_type,
					      $cyc_hash_ref, $num_cycles,
					      $read_bases_ref);

	print_info_content_by_cycle(*OSTRM, $data_type,
				    $all_bases_ref, $cyc_hash_ref,
				    $num_cycles);
    }

    if (!$is_all_reads) {
        print_most_common_patterns(*OSTRM, $data_type,
                                   $common_words_texts_ref);

	print_likelihood_scores(*OSTRM, $data_type, $cyc_hash_ref, $num_cycles,
				$read_bases_ref);

	print_cum_errors_by_cycle(*OSTRM, $data_type,
				  $cum_err_counts_by_cycle_ref,
				  $num_cycles, $max_errs_to_count_by_cycle);
    }
}

#------------------------------------------------------------------------------

sub write_quality_line($$$$$$$)
{
    local *OSTRM = shift;
    my $qual_val = shift;
    my $curr_err_count = shift;
    my $curr_freq = shift;
    my $cum_err_count = shift;
    my $cum_freq = shift;
    my $num_bases = shift;

    # % error expected if the quality values are 100% accurate
    my $expected_err = 100 / (1 + (10 ** ($qual_val / 10)));

    printf(OSTRM 
	   "%d\t%d\t%d\t%5.2f\t%5.2f\t|\t%d\t%d\t%5.2f\t%5.2f\t|\t%5.3f\n",
	   $qual_val,
	   $cum_err_count,
	   $cum_freq,
	   100 * $cum_err_count / $cum_freq,
	   100.0 * $cum_freq / $num_bases, 
	   $curr_err_count,
	   $curr_freq,
	   100 * $curr_err_count / $curr_freq,
	   100 * $curr_freq / $num_bases,
	   $expected_err);
}

#------------------------------------------------------------------------------

sub write_quality_report($$$$$$)
{
    my $qual_val_freqs_ref = shift;
    my $errs_by_qual_val_ref = shift;
    my $lane_num = shift;
    my $tile_freqs_ref = shift;
    my $multiple_reads = shift;
    my $read_num = shift;

    my @tile_nums = sort { $a <=> $b } keys %{$tile_freqs_ref};
    my @tile_names = map { sprintf("s_%d_%04d", $lane_num, $_); } @tile_nums;

    my $tile_list = '';
    my $num_tiles = 0;

    foreach my $tile_name (@tile_names) {
        $tile_list .= $tile_name;
        $tile_list .= (((++$num_tiles % 8) == 0) ? "\n" : " ");
    }

    my $read_str = ($multiple_reads ? ":$read_num" : '');
    my @qv_data_types = ($raw_str . $read_str, $QF_str . $read_str);

    $read_str = ($multiple_reads ? "_$read_num" : '');
    my $report_fn = "s_" . $lane_num . $read_str . "_qcalreport.txt";

    open (OSTRM, ">$report_fn")
	or die "Failed to open $report_fn for writing\n";

    foreach my $data_type (@qv_data_types) {
	if ($data_type =~ /^$raw_str/) {
	    print OSTRM " ** Quality value report **\n\n";
	    print OSTRM "Tiles analysed:\n", $tile_list;
	    print OSTRM "\n$num_tiles tiles in all\n\n"; 
	    print OSTRM " ** Results for raw data **\n"; 
	} else {
	    print OSTRM "\n\n\nResults for quality filtered data\n"; 
	}

	print OSTRM "-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n";
	print OSTRM 
	    "QV\tthis QV or higher\t\t|\tthis QV only\t\t\t|\texpected\n";
	print OSTRM "\t# error\t# bases\t% error\t% total\t|"
	    . "\t# error\t# bases\t% error\t% total\t|\t% error\n";
	print OSTRM "-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n";

	my $cum_err_count = 0;
	my $cum_freq = 0;
	my $num_bases = 0;

	for (keys %{$qual_val_freqs_ref->{$data_type}}) {
	    $num_bases += $qual_val_freqs_ref->{$data_type}{$_};
	}

	for (sort {$b <=> $a} keys %{$qual_val_freqs_ref->{$data_type}}) {
	    $errs_by_qual_val_ref->{$data_type}{$_} = 0
		unless defined($errs_by_qual_val_ref->{$data_type}{$_});

	    my $curr_err_count = $errs_by_qual_val_ref->{$data_type}{$_};
	    $cum_err_count += $curr_err_count;

	    my $curr_freq = $qual_val_freqs_ref->{$data_type}{$_};
	    $cum_freq += $curr_freq;

	    write_quality_line(*OSTRM, $_,
			       $curr_err_count, $curr_freq,
			       $cum_err_count, $cum_freq,
			       $num_bases);
	}
    }

    close (OSTRM);
}

#------------------------------------------------------------------------------
# main
#------------------------------------------------------------------------------

conditionally_init_sigpipe_handler();

my $input_type = $INPUT_TYPE_ALIGN;
my $input_is_saf = 0;
my $multiple_reads = 0;
my $max_errs_to_count_by_cycle = 5;

if (@ARGV > 0) {
    if ($ARGV[0] eq "-saf-input") {
        $input_type = $INPUT_TYPE_SAF;
        $input_is_saf = 1;
    } elsif ($ARGV[0] eq "-saf-multi-input") {
        $input_type = $INPUT_TYPE_SAF_MULTI;
        $input_is_saf = 1;
        $multiple_reads = 1;
    }
}

my $blank_base = ($input_is_saf ? 'N' : '.');

my @header_strs = ("#RUN_TIME " . scalar(localtime) . "\n",
                   "#SOFTWARE_VERSION $id\n");

my $lane_num = 0;
my %tile_freqs; # actually cluster counts but used just to get tile nums
my @tile_nums = ();
my $num_reads = 1;

my %allBases;
my %hashSeq;
my %hashBlank;
my %common_words_texts;
# my %cycBases;
my %allHash;
my %cycHash;

my %cycle_counts;
my %unique_align_counts;
my %align_score_sums;
my %base_counts;
my %blank_counts;
my %err_counts;
my %cum_err_counts_by_cycle;
my %qual_val_freqs;
my %errs_by_qual_val;

read_input($input_is_saf, $blank_base, $multiple_reads, \$num_reads,
           \@header_strs, \%allBases, \%hashSeq, \%hashBlank,
           \%common_words_texts, # \%cycBases,
           \%allHash, \%cycHash,
           \%cycle_counts, \%unique_align_counts, \%align_score_sums,
           \%base_counts, \%blank_counts, \%err_counts,	   
           \$lane_num, \%tile_freqs,
	   \%cum_err_counts_by_cycle, $max_errs_to_count_by_cycle,
	   \%qual_val_freqs, \%errs_by_qual_val);

my @base_data_types = ($raw_str);

if ($input_is_saf) {
    push(@base_data_types, $QF_str);
}

my @data_types;

foreach my $base_data_type (@base_data_types) {
    for (my $read_num = 1; $read_num <= $num_reads; ++$read_num) {
        my $read_str = ($multiple_reads ? ":$read_num" : '');

        push(@data_types, $base_data_type . $read_str);
    }

    if ($multiple_reads) {
        push(@data_types, join(':', ($base_data_type, $all_reads_str)));
    }
}

if ($input_is_saf) {
    my @per_tile_data_types = @data_types;
    @data_types = ();

    write_bad_tile_files(\%tile_freqs, $lane_num, \@per_tile_data_types,
                         \@header_strs);

    @tile_nums = sort { $a <=> $b} keys %tile_freqs;

    foreach my $tile_num (@tile_nums) {
        push(@data_types,
             map { join(':', ($tile_num, $_)) } @per_tile_data_types);
    }
}

my $last_filename = '';
*OSTRM = *STDOUT;

foreach my $data_type (@data_types) {
    if ($input_is_saf) {
        my $filename = filename_for_data_type($data_type, $lane_num);

        if ($filename ne $last_filename) {
            open(OSTRM, ">$filename") or die "Failed to open $filename.\n";
            $last_filename = $filename;
        }
    }

    write_one_stat_set(*OSTRM, $input_is_saf, $blank_base,
                       $lane_num, $data_type, \%cycle_counts,
                       \%unique_align_counts, \%align_score_sums,
                       \%base_counts, \%blank_counts, \%err_counts,
                       \@header_strs, \%cycHash, \%allHash,
                       \%allBases, \%common_words_texts,
		       \%cum_err_counts_by_cycle, $max_errs_to_count_by_cycle);
}

# if ($input_is_saf) {
#    for (my $read_num = 1; $read_num <= $num_reads; ++$read_num) {
#        write_quality_report(\%qual_val_freqs, \%errs_by_qual_val,
#                             $lane_num, \%tile_freqs,
#                             $multiple_reads, $read_num);
#    }
# }

conditionally_report_sigpipes();

#------------------------------------------------------------------------------

