#!/usr/bin/env python

#  Time-stamp: <1 Aug 2007 16:24 BST update_offsets.py km>

# Copyright (c) Solexa 2005, 2006
# Author: Klaus Maisinger
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

import sys, glob, os.path, math, time
GA_LIB_LIST=[os.path.join(os.path.dirname(os.path.realpath(sys.argv[0])), "..", "lib")]
sys.path = GA_LIB_LIST + sys.path
import runfolder, common

"""Utility to extract the x-,y-offsets (in pixels) between images
taken in different colours from the header files in a given analysis
directory. Throughout the program, offsets are stored as 8-tuples
where even elements are x-coordinates and odd elements are the
y-offsets: the first element is the x-offset of the first colour, the
second element is the y-offsets of the first colour, and so forth for
the other colours."""

npixels = 1000

def read_large_offset_file(filename):
    """Read offsets from a file produced by parse_offset_scaling.py
    and return a dictionary with all offsets per tile"""
    alloffsets = {}
    allscales  = {}
    for line in open(filename).readlines():
        if line[0] == "#": # skip comments
            continue
        elements = line.split()
        # discard deblocking cycles and first cycle
        if (int(elements[4]) == 1) or  (int(elements[2]) == 1):
            continue
        tile = runfolder.FileName.tilestr(runfolder.FileName.name("s", "int", int(elements[0])-1, int(elements[1])-1))
        if (int(elements[3]) == 0): # on first image
            offsets = []
            scales =  []
        offsets += map(float, elements[5:7])
        scales  += map(float, elements[7:9])
        if (int(elements[3]) == 3) and len(offsets) == 8: # on fourth image
            # Discard zero offsets - they are unlikely to be real
            if ((max(map(math.fabs, offsets)) != 0.) and
                (max(map(math.fabs, scales))  != 0)):
                alloffsets.setdefault(tile, []).append(offsets)
                allscales.setdefault(tile, []).append(scales)
    return (alloffsets, allscales)

def write_offsets(filename, offsets, scales = None, comments = []):
    lines = []
    for comment in comments:
        lines.append("# %s\n" % comment)
    for i in range(4):
        line = "%5.2f %5.2f" % (offsets[2*i]-offsets[0],
                                offsets[2*i+1]-offsets[1])
        if scales:
            line += " %8.5f %8.5f" % (scales[2*i]-scales[0],
                                      scales[2*i+1]-scales[1])
        line += "\n"
        lines.append(line)
    fd = open(filename, "w")
    if fd:
        fd.writelines(lines)
        fd.close()

def read_offsets(filename):
    offsets = []
    scales = []
    for line in open(filename).readlines():
        if line.find('#') >= 0:
            line = line[:line.find('#')]
        elements = map(float, line.split())
        if len(elements) >= 2:
            offsets += elements[0:2]
        if len(elements) >= 4:
            scales += elements[2:4]
    return offsets, scales

def subtract(arg1, arg2, arg3 = "", ndim=8):
    """If the second argument is an offset list, subtract two lists of
    offsets. If the second argument is a scalar, subtract it from all
    elements of the first list. If a third scalar argument is given,
    subtract it from every other element in the list (the
    y-coordinates)."""
    result = [ 0. ] * ndim
    if not type(arg1).__name__ == "list":
        return result
    if type(arg2).__name__ == "list":
        for i in range(ndim):
            result[i] = arg1[i] - arg2[i]
    if type(arg2).__name__ == "float":
        for i in range(ndim):
            result[i] = arg1[i] - arg2
    if type(arg3).__name__ == "float":
        for i in range(1,ndim,2):
             result[i] = arg1[i] - arg3
    return result
    
def absmax(list):
    """Return the maximum of the absolute values in a list"""
    return max([ max(list), -min(list)])

def validate_offsets(alloffsets, default_offsets, maxoffset):
    """Given a list of offsets, subtract the default offsets and
    verify that all corrected offsets are within a maximal distance
    <maxoffset> from each other.
    Returns the best estimate of the offsets, their standard
    deviations and the total number of validated samples."""
    count = 0
    mall = [ 0. ] * 8
    sall = [ 0. ] * 8
    for moffsets in alloffsets:
        residuals = subtract(moffsets, default_offsets)
        validated = 1
        for i in range(8):
            if (math.fabs(residuals[i%8] - residuals[(i+2)%8]) > maxoffset):
                validated = 0
                #print "Failed: ", offsets
        xsum = 0.
        ysum = 0.
        for i in range(4):
            xsum += moffsets[2*i]
            ysum += moffsets[2*i+1]
        xmean = xsum/4.
        ymean = ysum/4.
        offsets = subtract(moffsets, xmean, ymean)
        if validated:
            count += 1
            for i in range(8):
                mall[i] += offsets[i]
                sall[i] += offsets[i] * offsets[i]
    if count > 0:
        for i in range(8):
            mall[i] /= count
            sall[i] = sall[i] / count - mall[i]*mall[i]
            if sall[i] >= 0.:
                sall[i] = math.sqrt(sall[i])
            else:
                sall[i] = -math.sqrt(-sall[i])
    return mall, sall, count


def total_average(tiledic):
    """Given a dictionary with the results of validate_offsets()
    for each tile, generate the overall best weighted average and the
    corresponding error on the mean"""
    allaverage = [ 0. ] * 8
    allweight = [ 0. ] * 8
    for key in tiledic.keys():
        sigma = [ 0. ] * 8
        weight = [ 0. ] * 8
        for i in range(8):
            sigma[i] = tiledic[key][1][i]
            if sigma[i] > 0.:
                # form a weighted average. the weights are given by
                # N/(sigma*sigma)
                weight[i] = tiledic[key][2] / (sigma[i] * sigma[i])
                allweight[i] += weight[i]
                allaverage[i] += weight[i] * tiledic[key][0][i]
    for i in range(8):
        if allweight[i] > 0.:
            allaverage[i] /= allweight[i]
            allweight[i] = 1./math.sqrt(allweight[i])
    return allaverage, allweight

def print_stats_line(key, num, mean, error, factor = 1.):
    s = "%9s %5i " % (key, num)
    for i in range(4):
        s+= "%5.2f*%4.2f %5.2f*%4.2f " %\
            (mean[2*i]   * factor, error[2*i]   * factor,
             mean[2*i+1] * factor, error[2*i+1] * factor)
    print s
    

def print_tile_stats(stats, factor = 1.):
    """Print a summary table listing all tiles."""
    tiles = stats.keys()
    tiles.sort()
    for key in tiles:
        diff = subtract(stats[key][0], stats[key][0][0],
                        stats[key][0][1])
        print_stats_line(key, stats[key][2], diff, stats[key][1], factor) 

def print_outliers(stats, average, maxsigma, desc = ""):
    tiles = stats.keys()
    tiles.sort()
    for key in tiles:
        values, sdevs, count = stats[key]
        if count < 2 or min(sdevs) < 1.e-6:
            continue
        sigma = [ 0. ] * 8
        for i in range(8):
            # use standard deviation
            sigma[i] = (math.fabs(stats[key][0][i] - average[i])
                        / (stats[key][1][i]))
        if max(sigma) > maxsigma:
            common.warning("More than %4.1f sigma deviation:"
                           " %5.1f %s (%s)." %
                           (maxsigma, max(sigma), key, desc))
    
    
def usage():
    print ("Usage: update_offsets.py <offset summary file>"
           " [<outputfile> [<reference offsetfile>]]")
    

if __name__ == "__main__":

    # command-line options
    if len(sys.argv) < 2:
        usage
        sys.exit(1)
    offsetsummaryfile = sys.argv[1]
    if (len(sys.argv) > 2):
        new_offset_filename = sys.argv[2]
    else:
        new_offset_filename =\
                            os.path.join(runfolder.get_data_folder(sys.argv[1]),runfolder.offsetfile)
    if (len(sys.argv) > 3):
        old_offset_filename = sys.argv[3]
    else:
        old_offset_filename = runfolder.find_offset_file(sys.argv[1],
                             instrument = common.get_instrument(sys.argv[1]))


    alloffsets = {}
    allscales  = {}
    alloffsets, allscales = read_large_offset_file(offsetsummaryfile)
        
    tiles = alloffsets.keys()
    tiles.sort()
    offset_stats = {}
    scale_stats  = {}
    for tile in tiles:
        offsets = alloffsets[tile]
        # iteratively determine best offsets by narrowing down maximum
        # allowed deviation while simultaeneously updating the best
        # estimate
        mean = [ 0. ] * 8
        for i in range(6, 0, -1):
            mean, sdev, count = validate_offsets(offsets, mean, i*1.)
        offset_stats[tile] = [ mean, sdev, count ]

        scales  = allscales [tile]
        mean = [ 0. ] * 8
        for i in range(10, 0, -1):
            mean, sdev, count = validate_offsets(scales, mean,
                                                 i/float(npixels))
        # low standard deviation is likely to come from faulty data,
        # so downweight accordingly
        if max(sdev) < 0.001 / npixels:
            count = 0
            for i in range(len(sdev)):
                sdev[i] = 1.

        scale_stats[tile] = [ mean, sdev, count ]
    

    allcount = 0
    offset_stats_counts = map(lambda x: x[2], offset_stats.values())
    if (len(offset_stats_counts) > 0):
        allcount = reduce(lambda x, y: x + y, offset_stats_counts)
    #################################################################
    # print average offset estimates and their standard deviations
    #################################################################
    print_tile_stats(offset_stats)

    #################################################################
    # print overall estimates and its error (not s.dev.)
    #################################################################
    alloffsetaverage, alloffseterror = total_average(offset_stats)
    alloffsetref = subtract(alloffsetaverage,
                            alloffsetaverage[0], alloffsetaverage[1])
    print_stats_line("all", allcount, alloffsetref, alloffseterror)

    #################################################################
    # check the maximal deviation of individual tiles from the
    # overall best estimate
    #################################################################
    print_outliers(offset_stats, alloffsetaverage, 6., "offsets")

    #################################################################
    # repeat all stats for scales
    #################################################################
    print_tile_stats(scale_stats, npixels)

    allscaleaverage, allscaleerror = total_average(scale_stats)
    allscaleref = subtract(allscaleaverage,
                            allscaleaverage[0], allscaleaverage[1])
    print_stats_line("all", allcount, allscaleref, allscaleerror, npixels)

    print_outliers(scale_stats, allscaleaverage, 6., "scaling")

    #################################################################
    # compare best estimates to values in any old offset files
    #################################################################
    tolerance = 0.3

    #print old_offset_filename
    if old_offset_filename:
        print "Comparing to offset file %s" % old_offset_filename
        oldoffsets, oldscales = read_offsets(old_offset_filename)
        if len(oldoffsets) < 8:
            common.warning("Could not parse offsets file %s."
                           % old_offset_filename)
            oldoffsets += [ 0. ] * (8 - len(oldoffsets))
        if len(oldscales) < 8:
            oldscales  += [ 0. ] * (8 - len(oldscales))
        offsetdifference = subtract(oldoffsets, alloffsetref)
        scaledifference  = subtract(oldscales,  allscaleref)
        print "  New offsets   old offsets   difference",
        print "  new scales    old scales    difference"
        for i in range(4):
            print "   %5.2f %5.2f  %5.2f %5.2f  %5.2f %5.2f" % \
                  (alloffsetref[2*i],     alloffsetref[2*i+1],
                   oldoffsets[2*i],       oldoffsets[2*i+1],
                   offsetdifference[2*i], offsetdifference[2*i+1] ),
            print "   %5.2f %5.2f  %5.2f %5.2f  %5.2f %5.2f" % \
                  (allscaleref[2*i]*npixels, allscaleref[2*i+1]*npixels,
                   oldscales[2*i]*npixels,   oldscales[2*i+1]*npixels,
                   scaledifference[2*i]*npixels,
                   scaledifference[2*i+1]*npixels )
            
        if absmax(offsetdifference) > tolerance:
            common.warning("Old offsets deviate by more than %5.2f!"
                           % tolerance)
        if absmax(scaledifference*npixels) > tolerance:
            common.warning("Old scales deviate by more than %5.2f!"
                           % tolerance)

    #################################################################
    # save best estimates to a file
    #################################################################
    
    # verify that the errors on the estimate are less than a fraction
    # of a pixel
    tolerance = 0.03
    mincount  = 10
    mincycles = 4
    maxcycles = 0
    if (len(offset_stats_counts) > 0):
        maxcycles = max(offset_stats_counts)
    if ((absmax(alloffseterror) > tolerance) or (allcount < mincount)
        or (maxcycles < mincycles)):
        print "*** Data not good enough to use new values!"
        if (absmax(alloffseterror) > tolerance):
            print "*** Tolerance %4.2f > %4.2f" % (absmax(alloffseterror), tolerance)
        if (allcount < mincount):
            print "*** Number of good tiles %i < %i" % (allcount, mincount)
        if (maxcycles < mincycles):
            print "*** Number of cycles %i < %i" % (maxcycles, mincycles)
        print "Reverting to original offsets..."
        alloffsetaverage = oldoffsets
        allscaleaverage  = oldscales

        
    if new_offset_filename: 
        print "Updating offsets file %s" % new_offset_filename
        try:
            write_offsets(new_offset_filename, alloffsetaverage,
                          allscaleaverage,
                          comments =
                          [ "Default offsets generated on %s by %s"
                            % (time.ctime(), runfolder.get_login()),
                            "%s" % sys.argv[1]])
        except:
            common.error("Failed to write %s" % new_offset_filename)
