import re

class CycleSelector:
    def __init__(self, cycleSelectionString):
        """Convert the comma-separated list of cycle selections into a list of
        instances of CycleSelection. The list is then sorted, to ensure correct
        indexing of the reads"""

        cycleSelectionString = cycleSelectionString.lower()
        if filter(lambda s: s == 'auto', cycleSelectionString.split(',')) and len(cycleSelectionString.split(',')) > 1:
            raise ValueError("full auto cycle selection can't be combined with other selection options: --cycle=%s" % cycleSelectionString)
        m= re.match("^.*auto(?P<suffix>.*)$", cycleSelectionString)
        if m and m.group("suffix"):
            raise ValueError("auto cycle selection must be the last selection element: %s" % cycleSelectionString)
        self.cycleSelectionList = map(lambda s: CycleSelection(s), cycleSelectionString.split(','))
        if not self.cycleSelectionList:
            raise ValueError("no cycle selection provided: %s" % cycleSelectionString)
        if len(self.cycleSelectionList) > 1:
            self.cycleSelectionList.sort(lambda s1, s2: s1.firstCycle - s2.firstCycle)

    def isAutoCycleLast(self):
        return self.isAutoCycleAll() or self.cycleSelectionList[-1].autoLast
    
    def isAutoCycleAll(self):
        return self.cycleSelectionList[0].autoAll

    def getFirstCycle(self):
        if self.cycleSelectionList[0].autoAll:
            return 1
        else:
            return self.cycleSelectionList[0].firstCycle

    def isContiguous(self):
        if self.isAutoCycleAll():
            return True
        previous = 0
        for cycleSelection in self.cycleSelectionList:
            if cycleSelection.firstCycle != previous + 1:
                return False
            if cycleSelection.autoLast:
                return True
            previous = cycleSelection.lastCycle
        
        return True
    
    def getCycleList(self, imagingReadList, useLastCycle = True):
        """Get the list of cycles."""
        readList = self.getImagingReadSelection(imagingReadList, useLastCycle)
        cycleList = []
        for read in readList:
            cycleList += range(read.FirstCycle, read.LastCycle + 1)
        return cycleList

    def getReadList(self, imagingReadList, useLastCycle = True):
        """Build the complete read list for the given imaging read list and the
        current cycle selection."""
        from common import Read
        readList = []
        useLastCycleList = [True for i in range (0,len(imagingReadList))]
        useLastCycleList[-1] = useLastCycle
        for i in range(0, len(imagingReadList)):
            imagingRead = imagingReadList[i]
            cycleList = []
            for cycleSelection in self.cycleSelectionList:
                cycleList += cycleSelection.getCycleList(imagingRead, useLastCycleList[i])
            if cycleList:
                firstCycle = 1
                lastCycle = len(cycleList)
                if readList:
                    firstCycle += readList[-1].LastCycle
                    lastCycle  += readList[-1].LastCycle
                readList.append(Read(imagingRead.RunFolder, len(readList) + 1, firstCycle, lastCycle))
        return readList
                
        
    def getReadListx(self, imagingReadList, useLastCycle = True):
        """Build the complete read list for the given imaging read list and the
        current cycle selection."""
        readList = self.getImagingReadSelection(imagingReadList, useLastCycle)
        # shift the cycles of the logical reads to get a contiguous sequence of cycles
        firstCycle = 1
        for read in readList:
            read.LastCycle -= read.FirstCycle - firstCycle
            read.FirstCycle = firstCycle
            firstCycle = read.LastCycle + 1
        return readList
    
    def getImagingReadSelection(self, imagingReadList, useLastCycle):
        """Get the intersection of the imaging reads with the cycle selection"""
        readList = []
        useLastCycleList = [True for i in range (0,len(imagingReadList))]
        useLastCycleList[-1] = useLastCycle
        for i in range(0, len(imagingReadList)):
            # for imagingRead in imagingReadList:
            imagingRead = imagingReadList[i]
            readList += self.getSingleReadList(imagingRead, 1 + len(readList), useLastCycleList[i])
        return readList
        
    def getSingleReadList(self, imagingRead, index, useLastCycle = True):
        """Map the selection onto the given imaging read. Full auto will create
        a read from the first cycle to the last available cycle. Partial auto,
        when the first selected cycle is within the imagingRead cycle range,
        will create a read from the first selected cycle to the last available
        cycle. Manual selection will create a read for each manually selected
        interval. It is an error if there are overlaps between manual selection
        and partial auto selection. The last available cycle is either the last
        cycle of the imaging cycle (if useLastCycle == True) or the previous
        cycle. This allows intermediate analysis, where the last cycle would
        be incomplete"""
        readList = []
        for cycleSelection in self.cycleSelectionList:
            readList += cycleSelection.getReadList(imagingRead, index + len(readList), useLastCycle)
        return readList

class CycleSelection:
    def __init__(self, s):
        """Parse the cycle selection string. It can be i) 'auto', ii) a single digit,
        iii) two digits separated by a '-', iv) a digit followed by '-auto'."""
        s = s.lower()
        self.firstCycle = 0
        self.lastCycle  = 0
        self.autoAll = (s == 'auto')
        self.autoLast  = self.autoAll
        if not self.autoAll:
            bounds = s.split('-')
            if not bounds[0] or 2 == len(bounds) and not bounds[1] or 2 < len(bounds):
                raise ValueError("Invalid cycle selector: %s" % s)
            self.firstCycle = int(bounds[0])
            if self.firstCycle < 1:
                raise ValueError("First cycle must be strictly greater than 0: %s" % s)
            if len(bounds) < 2:
                self.lastCycle = self.firstCycle
            else:
                self.autoLast = (bounds[1].lower() == 'auto')
                if not self.autoLast:
                    self.lastCycle = int(bounds[1])
                    if self.firstCycle > self.lastCycle:
                        raise ValueError("First cycle must be less than or equal to last cycle: %s" % s)

    def getCycleList(self, imagingRead, useLastCycle):
        """Return the list of cycles from the given imaging read which are in the current selection"""
        firstCycle = imagingRead.FirstCycle
        lastCycle  = imagingRead.LastCycle
        if not useLastCycle:
            lastCycle -= 1
        if not self.autoAll:
            firstCycle = max(firstCycle, self.firstCycle)
            if not self.autoLast:
                lastCycle = min(lastCycle, self.lastCycle)
        if firstCycle <= lastCycle:
            return range(firstCycle, lastCycle+1)
        else:
            return []
        
    def getReadList(self, imagingRead, index, useLastCycle):
        """Return the read list for the given imaging read, according to the
        current cycle selection"""
        from common import Read
        runFolder  = imagingRead.RunFolder
        # initialize first and last cycles for the case of auto cycle
        firstCycle = imagingRead.FirstCycle
        lastCycle  = imagingRead.LastCycle
        if not useLastCycle:
            lastCycle -= 1
        if not self.autoAll:
            firstCycle = max(firstCycle, self.firstCycle)
            if not self.autoLast:
                lastCycle = min(lastCycle, self.lastCycle)
        if firstCycle <= lastCycle:
            return [Read(runFolder, index, firstCycle, lastCycle)]
        else:
            return []
                   
