from common import Read
import operator


class ImagingProtocol:
    """The list of ImagingFolders. This would typically be the list of top-level
    folders where the cycle (or deblock) subfolders can be found, but the actual
    management of the file system paths is left to the user."""
    def __init__(self):
        self.folderName = ""
        self.imagingFolderList = []
        
    def addImagingFolder(self, folderName, cycleFolderList, restartCycleList = []):
        """Add an imaging folder to the protocol, at the given name (path or any
        convenient access key for the client code) and the list of cycles (again,
        the path of the subdirectories, or at the discretion of the user code).
        The indices in restartCycleList are the cycle numbers (starting from 1) where
        new reads start. These values must be strictly greater than 1. They must
        be sorted. The values outside the cycle range for the current folder are
        ignored."""
        firstCycle = 1 + self.getNumberOfCycles()
        self.imagingFolderList.append(ImagingFolder(folderName, cycleFolderList, firstCycle, restartCycleList))
        
    def getImagingReadList(self):
        """Build the imaging reads from the current protocol information"""
        imagingReadList = []
        firstCycle = 1
        for folder in self.imagingFolderList:
            imagingReadList += folder.getImagingReadList(len(imagingReadList)+1, firstCycle);
            firstCycle = imagingReadList[-1].LastCycle + 1
        return imagingReadList
    
    def getNumberOfCycles(self):
        return reduce(operator.add, map(lambda imagingFolder: imagingFolder.getNumberOfCycles(), self.imagingFolderList), 0)

    def getCycleFolderList(self, folderName, indexList = []):
        found = False
        firstCycle = 1
        for folder in self.imagingFolderList:
            if folder.folderName == folderName:
                found = True
                break
            firstCycle += folder.getNumberOfCycles()
        if not found:
            raise ValueError("%s not found" % (folderName))
        cycleFolderList = folder.getCycleFolderList()
        if indexList:
            indexList = filter(lambda v: v >= firstCycle and v < firstCycle + len(cycleFolderList), indexList)
            return map(lambda i: cycleFolderList[i - firstCycle], indexList)
        else:
            return cycleFolderList

class ImagingFolder:
    """An ordered list of consecutive imaging cycles."""
    def __init__(self, folderName, cycleFolderList, firstCycle, restartCycleList):
        """Split the cycle folder ist into one or more ImagingCycles,
        according to the firstCycle and to the restart positions indicated by
        restartCycleList, if any."""
        self.checkParameters(folderName, cycleFolderList, firstCycle, restartCycleList)
        self.folderName = folderName
        boundaries = filter(lambda i: i > 0 and i < len(cycleFolderList), map(lambda cycle: cycle - firstCycle, restartCycleList))
        fullStartList = [0] + boundaries + [len(cycleFolderList)]
        self.imagingCycleList = map(lambda(i): ImagingCycle(cycleFolderList[fullStartList[i]:fullStartList[i+1]]),
                                    range(0, len(fullStartList)-1))

    def getCycleFolderList(self):
        cycleFolderList = []
        for imagingCycle in self.imagingCycleList:
            cycleFolderList += imagingCycle.cycleFolderList
        return cycleFolderList
    
    def getImagingReadList(self, index, firstCycle):
        """delegate the building of the ImagingReads to the Imaging cycles. Keeps track of the index and of the first read."""
        imagingReadList = []
        for imagingCycle in self.imagingCycleList:
            imagingReadList.append(imagingCycle.getImagingRead(self.folderName, index, firstCycle))
            index += 1;
            firstCycle = imagingReadList[-1].LastCycle + 1
        return imagingReadList

    def getNumberOfCycles(self):
        return reduce(operator.add, map(lambda imagingCycle: imagingCycle.getNumberOfCycles(), self.imagingCycleList))
        
    def checkParameters(self, folderName, cycleFolderList, firstCycle, restartCycleList):
        if not folderName:
            raise ValueError("A folder name is required")
        if not cycleFolderList:
            raise ValueError("A folder list is required")
        previous = 1;
        for v in restartCycleList:
            if previous < v: # and v < len(cycleFolderList):
                previous = v
            else:
                if v <= previous:
                    message = "Out of sequence restart positions: %i >= %i" % (previous, v)
                #if v >= len(cycleFolderList):
                #    message = "Invalid restart position: %i >= %i" % (v, len(cycleFolderList))
                raise ValueError(message)

class ImagingCycle:
    """An ordered list of subfolder names representing a range of consecutive
    cycles in an imaging protocol. These """
    def __init__(self, cycleFolderList):
        self.cycleFolderList = cycleFolderList;
    def getImagingRead(self, folderName, index, firstCycle):
        return Read(folderName, index, firstCycle, firstCycle + len(self.cycleFolderList) - 1)

    def getNumberOfCycles(self):
        return len(self.cycleFolderList)

