#!/usr/bin/env python
#
# Copyright (c) 2005, 2006 Solexa
# Author: Klaus Maisinger
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

import os, os.path, types, sys, shutil
import runfolder, goat_pipeline, bustard, gerald, common

"""Generate makefiles for the image analysis and base-calling software."""

GA_DATA_DIR           = os.path.join(os.path.dirname(os.path.realpath(sys.argv[0])), "..", "..", "share")
makefilepath          = "Makefiles"
num_calibration_tiles_per_lane = 6

def get_max_cycle(value, ncolours=4):
    if type(value) == types.IntType:
        return value
    if type(value) == types.ListType:
        return len(value)/ncolours
    raise "Wrong type for value in makefile.get_max_cycle()"

def get_var_tile(tilestr):
    sample, lane, tile = runfolder.FileName.split_tile(tilestr)
    return tilestr.replace(sample, "$(sample%i)" % (lane+1))

def get_lane_tiles(tiles):
    lanedic = {}
    for tilestr in tiles:
        sample, lane, tile = runfolder.FileName.split_tile(tilestr)
        lanedic.setdefault(lane, []).append(tilestr)
    for lane in lanedic.keys():
        lanedic[lane].sort()  
    return lanedic

def generate_tile_sampling(tiles, nlanetotal = 3):
    """Subsample a list of tiles, using only <nlanetotal> tiles per lane"""
    newtiles = []
    lanedic = get_lane_tiles(tiles)
    for lane in lanedic.keys():
        maxtiles = nlanetotal
        if maxtiles > len(lanedic[lane]):
            maxtiles = len(lanedic[lane])
        factor = len(lanedic[lane])/maxtiles
        for i in range(maxtiles):
            newtiles.append(lanedic[lane][i*factor])
    return newtiles
    

def generate_makefile(tiledic, analysis_folder, FileNameClass,
                      run_params, image_params,
                      matrix_cycles,
                      old_analysis_folder):
    filename = os.path.join(analysis_folder, "Makefile")
    lines = []
    tiles = tiledic.keys()
    tiles.sort()
    lanes = runfolder.get_lane_samples(tiles)
    if run_params.IparFlag:
        ipar_short_option  = "-I"
    else:
        ipar_short_option  = ""

    lines.append("# Makefile auto-generated by makefile.py (version %s)\n# %s\n\n"
                 % (goat_pipeline.image_software_version,
                    runfolder.get_date_time()))
    image_exe = runfolder.get_executable(goat_pipeline.image_executable)
    installation_dir = os.path.split(image_exe)[0]
    lines.append("installationdir:=%s\n" %  installation_dir)
    lines.append("image_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.image_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("offsets_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.offset_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("offsets2_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.offset2_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("sizes_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.sizes_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("sizes2_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.sizes2_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("sizes3_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.sizes3_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("matrix1_exe:=%s %s\n" %
             (runfolder.get_executable(goat_pipeline.matrix1_executable)\
                 .replace(installation_dir, "$(installationdir)"), ipar_short_option))
    #lines.append("matrix2_exe:=%s\n" %
    #         runfolder.get_executable(goat_pipeline.matrix2_executable)\
    #             .replace(installation_dir, "$(installationdir)"))
    lines.append("matrix3_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.matrix3_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("matrix4_exe:=%s\n\n" %
             runfolder.get_executable(goat_pipeline.matrix4_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("HAVE_BASH3:=$(shell sh --version | grep \"bash, version [345]\.\")\n"
                 "ifneq ($(HAVE_BASH3),)\n"
                 "PIPECMD:=set -o pipefail;\n"
                 "endif\n")

    lines.append("PASTE=paste\n"
                 "CP=cp\n"
                 "SAFEPIPETARGET= $@.tmp && mv $@.tmp $@\n\n")
    lines.append("analysisdir:=%s\n" % analysis_folder)
    for lane in range(runfolder.maxlane):
        lines.append("sample%i:=%s\n" % (lane+1, lanes.get(lane,
                                                 runfolder.defaultsample)))
    lines.append("analysis_subfolder:=%s\n" %
                 runfolder.analysissubdir)
    lines.append("adir2:=.\n")
    lines.append("lanes:=%s\n" % " ".join(map(lambda x: "$(sample%i)_%i" % (x+1, x+1),
                                              lanes.keys())))
    lines.append("lanefile:=lanes.txt\n"
                 "tilefile:=tiles.txt\n"
                 "calibrationfile:=tiles_cal.txt\n")
    folders = []
    for lane in lanes.keys():
        folders.append(runfolder.get_lane_subfolder("", "$(analysis_subfolder)",
                                                    "", lane))
                                              
        lines.append("analysisdir%i:=%s\n" % (lane+1, folders[-1]))
    lines.append("analysisdirs:=%s\n" % " ".join(folders))
    ncolours = 4
    allmaxcycle = get_max_cycle(tiledic.values()[0])
    lines.append("maxcycle:=%02i\n" % allmaxcycle)
    lines.append("cycles:=%s\n" % " ".join(map(lambda x: "%02i" % (x+1),
                                           range(allmaxcycle))))
    lines.append("subfolders:=$(wildcard %s*)\n" %
                 bustard.basecaller_software)
    lines.append("logfile:=$(analysis_subfolder)/$(analysis_subfolder).log\n")
    lines.append("focusdir:=%s\n" % "Focus")
    lines.append("sizedir:=%s\n" % runfolder.sizedir)
    lines.append("offsetdir:=%s\n" % runfolder.offsetdir)
    lines.append("offsetfile:=%s\n" % "offsets.txt")
    lines.append("offsetstats:=$(offsetdir)/offset_summary.txt\n\n")
    lines.append("calibrationflag?=false\n")
    if image_params.AutoOffsetFlag:
        lines.append("autooffsetflag?=true\n")
    else:
        lines.append("autooffsetflag?=false\n")
    if image_params.AutoSizeFlag:
        lines.append("autosizeflag?=true\n\n")
    else:
        lines.append("autosizeflag?=false\n\n")
    if old_analysis_folder:
        lines.append("loadfile:= -loadsuffix=.old\n")
    else:
        lines.append("loadfile:=\n")
    
    lines.append("tempdir:=%s\n" % runfolder.tempdir)
    lines.append("inputdir:=%s\n" % runfolder.inputdir)

    lines.append("default_offsets:=%s\n" % "default_offsets.txt")
    if ((image_params.OffsetFile != image_params.DataOffsetFile)
        and (image_params.OffsetFile != image_params.InstrumentOffsetFile)):
        lines.append("input_offsets:=%s\n" % image_params.OffsetFile)
    lines.append("data_offsets:=../$(default_offsets)\n")
    if image_params.InstrumentOffsetFile:
        lines.append("instrument_offsets:=%s\n" %
                     image_params.InstrumentOffsetFile)
    lines.append("offsets:=$(default_offsets)\n")

    lines.append("compression:=%s\n" % run_params.Compression)
    lines.append("compressionsuffix:=%s\n" % run_params.CompressionSuffix)
    lines.append("fwhm:=%f\n" %      image_params.Fwhm);
    lines.append("dist:=%f\n" %      image_params.RemappingDistance)
    lines.append("threshold:=%f\n" % image_params.Threshold)
    lines.append("image_flags:=%s\n" % image_params.ImageFlags)

    # set up look up table for deblock string and maxcycle per tile
    # also find lowest common denominator for deblock length
    difflengthtiles = {}
    diffdeblocktiles = {}
    for tilestr in tiles:
        sample, lane, tile = runfolder.FileName.split_tile(tilestr)
        firstcycle = 0
        ncycles, ncolour, deblock = goat_pipeline.get_cycle_params(\
            tiledic[tilestr][ncolours*firstcycle:], FileNameClass)
        
        tilestr2 = get_var_tile(tilestr)
        maxcycle = get_max_cycle(tiledic[tilestr])
        diffdeblocktiles.setdefault(deblock, []).append(tilestr2)
        if maxcycle != allmaxcycle:
            difflengthtiles.setdefault(maxcycle, []).append(tilestr2)
    mindeblock = diffdeblocktiles.keys()[0]
    for deblock in diffdeblocktiles.keys():
        if (len(deblock) < len(mindeblock)):
            mindeblock = deblock

    lines.append("matrixdir:=%s\n" % runfolder.matrixdir)
    # identify cycles to use for matrix estimation (skip deblock cycles)
    if (matrix_cycles == None or len(matrix_cycles) == 0):
        matrix_cycles = [0]
    
    for i in matrix_cycles:
        lines.append("matrixsummary%i:=s_%02i_summary.txt\n" %
                     (i+1, i+1))

    iparSuffix = "";
    if run_params.IparFlag:
        lines.append("iparsuffix:=.p\n")
        iparSuffix = "$(iparsuffix)"

    lines.append("jnlsuffix:=_jnl.xml\n"
                 "intsuffix:=_int.txt\n"
                 "tmpsuffix:=.tmp\n")
    lines.append("fullintsuffix:=$(intsuffix)%s$(compressionsuffix)\n\n" % iparSuffix)
    lines.append("# Hooks\ncmdf1?=echo \ncmdf2?=echo \n"
                 "cmdf3?=echo \ncmdf4?=echo \n")

    clean_targets = map(lambda x: "clean%i" % (x+1), lanes.keys()) 
    lines.append("\nvpath %%$(jnlsuffix) %s\n" % runfolder.inputdir)
    lines.append("vpath %_qcm.xml $(analysisdirs)\n")
    lines.append("vpath %_mat.txt $(matrixdir)\n")
    lines.append("vpath %_matrix.txt $(matrixdir)\n")
    lines.append("vpath %_summary.txt $(matrixdir)\n")
    lines.append("vpath %offsets.txt $(offsetdir)\n")
    lines.append("\n"
                 ".PHONY: all allmsg $(lanes) update_offsets matrix focus recursive $(subfolders)\n"
                 ".PHONY: clean clean_intermediate clean_matrix clean_offsets cleanfocus %s\n" % (" ".join(clean_targets)))
    lines.append(".PRECIOUS: %$(fullintsuffix)\n"
                 ".PRECIOUS: %_mat.txt\n"
                 ".PRECIOUS: %_matrix.txt\n"
                 ".PRECIOUS: %_offsets.txt\n"
                 ".PRECIOUS: %_summary.txt\n"
                 ".PRECIOUS: $(default_offsets)\n")
    lines.append(".INTERMEDIATE: calibration.1 calibration.2 "
                 "calibration.3 calibration.4 calibration.5\n\n")
    lines.append("all: ")
    if run_params.IparFlag:
        lines.append(" iparmsg matrix\n")
    else:
        lines.append(" allmsg $(lanes) update_offsets matrix $(sizedir)/sizes.xml\n")
    lines.append("\tfind $(analysisdirs) -regex \".*_\\(idx\\|int\\|nse\\)\\.txt\" | xargs $(RM) \n")
    lines.append("\tfind $(analysisdirs) -regex \".*_clu.txt\" | xargs -i mv {} $(tempdir) \n")
    lines.append("\t@echo Image analysis has completed successfully. && \\\n"
                 "\ttouch finished.txt && \\\n"
                 "\t$(cmdf3)\n")
    if run_params.IparFlag:
        lines.append("iparmsg:\n\t@echo Finishing analysis for lanes $(lanes)\n")
    else:
        lines.append("allmsg:\n\t@echo Analysing lanes $(lanes)\n")
    lines.append("\n")

    ##################################################################
    # generate rules for each lane
    ##################################################################
    for lane2 in range(runfolder.maxlane):
        tileids = []
        for tilestr in tiles:
            sample, lane, tile = runfolder.FileName.split_tile(tilestr)
            if lane2 == lane:
                tileids.append(get_var_tile(tilestr))
        lines.append("l%i_tiles:=%s\n" % (lane2+1, " ".join(tileids)))
    lines.append("tiles:=%s\n\n" %
                 " ".join(map(lambda x: "$(l%i_tiles)" % (x+1),
                              lanes.keys())))
    calibration_tiles = generate_tile_sampling(tiles, num_calibration_tiles_per_lane)
    lines.append("calibrationtiles:=%s\n\n" %
                 " ".join(calibration_tiles))
    for lane in range(runfolder.maxlane):
        lines.append("l%i_intfiles:=$(addsuffix $(fullintsuffix),$(l%i_tiles))\n"
                     % (lane+1, lane+1))
    lines.append("intfiles:=$(addsuffix $(fullintsuffix),$(tiles))\n\n")
    
    lines.append("ifeq ($(calibrationflag),true)\n")
    for lane in range(runfolder.maxlane):
        lines.append("l%i_intfiles:=$(addsuffix $(fullintsuffix),$(filter $(sample%i)_%i%%,$(calibrationtiles)))\n"
                     % (lane+1, lane+1, lane+1))
    lines.append("tilefile:=$(calibrationfile)\n"
                 "offsets:=$(default_offsets).tmp\n"
                 "default_offsets:=$(default_offsets).tmp\n"
                 "endif\n\n")

    if not run_params.IparFlag:
        lines.append("ifeq ($(calibrationflag),false)\n"
                     "$(intfiles): offsets_finished.txt\n"
                     "endif\n\n")
        
    lines.append("$(lanefile):\n"
                 "\techo \"$(tiles)\" > $(tilefile) && \\\n"
                 "\techo \"$(calibrationtiles)\" > $(calibrationfile) && \\\n"
                 "\techo \"$(lanes)\" > $(SAFEPIPETARGET)\n\n")

    ##################################################################
    # generate rules for int.txt files
    ##################################################################
    if not run_params.IparFlag:
        lines.append("# Generate intensity files\n")
        keys = difflengthtiles.keys()
        keys.sort()
        for cycle in keys:
            lines.append("%s: maxcycle:=%i\n" %
                         (" ".join(map(lambda x: "%$(fullintsuffix)" % x,
                                   difflengthtiles[cycle])), cycle))
        keys = diffdeblocktiles.keys()
        keys.sort()
        for deblock in keys:
            lines.append("%s: deblock:=%s\n" %
                         (" ".join(map(lambda x: "%s$(fullintsuffix)" % x,
                                       diffdeblocktiles[deblock])), deblock))

        lines.append("%$(fullintsuffix): tile=$(subst $(fullintsuffix),,$@)\n")
        for lane in lanes.keys():
            lines.append("$(l%i_intfiles): adir:=$(analysisdir%i)\n" %
                         (lane+1, lane+1)) 
        for lane in lanes.keys():
            lines.append("$(l%i_intfiles): params:=-params=params%i.xml\n" %
                         (lane+1, lane+1)) 
        lines.append("%%$(fullintsuffix): %%$(jnlsuffix) $(lanefile)\n"
                     "\t$(cmdprefix) $(image_exe) Input/$(tile)$(jnlsuffix) -fwhm=$(fwhm) -dist=$(dist) -threshold=$(threshold) -nc=0 -logfile=$(logfile) -prefix=$(adir)/ -prefix2=$(adir2)/ -offsets=$(offsets) -tile=$(tile) -deblocked=$(deblock) $(loadfile) -compression=$(compression) -savesuffix=.tmp $(params) $(image_flags) %s && \\\n"
                     "\tmv $(tile)$(intsuffix)$(tmpsuffix)$(compressionsuffix) $@ \n" %
                     " ".join(map(lambda x: "-matrix=%i" % (x+1), matrix_cycles)))
 
    ##################################################################
    # calibration procedure: generate default offsets / cluster sizes
    ##################################################################
    if not run_params.IparFlag:
        lines.append("\n# Calibrate default offsets and sizes\n")
        lines.append("calibration.1: $(lanefile)\n")
        lines.append("\techo Auto-generating default offsets... && \\\n")
        if (os.path.exists(image_params.OffsetFile)
            and (image_params.OffsetFile != image_params.DataOffsetFile)
            and (image_params.OffsetFile != image_params.InstrumentOffsetFile)):
            lines.append("\t$(CP) $(input_offsets) $(default_offsets).tmp\n")
        elif os.path.exists(image_params.DataOffsetFile):
            lines.append("\t$(CP) $(data_offsets) $(default_offsets).tmp\n")
        elif os.path.exists(image_params.InstrumentOffsetFile):
            lines.append("\t$(CP) $(instrument_offsets) $(default_offsets).tmp\n")

        lines.append("calibration.2: calibration.1\n"
                     "ifeq ($(autooffsetflag),true)\n"
                     "  ifeq ($(autosizeflag),true)\n"
                     "\t$(MAKE) calibrationflag=true update_offsets $(sizedir)/sizes.xml\n"
                     "  else\n"
                     "\t$(MAKE) calibrationflag=true update_offsets\n"
                     "  endif\n"
                     "else\n"
                     "  ifeq ($(autosizeflag),true)\n"
                     "\t$(MAKE) calibrationflag=true update_offsets $(sizedir)/sizes.xml\n"
                     "  else\n"
                     "\techo \"No automatic calibration used.\"\n"
                     "  endif\n"
                     "endif\n")
        lines.append("calibration.3: calibration.2\n"
                     "ifeq ($(autooffsetflag),true)\n"
                     "\techo \"Generating offsets.\" && \\\n"
                     "\t$(CP) $(data_offsets) $(default_offsets).tmp\n"
                     "endif\n")
        lines.append("calibration.4: calibration.3\n"
                     "ifeq ($(autosizeflag),true)\n")
        for lane in lanes.keys():
            lines.append("\techo \"Generating size parameters.\" && \\\n")
            lines.append("\t$(cmdprefix) $(sizes3_exe) $(sizedir)/$(sample%i)_%i_sizes.xml > params%i.xml.tmp && "
                         "mv params%i.xml.tmp params%i.xml\n" %
                         (lane+1, lane+1, lane+1, lane+1, lane+1))
        lines.append("endif\n")
        lines.append("calibration.5: calibration.4\n"
                     "ifeq ($(autooffsetflag),true)\n"
                     "\t$(MAKE) clean_lanes clean_offsets\n"
                     "endif\n"
                     "ifeq ($(autosizeflag),true)\n"
                     "\t$(MAKE) clean_lanes clean_sizes\n"
                     "endif\n")
        lines.append("$(default_offsets): calibration.5\n")
        lines.append("\tmv $(default_offsets).tmp $(default_offsets)\n")
        lines.append("offsets_finished.txt: $(default_offsets)\n"
                     "\ttouch $@ && $(cmdf1)\n")

    ##################################################################
    # estimate cluster sizes
    ##################################################################
    if not run_params.IparFlag:
        lines.append("\n# Generate cluster sizes\n")
        lanetargets = []
        for lane in lanes.keys():
            lanetargets.append("$(sizedir)/$(sample%i)_%i_sizes.txt" %
                               (lane+1, lane+1))
            lines.append("%s: $(l%i_intfiles)\n" % 
                         (lanetargets[-1], lane+1))
            lines.append("\t$(PIPECMD) $(cmdprefix) cat $(tilefile) | sed 's/ /\\n/g' | "
                         "grep $(sample%i)_%i | "
                         "perl -lane '{ printf (\"$(analysisdir%i)/$${F[0]}_01_qcm.xml\\n\");}' | " 
                         "$(sizes_exe) - > $(SAFEPIPETARGET)\n" %
                         (lane+1, lane+1, lane+1))
        lines.append("$(sizedir)/sizes.txt: %s\n" %
                     " ".join(lanetargets))
        lines.append("\tcat $^ > $(SAFEPIPETARGET)\n")
        lines.append("$(sizedir)/%_sizes.xml: $(sizedir)/%_sizes.txt\n"
                     "\t$(cmdprefix) $(sizes2_exe) $< > $(SAFEPIPETARGET)\n")
        lanetargets = []
        for lane in lanes.keys():
            lanetargets.append("$(sizedir)/$(sample%i)_%i_sizes.xml" %
                               (lane+1, lane+1))
        lines.append("$(sizedir)/sizes.xml: $(sizedir)/sizes.txt %s\n"
                     "\t$(cmdprefix) $(sizes2_exe) $< > $(SAFEPIPETARGET)\n"
                     % " ".join(lanetargets))

    ##################################################################
    # update offsets
    ##################################################################
    if not run_params.IparFlag:
        lines.append("\n# Generate offsets\n")
        for lane in lanes.keys():
            lines.append("$(offsetdir)/$(sample%i)_%i_offsets.txt:"
                         " $(l%i_intfiles)\n" % 
                         (lane+1, lane+1, lane+1))
            lines.append("\t$(PIPECMD) $(cmdprefix) cat $(tilefile) | sed 's/ /\\n/g' | "
                         "grep $(sample%i)_%i | "
                         "perl -lane '{ foreach $$i (qw/ $(cycles) /) { printf (\"$(analysisdir%i)/$${F[0]}_%%s_qcm.xml\\n\", $$i);}}' | " 
                         "$(offsets_exe) - > $(SAFEPIPETARGET)\n" %
                         (lane+1, lane+1, lane+1))

        lines.append("\n$(offsetdir)/$(offsetfile): %s\n" %
                     " ".join(map(lambda x:
                                  "$(offsetdir)/$(sample%i)_%i_offsets.txt" %
                                  (x+1, x+1), lanes.keys())))
        lines.append("\tcat $^ > $(SAFEPIPETARGET)\n")

        lines.append("\nupdate_offsets: $(offsetfile)\n")
        if (image_params.InstrumentOffsetFile
            and not os.path.exists(image_params.DataOffsetFile)):
            lines.append("\t$(CP) $(instrument_offsets) $(data_offsets) && \\\n")
        lines.append("\t$(cmdprefix) $(offsets2_exe) $(offsetdir)/$(offsetfile) $(data_offsets) $(default_offsets) > $(offsetstats) && \\\n")
        lines.append("\tcat $(offsetstats) && \\\n")
        if image_params.InstrumentOffsetFile:
            lines.append("\t$(CP) $(data_offsets) $(instrument_offsets)\n")
        else:
            lines.append("\techo Offsets updated\n")
        

    ##################################################################
    # matrix estimation
    ##################################################################
    # matrix estimation is performed on the first two cycles.
    # need to make sure that each tile has sequenced two cycles.
    lines.append("\n# Matrix estimation targets\n")
    # matrix targets by lane
    for lane2 in lanes.keys():
        for i in matrix_cycles:
            lines.append("l%i_%02i_matfiles:=$(addprefix $(matrixdir)/,$(addsuffix _%02i_mat.txt,$(l%i_tiles)))\n"
                         % (lane2+1, i+1, i+1, lane2+1))
        lines.append("l%i_matfiles:=%s\n"
                     % (lane2+1,
                        " ".join(map(lambda x: "$(l%i_%02i_matfiles)" %
                                     (lane2+1, x+1), matrix_cycles))))

    # main matrix target
    matrixsummary_targets = []
    for i in matrix_cycles:
        for lane2 in lanes.keys():
            target = "$(matrixdir)/$(sample%i)_%i_%02i_summary.txt" \
                     % (lane2+1, lane2+1, i+1)
            matrixsummary_targets.append(target)
            lines.append("%s: $(l%i_%02i_matfiles)\n"
                         % (target, lane2+1, i+1))
    lines.append("%s: \n" % " ".join(matrixsummary_targets))
    lines.append("\t$(RM) $@ && \\\n"
                 "\tfor f in $^; do \\\n"
                 "\t    echo $$f && \\\n"
                 "\t    $(cmdprefix) $(matrix3_exe) $$f >> $@.tmp; \\\n"
                 "\tdone && \\\n"
                 "\tmv $@.tmp $@\n\n" )
    for i in matrix_cycles:
        lines.append("$(matrixdir)/$(matrixsummary%i): %s\n"
                     % (i+1,
                        " ".join(map(lambda x:
                                     "$(matrixdir)/$(sample%i)_%i_%02i_summary.txt"
                                     % (x+1, x+1, i+1), lanes.keys()))))
        lines.append("\tcat $^ > $(SAFEPIPETARGET)\n")

    for lane2 in lanes.keys():
        for i in matrix_cycles:
            lines.append("$(matrixdir)/$(sample%i)_%i_%02i_matrix.txt: "
                         "$(matrixdir)/$(sample%i)_%i_%02i_summary.txt\n" %
                         (lane2+1, lane2+1, i+1, lane2+1, lane2+1, i+1))
            lines.append("\t$(cmdprefix) $(matrix4_exe) $^ "
                         "> $(SAFEPIPETARGET)\n")
    for i in matrix_cycles:
        lines.append("$(matrixdir)/s_%02i_matrix.txt: "
                     "$(matrixdir)/$(matrixsummary%i)\n" %
                     (i+1, i+1))
        lines.append("\t$(cmdprefix) $(matrix4_exe) $^ > $(SAFEPIPETARGET)\n")

    # compatibility modes
    lines.append("$(matrixdir)/s_matrix.txt: %s\n" %
                 " ".join(map(lambda x:
                              "$(matrixdir)/s_%02i_matrix.txt"
                              % (x+1), matrix_cycles)))
    lines.append("\t$(CP) $< $(SAFEPIPETARGET)\n")
    for lane2 in lanes.keys():
        lines.append("$(matrixdir)/s_%i_matrix.txt: %s\n" %
                     (lane2+1, 
                      " ".join(map(lambda x:
                                   "$(matrixdir)/s_%i_%02i_matrix.txt"
                                   % (lane2+1, x+1), matrix_cycles))))
        lines.append("\t$(CP) $< $(SAFEPIPETARGET)\n")
        

    lines.append("\nmatrix: $(matrixdir)/s_matrix.txt %s\n\n" %
                 " ".join(map(lambda x:
                              "$(matrixdir)/$(sample%i)_%i_matrix.txt"
                              % (x+1, x+1), lanes.keys())))

    # matrix targets by tile
    lines.append("$(matrixdir)/%_mat.txt: "
                 "tilecycle=$(subst _mat.txt,,$(notdir $@))\n")
    for lane in lanes.keys():
        lines.append("$(l%i_matfiles): adir:=$(analysisdir%i)\n" % (lane+1,
                                                                lane+1))
    for i in matrix_cycles:
        lines.append("$(matrixdir)/%%_%02i_mat.txt: %%$(fullintsuffix)\n"
                     % (i+1))
        lines.append("# version with old per-cycle intensity files\n"
                     "#\t$(cmdprefix) $(matrix1_exe) "
                     "$(adir)/$(tilecycle)$(intsuffix) "
                     "> $(SAFEPIPETARGET)\n")
        lines.append("\t$(cmdprefix) $(matrix1_exe) "
                     "-z $(compression) -c %i "
                     "$(subst $(fullintsuffix),$(intsuffix),$^) "
                     "> $(SAFEPIPETARGET)\n" % (i+1))

    ##################################################################
    # targets to make typing easier
    ##################################################################
    tiletargets = {}
    lines.append("\n# Some rules to make typing targets easier\n")
    for tilestr in tiles:
        sample, lane, tile = runfolder.FileName.split_tile(tilestr)
        tilestr2 = get_var_tile(tilestr)
        lines.append("%s: %s\n" %
                     (tilestr, " ".join(map(lambda x :"%s_%02i_mat.txt" %
                                            (tilestr2, x+1),
                                            matrix_cycles))))
        tiletargets.setdefault(tile, []).append(tilestr2)
    keys = tiletargets.keys()
    keys.sort()
    for key in keys:
        lines.append("_%04i: %s\n" % (key + 1, " ".join(tiletargets[key])))

    ##################################################################
    # generate rules for each lane
    ##################################################################
    lines.append("\n# Lane targets\n")
    lines.append("%_finished.txt:\n"
                 "\ttouch $@ && $(cmdf2) $(subst _finished.txt,,$@)\n")
    for lane in range(runfolder.maxlane):
        lines.append("$(sample%i)_%i: "
                         "$(sample%i)_%i_finished.txt l%i_msg\n" %
                         (lane+1, lane+1, lane+1, lane+1, lane+1))
        if not run_params.IparFlag:
            lines.append("$(sample%i)_%i_finished.txt: "
                         "offsets_finished.txt " % (lane+1, lane+1))
            if lane in lanes.keys():
                lines.append("%s" %
                             (" ".join(map(lambda x:
                                           "$(matrixdir)/$(sample%i)_%i_%02i_summary.txt" %
                                           (lane+1, lane+1, x+1),
                                           matrix_cycles))))
        #lines.append("\ttouch $@ && $(cmdf2) $(subst _finished.txt,,$@)\n")
        lines.append("\nl%i_msg:\n"
                     "\t@echo Analysing lane %i: $(l%i_intfiles)\n" %\
                     (lane+1, lane+1, lane+1))
    lines.append("\n")

    ##################################################################
    # clean targets
    ##################################################################
    lines.append("\n#Clean targets\n")
    lines.append("clean: clean_matrix clean_offsets clean_sizes "
                 "clean_focus clean_tags clean_lanes\n")
    # The next line means that a restart of the pipeline after a
    # make clean is not safe to give exactly the same result if the
    # offsets have changed. But then why do "make clean" if you don't
    # want to recalculate.
    lines.append("ifeq ($(autooffsetflag),true)\n"
                 "\t$(RM) $(default_offsets)\n"
                 "endif\n"
                 "ifeq ($(autosizeflag),true)\n"
                 "\t$(RM) params*.xml\n"
                 "endif\n")
    lines.append("clean_tags:\n"
                 "\t$(RM) *finished.txt $(lanefile) $(tilefile)\n")
    lines.append("clean_lanes: %s\n"
                 % " ".join(clean_targets))
    lines.append("\t$(RM) $(analysis_subfolder)/*.log\n")
    lines.append("\n")
    for lane in lanes.keys():
        lines.append("clean%i: clean%i_int clean%i_int_sub clean%i_all_sub\n"
                     % (lane+1, lane+1, lane+1, lane+1))
        
        lines.append("clean%i_int:\n"
                     "\tfind . -maxdepth 1 -regex \".*$(sample%i)_%i_.*_\\(int\\|nse\\)\\.txt\" | xargs $(RM)\n" % (lane+1, lane+1, lane+1))
        lines.append("\tfind . -maxdepth 1 -regex \".*$(sample%i)_%i_.*_\\(int\\|nse\\)\\.txt\\.gz\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("\tfind . -maxdepth 1 -regex \".*$(sample%i)_%i_.*_.*\\.tmp\" | xargs $(RM)\n" % (lane+1, lane+1))

        # clean subdirectories
        lines.append("clean%i_int_sub:\n"
                     "\tfind $(analysisdir%i) -regex \".*_\\(int\\|nse\\)\\.txt\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("clean%i_all_sub:\n"
                     "\tfind $(analysisdir%i) -regex \".*_\\(idx\\|pos\\|clu\\)\\.txt\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("\tfind $(analysisdir%i) -name \"*_qcm.xml\" | xargs $(RM)\n" % (lane+1))
    lines.append("clean_intermediate: %s\n" %
                 " ".join(map(lambda x: "clean%i_int_sub" % (x+1),
                              lanes.keys())))
    lines.append("clean_offsets:\n"
                 "\tfind $(offsetdir) -name \"*.txt\" | xargs $(RM)\n")
    lines.append("clean_sizes:\n"
                 "\tfind $(sizedir) -regex \".*\\.\\(txt\\|xml\\)\" | xargs $(RM)\n")
    lines.append("clean_focus:\n"
                 "\tfind $(focusdir) -name \"*.txt\" | xargs $(RM)\n"
                 "\tfind $(focusdir) -name \"*.csv\" | xargs $(RM)\n")
    lines.append("clean_matrix:\n")
    lines.append("\t$(RM) %s\n" %
                 " ".join(map(lambda x: "$(matrixsummary%i)" % (x+1),
                              matrix_cycles)))
    lines.append("\tfind $(matrixdir) -name \"*_mat.txt\" | xargs $(RM)\n")
    lines.append("\tfind $(matrixdir) -name \"*_cmat.txt\" | xargs $(RM)\n")
    lines.append("\tfind $(matrixdir) -name \"*_matrix.txt\" | xargs $(RM)\n")
    lines.append("\tfind $(matrixdir) -name \"*_summary.txt\" | xargs $(RM)\n")

    ##################################################################
    # targets in other Makefiles
    ##################################################################
    lines.append("compress uncompress compress_intermediate uncompress_intermediate:\n"
                 "\t$(MAKE) -f Makefile.compress.firecrest $@\n")
    lines.append("compress_images uncompress_images:\n"
                 "\t$(MAKE) -f Makefile.compress.images $@\n")

    lines.append("\nvpath %_image.txt $(focusdir)\n")
    lines.append("focus: $(focusdir)/focus.csv\n")
    lines.append("focuscmd?=$(focusdir)Quality\n")

    #lines.append("$(focusdir)/focus.csv: $(addprefix $(focusdir)/,$(subst _int.txt,_focus.csv,$(intfiles)))\n"
    #             "\t$(RM) $@ ; \\\n"
    #             "\tfor f in $^; do \\\n"
    #             "\t  cat $$f >> $@ ; \\\n"
    #             "\tdone \n"

    #             "$(focusdir)/%_image.txt: %_jnl.txt\n"
    #             "\tsed -n '{s/Load \"\\(.*\\)\"/\\1/p}' $< > $@\n"
    #             "$(focusdir)/%_focus.csv: $(focusdir)/%_image.txt\n"
    #             "\t$(focuscmd) -c $< $@\n")

    ##################################################################
    # recursive make
    ##################################################################
    lines.append("\nrecursive: $(subfolders) nonrecursive\n")
    lines.append("nonrecursive:\n")
    lines.append("\t$(MAKE) $(TARGET)\n")
    lines.append("$(subfolders): nonrecursive\n")
    lines.append("\t$(MAKE) --directory=$@ recursive TARGET=$(TARGET)\n")
    try:
        fd = open(filename, "w")
        if not fd:
            common.error("Could not open %s." % filename)
            return
        fd.writelines(lines)
        fd.close()
    except:
        common.error("Failed to write %s." % filename)

    makefilename = "Makefile.compress.images"
    makefilepath2 = GA_DATA_DIR
    try:
        shutil.copyfile(os.path.join(makefilepath2, makefilename),
                        os.path.join(analysis_folder, makefilename))
    except:
        common.error("Failed to copy %s." % makefilename)

    makefilename = "Makefile.compress.firecrest"
    try:
        shutil.copyfile(os.path.join(makefilepath2, makefilename),
                        os.path.join(analysis_folder, makefilename))
    except:
        common.error("Failed to copy %s." % makefilename)


def getAutoFilename(option, lane, prefix, suffix):
    if option.AutoFlag == 1:
        if option.AutoLane == 0:
            filename = (prefix + ("s_%02i" % option.Cycle) + suffix)
        else:
            filename = (prefix +
                        ("$(sample%i)_%i_%02i" %
                         (lane+1, option.AutoLane, option.Cycle))
                        + suffix)
                    
    elif option.AutoFlag == 2:
        filename = (prefix +
                    ("$(sample%i)_%i_%02i" %
                    (lane+1, lane+1, option.Cycle))
                    + suffix)
    else:
        filename = option.Path
    return filename
    
def getMatrixFilename(option, lane):
    return getAutoFilename(option, lane,
                           "$(analysisdir)/$(matrixdir)/",
                           "_matrix.txt")
    
def getPhasingFilename(option, lane):
    filename = getAutoFilename(option, lane,
                               "$(phasingdir)/", "_phasing.xml")
    if option.AutoFlag == 0 and not option.Path:
        filename = "params%i.xml" % (lane+1)
    return filename
    

def generate_basecaller_makefile(tiledic,
                                 analysis_folder, sequence_folder,
                                 run_params, base_params):
    phasing_rate     =  base_params.Phasing[0].PhasingRate
    prephasing_rate  =  base_params.Phasing[0].PrephasingRate
    autophasing      =  base_params.Phasing[0].AutoFlag
    autophasinglane  =  base_params.Phasing[0].AutoLane

    if run_params.IparFlag:
        ipar_long_option  = "--IPAR"
        ipar_short_option = "-I"
    else:
        ipar_long_option  = ""
        ipar_short_option = ""

    
    if run_params.IterativeMatrixFlag:
        iterative_matrix_option = "-iterative-matrix"
    else:
        iterative_matrix_option = ""

    makefilename = os.path.join(sequence_folder, "Makefile")
    lines = []
    
    tiles = tiledic.keys()
    tiles.sort()
    lanes = runfolder.get_lane_samples(tiles)

    lines.append("# Makefile auto-generated by makefile.py (version %s)\n# %s\n\n"
                 % (bustard.basecaller_software_version,
                    runfolder.get_date_time()))
    installation_dir = os.path.split(runfolder.get_executable(bustard.basecaller_executable))[0]
    lines.append("installationdir:=%s\n" %  installation_dir)
    lines.append("basecall_exe:=%s %s %s\n" % (runfolder.get_executable(bustard.basecaller_executable)\
                                                .replace(installation_dir, "$(installationdir)"),\
                                            ipar_long_option, iterative_matrix_option))
    lines.append("phasing1_exe:=%s %s\n" %
             (runfolder.get_executable(bustard.phasing1_executable)\
                 .replace(installation_dir, "$(installationdir)"),\
              ipar_long_option))
    lines.append("phasing2_exe:=%s\n" %
             runfolder.get_executable(bustard.phasing2_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("phasing3_exe:=%s\n\n" %
             runfolder.get_executable(bustard.phasing3_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    #lines.append("ilmn2srf_exe:=%s %s\n\n" % (bustard.ilmn2srf_executable, ipar_short_option) )
    lines.append("#analysisdir:=%s\n" % analysis_folder)
    lines.append("analysisdir:=%s\n" % '..')
    lines.append("sequencedir:=%s\n" % sequence_folder)
    for lane in range(runfolder.maxlane):
        lines.append("sample%i:=%s\n" % (lane+1, lanes.get(lane,
                                                 runfolder.defaultsample)))
    lines.append("analysis_subfolder:=%s\n" % runfolder.analysissubdir)
    lines.append("sequence_subfolder:=%s\n" % runfolder.sequencesubdir)
    lines.append("lanes:=%s\n" %" ".join(map(lambda x: "$(sample%i)_%i" %
                                            (x+1, x+1),
                                            lanes.keys())))
    for lane in lanes.keys():
        folder = runfolder.get_lane_subfolder("", "$(analysis_subfolder)",
                                              "", lane)
        lines.append("analysisdir%i:=%s\n" % (lane+1, folder))
    for lane in lanes.keys():
        folder = runfolder.get_lane_subfolder("", "$(sequence_subfolder)",
                                              "", lane)
        lines.append("sequencedir%i:=%s\n" % (lane+1, folder))
    ncolours = 4
    allmaxcycle = get_max_cycle(tiledic.values()[0])
    lines.append("maxcycle:=%i\n" % allmaxcycle)
    lines.append("cycles:=%s\n\n" % " ".join(map(lambda x: "%02i" % (x+1),
                                                 range(allmaxcycle))))
             
    lines.append("HAVE_BASH3:=$(shell sh --version | grep \"bash, version [345]\.\")\n"
                 "ifneq ($(HAVE_BASH3),)\n"
                 "PIPECMD:=set -o pipefail;\n"
                 "endif\n")
    lines.append("PASTE=paste\n"
                 "CP=cp\n"
                 "SAFEPIPETARGET= $@.tmp && mv $@.tmp $@\n\n")
    lines.append("subfolders:=$(wildcard %s*)\n" % "GERALD")
    
    lines.append("TOOLS_DIR:=%s\n" %
                 os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])),
                                 gerald.toolspath)))
    lines.append("SMT_DIR=%s\n" % runfolder.signalmeansdir)
    lines.append("QH:=%s\n" %
                 os.path.join("$(TOOLS_DIR)",
                              runfolder.get_platform_executable("quahog")))
    pure_bases, use_bases, chastity_threshold = gerald.getQhgInfo()
    if (run_params.ChastityThreshold > 0):
        chastity_threshold = run_params.ChastityThreshold
    elif chastity_threshold > 0:
        run_params.ChastityThreshold = chastity_threshold
    smt_filter, smt_relation, smt_threshold, chastity_threshold = common.getSignalMeansInfo()
    if (run_params.SmtFilter):
        smt_filter    = run_params.SmtFilter
        smt_relation  = run_params.SmtRelation
        smt_threshold = run_params.SmtThreshold
    # lines.append("USE_BASES:=-useBases %s\n" % use_bases)
    # lines.append("PURE_BASES:=-pureBases %s\n" % pure_bases)
    # lines.append("QH_PARAMS:=$(USE_BASES) $(PURE_BASES)\n\n")
    lines.append("USE_BASES:=-u %s\n" % use_bases)
    lines.append("PURE_BASES:=-p %s\n" % pure_bases)
    lines.append("CHASTITY_THRESHOLD:=--chastity-threshold %s\n" % chastity_threshold)
    lines.append("SMT_FILTER:=-f %s\n" % smt_filter)
    lines.append("SMT_RELATION:=-r %s\n" % smt_relation)
    lines.append("SMT_THRESHOLD:=-t %s\n" % smt_threshold)
    #lines.append("SMT_PREFIX:=--signal-means-prefix %s\n" % "SignalMeans/")
    lines.append("SMT_PREFIX:=%s\n" % "")
    lines.append("SMT_SUFFIX:=--signal-means-suffix %s\n" % "_all.txt.tmp")
    lines.append("SMT_PARAMS:= --with-signal-means $(SMT_FILTER) $(SMT_RELATION) $(SMT_THRESHOLD) $(SMT_PREFIX) $(SMT_SUFFIX)\n")
    lines.append("QH_PARAMS:=$(SMT_PARAMS) $(USE_BASES) $(PURE_BASES) $(CHASTITY_THRESHOLD)\n\n")
    
    lines.append("logfile:=$(sequence_subfolder)/$(sequence_subfolder).log\n")
    lines.append("matrixdir:=%s\n" % runfolder.matrixdir)
    lines.append("matrix:=%s\n" % "matrix.txt")
    for lane in lanes.keys():
        lines.append("matrix%i:=matrix%i.txt\n" % (lane+1, lane+1))
    for lane in lanes.keys():
        lines.append("params%i:=params%i.xml\n" % (lane+1, lane+1))
    lines.append("phasingdir:=%s\n" % runfolder.phasingdir)
    lines.append("phasing_summary:=$(phasingdir)/phasing.txt\n")
    if base_params.Matrix[0].AutoFlag:
        lines.append("automatrix?=true\n")
    else:
        lines.append("automatrix?=false\n")
    if autophasing:
        lines.append("autophasing?=true\n")
    else:
        lines.append("autophasing?=false\n")
    lines.append("compression:=%s\n" % run_params.Compression)
    lines.append("compressionsuffix:=%s\n" % run_params.CompressionSuffix)
    iparSuffix = "";
    if run_params.IparFlag:
        lines.append("iparsuffix:=.p\n")
        iparSuffix = "$(iparsuffix)"
    lines.append("seqsuffix:=_seq.txt\n"
                 "intsuffix:=_int.txt\n"
                 "qhgsuffix:=_qhg.txt\n"
                 "smtsuffix:=_all.txt\n"
                 "tmpsuffix:=.tmp\n")
    #lines.append("srfsuffix:=_traces.srf\n")
    lines.append("fullintsuffix:=$(intsuffix)%s$(compressionsuffix)\n\n" % iparSuffix)
    lines.append("# Hooks\ncmdb1?=echo \ncmdb2?=echo \n"
                 "cmdb3?=echo \ncmdb4?=echo \n")
        
    clean_targets = map(lambda x: "clean%i" % (x+1), lanes.keys()) 
    lines.append("\nvpath %$(fullintsuffix) $(analysisdir)\n")
    lines.append("vpath %phasing.txt $(phasingdir)\n")
    lines.append("vpath %phasing.xml $(phasingdir)\n")
    lines.append("vpath %_matrix.txt $(analysisdir)/$(matrixdir)\n")
    lines.append("\n"
                 ".PHONY: allmsg $(lanes) clean clean_phasing "
                 "clean_intermediate %s "
                 "recursive nonrecursive $(subfolders)\n" %
                 (" ".join(clean_targets)))
    lines.append(".PRECIOUS: %$(seqsuffix)\n"
                 ".PRECIOUS: %_phasing.txt\n"
                 ".PRECIOUS: %.xml\n")

    lines.append("all: allmsg $(lanes) phasing.xml\n"
                 "\ttouch finished.txt && \\\n"
                 "\techo \"Base-calling has completed successfully.\" && \\\n"
                 "\t$(cmdb4)\n")
    lines.append("allmsg:\n\t@echo Analysing lanes $(lanes)\n")
    lines.append("\n")

    for lane2 in range(runfolder.maxlane):
        seqfiles = []
        for tilestr in tiles:
            sample, lane, tile = runfolder.FileName.split_tile(tilestr)
            if lane2 == lane:
                seqfiles.append(get_var_tile(tilestr))
        lines.append("l%i_tiles:=%s\n" %     (lane2+1, " ".join(seqfiles)))
        lines.append("l%i_seqfiles:=$(addsuffix $(seqsuffix),$(l%i_tiles))\n"
                     %  (lane2 + 1, lane2 + 1))
        lines.append("l%i_qhgfiles:=$(addsuffix $(qhgsuffix),$(l%i_tiles))\n"
                     % (lane2+1, lane2+1))
        

    ##################################################################
    # generate rules for seq.txt files
    ##################################################################
    lines.append("# Auto-generate matrices\n")
    for lane in lanes.keys():
        lines.append("$(matrix%i):\n" % (lane+1))
        if base_params.Matrix[0].AutoFlag == 2:
            lines.append("\t$(CP) $(analysisdir)/$(matrixdir)/s_%i_matrix.txt"
                         " $(SAFEPIPETARGET)\n\n" % (lane+1))
        elif base_params.Matrix[0].AutoLane:
            lines.append("\t$(CP) $(analysisdir)/$(matrixdir)/s_%i_matrix.txt"
                         " $(SAFEPIPETARGET)\n\n" % (base_params.Matrix[0].AutoLane))
        else:
            lines.append("\t$(CP) $(analysisdir)/$(matrixdir)/s_matrix.txt "
                         "$(SAFEPIPETARGET)\n\n")
        
    lines.append("# Generate sequence and probability files\n")
    #lines.append("%$(seqsuffix): %$(fullintsuffix)\n")
    difflengthtiles = {}
    for tilestr in tiles:
        maxcycle = get_max_cycle(tiledic[tilestr])
        if maxcycle != allmaxcycle:
            difflengthtiles.setdefault(maxcycle, []).append(tilestr)
    keys = difflengthtiles.keys()
    keys.sort()
    for cycle in keys:
            lines.append("%s: maxcycle:=%i\n" %
                         (" ".join(map(lambda x: "%s$(seqsuffix)" % x,
                                       difflengthtiles[cycle])), cycle))
    lines.append("%$(seqsuffix): tile=$(subst $(seqsuffix),,$@)\n")
    matrix_options  = base_params.getMatrixOptions(run_params)
    phasing_options = base_params.getPhasingOptions(run_params)
    for lane in lanes.keys():
        lines.append("$(l%i_seqfiles): $(matrix%i) $(params%i)\n"
                     % (lane+1, lane+1, lane+1))
        lines.append("$(l%i_seqfiles): adir:=$(analysisdir)/$(analysisdir%i)\n"
                     % (lane+1, lane+1))
        lines.append("$(l%i_seqfiles): sdir:=.\n"
                     % (lane+1))
        # construct matrix options
        option_strings = []
        for option in matrix_options:
            filename = getMatrixFilename(option, lane)
            option_strings.append("-new-matrix=%i:%s" %
                                  (option.FirstCycle, filename))
        lines.append("$(l%i_seqfiles): matrixoptions:=%s\n"
                     % (lane+1, " ".join(option_strings)))
        # construct phasing options
        option_strings = []
        for option in phasing_options:
            filename = getPhasingFilename(option, lane)
            if option.Read == 1:
                option_strings.append("-params=%s" % filename)
            else:
                option_strings.append("-new-params=%i:%s" %
                                      (option.FirstCycle, filename))
        lines.append("$(l%i_seqfiles): phasingoptions:=%s\n"
                     % (lane+1, " ".join(option_strings)))

    lines.append("%$(seqsuffix):%$(fullintsuffix)\n"
                 "\t$(cmdprefix) $(basecall_exe) -nc=$(maxcycle) -logfile=$(logfile) -prefix=$(adir)/ -prefix2=$(sdir)/ -prefix3=../ -tile=$(tile) $(matrixoptions) $(phasingoptions) -compression=$(compression) -sig2compression=$(compression) -savesuffix=$(tmpsuffix) && \\\n"
                 "\tmv $@.tmp $@\n\n")

    #################################################################
    # qhg files - now generates signal means as well
    ################################################################
    lines.append("%$(qhgsuffix): %$(seqsuffix)\n"
                 "\t$(cmdprefix) $(QH) $(QH_PARAMS) "
                 #"$(subst seq,sig2,$<) > $(SAFEPIPETARGET) && mv $(@:%$(qhgsuffix)=%$(smtsuffix)).tmp $(SMT_DIR)/$(@:%$(qhgsuffix)=%$(smtsuffix))\n")
                 "$(subst seq,sig2,$<) -c $(compression) > $(SAFEPIPETARGET) && mv $(@:%$(qhgsuffix)=%$(smtsuffix)).tmp $(SMT_DIR)/$(@:%$(qhgsuffix)=%$(smtsuffix))\n")
    
    ##################################################################
    # phasing estimation
    ##################################################################
    lines.append("\n# Phasing correction\n")
    phasingcycles = [1] + base_params.getPhasingRestarts(run_params)
    #print phasingcycles
    for lane in lanes.keys():
        for cycle in phasingcycles:
            lines.append("l%i_%02i_phasingfiles:=$(addprefix $(phasingdir)/,$(addsuffix _%02i_cycle.txt,$(l%i_tiles)))\n"
                         % (lane+1, cycle, cycle, lane+1))
        lines.append("l%i_phasingfiles:=%s\n" %
                     (lane+1,
                      " ".join(map(lambda x: "$(l%i_%02i_phasingfiles)" %
                                   (lane+1, x), phasingcycles))))

    for lane in lanes.keys():
        #lines.append("$(phasingdir)/$(sample%i)_%i_%%_phasing.txt: "
        #             "$(matrix%i)\n" %
        #             (lane+1, lane+1, lane+1))
        for cycle in phasingcycles:
            for option in matrix_options:
                if ((cycle >= option.FirstCycle) and
                    (cycle <= option.LastCycle)):
                    break
            filename = getMatrixFilename(option, lane)
            lines.append("$(l%i_%02i_phasingfiles): "
                         " matrix:=%s\n" % (lane+1, cycle, filename))
            lines.append("$(l%i_%02i_phasingfiles): "
                         " firstcycle:=%i\n" % (lane+1, cycle, cycle))
    for cycle in phasingcycles:
        lines.append("$(phasingdir)/%%_%02i_cycle.txt: "
                     "tile=$(subst _%02i_cycle.txt,,$(notdir $@))\n"
                     % (cycle, cycle))
    for cycle in phasingcycles:
        lines.append("$(phasingdir)/%%_%02i_cycle.txt: %%$(fullintsuffix)\n"
                     % (cycle))
        lines.append("\t$(cmdprefix) $(phasing1_exe) "
                     "--compression=$(compression) "
                     "--firstcycle=$(firstcycle) "
                     "--matrix=$(matrix) $(analysisdir)/$(tile)$(intsuffix)"
                     " > $(SAFEPIPETARGET)\n")

    phasinglanetargets = []
    for lane in lanes.keys():
        for cycle in phasingcycles:
            target = "$(phasingdir)/$(sample%i)_%i_%02i_phasing.txt" %\
                     (lane+1, lane+1, cycle)
            #prereq = "$(l%i_%02i_phasingfiles) $(matrix%i)" % (lane+1, cycle, lane+1)
            prereq = "$(l%i_%02i_phasingfiles)" % (lane+1, cycle)
            lines.append("%s: %s\n" % (target, prereq))
            phasinglanetargets.append(target)
    lines.append("%s:\n" % " ".join(phasinglanetargets))
    lines.append("\tfor f in $^; do \\\n"
             "\t    echo $$f && \\\n"
             "\t    $(cmdprefix) $(phasing2_exe) $$f >> $@.tmp ; \\\n"
             "\tdone && \\\n"
             "\tmv $@.tmp $@\n")

    for cycle in phasingcycles:
        prereqs = []
        for lane in lanes.keys():
            prereqs.append("$(phasingdir)/$(sample%i)_%i_%02i_phasing.txt"
                           % (lane+1, lane+1, cycle))
        lines.append("$(phasingdir)/s_%02i_phasing.txt: %s\n"
                     "\tcat $^ > $(SAFEPIPETARGET)\n" %
                     (cycle, " ".join(prereqs)))

    lines.append("$(phasingdir)/%_phasing.xml: $(phasingdir)/%_phasing.txt\n"
                 "\t$(cmdprefix) $(phasing3_exe) $< > $(SAFEPIPETARGET)\n")
    for lane in lanes.keys():
        prereqs = []
        for cycle in phasingcycles:
            prereqs.append("$(phasingdir)/$(sample%i)_%i_%02i_phasing.xml" %
                           (lane+1, lane+1, cycle))
        lines.append("$(phasingdir)/$(sample%i)_%i_phasing.xml: %s\n"
                     "\t$(CP) $< $(SAFEPIPETARGET) && "
                     "$(cmdb1) $(sample%i)_%i\n" %
                     (lane+1, lane+1, " ".join(prereqs), lane+1, lane+1)) 

    prereqs = []
    for cycle in phasingcycles:
        prereqs.append("$(phasingdir)/s_%02i_phasing.xml" % cycle)
    for lane in lanes.keys():
        prereqs.append("$(phasingdir)/$(sample%i)_%i_phasing.xml" %
                       (lane+1, lane+1))
    lines.append("$(phasingdir)/phasing.xml: %s\n"
                 "\t$(CP) $< $(SAFEPIPETARGET) && $(cmdb2)\n" %
                 " ".join(prereqs))

    lines.append("ifeq ($(autophasing),true)\n")
    for lane in lanes.keys():
        if autophasing == 2:
                lines.append("$(params%i): $(phasingdir)/$(sample%i)_%i_phasing.xml\n"
                             % (lane+1, lane+1, lane+1))
        else:
            if autophasinglane:
                lines.append("$(params%i): $(phasingdir)/$(sample%i)_%i_phasing.xml $(phasingdir)/$(sample%i)_%i_phasing.xml\n"
                             % (lane+1, autophasinglane, autophasinglane,
                                lane+1, lane+1))
            else:
                lines.append("$(params%i): $(phasingdir)/phasing.xml $(phasingdir)/$(sample%i)_%i_phasing.xml\n"
                             % (lane+1, lane+1, lane+1))
            
        lines.append("\t$(CP) $< $(SAFEPIPETARGET)\n")
    lines.append("endif\n")

    ##################################################################
    # targets to make typing easier
    ##################################################################
    tiletargets = {}
    lines.append("\n# Some rules to make typing targets easier\n")
    for tilestr in tiles:
        lines.append("%s: %s$(qhgsuffix)\n" % (tilestr, get_var_tile(tilestr)))
        sample, lane, tile = runfolder.FileName.split_tile(tilestr)
        tiletargets.setdefault(tile, []).append(get_var_tile(tilestr))
    keys = tiletargets.keys()
    keys.sort()
    for key in keys:
        lines.append("_%04i: %s\n" % (key + 1, " ".join(tiletargets[key])))

    ##################################################################
    # generate rules for each lane
    ##################################################################
    lines.append("\n# Lane targets\n")
    lines.append("%_finished.txt:\n"
                 "\ttouch $@ && $(cmdb3) $(subst _finished.txt,,$@)\n")
    for lane in range(runfolder.maxlane):
        prefix = "$(sample%i)_%i" % (lane+1, lane+1)
        if len(run_params.Reads) > 1:
            second_pair = "-2 %i" % run_params.Reads[1].FirstCycle
        else:
            second_pair = ""
        lines.append("$(sample%i)_%i: l%i_msg $(sample%i)_%i_finished.txt\n" %
                     (lane+1, lane+1, lane+1, lane+1, lane+1))
        #lines.append("$(sample%i)_%i_finished.txt: $(l%i_qhgfiles) $(sample%i)_%i$(srfsuffix)\n" %
        #             (lane+1, lane+1, lane+1, lane+1, lane+1))
        lines.append("$(sample%i)_%i_finished.txt: $(l%i_qhgfiles) \n" %
                     (lane+1, lane+1, lane+1))
        lines.append("l%i_msg:\n\t@echo Analysing lane %i: $(l%i_qhgfiles)\n" %\
                     (lane+1, lane+1, lane+1))
        #lines.append("%s$(srfsuffix): $(l%i_seqfiles)\n" % (prefix, lane+1))
        #lines.append("\t$(ilmn2srf_exe) -r -o $@ $^\n")
    lines.append("\n")

    ##################################################################
    # clean targets
    ##################################################################
    lines.append("\n# clean targets\n")
    lines.append("clean: %s clean_phasing clean_all\n" % " ".join(clean_targets))
    lines.append("\t$(RM) $(sequence_subfolder)/*.log\n")
    lines.append("\t$(RM) ztrarc.hash\n"
                 "\t$(RM) *finished.txt\n")
    lines.append("    ifeq ($(automatrix),true)\n"
                 "\t$(RM) matrix*.txt\n"
                 "    endif\n")
    lines.append("    ifeq ($(autophasing),true)\n"
                 "\t$(RM) params*.xml\n"
                 "    endif\n")
    for lane in lanes.keys():
        lines.append("clean%i: clean%i_seq clean%i_sig "
                     "clean%i_sig2 clean%i_qhg clean%i_smt\n" %
                     (lane+1, lane+1, lane+1, lane+1, lane+1, lane+1))
        lines.append("clean%i_seq:\n" % (lane+1))
        lines.append("\tfind . -maxdepth 1 -regex \".*$(sample%i)_%i_.*_\\(seq\\|prb\\)\\.txt\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("\tfind . -maxdepth 1 -regex \".*$(sample%i)_%i_.*_\\(seq\\|prb\\)\\.txt\\.gz\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("\tfind . -maxdepth 1 -regex \".*$(sample%i)_%i_.*_.*\\.tmp\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("clean%i_sig:\n" % (lane+1))
        lines.append("\tfind . -maxdepth 1 -name \"$(sample%i)_%i_*_sig.txt\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("clean%i_qhg:\n" % (lane+1))
        lines.append("\tfind . -maxdepth 1 -name \"$(sample%i)_%i_*_qhg.txt\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("clean%i_smt:\n" % (lane+1))
        lines.append("\tfind . -maxdepth 1 -name \"$(sample%i)_%i_*_smt.txt\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("clean%i_sig2:\n" % (lane+1))
        lines.append("\tfind . -maxdepth 1 -name \"$(sample%i)_%i_*_sig2.txt\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("\tfind . -maxdepth 1 -name \"$(sample%i)_%i_*_sig2.txt.gz\" | xargs $(RM)\n" % (lane+1, lane+1))

    lines.append("clean_intermediate:\n"
                 "\t@echo Target clean_intermediate is obsolete.\n")
    lines.append("clean_phasing:\n"
                 "\tfind $(phasingdir) -regex \".*\\.\\(txt\\|xml\\)\" | xargs $(RM)\n\n")
    lines.append("clean_all:\n"
                 "\tfind $(SMT_DIR) -regex \".*\\.\\(txt\\|xml\\)\" | xargs $(RM)\n\n")

    lines.append("compress uncompress:\n"
                 "\t$(MAKE) -f Makefile.compress.bustard $@\n")

    ##################################################################
    # recursive make
    ##################################################################
    lines.append("# Recursive analysis: do alignment\n")
    lines.append("recursive: $(subfolders) nonrecursive\n")
    lines.append("nonrecursive:\n"
                 "\t$(MAKE) $(TARGET)\n")
    lines.append("$(subfolders): nonrecursive\n"
                 "    ifeq ($(words $(filter-out clean%,$(filter-out all,$(TARGET)))), 0)\n"
                 "\t$(MAKE) --directory=$@ $(TARGET)\n"
                 "    else\n"
                 "\t$(MAKE) --directory=$@ TILE=$(TARGET) all\n"
                 "    endif\n")

    # ztr archive
    lines.append("\n# Export trace file for gap4\n")
    lines.append("ztrarc.hash:\n")
    lines.append("\tilmn2ztr -p -o $@ *$(seqsuffix)\n")
    try:
        fd = open(makefilename, "w")
        if not fd:
            common.error("Could not open %s." % makefilename)
            return
        fd.writelines(lines)
        fd.close()
    except:
        common.error("Failed to write %s." % makefilename)

    makefilename = "Makefile.compress.bustard"
    makefilepath2 = GA_DATA_DIR
    try:
        shutil.copyfile(os.path.join(makefilepath2, makefilename),
                        os.path.join(sequence_folder, makefilename))
    except:
        common.error("Failed to copy %s." % makefilename)
