#! /usr/bin/env python

#
# Copyright (c) Solexa 2005-2007
# Author: Klaus Maisinger
#
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

#
# program name: 'Bustard'
# CVS - $Id: bustard.py,v 1.5 2008/07/25 12:03:38 rshaw Exp $      
#             
# script to configure the analysis environment and
# generate the Makefiles for:
#  . base-calling
#  . sequencing


import sys, os, os.path, glob, getopt, shutil, traceback
GA_LIB_LIST=[os.path.join(os.path.dirname(os.path.realpath(sys.argv[0])), "..", "lib")]
sys.path = GA_LIB_LIST + sys.path
import runfolder, common, makefile, XMLSimple, gerald

basecaller_software         = "Bustard"
basecaller_software_version = "1.9.6"
basecaller_executable = "callBases"
ilmn2srf_executable   = "ilmn2srf"
phasing1_executable   = "estimatePhasing"
phasing2_executable   = "phasing_summary.py"
phasing3_executable   = "select_phasing.py"

class ReadParams:
    def __init__(self,
                 AutoLane   =  0,
                 AutoFlag   =  1, # 0: Default, 1: Auto, 2: Lane
                 Read       =  0,
                 Duplicates =  0,
                 FirstCycle = -1,
                 LastCycle  = -1):
        self.AutoLane   = AutoLane
        self.AutoFlag   = AutoFlag
        self.Read       = Read
        self.Duplicates = Duplicates
        self.FirstCycle = FirstCycle
        self.LastCycle  = LastCycle
        self.Path       = ""
        self.CycleOffset= 0 # by default, use first cycle for estimation


class Matrix(ReadParams):
    """Holds information necessary to perform matrix estimation."""
    def __init__(self):
        ReadParams.__init__(self)
        self.CycleOffset = 1  # matrix estimation starts on second cycle
    pass

class Phasing(ReadParams):
    """Holds information necessary to perform phasing estimation."""
    def __init__(self):
        ReadParams.__init__(self)
        self.PhasingRate     = 0.
        self.PrephasingRate  = 0.


def getReadIndex(items, index):
    result = None
    for item in items:
        if item.Read == index:
            result = item
            break
    return result

def getReadIndexDefault(items, index):
    result = getReadIndex(items, index)
    if not result:
        result = getReadIndex(items, 0)
    return result

def parseReadArgument(arg):
    result = None
    parts = arg.split(":")
    if len(parts) > 2:
        common.error("Wrong argument %s=%s:"
                     " Too many colons." % arg)
        return result

    # identify read the argument applies to (0 means all)
    idx = 0
    if len(parts) == 2:
        try:
            idx = int(parts[0])
        except:
            common.error("Could not parse %s" % parts[0])
            return result
    return { idx: parts[-1] }


def getReadOptions(reads, objects, ObjectClass):
    """Condense a list of objects derived from Readparams
    so that there is exactly one fully expanded object per read."""
    result = []
    for read in reads:
        newobject            = ObjectClass()
        newobject.FirstCycle = read.FirstCycle
        newobject.LastCycle  = read.LastCycle
        newobject.Read       = read.Index
        myobject = getReadIndexDefault(objects, read.Index)
        if not myobject: continue
        read2 = read
        if myobject.Duplicates > 0:
            read2 = reads[myobject.Duplicates-1]
            myobject = getReadIndexDefault(objects, myobject.Duplicates)
            if not myobject: continue
        newobject.Cycle = read2.FirstCycle
        if read2.LastCycle >= read2.FirstCycle + myobject.CycleOffset:
            newobject.Cycle = read2.FirstCycle + myobject.CycleOffset
        newobject.AutoFlag  = myobject.AutoFlag
        newobject.AutoLane  = myobject.AutoLane
        newobject.Path      = myobject.Path
        result.append(newobject)
    return result

def printRead(name, read):
    if read.Read == 0:
        print "%-7s:      " % (name.capitalize()),
    else:
        print "%-7s rd. %i:" % (name.capitalize(), read.Read),
    if read.Duplicates > 0:
        print "Using same as read %i." % read.Duplicates
    else:
        if read.AutoFlag == 1:
            if read.AutoLane:
                print ("Using automatic %s estimation "
                       " (lane %i only)." % (name, read.AutoLane))
            else:
                print "Using automatic %s estimation." % name
        elif read.AutoFlag == 2:
            print "Using lane-specific %s." % name
        elif read.Path:
            print "Using file %s." % read.Path
    

class BaseCallParams(common.Params):
    def __init__(self):
        self.Phasing         = [ Phasing() ]
        self.Matrix          = [ Matrix() ]
        self.PhasingRestarts = []
        self.options         = [ "matrix=", "matrix-cycle=",
                                 "phasing=", "prephasing=",
                                 "reset-phasing-cycle=", "control-lane=",
                                 "srf" ]

    def completePath(self, run_params):
        for matrix in self.Matrix:
            if matrix.AutoFlag == 0:
                matrix.Path = common.get_matrix(matrix.Path,
                                                run_params.InstrumentFolder,
                                                run_params.RunFolder)
                if not matrix.Path:
                    common.error("Specified matrix file does not exist.")
                    sys.exit(1)

    def getMatrix(self, index):
        return getReadIndex(self.Matrix, index)

    def getPhasing(self, index):
        return getReadIndex(self.Phasing, index)

    def getMatrixOptions(self, run_params):
        return getReadOptions(run_params.Reads, self.Matrix, Matrix)

    def getPhasingOptions(self, run_params):
        return getReadOptions(run_params.Reads, self.Phasing, Phasing)

    def getMatrixCycles(self, run_params): # 0 indexed
        """Get all cycles for which matrices are estimated."""
        cycles = []
        matrices = self.getMatrixOptions(run_params)
        for matrix in matrices:
            cycles.append(matrix.Cycle-1)
        #for read in run_params.Reads:
        #    matrix = Matrix()
        #    cycles.append(read.FirstCycle + matrix.CycleOffset-1)
        cycles = common.unique_list(cycles)
        cycles.sort()
        return cycles

    def getRunRestarts(self, run_params):
        cycles = []
        for read in run_params.Reads[1:]:
            cycles.append(read.FirstCycle)
        return cycles

    def getPhasingRestarts(self, run_params):
        return common.unique_list(self.getRunRestarts(run_params)
                                  + self.PhasingRestarts)

    def parse(self, options):
        """Set values according to command-line options from
        getopt.getopt"""
        for o, a in options:
            if o == "--matrix-cycle":
                # split command-line arguments on ":" into a (read, value) pair
                argdic = parseReadArgument(a)
                if not argdic:
                    continue
                idx = argdic.keys()[0]
                arg = argdic[idx]
                # check if a matrix for the read exists - if not add a new one
                matrix = self.getMatrix(idx)
                if matrix == None:
                    self.Matrix.append(Matrix())
                    matrix = self.Matrix[-1]
                    matrix.Read = idx
                try: 
                    matrix.CycleOffset = int(arg) - 1
                except:
                    common.error("Cannot convert %s to integer" % arg)
                    sys.exit(1)

            if o == "--matrix":
                # split command-line arguments on ":" into a (read, value) pair
                argdic = parseReadArgument(a)
                if not argdic:
                    continue
                idx = argdic.keys()[0]
                arg = argdic[idx]
                # check if a matrix for the read exists - if not add a new one
                matrix = self.getMatrix(idx)
                if matrix == None:
                    self.Matrix.append(Matrix())
                    matrix = self.Matrix[-1]
                    matrix.Read = idx

                # parse option
                if (arg.lower().find("auto") == 0):
                    matrix.AutoFlag = 1
                    if (len(arg) == 5):
                        try:
                            matrix.AutoLane = int(arg[4])
                            if matrix.AutoLane > runfolder.maxlane:
                                matrix.AutoLane = runfolder.maxlane
                        except:
                            common.error("Cannot parse option %s=%s."
                                         % (o, a))
                elif (arg.lower().find("default") == 0):
                    matrix.AutoFlag = 0
                elif (arg.lower().find("lane") == 0):
                    matrix.AutoFlag = 2
                else:
                    try:
                        idx = int(arg)
                        matrix.Duplicates = idx
                    except:
                        matrix.Path = os.path.abspath(arg)
                        matrix.AutoFlag = 0
                
            if o == "--phasing":
                argdic = parseReadArgument(a)
                if not argdic:
                    continue
                idx = argdic.keys()[0]
                arg = argdic[idx]
                phasing = self.getPhasing(idx)
                if phasing == None:
                    self.Phasing.append(Phasing())
                    phasing = self.Phasing[-1]
                    phasing.Read = idx

                if (arg.lower().find("auto") == 0):
                    phasing.AutoFlag       = 1
                    phasing.PhasingRate    = 0.
                    phasing.PrephasingRate = 0.
                    if (len(arg) == 5):
                        try:
                            phasing.AutoLane = int(arg[4])
                            if phasing.AutoLane > runfolder.maxlane:
                                phasing.AutoLane = runfolder.maxlane
                        except:
                            common.error("Cannot parse option %s=%s."
                                         % (o, a))
                else:
                    phasing.AutoFlag = 0
                    try:
                        idx = int(arg)
                        phasing.Duplicates = idx
                    except:
                        try:
                            phasing.PhasingRate = float(arg)
                        except:
                            phasing.Path = arg

            if o == "--control-lane":
                matrix = self.getMatrix(0)
                if matrix:
                    matrix.AutoFlag = 1
                    matrix.AutoLane = int(a)
                phasing = self.getPhasing(0)
                if phasing:
                    phasing.AutoFlag = 1
                    phasing.AutoLane = int(a)
            if o == "--prephasing":
                phasing = self.getPhasing(0)
                phasing.AutoFlag = 0
                phasing.PrephasingRate = float(a)
            if o == "--reset-phasing-cycle":
                self.PhasingRestarts.append(int(a))

    def printMatrix(self):
        for matrix in self.Matrix:
            printRead("matrix", matrix)

    def printPhasing(self):
        for phasing in self.Phasing:
            printRead("phasing", phasing)
            if ((phasing.Duplicates == 0) and (not phasing.AutoFlag)
                and (not phasing.Path)):
                print "Phasing: %6.4f  prephasing: %6.4f" \
                      % (phasing.PhasingRate, phasing.PrephasingRate)
                if ((phasing.PhasingRate > 0.05) or
                    (phasing.PrephasingRate > 0.05)):
                    common.error("Phasing/prephasing values "
                                 "suspiciously high! A typo ??")
                    sys.exit(1)
            
    def printParams(self):
        self.printMatrix()
        self.printPhasing()

    def printRestarts(self, run_params):
        if self.PhasingRestarts:
            print "  Resetting phasing in cycle %s." %\
                  ",".join(map(lambda x: "%i" % x,
                               self.PhasingRestarts))
    def getDict(self, run_params=None):
        dic = common.Params.getDict(self)
        if run_params:
            dic["Phasing"] = common.getDict(self.getPhasingOptions(run_params))
            dic["Matrix"]  = common.getDict(self.getMatrixOptions(run_params))
        return dic
        


def print_tile_counts(tiledic):
    tiles = tiledic.keys()
    tiles.sort()
    for key in tiles:
        print "%10s %3i" % (key, tiledic[key])
    

def getTiles(run_params, analysisfolder):
    tiles = run_params.Tiles[:]
    run_params.Tiles = []
    maxcycle = run_params.MaxCycle
    tiledic = {}
    if tiles:
        for tile in tiles:
            tiledic[tile] = maxcycle
        return tiledic
    common.warning("Could not identify tiles from parameter file %s."
                   % paramsfile)
    return common.count_intensity_files(analysisfolder)


def getReadLength(tiledic):
    return min(tiledic.values())


def createBasecallerFolders(sequence_folder, tiles):
    lanes = runfolder.get_lane_samples(tiles)
    try:
        os.makedirs(sequence_folder)
        os.mkdir(os.path.join(sequence_folder, runfolder.sequencesubdir))
        os.mkdir(os.path.join(sequence_folder, runfolder.phasingdir))
        os.mkdir(os.path.join(sequence_folder, runfolder.signalmeansdir))
    except:
        common.error("Encountered error while creating folder %s. "
                     "Perhaps you do not have the permissions?"
                     % sequence_folder)
        traceback.print_exc()
        sys.exit(1)

def renameBasecallerFolders(old_sequence_folder, sequence_folder):
    try:
        oldfolder = os.path.join(os.path.split(sequence_folder)[0],
                                 os.path.split(old_sequence_folder)[1])
        print oldfolder
        print os.path.split(sequence_folder)[0]
        print os.path.split(old_sequence_folder[1])
        os.rename(oldfolder, sequence_folder)
    except:
        common.error("Failed to rename folder %s to %s!"
                     % (old_sequence_folder, sequence_folder))
        sys.exit(1)

def processParamsFiles(sequence_folder, tiles, base_params):
    if not base_params.Phasing[0].AutoFlag:
        for lane in runfolder.get_lane_samples(tiles).keys():
            paramsfile = os.path.join(sequence_folder,
                                      "params%i.xml" % (lane+1))
            if base_params.Phasing[0].Path:
                shutil.copyfile(base_params.Phasing[0].Path,
                                paramsfile)
            else:
                params = { "Phasing":    base_params.Phasing[0].PhasingRate,
                           "Prephasing": base_params.Phasing[0].PrephasingRate }
                XMLSimple.XMLout(params, RootName="Parameters",
                                 OutputFile = paramsfile)


def usage():
    print "Usage: bustard.py [--tiles=<tile1>[,<tile2>[,...]]] \
[--maxcycle=<number>] [--make] \
[--matrix=<matrix filename>|auto] \
[--phasing=<phasing rate>|auto|auto1|...] [--prephasing=<prephasing rate>] \
[--reset-phasing-cycle=<restart cycle>] \
[--new-read-cycle=<cycle> \
[--GERALD=<config file>] \
<image analysis directory>"


#
# -- main program --
#
if __name__ == "__main__":
    
    print "Starting the Bustard base-caller, %s version %s"\
          % (basecaller_software,
             basecaller_software_version)

    runfolder.validate_executables([ basecaller_executable,
                                     phasing1_executable ])

    run_params    = common.GeneralParams()
    base_params   = BaseCallParams()
    gerald_params = gerald.GeraldParams()

    # parsing command-line options
    try:
        (options, args) = getopt.gnu_getopt(sys.argv[1:], ":m",
                                            run_params.options
                                            + base_params.options
                                            + gerald_params.options)
        run_params.parseImageAnalysis(args)
    except:
        common.error("Failed to parse command-line options.")
        usage()
        sys.exit(1)
        
    #print options, args
    
    # parse command-line options
    run_params.parse(options)
    base_params.parse(options)
    gerald_params.parse(options)

    # checking the validity of the cycle selection
    if not run_params.CycleSelector.isAutoCycleAll() and 1 != run_params.CycleSelector.getFirstCycle():
        common.error("For base calling, the cycle selection must start at the first cycle")
        sys.exit(1)
    if not run_params.CycleSelector.isAutoCycleAll() and not run_params.CycleSelector.isContiguous():
        common.error("For base calling, the cycle selection must be contiguous")
        sys.exit(1)


    # Analysis folder
    analysis_folder = os.path.abspath(args[0])
    if os.path.split(analysis_folder)[1] == runfolder.analysissubdir:
        analysis_folder = os.path.split(analysis_folder)[0]

    # Identify IPAR analysis folder
    run_params.IparFlag = runfolder.is_ipar_analysis_folder(args[0])
    
    # grouping filenames by position on chip
    tiledic = getTiles(run_params, analysis_folder)
    tiledic = common.select_tiles(tiledic, run_params.SelectedTiles)
    if run_params.MaxCycle > 0:
        for tile in tiledic.keys():
            tiledic[tile] = min([ tiledic[tile], run_params.MaxCycle ])
        for read in run_params.Reads:
            if read.LastCycle == -1:
                read.LastCycle = min(tiledic.values())
    if (not tiledic) or (len(tiledic.values()) == 0):
        common.error("Could not locate image analysis run.")
        sys.exit(1)
    print_tile_counts(tiledic)

    # Sequence folder
    sequence_folder =\
                runfolder.get_new_sequence_folder(analysis_folder,
                                                  basecaller_software,
                                                  basecaller_software_version)
    run_folder = runfolder.get_run_folder(analysis_folder)
    print  "Analysis folder: %s" % analysis_folder
    print  "Sequence folder: %s" % sequence_folder

    # Instrument
    run_params.printInstrument()
    run_params.printReads()

    # Matrix
    base_params.completePath(run_params)
    base_params.printParams()
    base_params.printRestarts(run_params)

    #print run_params.getDict()
    #print base_params.getDict(run_params)
    #print base_params.getMatrixCycles(run_params)
    #print common.getDict(base_params.getMatrixOptions(run_params))
    #print common.getDict(base_params.getPhasingOptions(run_params))
    #print base_params.getPhasingRestarts(run_params)

    # Gerald configuration file
    if not run_params.MakeFlag:
        if (gerald.processGeraldConfig(run_params.MakeFlag,
                                       gerald_params.geraldconfigfiles, 
                                       run_params, sequence_folder) != 0):
            common.error("Failed to process Gerald configuration.")
            sys.exit(1)

    #--
    if not run_params.MakeFlag:
        print ("You can generate the Makefiles by using the \"--make\" "
               "option.")
        sys.exit(0)
    #--

    # Base-caller folders
    tiles = tiledic.keys()
    tiles.sort()
    createBasecallerFolders(sequence_folder, tiles)

    if base_params.Matrix[0] and not base_params.Matrix[0].AutoFlag:
        for lane in runfolder.get_lane_samples(tiledic.keys()).keys():
            matrixfile2 = os.path.join(sequence_folder,
                                       "matrix%i.txt" % (lane+1))
            shutil.copyfile(base_params.Matrix[0].Path, matrixfile2)

    processParamsFiles(sequence_folder, tiledic, base_params)

    print "Generating Makefiles and parameter files ..."

    # generating Makefile for the base-caller
    makefile.generate_basecaller_makefile(tiledic, analysis_folder,
                                          sequence_folder,
                                          run_params, base_params)

    # generating .params file for base-calls
    cyclerange = range(1, max(tiledic.values())+1)
    runfolder.write_sequence_params(analysis_folder, sequence_folder,
                                    tiles, cyclerange,
                                    basecaller_software,
                                    basecaller_software_version,
                                    runfolder.get_login(),
                                    run_params, base_params)

    # processing Gerald configuration file
    if (gerald.processGeraldConfig(run_params.MakeFlag,
                                   gerald_params.geraldconfigfiles, 
                                   run_params, sequence_folder) != 0):
        common.error("Failed to process Gerald configuration.")
        sys.exit(1)
